% LaTeX2e anthony style file

% #### IDENTIFICATION #### %

\NeedsTeXFormat{LaTeX2e}
\ProvidesClass{MasArt}[2003/03/25]


% #### INITIAL CODE #### %

\RequirePackage{float}
\RequirePackage{ifthen}
\newboolean{MasCommon@pdf}
\setboolean{MasCommon@pdf}{false}

\AtEndOfClass{%
  \MasCommon@PageLayout%
  \let\MAS@section=\section%
  \renewcommand\section{\newpage\MAS@section}%
  \advanceheading{0}%
}

% #### DECLARATION OF OPTIONS #### %

\DeclareOption{sf}{\AtEndOfClass{\renewcommand\familydefault{\sfdefault}}}
\DeclareOption{rm}{\AtEndOfClass{\renewcommand\familydefault{\rmdefault}}}

\DeclareOption{a4paper}{
  \PassOptionsToClass{a4paper}{article}
}

\DeclareOption{pdftex}{
  \setboolean{MasCommon@pdf}{true}
  \PassOptionsToClass{pdftex}{article}
}
\DeclareOption{links}{
  \AtEndOfClass{%
    \RequirePackage[pdftex,%
                    colorlinks=true,%
                    linkcolor={Red},%
                    citecolor={Green},%
                    bookmarks=true,%
                    pagebackref=false,%
                    debug=false,%
                    hyperindex=false]{hyperref}%
    \def\equationautorefname{Equation}%
    \def\footnoteautorefname{footnote}%
    \def\itemautorefname{item}%
    \def\figureautorefname{Figure}%
    \def\tableautorefname{Table}%
    \def\partautorefname{Part}%
    \def\appendixautorefname{Appendix}%
    \def\chapterautorefname{Chapter}%
    \def\sectionautorefname{Section}%
    \def\subsectionautorefname{Section}%
    \def\subsubsectionautorefname{Section}%
    \def\paragraphautorefname{Paragraph}%
    \def\subparagraphautorefname{Subparagraph}%
    \def\FancyVerbLineautorefname{Line}%
    \def\theoremautorefname{Theorem}%
    \def\asideautorefname{Aside}%
    \def\exampleautorefname{Example}%
  }
}

\DeclareOption{landscape}{\OptionNotUsed
  \typeout{Class does not support landscape mode}
}
\DeclareOption{fixfigure}{\AtEndOfClass{\floatplacement{figure}{H}}}
\DeclareOption{fixtable}{\AtEndOfClass{\floatplacement{table}{H}}}

\DeclareOption*{\PassOptionsToClass{\CurrentOption}{article}}

% #### EXECUTE OPTIONS #### %

\ExecuteOptions{a4paper}
\ProcessOptions

% #### PACKAGE LOADING #### %

\LoadClass{article}

% #### MAIN CODE #### %

\renewcommand\section{\@startsection%
  {section}{1}{\z@}{\z@}{2.3ex \@plus.2ex}%
  {\thispagestyle{hdrftrii}\normalfont\Large\bfseries}%
}%
\renewcommand{\sectionmark}[1]{\markboth{\textsl{\thesection\ #1}}{}}%
\let\Mas@bibliography=\bibliography
\def\bibliography#1{%
  \newpage%
  \refstepcounter{section}%
  \addcontentsline{toc}{section}{\refname}%
  \markboth{\textsl{\refname}}{}%
  \Mas@bibliography{#1}%
}

\RequirePackage{MasCommon}

\MasCommon@headinglevel=1
\MasCommon@setheading{\MasCommon@headinglevel}

\def\advanceheading#1{%
  \advance\MasCommon@headinglevel by #1
  \ifnum\MasCommon@headinglevel<1%
    \MasCommon@headinglevel=1
    \typeout{Warning tried to reset heading level before 1}%
  \else%
    \ifnum\MasCommon@headinglevel>3%
      \typeout{Warning heading levels after 3 treated as heading level 3}%
    \fi%
  \fi%
  \MasCommon@setheading{\MasCommon@headinglevel}%
}

\let\MasCommon@maketitle=\maketitle
\renewcommand{\maketitle}{%
  \gdef\@pdftitle{\MasCommon@title}%
  \gdef\@pdfauthor{\MasCommon@author}%
  \MasCommon@maketitle\thispagestyle{hdrftrii}%
}

\endinput
