package lmf.formula.csp.gui;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.event.ActionEvent;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.PushbackReader;

import javax.swing.AbstractAction;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextPane;
import javax.swing.JToolBar;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;

import csp2hc.gui.MyFileFilter;
import csp2hc.messages.HandelC;

import lmf.formula.csp.lexer.Lexer;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.Start;
import lmf.formula.csp.parser.Parser;
import lmf.formula.csp.semantic.CspAnalyserListener;
import lmf.formula.csp.semantic.CspContextualAnalyser;
import lmf.formula.csp.semantic.exception.CspAnalyserException;
import lmf.formula.csp.util.Location;
import lmf.formula.csp.util.Region;

/*
 * @SuppressWarnings("serial")
 */
public final class CspExplorer extends JFrame implements CspAnalyserListener {

	private AbstractAction newAction;
	private AbstractAction openAction;
	private AbstractAction saveAction;
	private AbstractAction closeAction;

	private JTextPane jTextPaneCspM;

	private CspJTree jTreeOutline;

	private JTextPane jTextPaneMessages;

	private JLabel statusLabel;

	public CspExplorer() {
		super("CSP Explorer");

		this.initialize();

		this.setDefaultCloseOperation(EXIT_ON_CLOSE);
	}

	private void initialize() {
		createMenu();

		JPanel jPanelOutline = new JPanel(new BorderLayout());
		JToolBar jToolbarOutline = new JToolBar("Outline");
		JTabbedPane jTabOutline = new JTabbedPane();
		jTreeOutline = new CspJTree();
		jTreeOutline.addTreeSelectionListener(new TreeSelectionListener() {
			public void valueChanged(TreeSelectionEvent e) {
				DefaultMutableTreeNode node =
					(DefaultMutableTreeNode) e.getPath().getLastPathComponent();
				if (node.getUserObject() instanceof Node) {
					//((Node) node.getUserObject());
				}
			}
		});
		jTabOutline.addTab("Outline", new JScrollPane(jTreeOutline));
		jToolbarOutline.add(jTabOutline);
		jPanelOutline.add(jToolbarOutline);

		JPanel jPanelEditor = new JPanel(new BorderLayout());
		jTextPaneCspM = new JTextPane();
		jPanelEditor.add(BorderLayout.CENTER, new JScrollPane(jTextPaneCspM));

		JToolBar jToolbarCSPm = new JToolBar("CSPm");
		JTabbedPane tabPane = new JTabbedPane();
		tabPane.addTab("CSPm", jPanelEditor);
		jToolbarCSPm.add(tabPane);

		JToolBar jToolbarMessages = new JToolBar("Messages");
		JTabbedPane tabPaneMessages = new JTabbedPane();
		jTextPaneMessages = new JTextPane();
		tabPaneMessages.addTab("Messages", jTextPaneMessages);
		jToolbarMessages.add(tabPaneMessages);

		JPanel jPanelCenter = new JPanel(new BorderLayout());
		jPanelCenter.add(BorderLayout.CENTER, jToolbarCSPm);
		jPanelCenter.add(BorderLayout.SOUTH, jToolbarMessages);

		statusLabel = new JLabel(" CSP Explorer");

		JSplitPane jSplitPane = new JSplitPane();
		jSplitPane.setOrientation(JSplitPane.HORIZONTAL_SPLIT);
		jSplitPane.setDividerLocation(200);
		jSplitPane.setOneTouchExpandable(true);
		jSplitPane.setLeftComponent(jPanelOutline);
		jSplitPane.setRightComponent(jPanelCenter);

		Container c = this.getContentPane();
		c.setLayout(new BorderLayout());
		c.add(BorderLayout.CENTER, jSplitPane);
		c.add(BorderLayout.SOUTH, statusLabel);
	}

	private void createMenu() {
		JMenuBar mb = new JMenuBar();

		JMenu fileMenu = new JMenu("File");

		newAction = new AbstractAction("New") {
			public void actionPerformed(ActionEvent arg0) {
				;
			}
		};
		fileMenu.add(newAction);
		openAction = new AbstractAction("Open") {
			public void actionPerformed(ActionEvent arg0) {
				open();
			}
		};
		fileMenu.add(openAction);
		saveAction = new AbstractAction("Save") {
			public void actionPerformed(ActionEvent arg0) {
				;
			}
		};
		fileMenu.add(saveAction);
		closeAction = new AbstractAction("Close") {
			public void actionPerformed(ActionEvent arg0) {
				;
			}
		};
		fileMenu.add(closeAction);

		mb.add(fileMenu);
		this.setJMenuBar(mb);
	}

	private void open() {
		JFileChooser fileDialog = new JFileChooser();
		fileDialog.setCurrentDirectory(new File("Z:/files/workspace/marcel/projetos/internalApp/csp2hc/examples"));
		
		//fileDialog.setCurrentDirectory(new File("C:\\Documents and Settings\\Marcel\\My Documents\\My Files\\workspace\\marcel\\projetos\\internalApp\\csp2hc\\examples"));

		int option = fileDialog.showOpenDialog(this);
		
		
		
		if (option == JFileChooser.APPROVE_OPTION) {
			File file = fileDialog.getSelectedFile();
			try {
				BufferedReader is = new BufferedReader(new FileReader(file));
				StringBuffer buffer = new StringBuffer();
				String line = is.readLine();
				while (line != null) {
					buffer.append(line + "\n");
					line = is.readLine();
				}
				jTextPaneCspM.setText(buffer.toString());

				Lexer lexer = new Lexer(new PushbackReader(
						new FileReader(file), 1000));
				Parser parser = new Parser(lexer);
				Start tree = parser.parse();

				// CspContextualAnalyser 
				CspContextualAnalyser analyser = new CspContextualAnalyser(this);
				analyser.setDebugging(true);
				tree.apply(analyser);

				// CspParserTreeUpdater
				jTreeOutline.setCellRenderer(new CspJTreeCellRenderer());
				tree.apply(new CspTreeGenerator(jTreeOutline));
			} catch (Exception e) {
				e.printStackTrace();
				DefaultMutableTreeNode root = new DefaultMutableTreeNode(
						"Error");
				root.add(new DefaultMutableTreeNode(e.getMessage()));
				DefaultTreeModel model = new DefaultTreeModel(root);
				jTreeOutline.setModel(model);

				statusLabel.setText(e.getMessage());
			}
			jTreeOutline.expandAll();
			jTreeOutline.updateUI();
		}
	}

	/** AnalysisListener methods */

	public void started() {
		statusLabel.setText("Type checking...");
	}

	public void analysing(Node node) {
		;
	}

	public void analysed(Node node) {
		;
	}

	public void catchException(CspAnalyserException e) {
		jTextPaneMessages.setText(e.getMessage());
	}

	public void warning(int line, int column, String message) {
		;
	}

	public void finished() {
		statusLabel.setText("Done!");
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		CspExplorer f = new CspExplorer();
		f.setBounds(0, 0, 600, 400);
		f.setVisible(true);
	}

}