package lmf.formula.csp.gui;

import java.util.Stack;

import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;

import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.node.ACspConstantCallExpr;
import lmf.formula.csp.node.ACspSpecification;
import lmf.formula.csp.node.EOF;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.Start;


/**
 * CspParserTreeUpdater.
 * 
 * @author Joabe Jesus (jbjj@cin.ufpe.br)
 */
public class CspTreeGenerator extends ExtendedDepthFirstAdapter {

	private JTree treeView;
	private DefaultMutableTreeNode rootNode;

	private Stack<DefaultMutableTreeNode> parents =
		new Stack<DefaultMutableTreeNode>();

	public CspTreeGenerator(JTree treeView) {
		this.treeView = treeView;
	}

	public void inStart(Start node) {
		DefaultMutableTreeNode rootNode =
			new DefaultMutableTreeNode("Specification");
		parents.push(rootNode);
	}

	public void outStart(Start node) {
		this.rootNode = parents.pop();
		this.treeView.setModel(new DefaultTreeModel(this.rootNode));
	}

	public void inACspSpecification(ACspSpecification node) {
	}

	public void outACspSpecification(ACspSpecification node) {
	}

	/*
	 * As we come across non terminals, push them onto the stack
	 */
	public void defaultIn(Node node) {
		String name = node.getClass().getSimpleName();

		DefaultMutableTreeNode thisNode;
		if (!name.startsWith("ACspBase")) {
			// Not ACspBase...
			name = name.substring(4); // Remove "ACsp"
			if (name.startsWith("Linked")) {
				if (!node.parent().getClass().getSimpleName().endsWith("List")) {
					name = name.substring(6); // Remove "Linked"
					thisNode = new DefaultMutableTreeNode(name);
					parents.push(thisNode);
				}
			} else {
				thisNode = new DefaultMutableTreeNode(name);
				parents.push(thisNode);
			}
		} else {
			// ACspBase...
			if (name.endsWith("List")
					&& !node.parent().getClass().getSimpleName().endsWith("List")) {
				name = name.substring(8); // Remove "ACspBase"
				thisNode = new DefaultMutableTreeNode(name);
				parents.push(thisNode);
			}
		}
	}

	/*
	 * As we leave a non terminal, it's parent is the node before it on the
	 * stack, so we pop it off and add it to that node
	 */
	public void defaultOut(Node node) {
		String name = node.getClass().getSimpleName();

		DefaultMutableTreeNode thisNode;
		if (!name.startsWith("ACspBase")) {
			name = name.substring(4); // Remove "ACsp"
			if (name.startsWith("Linked")) {
				if (!node.parent().getClass().getSimpleName().endsWith("List")) {
					thisNode = parents.pop();
					parents.peek().add(thisNode);
				}
			} else {
				thisNode = parents.pop();
				parents.peek().add(thisNode);
			}
		} else {
			if (name.endsWith("List")
					&& !node.parent().getClass().getSimpleName().endsWith("List")) {
				thisNode = parents.pop();
				parents.peek().add(thisNode);
			}
		}
	}

	public void inACspConstantCallExpr(ACspConstantCallExpr node) {
		DefaultMutableTreeNode thisNode = new DefaultMutableTreeNode(node);
		parents.push(thisNode);
	}

	/*
	 * Terminals - our parent is always on the top of the stack, so we add
	 * ourselves to it
	 */
	public void defaultCase(Node node) {
		DefaultMutableTreeNode thisNode = new DefaultMutableTreeNode(node);
		parents.peek().add(thisNode);
	}

	public void caseEOF(EOF node) {
	}

}