#
#  Blocks World Puzzle
#
#-----------------------------------------------------------------------
#
# Tcl/Tk implementation of the Blocks World Puzzle.
#
#	last update: 30.9.2003 Richard Harrison.
#

package require Soar

proc drawscreen {w} {
	catch {destroy $w}
	toplevel $w
	
	wm title $w "Blocks World Viewer"
	wm iconname $w "Blocks World"
	message $w.msg -font -Adobe-times-medium-r-normal--*-180* \
            	-aspect 400 -text "Blocks World Viewer"
	frame $w.frame -width 480 -height 320 -borderwidth 4 \
	    	-relief ridge -bg Grey70

	button $w.solve -text Solve -command "run"
	button $w.stop -text "Stop" -command "stop-soar"
        button $w.step -text "Step" -command "step"
	button $w.init -text "Init Soar" -command "init-soar; picture a_b_c"
	button $w.exch -text "Excise Chunks" -command "excise -chunks"
	button $w.ok -text Close -command "dismiss $w"
       
	pack $w.msg -side top
	pack $w.frame -side top -padx 20
	pack $w.solve $w.step $w.stop $w.init $w.exch -side left -padx 1m -pady 2m
	pack $w.ok -side right -padx 1m -pady 2m 	

	picture a_b_c
}

proc dismiss {w} {
    excise -all
    init-soar
    destroy $w
}

proc timer {pic} {
	global solveslow
	catch {destroy .blocks.frame.picture}
	image create photo test -file $pic
	label .blocks.frame.picture -image test -height 240 -width 320
	pack .blocks.frame.picture
	set m 1
	after 50 set m 0
	vwait m
}

proc picture {state} {
    global soar_library
    
    set filename "[file join images abc_$state.gif]"
    
    timer $filename
    
}

#-----------------------------------------------------------------------
#
# Initialization and GUI creation

# Load rules defining waterjug solution method

global soar_library
global interp_type

excise -all
source blocks_world.soar

drawscreen .blocks

###
### BLOCKS WORLD: 
### MONITOR STATE AND
### OPERATORS
###

# monitor 'block' events indicating a change in world state.
# three rules corresponding to initial, one stacked & all stacked:

sp {monitor*picture*start-state
   (state <s> ^block <a> <b> <c>)
   (<a> ^name a ^on table ^below o)
   (<b> ^name b ^on table ^below o)
   (<c> ^name c ^on table ^below o)
-->
   (tcl |picture a_b_c | )}
   
sp {monitor*picture*one-stacked
   (state <s> ^block <o1> <o2> <o3>)
   (<o1> ^name <n1> ^on table ^below o)
   (<o2> ^name <n2> ^on table ^below <n3>)
   (<o3> ^name <n3> ^on <n2> ^below o)
-->
   (tcl |picture | <n1> |_| <n2> <n3> )}
   
sp {monitor*picture*all-stacked
   (state <s> ^block <o1> <o2> <o3>)
   (<o1> ^name <n1> ^on table ^below <n2>)
   (<o2> ^name <n2> ^on <n1> ^below <n3>)
   (<o3> ^name <n3> ^on <n2> ^below o)
-->
   (tcl |picture | <n1> <n2> <n3> )}



#-----------------------------------------------------------------
# 
# sp {waterjug*monitor*state 
#     (state <s> ^jug <i> <j>)
#     (<i> ^volume 3 ^contents <icon>)
#     (<j> ^volume 5 ^contents <jcon>)
#     --> 
#     (tcl |picture | <icon> | | <jcon> | state|)
#     (write (crlf)
#            | 3:| <icon> 
#            | 5:| <jcon> )}
# 
# sp {waterjug*monitor*operator-application*empty
#     (state <s> ^operator <o>)
#     (<o> ^name empty
#     	 ^jug <i>)
#     (<i> ^volume <volume> ^contents <contents>)
# 
#     (<s> ^jug <x> <y>)
#     (<x> ^volume 3 ^contents <xcon>)
#     (<y> ^volume 5 ^contents <ycon>)
# 
# 
#     -->
#     (tcl | picture | <xcon> | | <ycon> | dump| <volume>)
#     (write (crlf) 
#     	   |  EMPTY(| <volume> 
#     	   |:| <contents> 
#     	   |)|)}
# 
# sp {waterjug*monitor*operator-application*fill
#     (state <s> ^operator <o>)
#     (<o> ^name fill 
#     	 ^jug <i>)
#     (<i> ^volume <volume> ^contents <contents>)
# 
#     (<s> ^jug <x> <y>)
#     (<x> ^volume 3 ^contents <xcon>)
#     (<y> ^volume 5 ^contents <ycon>)
#     -->
#     (tcl | picture | <xcon> | | <ycon> | fill| <volume>)
#     (write (crlf) 
#     	   |  FILL(| <volume> 
#     	   |:| <contents> 
#     	   |)|)}
# 
# sp {waterjug*monitor*operator-application*pour
#     (state <s> ^operator <o>)
#     (<o> ^name pour 
#     	 ^jug <i>
#     	 ^into <j>)
#     (<i> ^volume <ivol> ^contents <icon>)
#     (<j> ^volume <jvol> ^contents <jcon>)
# 
#     (<s> ^jug <x> <y>)
#     (<x> ^volume 3 ^contents <xcon>)
#     (<y> ^volume 5 ^contents <ycon>)
#     -->
#     (tcl | picture | <xcon> | | <ycon> | | <ivol> |pour| <jvol>)
#     (write (crlf) 
#     	   |  POUR(| <ivol> 
#     	   |:| <icon> 
#     	   |,| <jvol>
#     	   |:| <jcon>
#     	   |)|)}
# 
# sp {waterjug*monitor*tied-operator*empty
#     (state <s> ^impasse tie 
#                ^attribute operator 
#                ^item <item>)
#     (<item> ^name empty ^jug <i>)
#     (<i> ^volume <volume> ^contents <contents>)
# 
#     (<s> ^jug <x> <y>)
#     (<x> ^volume 3 ^contents <xcon>)
#     (<y> ^volume 5 ^contents <ycon>)
# 
#     -->
#     (tcl | picture | <xcon> | | <ycon> | dump| <volume>)
#     (write (crlf) 
#     	   |  | <item> 
#     	   |: empty(| <volume> 
#     	   |:| <contents> 
#     	   |)|)}
# 
# sp {waterjug*monitor*tied-operator*fill
#     (state <s> ^impasse tie 
#                ^attribute operator 
#                ^item <item>)
#     (<item> ^name fill ^jug <i>)
#     (<i> ^volume <volume> ^contents <contents>)
# 
#     (<s> ^jug <x> <y>)
#     (<x> ^volume 3 ^contents <xcon>)
#     (<y> ^volume 5 ^contents <ycon>)
#     -->
#     (tcl | picture | <xcon> | | <ycon> | fill| <volume>)
#     (write (crlf) 
#     	   |  | <item> 
#     	   |: fill(| <volume> 
#     	   |:| <contents> 
#     	   |)|)}
# 
# sp {waterjug*monitor*tied-operator*pour
#     (state <s> ^impasse tie 
#                ^attribute operator 
#                ^item <item>)
#     (<item> ^name pour ^jug <i> ^into <j>)
#     (<i> ^volume <ivol> ^contents <icon>)
#     (<j> ^volume <jvol> ^contents <jcon>)
# 
#     (<s> ^jug <x> <y>)
#     (<x> ^volume 3 ^contents <xcon>)
#     (<y> ^volume 5 ^contents <ycon>)
#     -->
#     (tcl | picture | <xcon> | | <ycon> | | <ivol> |pour| <jvol>)
#     (write (crlf) 
#     	   |  | <item> 
#     	   |: pour(| <ivol> 
#     	   |:| <icon> 
#     	   |,| <jvol>
#     	   |:| <jcon>
#     	   |)|)}
#-----------------------------------------------------------------

### eof of waterjug.soar (Version Type: Soar8)
