% LaTeX2e anthony style file

% #### IDENTIFICATION #### %

\NeedsTeXFormat{LaTeX2e}
\ProvidesClass{MasSlide}[2004/02/20]

% #### INITIAL CODE #### %

\RequirePackage{float}
\RequirePackage{ifthen}
\newboolean{MasCommon@pdf}
\setboolean{MasCommon@pdf}{false}

\AtEndOfClass{%
  \advanceheading{0}%
}

% #### DECLARATION OF OPTIONS #### %

\DeclareOption{sf}{\AtEndOfClass{\renewcommand\familydefault{\sfdefault}}}
\DeclareOption{rm}{\AtEndOfClass{\renewcommand\familydefault{\rmdefault}}}

\DeclareOption{a4paper}{
  \PassOptionsToClass{a4}{seminar}
}

\DeclareOption{a4}{
  \PassOptionsToClass{a4}{seminar}
}

\DeclareOption{landscape}{
  \PassOptionsToClass{landscape}{seminar}
}

\DeclareOption{pdftex}{
  \setboolean{MasCommon@pdf}{true}
}

\DeclareOption{links}{
  \AtEndOfClass{%
    \RequirePackage[pdftex,%
                    colorlinks=true,%
                    linkcolor={DarkRed},%
                    citecolor={DarkGreen},%
                    bookmarks=true,%
                    pagebackref=false,%
                    debug=false,%
                    hyperindex=false]{hyperref}%
    \def\equationautorefname{Equation}%
    \def\footnoteautorefname{footnote}%
    \def\itemautorefname{item}%
    \def\figureautorefname{Figure}%
    \def\tableautorefname{Table}%
    \def\partautorefname{Part}%
    \def\appendixautorefname{Appendix}%
    \def\chapterautorefname{Chapter}%
    \def\sectionautorefname{Section}%
    \def\subsectionautorefname{Section}%
    \def\subsubsectionautorefname{Section}%
    \def\paragraphautorefname{Paragraph}%
    \def\subparagraphautorefname{Subparagraph}%
    \def\FancyVerbLineautorefname{Line}%
    \def\theoremautorefname{Theorem}%
    \def\asideautorefname{Aside}%
    \def\exampleautorefname{Example}%
  }
}

\DeclareOption{fixfigure}{\AtEndOfClass{\floatplacement{figure}{H}}}
\DeclareOption{fixtable}{\AtEndOfClass{\floatplacement{table}{H}}}

\DeclareOption*{\PassOptionsToClass{\CurrentOption}{seminar}}

% #### EXECUTE OPTIONS #### %

\ExecuteOptions{a4}
\ProcessOptions

% #### PACKAGE LOADING #### %

\LoadClass{seminar}

% #### MAIN CODE #### %

\renewcommand{\sectionmark}[1]{\markboth{\textsl{\thesection\ #1}}{}}%
\let\Mas@bibliography=\bibliography
\def\bibliography#1{%
  \section*{\refname}%
  \addcontentsline{toc}{section}{\refname}%
  \renewcommand\section[2]{\relax}%
  \markboth{\textsl{\refname}}{}%
  \Mas@bibliography{#1}%
}

\RequirePackage{MasCommon}

\MasCommon@headinglevel=1
\MasCommon@setheading{\MasCommon@headinglevel}

\def\advanceheading#1{%
  \advance\MasCommon@headinglevel by #1
  \ifnum\MasCommon@headinglevel<1%
    \MasCommon@headinglevel=1
    \typeout{Warning tried to reset heading level before 1}%
  \else%
    \ifnum\MasCommon@headinglevel>3%
      \typeout{Warning heading levels after 3 treated as heading level 3}%
    \fi%
  \fi%
  \MasCommon@setheading{\MasCommon@headinglevel}%
}

\let\MasCommon@slidemaketitle=\slidemaketitle
\renewcommand{\slidemaketitle}{%
  \gdef\@pdftitle{\MasCommon@title}%
  \gdef\@pdfauthor{\MasCommon@author}%
  \MasCommon@slidemaketitle
}

\antpage{Slide \theslide}
\def\tab{{\hspace*{1.5em}}}

% Normal page setup
\newhdrftr{i}
\hdroddi  {}           {}           {}
\hdreveni {}           {}           {}
\ftroddi  {\@antdate}  {\@antcomm}  {\@antpage}
\ftreveni {\@antpage}  {\@antcomm}  {\@antdate}

% Main section (or chapter) page setup
\newhdrftr{ii}
\hdroddii  {}           {}           {}
\hdrevenii {}           {}           {}
\ftroddii  {\@antdate}  {\@antcomm}  {\@antpage}
\ftrevenii {\@antpage}  {\@antcomm}  {\@antdate}

\pagestyle{hdrftri}

\slidepagestyle{hdrftrii}
\newenvironment{titleslide}{%
  \centerslidestrue\begin{slide}%
}{%
  \end{slide}\centerslidesfalse%
}

\slidewidth  210mm
\slideheight 140mm

\renewcommand{\slideleftmargin}{1.2in}
\renewcommand{\sliderightmargin}{1.2in}
\renewcommand{\slidetopmargin}{0.8in}
\renewcommand{\slidebottommargin}{0.8in}

\advance\textheight by -1cm

\centerslidesfalse


\endinput
