package csp2hc.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JFileChooser;
import javax.swing.JInternalFrame;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.text.Document;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.UndoManager;



/**
 * Janela que exibi o texto 
 * @author rolango
 *
 */
public class EditorFrame extends JInternalFrame {
	/**
	 * 
	 */
	
	//private String caminhoArquivos = "C:\\Users\\IvanJunior\\Documents\\Mestrado\\Files_CSP2HC\\testes_explosao\\";
	private String caminhoArquivos = ".\\examples\\";
	
	private static final long serialVersionUID = 1L;

	/*
	 * Editor
	 */
	private JTextPane text;
	protected UndoAction undoAction;
	protected RedoAction redoAction;
	protected UndoManager undo = new UndoManager();

	private JScrollPane rolagem;
	private boolean edited;
	private File file;
	private String extension;
	private MainFrame parent;
	private String defaultTitle;
	private boolean saved;

	public EditorFrame(MainFrame parent, String title, boolean editable, String extension) {
		super(title, true, false, true, true);
		this.parent = parent;
		this.setTitle(title);
		this.defaultTitle = title;
		this.extension = extension;

		this.setSaved(false);
		this.setEdited(false);


		undoAction = new UndoAction();
		redoAction = new RedoAction();

		this.setLayout(new BorderLayout());
		text = new JTextPane();
		text.setBackground(Color.white);
		text.setBounds(0, 0, 334, 218);
		text.setFont(new Font("CZT", 4, 12));
		rolagem = new JScrollPane(text);
		this.add(rolagem, BorderLayout.CENTER);

		text.setEditable(editable);
		/*
		 * Deve possibilitar Undo e Redo;
		 */
		if (editable) {
			text.getStyledDocument().addUndoableEditListener(new MyUndoableEditListener());
		}
		add(rolagem);
		this.setVisible(true);

		this.text.addKeyListener(new KeyListener(){

			public void keyPressed(KeyEvent e) {
				setEdited(true);
			}

			public void keyReleased(KeyEvent e) {

			}

			public void keyTyped(KeyEvent e) {

			}});
		
		this.addHierarchyBoundsListener(new java.awt.event.HierarchyBoundsListener () {
			public void ancestorMoved(java.awt.event.HierarchyEvent e) {
			}
			public void ancestorResized(java.awt.event.HierarchyEvent e) {
				atualizarInfoTela();
			}
		}
		);

	}


	protected void atualizarInfoTela() {
		if (this.text.isEditable()) {
			parent.atualizarInfoCSPTela(this);
		}
		else {
			parent.atualizarInfoHandelCTela(this);
		}
		
	}


	protected class MyUndoableEditListener
	implements UndoableEditListener {
		public void undoableEditHappened(UndoableEditEvent e) {
//			Remember the edit and update the menus.
			undo.addEdit(e.getEdit());
			undoAction.updateUndoState();
			redoAction.updateRedoState();
		}
	}




	class UndoAction extends AbstractAction {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		public UndoAction() {
			super("Undo");
			setEnabled(false);
		}

		public void actionPerformed(ActionEvent e) {
			try {
				undo.undo();
			} catch (CannotUndoException ex) {
				System.out.println("Unable to undo: " + ex);
				ex.printStackTrace();
			}
			updateUndoState();
			redoAction.updateRedoState();
		}

		protected void updateUndoState() {
			if (undo.canUndo()) {
				parent.setUndoEnabled(true);
			} else {
				parent.setUndoEnabled(false);
			}
		}
	}

	class RedoAction extends AbstractAction {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		public RedoAction() {
			super("Redo");
			setEnabled(false);
		}

		public void actionPerformed(ActionEvent e) {
			try {
				undo.redo();
			} catch (CannotRedoException ex) {
				System.out.println("Unable to redo: " + ex);
				ex.printStackTrace();
			}
			updateRedoState();
			undoAction.updateUndoState();
		}

		protected void updateRedoState() {
			if (undo.canRedo()) {
				parent.setRedoEnabled(true);
			} else {
				parent.setRedoEnabled(false);
			}
		}
	}


	protected void setEdited(boolean newEdited) {
		if (!newEdited) {
			if (this.getTitle().endsWith("*")) {
				this.setTitle(this.getTitle().substring(0, this.getTitle().length()-1));
			}
		} else {
			if (!this.edited && this.text.isEditable()) {
				this.parent.keyPressed();
				this.setSaved(false);
				this.setTitle(this.getTitle()+"*");
			}
		}
		this.edited = newEdited;
	}

	public boolean getEdited() {
		return edited;
	}

	/**
	 * Define o texto ao textArea
	 * @param txt
	 */
	void setText(String txt, boolean opening) {
		text.setText(txt);
		text.setCaretPosition(0);
		if (!opening) {
			this.setEdited(true);
		}
	}

	void setFile(File file) throws IOException {
		this.file = file;
		String title = this.defaultTitle;
		if (file != null) {
			title = file.getCanonicalPath();
		}
		this.setTitle(title);
	}

	File getFile() {
		return this.file;
	}

	int getLineCount() {
		return 0;//this.text.getLineCount();
	}

	public void saveFile() throws IOException {
		//TODO
		if (this.file == null) {
			//File currentDir = new File(".");
			File currentDir = new File(caminhoArquivos); // Add - Ivan
			JFileChooser fileChooser = new JFileChooser(currentDir);
			fileChooser.setMultiSelectionEnabled(false);
			fileChooser.setDialogType(JFileChooser.SAVE_DIALOG);
			fileChooser.setDialogTitle("Save "+this.title);
			fileChooser.showSaveDialog(this);
			File file = fileChooser.getSelectedFile();
			this.file = file;
		}
		try {
			String fileName = this.file.getPath();
			File output;

			if (fileName.endsWith(this.extension)) {
				output = new File(fileName);
			} else{
				output = new File(fileName + this.extension);
			}

			FileOutputStream out = new FileOutputStream(output);
			byte [] data = this.text.getText().getBytes();
			out.write(data);
			out.close();
			this.setEdited(false);
			this.setTitle(output.getCanonicalPath());
			this.setSaved(true);
		}
		catch (Exception e) {
			/* Ação CANCELADA*/
		}
	}

	public boolean getSaved() {
		return saved;
	}

	private void setSaved(boolean newSaved) {
		saved = newSaved;
	}

	public void desfazer() {
		undoAction.actionPerformed(null);
	}

	public void refazer() {
		redoAction.actionPerformed(null);
	}

	public boolean hasText() {
		if (text.getText().length() != 0 ) {
			return true;
		}
		return false;
	}

	public void clearUndoRedo() {
		undo.discardAllEdits();
		parent.setUndoEnabled(false);
		parent.setRedoEnabled(false);
	}
}
