package csp2hc.gui;

import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.SystemColor;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PushbackReader;
import java.io.StringReader;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import javax.swing.DefaultDesktopManager;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JDesktopPane;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JSeparator;
import javax.swing.JToolBar;
import javax.swing.KeyStroke;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;

import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.lexer.Lexer;
import lmf.formula.csp.lexer.LexerException;
import lmf.formula.csp.node.ACspBaseEventList;
import lmf.formula.csp.node.ACspEvent;
import lmf.formula.csp.node.ACspInterleaveProcess;
import lmf.formula.csp.node.ACspLinkedEventList;
import lmf.formula.csp.node.ACspProcessCallExpr;
import lmf.formula.csp.node.ACspProcessDefinition;
import lmf.formula.csp.node.ACspProductionsSet;
import lmf.formula.csp.node.ACspRenamingProcess;
import lmf.formula.csp.node.ACspSharingParallelProcess;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.PArguments;
import lmf.formula.csp.node.PComunication;
import lmf.formula.csp.node.PEvent;
import lmf.formula.csp.node.PEventList;
import lmf.formula.csp.node.PProjection;
import lmf.formula.csp.node.Start;
import lmf.formula.csp.node.TComma;
import lmf.formula.csp.node.TName;
import lmf.formula.csp.parser.Parser;
import lmf.formula.csp.parser.ParserException;
import lmf.formula.csp.semantic.CspAnalyserListener;
import lmf.formula.csp.semantic.CspContextualAnalyser;
import lmf.formula.csp.semantic.exception.CspAnalyserException;
import lmf.formula.csp.util.CspNodeLocator;
import lmf.formula.csp.util.Location;


import com.sun.java.swing.plaf.windows.WindowsLookAndFeel;

import csp2hc.exception.AbstractDefinitionException;
import csp2hc.exception.BusUseException;
import csp2hc.exception.CouldNotInferTypeException;
import csp2hc.exception.InconsistentChannelProjectionUseException;
import csp2hc.exception.InvalidDirectiveException;
import csp2hc.exception.KeywordException;
import csp2hc.exception.NumberOfWritersException;
import csp2hc.exception.ReaderAndWriterParallelBranchException;
import csp2hc.exception.RepeatedConstantException;
import csp2hc.exception.UndefinedArgumentSetTypeException;
import csp2hc.exception.UndefinedArgumentTypeException;
import csp2hc.exception.UndefinedChannelInOutException;
import csp2hc.exception.UndefinedChannelTypeException;
import csp2hc.exception.UndefinedEmbeddedFunctionEvaluationException;
import csp2hc.exception.UndefinedEmbeddedFunctionTypeException;
import csp2hc.exception.UndefinedEmptySetTypeException;
import csp2hc.exception.UndefinedIntBitsException;
import csp2hc.exception.UnsupportedConstructorException;
import csp2hc.exception.UnsupportedEmbeddedFunctionException;
import csp2hc.exception.WrongNumberOfArgumentsException;
import csp2hc.messages.HandelC;
import csp2hc.messages.Messenger;
import csp2hc.messages.Warning;
import csp2hc.util.BetterPreProcessing;
import csp2hc.util.CodeFormatting;
import csp2hc.util.Config;
import csp2hc.util.ForcedInterleaveInsertion;
import csp2hc.util.ForcedInterleavingProtocol;
import csp2hc.util.ParserUtil;
import csp2hc.util.RenameProcessesUtil;
import csp2hc.visitors.ChannelsVisitor;
import csp2hc.visitors.CodeInsertionVisitor;
import csp2hc.visitors.ConstantsAndFunctionsVisitor;
import csp2hc.visitors.DatatypesVisitor;
import csp2hc.visitors.DirectivesVisitor;
import csp2hc.visitors.KeywordsVisitor;
import csp2hc.visitors.MultiSyncVisitor;
import csp2hc.visitors.ParallelismAnalyserVisitor;
import csp2hc.visitors.ProcessesInfoVisitor;
import csp2hc.visitors.ProcessesTranslatorVisitor;
import csp2hc.visitors.RenamingVisitor;
import csp2hc.visitors.RestrictionsVisitor;
import csp2hc.visitors.TailRecursionVisitor;

public class MainFrame extends JFrame implements CspAnalyserListener {
	
	//private String caminhoArquivos = "C:\\Users\\IvanJunior\\Documents\\Mestrado\\Files_CSP2HC\\";
	private String caminhoArquivos = ".\\examples\\";

	private static final String fileSep = System.getProperty("file.separator");
	private Config config;
	private String csp_files_dir;
	private Start start;
	private CspAnalyserException error;

	private static final String CSP = "CSP";
	private static final String HANDELC = "Handel-C";

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	private String path = "images/";


	private JDesktopPane theDesktopPane;
	private DefaultDesktopManager defaultDesktopManager;

	private JMenuBar menu;
	private JMenu file;
	private JMenuItem openCSP;
	private JMenuItem saveCSP;
	private JMenuItem translate;
	private JMenuItem saveHandelC;
	private JMenuItem exit;
	private JMenuItem undo;
	private JMenuItem redo;



	private GridBagConstraints cons; 
	private GridBagLayout gridBagLayout ;  

	/* Butões*/

	private JButton openButtonCSP;
	private JButton saveButtonCSP;
	private JButton saveButtonHandelC;
	private JButton translateButton;
	private JButton image_undo = new JButton();
	private JButton image_redo = new JButton();


	/* ToolBar de butoes de imagens*/
	private JToolBar imagensBar;


	/* Janelas Internas*/

	private EditorFrame cspFrame;
	private EditorFrame handelCFrame;
	private MessagesFrame msgsFrame;
	
	private InfoIFrames infoTelas;
	private boolean first = true;


	public MainFrame(){
		
		infoTelas = new InfoIFrames();
		
		cspFrame = new EditorFrame(this,"CSP", true, ".csp");
		handelCFrame = new EditorFrame(this,"Handel-C", false, ".hcc");
		msgsFrame = new MessagesFrame(this);

		/* instanciando tipo de Layout*/
		gridBagLayout = new GridBagLayout();
		cons = new GridBagConstraints();

		theDesktopPane = new JDesktopPane();
		defaultDesktopManager = new DefaultDesktopManager();
		theDesktopPane.setDesktopManager(defaultDesktopManager);

		setInternalFrames();
		startingGUI();
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		
		startingInfoTelas ();
		
		this.loadConfigurations();
	}

	private void startingInfoTelas() {
		infoTelas.setInfoCsp(cspFrame.getHeight(), cspFrame.getWidth()
				, cspFrame.getX(), cspFrame.getY());
		infoTelas.setInfoMsg(msgsFrame.getHeight(), msgsFrame.getWidth()
				, msgsFrame.getX(), msgsFrame.getY());	
		infoTelas.setInfoHandelC(handelCFrame.getHeight(), handelCFrame.getWidth()
				, handelCFrame.getX(), handelCFrame.getY());
	}

	private class TranslateThread extends Thread{

		public void run(){
			translate();
			msgsFrame.setListSelection();
		}
	}


	/** AnalysisListener methods */

	public void started() {
	}

	public void analysing(Node node) {
	}

	public void analysed(Node node) {
	}

	public void catchException(CspAnalyserException e) {
		this.error = e;
	}

	public void warning(int line, int column, String message) {
	}

	public void finished() {
	}
	/* End of methods */

	private void loadConfigurations(){
		Messenger.getInstance(this).start(Messenger.LOADING_CONFIG);
		try {
			this.config = Config.getInstance();
		} catch (Exception e) {
			e.printStackTrace();
		}
		//this.csp_files_dir = config.getConfig("CSP_FILES_DIR");
		Messenger.getInstance(this).finish(Messenger.LOADING_CONFIG);
	}

	/**
	 * Adiciona ao Desktop Manager as janelas internas 
	 * que comp�es o programa
	 */
	private void setInternalFrames() {
		this.theDesktopPane.add(cspFrame);
		this.theDesktopPane.add(handelCFrame);
		this.theDesktopPane.add(msgsFrame);
		this.theDesktopPane.setVisible(true);
		
		
	}

	/**
	 * Deixa a interface mais bonitinha
	 */
	private void setLookAndFeel() {
		if (isWindows()) {
			try {
				UIManager.setLookAndFeel( new WindowsLookAndFeel());
			} catch (UnsupportedLookAndFeelException e1) {
			}
		}

	}

	public MessagesFrame getMessageFrame() {
		return this.msgsFrame;
	}

	private static boolean isWindows() {
		String osName = System.getProperty("os.name");
		if (osName.startsWith("Windows")) {
			return true;
		}
		return false;
	}


	/**
	 * inicializa os componentes de GUI
	 */
	private void startingGUI() {

		this.setLocation(0,0);

		int width = getToolkit().getScreenSize().width;
		int height = getToolkit().getScreenSize().height;

		this.setSize(width , height*580/600) ;

		this.setTitle("CSP2HANDELC");
		this.getContentPane().setLayout(gridBagLayout);


		theDesktopPane.addHierarchyBoundsListener(new java.awt.event.HierarchyBoundsListener () {
			public void ancestorMoved(java.awt.event.HierarchyEvent e) {
			}
			public void ancestorResized(java.awt.event.HierarchyEvent e) {
				redimensionarJanelas(e);
			}
		}
		);
		
		this.addWindowListener(new WindowListener(){

			public void windowActivated(WindowEvent arg0) {
			}

			public void windowClosed(WindowEvent arg0) {
			}

			public void windowClosing(WindowEvent arg0) {
				exit();
			}

			public void windowDeactivated(WindowEvent arg0) {
			}

			public void windowDeiconified(WindowEvent arg0) {
			}

			public void windowIconified(WindowEvent arg0) {
			}
			public void windowOpened(WindowEvent arg0) {

			}

		});


		/* Tratando o menu file*/

		menu = new JMenuBar(); 

		file = new JMenu("File");

		openCSP  = new JMenuItem("Open CSP");
		openCSP.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				openCSP();
			}
		});


		saveCSP  = new JMenuItem("Save CSP");
		saveCSP.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent e) {
				save(MainFrame.CSP);
			}

		});
		/* Adiciona atalho CTRL + S*/
		saveCSP.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S, KeyEvent.CTRL_MASK));

		saveHandelC =  new JMenuItem("Save Handel-C");
		saveHandelC.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent e) {
				save(MainFrame.HANDELC);
			}

		});

		translate = new JMenuItem("Translate");
		translate.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent e) {
				new TranslateThread().start();
			}

		});



		undo = new JMenuItem("Undo");
		undo.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				desfazer();
			}

		});
		undo.setEnabled(false);

		/* Adiciona atalho CTRL + Z*/
		undo.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Z, KeyEvent.CTRL_MASK));

		redo = new JMenuItem("Redo");
		redo.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				refazer();	
			}

		});
		redo.setEnabled(false);
		/* Adiciona atalho CTRL + Y*/
		redo.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Y, KeyEvent.CTRL_MASK));


		/* Adiciona atalho CTRL + T*/
		translate.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_T, KeyEvent.CTRL_MASK));

		exit = new JMenuItem("Exit");
		exit.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent e) {
				exit();
			}

		});

		menu.add(file);
		file.add(openCSP);
		file.add(saveCSP);
		file.add(new JSeparator());
		file.add(undo);
		file.add(redo);
		file.add(new JSeparator());
		file.add(translate);
		file.add(new JSeparator());
		file.add(saveHandelC);
		file.add(new JSeparator());
		file.add(exit);

		this.setJMenuBar(menu);


		/* Definindo barra de but�es de imagem */

		imagensBar = new JToolBar();
		imagensBar.setBackground(SystemColor.scrollbar);

		openButtonCSP = new JButton();
		openButtonCSP.setIcon(new ImageIcon(path + "open.gif"));
		openButtonCSP.setMaximumSize(new Dimension(30, 27));
		openButtonCSP.setMinimumSize(new Dimension(30, 27));
		openButtonCSP.setPreferredSize(new Dimension(30, 27));
		openButtonCSP.setToolTipText("Open CSP");
		openButtonCSP.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent e) {
				openCSP();
			}

		});
		saveButtonCSP = new JButton();
		saveButtonCSP.setIcon(new ImageIcon(path + "save.gif"));
		saveButtonCSP.setMaximumSize(new Dimension(30, 27));
		saveButtonCSP.setMinimumSize(new Dimension(30, 27));
		saveButtonCSP.setPreferredSize(new Dimension(30, 27));
		saveButtonCSP.setToolTipText("Save CSP");
		saveButtonCSP.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				save(MainFrame.CSP);
			}
		});


		saveButtonHandelC = new JButton();
		saveButtonHandelC.setIcon(new ImageIcon(path + "save.gif"));
		saveButtonHandelC.setMaximumSize(new Dimension(30, 27));
		saveButtonHandelC.setMinimumSize(new Dimension(30, 27));
		saveButtonHandelC.setPreferredSize(new Dimension(30, 27));
		saveButtonHandelC.setToolTipText("Save Handel-C");
		saveButtonHandelC.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				save(MainFrame.HANDELC);
			}
		});


		translateButton = new JButton();
		translateButton.setIcon(new ImageIcon(path + "generate.gif"));
		translateButton.setMaximumSize(new Dimension(30, 27));
		translateButton.setMinimumSize(new Dimension(30, 27));
		translateButton.setPreferredSize(new Dimension(30, 27));
		translateButton.setToolTipText("Translate");
		translateButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				if (cspFrame.hasText()) {
					translate();
				}
			}
		});


		image_undo = new JButton();
		image_undo.setIcon(new ImageIcon(path + "undo.gif"));
		image_undo.setMaximumSize(new Dimension(30, 27));
		image_undo.setMinimumSize(new Dimension(30, 27));
		image_undo.setPreferredSize(new Dimension(30, 27));
		image_undo.setToolTipText("Undo");
		image_undo.setEnabled(false);
		image_undo.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				desfazer();
			}
		});


		image_redo = new JButton();
		image_redo.setIcon(new ImageIcon(path + "redo.gif"));
		image_redo.setMaximumSize(new Dimension(30, 27));
		image_redo.setMinimumSize(new Dimension(30, 27));
		image_redo.setPreferredSize(new Dimension(30, 27));
		image_redo.setToolTipText("Redo");
		image_redo.setEnabled(false);
		image_redo.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				refazer();
			}
		});




		/*
		 *Adicionando butoes no ImagensBar 
		 */
		imagensBar.setBorder(null);
		this.startImagensBar();


		setGUIBoundElements();
		this.disableTranslate();
	}


	private void redimensionarJanelas(java.awt.event.HierarchyEvent e) {
		
		if (first  && theDesktopPane.getHeight() != 0 && theDesktopPane.getWidth() != 0) {
			this.infoTelas.setInfoDesktop(theDesktopPane.getHeight(), theDesktopPane.getWidth());
			first = false;
		}
		/* As vezes ele chama esse metodo antes de inicializar os componentes*/
		else if (e.getChanged() == this && theDesktopPane.getHeight() != 0 && theDesktopPane.getWidth() != 0) {
			int newWidth = theDesktopPane.getWidth();
			int newHeight = theDesktopPane.getHeight();

			
			float fatorHeight = (float)newHeight/(float)this.infoTelas.getDesktopHeight();
			float fatorWidth = (float)newWidth/(float)this.infoTelas.getDesktopWidth();

			Float desAlt = this.infoTelas.getCspHeight()*fatorHeight;
			Float desLar = this.infoTelas.getCspWidth()*fatorWidth;

			Float desY = this.infoTelas.getCspY()*fatorHeight;
			Float desX = this.infoTelas.getCspX()*fatorWidth;

			Float codAlt = this.infoTelas.getHandelCHeight()*fatorHeight;
			Float codLar = this.infoTelas.getHandelCWidth()*fatorWidth;

			Float codY = this.infoTelas.getHandelCY()*fatorHeight;
			Float codX = this.infoTelas.getHandelCX()*fatorWidth;

			Float obriAlt = this.infoTelas.getMsgHeight()*fatorHeight;
			Float obriLar = this.infoTelas.getMsgWidth()*fatorWidth;

			Float obriY = this.infoTelas.getMsgY()*fatorHeight;
			Float obriX = this.infoTelas.getMsgX()*fatorWidth;


			//Redimensionando cspFrame
			defaultDesktopManager.setBoundsForFrame(cspFrame, desX.intValue()
					, desY.intValue(), desLar.intValue(), desAlt.intValue() );

			//Redimensionando HandelCFrame

			defaultDesktopManager.setBoundsForFrame(handelCFrame, codX.intValue() 
					, codY.intValue(), codLar.intValue(), codAlt.intValue());

			//Redimensionando MsgsFrames
			defaultDesktopManager.setBoundsForFrame(msgsFrame, obriX.intValue(),obriY.intValue() 
					, obriLar.intValue(), obriAlt.intValue());		


			this.infoTelas.AtualizarCsp(cspFrame);
			this.infoTelas.AtualizarHandelC(handelCFrame);
			this.infoTelas.AtualizarMsg(msgsFrame);
			this.infoTelas.AtualizarDesktop(theDesktopPane);

			for (JInternalFrame f : theDesktopPane.getAllFrames()) {
				//Se a mesma janela estiver na forma de icon;
				if (f.isIcon()) {
					JInternalFrame.JDesktopIcon icon =  f.getDesktopIcon();
					Float iconX = icon.getX()*fatorWidth;
					Float iconY = icon.getY()*fatorHeight;
					icon.setLocation(iconX.intValue(), iconY.intValue());
				}
			}

		}
		this.repaint();
	}

	protected void refazer() {
		cspFrame.refazer();

	}

	protected void desfazer() {
		cspFrame.desfazer();
	}

	/**
	 * Metodo que configurara e adicionara os componentes do ImagensBar
	 *
	 */
	private void startImagensBar() {

		JLabel label1 = new JLabel("CSP:");
		label1.setFont(new Font("",Font.BOLD,14));
		JLabel label2 = new JLabel("Handel-C:");
		label2.setFont(new Font("",Font.BOLD,14));
		JLabel label3 = new JLabel("Translate:");
		label3.setFont(new Font("",Font.BOLD,14));
		FlowLayout left = new FlowLayout();
		left.setAlignment(FlowLayout.LEFT);
		imagensBar.setLayout(left);
		imagensBar.add(label1);
		imagensBar.add(openButtonCSP);
		imagensBar.add(saveButtonCSP);
		imagensBar.add(label2);
		imagensBar.add(saveButtonHandelC);
		imagensBar.add(label3);
		imagensBar.add(translateButton);
		imagensBar.add(image_undo);
		imagensBar.add(image_redo);


	}

	private void openCSP() {
		openFile("Open CSP", cspFrame);
		if (this.cspFrame.getFile() != null) {
			enableTranslate();
		}
	}

	/**
	 * Adiciona elementos da gui a janela principal
	 */
	private void setGUIBoundElements() {

		int width = getToolkit().getScreenSize().width;
		int height = getToolkit().getScreenSize().height;

		int shiftx = 10;
		int shifty = 15;
		
		cspFrame.setLocation(shiftx, shifty);
		cspFrame.setSize(width*46/100, height*55/100);
		
		int widthSeta = 60;
		int heightSeta = 30;

		handelCFrame.setLocation(cspFrame.getX() + cspFrame.getWidth() + widthSeta, cspFrame.getY());
		handelCFrame.setSize(cspFrame.getSize());

		msgsFrame.setLocation(cspFrame.getX(), cspFrame.getY() + cspFrame.getHeight() + 10);
		msgsFrame.setSize(width*97/100, height * 25/100);
		msgsFrame.setGUIElements();


		this.setConstraints(0,0,2,62,1.00,0.0);
		cons.fill = cons.HORIZONTAL;
		this.getContentPane().add(imagensBar, cons);

		this.setConstraints(0, 3, 98, 62, 1.00, 1.00);
		cons.fill = cons.BOTH;
		this.getContentPane().add(theDesktopPane, cons);
	}

	/**
	 * Metodo que seta o astributos do Layout, no caso um GridBagLayout 
	 *
	 * @param gridx Permite especificar qual a posi��o absoluta horizontalmente que o componente ser� adicionado.
	 * @param gridy Permite especificar qual a posi��o absoluta verticalmente que o componente ser� adicionado.
	 * @param gridheight Especifica o n�mero de c�lulas que o componente ocupar� verticalmente
	 * @param gridwidth Especifica o n�mero de c�lulas que o componente ocupar� horizontalmente.
	 * @param weightx Peso em x. Especifica um peso no redimensionamento.
	 * @param weighty Peso em y. Especifica um peso no redimensionamento.
	 */	
	private void setConstraints (int gridx, int gridy,int gridheight,
			int gridwidth, double weightx, double weighty) {

		cons.gridx = gridx;
		cons.gridy = gridy;
		cons.gridheight = gridheight;
		cons.gridwidth = gridwidth;
		cons.weightx =weightx;
		cons.weighty = weighty;
		gridBagLayout.setConstraints(this, cons);
	}

	/**
	 * Deve realizar a tradu��o propriamente dita
	 */
	protected void translate() {
		this.save(MainFrame.CSP);
		this.msgsFrame.clear();
		try {
			String translation = this.translateCSP(this.cspFrame.getFile());
			this.handelCFrame.setText(translation, false);
			this.handelCFrame.setFile(null);
			this.translateButtonOK();
		} catch (FileNotFoundException e) {
			this.showKnownException(e);
		} catch (IOException e) {
			this.showKnownException(e);
		} catch (ParserException e) {
			this.showKnownException(e);
		} catch (LexerException e) {
			this.showKnownException(e);
		} catch (KeywordException e) {
			this.showKnownException(e);
		} catch (CspAnalyserException e) {
			this.showKnownException(e);
		} catch (UnsupportedConstructorException e) {
			this.showKnownException(e);
		} catch (UndefinedChannelInOutException e) {
			this.showKnownException(e);
		} catch (UndefinedArgumentTypeException e) {
			this.showKnownException(e);
		} catch (UndefinedChannelTypeException e) {
			this.showKnownException(e);
		} catch (UndefinedIntBitsException e) {
			this.showKnownException(e);
		} catch (AbstractDefinitionException e) {
			this.showKnownException(e);
		} catch (InvalidDirectiveException e) {
			this.showKnownException(e);
		} catch (InconsistentChannelProjectionUseException e) {
			this.showKnownException(e);
		} catch (RepeatedConstantException e) {
			this.showKnownException(e);
		} catch (WrongNumberOfArgumentsException e) {
			this.showKnownException(e);
		} catch (NumberOfWritersException e) {
			this.showKnownException(e);
		} catch (ReaderAndWriterParallelBranchException e) {
			this.showKnownException(e);
		} catch (BusUseException e) {
			this.showKnownException(e);
		} catch (CouldNotInferTypeException e) {
			this.showKnownException(e);
		} catch (UndefinedArgumentSetTypeException e) {
			this.showKnownException(e);
		} catch (UndefinedEmbeddedFunctionTypeException e) {
			this.showKnownException(e);
		} catch (UndefinedEmbeddedFunctionEvaluationException e) {
			this.showKnownException(e);
		} catch (UndefinedEmptySetTypeException e) {
			this.showKnownException(e);
		} catch (UnsupportedEmbeddedFunctionException e) {
			this.showKnownException(e);
		} catch (RuntimeException e) {
			this.showUnknownException(e);
		} finally {
			Messenger.getInstance(this).reset();
		}
	}


	/**
	 * Deve perguntar ao usuário se o mesmo deseja salvar o programa
	 * antes de finalizar a execução do programa.
	 * @param toQuit diz se é ou nao a opção de finalização do programa, ou seja
	 * , o evento da janela ter sido fechada.
	 */
	protected void exit() {

		if (cspFrame.getEdited() && !cspFrame.getSaved()) {
			int answer = JOptionPane.showConfirmDialog (null,
					Messenger.SAVE_CSP,
					Messenger.SAVE_CSP_TITLE,
					JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
			if (answer == JOptionPane.OK_OPTION) {
				this.save(MainFrame.CSP);
			}
		}

		if (handelCFrame.getEdited() && !handelCFrame.getSaved()) {
			int answer = JOptionPane.showConfirmDialog (null,
					Messenger.SAVE_HANDEL_C,
					Messenger.SAVE_HANDEL_C_TITLE,
					JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
			if (answer == JOptionPane.OK_OPTION) {
				this.save(MainFrame.HANDELC);
			}
		}

		System.exit(0);
	}

	/**
	 * Salva o arquivo de texto de CSP
	 */
	protected synchronized void save(String type) {
		try {
			if (type.equals(MainFrame.CSP)) {
				this.cspFrame.saveFile();
			} else if (type.equals(MainFrame.HANDELC)) {
				this.handelCFrame.saveFile();
			}
		} catch (IOException i) {
			Messenger.getInstance().error(i.getMessage());
		}
	}

	/**
	 * Abre um arquivo no editor passado como argumento
	 */
	protected void openFile(String title, EditorFrame frame) {
		try {
			//File currentDir = new File(".");
			File currentDir = new File(caminhoArquivos); // Add - Ivan
			JFileChooser fileChooser = new JFileChooser(currentDir);
			fileChooser.setMultiSelectionEnabled(false);
			fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
			fileChooser.setDialogTitle(title);
			int choice = fileChooser.showOpenDialog(this);
			File file = fileChooser.getSelectedFile();

			if (file != null && choice == JFileChooser.APPROVE_OPTION) {
				if(file != null){
					String sResult = "", tmp = "";
					// open the file for reading
					BufferedReader bread = new BufferedReader(new FileReader(file));
					// this reads the file in, one line at a time
					while((tmp = bread.readLine()) != null){
						sResult += tmp + "\n";
					}
					frame.setText(sResult, true);
					frame.clearUndoRedo();
				}        

			}
			frame.setFile(file);
		} catch (FileNotFoundException f) {
			f.printStackTrace();
		} catch (IOException i) {
			i.printStackTrace();
		}
	}
	
	private String translateCSP(File file) throws 
	FileNotFoundException, 
	IOException, 
	ParserException, 
	LexerException {  
		
		boolean keepGoing = false;
		
		long startTime = System.currentTimeMillis();

		//////////////////////
		// PREPROCESSING
		//////////////////////
		
		Messenger.getInstance(this).start(Messenger.PREPROCESSING);
		BetterPreProcessing preProcessing = new BetterPreProcessing();
		String preProcessingInsert = preProcessing.execute(file);
		Messenger.getInstance(this).finish(Messenger.PREPROCESSING);
		
		//Node start = ParserUtil.parser(file);
		
		Node start = ParserUtil.parser(file, preProcessingInsert);
		
		
		//////////////////////
		// LOAD DIRECTIVE INFO
		////////////////////// 
		Messenger.getInstance(this).start(Messenger.LOADING_DIRECTIVES);
		DirectivesVisitor directives = new DirectivesVisitor();
		start.apply(directives);
		Messenger.getInstance(this).finish(Messenger.LOADING_DIRECTIVES);

		///////////////////////////////
		// LOAD CONSTANTS AND FUNCTIONS
		///////////////////////////////
		Messenger.getInstance(this).start(Messenger.LOADING_CONSTANTS_AND_FUNCTIONS);
		ConstantsAndFunctionsVisitor constants = new ConstantsAndFunctionsVisitor(directives);
		start.apply(constants);
		Messenger.getInstance(this).finish(Messenger.LOADING_CONSTANTS_AND_FUNCTIONS);
		
		/////////////////
		// LOAD DATATYPES
		/////////////////
		Messenger.getInstance(this).start(Messenger.LOADING_DATATYPES);
		DatatypesVisitor types = new DatatypesVisitor(directives,constants);
		start.apply(types);
		Messenger.getInstance(this).finish(Messenger.LOADING_DATATYPES);
		
		String mainProc = HandelC.MAINP + HandelC.ASSIGNMENT + directives.getMain();
		start = ParserUtil.parser(file, preProcessingInsert + mainProc);
		
		////////////////////////////
		// LOAD CHANNELS INFORMATION
		////////////////////////////
		Messenger.getInstance(this).start(Messenger.LOADING_CHANNELS_INFO);
		ChannelsVisitor channels = new ChannelsVisitor(directives, types, constants);
		start.apply(channels);
		Messenger.getInstance(this).finish(Messenger.LOADING_CHANNELS_INFO);
		
		///////////////////////////////////////////
		// LOAD PROCESSES ARGUMENTS AND DEFINITIONS
		///////////////////////////////////////////
		Messenger.getInstance(this).start(Messenger.LOADING_PROCESSES_INFO);
		ProcessesInfoVisitor procInfo = new ProcessesInfoVisitor(directives, types, channels);
		start.apply(procInfo);
		Messenger.getInstance(this).finish(Messenger.LOADING_PROCESSES_INFO);
		
		
		// LOAD RENAMING
		Messenger.getInstance(this).start(Messenger.RENAMING_SUBSTITUTION);
		RenamingVisitor renaming = new RenamingVisitor(directives,constants,procInfo,channels,preProcessing.getForcedInterleaving(), preProcessing);
		start.apply(renaming);
		ParallelismAnalyserVisitor parallelismAnalyser = 
			new ParallelismAnalyserVisitor(procInfo, directives, constants);
		start.apply(parallelismAnalyser);
		RenameProcessesUtil renameProcesses = new RenameProcessesUtil(parallelismAnalyser, null, procInfo, preProcessing, directives, constants);
		start.apply(renameProcesses);
		procInfo = new ProcessesInfoVisitor(directives, types, channels);
		start.apply(procInfo);
		Messenger.getInstance(this).finish(Messenger.RENAMING_SUBSTITUTION);
		
		////////////////////////////////////
		// ANALYSE THE PARALLEL COMBINATIONS
		////////////////////////////////////
		Messenger.getInstance(this).start(Messenger.ANALISYNG_PARALLELISM);
		parallelismAnalyser = 
			new ParallelismAnalyserVisitor(procInfo, directives, constants);
		start.apply(parallelismAnalyser);
		Messenger.getInstance(this).finish(Messenger.ANALISYNG_PARALLELISM);
		
		
		/////////////////////
		// CHECK RESTRICTIONS
		///////////////////// 
		Messenger.getInstance(this).start(Messenger.CHECKING_PRIMARY_RESTRICTIONS);
		RestrictionsVisitor restrictions = new RestrictionsVisitor(types, channels, procInfo, 
				parallelismAnalyser, directives,
				constants);
		start.apply(restrictions);
		Messenger.getInstance(this).finish(Messenger.CHECKING_PRIMARY_RESTRICTIONS);
		
		///////////////////////
		// CHECK TAIL RECURSION
		///////////////////////
		Messenger.getInstance(this).start(Messenger.CHECKING_SECONDARY_RESTRICTIONS);
		TailRecursionVisitor recursion = new TailRecursionVisitor(procInfo);
		start.apply(recursion);
		Messenger.getInstance(this).finish(Messenger.CHECKING_SECONDARY_RESTRICTIONS);
		

		
		////////////////////////////
		// CHECK MULTI-SYNCHRONIZATION
		////////////////////////////
		Messenger.getInstance(this).start(Messenger.ANALISYNG_MULT_SYNCHRONIZATION);
		MultiSyncVisitor multisync = new MultiSyncVisitor(types,channels, parallelismAnalyser);
		start.apply(multisync);
		Messenger.getInstance(this).finish(Messenger.ANALISYNG_MULT_SYNCHRONIZATION);
		
		////////////////////////////
		// SEARCH KEYWORDS
		////////////////////////////
		Messenger.getInstance(this).start(Messenger.SEARCHING_KEYWORDS);
		KeywordsVisitor keywords = new KeywordsVisitor();
		start.apply(keywords);
		Messenger.getInstance(this).finish(Messenger.SEARCHING_KEYWORDS);
		
		//////////////////////
		// TRANSLATE PROCESSES
		//////////////////////
		Messenger.getInstance(this).start(Messenger.TRANSLATING_PROCESSES);
		String code = "";
		ProcessesTranslatorVisitor translator = 
			new ProcessesTranslatorVisitor(procInfo,directives,types,
					recursion,channels,constants,
					multisync,parallelismAnalyser);
		
		// As we do not have parallel in the specification, we are allowed to consider
		// that the value of the parallelism counter will not be altered in
		// the translation of the specification
		start.apply(translator);
		Messenger.getInstance(this).finish(Messenger.TRANSLATING_PROCESSES);
		
		
		//////////////////
		// FORMAT CODE
		////////////////// 
		String finalCode = translator.getHcCode();

		Messenger.getInstance(this).start(Messenger.FORMATTING_OUTPUT);
		code = CodeFormatting.format(finalCode);
		Messenger.getInstance(this).finish(Messenger.FORMATTING_OUTPUT);

		long endTime = System.currentTimeMillis();
		Messenger.getInstance().start(Messenger.TRANSLATION);
		long duration = endTime-startTime;
		Messenger.getInstance().finish(Messenger.TRANSLATION, duration);

		List<Warning> warnings = translator.getWarnings();
		for (int i=0; i<warnings.size(); i++) {
			Messenger.getInstance(this).warning(warnings.get(i));
		}
		return code;

		
	}
	
	
	private String translateCSP_OLD(File file) throws 
	FileNotFoundException, 
	IOException, 
	ParserException, 
	LexerException {  
		
		boolean keepGoing = false;
		
		long startTime = System.currentTimeMillis();

		/////////////////////////////////
		// PARSING AND CHECKING MAIN FILE
		/////////////////////////////////
		Node start = ParserUtil.parser(file);
		
		//////////////////////
		// LOAD DIRECTIVE INFO
		////////////////////// 
		Messenger.getInstance(this).start(Messenger.LOADING_DIRECTIVES);
		DirectivesVisitor directives = new DirectivesVisitor();
		start.apply(directives);
		Messenger.getInstance(this).finish(Messenger.LOADING_DIRECTIVES);


		//////////////////////
		// LOAD CODE INSERTION
		////////////////////// 

		Messenger.getInstance(this).start("INSERTION OF CODE");
		CodeInsertionVisitor codeInsertion = new CodeInsertionVisitor(directives);
		start.apply(codeInsertion);
		
		start = codeInsertion.getStartNode(file, directives);
		directives = codeInsertion.getNewDirectives();
		Messenger.getInstance(this).finish("INSERTION OF CODE");
		
		///////////////////////////////
		// LOAD CONSTANTS AND FUNCTIONS
		///////////////////////////////
		Messenger.getInstance(this).start(Messenger.LOADING_CONSTANTS_AND_FUNCTIONS);
		ConstantsAndFunctionsVisitor constants = new ConstantsAndFunctionsVisitor(directives);
		start.apply(constants);
		Messenger.getInstance(this).finish(Messenger.LOADING_CONSTANTS_AND_FUNCTIONS);
		
		/////////////////
		// LOAD DATATYPES
		/////////////////
		Messenger.getInstance(this).start(Messenger.LOADING_DATATYPES);
		DatatypesVisitor types = new DatatypesVisitor(directives,constants);
		start.apply(types);
		Messenger.getInstance(this).finish(Messenger.LOADING_DATATYPES);
		

		Messenger.getInstance(this).start("REMOVING FORCED INTERLEAVE");
		ForcedInterleavingProtocol forcedInterleavingProtocol = new ForcedInterleavingProtocol(file, directives, types, constants, codeInsertion);
		Node mainStart = forcedInterleavingProtocol.preRenaming();
		Messenger.getInstance(this).finish("REMOVING FORCED INTERLEAVE");
		
		Messenger.getInstance(this).start("RENAMING");
		//RenamingVisitor renaming = new RenamingVisitor(forcedInterleavingProtocol);
		//mainStart.apply(renaming);
		Messenger.getInstance(this).finish("RENAMING");

		////////////////////////////
		// LOAD CHANNELS INFORMATION
		////////////////////////////
		Messenger.getInstance(this).start(Messenger.LOADING_CHANNELS_INFO);
		ChannelsVisitor channels = new ChannelsVisitor(directives, types, constants);
		mainStart.apply(channels);
		Messenger.getInstance(this).finish(Messenger.LOADING_CHANNELS_INFO);
		
		///////////////////////////////////////////
		// LOAD PROCESSES ARGUMENTS AND DEFINITIONS
		///////////////////////////////////////////
		Messenger.getInstance(this).start(Messenger.LOADING_PROCESSES_INFO);
		ProcessesInfoVisitor procInfo = new ProcessesInfoVisitor(directives, types, channels, codeInsertion);
		mainStart.apply(procInfo);
		Messenger.getInstance(this).finish(Messenger.LOADING_PROCESSES_INFO);
		
		////////////////////////////////////
		// ANALYSE THE PARALLEL COMBINATIONS
		////////////////////////////////////
		Messenger.getInstance(this).start(Messenger.ANALISYNG_PARALLELISM);
		ParallelismAnalyserVisitor parallelismAnalyser = 
			new ParallelismAnalyserVisitor(procInfo, directives, constants);
		mainStart.apply(parallelismAnalyser);

		//RenameProcesses renameProcesses = new RenameProcesses(parallelismAnalyser, forcedInterleavingProtocol, procInfo);
		//mainStart.apply(renameProcesses);
		parallelismAnalyser = 
			new ParallelismAnalyserVisitor(procInfo, directives, constants);
		mainStart.apply(parallelismAnalyser);
		Messenger.getInstance(this).finish(Messenger.ANALISYNG_PARALLELISM);
				
		/////////////////////
		// CHECK RESTRICTIONS
		///////////////////// 
		Messenger.getInstance(this).start(Messenger.CHECKING_PRIMARY_RESTRICTIONS);
		RestrictionsVisitor restrictions = new RestrictionsVisitor(types, channels, procInfo, 
				parallelismAnalyser, directives,
				constants);
		mainStart.apply(restrictions);
		Messenger.getInstance(this).finish(Messenger.CHECKING_PRIMARY_RESTRICTIONS);
		
		///////////////////////
		// CHECK TAIL RECURSION
		///////////////////////
		Messenger.getInstance(this).start(Messenger.CHECKING_SECONDARY_RESTRICTIONS);
		TailRecursionVisitor recursion = new TailRecursionVisitor(procInfo);
		mainStart.apply(recursion);
		Messenger.getInstance(this).finish(Messenger.CHECKING_SECONDARY_RESTRICTIONS);
		
		////////////////////////////
		// CHECK MULTI-SYNCHRONIZATION
		////////////////////////////
		Messenger.getInstance(this).start("ANALISYNG_MULT-SYNCHRONIZATION");
		MultiSyncVisitor multisync = new MultiSyncVisitor(types,channels, parallelismAnalyser);
		mainStart.apply(multisync);
		Messenger.getInstance(this).finish("ANALISYNG_MULT-SYNCHRONIZATION");
		
		////////////////////////////
		// SEARCH KEYWORDS
		////////////////////////////
		Messenger.getInstance(this).start(Messenger.SEARCHING_KEYWORDS);
		KeywordsVisitor keywords = new KeywordsVisitor();
		mainStart.apply(keywords);
		Messenger.getInstance(this).finish(Messenger.SEARCHING_KEYWORDS);
		
		//////////////////////
		// TRANSLATE PROCESSES
		//////////////////////
		Messenger.getInstance(this).start(Messenger.TRANSLATING_PROCESSES);
		String code = "";
		ProcessesTranslatorVisitor translator = 
			new ProcessesTranslatorVisitor(procInfo,directives,types,
					recursion,channels,constants,
					multisync,parallelismAnalyser);
		
		// As we do not have parallel in the specification, we are allowed to consider
		// that the value of the parallelism counter will not be altered in
		// the translation of the specification
		mainStart.apply(translator);
		Messenger.getInstance(this).finish(Messenger.TRANSLATING_PROCESSES);
		
		
		//////////////////
		// FORMAT CODE
		////////////////// 
		String finalCode = translator.getHcCode();

		Messenger.getInstance(this).start(Messenger.FORMATTING_OUTPUT);
		code = CodeFormatting.format(finalCode);
		Messenger.getInstance(this).finish(Messenger.FORMATTING_OUTPUT);

		long endTime = System.currentTimeMillis();
		Messenger.getInstance().start(Messenger.TRANSLATION);
		long duration = endTime-startTime;
		Messenger.getInstance().finish(Messenger.TRANSLATION, duration);

		List<Warning> warnings = translator.getWarnings();
		for (int i=0; i<warnings.size(); i++) {
			Messenger.getInstance(this).warning(warnings.get(i));
		}
		return code;

		
	}
	
	public final class RenameProcesses extends ExtendedDepthFirstAdapter {
		private List <String> eventList;
		private boolean isRenamingSyncChannels;
		private List <String> renamingSyncChannels;
		private boolean hasChangedRSC;
		private Set <Integer> branches;
		private Set <Integer> leftBranches;
		private Set <Integer> rightBranches;
		private ParallelismAnalyserVisitor parallelismAnalyser;
		private ForcedInterleaveInsertion forcedInterleavingProtocol;
		private ProcessesInfoVisitor procInfo;
		public RenameProcesses(ParallelismAnalyserVisitor parallelismAnalyser, ForcedInterleaveInsertion forcedInterleavingProtocol, ProcessesInfoVisitor procInfo){
			this.eventList = null;
			this.isRenamingSyncChannels = false;
			branches = new HashSet <Integer>();
			this.parallelismAnalyser = parallelismAnalyser;
			this.forcedInterleavingProtocol = forcedInterleavingProtocol;
			this.procInfo = procInfo;
		}
		public void caseACspProcessDefinition(ACspProcessDefinition process) {
			String name = process.getName().toString().trim();
//			if (!this.forcedInterleavingProtocol.isProcessToBeRenamed(name)){
				this.procInfo.getProcDefinition(name).apply(this);
				//process.getExpression().apply(this);
//			}
		}
		public void caseACspRenamingProcess(ACspRenamingProcess process) {
		}
		public void caseACspProcessCallExpr(ACspProcessCallExpr node){
			int nodeId = this.parallelismAnalyser.getAbsoluteNodeBranchId(node);
			// Quando o 'branch' do n  igual a -1, o processo no faz parte do fluxo principal de execuo da especificao. Assim, no precisa ser renomeado.
			if (nodeId != -1) {
				String name = node.getName().getText().trim();
				this.branches.add(nodeId);
				if (this.forcedInterleavingProtocol.isProcessToBeRenamed(name)){
					if (node.getArguments() != null){
						ACspProcessCallExpr renamedNode = new ACspProcessCallExpr(
								new TName(name+"_RN"+nodeId, node.getName().getLine(), node.getName().getPos()),
								(PArguments) node.getArguments().clone());
						node.replaceBy(renamedNode);
					} else {
						ACspProcessCallExpr renamedNode = new ACspProcessCallExpr(
								new TName(name+"_RN"+nodeId, node.getName().getLine(), node.getName().getPos()));
						node.replaceBy(renamedNode);
					}
				} 
			}
		}
		
		public void caseACspInterleaveProcess(ACspInterleaveProcess node) {
			Set <Integer> leftBranchesLocal = new HashSet<Integer>();
			Set <Integer> rightBranchesLocal = new HashSet <Integer>();
			
			this.branches.clear();
			node.getLeftProcess().apply(this);
			leftBranchesLocal.addAll(this.branches);
			
			this.branches.clear();
			node.getRightProcess().apply(this);
			rightBranchesLocal.addAll(this.branches);
			
			this.branches.clear();
	        this.branches.addAll(rightBranchesLocal);
	        this.branches.addAll(leftBranchesLocal);
	    }
		
		public void caseACspSharingParallelProcess(ACspSharingParallelProcess node) {
			Set <Integer> leftBranchesLocal = new HashSet<Integer>();
			Set <Integer> rightBranchesLocal = new HashSet <Integer>();
			
			this.branches.clear();
			node.getLeftProcess().apply(this);
			leftBranchesLocal.addAll(this.branches);
			
			this.branches.clear();
			node.getRightProcess().apply(this);
			rightBranchesLocal.addAll(this.branches);
			
			this.leftBranches = leftBranchesLocal;
			this.rightBranches = rightBranchesLocal;
			this.isRenamingSyncChannels = true;
	        node.getExpression().apply(this);
	        this.isRenamingSyncChannels = false;
			
	        this.branches.clear();
	        this.branches.addAll(rightBranchesLocal);
	        this.branches.addAll(leftBranchesLocal);
	    }
		
		public void caseACspBaseEventList(ACspBaseEventList node) {  // Criado dia 01/12/2011
			
			if (this.isRenamingSyncChannels){
				
				PEvent event = node.getEvent();
				String channelName = ((ACspEvent) event).getChannelName().toString().trim();
				
				List <String> renamed = forcedInterleavingProtocol.getRenamedChannels(channelName);
				if (renamed.isEmpty()){
					eventList.add(channelName);
				} else {
					eventList.addAll(renamed);
				}
			}
		
			
		}
		
		public void caseACspLinkedEventList(ACspLinkedEventList node) {  // Criado dia 01/12/2011
			if (this.isRenamingSyncChannels){
				PEvent event = node.getEvent();
				String channelName = ((ACspEvent) event).getChannelName().toString().trim();
				
				List <String> renamed = forcedInterleavingProtocol.getRenamedChannels(channelName);
				if (renamed.isEmpty()){
					eventList.add(channelName);
				} else {
					for (String chan:renamed){
						eventList.add(chan);
					}
				}
			}
			node.getNext().apply(this);
			
		}
		
		
		public void caseACspProductionsSet(ACspProductionsSet node) {  // Criado dia 01/12/2011 
			if (this.isRenamingSyncChannels){
				
				eventList = new Vector<String>();
				node.getEventList().apply(this);
				
				PEventList events = null; 
				
				for (String chan:eventList){
					if (events == null) {
						ACspEvent event = new ACspEvent(new TName(chan), (List)new Vector<PProjection>(), (List)new Vector<PComunication>());
						events = new ACspBaseEventList(event);
					} else {
						ACspEvent event = new ACspEvent(new TName(chan), (List)new Vector<PProjection>(), (List)new Vector<PComunication>());
						events = new ACspLinkedEventList(event,new TComma(), events);
					}
					
				}
				node.setEventList(events);
				
			}
		}
		
		public void caseACspProductionsSet2(ACspProductionsSet node) {  // Alterado dia 01/12/2011 
			if (this.isRenamingSyncChannels){
				this.renamingSyncChannels = new Vector<String>();
				this.hasChangedRSC = false;
				node.getEventList().apply(this);
				if (this.hasChangedRSC){
					// Create new ElementList
					PEventList eventList = null;
					for (String chan:this.renamingSyncChannels){
						if (eventList == null){
							ACspEvent event = new ACspEvent();
							event.setChannelName(new TName(chan));
							eventList = new ACspBaseEventList(event);
						} else if (eventList instanceof ACspBaseEventList){
							ACspEvent event = new ACspEvent();
							event.setChannelName(new TName(chan));
							eventList = new ACspLinkedEventList(event,new TComma(),eventList);
						} else if (eventList instanceof ACspLinkedEventList){
							ACspEvent event = new ACspEvent();
							event.setChannelName(new TName(chan));
							eventList = new ACspLinkedEventList(event,new TComma(),eventList);
						}
					}
					node.setEventList(eventList);
				}
			}
		}
		public void caseACspEvent2(ACspEvent node) {  // Alterado dia 01/12/2011
			if (this.isRenamingSyncChannels){
				String name = node.getChannelName().getText().toString().trim();
				String keyName = HandelC.CHAN_DECL + HandelC.UNDERSCORE + name;
				boolean channelChanged = false;
				if (this.parallelismAnalyser.getRealSyncLUT().keySet().contains(keyName)
						&& this.parallelismAnalyser.getForcedInterleavingLUT().keySet().contains(keyName)){
					Set<Set<Integer>> realSync = this.parallelismAnalyser.getRealSyncLUT().get(keyName);
					for (Set<Integer> friends:realSync){
						Set<Integer> left = new HashSet<Integer>();
						left.addAll(this.leftBranches);
						left.retainAll(friends);
						if (!left.isEmpty()){
							Set<Integer> right = new HashSet<Integer>();
							right.addAll(this.rightBranches);
							right.retainAll(friends);
							if (!right.isEmpty()){
								this.renamingSyncChannels.add(name+HandelC.UNDERSCORE+friends.hashCode());
								channelChanged = true;
								this.hasChangedRSC = true;
							}
						}
					}
				}
				if (!channelChanged){
					this.renamingSyncChannels.add(name);
				}
			}
	    }
	}

	protected void translateButtonOK () {
		try {
			Thread.sleep(500);
		} catch(InterruptedException e) {
			e.printStackTrace();
		}
		this.translateButton.setIcon( new ImageIcon(path + "generategreen.gif"));
	}
	protected void translateButtonNOK () {
		try {
			Thread.sleep(100);
		} catch(InterruptedException e) {
			e.printStackTrace();
		}
		this.translateButton.setIcon( new ImageIcon(path + "generatered.gif"));
	}

	protected void keyPressed() {
		this.translateButtonNOK();
	}

	public void showKnownException(Throwable e) {
		Messenger.getInstance(this).error(e.getMessage());
		askTryAgain(e);
	}

	public void showUnknownException(Throwable e) {
		Messenger.getInstance(this).uerror(e.getMessage());
		askTryAgain(e);
	}

	private void askTryAgain(Throwable t){
		t.printStackTrace();
		String message = t.getMessage();
		int lineOfError = -1;
		int columnOfError = -1;
		int index =0;

		/////////////////////////////
		// Changes the indication of location of error if it happened in the main process

		if (t instanceof CspAnalyserException){
			Location l = CspNodeLocator.getLocation(((CspAnalyserException)t).getNode());
			lineOfError = l.getLine();
			columnOfError = l.getPos();
			index = message.indexOf("-")+1;
		} else if (t instanceof ParserException){
			lineOfError = ((ParserException)t).getToken().getLine();
			columnOfError = ((ParserException)t).getToken().getPos();
			index = message.indexOf("]")+1;
		}

		int lineCount = this.cspFrame.getLineCount();

		if (lineCount < lineOfError) {
			if (index > 0) {
				message = message.substring(index);
			}
			message = Messenger.IN_MAIN + message;
		}

		// Appends line and column information
		if (lineOfError >= 0 || columnOfError >=0) {
			message = message + " @ [";
			if (lineOfError != 0) {
				message = message + Messenger.LINE + lineOfError;
			}
			if (columnOfError != 0) {
				message = message + ", " + Messenger.COLUMN + columnOfError;
			}
			message = message + "]";
		}

		/////////////////////////////
		/////////////////////////////
		JOptionPane.showMessageDialog(null, Messenger.ERROR_MESSAGE+message, Messenger.ERROR_MESSAGE_TITLE, JOptionPane.ERROR_MESSAGE); 
	}

	private void disableTranslate() {
		translate.setEnabled(false);
		translateButton.setEnabled(false);
	}

	private void enableTranslate() {
		translate.setEnabled(true);
		translateButton.setEnabled(true);
		this.translateButtonNOK();
	}


	public static void main(String[] args) {
		MainFrame a = new MainFrame();
		a.setVisible(true);

	}

	public void setUndoEnabled(boolean b) {
		undo.setEnabled(b);
		image_undo.setEnabled(b);
	}
	public void setRedoEnabled(boolean b) {
		redo.setEnabled(b);
		image_redo.setEnabled(b);
	}

	public void atualizarInfoCSPTela(JInternalFrame frame) {
		infoTelas.AtualizarCsp(frame);
	}

	public void atualizarInfoHandelCTela(JInternalFrame frame) {
		infoTelas.AtualizarHandelC(frame);
	}
	public void atualizarInfoMsgsTela(JInternalFrame frame) {
		infoTelas.AtualizarMsg(frame);
	}

}
