package csp2hc.gui;

import java.awt.Color;
import java.awt.Font;

import javax.swing.ImageIcon;
import javax.swing.JInternalFrame;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableColumn;



/**
 * Janela interna que exibe as mensagens do tradutor
 * @author rolango
 *
 */
public class MessagesFrame extends JInternalFrame{
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	/* Atributos relacionados a tabela */
	private String[] colunas;
	private JTable table;
	private DefaultTableModel msgList;
	private String path = "images/";
	private JScrollPane roll;
	private MainFrame parent;
	

	public MessagesFrame(MainFrame parent) {
		
		super("Messages", true, false, true, true);
		this.setTitle("Messages");
		this.parent = parent; 
		
		//Configurando Tabela
		//Inicializando Colunas
		colunas = new String[2];
		colunas[0] = "Type";
		colunas[1] = "Message";
		
		Object[][] data = null;

		msgList = new DefaultTableModel(data, colunas){
			@Override 
			public boolean isCellEditable(int arg0, int arg1) {
				return false;//Torna as celulas nao editaveis
			}
		};
		table = new JTable(msgList){
			//  Returning the Class of each column will allow different
			//  renderers to be used based on Class
			public Class getColumnClass(int column)
			{
				if (column == 0) {
					return ImageIcon.class;
				}
				else {
					return String.class;
				}
			}
		};
		
		table.setPreferredScrollableViewportSize(table.getPreferredSize());
		table.setBackground(Color.white);
		table.setForeground(Color.black);
		table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		table.setFont(new Font("", Font.PLAIN, 12));
		table.setColumnSelectionAllowed(false);
		table.setRowSelectionAllowed(true);
		table.setRowHeight(25);
		
		
		roll = new JScrollPane(table);
		roll.setLocation(0,0);
		roll.setSize(this.getWidth(), this.getHeight());
		
		this.addHierarchyBoundsListener(new java.awt.event.HierarchyBoundsListener () {
			public void ancestorMoved(java.awt.event.HierarchyEvent e) {
			}
			public void ancestorResized(java.awt.event.HierarchyEvent e) {
				atualizarInfoTela();
			}
		}
		);

		this.add(roll);
		this.setVisible(true);
		
	}
	protected void atualizarInfoTela() {
		parent.atualizarInfoMsgsTela(this);
	}
	public void setGUIElements(){
		//Definindo tamanho das colunas
		TableColumn coluna1 = table.getColumnModel().getColumn(0);
		TableColumn coluna2 = table.getColumnModel().getColumn(1);
		int firstColumnWidth = 40;
		coluna1.setPreferredWidth(firstColumnWidth);
		coluna2.setPreferredWidth(this.getWidth()-firstColumnWidth);
	}
	
	
	private class AddRowThread extends Thread{
		
		private TypeMsg msg;
		private String txt;
		
		public AddRowThread(TypeMsg msg, String txt) {
			this.msg = msg;
			this.txt = txt;
		}
		public void run(){

			if (msg.equals(TypeMsg.ERROR)){
				msgList.addRow(new Object[]{new ImageIcon(path + "error.gif"), txt});
			} else if (msg.equals(TypeMsg.WARNING)) {
				msgList.addRow(new Object[]{new ImageIcon(path + "warning.gif"), txt});
			} else if (msg.equals(TypeMsg.INFO)) {
				msgList.addRow(new Object[]{new ImageIcon(path + "info.gif"), txt});
			}
			setListSelection();
		}
	}
	
	
	
private class AppendRowThread extends Thread{
		
		private String txt;
		
		public AppendRowThread(String txt) {
			this.txt = txt;
		}
		public void run(){

			int lastRow = msgList.getRowCount() - 1;
			String newMessage = msgList.getValueAt(lastRow, 1) + " " + txt; 
			msgList.setValueAt(newMessage, lastRow, 1);
			setListSelection();
		}
	}
	
	
	/**
	 * Adiciona a msg a janela de msgs
	 * @param msg
	 * @param txt
	 */
	public void addRow(TypeMsg msg, String txt) {
		
		SwingUtilities.invokeLater(new AddRowThread(msg,txt));
		
		/*if (msg.equals(TypeMsg.ERROR)){
			msgList.addRow(new Object[]{new ImageIcon(path + "error.gif"), txt});
		} else if (msg.equals(TypeMsg.WARNING)) {
			msgList.addRow(new Object[]{new ImageIcon(path + "warning.gif"), txt});
		} else if (msg.equals(TypeMsg.INFO)) {
			msgList.addRow(new Object[]{new ImageIcon(path + "info.gif"), txt});
		}
		this.setListSelection();*/
	}

	/**
	 * Adiciona uma msg a msg que ja esta na janela de msgs
	 * @param msg
	 * @param txt
	 */
	public void appendRow(String txt) {
		SwingUtilities.invokeLater(new AppendRowThread(txt));
		/*int lastRow = msgList.getRowCount() - 1;
		String newMessage = msgList.getValueAt(lastRow, 1) + " " + txt; 
		msgList.setValueAt(newMessage, lastRow, 1);
		this.setListSelection();*/
	}
	
	public void setListSelection() {
		SwingUtilities.invokeLater(new Runnable() {

			public void run() {
				roll.getVerticalScrollBar().setValue(roll.getVerticalScrollBar().getMaximum());
			}
			
		});
		
	}

	public void clear() {
		while (msgList.getRowCount() > 0) {
			msgList.removeRow(0);
		}
	}

}
