/*
 * HandelC.java
 *
 * Contains the Handel-C constructors used
 * Created on 18 January 2006, 18:17
 */

package csp2hc.messages;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import lmf.formula.csp.type.CspSetType;
import lmf.formula.csp.type.CspType;
import lmf.formula.csp.type.CspUnknownType;

import csp2hc.exception.InvalidSelectBlockArgumentsException;
import csp2hc.exception.UnsupportedConstructorException;
import csp2hc.util.CspUtil;
import csp2hc.util.ListUtil;
import csp2hc.util.StringUtil;

/**
 *
 * @author  marcel oliveira
 */
public class HandelC {
    // File extensions
    public static String CSP_EXTENSION = "csp";
    public static String EXTENSION = "hcc";
    public static String DOT = ".";
    public static String TEMPORARY_FILE = "___temp";

    // Handel-C Keywords
    public static String ADDRESS = "&";
    public static String ALTERNATION = "prialt";
    public static String ARRAY_BEGIN = "[";
    public static String ARRAY_END = "]";
    public static String ASSIGNMENT = "=";    
    public static String AUTO = "auto";
    public static String BINARY_PREFIX = "0b";    
    public static String BLOCK_BEGIN = "{";
    public static String BIT_RANGE = ":";
    public static String BLOCK_END = "}";
    public static String BREAK = "break";
    public static String BUS_IN = "bus_in";
    public static String BUS_OUT = "bus_out";
    public static String CASE = "case";
    public static String CASE_SEP = ":";
    public static String CHAN_DECL = "chan";
    public static String CHANIN = "chanin";
    public static String CHANOUT = "chanout";
    public static String CHAR = "char";
    public static String CLOCK = "clock";
    public static String CONDITIONAL_COND = "?";    
    public static String CONDITIONAL_SEP = ":";    
    public static String CONSTANT = "const";
    public static String CONTINUE = "continue";
    public static String DEFAULT = "default";
    public static String DEFINE = "#define";    
    public static String DELAY = "delay";
    public static String DATA = "data";
    public static String DO = "do";
    public static String DOUBLE = "double";
    public static String END_COMMAND = ";";
    public static String ENUM = "enum";    
    public static String EXPR = "expr";
    public static String EXTERN = "extern";
    public static String EXTERNAL_DIVIDE = "external_divide";
    public static String FAMILY = "family";
    public static String FLOAT = "float";
    public static String FOR = "for";
    public static String GOTO = "goto";
    public static String HC_MACRO = "macro";
    public static String IFSELECT = "ifselect";
    public static String IN = "in";
    public static String INLINE = "inline";
    public static String INPUT_COMMUNICATION = "?";
    public static String INT = "int";    
    public static String INT_WIDTH = "intwidth";
    public static String INTERFACE = "interface";
    public static String INTERNAL = "internal";
    public static String INTERNAL_DIVIDE = "internal_divide";
    public static String LINE_COMMENT = "//";
    public static String LOGICAL_AND = "&&";    
    public static String LOGICAL_BITWISE_AND = "&";    
    public static String LOGICAL_BITWISE_NOT = "~";    
    public static String LOGICAL_BITWISE_OR = "|";    
    public static String LOGICAL_BITWISE_XOR = "^";    
    public static String LOGICAL_NOT = "!";    
    public static String LOGICAL_OR = "||";    
    public static String CONCAT = "@";    
    public static String TAKE_LSB = "<-";    
    public static String LOGICAL_VALUES = "2";    
    public static String LONG = "long";    
    public static String MACRO_EXP = HC_MACRO + " " + EXPR;
    public static String PROC = "proc";    
    public static String MACRO_PROC = HC_MACRO + " " + PROC;
    public static String MAIN = "main";
    public static String MATH_INC = "++";
    public static String MATH_DEC = "--";    
    public static String MATH_DIFF = "-";    
    public static String MATH_DIV = "/";    
    public static String MATH_MULT = "*";    
    public static String MATH_NEG = "-";    
    public static String MATH_ONE = "1";    
    public static String MATH_POW = "**";    
    public static String MATH_REMAINDER = "%";    
    public static String MATH_SUM = "+";    
    public static String MATH_ZERO = "0";    
    public static String MPRAM = "mpram";    
    public static String OUTPUT_COMMUNICATION = "!";
    public static String PAR_BEGIN = "(";
    public static String PAR_END = ")";
    public static String PARALLEL = "par";
    public static String PART = "part";    
    public static String RAM = "ram";    
    public static String REFERENCE = "*";
    public static String REGISTER = "register";    
    public static String REL_EQUALS = "==";    
    public static String REL_GREATER = ">";    
    public static String REL_GREATER_EQUALS = ">=";    
    public static String REL_LESS = "<";    
    public static String REL_LESS_EQUALS = "<=";    
    public static String REL_NOT_EQUALS = "!=";    
    public static String RELEASE_SEMAPHOR = "releasesema";
    public static String RESET = "reset";    
    public static String RETURN = "return";
    public static String ROM = "rom";    
    public static String SELECT = "select";
    public static String SEMAPHOR = "sema";
    public static String SEP_ARGUMENTS = ",";
    public static String SEQUENCE = "seq";
    public static String SHARED = "shared";    
    public static String SHORT = "short";    
    public static String SIGNAL = "signal";    
    public static String SIGNED = "signed";    
    public static String SIZEOF = "sizeof";    
    public static String SPACE = " ";
    public static String STATIC = "static";
    public static String STRUCT = "struct";    
    public static String SWITCH = "switch";
    public static String TRY = "try";    
    public static String TRY_SEMAPHOR = "trysema";
    public static String TYPEDEF = "typedef";    
    public static String UNDEFINED = "undefined";    
    public static String UNSIGNED = "unsigned";    
    public static String VOID = "void";
    public static String VOLATILE = "volatile";    
    public static String WHILE = "while";    
    public static String WIDTH = "width";    
    public static String WITH = "with";    
    public static String WOM = "wom";

    // Handel-C and CSP keywords
    public static String EXTERNAL = "external";
    public static String IF = "if";    
    public static String ELSE = "else";    
    public static String LOGICAL_TRUE = "true";    
    public static String LOGICAL_FALSE = "false";    
    public static String LET = "let";    
    public static String SET = "set";    
    public static String SEQ = "seq";    
    public static String CSP_SET_UNION = "union";    
    public static String CSP_SET_INTER = "inter";    
    public static String CSP_SET_DIFF = "diff";    
    public static String POWERSET = "Set";
    
    // Translator Keywords
    public static String DUMB_CLOCK = "clock1";
    public static String UNDERSCORE = "_";

    public static String ACTION = "ACTION";
    public static String A_READ = "A_READ";
    public static String A_WRITE = "A_WRITE";
    public static String ARE_READERS_READY = "ARE_READERS_READY";
    public static String MAX_MULTI_SYNC_SET = "MAX_MULTI_SYNC_SET";
    public static String BOOLEAN = "boolean";    
    public static String BRANCH = "BRANCH";
    public static String BRANCH_SET = "BRANCH_SET";
    public static String WRITER_BRANCH = "WRITER_BRANCH";
    public static String BRANCH_ID = "BRANCH" + HandelC.UNDERSCORE + "ID";
    public static String CHANNEL = "CHANNEL";
    public static String CLOCK_DEF = "set clock = external \""+DUMB_CLOCK+"\" "+END_COMMAND;
    public static String CONFIRM_FINISHED = "CONFIRM_FINISHED";
    public static String CONFIRMED = "CONFIRMED";
    public static String CONFIRMED_END = "confirmed_end";
    public static String CONFIRM_READERS = "CONFIRM_READERS";
    public static String LINE_BREAK = "LINE_BREAK";
    public static String COMMENT_LINE_BREAK = UNDERSCORE+UNDERSCORE+LINE_BREAK;    
    public static String COMMENT_SPACES = UNDERSCORE+UNDERSCORE+"SPACES"+HandelC.UNDERSCORE;    
    public static String COMMENT_TAB = UNDERSCORE+UNDERSCORE+"TAB";    
    public static String CURRENT_SYNC_STATUS_LUT = "CURRENT_SYNC_STATUS_LUT";
    public static String DONE_EVENTS = "DONE_EVENTS";
    public static String FALSE_DEF = LOGICAL_FALSE + SPACE + MATH_ZERO;    
    public static String FORCED_SYNC_LUT = "FORCED_SYNC_LUT";
    public static String MULTI_SYNC_LUT = "MULTI_SYNC_LUT";
    public static String HAS_WRITER_LUT = "HAS_WRITER_LUT";
    public static String HAS_JOINED = "HAS_JOINED";
    public static String HAS_LEFT = "HAS_LEFT";
    public static String INEXISTENT_CHANNEL = "INEXISTENT_CHANNEL";    
    public static String INTEGERL = "integer";    
    public static String IS_EMPTY_SET = "IS_EMPTY_SET";
    public static String JOIN_SEMAPHOR = "JOIN_SEMAPHOR";
    public static String KEEP_LOOPING = "KEEP_LOOPING";
    public static String LEAVE_SEMAPHOR = "LEAVE_SEMAPHOR";
    public static String LOCAL_COPY = UNDERSCORE+"local"+UNDERSCORE;
    public static String LOCK_VAR = "LOCK_VARIABLE";
    public static String LOOKUP_TABLE = "LUT";
    public static String MACRO = "MACRO";
    public static String MAINP = "mainp";
    public static String IS_AVAILABLE = "IS_AVAILABLE";
    public static String NO_WRITER = "NO_WRITER";
    public static String OUTPORT = "OutPort";
    public static String MULTI_SYNC_READY = "MULTI_SYNC_READY";
    public static String MULTI_SYNC_READY_WITH_ME = "MULTI_SYNC_READY_WITH_ME";
    public static String IS_SYNC_READY = "IS_SYNC_READY";
    public static String MAY_LEAVE = "MAY_LEAVE";
    public static String MAY_JOIN = "MAY_JOIN";
    public static String IS_ALREADY = "IS_ALREADY";
    public static String BRANCH_MEMBER = "BRANCH_MEMBER";
    public static String MUTUAL_RECURSION = "MUTUAL_REC";
    public static String NEG = "neg"+ UNDERSCORE;    
    public static String OFFSET = UNDERSCORE+"offset";
    public static String PROGRAM_COUNTER = "PROGRAM_COUNTER";
    public static String SET_DIFF = "SET_DIFF";
    public static String SET_INTER = "SET_INTER";
    public static String SET_UNION = "SET_UNION";
    public static String SINGLETON = "singleton";
    public static String SYNC = "sync";    
    public static String SYNC_TYPE = SYNC.toUpperCase();
    public static String RANDOM = "random";
    public static String RANDOM_VAR = "random_var";
    public static String SLEEP = "SLEEP";
    public static String SLEEP_TIME = "SLEEP_TIME";
    public static String SYNC_DEF =  TYPEDEF + SPACE + UNSIGNED + SPACE + INT + SPACE + MATH_ONE + SPACE + HandelC.SYNC_TYPE + END_COMMAND; 
    public static String SYNC_IN = "syncin";    
    public static String SYNC_IN_DEF = SYNC_TYPE + SPACE + SYNC_IN + END_COMMAND;    
    public static String SYNC_OUT = "syncout";    
    public static String SYNC_OUT_DEF = CONSTANT + SPACE + SYNC_TYPE + SPACE + SYNC_OUT + SPACE + ASSIGNMENT + MATH_ZERO + END_COMMAND;
    public static String TRUE_DEF = LOGICAL_TRUE + SPACE + MATH_ONE;    
    public static String TRY_CONFIRM_READERS = "TRY_CONFIRM_READERS";
    public static String TRY_JOIN_SEMAPHOR = "TRY_JOIN_SEMAPHOR";
    public static String TRY_LEAVE_SEMAPHOR = "TRY_LEAVE_SEMAPHOR";
    public static String WITHDRAW_IN_COMM = "WITHDRAW_IN_COMM";
    public static String WITHDRAW_IN_CONF = "WITHDRAW_IN_CONFIRMATION";
    public static String WITHDRAW_IN_FINISHED = "WITHDRAW_IN_FINISHED";
    public static String WITHDRAW_SIGNALS = "WITHDRAW_SIGNALS";
    public static String SET_TYPE_PREFIX = HandelC.POWERSET+HandelC.PAR_BEGIN;
    public static String TYPEOF = "typeof";
    public static String TYPEOF_PREFIX = HandelC.TYPEOF+HandelC.PAR_BEGIN;
    
    // SUFFIXES/PREFIXES OF KEYWORDS
    private static String SET_SUFFIX = UNDERSCORE+"set";
    private static String CARDINALITY = UNDERSCORE+"card";
    private static String EMPTY_SET_SUFFIX = UNDERSCORE+"nil";
    private static String SEMAPHOR_SUFFIX = UNDERSCORE+"SEMA";
    private static String IN_BUS_PREFIX = "in"+UNDERSCORE+"bus";
    private static String OUT_BUS_PREFIX = "out"+UNDERSCORE+"bus";
    private static String BUS_VALUE_PREFIX = "bus";
    private static String BUS_VALUE_SUFFIX = "value";
    private static String IN_BUS_METHOD_PREFIX = "InBus";
    private static String OUT_BUS_METHOD_PREFIX = "OutBus";

    // Argument names for macros 
    public static String INT_ARG_I = "I";
    public static String SET_ARG_S = "S";
    public static String SET_ARG_T = "T";
    public static String BRANCH_ARG_B = "B";
    public static String CHANNEL_ARG_C = "C";
    public static String ACTION_ARG_A = "A";

    // Set of Handel-C keywords
    private static String[] hcKeywords = new String[]{
    	ALTERNATION,
    	AUTO,
    	BREAK,
    	BUS_IN,
    	BUS_OUT,
    	CASE,
    	CHAN_DECL,
    	CHANIN,
    	CHANOUT,
    	CHAR,
    	CLOCK,
    	CONSTANT,
    	CONTINUE,
    	DEFAULT,
    	DEFINE,
    	DELAY,
    	DO,
    	DOUBLE,
    	ENUM,
    	EXPR,
    	EXTERN,
    	EXTERNAL_DIVIDE,
    	FAMILY,
    	FLOAT,
    	FOR,
    	GOTO,
    	HC_MACRO,
    	IFSELECT,
    	IN,
    	INLINE,
    	INT,
    	INT_WIDTH,
    	INTERFACE,
    	INTERNAL,
    	INTERNAL_DIVIDE,
    	LONG,
    	MAIN,
    	PARALLEL,
    	PART,
    	PROC,
    	RAM,
    	//REFERENCE,
    	REGISTER,
    	RELEASE_SEMAPHOR,
    	RESET,
    	RETURN,
    	ROM,
    	SELECT,
    	SEMAPHOR,
    	SEQUENCE,
    	SHARED,
    	SHORT,
    	SIGNAL,
    	SIGNED,
    	SIZEOF,
    	STATIC,
    	STRUCT,
    	SWITCH,
    	TRY,
    	TRY_SEMAPHOR,
    	TYPEDEF,
    	UNDEFINED,
    	UNSIGNED,
    	VOID,
    	VOLATILE,
    	WHILE,
    	WIDTH,
    	WITH,   
    	WOM
    };

    // Set of translator's keywords
    private static String[] toolKeywords = new String[]{
        ACTION,
        ARE_READERS_READY,
        A_READ,
        A_WRITE, 
    	BOOLEAN,
    	BRANCH,
    	CHANNEL,
    	CONFIRM_FINISHED,
    	CONFIRMED,
    	CONFIRMED_END,
    	CONFIRM_READERS,
    	LINE_BREAK,
    	CURRENT_SYNC_STATUS_LUT,
    	DUMB_CLOCK,
    	FORCED_SYNC_LUT,
    	MULTI_SYNC_LUT,
    	HAS_WRITER_LUT,
    	HAS_JOINED,
    	HAS_LEFT,
    	INEXISTENT_CHANNEL,
    	INTEGERL,
    	IS_EMPTY_SET,
    	JOIN_SEMAPHOR,
    	KEEP_LOOPING,
    	LEAVE_SEMAPHOR,
    	LOCK_VAR,
    	IS_AVAILABLE,
    	NO_WRITER,
    	OUTPORT,
    	MULTI_SYNC_READY,
    	MULTI_SYNC_READY_WITH_ME,
    	IS_SYNC_READY,
    	MAY_LEAVE,
    	MAY_JOIN,
    	IS_ALREADY,
    	BRANCH_MEMBER,
        MAX_MULTI_SYNC_SET,
    	MUTUAL_RECURSION,
    	INTEGERL + OFFSET,
    	PROGRAM_COUNTER,
    	RANDOM,
    	RANDOM_VAR,
    	SET_DIFF,
    	SET_INTER,
    	SET_UNION,
    	SLEEP,
    	SLEEP_TIME,
    	SYNC,
    	SYNC_IN,
    	SYNC_OUT,
    	SYNC_TYPE,
    	TRY_CONFIRM_READERS,
    	TRY_JOIN_SEMAPHOR,
    	TRY_LEAVE_SEMAPHOR,
    	TYPEOF,
        WITHDRAW_IN_COMM,
        WITHDRAW_IN_CONF,
        WITHDRAW_IN_FINISHED,
        WITHDRAW_SIGNALS
    };
    
    private Set<String> furtherToolKeywords;
    private Vector<String> furtherToolKeywordsV;

    private Hashtable<String,String> embeddedFunctions;
    private Hashtable<String,List<CspType>> embeddedFunctionTypes;
    
    private static HandelC instance = null;
    private HandelC() {
    	this.furtherToolKeywords = new HashSet<String>();
    	this.furtherToolKeywordsV = new Vector<String>();
    	this.initialiseEmbeddedFunctions();
    }
    
    private void initialiseEmbeddedFunctions() {
    	this.embeddedFunctions = new Hashtable<String, String>();
    	this.embeddedFunctionTypes = new Hashtable<String, List<CspType>>();
    	
    	// Set binary functions
    	CspType argLeftType = new CspSetType(new CspUnknownType());
    	CspType argRightType = new CspSetType(new CspUnknownType());
    	CspType argReturnType = new CspSetType(new CspUnknownType());
    	List<CspType> setTypesBinary = new Vector();
    	setTypesBinary.add(argLeftType);
    	setTypesBinary.add(argRightType);
    	setTypesBinary.add(argReturnType);
    	
    	this.includeEmbeddedFunction(HandelC.CSP_SET_UNION, HandelC.SET_UNION, setTypesBinary);
    	this.includeEmbeddedFunction(HandelC.CSP_SET_INTER, HandelC.SET_INTER, setTypesBinary);
    	this.includeEmbeddedFunction(HandelC.CSP_SET_DIFF, HandelC.SET_DIFF, setTypesBinary);
    	
    	// New embedded CSP functions must be reflected in:
    	// ( 1 ) Types inference visitor: MaximalTypesInferenceVisitor.caseACspFunctionCallExpr
    	// ( 2 ) Alphabets visitor: MaximalTypesInferenceVisitor.caseACspFunctionCallExpr
    }
    
    private void includeEmbeddedFunction(String cspName, String hcName, List<CspType> types) {
    	this.embeddedFunctions.put(cspName, hcName);
    	this.embeddedFunctionTypes.put(cspName, types);
    }
    

    public boolean isEmbeddedFunction(String name) {
    	return this.embeddedFunctions.containsKey(name);
    }

    public int getEmbeddedFunctionCard(String name) {
    	return this.embeddedFunctionTypes.get(name).size()-1;
    }
    
    public static HandelC getInstance() {
       if(instance == null) {
          instance = new HandelC();
       }
       return instance;
    }
    
    public static Set<String> getHCKeywords(){
    	
    	// Build set of keywords
    	Set<String> keywordsSet = new HashSet<String>();
    	for (int i = 0; i < hcKeywords.length; i++) {
    		keywordsSet.add(hcKeywords[i]);
    	}

    	return keywordsSet;
    }

    public void addToolKeyword(String word) {
    	HandelC.getInstance().furtherToolKeywords.add(word);
    	HandelC.getInstance().furtherToolKeywordsV.add(word);
    }
    
    public Set<String> getToolKeywords(){
    	
    	// Build set of keywords
    	Set<String> keywordsSet = new HashSet<String>();
		keywordsSet.add(this.BRANCH_ID);
    	for (int i = 0; i < toolKeywords.length; i++) {
    		keywordsSet.add(toolKeywords[i]);
			//System.out.println(toolKeywords[i]);
    	}
		keywordsSet.addAll(furtherToolKeywords);
    	
		/*
		System.out.println("---------------------------------------------------");
		System.out.println("HandelC :: getToolKeywords -> further tool keywords");
		System.out.println("---------------------------------------------------");
		for (int i=0; i < HandelC.getInstance().furtherToolKeywordsV.size();i++) {
			System.out.println(HandelC.getInstance().furtherToolKeywordsV.get(i));
		}
		*/
		
    	return keywordsSet;
    }
    
    public String generateComment(String comment) {
        int length = comment.length();
        String beautyLine = HandelC.LINE_COMMENT + HandelC.SPACE;
        for (int i=0; i<length; i++) {
            beautyLine = beautyLine + HandelC.MATH_DIFF;
        }
        String commentLine = HandelC.LINE_COMMENT + HandelC.SPACE + comment;
        
        String result = HandelC.COMMENT_LINE_BREAK + beautyLine 
                        + HandelC.COMMENT_LINE_BREAK + commentLine 
                        + HandelC.COMMENT_LINE_BREAK + beautyLine 
                        + HandelC.COMMENT_LINE_BREAK;
        
        return result;
    }

    public String generateDefine(String definition) {
    	String define = HandelC.DEFINE + HandelC.SPACE + definition + HandelC.COMMENT_LINE_BREAK;
        return define;
    }
    
    public String getSetVarOf(String var){
    	String name = var + SET_SUFFIX;
    	if(!var.equals("")) {
        	HandelC.getInstance().addToolKeyword(name);
    	}
    	return name;
    }

    public String getSetOf(String var){
    	String name = var + SET_SUFFIX;
    	return name;
    }
    
    public static String getEmptySetVarOf(String var){
    	String name = HandelC.getInstance().getSetVarOf(var) + EMPTY_SET_SUFFIX; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }
    
    public String getCardVarOf(String var){
    	String name = var + CARDINALITY; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }

    public String getSemaVarOf(String var){
    	String name = var + SEMAPHOR_SUFFIX; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }
    
    public String getBranchName(int i){
    	String name = HandelC.BRANCH + HandelC.UNDERSCORE + i; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }
    
    public String getChannelName(String c) {
    	String name = HandelC.CHAN_DECL + HandelC.UNDERSCORE + c; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }

    public String getInBusVarName(String c) {
    	String name = HandelC.IN_BUS_PREFIX + HandelC.UNDERSCORE + c; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }

    public String getOutBusVarName(String c) {
    	String name = HandelC.OUT_BUS_PREFIX + HandelC.UNDERSCORE + c; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }

    public String getBusArgName(String c) {
    	String name = HandelC.BUS_VALUE_PREFIX+HandelC.UNDERSCORE+c+HandelC.UNDERSCORE+HandelC.BUS_VALUE_SUFFIX; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }

    public String getInBusMethodName(String c) {
    	String name = HandelC.IN_BUS_METHOD_PREFIX + HandelC.UNDERSCORE + c; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }

    public String getOutBusMethodName(String c) {
    	String name = HandelC.OUT_BUS_METHOD_PREFIX + HandelC.UNDERSCORE + c; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }
    
    public String getBusPinName(String c, int i) {
    	String name = c+HandelC.UNDERSCORE+i; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }

    public String getChannel(String channel_name) {
    	return channel_name.substring((HandelC.CHAN_DECL + HandelC.UNDERSCORE).length());
    }
    
    public String getSetUnionCode(List<String> elements, String type) {
		String retorno = HandelC.getInstance().getEmptySetVarOf(type);
    	if (elements.size() > 0) {
    		String head = elements.get(0);
    		elements.remove(0);
			retorno = HandelC.SET_UNION + HandelC.PAR_BEGIN
					+ head + HandelC.SEP_ARGUMENTS + getSetUnionCode(elements, type) 
					+ HandelC.PAR_END;
		}
		return retorno;
    }
    
    // Returns the name of the look up table for a given tag in a given datatype
    public String getLookupTableName(String type, String tag) {
    	String name = type + HandelC.UNDERSCORE + tag + HandelC.UNDERSCORE + HandelC.LOOKUP_TABLE;
    	if(!tag.equals("")) {
    		this.addToolKeyword(name);
    	}
    	return name;
    }

    // Returns the name of the singleton sets look up table for a given type
    public String getSingletonSetsLookupTableName(String type) {
    	String name = this.getLookupTableName(type,HandelC.getInstance().getSetVarOf(""));
    	this.addToolKeyword(name);
    	return name;
    }
    
    
    // Returns a conditional block for a given list of conditions and values
    // with same cardinality using the given identation (identation spaces).
    private String getConditionalBlock(List conditions, List values, int identation) {
        String select = "";
        
        if (conditions.size() > 0 && conditions.size() == values.size()) {
            int newIdentation = identation
                                + HandelC.MACRO.length();
            select = select + StringUtil.getSpace(identation) + HandelC.PAR_BEGIN;
            select = select + ListUtil.head(conditions);
            select = select + HandelC.CONDITIONAL_COND;
            select = select + HandelC.SPACE + ListUtil.head(values);
            select = select + HandelC.CONDITIONAL_SEP;
            select = select + HandelC.COMMENT_LINE_BREAK;

            // Creating the if-then-else block
            if (conditions.size() > 1) {
                select = select + this.getConditionalBlock(ListUtil.tail(conditions), ListUtil.tail(values), newIdentation);
            } else {
                select = select + StringUtil.getSpace(newIdentation) + ListUtil.head(values);
            }
            select = select + HandelC.PAR_END;
        } else {
            throw new InvalidSelectBlockArgumentsException();
        }
        
        return select;
    }
    
    public String getDoneEventsVarName(int diff){
    	String name = HandelC.DONE_EVENTS + HandelC.UNDERSCORE + diff;
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }

    public String getHasJoinedVarName(int diff){
    	String name = HandelC.HAS_JOINED + HandelC.UNDERSCORE + diff;
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }

    public String getArgName(String function, String arg) {
    	String name = function.toUpperCase() + HandelC.UNDERSCORE + arg;
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }

    public String getActionFromDirective(String directive) {
    	String action = "";
    	if (directive.equals(Directives.CHANNEL_IN)) {
    		action = HandelC.A_READ;
    	} else if (directive.equals(Directives.CHANNEL_OUT)) {
    		action = HandelC.A_WRITE;
    	}
    	
    	return action;
    }

    public String getHCFunction(String cspFunction){
    	String retorno = "";
    	if (this.isEmbeddedFunction(cspFunction)) {
    		retorno = this.embeddedFunctions.get(cspFunction);
    	}

    	return retorno;
    }
    
    /*
    public String getHandelCTypeName(String typeName) {
    	String hcType = "";
    	if (typeName.equals(CspUtil.getBoolType())) {
            hcType = HandelC.BOOLEAN;
    	} else if (typeName.equals(CspUtil.getIntType())) {
            hcType = HandelC.INTEGERL;
    	} else {
            hcType = typeName;
    	}
    	return hcType;
    }
    */

    public String getHandelCTypeName(String typeName) {
    	
    	String hcType = "";

    	if (this.isSetType(typeName)) {
    		hcType = this.getSetType(typeName);
    	} else {
    		hcType = typeName;
    	}
    	
    	if (hcType.equals(CspUtil.getBoolType())) {
            hcType = HandelC.BOOLEAN;
    	} else if (typeName.equals(CspUtil.getIntType())) {
            hcType = HandelC.INTEGERL;
    	}

    	if (this.isSetType(typeName)) {
    		hcType = HandelC.getInstance().getSetVarOf(hcType);
    	}
    	
    	return hcType;
    }

    public boolean isSetType(String typeName) {
    	return typeName.startsWith(HandelC.SET_TYPE_PREFIX);
    }
    
    public String getSetType(String typeName) {
    	if (this.isSetType(typeName)) {
    		int beginIndex = HandelC.SET_TYPE_PREFIX.length();
    		int endIndex = typeName.length()-1;
    		typeName = typeName.substring(beginIndex, endIndex);
    	}
    	return typeName;
    }

    public boolean isTypeOf(String typeName) {
    	return typeName.startsWith(HandelC.TYPEOF_PREFIX);
    }
    
    public String getTypeOfElement(String typeName) {
    	if (this.isTypeOf(typeName)) {
    		int beginIndex = HandelC.TYPEOF_PREFIX.length();
    		int endIndex = typeName.length()-1;
    		typeName = typeName.substring(beginIndex, endIndex);
    	}
    	return typeName;
    }

    // Gambiarra para solucionar problema do typechecker
    // quando ele deixar o tipo como typeof(x)
    public boolean isDeclarationLink(String typeName) {
    	return typeName.startsWith(HandelC.TYPEOF_PREFIX) && typeName.contains("=");
    }
    public String getTypeFromDeclarationLink(String typeName) {
    	if (this.isDeclarationLink(typeName)) {
    		int beginIndex = typeName.indexOf("=")+1;
    		int endIndex = typeName.length();
    		typeName = typeName.substring(beginIndex, endIndex).trim();
    	}
    	System.out.println("handelC :: getTypeFromDeclarationLink -> Type #"+typeName+"#");
    	return typeName;
    }

    
}
