package csp2hc.util;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Map;

import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.lexer.LexerException;
import lmf.formula.csp.node.ACspAbstractDefinition;
import lmf.formula.csp.node.ACspAlphabetisedParallelProcess;
import lmf.formula.csp.node.ACspInterleaveProcess;
import lmf.formula.csp.node.ACspPrefixProcess;
import lmf.formula.csp.node.ACspProcessDefinition;
import lmf.formula.csp.node.ACspRenamingProcess;
import lmf.formula.csp.node.ACspSharingParallelProcess;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.parser.ParserException;
import csp2hc.exception.UnsupportedConstructorException;
import csp2hc.messages.HandelC;
import csp2hc.messages.Messenger;
import csp2hc.visitors.ChannelsVisitor;
import csp2hc.visitors.CodeInsertionVisitor;
import csp2hc.visitors.ConstantsAndFunctionsVisitor;
import csp2hc.visitors.DatatypesVisitor;
import csp2hc.visitors.DirectivesVisitor;
import csp2hc.visitors.ForcedInterleavingVisitor;
import csp2hc.visitors.ParallelismAnalyserVisitor;
import csp2hc.visitors.ProcessesInfoVisitor;
import csp2hc.visitors.RenamingVisitor;
import csp2hc.visitors.SubstitutionVisitor;

public class BetterPreProcessing {
	
	private ForcedInterleavingVisitor forcedInterleaving;
	private ChannelsVisitor channels;
	private SubstitutionVisitor substitutions;

	public String execute(File file) throws FileNotFoundException, IOException, ParserException, LexerException {
	
		Node start = ParserUtil.parser(file);
		
		//////////////////////
		// LOAD DIRECTIVE INFO
		////////////////////// 
		DirectivesVisitor directives = new DirectivesVisitor();
		start.apply(directives);
	
		///////////////////////////////
		// LOAD CONSTANTS AND FUNCTIONS
		///////////////////////////////
		ConstantsAndFunctionsVisitor constants = new ConstantsAndFunctionsVisitor(directives);
		start.apply(constants);
		
		/////////////////
		// LOAD DATATYPES
		/////////////////
		DatatypesVisitor types = new DatatypesVisitor(directives,constants);
		start.apply(types);
		
		String mainProc = HandelC.MAINP + HandelC.ASSIGNMENT + directives.getMain();
		start = ParserUtil.parser(file, mainProc);
		
		////////////////////////////
		// LOAD CHANNELS INFORMATION
		////////////////////////////
		channels = new ChannelsVisitor(directives, types, constants);
		start.apply(channels);
		
		///////////////////////////////////////////
		// LOAD PROCESSES ARGUMENTS AND DEFINITIONS
		///////////////////////////////////////////
		ProcessesInfoVisitor procInfo = new ProcessesInfoVisitor(directives, types, channels);
		start.apply(procInfo);
		
		////////////////////////////////////
		// ANALYSE THE PARALLEL COMBINATIONS
		////////////////////////////////////
		ParallelismAnalyserVisitor parallelismAnalyser = 
			new ParallelismAnalyserVisitor(procInfo, directives, constants);
		start.apply(parallelismAnalyser);
		
		substitutions = new SubstitutionVisitor(directives, channels, procInfo);
		start.apply(substitutions);
		
		forcedInterleaving = new ForcedInterleavingVisitor(directives, channels, procInfo, parallelismAnalyser, substitutions);
		start.apply(forcedInterleaving);
		
		return forcedInterleaving.getCspCode();
	
	}

	public ForcedInterleavingVisitor getForcedInterleaving() {
		return forcedInterleaving;
	}
	
	public SubstitutionVisitor getSubstitutions() {
		return substitutions;
	}

}
