package csp2hc.util;

import java.io.File;
import java.io.IOException;
import java.util.Hashtable;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;
import java.util.Vector;

import lmf.formula.csp.lexer.LexerException;
import lmf.formula.csp.parser.ParserException;
import csp2hc.messages.HandelC;
import csp2hc.visitors.ChannelsVisitor;
import csp2hc.visitors.CodeInsertionVisitor;
import csp2hc.visitors.ConstantsAndFunctionsVisitor;
import csp2hc.visitors.DatatypesVisitor;
import csp2hc.visitors.DirectivesVisitor;
import csp2hc.visitors.ParallelismAnalyserVisitor;
import csp2hc.visitors.ProcessesInfoVisitor;

public class ForcedInterleaveInsertion {
	private Hashtable <String, List<String>> renamingProcesses;
	private Hashtable<String, List<String>> renamedChannels;
	private DirectivesVisitor directives;
	private DatatypesVisitor types;
	private ConstantsAndFunctionsVisitor constants;
	private ChannelsVisitor channels;
	private ProcessesInfoVisitor procInfo;
	private ParallelismAnalyserVisitor parallelismAnalyser;
	private File file;
	
	private String cspCode;
	
	public List <String> getRenamedChannels(String channelName){
		List <String> renamed;
		if (renamedChannels.containsKey(channelName)){
			renamed = renamedChannels.get(channelName);
		} else {
			renamed = new Vector<String>();
		}
		return renamed;
	}
	
	public DirectivesVisitor getDirectives() {
		return directives;
	}
	
	public ChannelsVisitor getChannels() {
		return channels;
	}

	public ParallelismAnalyserVisitor getParallelismAnalyser() {
		return parallelismAnalyser;
	}

	public ProcessesInfoVisitor getProcInfo() {
		return procInfo;
	}

	public ForcedInterleaveInsertion (DirectivesVisitor directives, ChannelsVisitor channels, ProcessesInfoVisitor procInfo, ParallelismAnalyserVisitor parallelismAnalyser) throws IOException, ParserException, LexerException{
		this.directives = directives;
		this.channels = channels;
		this.procInfo = procInfo;
		this.parallelismAnalyser = parallelismAnalyser;
		this.cspCode = "";
	}
	
	public boolean isProcessToBeRenamed (String name) {
		return this.renamingProcesses.keySet().contains(name);
	}
	
	public void preRenaming() {
		
		Hashtable <String,Set<Set<Integer>>> realSyncLUT = parallelismAnalyser.getRealSyncLUT();
		
		ForcedInterleavingOnChannel forcedInterleavingLUT = parallelismAnalyser.getForcedInterleavingLUT();
		
		Set <String> renamingChannels = forcedInterleavingLUT.keySet();
		
		this.renamingProcesses = new Hashtable <String, List<String>>();
		Set <String> processes = channels.getProcessChannels().keySet();
		for (String process : processes){
			List<String> channelsOfProcess = channels.getProcessChannels().get(process);
			ListIterator<String> channelsIter = channelsOfProcess.listIterator();
			while (channelsIter.hasNext()) {
				String next = channelsIter.next();
				if (renamingChannels.contains(HandelC.CHAN_DECL + HandelC.UNDERSCORE + next)){
					List <String> chanList;
					if (!renamingProcesses.containsKey(process)){
						chanList = new Vector <String>();
					} else {
						chanList = renamingProcesses.get(process);
					}
					chanList.add(next);
					renamingProcesses.put(process, chanList);
				}
			}
		}
		/*for (String process : this.renamingProcesses.keySet()) {
			this.procInfo.processesDefinitions.
		}*/
		
		System.out.println("\nRenamingProcess " + renamingProcesses);
		
		String rnProcess = "";
		String channelDeclaration = "";
		
		renamedChannels = new Hashtable <String, List<String>>();
		
		for (String process : this.renamingProcesses.keySet()){
			List <Integer> branches = this.parallelismAnalyser.getBranchesOfProcess(process);
			if (branches != null) {
				for (int branch : branches){
					String rnSubstitutionList = "";
					
					Hashtable<String, List<String>> substitutions = new Hashtable<String, List<String>>();
					
					ListIterator<String> channelsList = this.renamingProcesses.get(process).listIterator();
					while (channelsList.hasNext()){
						String nextChannel = channelsList.next();
						if (!realSyncLUT.containsKey(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel)){
							if (!rnSubstitutionList.equals("")){
								rnSubstitutionList = rnSubstitutionList + " , ";
							}
							rnSubstitutionList = rnSubstitutionList + nextChannel + " <- " + nextChannel + "_" + branch;
							
							if (substitutions.containsKey(nextChannel)){
								List<String> list = substitutions.get(nextChannel);
								list.add(nextChannel + "_" + branch);
								substitutions.put(nextChannel, list);
							} else {
								List<String> list = new Vector<String>();
								list.add(nextChannel + "_" + branch);
								substitutions.put(nextChannel, list);
							}
							
							String channelType = "";
							if (this.channels.containsChannelType(nextChannel)){
								List types = channels.getChannelType(nextChannel);
								if (!types.isEmpty()){
									channelType = channelType + " : ";
									for (Object type:types){
										if (!channelType.equals(" : ")){
											channelType = channelType + " . ";
										}
										channelType = channelType + type.toString();
									}
								}
							}
							if (!channelDeclaration.contains(nextChannel+ "_" + branch)){
								if (!renamedChannels.containsKey(nextChannel)){
									List <String> rem = new Vector<String>();
									rem.add(nextChannel + "_" + branch);
									renamedChannels.put(nextChannel, rem);
								} else{
									List <String> rem = renamedChannels.get(nextChannel);
									if (!rem.contains(nextChannel + "_" + branch)){
										rem.add(nextChannel + "_" + branch);
										renamedChannels.put(nextChannel, rem);
									}
								}
								
								channelDeclaration = channelDeclaration + "\nchannel " + nextChannel + "_" + branch + channelType;
							}
						} else {
							boolean branchIsInFriends = false;
							for (Set<Integer> friends:realSyncLUT.get(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel)){
								branchIsInFriends = branchIsInFriends || friends.contains(branch);
							}
							if (!branchIsInFriends){
								if (!rnSubstitutionList.equals("")){
									rnSubstitutionList = rnSubstitutionList + " , ";
								}
								rnSubstitutionList = rnSubstitutionList + nextChannel + " <- " + nextChannel + "_" + branch;
								
								if (substitutions.containsKey(nextChannel)){
									List<String> list = substitutions.get(nextChannel);
									list.add(nextChannel + "_" + branch);
									substitutions.put(nextChannel, list);
								} else {
									List<String> list = new Vector<String>();
									list.add(nextChannel + "_" + branch);
									substitutions.put(nextChannel, list);
								}
								
								String channelType = "";
								if (this.channels.containsChannelType(nextChannel)){
									List types = channels.getChannelType(nextChannel);
									if (!types.isEmpty()){
										channelType = channelType + " : ";
										for (Object type:types){
											if (!channelType.equals(" : ")){
												channelType = channelType + " . ";
											}
											channelType = channelType + type.toString();
										}
									}
								}
								if (!channelDeclaration.contains(nextChannel+ "_" + branch)){
									if (!renamedChannels.containsKey(nextChannel)){
										List <String> rem = new Vector<String>();
										rem.add(nextChannel + "_" + branch);
										renamedChannels.put(nextChannel, rem);
									} else{
										List <String> rem = renamedChannels.get(nextChannel);
										if (!rem.contains(nextChannel + "_" + branch)){
											rem.add(nextChannel + "_" + branch);
											renamedChannels.put(nextChannel, rem);
										}
									}
									channelDeclaration = channelDeclaration + "\nchannel " + nextChannel + "_" + branch + channelType;
								}
							} else { 
								Set <Set<Integer>> setOfFriends = realSyncLUT.get(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel);
								for (Set<Integer> friends: setOfFriends){
									if (friends.contains(branch)){
										if (!rnSubstitutionList.equals("")){
											rnSubstitutionList = rnSubstitutionList + " , ";
										}
										rnSubstitutionList = rnSubstitutionList + nextChannel + " <- " + nextChannel + "_" + friends.hashCode();
										
										if (substitutions.containsKey(nextChannel)){
											List<String> list = substitutions.get(nextChannel);
											list.add(nextChannel + "_" + friends.hashCode());
											substitutions.put(nextChannel, list);
										} else {
											List<String> list = new Vector<String>();
											list.add(nextChannel + "_" + friends.hashCode());
											substitutions.put(nextChannel, list);
										}
										
										
										String channelType = "";
										if (this.channels.containsChannelType(nextChannel)){
											List types = channels.getChannelType(nextChannel);
											if (!types.isEmpty()){
												channelType = channelType + " : ";
												for (Object type:types){
													if (!channelType.equals(" : ")){
														channelType = channelType + " . ";
													}
													channelType = channelType + type.toString();
												}
											}
										}
										if (!channelDeclaration.contains(nextChannel+ "_" + friends.hashCode())){
											if (!renamedChannels.containsKey(nextChannel)){
												List <String> rem = new Vector<String>();
												rem.add(nextChannel + "_" + friends.hashCode());
												renamedChannels.put(nextChannel, rem);
											} else{
												List <String> rem = renamedChannels.get(nextChannel);
												if (!rem.contains(nextChannel + "_" + friends.hashCode())){
													rem.add(nextChannel + "_" + friends.hashCode());
													renamedChannels.put(nextChannel, rem);
												}
											}
											channelDeclaration = channelDeclaration + "\nchannel " + nextChannel + "_" + friends.hashCode() + channelType;
										}
									}
								}
							}
						}
					}
					String arguments = "";
					if (this.procInfo.containsProcArguments(process)){
						arguments = this.procInfo.getProcArguments(process).toString().trim();
					}
					String directives = "";
					if (this.directives.containsProcArguments(process)){
						Hashtable args = this.directives.getProcArguments(process);
						for(Object arg: args.keySet()) {
							directives = "--!! arg " + arg + " " + args.get(arg) + " within " + process + "_RN" + branch + "\n" + directives;
						}
					}
					
					if (this.directives.containsProcChannelInOut(process)){
						Hashtable<String,String> chans = this.directives.getProcChannelInOut(process);
						for (String chan : chans.keySet()){
							directives = "--!! channel " + chan + " " + (HandelC.A_READ.equals(chans.get(chan))?"in":"out") 
													+ " within " +  process + "_RN" + branch + "\n" + directives;
						}
						for (String chan : substitutions.keySet()){
							if (chans.containsKey(chan)){
								for (String subs : substitutions.get(chan)){
									this.directives.putProcChannelInOut(process + "_RN" + branch, subs, chans.get(chan));
									directives = "--!! channel " + subs + " " + (HandelC.A_READ.equals(chans.get(chan))?"in":"out") 
											+ " within " +  process + "_RN" + branch + "\n" + directives;
								}
							}
						}
					}
					
					rnProcess = rnProcess + "\n" + "-- Processo " + process + " renomeado\n" + directives + process + "_RN" + branch + arguments + " = " + process + arguments + "[[" + rnSubstitutionList + "]]\n";
					
				}
			}
		}
		
		//this.cspCode = this.codeInsertion.getCspCode() + channelDeclaration + rnProcess ;
		this.cspCode = channelDeclaration + rnProcess ;
		
	}
	
	public void defineRenaming() {

		Hashtable <String,Set<Set<Integer>>> realSyncLUT = parallelismAnalyser.getRealSyncLUT();
		
		ForcedInterleavingOnChannel forcedInterleavingLUT = parallelismAnalyser.getForcedInterleavingLUT();
		
		Set <String> renamingChannels = forcedInterleavingLUT.keySet();
		
		this.renamingProcesses = new Hashtable <String, List<String>>();
		Set <String> processes = channels.getProcessChannels().keySet();
		for (String process : processes){
			List<String> channelsOfProcess = channels.getProcessChannels().get(process);
			ListIterator<String> channelsIter = channelsOfProcess.listIterator();
			while (channelsIter.hasNext()) {
				String next = channelsIter.next();
				if (renamingChannels.contains(HandelC.CHAN_DECL + HandelC.UNDERSCORE + next)){
					List <String> chanList;
					if (!renamingProcesses.containsKey(process)){
						chanList = new Vector <String>();
					} else {
						chanList = renamingProcesses.get(process);
					}
					chanList.add(next);
					renamingProcesses.put(process, chanList);
				}
			}
		}
		System.out.println("\nRenamingProcess " + renamingProcesses);
		
		String rnProcess = "";
		String channelDeclaration = "";
		
		renamedChannels = new Hashtable <String, List<String>>();
		
		for (String process : this.renamingProcesses.keySet()){
			List <Integer> branches = this.parallelismAnalyser.getBranchesOfProcess(process);
			if (branches != null) {
				for (int branch : branches){
					String rnSubstitutionList = "";
					ListIterator<String> channelsList = this.renamingProcesses.get(process).listIterator();
					while (channelsList.hasNext()){
						String nextChannel = channelsList.next();
						if (!realSyncLUT.containsKey(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel)){
							if (!rnSubstitutionList.equals("")){
								rnSubstitutionList = rnSubstitutionList + " , ";
							}
							rnSubstitutionList = rnSubstitutionList + nextChannel + " <- " + nextChannel + "_" + branch;
							
							String channelType = "";
							if (this.channels.containsChannelType(nextChannel)){
								List types = channels.getChannelType(nextChannel);
								if (!types.isEmpty()){
									channelType = channelType + " : ";
									for (Object type:types){
										if (!channelType.equals(" : ")){
											channelType = channelType + " . ";
										}
										channelType = channelType + type.toString();
									}
								}
							}
							if (!channelDeclaration.contains(nextChannel+ "_" + branch)){
								if (!renamedChannels.containsKey(nextChannel)){
									List <String> rem = new Vector<String>();
									rem.add(nextChannel + "_" + branch);
									renamedChannels.put(nextChannel, rem);
								} else{
									List <String> rem = renamedChannels.get(nextChannel);
									if (!rem.contains(nextChannel + "_" + branch)){
										rem.add(nextChannel + "_" + branch);
										renamedChannels.put(nextChannel, rem);
									}
								}
								
								channelDeclaration = channelDeclaration + "\nchannel " + nextChannel + "_" + branch + channelType;
							}
						} else {
							boolean branchIsInFriends = false;
							for (Set<Integer> friends:realSyncLUT.get(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel)){
								branchIsInFriends = branchIsInFriends || friends.contains(branch);
							}
							if (!branchIsInFriends){
								if (!rnSubstitutionList.equals("")){
									rnSubstitutionList = rnSubstitutionList + " , ";
								}
								rnSubstitutionList = rnSubstitutionList + nextChannel + " <- " + nextChannel + "_" + branch;
								
								String channelType = "";
								if (this.channels.containsChannelType(nextChannel)){
									List types = channels.getChannelType(nextChannel);
									if (!types.isEmpty()){
										channelType = channelType + " : ";
										for (Object type:types){
											if (!channelType.equals(" : ")){
												channelType = channelType + " . ";
											}
											channelType = channelType + type.toString();
										}
									}
								}
								if (!channelDeclaration.contains(nextChannel+ "_" + branch)){
									if (!renamedChannels.containsKey(nextChannel)){
										List <String> rem = new Vector<String>();
										rem.add(nextChannel + "_" + branch);
										renamedChannels.put(nextChannel, rem);
									} else{
										List <String> rem = renamedChannels.get(nextChannel);
										if (!rem.contains(nextChannel + "_" + branch)){
											rem.add(nextChannel + "_" + branch);
											renamedChannels.put(nextChannel, rem);
										}
									}
									channelDeclaration = channelDeclaration + "\nchannel " + nextChannel + "_" + branch + channelType;
								}
							} else { 
								Set <Set<Integer>> setOfFriends = realSyncLUT.get(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel);
								for (Set<Integer> friends: setOfFriends){
									if (friends.contains(branch)){
										if (!rnSubstitutionList.equals("")){
											rnSubstitutionList = rnSubstitutionList + " , ";
										}
										rnSubstitutionList = rnSubstitutionList + nextChannel + " <- " + nextChannel + "_" + friends.hashCode();
										
										String channelType = "";
										if (this.channels.containsChannelType(nextChannel)){
											List types = channels.getChannelType(nextChannel);
											if (!types.isEmpty()){
												channelType = channelType + " : ";
												for (Object type:types){
													if (!channelType.equals(" : ")){
														channelType = channelType + " . ";
													}
													channelType = channelType + type.toString();
												}
											}
										}
										if (!channelDeclaration.contains(nextChannel+ "_" + friends.hashCode())){
											if (!renamedChannels.containsKey(nextChannel)){
												List <String> rem = new Vector<String>();
												rem.add(nextChannel + "_" + friends.hashCode());
												renamedChannels.put(nextChannel, rem);
											} else{
												List <String> rem = renamedChannels.get(nextChannel);
												if (!rem.contains(nextChannel + "_" + friends.hashCode())){
													rem.add(nextChannel + "_" + friends.hashCode());
													renamedChannels.put(nextChannel, rem);
												}
											}
											channelDeclaration = channelDeclaration + "\nchannel " + nextChannel + "_" + friends.hashCode() + channelType;
										}
									}
								}
							}
						}
					}
					String arguments = "";
					if (this.procInfo.containsProcArguments(process)){
						arguments = this.procInfo.getProcArguments(process).toString().trim();
					}
					
					rnProcess = rnProcess + "\n" + "-- Processo " + process + " renomeado\n" + process + "_RN" + branch + arguments + " = " + process + arguments + "[[" + rnSubstitutionList + "]]\n";
					
				}
			}
		}
		
		this.cspCode = 
				//this.codeInsertion.getCspCode() + 
				channelDeclaration + rnProcess ;
	}

	public String getCspRenamingCode() {
		return this.cspCode;
	}
}
