package csp2hc.util;

import java.util.Hashtable;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;
import java.util.Vector;

import csp2hc.messages.HandelC;
import csp2hc.visitors.ChannelsVisitor;
import csp2hc.visitors.DirectivesVisitor;
import csp2hc.visitors.ParallelismAnalyserVisitor;
import csp2hc.visitors.ProcessesInfoVisitor;

/**
 * @author Ivan Soares
 *
 */
public class ForcedInterleaveUtil {
	
	private DirectivesVisitor directives;
	private ChannelsVisitor channels;
	private ProcessesInfoVisitor procInfo;
	private ParallelismAnalyserVisitor parallelismAnalyser;
	
	private Hashtable<String, List<String>> renamedProcesses;
	private Hashtable<String, List<String>> renamedChannels;
	private String cspCode;

	/**
	 * @param directives
	 * @param channels
	 * @param procInfo
	 * @param parallelismAnalyser
	 */
	public ForcedInterleaveUtil(DirectivesVisitor directives,
			ChannelsVisitor channels, ProcessesInfoVisitor procInfo,
			ParallelismAnalyserVisitor parallelismAnalyser) {
		this.directives = directives;
		this.channels = channels;
		this.procInfo = procInfo;
		this.parallelismAnalyser = parallelismAnalyser;
	}
	
	public void defineRenamedProcesses(){
		
		ForcedInterleavingOnChannel forcedInterleavingLUT = parallelismAnalyser
				.getForcedInterleavingLUT();

		Set<String> renamingChannels = forcedInterleavingLUT.keySet();

		this.renamedProcesses = new Hashtable<String, List<String>>();
		Set<String> processes = channels.getProcessChannels().keySet();
		for (String process : processes) {
			List<String> channelsOfProcess = channels.getProcessChannels().get(
					process);
			ListIterator<String> channelsIter = channelsOfProcess
					.listIterator();
			while (channelsIter.hasNext()) {
				String next = channelsIter.next();
				if (renamingChannels.contains(HandelC.CHAN_DECL
						+ HandelC.UNDERSCORE + next)) {
					List<String> chanList;
					if (!this.renamedProcesses.containsKey(process)) {
						chanList = new Vector<String>();
					} else {
						chanList = this.renamedProcesses.get(process);
					}
					chanList.add(next);
					this.renamedProcesses.put(process, chanList);
				}
			}
		}
		
		//System.out.println("\nRenamingProcess " + this.renamedProcesses);
	}

	public void defineRenamedChannels(){
		
	}
	
	@SuppressWarnings({ "rawtypes", "unchecked" })
	public void buildForcedInterleave(){
		
		defineRenamedProcesses();
		
		Hashtable<String, Set<Set<Integer>>> realSyncLUT = parallelismAnalyser
				.getRealSyncLUT();

		String rnProcess = "";
		String channelDeclaration = "";
		
		this.renamedChannels = new Hashtable <String, List<String>>();
		
		for (String process : this.renamedProcesses.keySet()){
			List <Integer> branches = this.parallelismAnalyser.getBranchesOfProcess(process);
			if (branches != null) {
				for (int branch : branches){
					String rnSubstitutionList = "";
					
					Hashtable<String, List<String>> substitutions = new Hashtable<String, List<String>>();
					
					ListIterator<String> channelsList = this.renamedProcesses.get(process).listIterator();
					while (channelsList.hasNext()){
						String nextChannel = channelsList.next();
						if (!realSyncLUT.containsKey(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel)){
							if (!rnSubstitutionList.equals("")){
								rnSubstitutionList = rnSubstitutionList + " , ";
							}
							rnSubstitutionList = rnSubstitutionList + nextChannel + " <- " + nextChannel + "_" + branch;
							
							if (substitutions.containsKey(nextChannel)){
								List<String> list = substitutions.get(nextChannel);
								list.add(nextChannel + "_" + branch);
								substitutions.put(nextChannel, list);
							} else {
								List<String> list = new Vector<String>();
								list.add(nextChannel + "_" + branch);
								substitutions.put(nextChannel, list);
							}
							
							String channelType = "";
							if (this.channels.containsChannelType(nextChannel)){
								List types = channels.getChannelType(nextChannel);
								if (!types.isEmpty()){
									channelType = channelType + " : ";
									for (Object type:types){
										if (!channelType.equals(" : ")){
											channelType = channelType + " . ";
										}
										channelType = channelType + type.toString();
									}
								}
							}
							if (!channelDeclaration.contains(nextChannel+ "_" + branch)){
								if (!this.renamedChannels.containsKey(nextChannel)){
									List <String> rem = new Vector<String>();
									rem.add(nextChannel + "_" + branch);
									this.renamedChannels.put(nextChannel, rem);
								} else{
									List <String> rem = this.renamedChannels.get(nextChannel);
									if (!rem.contains(nextChannel + "_" + branch)){
										rem.add(nextChannel + "_" + branch);
										this.renamedChannels.put(nextChannel, rem);
									}
								}
								
								channelDeclaration = channelDeclaration + "\nchannel " + nextChannel + "_" + branch + channelType;
							}
						} else {
							boolean branchIsInFriends = false;
							for (Set<Integer> friends:realSyncLUT.get(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel)){
								branchIsInFriends = branchIsInFriends || friends.contains(branch);
							}
							if (!branchIsInFriends){
								if (!rnSubstitutionList.equals("")){
									rnSubstitutionList = rnSubstitutionList + " , ";
								}
								rnSubstitutionList = rnSubstitutionList + nextChannel + " <- " + nextChannel + "_" + branch;
								
								if (substitutions.containsKey(nextChannel)){
									List<String> list = substitutions.get(nextChannel);
									list.add(nextChannel + "_" + branch);
									substitutions.put(nextChannel, list);
								} else {
									List<String> list = new Vector<String>();
									list.add(nextChannel + "_" + branch);
									substitutions.put(nextChannel, list);
								}
								
								String channelType = "";
								if (this.channels.containsChannelType(nextChannel)){
									List types = channels.getChannelType(nextChannel);
									if (!types.isEmpty()){
										channelType = channelType + " : ";
										for (Object type:types){
											if (!channelType.equals(" : ")){
												channelType = channelType + " . ";
											}
											channelType = channelType + type.toString();
										}
									}
								}
								if (!channelDeclaration.contains(nextChannel+ "_" + branch)){
									if (!this.renamedChannels.containsKey(nextChannel)){
										List <String> rem = new Vector<String>();
										rem.add(nextChannel + "_" + branch);
										this.renamedChannels.put(nextChannel, rem);
									} else{
										List <String> rem = this.renamedChannels.get(nextChannel);
										if (!rem.contains(nextChannel + "_" + branch)){
											rem.add(nextChannel + "_" + branch);
											this.renamedChannels.put(nextChannel, rem);
										}
									}
									channelDeclaration = channelDeclaration + "\nchannel " + nextChannel + "_" + branch + channelType;
								}
							} else { 
								Set <Set<Integer>> setOfFriends = realSyncLUT.get(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel);
								for (Set<Integer> friends: setOfFriends){
									if (friends.contains(branch)){
										if (!rnSubstitutionList.equals("")){
											rnSubstitutionList = rnSubstitutionList + " , ";
										}
										rnSubstitutionList = rnSubstitutionList + nextChannel + " <- " + nextChannel + "_" + friends.hashCode();
										
										if (substitutions.containsKey(nextChannel)){
											List<String> list = substitutions.get(nextChannel);
											list.add(nextChannel + "_" + friends.hashCode());
											substitutions.put(nextChannel, list);
										} else {
											List<String> list = new Vector<String>();
											list.add(nextChannel + "_" + friends.hashCode());
											substitutions.put(nextChannel, list);
										}
										
										
										String channelType = "";
										if (this.channels.containsChannelType(nextChannel)){
											List types = channels.getChannelType(nextChannel);
											if (!types.isEmpty()){
												channelType = channelType + " : ";
												for (Object type:types){
													if (!channelType.equals(" : ")){
														channelType = channelType + " . ";
													}
													channelType = channelType + type.toString();
												}
											}
										}
										if (!channelDeclaration.contains(nextChannel+ "_" + friends.hashCode())){
											if (!this.renamedChannels.containsKey(nextChannel)){
												List <String> rem = new Vector<String>();
												rem.add(nextChannel + "_" + friends.hashCode());
												this.renamedChannels.put(nextChannel, rem);
											} else{
												List <String> rem = this.renamedChannels.get(nextChannel);
												if (!rem.contains(nextChannel + "_" + friends.hashCode())){
													rem.add(nextChannel + "_" + friends.hashCode());
													this.renamedChannels.put(nextChannel, rem);
												}
											}
											channelDeclaration = channelDeclaration + "\nchannel " + nextChannel + "_" + friends.hashCode() + channelType;
										}
									}
								}
							}
						}
					}
					String arguments = "";
					if (this.procInfo.containsProcArguments(process)){
						arguments = this.procInfo.getProcArguments(process).toString().trim();
					}
					
					if (this.directives.containsProcChannelInOut(process)){
						Hashtable<String,String> chans = this.directives.getProcChannelInOut(process);
						for (String chan : substitutions.keySet()){
							if (chans.containsKey(chan)){
								for (String subs : substitutions.get(chan)){
									this.directives.putProcChannelInOut(process + "_RN" + branch, subs, chans.get(chan));
								}
							}
						}
					}
					
					rnProcess = rnProcess + "\n" + "-- Processo " + process + " renomeado\n" + process + "_RN" + branch + arguments + " = " + process + arguments + "[[" + rnSubstitutionList + "]]\n";
					
				}
			}
		}
		
		setCspCode(channelDeclaration + rnProcess);
		
	}

	public String getCspCode() {
		return cspCode;
	}

	public void setCspCode(String cspCode) {
		this.cspCode = cspCode;
	}
	
}
