package csp2hc.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PushbackReader;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;
import java.util.Vector;

import com.sun.org.apache.xerces.internal.impl.xs.SubstitutionGroupHandler;

import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.lexer.Lexer;
import lmf.formula.csp.lexer.LexerException;
import lmf.formula.csp.node.ACspAbstractDefinition;
import lmf.formula.csp.node.ACspParameters;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.Start;
import lmf.formula.csp.parser.Parser;
import lmf.formula.csp.parser.ParserException;
import lmf.formula.csp.semantic.CspAnalyserListener;
import lmf.formula.csp.semantic.CspContextualAnalyser;
import lmf.formula.csp.semantic.exception.CspAnalyserException;
import csp2hc.exception.AbstractDefinitionException;
import csp2hc.messages.HandelC;
import csp2hc.visitors.ChannelsVisitor;
import csp2hc.visitors.CodeInsertionVisitor;
import csp2hc.visitors.ConstantsAndFunctionsVisitor;
import csp2hc.visitors.DatatypesVisitor;
import csp2hc.visitors.DirectivesVisitor;
import csp2hc.visitors.ParallelismAnalyserVisitor;
import csp2hc.visitors.ProcessesInfoVisitor;

public class ForcedInterleavingProtocol// extends ExtendedDepthFirstAdapter// implements CspAnalyserListener 
	{
	private Hashtable <String, List<String>> renamingProcesses;
	private Hashtable<String, List<String>> renamedChannels;
	private DirectivesVisitor directives;
	private DatatypesVisitor types;
	private ConstantsAndFunctionsVisitor constants;
	private CodeInsertionVisitor codeInsertion;
	private ChannelsVisitor channels;
	private ProcessesInfoVisitor procInfo;
	private ParallelismAnalyserVisitor parallelismAnalyser;
	private File file;
	
	private String cspCode;
	
	public List <String> getRenamedChannels(String channelName){
		List <String> renamed;
		if (renamedChannels.containsKey(channelName)){
			renamed = renamedChannels.get(channelName);
		} else {
			renamed = new Vector<String>();
		}
		return renamed;
	}
	
	public DirectivesVisitor getDirectives() {
		return directives;
	}
	
	public ChannelsVisitor getChannels() {
		return channels;
	}

	public ParallelismAnalyserVisitor getParallelismAnalyser() {
		return parallelismAnalyser;
	}

	public ProcessesInfoVisitor getProcInfo() {
		return procInfo;
	}

	public ForcedInterleavingProtocol (File file, DirectivesVisitor directives, DatatypesVisitor types, ConstantsAndFunctionsVisitor constants, CodeInsertionVisitor codeInsertion) throws IOException, ParserException, LexerException{
		this.file = file;
		this.directives = directives;
		this.types = types;
		this.constants = constants;
		this.codeInsertion = codeInsertion;
		
		this.cspCode = "";
		
		String mainProc = HandelC.MAINP + HandelC.ASSIGNMENT + directives.getMain();
		Node newStart = ParserUtil.parser(file, this.codeInsertion.getCspCode() + mainProc);

		//TODO Remove
		newStart.apply(
				new ExtendedDepthFirstAdapter() {
		        	@Override
		        	public void caseACspAbstractDefinition(ACspAbstractDefinition node) {
		            	//System.out.println("caseACspAbstractDefinition -> node: "+node);
		        		throw new AbstractDefinitionException();
		            }
		        });

		////////////////////////////
		// LOAD CHANNELS INFORMATION
		////////////////////////////
		this.channels = new ChannelsVisitor(directives, types, constants);
		newStart.apply(this.channels);
		
		///////////////////////////////////////////
		// LOAD PROCESSES ARGUMENTS AND DEFINITIONS
		///////////////////////////////////////////
		this.procInfo = new ProcessesInfoVisitor(directives, types, channels, codeInsertion);
		newStart.apply(this.procInfo);
		
		////////////////////////////////////
		// ANALYSE THE PARALLEL COMBINATIONS
		////////////////////////////////////
		this.parallelismAnalyser = 
			new ParallelismAnalyserVisitor(procInfo, directives, constants);
		newStart.apply(this.parallelismAnalyser);

	}
	
	public boolean isProcessToBeRenamed (String name) {
		return this.renamingProcesses.keySet().contains(name);
	}
	
	public Node preRenaming() 
			throws 
			FileNotFoundException, 
			IOException, 
			ParserException, 
			LexerException{
		
		Hashtable <String,Set<Set<Integer>>> realSyncLUT = parallelismAnalyser.getRealSyncLUT();
		
		ForcedInterleavingOnChannel forcedInterleavingLUT = parallelismAnalyser.getForcedInterleavingLUT();
		
		Set <String> renamingChannels = forcedInterleavingLUT.keySet();
		
		this.renamingProcesses = new Hashtable <String, List<String>>();
		Set <String> processes = channels.getProcessChannels().keySet();
		for (String process : processes){
			List<String> channelsOfProcess = channels.getProcessChannels().get(process);
			ListIterator<String> channelsIter = channelsOfProcess.listIterator();
			while (channelsIter.hasNext()) {
				String next = channelsIter.next();
				if (renamingChannels.contains(HandelC.CHAN_DECL + HandelC.UNDERSCORE + next)){
					List <String> chanList;
					if (!renamingProcesses.containsKey(process)){
						chanList = new Vector <String>();
					} else {
						chanList = renamingProcesses.get(process);
					}
					chanList.add(next);
					renamingProcesses.put(process, chanList);
				}
			}
		}
		/*for (String process : this.renamingProcesses.keySet()) {
			this.procInfo.processesDefinitions.
		}*/
		
		System.out.println("\nRenamingProcess " + renamingProcesses);
		
		String rnProcess = "";
		String channelDeclaration = "";
		
		renamedChannels = new Hashtable <String, List<String>>();
		
		for (String process : this.renamingProcesses.keySet()){
			List <Integer> branches = this.parallelismAnalyser.getBranchesOfProcess(process);
			if (branches != null) {
				for (int branch : branches){
					String rnSubstitutionList = "";
					
					Hashtable<String, List<String>> substitutions = new Hashtable<String, List<String>>();
					
					ListIterator<String> channelsList = this.renamingProcesses.get(process).listIterator();
					while (channelsList.hasNext()){
						String nextChannel = channelsList.next();
						if (!realSyncLUT.containsKey(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel)){
							if (!rnSubstitutionList.equals("")){
								rnSubstitutionList = rnSubstitutionList + " , ";
							}
							rnSubstitutionList = rnSubstitutionList + nextChannel + " <- " + nextChannel + "_" + branch;
							
							if (substitutions.containsKey(nextChannel)){
								List<String> list = substitutions.get(nextChannel);
								list.add(nextChannel + "_" + branch);
								substitutions.put(nextChannel, list);
							} else {
								List<String> list = new Vector<String>();
								list.add(nextChannel + "_" + branch);
								substitutions.put(nextChannel, list);
							}
							
							String channelType = "";
							if (this.channels.containsChannelType(nextChannel)){
								List types = channels.getChannelType(nextChannel);
								if (!types.isEmpty()){
									channelType = channelType + " : ";
									for (Object type:types){
										if (!channelType.equals(" : ")){
											channelType = channelType + " . ";
										}
										channelType = channelType + type.toString();
									}
								}
							}
							if (!channelDeclaration.contains(nextChannel+ "_" + branch)){
								if (!renamedChannels.containsKey(nextChannel)){
									List <String> rem = new Vector<String>();
									rem.add(nextChannel + "_" + branch);
									renamedChannels.put(nextChannel, rem);
								} else{
									List <String> rem = renamedChannels.get(nextChannel);
									if (!rem.contains(nextChannel + "_" + branch)){
										rem.add(nextChannel + "_" + branch);
										renamedChannels.put(nextChannel, rem);
									}
								}
								
								channelDeclaration = channelDeclaration + "\nchannel " + nextChannel + "_" + branch + channelType;
							}
						} else {
							boolean branchIsInFriends = false;
							for (Set<Integer> friends:realSyncLUT.get(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel)){
								branchIsInFriends = branchIsInFriends || friends.contains(branch);
							}
							if (!branchIsInFriends){
								if (!rnSubstitutionList.equals("")){
									rnSubstitutionList = rnSubstitutionList + " , ";
								}
								rnSubstitutionList = rnSubstitutionList + nextChannel + " <- " + nextChannel + "_" + branch;
								
								if (substitutions.containsKey(nextChannel)){
									List<String> list = substitutions.get(nextChannel);
									list.add(nextChannel + "_" + branch);
									substitutions.put(nextChannel, list);
								} else {
									List<String> list = new Vector<String>();
									list.add(nextChannel + "_" + branch);
									substitutions.put(nextChannel, list);
								}
								
								String channelType = "";
								if (this.channels.containsChannelType(nextChannel)){
									List types = channels.getChannelType(nextChannel);
									if (!types.isEmpty()){
										channelType = channelType + " : ";
										for (Object type:types){
											if (!channelType.equals(" : ")){
												channelType = channelType + " . ";
											}
											channelType = channelType + type.toString();
										}
									}
								}
								if (!channelDeclaration.contains(nextChannel+ "_" + branch)){
									if (!renamedChannels.containsKey(nextChannel)){
										List <String> rem = new Vector<String>();
										rem.add(nextChannel + "_" + branch);
										renamedChannels.put(nextChannel, rem);
									} else{
										List <String> rem = renamedChannels.get(nextChannel);
										if (!rem.contains(nextChannel + "_" + branch)){
											rem.add(nextChannel + "_" + branch);
											renamedChannels.put(nextChannel, rem);
										}
									}
									channelDeclaration = channelDeclaration + "\nchannel " + nextChannel + "_" + branch + channelType;
								}
							} else { 
								Set <Set<Integer>> setOfFriends = realSyncLUT.get(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel);
								for (Set<Integer> friends: setOfFriends){
									if (friends.contains(branch)){
										if (!rnSubstitutionList.equals("")){
											rnSubstitutionList = rnSubstitutionList + " , ";
										}
										rnSubstitutionList = rnSubstitutionList + nextChannel + " <- " + nextChannel + "_" + friends.hashCode();
										
										if (substitutions.containsKey(nextChannel)){
											List<String> list = substitutions.get(nextChannel);
											list.add(nextChannel + "_" + friends.hashCode());
											substitutions.put(nextChannel, list);
										} else {
											List<String> list = new Vector<String>();
											list.add(nextChannel + "_" + friends.hashCode());
											substitutions.put(nextChannel, list);
										}
										
										
										String channelType = "";
										if (this.channels.containsChannelType(nextChannel)){
											List types = channels.getChannelType(nextChannel);
											if (!types.isEmpty()){
												channelType = channelType + " : ";
												for (Object type:types){
													if (!channelType.equals(" : ")){
														channelType = channelType + " . ";
													}
													channelType = channelType + type.toString();
												}
											}
										}
										if (!channelDeclaration.contains(nextChannel+ "_" + friends.hashCode())){
											if (!renamedChannels.containsKey(nextChannel)){
												List <String> rem = new Vector<String>();
												rem.add(nextChannel + "_" + friends.hashCode());
												renamedChannels.put(nextChannel, rem);
											} else{
												List <String> rem = renamedChannels.get(nextChannel);
												if (!rem.contains(nextChannel + "_" + friends.hashCode())){
													rem.add(nextChannel + "_" + friends.hashCode());
													renamedChannels.put(nextChannel, rem);
												}
											}
											channelDeclaration = channelDeclaration + "\nchannel " + nextChannel + "_" + friends.hashCode() + channelType;
										}
									}
								}
							}
						}
					}
					String arguments = "";
					if (this.procInfo.containsProcArguments(process)){
						arguments = this.procInfo.getProcArguments(process).toString().trim();
					}
					
					if (this.directives.containsProcChannelInOut(process)){
						Hashtable<String,String> chans = this.directives.getProcChannelInOut(process);
						for (String chan : substitutions.keySet()){
							if (chans.containsKey(chan)){
								for (String subs : substitutions.get(chan)){
									this.directives.putProcChannelInOut(process + "_RN" + branch, subs, chans.get(chan));
								}
							}
						}
					}
					
					rnProcess = rnProcess + "\n" + "-- Processo " + process + " renomeado\n" + process + "_RN" + branch + arguments + " = " + process + arguments + "[[" + rnSubstitutionList + "]]\n";
					
				}
			}
		}
		
		this.cspCode = this.codeInsertion.getCspCode() + channelDeclaration + rnProcess ;
		
		/*String letLocalDef = ""; // this.getLetDefinitions();
		
		/////////////////////////////////
		// PARSING AND CHECKING TEMP FILE
		/////////////////////////////////
		String tempFileName = file.toString()+HandelC.TEMPORARY_FILE;
		
		// Read Original File
		BufferedReader in = new BufferedReader(new FileReader(file.toString()));
		String original = "";
		String str;
		while ((str = in.readLine()) != null) {
			original = original + str + "\n";
		}
		in.close();
		
		// Write Temporary File (original file + main process)
		String mainProc = HandelC.MAINP + HandelC.ASSIGNMENT + directives.getMain();
		String contents = original + this.codeInsertion.getCspCode() + channelDeclaration + rnProcess + letLocalDef + mainProc;
		
		File fileTemp = new File(tempFileName);
		fileTemp.deleteOnExit();
		FileWriter writer = new FileWriter(fileTemp);
		writer.write(contents);
		writer.close();
		
		Node mainStart = ParserAndAnalyser(fileTemp);
		*/
		
		String mainProc = HandelC.MAINP + HandelC.ASSIGNMENT + directives.getMain();
		Node newStart = ParserUtil.parser(file, this.cspCode + mainProc);

		
		newStart.apply(
				new ExtendedDepthFirstAdapter() {
		        	@Override
		        	public void caseACspAbstractDefinition(ACspAbstractDefinition node) {
		            	//System.out.println("caseACspAbstractDefinition -> node: "+node);
		        		throw new AbstractDefinitionException();
		            }
		        });
		
		////////////////////////////
		// RELOAD CHANNELS INFORMATION
		////////////////////////////
		this.channels = new ChannelsVisitor(directives, types, constants);
		newStart.apply(this.channels);
		
		///////////////////////////////////////////
		// RELOAD PROCESSES ARGUMENTS AND DEFINITIONS
		///////////////////////////////////////////
		this.procInfo = new ProcessesInfoVisitor(directives,types, channels,codeInsertion);
		newStart.apply(this.procInfo);
		
		return newStart;

	}
	
	public void defineRenaming() {

		Hashtable <String,Set<Set<Integer>>> realSyncLUT = parallelismAnalyser.getRealSyncLUT();
		
		ForcedInterleavingOnChannel forcedInterleavingLUT = parallelismAnalyser.getForcedInterleavingLUT();
		
		Set <String> renamingChannels = forcedInterleavingLUT.keySet();
		
		this.renamingProcesses = new Hashtable <String, List<String>>();
		Set <String> processes = channels.getProcessChannels().keySet();
		for (String process : processes){
			List<String> channelsOfProcess = channels.getProcessChannels().get(process);
			ListIterator<String> channelsIter = channelsOfProcess.listIterator();
			while (channelsIter.hasNext()) {
				String next = channelsIter.next();
				if (renamingChannels.contains(HandelC.CHAN_DECL + HandelC.UNDERSCORE + next)){
					List <String> chanList;
					if (!renamingProcesses.containsKey(process)){
						chanList = new Vector <String>();
					} else {
						chanList = renamingProcesses.get(process);
					}
					chanList.add(next);
					renamingProcesses.put(process, chanList);
				}
			}
		}
		System.out.println("\nRenamingProcess " + renamingProcesses);
		
		String rnProcess = "";
		String channelDeclaration = "";
		
		renamedChannels = new Hashtable <String, List<String>>();
		
		for (String process : this.renamingProcesses.keySet()){
			List <Integer> branches = this.parallelismAnalyser.getBranchesOfProcess(process);
			if (branches != null) {
				for (int branch : branches){
					String rnSubstitutionList = "";
					ListIterator<String> channelsList = this.renamingProcesses.get(process).listIterator();
					while (channelsList.hasNext()){
						String nextChannel = channelsList.next();
						if (!realSyncLUT.containsKey(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel)){
							if (!rnSubstitutionList.equals("")){
								rnSubstitutionList = rnSubstitutionList + " , ";
							}
							rnSubstitutionList = rnSubstitutionList + nextChannel + " <- " + nextChannel + "_" + branch;
							
							String channelType = "";
							if (this.channels.containsChannelType(nextChannel)){
								List types = channels.getChannelType(nextChannel);
								if (!types.isEmpty()){
									channelType = channelType + " : ";
									for (Object type:types){
										if (!channelType.equals(" : ")){
											channelType = channelType + " . ";
										}
										channelType = channelType + type.toString();
									}
								}
							}
							if (!channelDeclaration.contains(nextChannel+ "_" + branch)){
								if (!renamedChannels.containsKey(nextChannel)){
									List <String> rem = new Vector<String>();
									rem.add(nextChannel + "_" + branch);
									renamedChannels.put(nextChannel, rem);
								} else{
									List <String> rem = renamedChannels.get(nextChannel);
									if (!rem.contains(nextChannel + "_" + branch)){
										rem.add(nextChannel + "_" + branch);
										renamedChannels.put(nextChannel, rem);
									}
								}
								
								channelDeclaration = channelDeclaration + "\nchannel " + nextChannel + "_" + branch + channelType;
							}
						} else {
							boolean branchIsInFriends = false;
							for (Set<Integer> friends:realSyncLUT.get(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel)){
								branchIsInFriends = branchIsInFriends || friends.contains(branch);
							}
							if (!branchIsInFriends){
								if (!rnSubstitutionList.equals("")){
									rnSubstitutionList = rnSubstitutionList + " , ";
								}
								rnSubstitutionList = rnSubstitutionList + nextChannel + " <- " + nextChannel + "_" + branch;
								
								String channelType = "";
								if (this.channels.containsChannelType(nextChannel)){
									List types = channels.getChannelType(nextChannel);
									if (!types.isEmpty()){
										channelType = channelType + " : ";
										for (Object type:types){
											if (!channelType.equals(" : ")){
												channelType = channelType + " . ";
											}
											channelType = channelType + type.toString();
										}
									}
								}
								if (!channelDeclaration.contains(nextChannel+ "_" + branch)){
									if (!renamedChannels.containsKey(nextChannel)){
										List <String> rem = new Vector<String>();
										rem.add(nextChannel + "_" + branch);
										renamedChannels.put(nextChannel, rem);
									} else{
										List <String> rem = renamedChannels.get(nextChannel);
										if (!rem.contains(nextChannel + "_" + branch)){
											rem.add(nextChannel + "_" + branch);
											renamedChannels.put(nextChannel, rem);
										}
									}
									channelDeclaration = channelDeclaration + "\nchannel " + nextChannel + "_" + branch + channelType;
								}
							} else { 
								Set <Set<Integer>> setOfFriends = realSyncLUT.get(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel);
								for (Set<Integer> friends: setOfFriends){
									if (friends.contains(branch)){
										if (!rnSubstitutionList.equals("")){
											rnSubstitutionList = rnSubstitutionList + " , ";
										}
										rnSubstitutionList = rnSubstitutionList + nextChannel + " <- " + nextChannel + "_" + friends.hashCode();
										
										String channelType = "";
										if (this.channels.containsChannelType(nextChannel)){
											List types = channels.getChannelType(nextChannel);
											if (!types.isEmpty()){
												channelType = channelType + " : ";
												for (Object type:types){
													if (!channelType.equals(" : ")){
														channelType = channelType + " . ";
													}
													channelType = channelType + type.toString();
												}
											}
										}
										if (!channelDeclaration.contains(nextChannel+ "_" + friends.hashCode())){
											if (!renamedChannels.containsKey(nextChannel)){
												List <String> rem = new Vector<String>();
												rem.add(nextChannel + "_" + friends.hashCode());
												renamedChannels.put(nextChannel, rem);
											} else{
												List <String> rem = renamedChannels.get(nextChannel);
												if (!rem.contains(nextChannel + "_" + friends.hashCode())){
													rem.add(nextChannel + "_" + friends.hashCode());
													renamedChannels.put(nextChannel, rem);
												}
											}
											channelDeclaration = channelDeclaration + "\nchannel " + nextChannel + "_" + friends.hashCode() + channelType;
										}
									}
								}
							}
						}
					}
					String arguments = "";
					if (this.procInfo.containsProcArguments(process)){
						arguments = this.procInfo.getProcArguments(process).toString().trim();
					}
					
					rnProcess = rnProcess + "\n" + "-- Processo " + process + " renomeado\n" + process + "_RN" + branch + arguments + " = " + process + arguments + "[[" + rnSubstitutionList + "]]\n";
					
				}
			}
		}
		
		this.cspCode = this.codeInsertion.getCspCode() + channelDeclaration + rnProcess ;
	}

	public String getCspRenamingCode() {
		return this.cspCode;
	}
	
	
/*
	private Start ParserAndAnalyser(File file) throws FileNotFoundException, 
	IOException, 
	ParserException,
	LexerException{

		BufferedReader is = new BufferedReader(new FileReader(file));
		StringBuffer buffer = new StringBuffer();
		String line = is.readLine();
		while (line != null) {
			buffer.append(line + "\n");
			line = is.readLine();
		}

		Lexer lexer = new Lexer(new PushbackReader(
				new FileReader(file), 1000));
		Parser parser = new Parser(lexer);
		Start tree = parser.parse();
		
		// CspContextualAnalyser 
		
		CspContextualAnalyser analyser = new CspContextualAnalyser(this);
		analyser.setDebugging(true);
		analyser.setStopOnError(true);
		tree.apply(analyser);

		return tree;
	}
	public void analysed(Node arg0) {
		// TODO Auto-generated method stub
		
	}
	public void analysing(Node arg0) {
		// TODO Auto-generated method stub
		
	}
	public void catchException(CspAnalyserException arg0) {
		// TODO Auto-generated method stub
		
	}
	public void finished() {
		// TODO Auto-generated method stub
		
	}
	public void started() {
		// TODO Auto-generated method stub
		
	}
	public void warning(int arg0, int arg1, String arg2) {
		// TODO Auto-generated method stub
		
	}*/
}