package csp2hc.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PushbackReader;

import csp2hc.messages.HandelC;

import lmf.formula.csp.lexer.Lexer;
import lmf.formula.csp.lexer.LexerException;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.Start;
import lmf.formula.csp.parser.Parser;
import lmf.formula.csp.parser.ParserException;
import lmf.formula.csp.semantic.CspAnalyserListener;
import lmf.formula.csp.semantic.CspContextualAnalyser;
import lmf.formula.csp.semantic.exception.CspAnalyserException;

/**
*
* @author  Ivan Soares
*/
public class ParserUtil{
	
	public static Start parser(File file) throws FileNotFoundException, 
				IOException, 
				ParserException,
				LexerException{

		BufferedReader is = new BufferedReader(new FileReader(file));
		StringBuffer buffer = new StringBuffer();
		String line = is.readLine();
		while (line != null) {
			buffer.append(line + "\n");
			line = is.readLine();
		}

		////////////////////////////
		// PARSING AND CHECKING FILE
		////////////////////////////
		Lexer lexer = new Lexer(new PushbackReader(
				new FileReader(file), 1000));
		Parser parser = new Parser(lexer);
		Start tree = parser.parse();
		
		// CspContextualAnalyser
		CspAnalyserListener listener = new CspAnalyserListener() {
			public void warning(int arg0, int arg1, String arg2) {}
			public void started() {}
			public void finished() {}
			public void catchException(CspAnalyserException arg0) {}
			public void analysing(Node arg0) {}
			public void analysed(Node arg0) {}
		};
		
		CspContextualAnalyser analyser = new CspContextualAnalyser(listener);
		analyser.setDebugging(true);
		analyser.setStopOnError(true);
		tree.apply(analyser);

		return tree;
	}
	
	public static Start parser(File file, String insert) throws FileNotFoundException, 
				IOException, 
				ParserException,
				LexerException{

		// Read Original File
		BufferedReader in = new BufferedReader(new FileReader(file.toString()));
		String original = "";
		String str;
		while ((str = in.readLine()) != null) {
			original = original + str + "\n";
		}
		in.close();

		// To write in temporary File (original file + added content)
		String contents = original + insert;

		// Preparing temporary file
		String tempFileName = file.toString()+HandelC.TEMPORARY_FILE;

		File fileTemp = new File(tempFileName);
		fileTemp.deleteOnExit();
		FileWriter writer = new FileWriter(fileTemp);
		writer.write(contents);
		writer.close();
		
		return (parser(fileTemp));
	}
	

}
