package csp2hc.util;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import com.sun.xml.internal.ws.api.pipe.NextAction;

import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.node.ACspAlphabetisedParallelProcess;
import lmf.formula.csp.node.ACspBaseCallExpr;
import lmf.formula.csp.node.ACspBaseChoiceProcess;
import lmf.formula.csp.node.ACspBaseDotExpr;
import lmf.formula.csp.node.ACspBaseEventList;
import lmf.formula.csp.node.ACspBaseExpr;
import lmf.formula.csp.node.ACspBaseExpression;
import lmf.formula.csp.node.ACspBaseGuardedProcess;
import lmf.formula.csp.node.ACspBaseHidingProcess;
import lmf.formula.csp.node.ACspBaseInterleaveProcess;
import lmf.formula.csp.node.ACspBaseInterruptProcess;
import lmf.formula.csp.node.ACspBaseLengthExpr;
import lmf.formula.csp.node.ACspBaseLogicalExpr;
import lmf.formula.csp.node.ACspBaseMath1Expr;
import lmf.formula.csp.node.ACspBaseMathExpr;
import lmf.formula.csp.node.ACspBaseMinusExpr;
import lmf.formula.csp.node.ACspBaseNotExpr;
import lmf.formula.csp.node.ACspBaseParallelProcess;
import lmf.formula.csp.node.ACspBasePrefixProcess;
import lmf.formula.csp.node.ACspBaseProcess;
import lmf.formula.csp.node.ACspBaseRelationalExpr;
import lmf.formula.csp.node.ACspBaseRenamingProcess;
import lmf.formula.csp.node.ACspBaseReplicatedProcess;
import lmf.formula.csp.node.ACspBaseSequenceExpr;
import lmf.formula.csp.node.ACspBaseSequentialProcess;
import lmf.formula.csp.node.ACspBaseTimeoutProcess;
import lmf.formula.csp.node.ACspEvent;
import lmf.formula.csp.node.ACspHidingProcess;
import lmf.formula.csp.node.ACspInterleaveProcess;
import lmf.formula.csp.node.ACspLinkedEventList;
import lmf.formula.csp.node.ACspParenthesedValue;
import lmf.formula.csp.node.ACspProcessCallExpr;
import lmf.formula.csp.node.ACspProcessDefinition;
import lmf.formula.csp.node.ACspProductionsSet;
import lmf.formula.csp.node.ACspRenamingProcess;
import lmf.formula.csp.node.ACspSetValue;
import lmf.formula.csp.node.ACspSharingParallelProcess;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.PArguments;
import lmf.formula.csp.node.PCallExpr;
import lmf.formula.csp.node.PComunication;
import lmf.formula.csp.node.PEvent;
import lmf.formula.csp.node.PEventList;
import lmf.formula.csp.node.PExpr;
import lmf.formula.csp.node.PExpression;
import lmf.formula.csp.node.PProjection;
import lmf.formula.csp.node.Start;
import lmf.formula.csp.node.TComma;
import lmf.formula.csp.node.TLPar;
import lmf.formula.csp.node.TLProds;
import lmf.formula.csp.node.TName;
import lmf.formula.csp.node.TRPar;
import lmf.formula.csp.node.TRProds;

import csp2hc.messages.HandelC;
import csp2hc.visitors.AlphabetsVisitor;
import csp2hc.visitors.ConstantsAndFunctionsVisitor;
import csp2hc.visitors.DirectivesVisitor;
import csp2hc.visitors.ForcedInterleavingVisitor;
import csp2hc.visitors.ParallelismAnalyserVisitor;
import csp2hc.visitors.ProcessesInfoVisitor;
import csp2hc.visitors.SubstitutionVisitor;

public class RenameProcessesUtil  extends ExtendedDepthFirstAdapter  {
	private ParallelismAnalyserVisitor parallelismAnalyser;
	private ForcedInterleavingVisitor forcedInterleaving;
	private ProcessesInfoVisitor procInfoVisitor;
	private String currentDefinition;
	private String currenAnalysedProcess;
	private boolean isGettingBranch;
	private int nBranches;
	private int currentBranch;
	private SubstitutionVisitor substitutions;
	private DirectivesVisitor directivesVisitor;
	private ConstantsAndFunctionsVisitor constantsVisitor;
	
	public RenameProcessesUtil(ParallelismAnalyserVisitor parallelismAnalyser, 
			ForcedInterleaveInsertion forcIntInsertion, ProcessesInfoVisitor procInfo, 
			BetterPreProcessing preProcessing, DirectivesVisitor directivesVisitor, 
			ConstantsAndFunctionsVisitor constantsVisitor){
		this.parallelismAnalyser = parallelismAnalyser;
		this.forcedInterleaving = preProcessing.getForcedInterleaving();
		this.substitutions = preProcessing.getSubstitutions();
		this.directivesVisitor = directivesVisitor;
		this.constantsVisitor = constantsVisitor;
		this.procInfoVisitor = procInfo;
		this.isGettingBranch = false;
		this.nBranches = this.parallelismAnalyser.getNumberOfBranches(); 
	}
	
	public void caseStart(Start node) {
		node.getPSpecification().apply(this);
	}
	
	public void caseACspProcessDefinition(ACspProcessDefinition process) {
		String name = process.getName().toString().trim();
//		if (!this.forcedInterleavingProtocol.isProcessToBeRenamed(name)){
			//this.procInfo.getProcDefinition(name).apply(this);
			this.currentDefinition = name;
			if (isToBeRenamed()) {
				this.currentBranch = extractProcessBranch(this.currentDefinition);
				this.currenAnalysedProcess = name;
				process.getExpression().apply(this);
			}
//		}
	}
	
	@Override
	public void caseACspHidingProcess(ACspHidingProcess node) {
		node.getProcess().apply(this);
		renameHiddenSet(node);
	}
	
	public void caseACspSharingParallelProcess(ACspSharingParallelProcess node) {
		analyseConcurrency(node);
    }
    
    public void caseACspAlphabetisedParallelProcess(ACspAlphabetisedParallelProcess node) {
    	analyseConcurrency(node);
    }
    
    public void caseACspInterleaveProcess(ACspInterleaveProcess node) {
    	analyseConcurrency(node);
    }
    
    private void analyseConcurrency(Node node){
        // Checking the branches
        Node left = null;
        Node right = null;

        if (node instanceof ACspSharingParallelProcess) {
            left = ((ACspSharingParallelProcess)node).getLeftProcess();
            right = ((ACspSharingParallelProcess)node).getRightProcess();
    	} else if (node instanceof ACspAlphabetisedParallelProcess) {
            left = ((ACspAlphabetisedParallelProcess)node).getLeftProcess();
            right = ((ACspAlphabetisedParallelProcess)node).getRightProcess();
    	} else if (node instanceof ACspInterleaveProcess) {
            left = ((ACspInterleaveProcess)node).getLeftProcess();
            right = ((ACspInterleaveProcess)node).getRightProcess();
    	}
        left.apply(this);
        this.currentBranch ++;
        right.apply(this);

        if (node instanceof ACspSharingParallelProcess) {
        	renameSyncSet((ACspSharingParallelProcess) node);
        }
    }

    public void renameHiddenSet(ACspHidingProcess node) {
        PExpr hidExp = node.getHidenEvents();
        
        AlphabetsVisitor alphabets = new AlphabetsVisitor(this.procInfoVisitor, this.directivesVisitor, this.constantsVisitor);
        alphabets.setCurrentAnalysedProcess("");
        
        Set<String> hiddenChannels = alphabets.analyseChannelSet(hidExp);

        alphabets.resetChannels();
        node.getProcess().apply(alphabets);
        Set<String> visibleChannels = alphabets.getVisibleChannels();
        alphabets.resetChannels();

        Map <String, List<String>> hidSet = new Hashtable<String, List<String>>();
        for (String c: visibleChannels) {
        	int pos = c.indexOf("_BR_");
        	if (pos <= 0) pos = c.indexOf("_SYNC_");
			if (pos >= 0){
				String chan = c.substring(0, pos);
				if (hiddenChannels.contains(chan)){
					if (hidSet.containsKey(chan)){
						List<String> list = hidSet.get(chan);
						list.add(c);
						hidSet.put(chan, list);
					} else {
						List<String> list = new Vector<String>();
						list.add(c);
						hidSet.put(chan, list);
					}
				}
			}
		}
        if (!hidSet.isEmpty()){
        	hiddenChannels.removeAll(hidSet.keySet());
        	for (String chan : hidSet.keySet()) {
        		hiddenChannels.addAll(hidSet.get(chan));
        	}
        	
        	node.setHidenEvents(createACspBaseExpr(createPEventList(hiddenChannels)));
        }
    }


    public void renameSyncSet(ACspSharingParallelProcess node) {
        PExpression syncExp = ((ACspSharingParallelProcess)node).getExpression();
        
        AlphabetsVisitor alphabets = new AlphabetsVisitor(this.procInfoVisitor, this.directivesVisitor, this.constantsVisitor);
        alphabets.setCurrentAnalysedProcess("");
        
        Set<String> channels = alphabets.analyseChannelSet(syncExp);

        Node left = ((ACspSharingParallelProcess)node).getLeftProcess();
        Node right = ((ACspSharingParallelProcess)node).getRightProcess();
        
        alphabets.resetChannels();
        left.apply(alphabets);
        Set<String> leftChannels = alphabets.getVisibleChannels();
        alphabets.resetChannels();

        right.apply(alphabets);
        Set<String> rightChannels = alphabets.getVisibleChannels();
        alphabets.resetChannels();

        rightChannels.retainAll(leftChannels); // rightUsedChannels intersection leftUsedChannels
        //rightChannels.retainAll(channels); // rightUsedChannels minus sync

        Map <String, List<String>> syncSet = new Hashtable<String, List<String>>();
        for (String c: rightChannels) {
        	int pos = c.indexOf("_BR_");
        	if (pos <= 0) pos = c.indexOf("_SYNC_");
			if (pos >= 0){
				String chan = c.substring(0, pos);
				if (channels.contains(chan)){
					if (syncSet.containsKey(chan)){
						List<String> list = syncSet.get(chan);
						list.add(c);
						syncSet.put(chan, list);
					} else {
						List<String> list = new Vector<String>();
						list.add(c);
						syncSet.put(chan, list);
					}
				}
			}
		}
        if (!syncSet.isEmpty()){
        	channels.removeAll(syncSet.keySet());
        	for (String chan : syncSet.keySet()) {
        		channels.addAll(syncSet.get(chan));
        	}
        	
        	node.setExpression(createACspBaseExpression(createPEventList(channels)));
        }
    }

    public void caseACspRenamingProcess(ACspRenamingProcess process) {
		ACspProcessCallExpr procCallExpr = getRenamedProc(process);
		String pname = procCallExpr.getName().getText().trim();
		if (this.substitutions.isRealSubstitution(pname, currentBranch))
			process.replaceBy(procCallToRenaming(procCallExpr));
			
	}
	public void caseACspProcessCallExpr(ACspProcessCallExpr node){
		String name = node.getName().getText().trim();
		if (!name.equals(CspUtil.getSkip()) 
        		&& !name.equals(CspUtil.getStop())) {
			if (this.isGettingBranch ) {
				if (!this.currenAnalysedProcess.equals(name)){
						String processDefinition = this.currenAnalysedProcess;
						this.currenAnalysedProcess = name;
						this.procInfoVisitor.getProcDefinition(name).apply(this);
						this.currenAnalysedProcess = processDefinition;
					}
			} else {
				// Quando o 'branch' do n  igual a -1, o processo no faz parte do fluxo principal de execuo da especificao. Assim, no precisa ser renomeado.
				if (this.currentBranch >= 0 && !this.currentDefinition.equals(node.getName().getText().trim())) {
					//this.branches.add(nodeId);
					if (this.forcedInterleaving.isProcessToBeRenamed(name)){
						if (node.getArguments() != null){
							ACspProcessCallExpr renamedNode = new ACspProcessCallExpr(
									new TName(name+"_RN_"+this.currentBranch, node.getName().getLine(), node.getName().getPos()),
									(PArguments) node.getArguments().clone());
							node.replaceBy(renamedNode);
						} else {
							ACspProcessCallExpr renamedNode = new ACspProcessCallExpr(
									new TName(name+"_RN_"+this.currentBranch, node.getName().getLine(), node.getName().getPos()));
							node.replaceBy(renamedNode);
						}
					} 
				}
				String processDefinition = this.currenAnalysedProcess;
				this.currenAnalysedProcess = name;
				this.isGettingBranch = true;
				this.procInfoVisitor.getProcDefinition(name).apply(this);
				this.isGettingBranch = false;
				this.currenAnalysedProcess = processDefinition;
			}
		}
		
	}
	
	public void caseACspProcessCallExpr_OLD(ACspProcessCallExpr node){
		//int nodeId = this.parallelismAnalyser.getNodeBranchId(node);
		int nodeId = this.parallelismAnalyser.getNodeBranchId(node);
		// Quando o 'branch' do n  igual a -1, o processo no faz parte do fluxo principal de execuo da especificao. Assim, no precisa ser renomeado.
		if (nodeId != -1 && !this.currentDefinition.equals(node.getName().getText().trim())) {
			String name = node.getName().getText().trim();
			//this.branches.add(nodeId);
			if (this.forcedInterleaving.isProcessToBeRenamed(name)){
				if (node.getArguments() != null){
					ACspProcessCallExpr renamedNode = new ACspProcessCallExpr(
							new TName(name+"_RN_"+nodeId, node.getName().getLine(), node.getName().getPos()),
							(PArguments) node.getArguments().clone());
					node.replaceBy(renamedNode);
				} else {
					ACspProcessCallExpr renamedNode = new ACspProcessCallExpr(
							new TName(name+"_RN_"+nodeId, node.getName().getLine(), node.getName().getPos()));
					node.replaceBy(renamedNode);
				}
			} 
		}
	}
	
	private boolean isToBeRenamed() {
		for (int i = 0;  i < nBranches; i++) {
			if (this.currentDefinition.endsWith("_RN_"+i))
				return true;
		}
		if (HandelC.getInstance().MAINP.equals(this.currentDefinition))
			return true;
		return false;
	}
	
	private Integer extractProcessBranch(String pname){
		if (HandelC.getInstance().MAINP.equals(this.currentDefinition))
			return 0;
		int index = pname.lastIndexOf("_RN_") + 4;
		int branch = Integer.parseInt(pname.substring(index));
		return branch;
	}
	
	private ACspProcessCallExpr getRenamedProc(Node proc){
		return new ExtendedDepthFirstAdapter(){
			private ACspProcessCallExpr pcall;
			public ACspProcessCallExpr getRenamedProc(Node p){
				p.apply(this);
				return pcall;
			}
			public void caseACspProcessCallExpr(ACspProcessCallExpr node) {pcall = node;};
		}.getRenamedProc(proc);
	}
	
	private ACspBaseRenamingProcess procCallToRenaming(ACspProcessCallExpr pcall){
		return new ACspBaseRenamingProcess(
				new ACspBaseProcess(
				 new ACspBaseExpr(
				  new ACspBaseLogicalExpr(
				   new ACspBaseRelationalExpr(
					new ACspBaseMathExpr(
					 new ACspBaseMath1Expr(
					  new ACspBaseSequenceExpr(
					   new ACspBaseNotExpr(
						new ACspBaseMinusExpr(
						 new ACspBaseLengthExpr(
						  new ACspBaseDotExpr(
						   (PCallExpr) pcall.clone()))))))))))));
	}
	
	private ACspEvent createACspEvent(String name){
    	return new ACspEvent(new TName(name), new Vector<PProjection>(), new Vector<PComunication>());
    }
    private ACspBaseEventList createACspBaseEventList(String name){
    	return new ACspBaseEventList(createACspEvent(name));
    }
    
    private PEventList createPEventList(Set<String> list){
    	PEventList evList = new ACspBaseEventList();
    	if (!list.isEmpty()){
    		Iterator<String> it = list.iterator();
    		evList = createACspBaseEventList(it.next());
    		while(it.hasNext()){
    			evList = new ACspLinkedEventList(createACspEvent(it.next()), new TComma(), evList);
    		}
    	}
    	return evList;
    }
    
    private ACspBaseExpression createACspBaseExpression(PEventList evList){
		//return new ACspBaseExpression(new ACspBaseInterleaveProcess(new ACspBasePara));
    	return new ACspBaseExpression(
			new ACspBaseInterleaveProcess(
			new ACspBaseParallelProcess(
			new ACspBaseChoiceProcess(
			new ACspBaseTimeoutProcess(
			new ACspBaseInterruptProcess(
			new ACspBaseSequentialProcess(
			new ACspBaseReplicatedProcess(
			new ACspBasePrefixProcess(
			new ACspBaseGuardedProcess(
			new ACspBaseHidingProcess(
			new ACspBaseRenamingProcess(
			new ACspBaseProcess(
			new ACspBaseExpr(
			new ACspBaseLogicalExpr(
			new ACspBaseRelationalExpr(
			new ACspBaseMathExpr(
			new ACspBaseMath1Expr(
			new ACspBaseSequenceExpr(
			new ACspBaseNotExpr(
			new ACspBaseMinusExpr(
			new ACspBaseLengthExpr(
			new ACspBaseDotExpr(
			new ACspBaseCallExpr(
			new ACspSetValue(
			new ACspProductionsSet(
					new TLProds(),
					evList,
					new TRProds()
			))))))))))))))))))))))))));
    }
    
    private ACspBaseExpr createACspBaseExpr(PEventList evList){
		//return new ACspBaseExpression(new ACspBaseInterleaveProcess(new ACspBasePara));
    	return new ACspBaseExpr(
			new ACspBaseLogicalExpr(
			new ACspBaseRelationalExpr(
			new ACspBaseMathExpr(
			new ACspBaseMath1Expr(
			new ACspBaseSequenceExpr(
			new ACspBaseNotExpr(
			new ACspBaseMinusExpr(
			new ACspBaseLengthExpr(
			new ACspBaseDotExpr(
			new ACspBaseCallExpr(
			new ACspSetValue(
			new ACspProductionsSet(
					new TLProds(),
					evList,
					new TRProds()
			)))))))))))));
    }
}
