package csp2hc.visitors;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;
import mathCollection.HashMathSet;
import mathCollection.MathSet;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class CodeInsertionVisitor extends ExtendedDepthFirstAdapter {
	
    // Local Processes Definition -- processName:String -> localsProcessesNames:List<String>
    private Hashtable<String,List<String>> localProcessesDefinitions;
    
    // Local Constants and Functions Definition -- processName:String -> definition:PExpression
    private Hashtable<String,PExpression> localConstFuncDefinitions;
    
    // Processes Arguments -- processName:String -> parameters:ACspParameters
    private Hashtable<String,List<String>> localConstFuncArguments;

	private String currentDefinition;
	
	private ACspParameters globalParameters;
	
	private boolean isLetDefinition;
	
	private Hashtable<String, PExpression> currentProcDefinitions;
	
	private Hashtable<String, ACspParameters> currentArguments;
	
	private Hashtable <String, PExpression> currentConstAndFunc;
	
	private String cspCode;

	private DirectivesVisitor directives;
	
	private boolean isAfterRenaming;
	
    public CodeInsertionVisitor(DirectivesVisitor directives) {
    	this.directives = directives;
		this.localProcessesDefinitions = new Hashtable<String,List<String>>();
		this.localConstFuncArguments = new Hashtable<String, List<String>>();
		this.localConstFuncDefinitions = new Hashtable<String, PExpression>();
		this.currentDefinition = "";
		this.globalParameters = null;
		this.isLetDefinition = false;
		this.cspCode = "";
		this.isAfterRenaming = false;
	}
    
    public boolean isLocalDefinition(String p1, String p2){
    	if (this.localProcessesDefinitions.containsKey(p2)){
    		List <String> locals = this.localProcessesDefinitions.get(p2);
    		if (locals.contains(p1)) return true;
    	}
    	return false;
    }
    
    public void caseACspAbstractCallExpr (ACspAbstractCallExpr node) {
    	throw new AbstractDefinitionException();
    }
    
    public void caseACspProcessCallExpr (ACspProcessCallExpr node) {
    	if (this.isLetDefinition){
	    	String pname = node.getName().getText().trim();
	    	if (this.currentProcDefinitions.containsKey(this.currentDefinition + "_Local_" + pname)) {
	    		node.setName(new TName(this.currentDefinition + "_Local_" + pname));
	    		
	    		if (this.globalParameters != null){
	    			PArgumentList argumentList;
	    			if (node.getArguments() != null) {
	    				argumentList = this.createPArgumentList(this.globalParameters.getParamList(), ((ACspArguments) node.getArguments()).getArgumentList());
	    			} else {
	    				argumentList = this.createPArgumentList(this.globalParameters.getParamList());
	    			}
	    			node.setArguments(new ACspArguments(new TLPar(),argumentList,new TRPar()));
	    		}
	    	}
    	}
    }
    
    public void caseACspConstantCallExpr (ACspConstantCallExpr node) {
    	if (this.isLetDefinition){
	    	String cname = node.getName().getText().trim();
	    	if (this.currentConstAndFunc.containsKey(this.currentDefinition + "_Local_" + cname)) {
	    		node.setName(new TName(this.currentDefinition + "_Local_" + cname));
	    	}
    	}
    }
    
    public void caseACspFunctionCallExpr (ACspFunctionCallExpr node) {
    	if (this.isLetDefinition){
	    	String fname = node.getName().getText().trim();
	    	if (this.currentConstAndFunc.containsKey(this.currentDefinition + "_Local_" + fname)) {
	    		node.setName(new TName(this.currentDefinition + "_Local_" + fname));
	    	}
    	}
    }

	public void caseACspDefinitionParagraph(ACspDefinitionParagraph node){
    	PDefinition definition = node.getDefinition();
        if (definition instanceof ACspProcessDefinition){
        	TName name = ((ACspProcessDefinition)definition).getName();
        	this.currentDefinition = name.toString().trim();
        	this.globalParameters = (ACspParameters)((ACspProcessDefinition)definition).getParameters();
        	PExpression expression = ((ACspProcessDefinition)definition).getExpression();
            expression.apply(this);
        }
    }
	
	public void caseACspLetExpr (ACspLetExpr node){
		if (isLetDefinition){
			throw new UnsupportedConstructorException(Messenger.LET_WITHIN,this.currentDefinition);
		} else {
			this.currentProcDefinitions = new Hashtable<String, PExpression>();
			this.currentArguments = new Hashtable<String, ACspParameters>();
			this.currentConstAndFunc = new Hashtable<String, PExpression>();
			
			this.isLetDefinition = true;
			this.checkDefinitionList(node.getDefinitionList());
//			node.getDefinitionList().apply(this);
			node.getExpression().apply(this);
			
			for (String constantName : this.currentConstAndFunc.keySet()){
				PExpression localExpr = this.currentConstAndFunc.get(constantName);
				localExpr.apply(this);
				
				this.localConstFuncDefinitions.put(constantName, (PExpression) localExpr.clone());
				if (this.currentArguments.containsKey(constantName)){
					List <String> args = new Vector<String>(); 
					PParamList params = this.currentArguments.get(constantName).getParamList();
					while (params != null){
						if (params instanceof ACspBaseParamList){
							args.add(((ACspBaseParamList) params).getParam().toString().trim());
							params = null;
						} else if (params instanceof ACspLinkedParamList){
							args.add(((ACspLinkedParamList) params).getParam().toString().trim());
							params = ((ACspLinkedParamList) params).getNext();
						}
					}
					this.localConstFuncArguments.put(constantName, args);
				}
				this.cspCode = this.cspCode + constantName;
				if (this.currentArguments.containsKey(constantName)) {
					this.cspCode = this.cspCode + this.currentArguments.get(constantName).toString();
				}
				this.cspCode = this.cspCode + "= " + localExpr.toString() + " \n";
			}
			
			for (String procName : this.currentProcDefinitions.keySet()){
				PExpression localExpr = this.currentProcDefinitions.get(procName);
				localExpr.apply(this);
			
				String procParent = procName.substring(0,procName.indexOf("_Local_"));
				String originalProcName = procName.substring(procName.indexOf("_Local_") + 7) ;
				if (this.directives.containsProcArguments(procParent)){
					Hashtable<String,String> args = this.directives.getProcArguments(procParent);
					for (String arg : args.keySet()){
						this.cspCode = this.cspCode + "--!! arg " + arg + " " + args.get(arg) + " within " + procName + " \n";
					}
				}
				if (this.directives.containsProcArguments(originalProcName)){
					Hashtable<String,String> args = this.directives.getProcArguments(originalProcName);
					for (String arg : args.keySet()){
						this.cspCode = this.cspCode + "--!! arg " + arg + " " + args.get(arg) + " within " + procName + " \n";
					}
				}
				if (this.directives.containsProcChannelInOut(procParent)){
					Hashtable<String,String> chans = this.directives.getProcChannelInOut(procParent);
					for (String chan : chans.keySet()){
						this.cspCode = this.cspCode + "--!! channel " + chan + " ";
						this.cspCode = this.cspCode + (HandelC.A_READ.equals(chans.get(chan))?"in":"out") + " within " + procName + " \n";
					}
				}
				if (this.directives.containsProcChannelInOut(originalProcName)){
					Hashtable<String,String> chans = this.directives.getProcChannelInOut(originalProcName);
					for (String chan : chans.keySet()){
						this.cspCode = this.cspCode + "--!! channel " + chan + " ";
						this.cspCode = this.cspCode + (HandelC.A_READ.equals(chans.get(chan))?"in":"out") + " within " + procName + " \n";
					}
				}
				this.cspCode = this.cspCode + procName;
				if (this.currentArguments.containsKey(procName)) {
					this.cspCode = this.cspCode + this.currentArguments.get(procName).toString();
				}
				this.cspCode = this.cspCode + "= " + localExpr.toString() + " \n";
			}
			this.isLetDefinition = false;
		}
	}
	
	private void checkDefinitionList (PDefinitionList definition){
		if (definition instanceof ACspBaseDefinitionList)
			checkDefinitionsInLet(((ACspBaseDefinitionList) definition).getDefinition());
		if (definition instanceof ACspLinkedDefinitionList) {
			checkDefinitionsInLet(((ACspLinkedDefinitionList) definition).getDefinition());
			checkDefinitionList(((ACspLinkedDefinitionList) definition).getNext());
		}
	}
	
	
	private void checkDefinitionsInLet (PDefinition definition){
		if (definition instanceof ACspProcessDefinition){
			checkProcessInLet((ACspProcessDefinition) definition);
		} else if (definition instanceof ACspConstantDefinition){
			checkConstantInLet((ACspConstantDefinition) definition);
		} else if (definition instanceof ACspFunctionDefinition){
			checkFunctionInLet((ACspFunctionDefinition) definition);
		} else {
			throw new UnsupportedConstructorException(Messenger.LET_WITHIN_PROHIBITION, this.currentDefinition);
		}
	}
	
	private void checkProcessInLet (ACspProcessDefinition def) {
		TName name = def.getName();
        ACspParameters localParameters = (ACspParameters) def.getParameters();
        PExpression expression = def.getExpression();
        
        List <ACspParameters> parameters = new Vector<ACspParameters>();
        if (this.globalParameters != null) parameters.add(this.globalParameters);
        if (localParameters != null) parameters.add(localParameters);
        
        String pname = name.toString().trim();
        pname = this.currentDefinition + "_Local_" + pname;
        
        this.currentProcDefinitions.put(pname, (PExpression)expression.clone());
        if (localParameters != null && this.globalParameters != null){
        	PParamList paramList = localParameters.getParamList();
        	PParamList globalParamList = this.globalParameters.getParamList();
        	while (!(globalParamList instanceof ACspBaseParamList)){
        		ACspParam globalParam = (ACspParam) ((ACspLinkedParamList)globalParamList).getParam().clone();
        		paramList = new ACspLinkedParamList(globalParam, new TComma(), paramList);
        		globalParamList = ((ACspLinkedParamList)globalParamList).getNext();
        	}
        	if (globalParamList instanceof ACspBaseParamList){
        		ACspParam globalParam = (ACspParam) ((ACspBaseParamList)globalParamList).getParam().clone();
        		paramList = new ACspLinkedParamList(globalParam, new TComma(), paramList);
        	}
        	this.currentArguments.put(pname, new ACspParameters(new TLPar(), paramList, new TRPar()));
        } else if (localParameters != null) {
        	this.currentArguments.put(pname, localParameters);
        } else if (this.globalParameters != null){
        	this.currentArguments.put(pname, this.globalParameters);
        }
        
        if (this.localProcessesDefinitions.containsKey(currentDefinition)){
        	List <String> defs = this.localProcessesDefinitions.get(currentDefinition);
        	defs.add(name.toString().trim());
        	this.localProcessesDefinitions.put(currentDefinition, defs);
        } else {
        	List <String> defs = new Vector<String>();
        	defs.add(name.toString().trim());
        	this.localProcessesDefinitions.put(currentDefinition, defs);
        }
	}
	
	private void checkFunctionInLet (ACspFunctionDefinition def) {
		String name = def.getName().toString().trim();
        PExpression expression = def.getExpression();
        String fname = this.currentDefinition + "_Local_" + name;
        if (this.currentConstAndFunc.containsKey(fname)) {
            throw new RepeatedConstantException(name);
        } else {
            PParameters params = def.getParameters();
            if (params != null) {
            	this.currentArguments.put(fname, (ACspParameters) params);
            }
            this.currentConstAndFunc.put(fname, (PExpression) expression.clone());
        }
        if (this.localProcessesDefinitions.containsKey(currentDefinition)){
        	List <String> defs = this.localProcessesDefinitions.get(currentDefinition);
        	defs.add(name.toString().trim());
        	this.localProcessesDefinitions.put(currentDefinition, defs);
        } else {
        	List <String> defs = new Vector<String>();
        	defs.add(name.toString().trim());
        	this.localProcessesDefinitions.put(currentDefinition, defs);
        }
	}
	
	private void checkConstantInLet (ACspConstantDefinition def) {
		String name = def.getName().toString().trim();
        PExpression expression = def.getExpression();

        String cname = this.currentDefinition + "_Local_" + name;
        if (this.currentConstAndFunc.containsKey(cname)) {
            throw new RepeatedConstantException(cname);
        } else {
            this.currentConstAndFunc.put(cname, (PExpression) expression.clone());
        }
        if (this.localProcessesDefinitions.containsKey(currentDefinition)){
        	List <String> defs = this.localProcessesDefinitions.get(currentDefinition);
        	defs.add(name);
        	this.localProcessesDefinitions.put(currentDefinition, defs);
        } else {
        	List <String> defs = new Vector<String>();
        	defs.add(name);
        	this.localProcessesDefinitions.put(currentDefinition, defs);
        }
	}
	
	public PArgumentList createPArgumentList (PParamList paramList){
		PArgumentList argList;
		String param = "";
		if (paramList instanceof ACspBaseParamList){
			param = ((ACspBaseParamList) paramList).getParam().toString().trim();
			return (new ACspBaseArgumentList(createArgument(param)));
		} else if (paramList instanceof ACspLinkedParamList){
			param = ((ACspLinkedParamList) paramList).getParam().toString().trim();
			return (new ACspLinkedArgumentList(createArgument(param),new TComma(), createPArgumentList(((ACspLinkedParamList) paramList).getNext())));
		} else return null;
	}
	
	public PArgumentList createPArgumentList (PParamList paramList, PArgumentList argList){
		String param = "";
		if (paramList instanceof ACspBaseParamList){
			param = ((ACspBaseParamList) paramList).getParam().toString().trim();
			return (new ACspLinkedArgumentList(createArgument(param),new TComma(), (PArgumentList)argList.clone()));
		} else if (paramList instanceof ACspLinkedParamList){
			param = ((ACspLinkedParamList) paramList).getParam().toString().trim();
			return ( createPArgumentList(
						((ACspLinkedParamList) paramList).getNext(), 
						new ACspLinkedArgumentList(createArgument(param),new TComma(), createPArgumentList(((ACspLinkedParamList) paramList).getNext()))
						));
		} else return null;
	}
	
	private ACspArgument createArgument (String expr){
		return new ACspArgument(
				new ACspBaseExpression(
				  new ACspBaseInterleaveProcess(
					new ACspBaseParallelProcess(
					  new ACspBaseChoiceProcess(
						new ACspBaseTimeoutProcess(
						  new ACspBaseInterruptProcess(
							new ACspBaseSequentialProcess(
							  new ACspBaseReplicatedProcess(
								new ACspBasePrefixProcess(
								  new ACspBaseGuardedProcess(
									new ACspBaseHidingProcess(
									  new ACspBaseRenamingProcess(
										new ACspBaseProcess(
										  new ACspBaseExpr(
											new ACspBaseLogicalExpr(
											  new ACspBaseRelationalExpr(
												new ACspBaseMathExpr(
												  new ACspBaseMath1Expr(
													new ACspBaseSequenceExpr(
													  new ACspBaseNotExpr(
														new ACspBaseMinusExpr(
														  new ACspBaseLengthExpr(
															new ACspBaseDotExpr(
															  new ACspNameCallExpr(new TName(expr))))))))))))))))))))))))));
	}

	public List<String> getConstAndFuncNames(){
		return ListUtil.createList(this.localConstFuncDefinitions.keySet()); 
	}
	public PExpression getConstAndFuncExpr(String definitionName){
		PExpression expression = null;
		if (this.localConstFuncDefinitions.containsKey(definitionName)){
			expression = this.localConstFuncDefinitions.get(definitionName);
		}
		return expression;
	}
	/*public List<String> getArguments(String definitionName){
		
		if (this.currentArguments.contains(definitionName)){
			ACspParameters arguments = this.currentArguments.get(definitionName);
			arguments.getParamList();
		}
		return expression;
	}*/
	
	public String getCspCode (){
		return this.cspCode;
	}

	public List<String> getConstAndFuncArgs(String defName) {
		List args = null;
		if (this.localConstFuncArguments.containsKey(defName)){
			args = this.localConstFuncArguments.get(defName);
		}
		return args;
	}

	public boolean containsArgs(String def) {
		return this.localConstFuncArguments.containsKey(def);
	}
	
	public Node getStartNode (File file, DirectivesVisitor directives) throws IOException, ParserException, LexerException{

		/////////////////////////////////
		// PARSING AND CHECKING TEMP FILE
		/////////////////////////////////
		String tempFileName = file.toString()+HandelC.TEMPORARY_FILE;

		// Read Original File
		BufferedReader in = new BufferedReader(new FileReader(file.toString()));
		String original = "";
		String str;
		while ((str = in.readLine()) != null) {
			original = original + str + "\n";
		}
		in.close();

		// Write Temporary File (original file + update)
		String contents = original + this.getCspCode();

		File fileTemp = new File(tempFileName);
		fileTemp.deleteOnExit();
		FileWriter writer = new FileWriter(fileTemp);
		writer.write(contents);
		writer.close();
		
		// Parser Temporary File
		Node start = ParserUtil.parser(fileTemp);

		// Type Check Temporary File
		
		start.apply(
				new ExtendedDepthFirstAdapter() {
		        	@Override
		        	public void caseACspAbstractDefinition(ACspAbstractDefinition node) {
		            	//System.out.println("caseACspAbstractDefinition -> node: "+node);
		        		throw new AbstractDefinitionException();
		            }
		        });
		
		//////////////////////
		// LOAD DIRECTIVE INFO
		////////////////////// 
		this.directives = new DirectivesVisitor();
		start.apply(this.directives);
		
		return start;
	}
	
	public DirectivesVisitor getNewDirectives(){
		return this.directives;
	}
}

