package csp2hc.visitors;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.type.CspIntType;
import lmf.formula.csp.type.CspType;
import lmf.formula.csp.type.CspUnknownType;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class ExpressionsVisitor extends ExtendedDepthFirstAdapter {

    // Visitor that has the information about the directies
    protected DirectivesVisitor directives;

    // Resulting handel-c code
    protected String hcCode = "";

    // Current definition being read
    protected String currentDefinition;

    // Indicates that we are analysing a set
    private boolean isSet;

    // Indicates that we are analysing an array index
    private boolean isArrayIndex;
    
    // Indicates if this is being used to translate macro expressions
    private boolean isExpression;
    
    protected List<String> inputVarNamesOfMultiComm;
    protected List<String> inputVarTypesOfMultiComm;
    protected List<String> inputVarValuesOfMultiComm;
    
    
    public ExpressionsVisitor(DirectivesVisitor directives, boolean isExpression) {
		this.directives = directives;
		this.currentDefinition = "";
		this.hcCode = "";
		
		this.isExpression = isExpression;
		this.isSet = false;
		this.isArrayIndex = false;
    }

    public boolean getIsExpression() { return this.isExpression; }
    public void setIsExpression(boolean newIsExpression) { this.isExpression = newIsExpression; }

    public String getHcCode() { return this.hcCode; }
    public void resetHcCode() { this.hcCode = ""; }
    
    public void setCurrentDefinition(String def){ 
        this.currentDefinition = def; 
    }
    
    public void caseACspChannelDefinition(ACspChannelDefinition node){
    }

    public void caseACspDatatypeDefinition(ACspDatatypeDefinition node){
    }
    
   	// Booleans
    public void caseACspAndLogicalExpr(ACspAndLogicalExpr node) {
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.SPACE + HandelC.LOGICAL_AND + HandelC.SPACE;
        node.getRightExpr().apply(this);
    }

    public void caseACspOrLogicalExpr(ACspOrLogicalExpr node) {
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.SPACE + HandelC.LOGICAL_OR + HandelC.SPACE;
        node.getRightExpr().apply(this);
    }
    
    public void caseACspNotExpr(ACspNotExpr node) {
        this.hcCode = this.hcCode + HandelC.LOGICAL_NOT;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getExpression().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    // Numbers 

    public void caseACspNumberValue(ACspNumberValue node) {
        this.hcCode = this.hcCode + node.toString().trim();
    }

    public void caseACspAddMathExpr(ACspAddMathExpr node) {
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.MATH_SUM;
        node.getRightExpr().apply(this);
    }

    public void caseACspDiffMathExpr(ACspDiffMathExpr node) {
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.MATH_DIFF;
        node.getRightExpr().apply(this);
    }

    public void caseACspMulMath1Expr(ACspMulMath1Expr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.MATH_MULT;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    public void caseACspDivMath1Expr(ACspDivMath1Expr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.MATH_DIV;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    public void caseACspRemainderMath1Expr(ACspRemainderMath1Expr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.MATH_REMAINDER;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    public void caseACspMinusExpr(ACspMinusExpr node) {
        this.hcCode = this.hcCode + HandelC.MATH_NEG;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getExpression().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    // Relations
    public void caseACspEqualRelationalExpr(ACspEqualRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_EQUALS;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    public void caseACspNotEqualRelationalExpr(ACspNotEqualRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_NOT_EQUALS;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    public void caseACspGtRelationalExpr(ACspGtRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_GREATER;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    public void caseACspGteRelationalExpr(ACspGteRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_GREATER_EQUALS;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    public void caseACspLtRelationalExpr(ACspLtRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_LESS;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    public void caseACspLteRelationalExpr(ACspLteRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_LESS_EQUALS;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    
    // Tuple
    public void caseACspTupleValue(ACspTupleValue node) {
    }

    // Sets
    public void caseACspSetValue(ACspSetValue node) {
        this.isSet = true;
    	//System.out.println("ExpressionsVisitor :: caseACspSetValue -> node "+node);
        node.getSet().apply(this);
    	//System.out.println("ExpressionsVisitor :: caseACspSetValue -> FIM ");
        this.isSet = false;
    }

    public void caseACspSet(ACspSet node) {
    	PElementList elements = node.getElementList();
    	if (elements != null) {
			elements.apply(this);
		} else {
	        String emptySet = this.getCurrentEmptySet();
			this.hcCode = this.hcCode + emptySet;
		}
    }
    
    public void caseACspLinkedElementList(ACspLinkedElementList node) {
        //System.out.println("ExpressionsVisitor :: caseACspLinkedElementList -> INICIO para #"+node+"# em #"+this.currentDefinition+"#");

        PCollectionElement element = node.getCollectionElement();
		
		if (element instanceof ACspRangeCollectionElement) {
			// It could be a range but we do not support it yet
			// EXCEPTION RAISED BY RESTRICTION VISITOR
		} else if (element instanceof ACspCollectionElement){
			this.hcCode = this.hcCode 
						+ HandelC.SET_UNION + HandelC.PAR_BEGIN;
			element.apply(this);
			this.hcCode = this.hcCode 
						+ HandelC.SEP_ARGUMENTS;
			node.getNext().apply(this);
			this.hcCode = this.hcCode 
						+ HandelC.PAR_END;
		}
    }

    // Sequences

    public void caseACspArgument(ACspArgument node) {
        String macro = HandelC.getInstance().getSetVarOf(HandelC.SINGLETON) + HandelC.UNDERSCORE + HandelC.MACRO;
        if(this.isSet) {
            this.hcCode = this.hcCode + macro + HandelC.PAR_BEGIN;
        }
        node.getExpression().apply(this);
        if(this.isSet) {
            this.hcCode = this.hcCode + HandelC.PAR_END;
        }

    }

    public void caseACspLinkedArgumentList(ACspLinkedArgumentList node) {
        node.getArgument().apply(this);
        this.hcCode = this.hcCode + HandelC.SEP_ARGUMENTS;
        node.getNext().apply(this);
    }
    
    public void caseACspNameCallExpr(ACspNameCallExpr node){
    	//System.out.println("ExpressionsVisitor :: caseACspNameCallExpr -> Node #"+node+"# em #"+this.currentDefinition+"#");
    	this.hcCode = this.hcCode + node.toString().trim();
    }
    
    public void caseACspConstantCallExpr(ACspConstantCallExpr node){
    	String name = node.getName().getText().trim();
    	this.hcCode = this.hcCode + name;
    }
    
	public void caseACspProductionsSet(ACspProductionsSet node){
    	PEventList elements = node.getEventList();
    	if (elements != null) {
			elements.apply(this);
		} else {
	        this.hcCode = this.hcCode + HandelC.getEmptySetVarOf(HandelC.CHANNEL);
		}

    }

    public void caseACspLinkedEventList(ACspLinkedEventList node) {
        PEvent event = node.getEvent();
		
		this.hcCode = this.hcCode 
					+ HandelC.SET_UNION + HandelC.PAR_BEGIN;
		event.apply(this);
		this.hcCode = this.hcCode 
					+ HandelC.SEP_ARGUMENTS;
		node.getNext().apply(this);
		this.hcCode = this.hcCode 
					+ HandelC.PAR_END;
    }

    public void caseACspBaseEventList(ACspBaseEventList node) {
        PEvent event = node.getEvent();
		event.apply(this);
    }

    // This should be only accessible in constants and functions definitions 
    public void caseACspEvent(ACspEvent node) {
        String event = node.getChannelName().getText().trim();
        
    	String lut = HandelC.getInstance().getSingletonSetsLookupTableName(HandelC.CHANNEL);
    	String channel = HandelC.getInstance().getChannelName(event);
		this.hcCode = this.hcCode + lut + HandelC.ARRAY_BEGIN + channel + HandelC.ARRAY_END;
    }
    
    public void caseACspIfExpr(ACspIfExpr node){
    	
        if (this.isExpression) {
            this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
            this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
            node.getCondition().apply(this);
            this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.CONDITIONAL_COND + HandelC.SPACE;
            this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
            node.getThenExpr().apply(this);
            this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.CONDITIONAL_SEP + HandelC.SPACE;
            this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
            node.getElseExpr().apply(this);
            this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.SPACE;
            this.hcCode = this.hcCode + HandelC.PAR_END;
        } else {
            this.hcCode = this.hcCode + HandelC.IF + HandelC.PAR_BEGIN;
            this.setIsExpression(true);
            this.hcCode = this.hcCode + HandelC.PAR_BEGIN + HandelC.BOOLEAN + HandelC.PAR_END + HandelC.PAR_BEGIN;
            node.getCondition().apply(this);
            this.hcCode = this.hcCode + HandelC.PAR_END;
            this.setIsExpression(false);

            this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.BLOCK_BEGIN;
            node.getThenExpr().apply(this);
            this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.SPACE;
            this.hcCode = this.hcCode + HandelC.ELSE + HandelC.SPACE + HandelC.BLOCK_BEGIN;
            node.getElseExpr().apply(this);
            this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.SPACE;
        }
    }

    public void caseACspAssert(ACspAssert node) {
    }

    public void caseACspFunctionCallExpr(ACspFunctionCallExpr node){
        // this is a constant
    	String functionName = node.getName().getText();
    	
    	// Checking if it a FDR pre-defined function
    	String hcName = HandelC.getInstance().getHCFunction(functionName);
    	if (hcName.equals("")) {
    		hcName = functionName;
    	}
    	
        this.hcCode = this.hcCode + hcName;
        // it might have arguments, so...
        if (node.getArguments() != null) {
            this.hcCode = this.hcCode + HandelC.PAR_BEGIN;

            boolean oldIsExpression = this.isExpression;
            this.setIsExpression(true);
            node.getArguments().apply(this);
            this.setIsExpression(oldIsExpression);
            
            this.hcCode = this.hcCode + HandelC.PAR_END;
        }
    }	
    
    // when translating the accesses to an element of an array of channel
    // this function is invoked in order to guarantee that, if the type
    // is an integer, it uses the unsigned representation of that
    // integer; otherwise, we could have access to negative numbers
    protected String getUnsignedCastForInteger() {
        return HandelC.PAR_BEGIN + HandelC.UNSIGNED + HandelC.SPACE + HandelC.INT + HandelC.PAR_END + HandelC.PAR_BEGIN;
    }

    private String getCurrentEmptySet() {
    	String type = this.getCurrentEmptySetType();
		String emptySet = HandelC.getInstance().getEmptySetVarOf(type);
		return emptySet;
    }

    private String getCurrentEmptySetType() {
    	String type = "";
        if (this.directives.containsSetType(this.currentDefinition)) {
            type = (String)this.directives.getSetType(this.currentDefinition);
        } else {
            throw new UndefinedEmptySetTypeException(this.currentDefinition);
        }
		return type;
    }
    
    public void caseACspCollectionElement(ACspCollectionElement node) {
    	
    	MaximalTypesInferenceVisitor inference = new MaximalTypesInferenceVisitor();
    	node.apply(inference);
    	CspType type = inference.getType();

		String name = node.toString().trim();
    	if (type instanceof CspUnknownType) {
			// in this case, the type checker did not work
    		if (this.directives.containsProcArguments(this.currentDefinition) 
    	        && this.directives.getProcArguments(this.currentDefinition).containsKey(name)) {
    			String hcType = ((String)this.directives.getProcArguments(this.currentDefinition).get(name)).trim();
				boolean isInteger = hcType.equals(HandelC.INTEGERL);
	        	this.translateCollectionElement(hcType, isInteger, node);
    		} else {
    			throw new CouldNotInferTypeException(name,this.currentDefinition);
    		}
    	} else {
        	boolean isInteger = (type instanceof CspIntType);
        	String hcType = HandelC.getInstance().getHandelCTypeName(type.toString());
        	this.translateCollectionElement(hcType, isInteger, node);
    	}
    }

    protected void translateCollectionElement(String hcType, boolean isInteger, ACspCollectionElement node) {
    	String lut = HandelC.getInstance().getSingletonSetsLookupTableName(hcType);
		
		this.hcCode = this.hcCode + lut + HandelC.ARRAY_BEGIN;
		if (isInteger) {
			this.hcCode = this.hcCode + this.getUnsignedCastForInteger();
		}
		node.getMathExpr().apply(this);
		if (isInteger) {
			this.hcCode = this.hcCode + HandelC.PAR_END;
		}
		this.hcCode = this.hcCode + HandelC.ARRAY_END;
    }
    
}
