package csp2hc.visitors;

import java.util.EmptyStackException;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.Vector;

import csp2hc.messages.HandelC;
import csp2hc.util.CspUtil;
import csp2hc.util.ForcedInterleavingOnChannel;
import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.node.ACspProcessCallExpr;
import lmf.formula.csp.node.ACspProcessDefinition;
import lmf.formula.csp.node.ACspRenamingProcess;
import lmf.formula.csp.node.ACspSubstitution;
import lmf.formula.csp.node.PExpression;
import lmf.formula.csp.node.Start;

public class ForcedInterleavingVisitor extends ExtendedDepthFirstAdapter {
	
	private ChannelsVisitor channelsVisitor;
	
	private ProcessesInfoVisitor procsInfoVisitor;

	private ParallelismAnalyserVisitor parallelismVisitor;
	
	private DirectivesVisitor directives;

	private SubstitutionVisitor substVisitor;

	// nodeId -> [channel -> channelList]
	private Map<Integer, Map<String,List<String>>> processesRenaming;
	
	private List <Integer> affectedProcessesRenaming;

	//Map <Integer, List<Integer>> processCalls;
	
	//private Stack<Integer> processCalls;
	
	private String currentAnalysedProcess;
	
	private Map <Integer, String> processesName;
	
	private String cspCode;

	private Stack<Integer> ctx;
	
	private Map<Integer, Map<String, String>> realSubstitution;
	
	//private Map<Integer,List<String>> forcedRenaming;
	
	//private Map<Integer,List<String>> realRenaming;
	
	//private List<List> forcedSubst;
	
	private boolean forcedRenaming;
	
	public ForcedInterleavingVisitor(DirectivesVisitor directives,
			ChannelsVisitor channelsVisitor,
			ProcessesInfoVisitor procsInfoVisitor,
			ParallelismAnalyserVisitor parallelismVisitor, 
			SubstitutionVisitor substVisitor) {
		this.directives = directives;
		this.channelsVisitor = channelsVisitor;
		this.procsInfoVisitor = procsInfoVisitor;
		this.parallelismVisitor = parallelismVisitor;
		this.substVisitor = substVisitor;
		initialise();
	}

	private void initialise() {
		setProcessesRenaming(new Hashtable<Integer, Map<String,List<String>>>());
		setAffectedProcessesRenaming(new Vector<Integer>());
		//setProcessCalls(new Stack<Integer>());
		setCurrentAnalysedProcess("");
		setProcessesName(new Hashtable<Integer, String>());
		this.ctx = new Stack<Integer>();
		this.realSubstitution = new Hashtable<Integer, Map<String,String>>();
		//this.forcedRenaming = new Hashtable<Integer, List<String>>();
		//this.realRenaming = new Hashtable<Integer, List<String>>();
		
		//this.forcedSubst = new Vector<List>();
		
		forcedRenaming = false;
	}
	
	public void caseStart(Start node) {
		node.getPSpecification().apply(this);
		
		cspCode = "";
		
		List<String> affect = new Vector<String>();
		for (int nodeId : this.affectedProcessesRenaming) {
			int pBranch = this.parallelismVisitor.getCtxNodeBranchId(nodeId);
			String pName = this.processesName.get(nodeId); 
			
			if (this.directives.containsProcChannelInOut(pName)){
				Map<String,String> chansInOut = this.directives.getProcChannelInOut(pName);
				for (String chan : chansInOut.keySet()) {
					cspCode = cspCode + 
								"--!! channel " + chan + (chansInOut.get(chan).equals(HandelC.A_READ)?" in":" out") +
								" within " + pName + "_RN_" + pBranch + " \n";
				}
			}
			
			if (this.directives.containsProcArguments(pName)){
				Map<String,String> args = this.directives.getProcArguments(pName);
				for (String arg : args.keySet()) {
					cspCode = cspCode + 
								"--!! arg " + arg + " " + args.get(arg) +
								" within " + pName + "_RN_" + pBranch + " \n";
				}
			}
			
			cspCode = cspCode + pName + "_RN_" + pBranch;
			if (this.procsInfoVisitor.containsProcArguments(pName)){
				cspCode = cspCode + " " + this.procsInfoVisitor.getProcArguments(pName).toString().trim();
			}
			cspCode = cspCode + " = " + this.procsInfoVisitor.getProcDefinition(pName).toString() + " \n\n";
			
			affect.add(pName);
		}
		//System.out.println("\nAffected Processes Renaming -> "+affect);
		
	}
	
	public void caseACspRenamingProcess(ACspRenamingProcess node) {
		int ctxId = this.ctx.hashCode();
		int nodeId = node.hashCode();
		//String pname = node.getName().getText().trim();
		
		String pname = new ExtendedDepthFirstAdapter() {
			private String name;
			public void caseACspProcessCallExpr(ACspProcessCallExpr node) { name = node.getName().getText().trim();}
			public String getProcessName(ACspRenamingProcess node){node.apply(this); return name;}
		}.getProcessName(node);
		
		//int pbranch = this.parallelismVisitor.getNodeBranchId(node);
		int pbranch = this.parallelismVisitor.getCtxNodeBranchId(node, this.ctx.hashCode());
		Map<String,List<String>> substitutions = this.substVisitor.getSubstitutionList(pname, pbranch);
		
		if (!pname.equals(CspUtil.getSkip()) 
        		&& !pname.equals(CspUtil.getStop())
        		&& !getCurrentAnalysedProcess().equals(pname)) {
			this.processesName.put(nodeId+ctxId, pname);
		
			ForcedInterleavingOnChannel forcedInterleavingLUT = this.parallelismVisitor.getForcedInterleavingLUT();
			Hashtable<String, Set<Set<Integer>>> realSyncLUT = this.parallelismVisitor.getRealSyncLUT();
			
			if (this.channelsVisitor.getProcessChannels().containsKey(pname)) {
				List<String> pchannels = this.channelsVisitor.getProcessChannels().get(pname);
				List<String> pchannelSubs = new Vector<String>();
				for (String chan : pchannels) {
					if (substitutions.containsKey(chan)){
						pchannelSubs.addAll(substitutions.get(chan));
						Map<String, String> s;
						if (realSubstitution.containsKey(nodeId+ctxId))
							s = realSubstitution.get(nodeId+ctxId);
						else s = new Hashtable<String, String>();
						for (String string : substitutions.get(chan)) {
							s.put(string,chan);
						}
						realSubstitution.put(nodeId+ctxId, s);
					}
					else pchannelSubs.add(chan);
				}
				for (String cname : pchannelSubs) {
					String cnameKey = HandelC.getInstance().getChannelName(cname);
					//if (forcedInterleavingLUT.containsKey(cnameKey) && !realSyncLUT.containsKey(cnameKey)) {
					if (forcedInterleavingLUT.containsKey(cnameKey) && !isSyncBranch(cnameKey, pbranch)) {
						//insertSubstChannel(nodeId+ctxId, cname, "_BR_"+pbranch);
						
						changeSubstitutionList(pname,pbranch,cname,"_BR_"+pbranch);
						
					} else if (forcedInterleavingLUT.containsKey(cnameKey)) {
						// cnameKey is in realSyncLUT and pbranch is in one of the sync sets
						List <String> list = new Vector<String>();
						Set <Set<Integer>> syncs = realSyncLUT.get(cnameKey);
						for (Set<Integer> sync: syncs){
							if (sync.contains(pbranch)) {
								//insertSubstChannel(nodeId+ctxId, cname, "_SYNC_"+sync.hashCode());
								
								list.add("_SYNC_"+sync.hashCode());
							}
						}
						changeSubstitutionList(pname,pbranch,cname,list);
					}
				}
			}
			
			String parent = getCurrentAnalysedProcess();
			/*if (!isMain(parent)){
				//int parentDefinition = this.processCalls.peek();
				
				int parentNode = this.ctx.pop();
				int parentDefinition = this.ctx.hashCode() + parentNode;
				this.ctx.push(parentNode);
				
				if ( //this.processesRenaming.containsKey(ctxId + nodeId)
						this.substVisitor.isChangedSubst(pname, pbranch)
						&& !this.processesRenaming.containsKey(parentDefinition)
						&& !this.affectedProcessesRenaming.contains(parentDefinition))
					this.affectedProcessesRenaming.add(parentDefinition);
			}*/
			setCurrentAnalysedProcess(pname);
			
			boolean forcedRenamingAux = this.forcedRenaming;
			this.forcedRenaming = false;
			
			//this.processCalls.push(nodeId);
			this.ctx.push(nodeId);
			PExpression proc = (PExpression)this.procsInfoVisitor.getProcDefinition(pname);
	    	proc.apply(this);
	    	this.ctx.pop();
	    	//if (this.processCalls.pop() != nodeId) {throw new java.lang.RuntimeException();}
	    	
	    	this.forcedRenaming = forcedRenamingAux || this.forcedRenaming || this.substVisitor.isChangedSubst(pname, pbranch);
	    	
	    	setCurrentAnalysedProcess(parent);
		}
	}

	public void caseACspProcessDefinition(ACspProcessDefinition node) {
		String pname = node.getName().getText().trim();
		setCurrentAnalysedProcess(pname);
		if (isMain(pname)){
			this.ctx = new Stack<Integer>();
			this.ctx.push(node.hashCode());
			PExpression expression = node.getExpression();
			expression.apply(this);
		}
	}
	
	public void caseACspProcessCallExpr(ACspProcessCallExpr node) {
		int ctxId = this.ctx.hashCode();
		int nodeId = node.hashCode();
		String pname = node.getName().getText().trim();
		//int pbranch = this.parallelismVisitor.getNodeBranchId(node);
		int pbranch = this.parallelismVisitor.getCtxNodeBranchId(node, this.ctx.hashCode());
		if (!pname.equals(CspUtil.getSkip()) 
        		&& !pname.equals(CspUtil.getStop())
        		&& !getCurrentAnalysedProcess().equals(pname)) {
			this.processesName.put(nodeId+ctxId, pname);
		
			ForcedInterleavingOnChannel forcedInterleavingLUT = this.parallelismVisitor.getForcedInterleavingLUT();
			Hashtable<String, Set<Set<Integer>>> realSyncLUT = this.parallelismVisitor.getRealSyncLUT();
			
			if (this.channelsVisitor.getProcessChannels().containsKey(pname)) {
				List<String> pchannels = this.channelsVisitor.getProcessChannels().get(pname);
				
				//boolean forced = false;
				for (String cname : pchannels) {
					String cnameKey = HandelC.getInstance().getChannelName(cname);
					//if (forcedInterleavingLUT.containsKey(cnameKey) && !realSyncLUT.containsKey(cnameKey)) {
					if (forcedInterleavingLUT.containsKey(cnameKey) && !isSyncBranch(cnameKey, pbranch)) {
						insertSubstChannel(nodeId+ctxId, cname, "_BR_"+pbranch);
						//forced = true;
						forcedRenaming = true;
					} else if (forcedInterleavingLUT.containsKey(cnameKey)) {
						// cnameKey is in realSyncLUT and pbranch is in one of the sync sets
						Set <Set<Integer>> syncs = realSyncLUT.get(cnameKey);
						for (Set<Integer> sync: syncs){
							if (sync.contains(pbranch)){
								insertSubstChannel(nodeId+ctxId, cname, "_SYNC_"+sync.hashCode());
								//forced = true;
								forcedRenaming = true;
							}
						}
					}
				}
				/*if (forced){
					List<String> procs;
					if (this.forcedRenaming.containsKey(pbranch)){
						procs = this.forcedRenaming.get(pbranch);
						if (!procs.contains(pname))
							procs.add(pname);
					}else {
						procs = new Vector<String>();
						procs.add(pname);
					}
					this.forcedRenaming.put(pbranch, procs);
				}*/
			}
			
			String parent = getCurrentAnalysedProcess();
			setCurrentAnalysedProcess(pname);
			
			boolean forcedRenamingAux = this.forcedRenaming;
			this.forcedRenaming = false;
			
			//this.processCalls.push(nodeId);
			this.ctx.push(nodeId);
			PExpression proc = (PExpression)this.procsInfoVisitor.getProcDefinition(pname);
	    	proc.apply(this);
	    	this.ctx.pop();
	    	//if (this.processCalls.pop() != nodeId) {throw new java.lang.RuntimeException();}
	    	
	    	if ( this.forcedRenaming
	    			&& !this.processesRenaming.containsKey(nodeId + ctxId)
					&& !this.affectedProcessesRenaming.contains(nodeId + ctxId))
				this.affectedProcessesRenaming.add(nodeId + ctxId);
	    	this.forcedRenaming = forcedRenamingAux || this.forcedRenaming;
	    	
	    	setCurrentAnalysedProcess(parent);
		}
	}
	
	private boolean isSyncBranch (String channel, int branch){
		boolean result = false;
		Hashtable<String, Set<Set<Integer>>> realSyncLUT = this.parallelismVisitor.getRealSyncLUT();
		if (realSyncLUT.containsKey(channel)) {
			Set<Set<Integer>> syncs = realSyncLUT.get(channel);
			for (Set<Integer> sync : syncs) {
				result = result || sync.contains(branch);
			}
		}
		return result;
	}
	
	private void changeSubstitutionList(String pname, int pbranch,
			String cname, String suffix) {
		this.substVisitor.changeSubstitutions(pname, pbranch, cname, suffix);
	}
	
	private void changeSubstitutionList(String pname, int pbranch,
			String cname, List<String> list) {
		this.substVisitor.changeSubstitutions(pname, pbranch, cname, list);
	}
	
	private void insertSubstChannel(int nodeId, String cname, String suffix){
		
		updateChannels(cname, suffix);
		
		if (this.processesRenaming.containsKey(nodeId)){
			Map<String, List<String>> cRenames = this.processesRenaming.get(nodeId);
			if (cRenames.containsKey(cname)){
				List <String> cList = cRenames.get(cname);
				cList.add(cname+suffix);
				cRenames.put(cname, cList);
			} else{
				List <String> cList = new Vector<String>();
				cList.add(cname+suffix);
				cRenames.put(cname, cList);
			}
			this.processesRenaming.put(nodeId, cRenames);
		} else {
			List <String> cList = new Vector<String>();
			cList.add(cname+suffix);
			
			Map<String, List<String>> cRenames = new Hashtable<String, List<String>>();
			cRenames.put(cname, cList);
			
			this.processesRenaming.put(nodeId, cRenames);
		}
		
	}
	
	private void updateChannels(String cname, String suffix) {
		if (this.channelsVisitor.isCommChannel(cname))
			this.channelsVisitor.insertCommChannel(cname+suffix);
		if (this.channelsVisitor.isProjChannel(cname))
			this.channelsVisitor.insertProjChannel(cname+suffix);
	}

	public String getCspCode(){
		
		String substDeclaration = this.substVisitor.getChangedSubstDecl();
		
		List<String> cspDeclarations = new Vector<String>();
		
		
		
		
		// cspCode = cspCode + ;
		for (int nodeId : this.processesRenaming.keySet()) {
			//int branch = this.parallelismVisitor.getNodeBranchId(nodeId);
			int branch = this.parallelismVisitor.getCtxNodeBranchId(nodeId);
			String pname = this.processesName.get(nodeId);
			String pnameRN = pname + "_RN_" + branch;
			String cspDirectives = "";
			String cspSubstitution = "";
			
			Map<String, List<String>> substitution = this.processesRenaming.get(nodeId);
			for (String originalChannel : substitution.keySet()) {
				String action = " ";
				String realOriginalChannel;
				if (realSubstitution.containsKey(nodeId) 
						&& realSubstitution.get(nodeId).containsKey(originalChannel)){
					realOriginalChannel = realSubstitution.get(nodeId).get(originalChannel);
				} else realOriginalChannel = originalChannel;
				if (this.directives.containsProcChannelInOut(pname)){
					Hashtable<String,String> actions = this.directives.getProcChannelInOut(pname);
					if (actions.containsKey(realOriginalChannel)){
						action = action + (actions.get(realOriginalChannel).equals(HandelC.A_READ)?"in":"out");
					}
				}
				String channelType = "";
				if (this.channelsVisitor.containsChannelType(originalChannel)){
					List types = channelsVisitor.getChannelType(originalChannel);
					if (!types.isEmpty()){
						channelType = channelType + " : ";
						for (Object type:types){
							if (!channelType.equals(" : ")){
								channelType = channelType + " . ";
							}
							channelType = channelType + type.toString();
						}
					}
				}
				List<String> chanList = substitution.get(originalChannel);
				for (String replacedChannel : chanList) {
					String cspDeclaration = "channel " + replacedChannel + channelType + " \n";
					if (!cspDeclarations.contains(cspDeclaration)
							&& !this.substVisitor.isDeclaredChannel(replacedChannel))
						cspDeclarations.add(cspDeclaration);
					//cspDeclarations = cspDeclarations + "channel " + replacedChannel + channelType + " \n";
					cspDirectives = cspDirectives + "--!! channel " + replacedChannel + action  + " within " + pnameRN + " \n"; 
					if (!cspSubstitution .equals("")) cspSubstitution = cspSubstitution + " , ";
					cspSubstitution = cspSubstitution + realOriginalChannel + " <- " + replacedChannel;
					
				}
			}
			String arguments = "";
			if (this.procsInfoVisitor.containsProcArguments(pname))
				arguments = this.procsInfoVisitor.getProcArguments(pname).toString().trim();
			
			if (this.directives.containsProcArguments(pname)){
				Hashtable<String,String> args = this.directives.getProcArguments(pname);
				for (String arg : args.keySet()){
					cspDirectives = cspDirectives + 
							"--!! arg " + arg + " " + this.directives.getProcArguments(pname).get(arg) + " within " + pnameRN + " \n";
				}
			}
			
			
			cspCode = cspCode + cspDirectives;
			cspCode = cspCode + pnameRN + arguments + " = " + pname + arguments + " [[" + cspSubstitution + "]] \n\n";
		}
		
		for (String decl : cspDeclarations) {
			cspCode = decl + cspCode;
		}

		return (substDeclaration + cspCode);
	}
	
	public boolean isProcessToBeRenamed (String name){
		
		for (int id : this.getAffectedProcessesRenaming()) {
			if (name.equals(this.processesName.get(id))){
				return true;
			}
		}
		for (int id: this.processesRenaming.keySet()){
			if (name.equals(this.processesName.get(id))){
				return true;
			}
		}
		return false;
	}
	
	private boolean isMain(String pname) {
    	return HandelC.MAINP.equals(pname);
	}

	public Map<Integer, Map<String, List<String>>> getProcessesRenaming() {
		return processesRenaming;
	}

	public void setProcessesRenaming(
			Map<Integer, Map<String, List<String>>> processesRenaming) {
		this.processesRenaming = processesRenaming;
	}

	public List<Integer> getAffectedProcessesRenaming() {
		return affectedProcessesRenaming;
	}

	public void setAffectedProcessesRenaming(List<Integer> affectedProcessesRenaming) {
		this.affectedProcessesRenaming = affectedProcessesRenaming;
	}

	/*public Stack<Integer> getProcessCalls() {
		return processCalls;
	}

	public void setProcessCalls(Stack<Integer> processCalls) {
		this.processCalls = processCalls;
	}
*/
	public String getCurrentAnalysedProcess() {
		return currentAnalysedProcess;
	}

	public void setCurrentAnalysedProcess(String currentAnalysedProcess) {
		this.currentAnalysedProcess = currentAnalysedProcess;
	}

	public Map<Integer, String> getProcessesName() {
		return processesName;
	}

	public void setProcessesName(Map<Integer, String> processesName) {
		this.processesName = processesName;
	}

	public ChannelsVisitor getChannelsVisitor() {
		return channelsVisitor;
	}

}
