package csp2hc.visitors;

import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;
import java.util.Stack;
import java.util.Vector;

import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.node.ACspAbstractDefinition;
import lmf.formula.csp.node.ACspAlphabetisedParallelProcess;
import lmf.formula.csp.node.ACspAlphabetisedReplicatedProcess;
import lmf.formula.csp.node.ACspAssert;
import lmf.formula.csp.node.ACspCatSequenceExpr;
import lmf.formula.csp.node.ACspChannelDefinition;
import lmf.formula.csp.node.ACspClosedRange;
import lmf.formula.csp.node.ACspComprehensionSet;
import lmf.formula.csp.node.ACspConstantCallExpr;
import lmf.formula.csp.node.ACspConstantDefinition;
import lmf.formula.csp.node.ACspDatatypeDefinition;
import lmf.formula.csp.node.ACspDotExpr;
import lmf.formula.csp.node.ACspEvent;
import lmf.formula.csp.node.ACspExternalChoiceProcess;
import lmf.formula.csp.node.ACspExternalDefinition;
import lmf.formula.csp.node.ACspExternalReplicatedProcess;
import lmf.formula.csp.node.ACspFunctionCallExpr;
import lmf.formula.csp.node.ACspGuardedProcess;
import lmf.formula.csp.node.ACspHidingProcess;
import lmf.formula.csp.node.ACspInputComunication;
import lmf.formula.csp.node.ACspInterleaveProcess;
import lmf.formula.csp.node.ACspInterleaveReplicatedProcess;
import lmf.formula.csp.node.ACspInternalChoiceProcess;
import lmf.formula.csp.node.ACspInternalReplicatedProcess;
import lmf.formula.csp.node.ACspInterruptProcess;
import lmf.formula.csp.node.ACspLengthExpr;
import lmf.formula.csp.node.ACspLinkedParallelProcess;
import lmf.formula.csp.node.ACspLinkedReplicatedProcess;
import lmf.formula.csp.node.ACspNametypeDefinition;
import lmf.formula.csp.node.ACspOpenRange;
import lmf.formula.csp.node.ACspOutputComunication;
import lmf.formula.csp.node.ACspParameters;
import lmf.formula.csp.node.ACspPrefixProcess;
import lmf.formula.csp.node.ACspProcessCallExpr;
import lmf.formula.csp.node.ACspProcessDefinition;
import lmf.formula.csp.node.ACspProjection;
import lmf.formula.csp.node.ACspRangeCollectionElement;
import lmf.formula.csp.node.ACspRenamingProcess;
import lmf.formula.csp.node.ACspSequenceValue;
import lmf.formula.csp.node.ACspSequentialProcess;
import lmf.formula.csp.node.ACspSequentialReplicatedProcess;
import lmf.formula.csp.node.ACspSet;
import lmf.formula.csp.node.ACspSetType;
import lmf.formula.csp.node.ACspSharingParallelProcess;
import lmf.formula.csp.node.ACspSharingReplicatedProcess;
import lmf.formula.csp.node.ACspSubtypeDefinition;
import lmf.formula.csp.node.ACspTransparentDefinition;
import lmf.formula.csp.node.ACspTupleType;
import lmf.formula.csp.node.ACspTupleValue;
import lmf.formula.csp.node.ACspType;
import lmf.formula.csp.node.ACspUntimedTimeoutProcess;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.PArguments;
import lmf.formula.csp.node.PChannelList;
import lmf.formula.csp.node.PElementList;
import lmf.formula.csp.node.PEvent;
import lmf.formula.csp.node.PExpression;
import lmf.formula.csp.node.PHelp;
import lmf.formula.csp.node.PProjection;
import lmf.formula.csp.node.PTypeSpecifier;
import lmf.formula.csp.node.Start;
import lmf.formula.csp.type.CspSetType;
import lmf.formula.csp.type.CspType;
import lmf.formula.csp.type.CspUnknownType;
import mathCollection.MathSet;
import csp2hc.exception.AmbiguousBusUseException;
import csp2hc.exception.BusUseInChannelSetException;
import csp2hc.exception.BusUseInExternalChoiceException;
import csp2hc.exception.CouldNotInferTypeException;
import csp2hc.exception.NumberOfWritersException;
import csp2hc.exception.ProjectionBusUseException;
import csp2hc.exception.ReaderAndWriterParallelBranchException;
import csp2hc.exception.UndefinedBusTypeException;
import csp2hc.exception.UnsupportedConstructorException;
import csp2hc.exception.UnsupportedEmbeddedFunctionException;
import csp2hc.exception.UnusedBusException;
import csp2hc.exception.WrongBusTypeException;
import csp2hc.messages.HandelC;
import csp2hc.messages.Messenger;
import csp2hc.util.Branch2ActionSetLUT;
import csp2hc.util.BranchActionsOnChannel;
import csp2hc.util.CspUtil;
import csp2hc.util.ForcedInterleavingOnChannel;

public class ForcedInterleavingVisitor_OLD extends ExtendedDepthFirstAdapter {
    
    // Visitor containing info about types
    private DatatypesVisitor types;

    // Visitor that has the information about the processes
    private ProcessesInfoVisitor procsInfo;

    // Visitor that has the information about the parallel compositions
    private ParallelismAnalyserVisitor parallelismAnalyser;

    // Visitor that has the information about the directives
    private DirectivesVisitor directives;

    // Visitor that gets the information about the alphabets
	private AlphabetsVisitor alphabets; 

    // Visitor that gets the information about the channels
	private ChannelsVisitor channels; 

    // Visitor that gets the information about the constants and functions
	private ConstantsAndFunctionsVisitor constants; 
	
	// The name of the current definition being checked
    String currentDefinition;

	// The name of the current process being checked
    String currentAnalysedProcess;
    
    /*
     * Used in the verification of the types of communication accepted. Indicates that a 
     * communication of a value has been found and therefore, this event should neither
     * output nor synchornise in anything else.
    */ 
    boolean commPrefixFound;
    
    /*
     * Used in the verification of the types of communication accepted. Stores the current
     * prefixing being analysed
    */ 
    ACspPrefixProcess currentPrefix;

    // Stores the number of non-projections used for each channel
    @SuppressWarnings("rawtypes")
	private Hashtable channelNonProjections;

    // Counts the number of non-projections for the current channel
    private int nonProjectionCounter;

    // Counts the number of arguments left in a projection that uses datatypes tags
    private int tagsArgsLeft;
    
    // Indicates that we are checking synchronisation channel sets
    private boolean checkingChannelSet;

    // Stack of sets of events whose interleaving must be forced
    @SuppressWarnings("rawtypes")
	private Stack forcedInterleaving;

    // Set of processes names that already been analysed while 
    // checking a concurrent process
    private Set<String> checkedProcesses;

    // Set of events currently being hidden 
    private Stack<Set<String>> currentHiddenChannels;

    // Number of parallel composition we have above in the tree 
    private int parallelismCounter;
    
    // Indicates that we are checking an external choice
    private boolean checkingExternalChoice;

    // Indicates that we are checking a channel type
    private boolean checkingChannelType;

    // Indicates that we are checking a datatype definition
    private boolean checkingDatatype;

    // Indicates that we are checking a set
    private boolean checkingSet;

	private int nextBranchId;

	private Stack<Integer> maximumParallelIds;

	private Hashtable<String, List<String>> renamingProcesses;
    
    
    
    
    @SuppressWarnings("rawtypes")
	public ForcedInterleavingVisitor_OLD(DatatypesVisitor types,
    							ChannelsVisitor channels,
    							ProcessesInfoVisitor procsInfo,
    							ParallelismAnalyserVisitor parallelismAnalyser,
    							DirectivesVisitor directives,
    							ConstantsAndFunctionsVisitor constants) {
        this.types = types;
        this.procsInfo = procsInfo;
        this.parallelismAnalyser = parallelismAnalyser;
        this.directives = directives;
        this.alphabets = new AlphabetsVisitor(this.procsInfo, this.directives, constants);
        this.channels = channels;
        this.constants = constants;

        this.commPrefixFound = false;
        this.channelNonProjections = new Hashtable();
        this.tagsArgsLeft = 0;
        this.checkingChannelSet = false;
        this.checkedProcesses = new HashSet<String>();
        this.currentHiddenChannels = new Stack<Set<String>>();
        this.parallelismCounter = 0;
        this.checkingExternalChoice = false;
        this.checkingChannelType = false;
        this.checkingDatatype = false;
        this.checkingSet = false;
    }
    
    private void initialiseForMain() {
        this.nextBranchId = 0;
        this.maximumParallelIds = new Stack<Integer>();
        this.maximumParallelIds.push(new Integer(0));
    }
    
    private void setCurrentDefinition(String s) {
    	this.currentDefinition = s;
    	//this.setAlphabetsVisitorDefinition(s);
    }
    private void setCurrentAnalysedProcess(String s) {
    	this.currentAnalysedProcess = s;
    	this.setAlphabetsVisitorDefinition(s);
    }

    private void setAlphabetsVisitorDefinition(String s) {
    	this.alphabets.setCurrentAnalysedProcess(s);
    }
    
    public void caseStart(Start node){
    	
    	this.renamingProcesses = new Hashtable <String, List<String>>();
		createRenamingProcess(node);
    	   	
    	node.getPSpecification().apply(this);
    	
    	//this.checkParallelCompositionRestrictions();
    	//this.checkMultiSync();
    	//this.checkBuses();
    }
    
    private void checkParallelCompositionRestrictions() {
    	BranchActionsOnChannel branches2Actions = this.parallelismAnalyser.getBranchActionsOnChannel();
    	
    	Enumeration<String> channels = branches2Actions.keys();
    	
		// NO BRANCH IS READER AND WRITER ON THE SAME CHANNEL
    	while(channels.hasMoreElements()) {
    		String channel = channels.nextElement();
    		
    		Branch2ActionSetLUT branch2Actions = branches2Actions.get(channel);
    		
        	Enumeration<Integer> branches = branch2Actions.keys();
        	
        	boolean foundProblem = false;
        	while(branches.hasMoreElements()) {
        		Integer branch = branches.nextElement();
        		Set<String> actions = branch2Actions.get(branch);

        		if (actions.contains(HandelC.A_READ) && actions.contains(HandelC.A_WRITE)) {
    	            throw new ReaderAndWriterParallelBranchException(
    	            		HandelC.getInstance().getChannel(channel), this.currentDefinition);
        		}
        		
        	}	
    	}
    	
		// EXACTLY ONE BRANCH MUST BE A WRITER IN A PARALLEL COMPOSITION
    	// BASED ON THE PREVIOUS RESTRICTION: NO BRANCH IS READER AND WRITER OM THE SAME CHANNEL
    	Hashtable<String,Set<Set<Integer>>> realSync = this.parallelismAnalyser.getRealSyncLUT();

    	Enumeration<String> rsChannels = realSync.keys();
    	
    	while(rsChannels.hasMoreElements()) {
    		String channel = rsChannels.nextElement();
    		Set<Set<Integer>> syncs = realSync.get(channel);
    		
    		Iterator<Set<Integer>> iterator = syncs.iterator();
    		while (iterator.hasNext()) {
				Set<Integer> syncChannels = iterator.next();

				int readers = 0;
				int writers = 0;
				
	    		Iterator<Integer> channelIter = syncChannels.iterator();
	    		while (channelIter.hasNext()) {
					Integer branch = channelIter.next();
				    Set<String> actions = this.parallelismAnalyser.getBranchActionOnChannel(channel, branch);
				    
	        		if (actions.contains(HandelC.A_READ)) {
	        			readers++;
	        		} else if (actions.contains(HandelC.A_WRITE)) {
	        			writers++;
	        		}
	    		}
	        	if ( writers < 1 ) {
	        		throw new NumberOfWritersException(HandelC.getInstance().getChannel(channel),NumberOfWritersException.NONE, this.currentDefinition);
				} else if ( writers > 1 ) {
		            throw new NumberOfWritersException(HandelC.getInstance().getChannel(channel),NumberOfWritersException.MULTIPLE, this.currentDefinition);
				}
	    		
			}
    	}
    	
    }
    
    private void checkMultiSync() throws UnsupportedConstructorException{
    	if(this.parallelismAnalyser.isMultiSync()) {
    		Set<String> ms = this.parallelismAnalyser.getMultiSyncChannels();
            throw new UnsupportedConstructorException(
            		Messenger.MULTI_SYNCHRONISATION + ms, 
            		this.currentDefinition);
    	}
    }
    
    @SuppressWarnings("rawtypes")
	private void checkBuses() {
    	
    	// Projection on buses are not allowed
    	if (this.channels.getProjectionBuses().size() > 0) {
            throw new ProjectionBusUseException(this.channels.getProjectionBuses().toString());
    	}

    	// Buses used as both input and output are not allowed
    	MathSet intersection = this.channels.getInputBuses().intersection(this.channels.getOutputBuses()); 
    	if (!intersection.isEmpty()) {
            throw new AmbiguousBusUseException(intersection.toString());
    	}
    	
    	// Bus has been declared but never used
    	MathSet busesUsed = this.channels.getBuses();
    	MathSet busesDeclared = this.directives.getBuses();
    	MathSet difference = busesDeclared.symmetricDifference(busesUsed); 

    	if (!difference.isEmpty()) {
        	throw new UnusedBusException(difference.toString());
    	}
    	
    	// Buses have exactly one type
        MathSet buses = this.channels.getBuses();
        for (Object bus : buses) {
        	String busStr = (String)bus;
        	boolean contains = this.channels.containsChannelType(busStr); 

        	if (contains) {
                List types = this.channels.getChannelType(busStr);
                if (types.size() != 1) {
                	throw new WrongBusTypeException(busStr);
                }
            } else {
            	throw new UndefinedBusTypeException(busStr);
            }
    	}
    }
    
    public void caseACspSubtypeDefinition(ACspSubtypeDefinition node) {
        throw new UnsupportedConstructorException(Messenger.SUBTYPE,node.getName().toString());
    }
    public void caseACspNametypeDefinition(ACspNametypeDefinition node) {
        throw new UnsupportedConstructorException(Messenger.NAMETYPE,node.getName().toString());
    }
    public void caseACspExternalDefinition(ACspExternalDefinition node) {
        throw new UnsupportedConstructorException(Messenger.EXTERNAL,node.getNameList().toString());
    }
    public void caseACspTransparentDefinition(ACspTransparentDefinition node) {
        throw new UnsupportedConstructorException(Messenger.TRANSPARENT,node.getNameList().toString());
    }
    
    public void getRenamingChannels(Node process){
    	String pname = ((ACspProcessDefinition) process).getName().getText().trim();
    	
    	this.alphabets.resetChannels();
    	process.apply(this.alphabets);
    	Set <String> alphabet = new HashSet<String>(); 
    	alphabet.addAll(this.alphabets.getVisibleChannels());
    	alphabet.addAll(this.alphabets.getHiddenChannels());
    	
    	//alphabet.retainAll(this.parallelismAnalyser.getRealSyncLUT().keySet());
    	
    	/*for (String channel : alphabet) {
    		this.parallelismAnalyser.getRealSyncLUT().get(arg0)
			
		}*/
    	
    	
    	this.parallelismAnalyser.getNodeBranchId(process);
    	
    	this.parallelismAnalyser.getRealSyncLUT();
    	
    }
    
    public void createRenamingProcess(Node node) {
    	ForcedInterleavingOnChannel forcedInterleavingLUT = parallelismAnalyser.getForcedInterleavingLUT();
		Set <String> renamingChannels = forcedInterleavingLUT.keySet();
		
		Set <String> processes = this.channels.getProcessChannels().keySet();
		for (String process : processes){
			List<String> channelsOfProcess = this.channels.getProcessChannels().get(process);
			ListIterator<String> channelsIter = channelsOfProcess.listIterator();
			while (channelsIter.hasNext()) {
				String next = channelsIter.next();
				if (renamingChannels.contains(HandelC.getInstance().getChannelName(next))){
					List <String> chanList;
					if (!this.renamingProcesses.containsKey(process)){
						chanList = new Vector <String>();
					} else {
						chanList = this.renamingProcesses.get(process);
					}
					chanList.add(next);
					this.renamingProcesses.put(process, chanList);
				}
			}
		}
    }

    public void caseACspProcessCallExpr(ACspProcessCallExpr node) {
    	
    	String pname = node.getName().toString().trim();

        if (!pname.equals(CspUtil.getSkip()) 
        		&& !pname.equals(CspUtil.getStop())
        		&& !this.checkedProcesses.contains(pname)) {
    		
    		this.checkedProcesses.add(pname);
        	this.setCurrentAnalysedProcess(pname);
    		
        	
        	
        	
        	PExpression proc = (PExpression)this.procsInfo.getProcDefinition(pname);
        	proc.apply(this);
    		this.checkedProcesses.remove(pname);
    			
    	}
    	
    	
    	/*
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}

		PArguments arguments = node.getArguments();
		if (arguments != null) {
			arguments.apply(this);
		}
		
		String pname = node.getName().toString().trim();
        if (!pname.equals(CspUtil.getSkip()) 
        		&& !pname.equals(CspUtil.getStop())
        		&& !this.checkedProcesses.contains(pname)) {
        
    		this.checkedProcesses.add(pname);
        	PExpression proc = this.procsInfo.getProcDefinition(pname);

        	String previousAnalysedProcess = this.currentAnalysedProcess;
    		this.setCurrentAnalysedProcess(pname);
    		
        	proc.apply(this);
        	
        	this.setCurrentAnalysedProcess(previousAnalysedProcess);

        	this.checkedProcesses.remove(pname);
    	}*/

    }
    
    public void caseACspProcessDefinition(ACspProcessDefinition definition) {
        String name = definition.getName().getText().trim();
        //System.out.println("RestrictionsVisitor :: caseACspProcessDefinition -> process "+name);

        this.setCurrentDefinition(name);
		this.setCurrentAnalysedProcess(name);

		initialiseForMain();
		if (isMain()){
			PExpression expression = definition.getExpression();
			expression.apply(this);
			
		}
		
/*        ACspParameters parameters = (ACspParameters)definition.getParameters();
        if (parameters == null) {
        	parameters = new ACspParameters();
        }
*/
    }

    public void caseACspConstantCallExpr(ACspConstantCallExpr node){
    	String name = node.getName().getText().trim();
    	if (this.checkingChannelSet){
    		this.constants.getConstantsDefinitions(name).apply(this);
    	}
    }

    public void caseACspFunctionCallExpr(ACspFunctionCallExpr node){
    	
    	String fname = node.getName().getText().trim();

    	if (this.checkingChannelSet){
        	if (HandelC.getInstance().isEmbeddedFunction(fname)) {
        		node.getArguments().apply(this);
        	} else {
        		throw new UnsupportedConstructorException(Messenger.SYNCHRONISATION_CHANNEL_SET_FUNCTION,this.currentDefinition);
        	}
    	} 
    	if (this.constants.containsConstant(fname)
    			|| HandelC.getInstance().isEmbeddedFunction(fname)) {
    		node.getArguments().apply(this);
    	} else {
    		throw new UnsupportedEmbeddedFunctionException(fname);
    	}
    }
    
    public void caseACspChannelDefinition(ACspChannelDefinition node){
        PChannelList channels = node.getChannelList();
        PTypeSpecifier type = node.getTypeSpecifier();
        
        this.setCurrentDefinition(channels.toString());
        if (type != null) {
        	this.checkingChannelType = true;
            type.apply(this);
        	this.checkingChannelType = false;
        }
    }
    
    public void caseACspConstantDefinition(ACspConstantDefinition node) {
    	this.setCurrentDefinition(node.getName().toString().trim());
        node.getExpression().apply(this);
    }
  
    public void caseACspDatatypeDefinition(ACspDatatypeDefinition node){
        this.setCurrentDefinition(node.getName().toString().trim());

    	this.checkingDatatype = true;
        node.getBranchList().apply(this);
    	this.checkingDatatype = false;
        
    }
    public void caseACspAbstractDefinition(ACspAbstractDefinition definition) {
		//throw new AbstractDefinitionException();
    	String name = definition.getName().toString().trim();
        //System.out.println("RestrictionsVisitor :: caseACspProcessDefinition -> process "+name);

        this.setCurrentDefinition(name);
		this.setCurrentAnalysedProcess(name);

        ACspParameters parameters = (ACspParameters)definition.getParameters();
        PExpression expression = definition.getExpression();

        if (parameters == null) {
            parameters = new ACspParameters();
        }
        expression.apply(this);
    }  //Add - Ivan
    
    @SuppressWarnings("rawtypes")
	public void caseACspEvent(ACspEvent node) {
        String name = node.getChannelName().toString().trim();
        LinkedList projections = node.getProjection();
        LinkedList comm = node.getComunication();

        if (this.checkingChannelSet && this.channels.getBuses().contains(name)) {
        	throw new BusUseInChannelSetException(name,this.currentDefinition);
        }
        
        for (Iterator iter = projections.iterator(); iter.hasNext();) {
			PProjection element = (PProjection) iter.next();
			element.apply(this);
		}
        
        ListIterator iterator = comm.listIterator(0);

        while(iterator.hasNext()) {
            Object next = iterator.next();
            if (next instanceof ACspInputComunication) {
                ((ACspInputComunication)next).apply(this);
            } else if (next instanceof ACspOutputComunication) {
                ((ACspOutputComunication)next).apply(this);
            } else if (next instanceof ACspProjection) {
                ((ACspProjection)next).apply(this);
            }
        }
    }
    
    
    @SuppressWarnings("unchecked")
	public void caseACspPrefixProcess(ACspPrefixProcess node){

        this.currentPrefix = node;
        String channelName = ((ACspEvent)this.currentPrefix.getEvent()).getChannelName().toString().trim();

        this.nonProjectionCounter = 0;
        
        PEvent event = node.getEvent();
        event.apply(this);
        
        /*
        LinkedList projections = ((ACspEvent)event).getProjection();
        LinkedList comm = ((ACspEvent)event).getComunication();

        for (Iterator iter = projections.iterator(); iter.hasNext();) {
			PProjection element = (PProjection) iter.next();
			element.apply(this);
		}
        
        ListIterator iterator = comm.listIterator(0);

        while(iterator.hasNext()) {
            Object next = iterator.next();
            if (next instanceof ACspInputComunication) {
                ((ACspInputComunication)next).apply(this);
            } else if (next instanceof ACspOutputComunication) {
                ((ACspOutputComunication)next).apply(this);
            } else if (next instanceof ACspProjection) {
                ((ACspProjection)next).apply(this);
            }
        }
        */
        this.commPrefixFound = false;
        
        // Checking the consistent use of the channel projections
        if (this.channelNonProjections.containsKey(channelName)) {
            int counter = ((Integer)this.channelNonProjections.get(channelName)).intValue();
            if(this.nonProjectionCounter != counter) {
                //throw new InconsistentChannelProjectionUseException(channelName, this.currentDefinition); //Add - Ivan
            }
        } else {
            this.channelNonProjections.put(channelName, new Integer(this.nonProjectionCounter));
        }

        // Checking hiding restrictions

        // We are hiding channels
        if (!this.currentHiddenChannels.isEmpty()) {
        	Set<String> hiddenChannels = this.currentHiddenChannels.peek();
        	// We are not in a parallel composition
        	// and the current event is being hidden
        	if (this.parallelismCounter <=0 && hiddenChannels.contains(channelName)) {
                throw new UnsupportedConstructorException(Messenger.HIDING_CHANNEL_WITH_NO_PARALLEL_COMPOSITION_1
                										+ channelName
                										+ Messenger.HIDING_CHANNEL_WITH_NO_PARALLEL_COMPOSITION_2,this.currentDefinition);
        	}
        }

        this.checkingExternalChoice = false;

        PHelp proc = node.getProcess();
        proc.apply(this);
    }
    
    public void caseACspInputComunication(ACspInputComunication node) {
    	if (this.checkingChannelSet){
            throw new UnsupportedConstructorException(Messenger.SYNCHRONISATION_CHANNEL_SET_INPUT,this.currentDefinition);
    	} else if (this.commPrefixFound) {
            //throw new UnsupportedConstructorException(this.currentPrefix.toString(),this.currentDefinition);  //Add - Ivan
        } else {
        	this.nonProjectionCounter++;
            this.commPrefixFound = true;
            node.getParam().apply(this);
        }
    }

    public void caseACspOutputComunication(ACspOutputComunication node) {
    	if (this.checkingChannelSet){
            throw new UnsupportedConstructorException(Messenger.SYNCHRONISATION_CHANNEL_SET_OUTPUT,this.currentDefinition);
    	} else if (this.commPrefixFound) {
            //throw new UnsupportedConstructorException(this.currentPrefix.toString(),this.currentDefinition);  //Add - Ivan
        } else {
        	this.nonProjectionCounter++;
            this.commPrefixFound = true;
            node.getExpression().apply(this);
        }
    }

    public void caseACspProjection(ACspProjection node) {
    	if (this.checkingChannelSet){
            throw new UnsupportedConstructorException(Messenger.SYNCHRONISATION_CHANNEL_SET_PROJECTION,this.currentDefinition);
    	} else {
	        if (this.tagsArgsLeft > 0) {
	            this.tagsArgsLeft--;
	        } else {
	            if (this.commPrefixFound) {
	                //throw new UnsupportedConstructorException(this.currentPrefix.toString(),this.currentDefinition);  //Add - Ivan
	            }
	        }
	        node.getCallExpr().apply(this);
    	}
    }
    
    @SuppressWarnings("rawtypes")
	public void caseACspDotExpr(ACspDotExpr node){
        if (node.getCallExpr() instanceof ACspConstantCallExpr) {
        	ACspConstantCallExpr consCall = (ACspConstantCallExpr) node.getCallExpr();
        	String name = consCall.getName().toString().trim();
            if (this.types.isTag(name)) {
                this.tagsArgsLeft = this.tagsArgsLeft + this.types.getTagValues(name).size();
            } else {
            	consCall.apply(this);
            }
        }
        
        List projections = node.getProjection();
        for (Iterator iter = projections.iterator(); iter.hasNext();) {
			PProjection element = (PProjection) iter.next();
			element.apply(this);
		}
    }

    public void caseACspSharingParallelProcess(ACspSharingParallelProcess node) {
        this.checkConcurrency(node);
    }
    
    public void caseACspAlphabetisedParallelProcess(ACspAlphabetisedParallelProcess node) {
        this.checkConcurrency(node);
    }
    
    public void caseACspInterleaveProcess(ACspInterleaveProcess node) {
        this.checkConcurrency(node);
    }

    public void caseACspHidingProcess(ACspHidingProcess node) {
    	
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}

		// Verifiy the channel set
        this.checkingChannelSet = true;
        
        
        Node hiddenChanExp = ((ACspHidingProcess)node).getHidenEvents();

        System.out.println("RestrictionsVisitor :: caseACspHidingProcess -> Verifying "+hiddenChanExp);
        
        hiddenChanExp.apply(this);
        this.checkingChannelSet = false;

        // Getting used, hidden channels in process
        System.out.println("RestrictionsVisitor :: caseACspHidingProcess -> Getting used, hidden channels in process "+hiddenChanExp);

        Node process = ((ACspHidingProcess)node).getProcess();
        process.apply(alphabets);
        Set processReferencedChannels = alphabets.getReferencedChannels();
        Set processWrittenChannels = alphabets.getWrittenChannels();
        Set processReadChannels = alphabets.getReadChannels();
        Set processHiddenChannels = alphabets.getHiddenChannels();
        Set processVisibleChannels = alphabets.getVisibleChannels();
        alphabets.resetChannels();
        
        // Getting hidden channels
        Set<String> hiddenChannels = alphabets.analyseChannelSet(hiddenChanExp);

        hiddenChannels.retainAll(processReferencedChannels);

        if (hiddenChannels.isEmpty()) {
            // If hidden channels are not used in the process - ignore it
            process.apply(this);
        } else {
            // Update the current hidden events
            if (!this.currentHiddenChannels.isEmpty()) {
            	hiddenChannels.addAll(this.currentHiddenChannels.peek());
            }        	
            this.currentHiddenChannels.push(hiddenChannels);

            // Verifiy the remaing process
            process.apply(this);
            
            this.currentHiddenChannels.pop();
        }
    }
    
    @SuppressWarnings({ "rawtypes", "unchecked" })
	private void checkConcurrency(Node node) {
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}

		PExpression leftExp;
        PExpression rightExp;

        Set<String> syncSet = null;
        Node left = null;
        Node right = null;
        
        if (node instanceof ACspSharingParallelProcess) {
            syncSet = this.parallelismAnalyser.getSyncSet((ACspSharingParallelProcess)node);
            left = ((ACspSharingParallelProcess)node).getLeftProcess();
            right = ((ACspSharingParallelProcess)node).getRightProcess();
    	} else if (node instanceof ACspAlphabetisedParallelProcess) {
            syncSet = this.parallelismAnalyser.getSyncSet((ACspAlphabetisedParallelProcess)node);
            left = ((ACspAlphabetisedParallelProcess)node).getLeftProcess();
            right = ((ACspAlphabetisedParallelProcess)node).getRightProcess();
    	} else if (node instanceof ACspInterleaveProcess) {
            syncSet = this.parallelismAnalyser.getSyncSet((ACspInterleaveProcess)node);
            left = ((ACspInterleaveProcess)node).getLeftProcess();
            right = ((ACspInterleaveProcess)node).getRightProcess();
    	} else {
    		throw new UnsupportedConstructorException(Messenger.UNKNOWN_PARALLEL_COMPOSITION, this.currentDefinition);
        } 

        // Checking restrictions on the synchronisation channel set
        this.checkingChannelSet = true;
        if (node instanceof ACspSharingParallelProcess) {
        	leftExp = ((ACspSharingParallelProcess)node).getExpression();
        	leftExp.apply(this);
    	} else if (node instanceof ACspAlphabetisedParallelProcess) {
    		leftExp = ((ACspAlphabetisedParallelProcess)node).getLeftExpr();
    		leftExp.apply(this);
    		rightExp = ((ACspAlphabetisedParallelProcess)node).getRightExpr();
    		rightExp.apply(this);
    	} 
        this.checkingChannelSet = false;
        
        alphabets.resetChannels();

        // Retrieve info about channels usage on each branch
        left.apply(alphabets);
        Set leftReferencedChannels = alphabets.getReferencedChannels();
        Set leftHiddenChannels = alphabets.getHiddenChannels();
        Set leftVisibleChannels = alphabets.getVisibleChannels();
        Set leftWrittenChannels = alphabets.getWrittenChannels();
        Set leftReadChannels = alphabets.getReadChannels();
        alphabets.resetChannels();

        right.apply(alphabets);
        Set rightReferencedChannels = alphabets.getReferencedChannels();
        Set rightHiddenChannels = alphabets.getHiddenChannels();
        Set rightVisibleChannels = alphabets.getVisibleChannels();
        Set rightWrittenChannels = alphabets.getWrittenChannels();
        Set rightReadChannels = alphabets.getReadChannels();
        alphabets.resetChannels();

        //////////////////////
        // HIDING RESTRICTIONS
        //////////////////////
        
        // There is no channel that is hidden in the left, in CS, and used in the right
        leftHiddenChannels.retainAll(syncSet);
        leftHiddenChannels.retainAll(rightReferencedChannels);
        if (!leftHiddenChannels.isEmpty()) {
            throw new UnsupportedConstructorException(
            		Messenger.HIDING_CHANNEL_WITH_COMMUNICATION_1
					+ leftHiddenChannels
					+ Messenger.HIDING_CHANNEL_WITH_COMMUNICATION_2
					+ syncSet
					+ Messenger.HIDING_CHANNEL_WITH_COMMUNICATION_3,this.currentDefinition);
        }

        // There is no channel that is hidden in the right, in CS, and used in the left
        rightHiddenChannels.retainAll(syncSet);
        rightHiddenChannels.retainAll(leftReferencedChannels);
        if (!rightHiddenChannels.isEmpty()) {
            throw new UnsupportedConstructorException(
            		Messenger.HIDING_CHANNEL_WITH_COMMUNICATION_1
					+ rightHiddenChannels
					+ Messenger.HIDING_CHANNEL_WITH_COMMUNICATION_2
					+ syncSet
					+ Messenger.HIDING_CHANNEL_WITH_COMMUNICATION_3,this.currentDefinition);
        }
        
        // S1 = (Visible(P) \cap Written(P)) \cap CS \cap (Visible(Q) \cap Read(Q))
        // Visible written channels that are on the left, on CS, and visible and read on the right
        leftWrittenChannels.retainAll(leftVisibleChannels);
        leftWrittenChannels.retainAll(syncSet);
        leftWrittenChannels.retainAll(rightVisibleChannels);
        leftWrittenChannels.retainAll(rightReadChannels);
        
        // S2 = (Visible(Q) \cap Written(Q)) \cap CS \cap (Visible(P) \cap Read(P))
        // Visible written channels that are on the right, on CS, and visible and read on the left
        rightWrittenChannels.retainAll(rightVisibleChannels);
        rightWrittenChannels.retainAll(syncSet);
        rightWrittenChannels.retainAll(leftVisibleChannels);
        rightWrittenChannels.retainAll(leftReadChannels);
        
        // CCS \cap (Alpha(P) \cup Alpha(Q))
        // Current hiding must be either in S1 or S2
        Set<String> hiddenChannels = new HashSet<String>();
        if (!this.currentHiddenChannels.isEmpty()) {
        	hiddenChannels.addAll(this.currentHiddenChannels.peek());
        }
        leftReferencedChannels.addAll(rightReferencedChannels);
        hiddenChannels.retainAll(leftReferencedChannels);

        // S1 \cup S2
        leftWrittenChannels.addAll(rightWrittenChannels);
        
        // For all effectively hidden channel there must have a communication
        // between both branches; otherwise, error.
        // (CCS \cap Alpha(P) \cap Alpha(Q)) \subseteq S1 \cup S2
        if (!leftWrittenChannels.containsAll(hiddenChannels)) {
        	hiddenChannels.removeAll(leftWrittenChannels); 
            throw new UnsupportedConstructorException(Messenger.HIDING_CHANNEL_WITH_NO_COMMUNICATION_1
            										+ hiddenChannels
            										+ Messenger.HIDING_CHANNEL_WITH_NO_COMMUNICATION_2,this.currentDefinition);
        } else {
            // Go on checking, but ignore the hiding
            if (!this.currentHiddenChannels.isEmpty()) {
                Set<String> oldHiddenChannels = this.currentHiddenChannels.pop();
            	this.analyseBothBranches(left, right);
            	this.currentHiddenChannels.push(oldHiddenChannels);
            } else {
            	this.analyseBothBranches(left, right);
            }
        }
    }

    private void analyseBothBranches(Node left, Node right) {
        this.parallelismCounter++;
        left.apply(this);
        right.apply(this);
        this.parallelismCounter--;
    }
    
    @SuppressWarnings("rawtypes")
	public void caseACspExternalChoiceProcess(ACspExternalChoiceProcess node) {
        // Verify that there are only prefixed processes in the choice
    	
        InitialsExtractorVisitor initialsExtractor = new InitialsExtractorVisitor();
        node.apply(initialsExtractor);

        List keys = initialsExtractor.keyList();

        Vector inputVars = new Vector();
        Vector inputComm = new Vector(); 
        Vector outputComm = new Vector(); 
        
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            String channelName = ((ACspEvent)event).getChannelName().toString().trim();

            // Verify that buses are not used in external choices
            if (this.channels.getBuses().contains(channelName)) {
                throw new BusUseInExternalChoiceException(channelName,this.currentDefinition);
            } 
            
            LinkedList comm = initialsExtractor.getCommunication(event);

            ListIterator iterator = comm.listIterator(0);

            while(iterator.hasNext()) {
                Object next = iterator.next();

                // Verify that there are no two input variables with the same name
                if (next instanceof ACspInputComunication) {
                    String var = ((ACspInputComunication)next).getParam().toString();

                    if (inputVars.contains(var)) {
                        throw new UnsupportedConstructorException(Messenger.INPUT_PREFIX_SAME_VAR_NAME + var,this.currentDefinition);
                    } else {
                        inputVars.add(var);
                    }
                
                }

                // Verify that there are no output on multi-synchronised channels
                if (next instanceof ACspOutputComunication) {
                    if (this.parallelismAnalyser.isMultiSyncChannel(channelName)) {
                        throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_WITH_OUTPUT_ON_MULTISYNC + channelName,this.currentDefinition);
                    }
                }
            }
        }
        
        this.checkingExternalChoice = true;
        node.getLeftProcess().apply(this);
        this.checkingExternalChoice = true;
        node.getRightProcess().apply(this);
    }

    public void caseACspGuardedProcess(ACspGuardedProcess node) {
		/*
    	if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}
		*/
    	node.getProcess().apply(this);
    }

    public void caseACspRenamingProcess(ACspRenamingProcess node) {
        throw new UnsupportedConstructorException(Messenger.RENAMING,this.currentDefinition);
        /*
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}
        */
    }

    public void caseACspRangeCollectionElement(ACspRangeCollectionElement node) {
    	if (!checkingChannelType && !checkingDatatype) {
    		throw new UnsupportedConstructorException(Messenger.RANGE,this.currentDefinition);
    	}
    }
    
    public void caseACspLinkedParallelProcess(ACspLinkedParallelProcess node) {
        throw new UnsupportedConstructorException(Messenger.LINKED_PARALLEL,this.currentDefinition);
        /*
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}
        */
    }

    public void caseACspInterruptProcess(ACspInterruptProcess node) {
        throw new UnsupportedConstructorException(Messenger.INTERRUPT,this.currentDefinition);
        /*
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}
        */
    }
    
    public void caseACspInternalChoiceProcess(ACspInternalChoiceProcess node) {
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}
    }

    public void caseACspUntimedTimeoutProcess(ACspUntimedTimeoutProcess node) {
        throw new UnsupportedConstructorException(Messenger.UNTIMED_TIMEOUT,this.currentDefinition);
        /*
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}
        */
    }

    public void caseACspSequentialProcess(ACspSequentialProcess node) {
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}
    }
    
    public void caseACspSequentialReplicatedProcess(ACspSequentialReplicatedProcess node) {
        throw new UnsupportedConstructorException(Messenger.REPLICATED_SEQUENTIAL_COMPOSITION,this.currentDefinition);
        /*
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}
        */
    }
    
    public void caseACspExternalReplicatedProcess(ACspExternalReplicatedProcess node) {
        throw new UnsupportedConstructorException(Messenger.REPLICATED_EXTERNAL_CHOICE,this.currentDefinition);
        /*
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}
        */
    }

    public void caseACspInternalReplicatedProcess(ACspInternalReplicatedProcess node) {
        throw new UnsupportedConstructorException(Messenger.REPLICATED_INTERNAL_CHOICE,this.currentDefinition);
        /*
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}
        */
    }

    public void caseACspInterleaveReplicatedProcess(ACspInterleaveReplicatedProcess node) {
        throw new UnsupportedConstructorException(Messenger.REPLICATED_INTERLEAVE,this.currentDefinition);
        /*
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}
        */
    }

    public void caseACspSharingReplicatedProcess(ACspSharingReplicatedProcess node) {
        throw new UnsupportedConstructorException(Messenger.REPLICATED_SHARING,this.currentDefinition);
        /*
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}
        */
    }

    public void caseACspAlphabetisedReplicatedProcess(ACspAlphabetisedReplicatedProcess node) {
        throw new UnsupportedConstructorException(Messenger.REPLICATED_ALPHABETISED_PARALLEL,this.currentDefinition);
        /*
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}
        */
    }

    public void caseACspLinkedReplicatedProcess(ACspLinkedReplicatedProcess node) {
        throw new UnsupportedConstructorException(Messenger.REPLICATED_LINKED_PARALLEL,this.currentDefinition);
        /*
		if (this.checkingExternalChoice) {
			throw new UnsupportedConstructorException(Messenger.EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES,this.currentDefinition);
		}
        */
    }
    
    
    // Tuple
    public void caseACspTupleValue(ACspTupleValue node) {
        throw new UnsupportedConstructorException(Messenger.TUPLE_VALUE,this.currentDefinition);
    }
    public void caseACspTupleType(ACspTupleType node) {
        throw new UnsupportedConstructorException(Messenger.TUPLE_TYPE,this.currentDefinition);
    }

    // Sets
    public void caseACspComprehensionSet(ACspComprehensionSet node) {
        throw new UnsupportedConstructorException(Messenger.SET_COMPREHENSION,this.currentDefinition);
        /*
    	if (this.checkingSyncSet){
            throw new UnsupportedConstructorException(Messenger.SYNCHRONISATION_CHANNEL_SET_COMPREHENSION,this.currentDefinition);
    	}
    	*/
    }
    public void caseACspClosedRange(ACspClosedRange node) {
        throw new UnsupportedConstructorException(Messenger.CLOSED_RANGE,this.currentDefinition);
    }
    public void caseACspOpenRange(ACspOpenRange node) {
        throw new UnsupportedConstructorException(Messenger.OPEN_RANGE,this.currentDefinition);
    }
    public void caseACspSetType(ACspSetType node){
    	MaximalTypesInferenceVisitor inference = new MaximalTypesInferenceVisitor();
    	node.getSet().apply(inference);
    	CspType type = inference.getType();

    	if (type instanceof CspUnknownType) {
    		throw new CouldNotInferTypeException(node.toString(),this.currentDefinition.toString());
    	} else if (type instanceof CspSetType) {
    		// It is a set type, checking elements
    		CspType elementsType = ((CspSetType)type).getElemType();
    		
        	if (elementsType instanceof CspUnknownType) {
        		throw new CouldNotInferTypeException(node.toString(),this.currentDefinition.toString());
        	} else if (elementsType instanceof CspSetType) {
        		// Elements of the set are set, this is only allowed for channel declaration
        		if (this.checkingChannelType) {
        			// Confirm that they are not sets
        			CspType elementsType2 = ((CspSetType)elementsType).getElemType();
            		
                	if (elementsType2 instanceof CspUnknownType) {
                		throw new CouldNotInferTypeException(node.toString(),this.currentDefinition.toString());
                	} else if (elementsType2 instanceof CspSetType) {
                        throw new UnsupportedConstructorException(Messenger.TYPE_SETS_OF_SETS_CHANNELS,this.currentDefinition);
                	}
        		} else if (this.checkingDatatype) {
                    throw new UnsupportedConstructorException(Messenger.TYPE_SETS_OF_SETS_DATATYPES,this.currentDefinition);
        		}
        	} else {
        		node.getSet().apply(this);
        	}
    	}
    	MaximalTypesInferenceVisitor.applyToMaximalType(this, node);
    }
    public void caseACspSet(ACspSet node) {
    	if (this.checkingSet) {
            throw new UnsupportedConstructorException(Messenger.EXPRESSION_SETS_OF_SETS,this.currentDefinition);
    	} else {
    		this.checkingSet = true;
    		PElementList elements = node.getElementList();
    		if (elements != null) {
    			elements.apply(this);
    		}
    		this.checkingSet = false;
    	}
    }

    public void caseACspType(ACspType node){
    	String typeName = node.getName().getText(); 
        this.checkType(typeName);
    }

    private void checkType(String typeName) {
    	if (HandelC.getInstance().isSetType(typeName)) {
    		typeName = HandelC.getInstance().getSetType(typeName);
    	}

    	if (!typeName.equals(CspUtil.getBoolType())
    			&& !typeName.equals(CspUtil.getIntType())
    			&& !this.types.isDatatype(typeName)
    			&& !this.constants.containsConstant(typeName) ) {
            throw new UnsupportedConstructorException(Messenger.INVALID_TYPE_REFERENCE,this.currentDefinition.toString());
    	}
    }
    
    // Sequences
    public void caseACspSequenceValue(ACspSequenceValue node) {
        throw new UnsupportedConstructorException(Messenger.SEQUENCE_VALUE,this.currentDefinition);
    }
    public void caseACspCatSequenceExpr(ACspCatSequenceExpr node) {
        throw new UnsupportedConstructorException(Messenger.SEQUENCE_CONCATENATION,this.currentDefinition);
    }
    public void caseACspLengthExpr(ACspLengthExpr node) {
        throw new UnsupportedConstructorException(Messenger.SEQUENCE_LENGTH,this.currentDefinition);
    }

    public void caseACspAssert(ACspAssert node) {
    }
    
    public boolean isMain(){
    	return HandelC.MAINP.equals(this.currentDefinition);
    }
    
}

