package csp2hc.visitors;

import java.util.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.analysis.*;

import csp2hc.util.InsertionOrderHashMap;

public class InitialsExtractorVisitor extends ExtendedDepthFirstAdapter {
    private static int COMMUNICATION_INDEX = 0;
    private static int PROJECTION_INDEX = 1;
    private static int PROCESS_INDEX = 2;
    private static int GUARD_INDEX = 3;
    
    private PExpr currentGuard;
    
    private InsertionOrderHashMap initials;
    
    public InitialsExtractorVisitor() {
    	this.currentGuard = null;
        this.initials = new InsertionOrderHashMap();
    }

    public List keyList() { return this.initials.keyList(); }
    public LinkedList getCommunication(PEvent event) {
        return (LinkedList)((List)this.initials.get(event)).get(COMMUNICATION_INDEX);
    }
    public LinkedList getCommunication(String channel) {
    	LinkedList comm = new LinkedList();
        List keys = this.keyList();        
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            String name = ((ACspEvent)event).getChannelName().toString().trim();
            if (channel.equals(name)) {
            	comm = (LinkedList)((List)this.initials.get(event)).get(COMMUNICATION_INDEX); 
            }
        }
        return comm;
    }

    public LinkedList getProjection(PEvent event) {
        return (LinkedList)((List)this.initials.get(event)).get(PROJECTION_INDEX);
    }
    public LinkedList getProjection(String channel) {
    	LinkedList proj = new LinkedList();
        List keys = this.keyList();        
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            String name = ((ACspEvent)event).getChannelName().toString().trim();
            if (channel.equals(name)) {
            	proj = (LinkedList)((List)this.initials.get(event)).get(PROJECTION_INDEX); 
            }
        }
        return proj;
    }

    public PHelp getProcess(PEvent event) {
    	PHelp p = (PHelp)((List)this.initials.get(event)).get(PROCESS_INDEX);
    	return p;
    }
    
    public PExpr getGuard(PEvent event) {
    	PExpr g = (PExpr)((List)this.initials.get(event)).get(GUARD_INDEX);
    	return g;
    }
    
    // returns the keys on the channels belonging to the given list of names
    public List<String> getKeyNamesInList(List names) {
    	List keysInList = new Vector();
    	
        List keys = this.keyList();        
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            String name = ((ACspEvent)event).getChannelName().toString().trim();
            if (names.contains(name)) {
            	keysInList.add(name);
            }
        }
    	return keysInList;
    }

    // returns the keys on th channels not belongging to the given list of names
    public List<String> getKeyNamesNotInList(List names) {
    	List<String> keysNotInList = new Vector<String>();
    	
        List keys = this.keyList();        
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            String name = ((ACspEvent)event).getChannelName().toString().trim();
            if (!names.contains(name)) {
            	keysNotInList.add(name);
            }
        }
    	return keysNotInList;
    }

    // returns the name of the channels used as keys 
    public List<String> getKeyNames() {
    	return this.getKeyNamesNotInList(new Vector<String>());
    }
    
    // returns the name of the channels used as keys 
    public List<PEvent> getGuardedKeyList() {
    	List guardedKeys = new Vector();
    	
    	List keys = this.keyList();        
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            if (((List)this.initials.get(event)).get(GUARD_INDEX) != null) {
            	//String name = ((ACspEvent)event).getChannelName().toString().trim();
            	guardedKeys.add(event);
            }
        }
        return guardedKeys;
    }
    
    // returns the name of the channels used as keys 
    public List<PEvent> getNotGuardedKeyList() {
    	List notGuardedKeys = new Vector();
    	
    	List keys = this.keyList();        
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            if (((List)this.initials.get(event)).get(GUARD_INDEX) == null) {
            	//String name = ((ACspEvent)event).getChannelName().toString().trim();
            	notGuardedKeys.add(event);
            }
        }
        return notGuardedKeys;
    }
    
    public void caseACspPrefixProcess(ACspPrefixProcess node){
        PEvent event = node.getEvent();
        LinkedList comm = ((ACspEvent)node.getEvent()).getComunication();
        LinkedList proj = ((ACspEvent)node.getEvent()).getProjection();
        PHelp proc = node.getProcess();
        
        Vector data = new Vector();
        data.add(COMMUNICATION_INDEX,comm);
        data.add(PROJECTION_INDEX,proj);
        data.add(PROCESS_INDEX,proc);
        data.add(GUARD_INDEX,this.currentGuard);
        this.currentGuard = null;
        this.initials.put(event,data);
    }
    
/*    public void caseACspBasePrefixProcess(ACspBasePrefixProcess node){
        
    }*/
    
    public void caseACspGuardedProcess(ACspGuardedProcess node){
        this.currentGuard = node.getGuardExpr();
    	node.getProcess().apply(this);
    }
    
    public void caseACspAssert(ACspAssert node) {
    }
    
}
