package csp2hc.visitors;

import csp2hc.exception.CouldNotInferTypeException;
import csp2hc.exception.UndefinedArgumentSetTypeException;
import csp2hc.exception.UndefinedArgumentTypeException;
import csp2hc.exception.UndefinedEmbeddedFunctionTypeException;
import csp2hc.exception.UnsupportedConstructorException;
import csp2hc.messages.HandelC;
import csp2hc.messages.Messenger;
import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.node.ACspAbstractType;
import lmf.formula.csp.node.ACspAddMathExpr;
import lmf.formula.csp.node.ACspAndLogicalExpr;
import lmf.formula.csp.node.ACspBaseArgumentList;
import lmf.formula.csp.node.ACspCollectionElement;
import lmf.formula.csp.node.ACspConstantCallExpr;
import lmf.formula.csp.node.ACspDiffMathExpr;
import lmf.formula.csp.node.ACspDivMath1Expr;
import lmf.formula.csp.node.ACspEqualRelationalExpr;
import lmf.formula.csp.node.ACspFunctionCallExpr;
import lmf.formula.csp.node.ACspGtRelationalExpr;
import lmf.formula.csp.node.ACspGteRelationalExpr;
import lmf.formula.csp.node.ACspIfExpr;
import lmf.formula.csp.node.ACspLinkedArgumentList;
import lmf.formula.csp.node.ACspLinkedElementList;
import lmf.formula.csp.node.ACspLtRelationalExpr;
import lmf.formula.csp.node.ACspLteRelationalExpr;
import lmf.formula.csp.node.ACspMinusExpr;
import lmf.formula.csp.node.ACspMulMath1Expr;
import lmf.formula.csp.node.ACspNotEqualRelationalExpr;
import lmf.formula.csp.node.ACspNotExpr;
import lmf.formula.csp.node.ACspNumberValue;
import lmf.formula.csp.node.ACspOrLogicalExpr;
import lmf.formula.csp.node.ACspProjection;
import lmf.formula.csp.node.ACspRangeCollectionElement;
import lmf.formula.csp.node.ACspRemainderMath1Expr;
import lmf.formula.csp.node.ACspSet;
import lmf.formula.csp.node.ACspSetType;
import lmf.formula.csp.node.ACspTupleValue;
import lmf.formula.csp.node.ACspType;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.PCollectionElement;
import lmf.formula.csp.node.PElementList;
import lmf.formula.csp.node.PType;
import lmf.formula.csp.node.TName;
import lmf.formula.csp.type.CspBoolType;
import lmf.formula.csp.type.CspDataType;
import lmf.formula.csp.type.CspFunctionType;
import lmf.formula.csp.type.CspIntRangeType;
import lmf.formula.csp.type.CspIntType;
import lmf.formula.csp.type.CspSetType;
import lmf.formula.csp.type.CspType;
import lmf.formula.csp.type.CspUnknownType;

public class MaximalTypesInferenceVisitor extends ExtendedDepthFirstAdapter {

    // Type of expression
    private CspType type;

    public MaximalTypesInferenceVisitor() {
    	this.setType(new CspUnknownType());
    }

    public CspType getType() { return this.type; }
    private void setType(CspType newType) { this.type = newType; }

    // Booleans
    public void caseACspAndLogicalExpr(ACspAndLogicalExpr node) {
    	this.setType(new CspBoolType());
    }

    public void caseACspOrLogicalExpr(ACspOrLogicalExpr node) {
    	this.setType(new CspBoolType());
    }
    
    public void caseACspNotExpr(ACspNotExpr node) {
    	this.setType(new CspBoolType());
    }
    
    // Numbers 

    public void caseACspNumberValue(ACspNumberValue node) {
    	this.setType(new CspIntType());
    }

    public void caseACspAddMathExpr(ACspAddMathExpr node) {
    	this.setType(new CspIntType());
    }

    public void caseACspDiffMathExpr(ACspDiffMathExpr node) {
    	this.setType(new CspIntType());
    }

    public void caseACspMulMath1Expr(ACspMulMath1Expr node) {
    	this.setType(new CspIntType());
    }

    public void caseACspDivMath1Expr(ACspDivMath1Expr node) {
    	this.setType(new CspIntType());
    }

    public void caseACspRemainderMath1Expr(ACspRemainderMath1Expr node) {
    	this.setType(new CspIntType());
    }
    
    public void caseACspMinusExpr(ACspMinusExpr node) {
    	this.setType(new CspIntType());
    }

    // Relations
    public void caseACspEqualRelationalExpr(ACspEqualRelationalExpr node) {
    	this.setType(new CspBoolType());
    }
    
    public void caseACspNotEqualRelationalExpr(ACspNotEqualRelationalExpr node) {
    	this.setType(new CspBoolType());
    }

    public void caseACspGtRelationalExpr(ACspGtRelationalExpr node) {
    	this.setType(new CspBoolType());
    }

    public void caseACspGteRelationalExpr(ACspGteRelationalExpr node) {
    	this.setType(new CspBoolType());
    }
    
    public void caseACspLtRelationalExpr(ACspLtRelationalExpr node) {
    	this.setType(new CspBoolType());
    }

    public void caseACspLteRelationalExpr(ACspLteRelationalExpr node) {
    	this.setType(new CspBoolType());
    }
    
    
    // Tuple

    // Sets
    public void caseACspSet(ACspSet node) {
    	PElementList elements = node.getElementList();
    	if (elements != null) {
			elements.apply(this);
	        CspType elementsType = this.getType();
	        this.setType(new CspSetType(elementsType));
		} else {
			this.setType(new CspSetType(new CspUnknownType()));
		}
    }
    
    public void caseACspLinkedElementList(ACspLinkedElementList node) {
        PCollectionElement element = node.getCollectionElement();
		
		element.apply(this);
        if (this.getType() instanceof CspUnknownType) {
			node.getNext().apply(this);
        }
    }

    public void caseACspRangeCollectionElement(ACspRangeCollectionElement node) {
		this.setType(new CspIntRangeType());
    }
    
    public void caseACspCollectionElement(ACspCollectionElement node) {
		node.getMathExpr().apply(this);
    }
    
    public void caseACspConstantCallExpr(ACspConstantCallExpr node){
    	
    	//System.out.println("MaximalTypesInference :: caseACspConstantCallExpr -> Node "+node);
    	
    	CspType nodeType = node.getCspType(); 
    	if (nodeType instanceof CspFunctionType) {
    		nodeType = ((CspFunctionType)nodeType).getReturnType();
    	}

    	//////////////////////////////////////////////////////
    	// This piece of code is a solution for the cases
    	// in which the type of the set elements was
    	// determined as typeof(var)
    	// This is AN ERROR in the type CHECKER
    	//////////////////////////////////////////////////////
    	String nodeTypeStr = nodeType.toString();
    	if (HandelC.getInstance().isDeclarationLink(nodeTypeStr)) {
    		nodeType = new CspType(HandelC.getInstance().getTypeFromDeclarationLink(nodeTypeStr));
    	} else if (nodeType instanceof CspSetType) {
    		CspType elemType = ((CspSetType)nodeType).getElemType();
        	String elemTypeStr = elemType.toString();
        	if (HandelC.getInstance().isDeclarationLink(elemTypeStr)) {
            	//System.out.println("MaximalTypesInference :: caseACspConstantCallExpr -> Type "+elemType);
            	//System.out.println("MaximalTypesInference :: caseACspConstantCallExpr -> Type Class "+elemType.getClass().getName());
        		elemType = new CspType(HandelC.getInstance().getTypeFromDeclarationLink(elemTypeStr));
        		nodeType = new CspSetType(elemType);
        	}
    	}
    	//////////////////////////////////////////////////////
    	//////////////////////////////////////////////////////
    	
    	//System.out.println("MaximalTypesInference :: caseACspConstantCallExpr -> Type "+nodeType);
    	this.setType(nodeType);
    }

    public void caseACspProjection(ACspProjection node) {
    }
    
    public void caseACspIfExpr(ACspIfExpr node){
        node.getThenExpr().apply(this);
        if (this.getType() instanceof CspUnknownType) {
            node.getElseExpr().apply(this);
        }
    }

    public void caseACspFunctionCallExpr(ACspFunctionCallExpr node){
    	String fname = node.getName().getText().trim();
    	
    	if (HandelC.getInstance().isEmbeddedFunction(fname)) {
    		if (fname.equals(HandelC.CSP_SET_UNION)
    				|| fname.equals(HandelC.CSP_SET_DIFF)
    				|| fname.equals(HandelC.CSP_SET_INTER)) {
    			
    			// Attention: for other functions, arguments may be NULL
    			node.getArguments().apply(this);
    		} else {
    			throw new UndefinedEmbeddedFunctionTypeException(fname);
    		}
    	} else {
        	this.setType(node.getCspType());
    	}
    }	

    public void caseACspBaseArgumentList(ACspBaseArgumentList node){
        node.getArgument().apply(this);
    }	
    public void caseACspLinkedArgumentList(ACspLinkedArgumentList node){
        node.getArgument().apply(this);
        if (this.getType() instanceof CspUnknownType) {
            node.getNext().apply(this);
        }
    }	

    public void caseACspAbstractType(ACspAbstractType node) {
    	ACspFunctionCallExpr fcNode = new ACspFunctionCallExpr(node.getName(), node.getArguments());
    	fcNode.apply(this);
    }
    
    public static void applyToMaximalType(ExtendedDepthFirstAdapter visitor, Node expression){
    	String expressionStr = expression.toString();
    	
    	MaximalTypesInferenceVisitor inference = new MaximalTypesInferenceVisitor();
    	expression.apply(inference);
    	CspType cspType = inference.getType();
    	
    	MaximalTypesInferenceVisitor.applyToMaximalType(visitor, cspType, expressionStr);
    }

    private static void applyToMaximalType(ExtendedDepthFirstAdapter visitor, CspType cspType, String expression){
    	if (cspType != null) {
        	String typeName = cspType.toString().trim(); 
        	if (cspType instanceof CspSetType) {
        		CspType elementsType = ((CspSetType)cspType).getElemType();
        		if (elementsType instanceof CspUnknownType) {
        			throw new CouldNotInferTypeException(expression);
            	} else {
            		typeName = elementsType.toString().trim();
            	}
        	}
        	ACspType aCspType = new ACspType(new TName(typeName));
        	aCspType.apply(visitor);
    	} else {
			throw new CouldNotInferTypeException(expression);
    	}
    }

    
}
