package csp2hc.visitors;

import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;
import java.util.Stack;
import java.util.Vector;

import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.node.ACspAlphabetisedParallelProcess;
import lmf.formula.csp.node.ACspAssert;
import lmf.formula.csp.node.ACspEvent;
import lmf.formula.csp.node.ACspExternalChoiceProcess;
import lmf.formula.csp.node.ACspInputComunication;
import lmf.formula.csp.node.ACspInterleaveProcess;
import lmf.formula.csp.node.ACspLetExpr;
import lmf.formula.csp.node.ACspOutputComunication;
import lmf.formula.csp.node.ACspPrefixProcess;
import lmf.formula.csp.node.ACspProcessCallExpr;
import lmf.formula.csp.node.ACspProcessDefinition;
import lmf.formula.csp.node.ACspProjection;
import lmf.formula.csp.node.ACspSequentialProcess;
import lmf.formula.csp.node.ACspSharingParallelProcess;
import lmf.formula.csp.node.ACspType;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.PDefinitionList;
import lmf.formula.csp.node.PEvent;
import lmf.formula.csp.node.PExpression;
import lmf.formula.csp.node.PHelp;
import lmf.formula.csp.node.PType;
import lmf.formula.csp.node.Start;
import mathCollection.HashMathSet;
import mathCollection.MathSet;
import mathCollection.SetOfSets;
import csp2hc.exception.BusUseInExternalChoiceException;
import csp2hc.exception.UndefinedChannelInOutException;
import csp2hc.exception.UnsupportedConstructorException;
import csp2hc.messages.HandelC;
import csp2hc.messages.Messenger;
import csp2hc.util.Branch2ActionSetLUT;
import csp2hc.util.Branch2BranchSetLUT;
import csp2hc.util.BranchActionsOnChannel;
import csp2hc.util.CspUtil;
import csp2hc.util.ForcedInterleavingOnChannel;
import csp2hc.util.ListUtil;
import csp2hc.util.MathUtil;
import csp2hc.util.SetUtil;

public class MultiSyncVisitor extends ExtendedDepthFirstAdapter {
    
	private Hashtable <String,List<MultiSyncClient>> multiSyncProc;
	
	private Set<String> multiSyncChan;
	
	private int largerCommInMultiSync;
	
	private int largerCommInterrupt;
	
	private int largerClient;
	
    // Visitor that has the information about Parallelism properties of the system
    private ParallelismAnalyserVisitor parallelismAnalyser;

    private ChannelsVisitor channelsVisitor;
    
    private DatatypesVisitor types;
	
    // Stores the name of the current process definition being analysed
    private String currentProcessDefinition;

    private Hashtable<Integer,List<Integer>> versionsOfMSClients;
    
    // Set of Handel-C keywords for Multi-synchronisation
    private final String[] msKeywords = new String[]{ 
    	"bits_of_MSYNC_ID",
    	"MSYNC_ID",
    	"integer_message",
    	"integer_value",
    	"MSYNC_VALUE",
    	"MSYNC_INTR_VALUE",
    	"integer_MS_CLIENT",
    	"MessageValue",
    	"sMessageValue",
    	"M", "MV", 
    	"BoolValue",
    	"sBoolValue",
    	"B", "BV",
    	"ChannelOfMultiSync",
    	"sChannelOfMultiSync",
    	"psChannelOfMultiSync",
    	"InterruptOfMultiSync",
    	"sInterruptOfMultiSync",
    	"psInterruptOfMultiSync",
    	"FLIP",
    	"endSystem"
    };

	public MultiSyncVisitor(//ProcessesInfoVisitor procsInfo, DirectivesVisitor directives, ConstantsAndFunctionsVisitor constants, 
			DatatypesVisitor types, ChannelsVisitor channelsVisitor,  ParallelismAnalyserVisitor parallelismAnalyser) {
		//this.procsInfo = procsInfo; //this.directives = directives; //this.constants = constants; //this.alphabets = new AlphabetsVisitor(this.procsInfo, this.directives, this.constants);
		this.types = types;
		this.channelsVisitor = channelsVisitor;
		this.parallelismAnalyser = parallelismAnalyser;
		this.initialise();
    }

    private void initialise() {
		this.largerClient = 0;
		this.largerCommInMultiSync = 0;
		this.largerCommInterrupt = 0;
		this.currentProcessDefinition = "";
		this.multiSyncProc = new Hashtable<String, List<MultiSyncClient>>();
		this.versionsOfMSClients = new Hashtable<Integer, List<Integer>>();
		this.findMSyncChan();
		if (!this.multiSyncChan.isEmpty()){
			HandelC hcInstance = HandelC.getInstance();
			for (String keyword : this.msKeywords) {
				hcInstance.addToolKeyword(keyword);
			}
			
			
		}
	}

    // Mtodo que retorna quantos e quais verses de MultiSyncControl so necessrias
    public List <Integer> getListOfControls(){
    	Vector <Integer> controls = new Vector<Integer>();
    	for (String channel: this.parallelismAnalyser.getRealSyncLUT().keySet()){
    		Set<Set<Integer>> syncs = this.parallelismAnalyser.getRealSyncLUT().get(channel);
    		Iterator <Set<Integer>> sync = syncs.iterator();
    		int card = sync.next().size();
    		if (card > 2 && !controls.contains(card)){
    			controls.add(card);
    		}
    	}
    	return controls;
    }
    
    public void caseACspProcessDefinition(ACspProcessDefinition node){
    	this.setCurrentProcessDefinition(node.getName().toString().trim());
    	PExpression expression = node.getExpression();
    	expression.apply(this);
    }

    public void caseACspPrefixProcess(ACspPrefixProcess node){
    	PEvent event = node.getEvent();
    	String channelName = ((ACspEvent)event).getChannelName().toString().trim();

  		if (this.parallelismAnalyser.isMultiSyncChannel(channelName)){
  			MultiSyncChannel msChannel = new MultiSyncChannel(channelName, ListUtil.clone(this.channelsVisitor.getChannelType(channelName)));
  			MultiSyncClient msClient = new MultiSyncClient(node.hashCode());
  			msClient.addMSChan(msChannel);
  			
  			if (this.multiSyncProc.containsKey(this.currentProcessDefinition)){
  				List msClients = this.multiSyncProc.get(this.currentProcessDefinition);
  				msClients.add(msClient);
  				this.multiSyncProc.put(currentProcessDefinition, msClients);
  			} else {
  				List msClients = new Vector();
  				msClients.add(msClient);
  				this.multiSyncProc.put(currentProcessDefinition, msClients);
  			}
  			if (this.largerClient == 0) this.largerClient = 1;
  			String commType = this.channelsVisitor.getChannelCommType(channelName);
  			int commTypeLength = this.types.getTypeLength(commType);
  			if (this.largerCommInMultiSync < commTypeLength) {
  				this.largerCommInMultiSync = commTypeLength;
  			}
		}
		node.getProcess().apply(this);
    }
    
    public void caseACspExternalChoiceProcess(ACspExternalChoiceProcess node) {
    	
    	InitialsExtractorVisitor initialsExtractor = new InitialsExtractorVisitor();
    	node.apply(initialsExtractor);
    	List keys = initialsExtractor.keyList();
    	
    	List <MultiSyncChannel> msChannels = new Vector<MultiSyncChannel>();
    	List <AnyChannel>otherChannels = new Vector<AnyChannel>();
    	
    	int numbOf_MsChanWithGuard = 0;
    	int numbOf_InChanWithGuard = 0;
    	int numbOf_OutChanWithGuard = 0;
    	
    	int largerCommInterrupt = 0;
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            LinkedList comm = ((ACspEvent)event).getComunication();
            LinkedList proj = ((ACspEvent)event).getProjection();
            String channelName = ((ACspEvent)event).getChannelName().toString().trim();
            if (this.parallelismAnalyser.isMultiSyncChannel(channelName)) {
            	MultiSyncChannel msChannel = new MultiSyncChannel(channelName, ListUtil.clone(this.channelsVisitor.getChannelType(channelName)));
      			msChannels.add(msChannel);
      			
      			if (initialsExtractor.getGuard(event) != null) numbOf_MsChanWithGuard ++;

      			String commType = this.channelsVisitor.getChannelCommType(channelName);
      			int commTypeLength;
      			if (commType.equals(HandelC.SYNC_TYPE)){
      				commTypeLength = 1;
      			} else {
      				commTypeLength = this.types.getTypeLength(commType);
      			}
      			if (this.largerCommInMultiSync < commTypeLength) {
      				this.largerCommInMultiSync = commTypeLength;
      			}
            }
    		else {
    			String channel_name = HandelC.getInstance().getChannelName(channelName);
    	        String action = this.parallelismAnalyser.getCommunicationAction(this.currentProcessDefinition, 
    	        																channel_name, 
    	        																comm);
    	        AnyChannel otherChannel = new AnyChannel(channelName, action, ListUtil.clone(this.channelsVisitor.getChannelType(channelName)));
    	        otherChannels.add(otherChannel);
    	        
    	        if (initialsExtractor.getGuard(event) != null) {
    	        	if (action.equals(HandelC.A_READ)) numbOf_InChanWithGuard ++;
    	        	else if (action.equals(HandelC.A_WRITE)) numbOf_OutChanWithGuard ++;
    	        }
    	        
    	        String commType = this.channelsVisitor.getChannelCommType(channelName);
      			int commTypeLength;
      			if (commType.equals(HandelC.SYNC_TYPE)){
      				commTypeLength = 1;
      			} else {
      				commTypeLength = this.types.getTypeLength(commType);
      			}
      			if (largerCommInterrupt < commTypeLength) {
      				largerCommInterrupt = commTypeLength;
      			}
    		}
        }
        
        if (!msChannels.isEmpty()){
        	MultiSyncClient msClient = new MultiSyncClient(node.hashCode());
        	List <MultiSyncClient> msClientGuarded;
        	for (MultiSyncChannel chan:msChannels){
        		msClient.addMSChan(chan);
        	}
        	for (AnyChannel chan:otherChannels){
        		msClient.addAnyChan(chan);
        	}
        	msClient.setGuards(numbOf_MsChanWithGuard, numbOf_InChanWithGuard, numbOf_OutChanWithGuard);
        	
        	
        	
        	if (this.multiSyncProc.containsKey(this.currentProcessDefinition)){
        		List msClients = this.multiSyncProc.get(this.currentProcessDefinition);
        		msClients.add(msClient);
        		this.multiSyncProc.put(currentProcessDefinition, msClients);
        	} else {
        		List msClients = new Vector();
        		msClients.add(msClient);
        		this.multiSyncProc.put(currentProcessDefinition, msClients);
        	}
        	
        	if (this.largerClient < keys.size()) this.largerClient = keys.size();
        	
        	if (this.largerCommInterrupt < largerCommInterrupt) {
  				this.largerCommInterrupt = largerCommInterrupt;
  			}
        }

        for(int i=0; i < keys.size(); i++) {
        	PEvent event = (PEvent)keys.get(i);
        	Node process = initialsExtractor.getProcess(event);
        	process.apply(this);
        }
    }
    
    public void caseACspAssert(ACspAssert node) {
    }
    
    public String getChannelDeclaration(String chan) {
        String arrayDecl = "";
        // Declaring the value communicated via this channel
        if (this.channelsVisitor.containsChannelType(chan)) {
            List channelType = this.channelsVisitor.getChannelType(chan);
            int sizeOfChannelType = (!this.channelsVisitor.isCommChannel(chan)?channelType.size():channelType.size()-1);
            if(this.channelsVisitor.isProjChannel(chan)){
                for (int i = 0; i < sizeOfChannelType; i++) {
                    PType type = (PType)channelType.get(i);
                    
                    String nodeTypeName = ((ACspType) type).getName().getText().trim(); 
                	String hcType = HandelC.getInstance().getHandelCTypeName(nodeTypeName);
                    String typeCard = HandelC.getInstance().getCardVarOf(hcType);
                    
                    arrayDecl = arrayDecl + HandelC.ARRAY_BEGIN;
                    arrayDecl = arrayDecl + typeCard;
                    arrayDecl = arrayDecl + HandelC.ARRAY_END;
                }
            }
        }
        
        Set <Set<Integer>> sync_groups = this.parallelismAnalyser.getRealSyncLUT().get(HandelC.getInstance().getChannelName(chan));
        int numberOfChannels = sync_groups.iterator().next().size();
        String decl = 
        	HandelC.CHAN_DECL + HandelC.SPACE + "sMessageValue" + HandelC.SPACE + "to_" + chan + arrayDecl +
        	HandelC.ARRAY_BEGIN + numberOfChannels + HandelC.ARRAY_END + HandelC.END_COMMAND +
        	HandelC.CHAN_DECL + HandelC.SPACE + "sBoolValue" + HandelC.SPACE + "from_" + chan + arrayDecl +
        	HandelC.ARRAY_BEGIN + numberOfChannels + HandelC.ARRAY_END + HandelC.END_COMMAND;
        return decl;
    }
    
    
    public List <List<Integer>> getClients (){
    	List msClients = new Vector();
    	for (String proc: this.multiSyncProc.keySet()){
    		for (MultiSyncClient client:this.multiSyncProc.get(proc)){
    			List <Integer> msClient = client.getCardinality();

    			if (!msClients.contains(msClient))
    				msClients.add(msClient);
    			
    			List <Integer> guards = client.getGuards();
    			if (!guards.isEmpty()){
	    			int numbOf_MsChanWithGuard = guards.get(0);
	    			int numbOf_InChanWithGuard = guards.get(1);
	    			int numbOf_OutChanWithGuard = guards.get(2);
	    			
	    			
	    			for(int i = 0; i <= numbOf_MsChanWithGuard; i++){
	    				for(int j = 0; j <= numbOf_InChanWithGuard; j++){
	    					for(int k = 0; k <= numbOf_OutChanWithGuard; k++){
	    						if ((msClient.get(0) - i) > 0) {
	    							List ms = new Vector(3); 
	    							ms.add(0, msClient.get(0) - i);
	    							ms.add(1, msClient.get(1) - j);
	    							ms.add(2, msClient.get(2) - k);
	    							
	    							if (!msClients.contains(ms))
	    								msClients.add(ms);
	    						}
	    					}
	    				}
	    			}
    			}
    		}
    	}
    	return msClients;
    }
    
    public List <String> getMSChannels(String proc, int clientId){
    	List <String> msChannels = new Vector<String>();
    	if (this.multiSyncProc.containsKey(proc)){
    		for (MultiSyncClient client : this.multiSyncProc.get(proc)){
    			if (client.nodeId == clientId){
    				for (MultiSyncChannel channel : client.msChannels){
    					msChannels.add(channel.name);
    				}
    			}
    		}	
    	}
    	return msChannels;
    }
    
    public List <String> getInputChannels(String proc, int clientId){
    	List <String> inChannels = new Vector<String>();
    	if (this.multiSyncProc.containsKey(proc)){
    		for (MultiSyncClient client : this.multiSyncProc.get(proc)){
    			if (client.nodeId == clientId
    					&& client.otherChanIn > 0){
    				for (AnyChannel channel : client.otherChannels){
    					if (channel.action.equals(HandelC.A_READ)){
    						inChannels.add(channel.name);
    					}
    				}
    			}
    		}	
    	}
    	return inChannels;
    }
    
    public List <String> getOutputChannels(String proc, int clientId){
    	List <String> outChannels = new Vector<String>();
    	if (this.multiSyncProc.containsKey(proc)){
    		for (MultiSyncClient client : this.multiSyncProc.get(proc)){
    			if (client.nodeId == clientId
    					&& client.otherChanOut > 0){
    				for (AnyChannel channel : client.otherChannels){
    					if (channel.action.equals(HandelC.A_WRITE)){
    						outChannels.add(channel.name);
    					}
    				}
    			}
    		}	
    	}
    	return outChannels;
    }
    
    public String getHcClient2(String proc, int nodeId){
    	String hcCode = "";
    	if (this.multiSyncProc.containsKey(proc)){
    		for (MultiSyncClient client : this.multiSyncProc.get(proc)){
    			if (client.nodeId == nodeId){
    				String hcMSChannels = "";
    	    		String hcInputChannels = "";
    	    		String hcOutputChannels = "";
    	    		
    	    		int countSync = 0;
    				for (MultiSyncChannel msChan:client.msChannels){
    					String hcPrefix = "seqOfSync" + ((client.msChannels.size() > 1)?"["+countSync+"]":"");
    					hcMSChannels = hcMSChannels + 
    						hcPrefix + ".to_sync = to_"+msChan.name+"; "+
    						hcPrefix + ".from_sync = from_"+msChan.name+"; "+
    						hcPrefix + ".id = MULTI_SYNC_ID_LUT[chan_"+msChan.name+"][BRANCH_ID]; " +
    						hcPrefix + ".channel = chan_"+msChan.name+"; ";
    					countSync++;
    				}
    				int countInput = 0, countOutput = 0;
    				for (AnyChannel anyChan:client.otherChannels){
    					if (anyChan.action.equals(HandelC.A_READ)){
    						String hcPrefix = "seqOfInputNotSync" + ((client.otherChanIn > 1)?"["+countInput+"]":"");
    						hcInputChannels = hcInputChannels + 
	    						hcPrefix + ".interrupt = &"+anyChan.name+"; ";
    						countInput++;
    					} else if (anyChan.action.equals(HandelC.A_WRITE)){
    						String hcPrefix = "seqOfOutputNotSync" + ((client.otherChanOut > 1)?"["+countOutput+"]":"");
    						hcOutputChannels = hcOutputChannels + 
	    						hcPrefix + ".interrupt = &"+anyChan.name+"; ";
    						countOutput++;
    					}
    				}
    				
    				String hcDeclaration = "";
    				hcDeclaration = hcDeclaration + 
				    				"sChannelOfMultiSync seqOfSync " + 
				    				((client.msChannels.size() > 1)?"["+client.msChannels.size()+"]":"") +
				    				"; ";
    				if (client.otherChanIn > 0) {
	    				hcDeclaration = hcDeclaration + 
					    				"sInterruptOfMultiSync seqOfInputNotSync " + 
					    				((client.otherChanIn > 1)?"["+client.otherChanIn+"]":"") +
					    				"; ";
    				}
    				if (client.otherChanOut > 0) {
	    				hcDeclaration = hcDeclaration + 
	    								"sInterruptOfMultiSync seqOfOutputNotSync " + 
					    				((client.otherChanOut > 1)?"["+client.otherChanOut+"]":"") +
					    				"; ";
    				}
    				
    				hcCode = hcDeclaration +
		    				"integer_MS_CLIENT id;"+
		    	    		"integer_value value;"+
		    	    		"par {"+
			    	    		hcMSChannels +
			    	    		hcInputChannels +
			    	    		hcOutputChannels +
		    	    			"value = 0;"+
		    	    		"}";

    				hcCode = hcCode + "id = MultiSyncClient_"+client.msChannels.size()+"x"+client.otherChanIn+"x"+client.otherChanOut+"("+ 
    						(client.msChannels.size()>1 ? "":"&") + "seqOfSync, "+
    						(client.otherChanIn == 1 ? "&":"") + (client.otherChanIn>0 ? "seqOfInputNotSync, ":"") +
    						(client.otherChanOut == 1 ? "&":"") + (client.otherChanOut>0?"seqOfOutputNotSync, ":"")+
    						"&value);";
    			}
    		}
    		return "";
    	} else return "";
    }

    private void findMSyncChan() {
		this.multiSyncChan = new HashSet<String>();
    	for (String channel_name : this.parallelismAnalyser.getRealSyncLUT().keySet()){
    		for (Set <Integer> syncSet : this.parallelismAnalyser.getRealSyncLUT().get(channel_name)){
    			if (syncSet.size() > 2) {
    				this.multiSyncChan.add(HandelC.getInstance().getChannel(channel_name));
    			}
    		}
    	}
	}
    
    public int getLargerMultiSyncSet() {
    	int larger = 0;
    	Enumeration<String> channels = this.parallelismAnalyser.getRealSyncLUT().keys();
    	while(channels.hasMoreElements()) {
    		String channel = channels.nextElement();
    		Set<Set<Integer>> syncs = this.parallelismAnalyser.getRealSyncLUT().get(channel);
    		for (Set<Integer> sync:syncs){
    			int size = sync.size();
    			if (size > larger) {
    				larger = size;
    			}
    		}
    	}
    	return larger;
    }
    
    public List<String> getHcClientsPrototypes(){
    	List <String> prototypes = new Vector<String>();
    	String prototype = "";
    	
    	for (List <Integer> client: this.getClients()){
    		if (client.size() == 3) { 
    			int syncCard = client.get(0);
    			int cardMSInterruptIn = client.get(1);
    			int cardMSInterruptOut = client.get(2);
    	    	prototype = "inline integer_MS_CLIENT MultiSyncClient" + HandelC.UNDERSCORE + syncCard + "("
			    	+"psChannelOfMultiSync seqOfSync,"
			    	+(cardMSInterruptIn>0 ? "psInterruptOfMultiSync seqOfInputNotSync,":"")
			    	+(cardMSInterruptOut>0 ? "psInterruptOfMultiSync seqOfOutputNotSync,":"")
			    	+"MSYNC_VALUE * value)";
    		}
    	}
    	return prototypes;
    }

	private void setCurrentProcessDefinition(String s) {
    	this.currentProcessDefinition = s;
    }

    public Set<String> getMultiSyncChannels() {
    	return this.multiSyncChan;
    }
    
    public int getLargerClient (){
    	return this.largerClient;
    }
    
    public int getLargerCommInMultiSync() {
		return this.largerCommInMultiSync;
	}
    
    public int getLargerCommInterrupt() {
		return this.largerCommInterrupt;
	}
    
    public Hashtable<Integer,List<Integer>> getVersionsOfMSClients (Node node){
    	return this.versionsOfMSClients;
    }
    
    private class MultiSyncChannel {
		private String name;
		private List <String> type;
		public MultiSyncChannel(String name, List<String> type) {
			super();
			this.name = name;
			this.type = type;
		}
	}
	private class AnyChannel {
		private String name;
		private String action; // Write or Read
		private List <String> type;
		public AnyChannel(String name, String action, List<String> type) {
			super();
			this.name = name;
			this.action = action;
			this.type = type;
		}
	}
	private class MultiSyncClient {
		private List<MultiSyncChannel> msChannels;
		private List<AnyChannel> otherChannels;
		private int nodeId;
		private List <Integer> guards;
		
		private int otherChanIn;
		private int otherChanOut;
		public MultiSyncClient(int nodeId) {
			super();
			this.nodeId = nodeId;
			this.otherChanIn = 0;
			this.otherChanOut = 0;
			this.guards = new Vector<Integer>(3);
		}
		public void addMSChan(MultiSyncChannel chan){
			if (msChannels == null){
				msChannels = new Vector <MultiSyncChannel>();
			} 
			msChannels.add(chan);
		}
		public void addAnyChan(AnyChannel chan){
			if (otherChannels == null){
				otherChannels = new Vector <AnyChannel>();
			} 
			if (chan.action.equals(HandelC.A_WRITE)){
				this.otherChanOut++;
			} else if (chan.action.equals(HandelC.A_READ)){
				this.otherChanIn++;
			} 
			otherChannels.add(chan);
		}
		
		public void setGuards(int ms, int in, int out) {
			this.guards.add(0, ms);
			this.guards.add(1, in);
			this.guards.add(2, out);
		}
		
		public List <Integer> getGuards() {
			return this.guards;
		}
		
		public List <Integer> getCardinality() {
			List <Integer> cardinality = new Vector<Integer>();
			cardinality.add(this.msChannels.size());
			cardinality.add(this.otherChanIn);
			cardinality.add(this.otherChanOut);
			return cardinality;
		}
	}
    
}

