package csp2hc.visitors;

import java.util.List;

import csp2hc.exception.AbstractDefinitionException;

import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.node.ACspAbstractDefinition;
import lmf.formula.csp.node.ACspDefinitionParagraph;
import lmf.formula.csp.node.ACspProcessDefinition;
import lmf.formula.csp.node.ACspSpecification;
import lmf.formula.csp.node.PDefinition;
import lmf.formula.csp.node.PParagraph;
import lmf.formula.csp.node.Start;


/**
 * ProcessFinder.
 */
public final class ProcessFinder extends ExtendedDepthFirstAdapter {

    private String processName;
    private PDefinition processNode;

    public PDefinition findProcess(String processName, Start node) {
        this.processNode = null;
        this.processName = processName;
        //System.out.println("ProcessFinder: Start is "+node);
        node.apply(this);
        return processNode;
    }
    @Override
    public void caseACspSpecification(ACspSpecification node) {
        List<PParagraph> temp = node.getParagraph();

        for (PParagraph paragraph : temp) {
            if (paragraph instanceof ACspDefinitionParagraph) {
                PDefinition definition =
                    ((ACspDefinitionParagraph) paragraph).getDefinition();
                if (definition instanceof ACspProcessDefinition) {
                    definition.apply(this);
                    
                    if (processNode != null) {
                        break;
                    }
                } else if (definition instanceof ACspAbstractDefinition) {
                    definition.apply(this);
                } 
            }
        }
    }
    @Override
    public void caseACspProcessDefinition(ACspProcessDefinition node) {
        String name = node.getName().getText();
        //System.out.println("caseACspProcessDefinition::ProcessFinder: "+node);
        if (name.equals(processName)) {
            this.processNode = node;
        }
    }
    @Override
    public void caseACspAbstractDefinition(ACspAbstractDefinition node) {
    	//System.out.println("caseACspAbstractDefinition::ProcessFinder: "+node);
        throw new AbstractDefinitionException();
    }
}