package csp2hc.visitors;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class ProcessesInfoVisitor extends ExtendedDepthFirstAdapter {
    // Processes Arguments -- processName:String -> parameters:ACspParameters
    public Hashtable processesArguments;

    // Processes Definition -- processName:String -> definition:PExpression
    public Hashtable processesDefinitions;
    
    private ACspLetExpr letExpr;

	private boolean isParameters;

	private List <String> currentParams;

	private boolean checkingParamTypes;
	
	private DirectivesVisitor directives;
	private DatatypesVisitor types;
	private ChannelsVisitor channels;

	private String currentDefinition;
	
	private PExpression letExpression;

	private CodeInsertionVisitor insertion;
    
	public ProcessesInfoVisitor(DirectivesVisitor directives,
			DatatypesVisitor types, ChannelsVisitor channels) {
        initialise(directives, types, channels, null);
	}
	
    public ProcessesInfoVisitor(DirectivesVisitor directives, DatatypesVisitor types, ChannelsVisitor channels, CodeInsertionVisitor insertion) {
        initialise(directives, types, channels, insertion);
    }

	private void initialise(DirectivesVisitor directives, DatatypesVisitor types, ChannelsVisitor channels, CodeInsertionVisitor insertion) {
		this.processesArguments = new Hashtable();
        this.processesDefinitions = new Hashtable();
    
        this.directives = directives;
        this.types = types;
        this.channels = channels;
        this.insertion = insertion;
        this.isParameters = false;
	}

	public boolean containsProcArguments(String procName) {
        return this.processesArguments.containsKey(procName);
    }
    
    public ACspParameters getProcArguments(String procName){
        return (ACspParameters)this.processesArguments.get(procName);
    }

    public boolean containsProcDefinition(String procName) {
        return this.processesDefinitions.containsKey(procName);
    }
    
    public PExpression getProcDefinition(String procName){
    	if (this.processesDefinitions.containsKey(procName)) {
            return (PExpression)this.processesDefinitions.get(procName);
    	} else {
    		throw new UndefinedProcessDefinitionException(procName);
    	}
    }
    
    public Enumeration keys() {
        return this.processesDefinitions.keys();
    }
    
    public int size() {
        return this.processesDefinitions.size();
    }
    
    public boolean isArgumentOf(String argName, String procName) {
        boolean answer = false;

        if (this.containsProcArguments(procName)) {
            ACspParameters parameters = this.getProcArguments(procName);
            if (parameters.getParamList() != null){
                Node paramsList = ((ACspParameters)parameters).getParamList();

                if (paramsList instanceof ACspBaseParamList) {
                    String parameter = ((ACspBaseParamList)paramsList).getParam().toString().trim();
                    if (argName.equals(parameter)) {
                        answer = true;
                    }
                } else if (paramsList instanceof ACspLinkedParamList) {
                    String parameter = ((ACspLinkedParamList)paramsList).getParam().toString().trim();
                    if (argName.equals(parameter)) {
                        answer = true;
                    }
                    while((((ACspLinkedParamList)paramsList)).getNext() instanceof ACspLinkedParamList){
                        paramsList = ((ACspLinkedParamList)paramsList).getNext();
                        parameter = ((ACspLinkedParamList)paramsList).getParam().toString().trim();
                        if (argName.equals(parameter)) {
                            answer = true;
                        }
                    }
                    paramsList = ((ACspLinkedParamList)paramsList).getNext();
                    parameter = ((ACspBaseParamList)paramsList).getParam().toString().trim();
                    if (argName.equals(parameter)) {
                        answer = true;
                    }
                }
            }
        }
        return answer;
    }

    public void caseACspDefinitionParagraph(ACspDefinitionParagraph node){
        PDefinition definition = node.getDefinition();
        if (definition instanceof ACspProcessDefinition){
        	TName name = ((ACspProcessDefinition)definition).getName();
            ACspParameters parameters = (ACspParameters)((ACspProcessDefinition)definition).getParameters();
            PExpression expression = ((ACspProcessDefinition)definition).getExpression();
            
            this.currentDefinition = name.toString().trim();
            
            this.currentParams = new Vector<String>();
            if (parameters == null) {
            	parameters = new ACspParameters();
            } else {
            	this.isParameters = true;
            	parameters.apply(this);
            	this.isParameters = false;
            }
            String pname = name.toString().trim();
            this.processesArguments.put(pname,parameters);
            
            if (!currentParams.isEmpty()) this.checkingParamTypes = true;
            
            this.letExpr = null;
            expression.apply(this);
            
            this.checkingParamTypes = false;
            
            if (this.letExpr != null){
            	//expression = this.letExpr.getExpression();
            	this.letExpression.apply(this);
            	expression = this.letExpression;
            }
            
            this.processesDefinitions.put(pname,expression);
        }
        //definition.apply(this);
        /*if (this.letExpr != null){
        	this.letExpr.getDefinitionList().apply(this);
        }*/
        
    }
    
    public void caseACspLetExpr (ACspLetExpr node){
    	this.letExpr = node;
    	this.letExpression = (PExpression) node.getExpression().clone();
    }
    
    public void caseACspLinkedDefinitionList (ACspLinkedDefinitionList node){
    	boolean teste = false;
    	PDefinition definition = node.getDefinition();
        if (definition instanceof ACspProcessDefinition){
            TName name = ((ACspProcessDefinition)definition).getName();
            ACspParameters parameters = (ACspParameters)((ACspProcessDefinition)definition).getParameters();
            PExpression expression = ((ACspProcessDefinition)definition).getExpression();

            if (parameters == null) {
                parameters = new ACspParameters();
            }
            
            String pname = name.toString().trim();
            
            this.processesArguments.put(pname,parameters);
            this.processesDefinitions.put(pname,expression);
        }  
        
        node.getNext().apply(this);
    }
    
    public void caseACspBaseDefinitionList (ACspBaseDefinitionList node){
    	//System.out.println("ABaseDefinition " + node);
    	PDefinition definition = node.getDefinition();
        if (definition instanceof ACspProcessDefinition){
            TName name = ((ACspProcessDefinition)definition).getName();
            ACspParameters parameters = (ACspParameters)((ACspProcessDefinition)definition).getParameters();
            PExpression expression = ((ACspProcessDefinition)definition).getExpression();

            this.letExpr = null;
            expression.apply(this);
            if (this.letExpr != null){
            	expression = this.letExpr.getExpression();
            }
            
            if (parameters == null) {
                parameters = new ACspParameters();
            }
            
            String pname = name.toString().trim();
            
            this.processesArguments.put(pname,parameters);
            this.processesDefinitions.put(pname,expression);
        }
        
        //definition.apply(this);
        if (this.letExpr != null){
        	this.letExpr.getDefinitionList().apply(this);
        }
    }

    public void caseACspAssert(ACspAssert node) {
    }
    
    public void caseACspParam(ACspParam node){
    	if (this.isParameters) {
    		this.currentParams.add(node.getExpression().toString().trim());
    	}
    }
    
    /*public void caseACspEvent(ACspEvent node){
    	
    	String channelName = node.getChannelName().getText().trim();
    	LinkedList<PProjection> proj = node.getProjection();
    	LinkedList<PComunication> comm = node.getComunication();
    	
    	ListIterator<PProjection> projIter = proj.listIterator();
    	int index = 0;
    	while (projIter.hasNext()){
    		PProjection p = projIter.next();
    		if (this.currentParams.contains(((ACspProjection) p).getCallExpr().toString().trim())){
    			String param = ((ACspProjection) p).getCallExpr().toString().trim();
    			if (this.channels.containsChannelType(channelName)){
    				this.currentParams.remove(param);
    				List chanTypes = this.channels.getChannelType(channelName);
    				PType type = (PType) chanTypes.get(index);
    				
    				this.directives.putProcArguments(currentDefinition, param, type.toString().trim());
    			}
    		}
    		index++;
    	}
    	
    	ListIterator commIter = comm.listIterator();
    	if (commIter.hasNext()) {
        	Object next = commIter.next();
            if (next instanceof ACspOutputComunication) {
            	if (this.currentParams.contains(((ACspOutputComunication) next).getExpression().toString().trim())){
        			String param = ((ACspOutputComunication) next).getExpression().toString().trim();
        			if (this.channels.containsChannelType(channelName)){
        				this.currentParams.remove(param);
        				List chanTypes = this.channels.getChannelType(channelName);
        				//PType type = (PType) chanTypes.get(index);
        				
        				PType type = (PType)ListUtil.last(chanTypes);
        				
        				this.directives.putProcArguments(currentDefinition, param, type.toString().trim());
        			}
        		}
            }
        }
    }*/
    
    public void caseACspProcessCallExpr (ACspProcessCallExpr node) {
    	String pname = node.getName().getText().trim();
    	if (this.letExpr != null){
    		if (this.insertion == null) {
    			// The visitor is not called in the preprocessing step  
    		} else if (this.insertion.isLocalDefinition(pname, currentDefinition)){
    			node.setName(new TName(this.currentDefinition + "_Local_" + pname));
    			if (this.processesArguments.containsKey(currentDefinition)){
    			//if (this.globalParameters != null){
    				ACspParameters parentParams = (ACspParameters) this.processesArguments.get(currentDefinition);
        			PArgumentList argumentList;
        			if (node.getArguments() != null) {
        				argumentList = this.insertion.createPArgumentList(parentParams.getParamList(), ((ACspArguments) node.getArguments()).getArgumentList());
        			} else {
        				argumentList = this.insertion.createPArgumentList(parentParams.getParamList());
        			}
        			node.setArguments(new ACspArguments(new TLPar(),argumentList,new TRPar()));
        		}
    		}
    	}
    }
    
    public void caseACspFunctionCallExpr (ACspFunctionCallExpr node) {
    	String pname = node.getName().getText().trim();
    	if (this.letExpr != null){
    		if (this.insertion == null) {
    			// The visitor is not called in the preprocessing step  
    		} else if (this.insertion.isLocalDefinition(pname, currentDefinition)){
    			node.setName(new TName(this.currentDefinition + "_Local_" + pname));
    		}
    	}
    }
    
    public Set<String> getProcessesNames(){
    	return this.processesDefinitions.keySet();
    }
    
    /*public void caseACspOutputComunication(ACspOutputComunication node) {
    	if (this.currentParams.contains(node.getExpression().toString().trim())){
			String param = node.getExpression().toString().trim();
			if (this.channels.containsChannelType(channelName)){
				this.currentParams.remove(param);
				List chanTypes = this.channels.getChannelType(channelName);
				PType type = (PType) chanTypes.get(index);
				
				this.directives.putProcArguments(currentDefinition, param, type.toString().trim());
			}
		} 
    }*/

    /*public void caseACspIfProcess(ACspIfProcess node){
    	this.isCondition
    	node.getCondition().apply(this);
    	
    	if 
    	
    	node.getThenProcess().apply(this);
    	node.getElseProcess().apply(this);
    }*/
    
    /*public void caseACspProcessDefinition(ACspProcessDefinition definition){
    	this.currentDefinition = definition.getName().getText().trim();
    	definition.getParameters().apply(this);
    }
    
    
    private A
    
    */
    
    
    //public void outStart (Start node){
    //	System.out.println("ProcessesInfoVisitor::outStart : ProcessesDefinitions = " + this.processesDefinitions);
    //}
    
}
