package csp2hc.visitors;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;
import mathCollection.HashMathSet;
import mathCollection.MathSet;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class ReaderWriterVisitor extends ExtendedDepthFirstAdapter {
	
	private boolean isExtChoice = false;
	
	// Visitor that has the information about the channels
    private ChannelsVisitor channels;

    // Visitor that has the information about Parallelism properties of the system
    private ParallelismAnalyserVisitor parallelismAnalyser;
    
    // Channel Input or Output -- procName:String -> (channelName:String -> A_WRITE/A_READ:String)
    private Hashtable channelInOut;
    
    private void defineChannelInOut () {
    	List projChannels = channels.getProjChannels();
    	
    	Iterator channelIter = projChannels.iterator();
    	while (channelIter.hasNext()){
    		String channelName = (String)channelIter.next();
    		parallelismAnalyser.getBranchActionsOnChannel().containsKey(channelName);
    	}
    }
	
	/*public void caseACspConstantCallExpr (ACspConstantCallExpr node){
		System.out.println("ACspConstantCallExpr: " + node);
	}*/
	
	//public void caseACspProcessCallExpr (ACspProcessCallExpr node){
	//	System.out.println("ACspProcessCallExpr: " + node);
	//}
	
	/*public void caseACspBaseExpression (ACspBaseExpression node){
		//System.out.println("ACspBaseExpression: " + node);
		node.getInterleaveProcess().apply(this);
	}
	public void caseACspIfExpression (ACspIfExpression node){
		System.out.println("ACspIfExpression: " + node);
	}
	public void caseACspLetExpression (ACspLetExpression node){
		System.out.println("ACspLetExpression: " + node);
	}
	public void caseACspLambdaExpression (ACspLambdaExpression node){
		System.out.println("ACspLambdaExpression: " + node);
	}
	public void caseACspAlphabetisedParallelProcess (ACspAlphabetisedParallelProcess node){
		System.out.println("ACspAlphabetisedParallelProcess: " + node);
	}
	public void caseACspBaseParallelProcess (ACspBaseParallelProcess node){
		//System.out.println("ACspBaseParallelProcess: " + node);
		node.getChoiceProcess().apply(this);
	}
	public void caseACspLinkedParallelProcess (ACspLinkedParallelProcess node){
		System.out.println("ACspLinkedParallelProcess: " + node);
	}
	public void caseACspSharingParallelProcess (ACspSharingParallelProcess node){
		System.out.println("ACspSharingParallelProcess: " + node);
	}
	public void caseACspBaseCallExpr (ACspBaseCallExpr node){
		System.out.println("ACspBaseCallExpr: " + node);
	}
	public void caseACspAbstractCallExpr (ACspAbstractCallExpr node){
		System.out.println("ACspAbstractCallExpr: " + node);
	}public void caseACspCallExpr (ACspCallExpr node){
		System.out.println("ACspCallExpr: " + node);
		
	}public void caseACspFunctionCallExpr (ACspFunctionCallExpr node){
		System.out.println("ACspFunctionCallExpr: " + node);
	}public void caseACspNameCallExpr (ACspNameCallExpr node){
		System.out.println("ACspNameCallExpr: " + node);
	}
    public void caseACspAbstractDefinition (ACspAbstractDefinition node){
		System.out.println("ACspAbstractDefinition: " + node);
		String name = node.getName().toString().trim();
        
        if (name.equals(HandelC.MAINP)) {
        	System.out.println("Processo MAINP");
            PExpression expression = node.getExpression();
            expression.apply(this);
    		
        }
	}	
	public void caseACspChannelDefinition (ACspChannelDefinition node){
		//System.out.println("ACspChannelDefinition");
	}	
	public void caseACspConstantDefinition (ACspConstantDefinition node){
		System.out.println("ACspConstantDefinition");
	}
	public void caseACspDatatypeDefinition (ACspDatatypeDefinition node){
		//System.out.println("ACspDatatypeDefinition");
	}
	public void caseACspExternalDefinition (ACspExternalDefinition node){
		System.out.println("ACspExternalDefinition");
	}
	public void caseACspFunctionDefinition (ACspFunctionDefinition node){
		System.out.println("ACspFunctionDefinition");
	}
	public void caseACspNametypeDefinition (ACspNametypeDefinition node){
		System.out.println("ACspNametypeDefinition");
	}    
	public void caseACspProcessDefinition (ACspProcessDefinition node){
		System.out.println("ACspProcessDefinition");
	}
	public void caseACspSubtypeDefinition (ACspSubtypeDefinition node){
		System.out.println("ACspSubtypeDefinition");
	}
	public void caseACspTransparentDefinition (ACspTransparentDefinition node){
		System.out.println("ACspTransparentDefinition");
	}
	
	public void caseACspSpecification(ACspSpecification node){
		System.out.println(node);
	}*/

    /*
     public void case ( node){
		
	}
     */
	
	public void caseACspProcessCallExpr (ACspProcessCallExpr node){
		//printParents(node);		
	}

	public void caseACspAbstractCallExpr (ACspAbstractCallExpr node){
		//printParents(node);
	}
	
	public void caseACspNameCallExpr (ACspNameCallExpr node){
		//printParents(node);
	}
	
	public void caseACspIfExpression(ACspIfExpression node) {
		//printParents(node);
		printChildren(node);
	}
	
	public void printParents (Node node){
		System.out.print("Node: " + node + ", is ");
		printTypeOfNode(node);
		Object pai = node.parent();
		System.out.print("Node: " + pai + ", is ");
		printTypeOfNode(pai);
		while (!(pai instanceof ACspDefinitionParagraph)){
			pai = ((Node)pai).parent();
			System.out.print("Node: " + pai + ", is ");
			printTypeOfNode(pai);
		}
	}
	
	public void caseACspPrefixProcess (ACspPrefixProcess node){
		//prefixProcessBuscaPai(node);
		//System.out.print("Node: " + node + ", is ");
		//printTypeOfNode(node);
		//printParents(node);
		node.getProcess().apply(this);
		
	}
	
	public void prefixProcessBuscaPai(ACspPrefixProcess node){
		System.out.println("Encontrei um Prefix: " + node);
    	Object pai = node.parent();
    	printTypeOfNode(pai);
		while (!(pai instanceof PHelp) && !(pai instanceof ACspSpecification)){
			pai = ((Node)pai).parent();
			printTypeOfNode(pai);
		}
		if (pai instanceof ACspSpecification) {
			System.out.println("Este Prefix no tem um pai PHelp!\n" +
					"Buscando outro Prefix...");
		}
    	node.getProcess().apply(this);
	}
    
    public void caseACspExternalChoiceProcess (ACspExternalChoiceProcess node){
    	//externalChoiceProcessBuscaPai (node);
    	//printParents(node);
    	node.getLeftProcess().apply(this);
    	node.getRightProcess().apply(this);
	}
    
    private void externalChoiceProcessBuscaPai (ACspExternalChoiceProcess node){
    	System.out.println("Encontrei um External Choice: " + node);
    	/*System.out.println("Verificando lado esquerdo...");
		this.isExtChoice = true;
		node.getLeftProcess().apply(this);
		this.isExtChoice = false;
		*/
    	Object pai = node.parent();
    	printTypeOfNode(pai);
		while (!(pai instanceof PHelp) && !(pai instanceof ACspSpecification)){
			pai = ((Node)pai).parent();
			printTypeOfNode(pai);
		}
		if (pai instanceof ACspSpecification) {
			System.out.println("Este External Choice no tem um pai PHelp!");
		}
	}
    
	public void printTypeOfNode (Object node) {
		if (node instanceof ACspAbstractCallExpr){
			System.out.println("ACspAbstractCallExpr");
		} else if (node instanceof ACspAbstractDefinition){
			System.out.println("ACspAbstractDefinition");
		} else if (node instanceof ACspAbstractType){
			System.out.println("ACsp3");
		} else if (node instanceof ACspAddMathExpr){
			System.out.println("ACsp4");
		} else if (node instanceof ACspAlphabetisedParallelProcess){
			System.out.println("ACsp5");
		} else if (node instanceof ACspAlphabetisedReplicatedProcess){
			System.out.println("ACsp6");
		} else if (node instanceof ACspAndLogicalExpr){
			System.out.println("ACsp7");
		} else if (node instanceof ACspArgument){
			System.out.println("ACsp08");
		} else if (node instanceof ACspArguments){
			System.out.println("ACsp09");
		} else if (node instanceof ACspAssert){
			System.out.println("ACsp10");
		} else if (node instanceof ACspAssertParagraph){
			System.out.println("ACsp11");
		} else if (node instanceof ACspBaseArgumentList){
			System.out.println("ACsp12");
		} else if (node instanceof ACspBaseBranchList){
			System.out.println("ACsp13");
		} else if (node instanceof ACspBaseCallExpr){
			System.out.println("ACspBaseCallExpr");
		} else if (node instanceof ACspBaseChannelList){
			System.out.println("ACsp15");
		} else if (node instanceof ACspBaseCheckExpr){
			System.out.println("ACsp16");
		} else if (node instanceof ACspBaseChoiceProcess){
			System.out.println("ACspBaseChoiceProcess");
		} else if (node instanceof ACspBaseCompositeType){
			System.out.println("ACsp18");
		} else if (node instanceof ACspBaseConstraintList){
			System.out.println("ACsp19");
		} else if (node instanceof ACspBaseDefinitionList){
			System.out.println("ACsp20");
		} else if (node instanceof ACspBaseDotExpr){
			System.out.println("ACspBaseDotExpr");
		} else if (node instanceof ACspBaseElementList){
			System.out.println("ACsp22");
		} else if (node instanceof ACspBaseEventList){
			System.out.println("ACsp23");
		} else if (node instanceof ACspBaseExpr){
			System.out.println("ACspBaseExpr");
		} else if (node instanceof ACspBaseExpression){
			System.out.println("ACspBaseExpression");
		} else if (node instanceof ACspBaseGuardedProcess){
			System.out.println("ACspBaseGuardedProcess");
		} else if (node instanceof ACspBaseHidingProcess){
			System.out.println("ACspBaseHidingProcess");
		} else if (node instanceof ACspBaseIfHelp){
			System.out.println("ACsp28");
		} else if (node instanceof ACspBaseInterleaveProcess){
			System.out.println("ACspBaseInterleaveProcess");
		} else if (node instanceof ACspBaseInterruptProcess){
			System.out.println("ACspBaseInterruptProcess");
		} else if (node instanceof ACspBaseLambdaHelp){
			System.out.println("ACsp31");
		} else if (node instanceof ACspBaseLengthExpr){
			System.out.println("ACspBaseLengthExpr");
		} else if (node instanceof ACspBaseLetHelp){
			System.out.println("ACsp33");
		} else if (node instanceof ACspBaseLinkList){
			System.out.println("ACsp34");
		} else if (node instanceof ACspBaseLogicalExpr){
			System.out.println("ACspBaseLogicalExpr");
		} else if (node instanceof ACspBaseMath1Expr){
			System.out.println("ACspBaseMath1Expr");
		} else if (node instanceof ACspBaseMathExpr){
			System.out.println("ACspBaseMathExpr");
		} else if (node instanceof ACspBaseMinusExpr){
			System.out.println("ACspBaseMinusExpr");
		} else if (node instanceof ACspBaseNameList){
			System.out.println("ACsp39");
		} else if (node instanceof ACspBaseNotExpr){
			System.out.println("ACspBaseNotExpr");
		} else if (node instanceof ACspBaseParallelProcess){
			System.out.println("ACspBaseParallelProcess");
		} else if (node instanceof ACspBaseParamList){
			System.out.println("ACsp42");
		} else if (node instanceof ACspBasePrefixProcess){
			System.out.println("ACspBasePrefixProcess");
		} else if (node instanceof ACspBaseProcess){
			System.out.println("ACspBaseProcess");
		} else if (node instanceof ACspBaseRelationalExpr){
			System.out.println("ACspBaseRelationalExpr");
		} else if (node instanceof ACspBaseRenamingProcess){
			System.out.println("ACspBaseRenamingProcess");
		} else if (node instanceof ACspBaseReplicatedHelp){
			System.out.println("ACspBaseReplicatedHelp");
		} else if (node instanceof ACspBaseReplicatedProcess){
			System.out.println("ACspBaseReplicatedProcess");
		} else if (node instanceof ACspBaseSequenceExpr){
			System.out.println("ACspBaseSequenceExpr");
		} else if (node instanceof ACspBaseSequentialProcess){
			System.out.println("ACspBaseSequentialProcess");
		} else if (node instanceof ACspBaseSubstitutionList){
			System.out.println("ACsp51");
		} else if (node instanceof ACspBaseTimeoutProcess){
			System.out.println("ACspBaseTimeoutProcess");
		} else if (node instanceof ACspBaseVariableList){
			System.out.println("ACsp53");
		} else if (node instanceof ACspCallExpr){
			System.out.println("ACsp54");
		} else if (node instanceof ACspCatSequenceExpr){
			System.out.println("ACsp55");
		} else if (node instanceof ACspChannel){
			System.out.println("ACsp56");
		} else if (node instanceof ACspChannelDefinition){
			System.out.println("ACsp57");
		} else if (node instanceof ACspCheckExpr){
			System.out.println("ACsp58");
		} else if (node instanceof ACspClosedRange){
			System.out.println("ACsp59");
		} else if (node instanceof ACspCollectionElement){
			System.out.println("ACsp60");
		} else if (node instanceof ACspCommentedPrefixProcess){
			System.out.println("ACsp61");
		} else if (node instanceof ACspCommentParagraph){
			System.out.println("ACsp62");
		} else if (node instanceof ACspComplexBranch){
			System.out.println("ACsp63");
		} else if (node instanceof ACspCompositeType){
			System.out.println("ACsp64");
		} else if (node instanceof ACspComprehensionSequence){
			System.out.println("ACsp65");
		} else if (node instanceof ACspComprehensionSet){
			System.out.println("ACsp66");
		} else if (node instanceof ACspConstantCallExpr){
			System.out.println("ACsp67");
		} else if (node instanceof ACspConstantDefinition){
			System.out.println("ACsp68");
		} else if (node instanceof ACspConstrainedInputComunication){
			System.out.println("ACsp69");
		} else if (node instanceof ACspConstraint){
			System.out.println("ACsp70");
		} else if (node instanceof ACspConstraints){
			System.out.println("ACsp71");
		} else if (node instanceof ACspConstraintVariable){
			System.out.println("ACsp72");
		} else if (node instanceof ACspDatatypeDefinition){
			System.out.println("ACsp73");
		} else if (node instanceof ACspDefinitionParagraph){
			System.out.println("ACspDefinitionParagraph");
		} else if (node instanceof ACspDiffMathExpr){
			System.out.println("ACsp75");
		} else if (node instanceof ACspDivMath1Expr){
			System.out.println("ACsp76");
		} else if (node instanceof ACspDotExpr){
			System.out.println("ACsp77");
		} else if (node instanceof ACspEqualRelationalExpr){
			System.out.println("ACsp78");
		} else if (node instanceof ACspEvent){
			System.out.println("ACsp79");
		} else if (node instanceof ACspExternalChoiceProcess){
			System.out.println("ACspExternalChoiceProcess");
		} else if (node instanceof ACspExternalDefinition){
			System.out.println("ACsp81");
		} else if (node instanceof ACspExternalReplicatedProcess){
			System.out.println("ACsp82");
		} else if (node instanceof ACspFunctionCallExpr){
			System.out.println("ACsp83");
		} else if (node instanceof ACspFunctionDefinition){
			System.out.println("ACsp84");
		} else if (node instanceof ACspGteRelationalExpr){
			System.out.println("ACsp85");
		} else if (node instanceof ACspGtRelationalExpr){
			System.out.println("ACsp86");
		} else if (node instanceof ACspGuardedProcess){
			System.out.println("ACsp87");
		} else if (node instanceof ACspHidingProcess){
			System.out.println("ACsp88");
		} else if (node instanceof ACspIfExpr){
			System.out.println("ACspIfExpr");
		} else if (node instanceof ACspIfExpression){
			System.out.println("ACspIfExpression");
		} else if (node instanceof ACspIfProcess){
			System.out.println("ACsp91");
		} else if (node instanceof ACspInclude){
			System.out.println("ACsp92");
		} else if (node instanceof ACspIncludeParagraph){
			System.out.println("ACsp93");
		} else if (node instanceof ACspInputComunication){
			System.out.println("ACsp94");
		} else if (node instanceof ACspInstance){
			System.out.println("ACsp95");
		} else if (node instanceof ACspInstanceParagraph){
			System.out.println("ACsp96");
		} else if (node instanceof ACspInterleaveProcess){
			System.out.println("ACsp97");
		} else if (node instanceof ACspInterleaveReplicatedProcess){
			System.out.println("ACsp98");
		} else if (node instanceof ACspInternalChoiceProcess){
			System.out.println("ACsp99");
		} else if (node instanceof ACspInternalReplicatedProcess){
			System.out.println("ACsp100");
		} else if (node instanceof ACspInterruptProcess){
			System.out.println("ACsp101");
		} else if (node instanceof ACspLambdaExpr){
			System.out.println("ACsp102");
		} else if (node instanceof ACspLambdaExpression){
			System.out.println("ACsp103");
		} else if (node instanceof ACspLambdaProcess){
			System.out.println("ACsp104");
		} else if (node instanceof ACspLengthExpr){
			System.out.println("ACsp105");
		} else if (node instanceof ACspLetExpr){
			System.out.println("ACsp106");
		} else if (node instanceof ACspLetExpression){
			System.out.println("ACsp107");
		} else if (node instanceof ACspLetProcess){
			System.out.println("ACsp108");
		} else if (node instanceof ACspLineComment){
			System.out.println("ACsp109");
		} else if (node instanceof ACspLink){
			System.out.println("ACsp110");
		} else if (node instanceof ACspLinkedArgumentList){
			System.out.println("ACsp111");
		} else if (node instanceof ACspLinkedBranchList){
			System.out.println("ACsp112");
		} else if (node instanceof ACspLinkedChannelList){
			System.out.println("ACsp113");
		} else if (node instanceof ACspLinkedConstraintList){
			System.out.println("ACsp114");
		} else if (node instanceof ACspLinkedDefinitionList){
			System.out.println("ACsp115");
		} else if (node instanceof ACspLinkedElementList){
			System.out.println("ACsp116");
		} else if (node instanceof ACspLinkedEventList){
			System.out.println("ACsp117");
		} else if (node instanceof ACspLinkedLinkList){
			System.out.println("ACsp118");
		} else if (node instanceof ACspLinkedNameList){
			System.out.println("ACsp119");
		} else if (node instanceof ACspLinkedParallelProcess){
			System.out.println("ACsp120");
		} else if (node instanceof ACspLinkedParamList){
			System.out.println("ACsp121");
		} else if (node instanceof ACspLinkedReplicatedProcess){
			System.out.println("ACsp122");
		} else if (node instanceof ACspLinkedSubstitutionList){
			System.out.println("ACsp123");
		} else if (node instanceof ACspLinkedVariableList){
			System.out.println("ACsp124");
		} else if (node instanceof ACspLteRelationalExpr){
			System.out.println("ACsp125");
		} else if (node instanceof ACspLtRelationalExpr){
			System.out.println("ACsp126");
		} else if (node instanceof ACspMinusExpr){
			System.out.println("ACsp127");
		} else if (node instanceof ACspModel){
			System.out.println("ACsp128");
		} else if (node instanceof ACspModelReference){
			System.out.println("ACsp129");
		} else if (node instanceof ACspModule){
			System.out.println("ACsp130");
		} else if (node instanceof ACspModuleCall){
			System.out.println("ACsp131");
		} else if (node instanceof ACspModuleEvent){
			System.out.println("ACsp132");
		} else if (node instanceof ACspModuleParagraph){
			System.out.println("ACsp133");
		} else if (node instanceof ACspMulMath1Expr){
			System.out.println("ACsp134");
		} else if (node instanceof ACspMultiLineComment){
			System.out.println("ACsp135");
		} else if (node instanceof ACspNameCallExpr){
			System.out.println("ACspNameCallExpr");
		} else if (node instanceof ACspNametypeDefinition){
			System.out.println("ACsp137");
		} else if (node instanceof ACspNotEqualRelationalExpr){
			System.out.println("ACsp138");
		} else if (node instanceof ACspNotExpr){
			System.out.println("ACsp139");
		} else if (node instanceof ACspNumberValue){
			System.out.println("ACsp140");
		} else if (node instanceof ACspOpenRange){
			System.out.println("ACsp141");
		} else if (node instanceof ACspOrLogicalExpr){
			System.out.println("ACsp142");
		} else if (node instanceof ACspOutputComunication){
			System.out.println("ACsp143");
		} else if (node instanceof ACspParam){
			System.out.println("ACsp144");
		} else if (node instanceof ACspParameters){
			System.out.println("ACsp145");
		} else if (node instanceof ACspParenthesedValue){
			System.out.println("ACspParenthesedValue");
		} else if (node instanceof ACspPrefixProcess){
			System.out.println("ACspPrefixProcess");
		} else if (node instanceof ACspProcessCallExpr){
			System.out.println("ACsp148");
		} else if (node instanceof ACspProcessDefinition){
			System.out.println("ACsp149");
		} else if (node instanceof ACspProductionsSet){
			System.out.println("ACsp150");
		} else if (node instanceof ACspProjection){
			System.out.println("ACsp151");
		} else if (node instanceof ACspRangeCollectionElement){
			System.out.println("ACsp152");
		} else if (node instanceof ACspRefinementCheckExpr){
			System.out.println("ACsp153");
		} else if (node instanceof ACspRemainderMath1Expr){
			System.out.println("ACsp154");
		} else if (node instanceof ACspRenamingProcess){
			System.out.println("ACsp155");
		} else if (node instanceof ACspSequence){
			System.out.println("ACsp156");
		} else if (node instanceof ACspSequenceValue){
			System.out.println("ACsp157");
		} else if (node instanceof ACspSequentialProcess){
			System.out.println("ACsp158");
		} else if (node instanceof ACspSequentialReplicatedProcess){
			System.out.println("ACsp159");
		} else if (node instanceof ACspSet){
			System.out.println("ACsp160");
		} else if (node instanceof ACspSetType){
			System.out.println("ACsp161");
		} else if (node instanceof ACspSetValue){
			System.out.println("ACsp162");
		} else if (node instanceof ACspSharingParallelProcess){
			System.out.println("ACsp163");
		} else if (node instanceof ACspSharingReplicatedProcess){
			System.out.println("ACsp164");
		} else if (node instanceof ACspSimpleBranch){
			System.out.println("ACsp165");
		} else if (node instanceof ACspSourceConstraint){
			System.out.println("ACsp166");
		} else if (node instanceof ACspSpecification){
			System.out.println("ACsp167");
		} else if (node instanceof ACspSubstitution){
			System.out.println("ACsp168");
		} else if (node instanceof ACspSubtypeDefinition){
			System.out.println("ACsp169");
		} else if (node instanceof ACspTransparentDefinition){
			System.out.println("ACsp170");
		} else if (node instanceof ACspTuple){
			System.out.println("ACsp171");
		} else if (node instanceof ACspTupleElements){
			System.out.println("ACsp172");
		} else if (node instanceof ACspTupleType){
			System.out.println("ACsp173");
		} else if (node instanceof ACspTupleValue){
			System.out.println("ACsp174");
		} else if (node instanceof ACspType){
			System.out.println("ACsp175");
		} else if (node instanceof ACspTypeSpecifier){
			System.out.println("ACsp176");
		} else if (node instanceof ACspUntimedTimeoutProcess){
			System.out.println("ACsp177");
		} else if (node instanceof ACspVariable){
			System.out.println("ACsp178");
		}
	}

	public void printChildren (Object node) {
		String abcd;
		if (node instanceof ACspAbstractCallExpr){
		System.out.println("Node: " + node + " is ACspAbstractCallExpr");
		abcd = "";
		}else if (node instanceof ACspAbstractDefinition){
		System.out.println("Node: " + node + " is ACspAbstractDefinition");
		abcd = "";
		}else if (node instanceof ACspAbstractType){
		System.out.println("Node: " + node + " is ACspAbstractType");
		abcd = "";
		}else if (node instanceof ACspAddMathExpr){
		System.out.println("Node: " + node + " is ACspAddMathExpr");
		abcd = "";
		}else if (node instanceof ACspAlphabetisedParallelProcess){
		System.out.println("Node: " + node + " is ACspAlphabetisedParallelProcess");
		abcd = "";
		}else if (node instanceof ACspAlphabetisedReplicatedProcess){
		System.out.println("Node: " + node + " is ACspAlphabetisedReplicatedProcess");
		abcd = "";
		}else if (node instanceof ACspAndLogicalExpr){
		System.out.println("Node: " + node + " is ACspAndLogicalExpr");
		abcd = "";
		}else if (node instanceof ACspArgument){
		System.out.println("Node: " + node + " is ACspArgument");
		abcd = "";
		}else if (node instanceof ACspArguments){
		System.out.println("Node: " + node + " is ACspArguments");
		abcd = "";
		}else if (node instanceof ACspAssert){
		System.out.println("Node: " + node + " is ACspAssert");
		abcd = "";
		}else if (node instanceof ACspAssertParagraph){
		System.out.println("Node: " + node + " is ACspAssertParagraph");
		abcd = "";
		}else if (node instanceof ACspBaseArgumentList){
		System.out.println("Node: " + node + " is ACspBaseArgumentList");
		abcd = "";
		}else if (node instanceof ACspBaseBranchList){
		System.out.println("Node: " + node + " is ACspBaseBranchList");
		abcd = "";
		}else if (node instanceof ACspBaseCallExpr){
		System.out.println("Node: " + node + " is ACspBaseCallExpr");
		abcd = "";
		}else if (node instanceof ACspBaseChannelList){
		System.out.println("Node: " + node + " is ACspBaseChannelList");
		abcd = "";
		}else if (node instanceof ACspBaseCheckExpr){
		System.out.println("Node: " + node + " is ACspBaseCheckExpr");
		abcd = "";
		}else if (node instanceof ACspBaseChoiceProcess){
		System.out.println("Node: " + node + " is ACspBaseChoiceProcess");
		abcd = "";
		}else if (node instanceof ACspBaseCompositeType){
		System.out.println("Node: " + node + " is ACspBaseCompositeType");
		abcd = "";
		}else if (node instanceof ACspBaseConstraintList){
		System.out.println("Node: " + node + " is ACspBaseConstraintList");
		abcd = "";
		}else if (node instanceof ACspBaseDefinitionList){
		System.out.println("Node: " + node + " is ACspBaseDefinitionList");
		abcd = "";
		}else if (node instanceof ACspBaseDotExpr){
		System.out.println("Node: " + node + " is ACspBaseDotExpr");
		abcd = "";
		}else if (node instanceof ACspBaseElementList){
		System.out.println("Node: " + node + " is ACspBaseElementList");
		abcd = "";
		}else if (node instanceof ACspBaseEventList){
		System.out.println("Node: " + node + " is ACspBaseEventList");
		abcd = "";
		}else if (node instanceof ACspBaseExpr){
		System.out.println("Node: " + node + " is ACspBaseExpr");
		abcd = "";
		}else if (node instanceof ACspBaseExpression){
		System.out.println("Node: " + node + " is ACspBaseExpression");
		abcd = "";
		}else if (node instanceof ACspBaseGuardedProcess){
		System.out.println("Node: " + node + " is ACspBaseGuardedProcess");
		abcd = "";
		}else if (node instanceof ACspBaseHidingProcess){
		System.out.println("Node: " + node + " is ACspBaseHidingProcess");
		abcd = "";
		}else if (node instanceof ACspBaseIfHelp){
		System.out.println("Node: " + node + " is ACspBaseIfHelp");
		abcd = "";
		}else if (node instanceof ACspBaseInterleaveProcess){
		System.out.println("Node: " + node + " is ACspBaseInterleaveProcess");
		abcd = "";
		}else if (node instanceof ACspBaseInterruptProcess){
		System.out.println("Node: " + node + " is ACspBaseInterruptProcess");
		abcd = "";
		}else if (node instanceof ACspBaseLambdaHelp){
		System.out.println("Node: " + node + " is ACspBaseLambdaHelp");
		abcd = "";
		}else if (node instanceof ACspBaseLengthExpr){
		System.out.println("Node: " + node + " is ACspBaseLengthExpr");
		abcd = "";
		}else if (node instanceof ACspBaseLetHelp){
		System.out.println("Node: " + node + " is ACspBaseLetHelp");
		abcd = "";
		}else if (node instanceof ACspBaseLinkList){
		System.out.println("Node: " + node + " is ACspBaseLinkList");
		abcd = "";
		}else if (node instanceof ACspBaseLogicalExpr){
		System.out.println("Node: " + node + " is ACspBaseLogicalExpr");
		abcd = "";
		}else if (node instanceof ACspBaseMath1Expr){
		System.out.println("Node: " + node + " is ACspBaseMath1Expr");
		abcd = "";
		}else if (node instanceof ACspBaseMathExpr){
		System.out.println("Node: " + node + " is ACspBaseMathExpr");
		abcd = "";
		}else if (node instanceof ACspBaseMinusExpr){
		System.out.println("Node: " + node + " is ACspBaseMinusExpr");
		abcd = "";
		}else if (node instanceof ACspBaseNameList){
		System.out.println("Node: " + node + " is ACspBaseNameList");
		abcd = "";
		}else if (node instanceof ACspBaseNotExpr){
		System.out.println("Node: " + node + " is ACspBaseNotExpr");
		abcd = "";
		}else if (node instanceof ACspBaseParallelProcess){
		System.out.println("Node: " + node + " is ACspBaseParallelProcess");
		abcd = "";
		}else if (node instanceof ACspBaseParamList){
		System.out.println("Node: " + node + " is ACspBaseParamList");
		abcd = "";
		}else if (node instanceof ACspBasePrefixProcess){
		System.out.println("Node: " + node + " is ACspBasePrefixProcess");
		abcd = "";
		}else if (node instanceof ACspBaseProcess){
		System.out.println("Node: " + node + " is ACspBaseProcess");
		abcd = "";
		}else if (node instanceof ACspBaseRelationalExpr){
		System.out.println("Node: " + node + " is ACspBaseRelationalExpr");
		abcd = "";
		}else if (node instanceof ACspBaseRenamingProcess){
		System.out.println("Node: " + node + " is ACspBaseRenamingProcess");
		abcd = "";
		}else if (node instanceof ACspBaseReplicatedHelp){
		System.out.println("Node: " + node + " is ACspBaseReplicatedHelp");
		abcd = "";
		}else if (node instanceof ACspBaseReplicatedProcess){
		System.out.println("Node: " + node + " is ACspBaseReplicatedProcess");
		abcd = "";
		}else if (node instanceof ACspBaseSequenceExpr){
		System.out.println("Node: " + node + " is ACspBaseSequenceExpr");
		abcd = "";
		}else if (node instanceof ACspBaseSequentialProcess){
		System.out.println("Node: " + node + " is ACspBaseSequentialProcess");
		abcd = "";
		}else if (node instanceof ACspBaseSubstitutionList){
		System.out.println("Node: " + node + " is ACspBaseSubstitutionList");
		abcd = "";
		}else if (node instanceof ACspBaseTimeoutProcess){
		System.out.println("Node: " + node + " is ACspBaseTimeoutProcess");
		abcd = "";
		}else if (node instanceof ACspBaseVariableList){
		System.out.println("Node: " + node + " is ACspBaseVariableList");
		abcd = "";
		}else if (node instanceof ACspCallExpr){
		System.out.println("Node: " + node + " is ACspCallExpr");
		abcd = "";
		}else if (node instanceof ACspCatSequenceExpr){
		System.out.println("Node: " + node + " is ACspCatSequenceExpr");
		abcd = "";
		}else if (node instanceof ACspChannel){
		System.out.println("Node: " + node + " is ACspChannel");
		abcd = "";
		}else if (node instanceof ACspChannelDefinition){
		System.out.println("Node: " + node + " is ACspChannelDefinition");
		abcd = "";
		}else if (node instanceof ACspCheckExpr){
		System.out.println("Node: " + node + " is ACspCheckExpr");
		abcd = "";
		}else if (node instanceof ACspClosedRange){
		System.out.println("Node: " + node + " is ACspClosedRange");
		abcd = "";
		}else if (node instanceof ACspCollectionElement){
		System.out.println("Node: " + node + " is ACspCollectionElement");
		abcd = "";
		}else if (node instanceof ACspCommentedPrefixProcess){
		System.out.println("Node: " + node + " is ACspCommentedPrefixProcess");
		abcd = "";
		}else if (node instanceof ACspCommentParagraph){
		System.out.println("Node: " + node + " is ACspCommentParagraph");
		abcd = "";
		}else if (node instanceof ACspComplexBranch){
		System.out.println("Node: " + node + " is ACspComplexBranch");
		abcd = "";
		}else if (node instanceof ACspCompositeType){
		System.out.println("Node: " + node + " is ACspCompositeType");
		abcd = "";
		}else if (node instanceof ACspComprehensionSequence){
		System.out.println("Node: " + node + " is ACspComprehensionSequence");
		abcd = "";
		}else if (node instanceof ACspComprehensionSet){
		System.out.println("Node: " + node + " is ACspComprehensionSet");
		abcd = "";
		}else if (node instanceof ACspConstantCallExpr){
		System.out.println("Node: " + node + " is ACspConstantCallExpr");
		abcd = "";
		}else if (node instanceof ACspConstantDefinition){
		System.out.println("Node: " + node + " is ACspConstantDefinition");
		abcd = "";
		}else if (node instanceof ACspConstrainedInputComunication){
		System.out.println("Node: " + node + " is ACspConstrainedInputComunication");
		abcd = "";
		}else if (node instanceof ACspConstraint){
		System.out.println("Node: " + node + " is ACspConstraint");
		abcd = "";
		}else if (node instanceof ACspConstraints){
		System.out.println("Node: " + node + " is ACspConstraints");
		abcd = "";
		}else if (node instanceof ACspConstraintVariable){
		System.out.println("Node: " + node + " is ACspConstraintVariable");
		abcd = "";
		}else if (node instanceof ACspDatatypeDefinition){
		System.out.println("Node: " + node + " is ACspDatatypeDefinition");
		abcd = "";
		}else if (node instanceof ACspDefinitionParagraph){
		System.out.println("Node: " + node + " is ACspDefinitionParagraph");
		abcd = "";
		}else if (node instanceof ACspDiffMathExpr){
		System.out.println("Node: " + node + " is ACspDiffMathExpr");
		abcd = "";
		}else if (node instanceof ACspDivMath1Expr){
		System.out.println("Node: " + node + " is ACspDivMath1Expr");
		abcd = "";
		}else if (node instanceof ACspDotExpr){
		System.out.println("Node: " + node + " is ACspDotExpr");
		abcd = "";
		}else if (node instanceof ACspEqualRelationalExpr){
		System.out.println("Node: " + node + " is ACspEqualRelationalExpr");
		abcd = "";
		}else if (node instanceof ACspEvent){
		System.out.println("Node: " + node + " is ACspEvent");
		abcd = "";
		}else if (node instanceof ACspExternalChoiceProcess){
		System.out.println("Node: " + node + " is ACspExternalChoiceProcess");
		abcd = "";
		}else if (node instanceof ACspExternalDefinition){
		System.out.println("Node: " + node + " is ACspExternalDefinition");
		abcd = "";
		}else if (node instanceof ACspExternalReplicatedProcess){
		System.out.println("Node: " + node + " is ACspExternalReplicatedProcess");
		abcd = "";
		}else if (node instanceof ACspFunctionCallExpr){
		System.out.println("Node: " + node + " is ACspFunctionCallExpr");
		abcd = "";
		}else if (node instanceof ACspFunctionDefinition){
		System.out.println("Node: " + node + " is ACspFunctionDefinition");
		abcd = "";
		}else if (node instanceof ACspGteRelationalExpr){
		System.out.println("Node: " + node + " is ACspGteRelationalExpr");
		abcd = "";
		}else if (node instanceof ACspGtRelationalExpr){
		System.out.println("Node: " + node + " is ACspGtRelationalExpr");
		abcd = "";
		}else if (node instanceof ACspGuardedProcess){
		System.out.println("Node: " + node + " is ACspGuardedProcess");
		abcd = "";
		}else if (node instanceof ACspHidingProcess){
		System.out.println("Node: " + node + " is ACspHidingProcess");
		abcd = "";
		}else if (node instanceof ACspIfExpr){
		System.out.println("Node: " + node + " is ACspIfExpr");
		printChildren(((ACspIfExpr) node).getCondition());
		printChildren(((ACspIfExpr) node).getThenExpr());
		printChildren(((ACspIfExpr) node).getElseExpr());
		abcd = "";
		}else if (node instanceof ACspIfExpression){
		System.out.println("Node: " + node + " is ACspIfExpression");
		printChildren(((ACspIfExpression) node).getIfExpr());
		abcd = "";
		}else if (node instanceof ACspIfProcess){
		System.out.println("Node: " + node + " is ACspIfProcess");
		abcd = "";
		}else if (node instanceof ACspInclude){
		System.out.println("Node: " + node + " is ACspInclude");
		abcd = "";
		}else if (node instanceof ACspIncludeParagraph){
		System.out.println("Node: " + node + " is ACspIncludeParagraph");
		abcd = "";
		}else if (node instanceof ACspInputComunication){
		System.out.println("Node: " + node + " is ACspInputComunication");
		abcd = "";
		}else if (node instanceof ACspInstance){
		System.out.println("Node: " + node + " is ACspInstance");
		abcd = "";
		}else if (node instanceof ACspInstanceParagraph){
		System.out.println("Node: " + node + " is ACspInstanceParagraph");
		abcd = "";
		}else if (node instanceof ACspInterleaveProcess){
		System.out.println("Node: " + node + " is ACspInterleaveProcess");
		abcd = "";
		}else if (node instanceof ACspInterleaveReplicatedProcess){
		System.out.println("Node: " + node + " is ACspInterleaveReplicatedProcess");
		abcd = "";
		}else if (node instanceof ACspInternalChoiceProcess){
		System.out.println("Node: " + node + " is ACspInternalChoiceProcess");
		abcd = "";
		}else if (node instanceof ACspInternalReplicatedProcess){
		System.out.println("Node: " + node + " is ACspInternalReplicatedProcess");
		abcd = "";
		}else if (node instanceof ACspInterruptProcess){
		System.out.println("Node: " + node + " is ACspInterruptProcess");
		abcd = "";
		}else if (node instanceof ACspLambdaExpr){
		System.out.println("Node: " + node + " is ACspLambdaExpr");
		abcd = "";
		}else if (node instanceof ACspLambdaExpression){
		System.out.println("Node: " + node + " is ACspLambdaExpression");
		abcd = "";
		}else if (node instanceof ACspLambdaProcess){
		System.out.println("Node: " + node + " is ACspLambdaProcess");
		abcd = "";
		}else if (node instanceof ACspLengthExpr){
		System.out.println("Node: " + node + " is ACspLengthExpr");
		abcd = "";
		}else if (node instanceof ACspLetExpr){
		System.out.println("Node: " + node + " is ACspLetExpr");
		abcd = "";
		}else if (node instanceof ACspLetExpression){
		System.out.println("Node: " + node + " is ACspLetExpression");
		abcd = "";
		}else if (node instanceof ACspLetProcess){
		System.out.println("Node: " + node + " is ACspLetProcess");
		abcd = "";
		}else if (node instanceof ACspLineComment){
		System.out.println("Node: " + node + " is ACspLineComment");
		abcd = "";
		}else if (node instanceof ACspLink){
		System.out.println("Node: " + node + " is ACspLink");
		abcd = "";
		}else if (node instanceof ACspLinkedArgumentList){
		System.out.println("Node: " + node + " is ACspLinkedArgumentList");
		abcd = "";
		}else if (node instanceof ACspLinkedBranchList){
		System.out.println("Node: " + node + " is ACspLinkedBranchList");
		abcd = "";
		}else if (node instanceof ACspLinkedChannelList){
		System.out.println("Node: " + node + " is ACspLinkedChannelList");
		abcd = "";
		}else if (node instanceof ACspLinkedConstraintList){
		System.out.println("Node: " + node + " is ACspLinkedConstraintList");
		abcd = "";
		}else if (node instanceof ACspLinkedDefinitionList){
		System.out.println("Node: " + node + " is ACspLinkedDefinitionList");
		abcd = "";
		}else if (node instanceof ACspLinkedElementList){
		System.out.println("Node: " + node + " is ACspLinkedElementList");
		abcd = "";
		}else if (node instanceof ACspLinkedEventList){
		System.out.println("Node: " + node + " is ACspLinkedEventList");
		abcd = "";
		}else if (node instanceof ACspLinkedLinkList){
		System.out.println("Node: " + node + " is ACspLinkedLinkList");
		abcd = "";
		}else if (node instanceof ACspLinkedNameList){
		System.out.println("Node: " + node + " is ACspLinkedNameList");
		abcd = "";
		}else if (node instanceof ACspLinkedParallelProcess){
		System.out.println("Node: " + node + " is ACspLinkedParallelProcess");
		abcd = "";
		}else if (node instanceof ACspLinkedParamList){
		System.out.println("Node: " + node + " is ACspLinkedParamList");
		abcd = "";
		}else if (node instanceof ACspLinkedReplicatedProcess){
		System.out.println("Node: " + node + " is ACspLinkedReplicatedProcess");
		abcd = "";
		}else if (node instanceof ACspLinkedSubstitutionList){
		System.out.println("Node: " + node + " is ACspLinkedSubstitutionList");
		abcd = "";
		}else if (node instanceof ACspLinkedVariableList){
		System.out.println("Node: " + node + " is ACspLinkedVariableList");
		abcd = "";
		}else if (node instanceof ACspLteRelationalExpr){
		System.out.println("Node: " + node + " is ACspLteRelationalExpr");
		abcd = "";
		}else if (node instanceof ACspLtRelationalExpr){
		System.out.println("Node: " + node + " is ACspLtRelationalExpr");
		abcd = "";
		}else if (node instanceof ACspMinusExpr){
		System.out.println("Node: " + node + " is ACspMinusExpr");
		abcd = "";
		}else if (node instanceof ACspModel){
		System.out.println("Node: " + node + " is ACspModel");
		abcd = "";
		}else if (node instanceof ACspModelReference){
		System.out.println("Node: " + node + " is ACspModelReference");
		abcd = "";
		}else if (node instanceof ACspModule){
		System.out.println("Node: " + node + " is ACspModule");
		abcd = "";
		}else if (node instanceof ACspModuleCall){
		System.out.println("Node: " + node + " is ACspModuleCall");
		abcd = "";
		}else if (node instanceof ACspModuleEvent){
		System.out.println("Node: " + node + " is ACspModuleEvent");
		abcd = "";
		}else if (node instanceof ACspModuleParagraph){
		System.out.println("Node: " + node + " is ACspModuleParagraph");
		abcd = "";
		}else if (node instanceof ACspMulMath1Expr){
		System.out.println("Node: " + node + " is ACspMulMath1Expr");
		abcd = "";
		}else if (node instanceof ACspMultiLineComment){
		System.out.println("Node: " + node + " is ACspMultiLineComment");
		abcd = "";
		}else if (node instanceof ACspNameCallExpr){
		System.out.println("Node: " + node + " is ACspNameCallExpr");
		abcd = "";
		}else if (node instanceof ACspNametypeDefinition){
		System.out.println("Node: " + node + " is ACspNametypeDefinition");
		abcd = "";
		}else if (node instanceof ACspNotEqualRelationalExpr){
		System.out.println("Node: " + node + " is ACspNotEqualRelationalExpr");
		abcd = "";
		}else if (node instanceof ACspNotExpr){
		System.out.println("Node: " + node + " is ACspNotExpr");
		abcd = "";
		}else if (node instanceof ACspNumberValue){
		System.out.println("Node: " + node + " is ACspNumberValue");
		abcd = "";
		}else if (node instanceof ACspOpenRange){
		System.out.println("Node: " + node + " is ACspOpenRange");
		abcd = "";
		}else if (node instanceof ACspOrLogicalExpr){
		System.out.println("Node: " + node + " is ACspOrLogicalExpr");
		abcd = "";
		}else if (node instanceof ACspOutputComunication){
		System.out.println("Node: " + node + " is ACspOutputComunication");
		abcd = "";
		}else if (node instanceof ACspParam){
		System.out.println("Node: " + node + " is ACspParam");
		abcd = "";
		}else if (node instanceof ACspParameters){
		System.out.println("Node: " + node + " is ACspParameters");
		abcd = "";
		}else if (node instanceof ACspParenthesedValue){
		System.out.println("Node: " + node + " is ACspParenthesedValue");
		abcd = "";
		}else if (node instanceof ACspPrefixProcess){
		System.out.println("Node: " + node + " is ACspPrefixProcess");
		abcd = "";
		}else if (node instanceof ACspProcessCallExpr){
		System.out.println("Node: " + node + " is ACspProcessCallExpr");
		abcd = "";
		}else if (node instanceof ACspProcessDefinition){
		System.out.println("Node: " + node + " is ACspProcessDefinition");
		abcd = "";
		}else if (node instanceof ACspProductionsSet){
		System.out.println("Node: " + node + " is ACspProductionsSet");
		abcd = "";
		}else if (node instanceof ACspProjection){
		System.out.println("Node: " + node + " is ACspProjection");
		abcd = "";
		}else if (node instanceof ACspRangeCollectionElement){
		System.out.println("Node: " + node + " is ACspRangeCollectionElement");
		abcd = "";
		}else if (node instanceof ACspRefinementCheckExpr){
		System.out.println("Node: " + node + " is ACspRefinementCheckExpr");
		abcd = "";
		}else if (node instanceof ACspRemainderMath1Expr){
		System.out.println("Node: " + node + " is ACspRemainderMath1Expr");
		abcd = "";
		}else if (node instanceof ACspRenamingProcess){
		System.out.println("Node: " + node + " is ACspRenamingProcess");
		abcd = "";
		}else if (node instanceof ACspSequence){
		System.out.println("Node: " + node + " is ACspSequence");
		abcd = "";
		}else if (node instanceof ACspSequenceValue){
		System.out.println("Node: " + node + " is ACspSequenceValue");
		abcd = "";
		}else if (node instanceof ACspSequentialProcess){
		System.out.println("Node: " + node + " is ACspSequentialProcess");
		abcd = "";
		}else if (node instanceof ACspSequentialReplicatedProcess){
		System.out.println("Node: " + node + " is ACspSequentialReplicatedProcess");
		abcd = "";
		}else if (node instanceof ACspSet){
		System.out.println("Node: " + node + " is ACspSet");
		abcd = "";
		}else if (node instanceof ACspSetType){
		System.out.println("Node: " + node + " is ACspSetType");
		abcd = "";
		}else if (node instanceof ACspSetValue){
		System.out.println("Node: " + node + " is ACspSetValue");
		abcd = "";
		}else if (node instanceof ACspSharingParallelProcess){
		System.out.println("Node: " + node + " is ACspSharingParallelProcess");
		abcd = "";
		}else if (node instanceof ACspSharingReplicatedProcess){
		System.out.println("Node: " + node + " is ACspSharingReplicatedProcess");
		abcd = "";
		}else if (node instanceof ACspSimpleBranch){
		System.out.println("Node: " + node + " is ACspSimpleBranch");
		abcd = "";
		}else if (node instanceof ACspSourceConstraint){
		System.out.println("Node: " + node + " is ACspSourceConstraint");
		abcd = "";
		}else if (node instanceof ACspSpecification){
		System.out.println("Node: " + node + " is ACspSpecification");
		abcd = "";
		}else if (node instanceof ACspSubstitution){
		System.out.println("Node: " + node + " is ACspSubstitution");
		abcd = "";
		}else if (node instanceof ACspSubtypeDefinition){
		System.out.println("Node: " + node + " is ACspSubtypeDefinition");
		abcd = "";
		}else if (node instanceof ACspTransparentDefinition){
		System.out.println("Node: " + node + " is ACspTransparentDefinition");
		abcd = "";
		}else if (node instanceof ACspTuple){
		System.out.println("Node: " + node + " is ACspTuple");
		abcd = "";
		}else if (node instanceof ACspTupleElements){
		System.out.println("Node: " + node + " is ACspTupleElements");
		abcd = "";
		}else if (node instanceof ACspTupleType){
		System.out.println("Node: " + node + " is ACspTupleType");
		abcd = "";
		}else if (node instanceof ACspTupleValue){
		System.out.println("Node: " + node + " is ACspTupleValue");
		abcd = "";
		}else if (node instanceof ACspType){
		System.out.println("Node: " + node + " is ACspType");
		abcd = "";
		}else if (node instanceof ACspTypeSpecifier){
		System.out.println("Node: " + node + " is ACspTypeSpecifier");
		abcd = "";
		}else if (node instanceof ACspUntimedTimeoutProcess){
		System.out.println("Node: " + node + " is ACspUntimedTimeoutProcess");
		abcd = "";
		}else if (node instanceof ACspVariable){
		System.out.println("Node: " + node + " is ACspVariable");
		abcd = "";
		}

	}
}
