package csp2hc.visitors;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import sun.reflect.generics.visitor.Reifier;

import csp2hc.exception.CouldNotInferTypeException;
import csp2hc.exception.UnsupportedConstructorException;
import csp2hc.exception.UnsupportedSubstitutionException;
import csp2hc.messages.HandelC;
import csp2hc.util.BetterPreProcessing;
import csp2hc.util.BranchUtil;

import lmf.formula.csp.node.*;
import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.node.ACspAlphabetisedParallelProcess;
import lmf.formula.csp.node.ACspArguments;
import lmf.formula.csp.node.ACspBaseCallExpr;
import lmf.formula.csp.node.ACspBaseChoiceProcess;
import lmf.formula.csp.node.ACspBaseDotExpr;
import lmf.formula.csp.node.ACspBaseEventList;
import lmf.formula.csp.node.ACspBaseExpr;
import lmf.formula.csp.node.ACspBaseExpression;
import lmf.formula.csp.node.ACspBaseGuardedProcess;
import lmf.formula.csp.node.ACspBaseHidingProcess;
import lmf.formula.csp.node.ACspBaseInterleaveProcess;
import lmf.formula.csp.node.ACspBaseInterruptProcess;
import lmf.formula.csp.node.ACspBaseLengthExpr;
import lmf.formula.csp.node.ACspBaseLogicalExpr;
import lmf.formula.csp.node.ACspBaseMath1Expr;
import lmf.formula.csp.node.ACspBaseMathExpr;
import lmf.formula.csp.node.ACspBaseMinusExpr;
import lmf.formula.csp.node.ACspBaseNotExpr;
import lmf.formula.csp.node.ACspBaseParallelProcess;
import lmf.formula.csp.node.ACspBasePrefixProcess;
import lmf.formula.csp.node.ACspBaseProcess;
import lmf.formula.csp.node.ACspBaseRelationalExpr;
import lmf.formula.csp.node.ACspBaseRenamingProcess;
import lmf.formula.csp.node.ACspBaseReplicatedHelp;
import lmf.formula.csp.node.ACspBaseReplicatedProcess;
import lmf.formula.csp.node.ACspBaseSequenceExpr;
import lmf.formula.csp.node.ACspBaseSequentialProcess;
import lmf.formula.csp.node.ACspBaseTimeoutProcess;
import lmf.formula.csp.node.ACspEvent;
import lmf.formula.csp.node.ACspExternalChoiceProcess;
import lmf.formula.csp.node.ACspIfProcess;
import lmf.formula.csp.node.ACspInterleaveProcess;
import lmf.formula.csp.node.ACspLinkedEventList;
import lmf.formula.csp.node.ACspPrefixProcess;
import lmf.formula.csp.node.ACspProcessCallExpr;
import lmf.formula.csp.node.ACspProcessDefinition;
import lmf.formula.csp.node.ACspRenamingProcess;
import lmf.formula.csp.node.ACspSharingParallelProcess;
import lmf.formula.csp.node.ACspSubstitution;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.PArguments;
import lmf.formula.csp.node.PChoiceProcess;
import lmf.formula.csp.node.PDefinition;
import lmf.formula.csp.node.PEvent;
import lmf.formula.csp.node.PExpression;
import lmf.formula.csp.node.PHelp;
import lmf.formula.csp.node.PProcess;
import lmf.formula.csp.node.Start;
import lmf.formula.csp.node.TArrow;
import lmf.formula.csp.node.TLPar;
import lmf.formula.csp.node.TName;
import lmf.formula.csp.node.TRPar;

public class RenamingVisitor extends ExtendedDepthFirstAdapter{
    
	private ProcessFinder finder = new ProcessFinder();
    private Start start;
    private Hashtable<String, List<String>> renamings;
    private boolean expanding = false;
    
    private String currentDefinition;
    private String processToBeRenamed;
    private PArguments arguments;
    private boolean isGettingRenProcName;
    private ACspBaseCallExpr baseProcessToBeRenamed;
    private boolean unnamedRenaming;
    
    private DirectivesVisitor directives;
    private ChannelsVisitor channels;
    
    private int currentBranch;
	private ConstantsAndFunctionsVisitor constants;
	private ProcessesInfoVisitor procInfo;
	private BetterPreProcessing preProcessing;
    
    public RenamingVisitor(DirectivesVisitor directives, ConstantsAndFunctionsVisitor constants, ProcessesInfoVisitor procInfo, ChannelsVisitor channels, ForcedInterleavingVisitor forcInterleaving, BetterPreProcessing preProcessing){
		this.directives = directives;
		this.constants = constants; 
		this.procInfo = procInfo;
		this.channels = channels;
		this.preProcessing = preProcessing;
	    this.currentDefinition = "";
	    this.isGettingRenProcName = false;
	    this.baseProcessToBeRenamed = null;
	    this.unnamedRenaming = false;
	}

	@Override
    public void caseStart(Start node) {
        this.start = node;
        
        node.getPSpecification().apply(this);
    }
	
    @Override
    public void caseACspProcessDefinition (ACspProcessDefinition node) {
    	this.currentDefinition = node.getName().toString().trim();
    	this.currentBranch = extractProcessBranch(currentDefinition);
    	node.getExpression().apply(this);
    }
    
    @Override
    public void caseACspProcessCallExpr (ACspProcessCallExpr node){
    	if (this.isGettingRenProcName || this.unnamedRenaming){
    		this.processToBeRenamed = node.getName().toString().trim();
    		this.arguments = node.getArguments();
    	} 
    }
    
    @Override
    public void caseACspIfProcess(ACspIfProcess node) {
    	if (unnamedRenaming)
    		throw new UnsupportedConstructorException("Renaming with a unnamed process", currentDefinition);
    	node.getThenProcess().apply(this);
    	node.getElseProcess().apply(this);
    }
    
    @Override
    public void caseACspBaseCallExpr(ACspBaseCallExpr node) {
    	if (isGettingRenProcName) {
    		throw new UnsupportedConstructorException("Renaming with a unnamed process or nested renaming", currentDefinition);
    	} else {
    		node.getValue().apply(this);
    	}
    }
    
    @Override
    public void caseACspRenamingProcess(ACspRenamingProcess node) {
    	
    	if (this.unnamedRenaming){
    		throw new UnsupportedConstructorException("Nested Renaming", currentDefinition);
    	}
    	
    	this.isGettingRenProcName = true;
    	node.getProcess().apply(this);
    	this.isGettingRenProcName = false;
    	
    	if (this.baseProcessToBeRenamed != null) {
    		this.unnamedRenaming = true;
    		this.baseProcessToBeRenamed.getValue().apply(this);
    		this.unnamedRenaming = false;
    		
    	}
    	
    	/*ACspProcessCallExpr procCallExpr = getRenamedProc(node);
    	String pname = procCallExpr.getName().getText().trim();
    	if (this.substVisitor.isRealSubstitution(pname, currentBranch))
    		node.replaceBy(procCallToRenaming(procCallExpr));
    	else{*/
    	
    		if (this.preProcessing.getSubstitutions().isChangedSubst(processToBeRenamed, currentBranch))
    			this.renamings = (Hashtable<String, List<String>>) this.preProcessing.getSubstitutions().getSubstitutionList(processToBeRenamed, currentBranch);
    		else {
		    	this.renamings = new Hashtable<String, List<String>>(10, 10);
		
		    	// Get the renamings.
		        node.getSubstitutionList().apply(this);
		        }
	        
	        // Find the process definition to be renamed.
			PDefinition process = finder.findProcess(
	                processToBeRenamed, start);
	        
	        // Then, start expanding.
	        expanding = true;
	        
	        // 1 - Clone the definition to avoid changing the original one.
	        ACspProcessDefinition cloneDefOfProcessToBeRenamed = (ACspProcessDefinition)process.clone();
			
	        // 2 - MAIN STEP: Apply the renaming in the new definition.
	        String currentDefinitionAux = this.currentDefinition;
	        cloneDefOfProcessToBeRenamed.getExpression().apply(this);
	        this.currentDefinition = currentDefinitionAux;
	        
	        // 3 - Refactor the clone definition to be a new local definition.
	        TName processTName = cloneDefOfProcessToBeRenamed.getName();
	        String srcProcessName = processTName.getText().trim();
			String newProcessName = this.currentDefinition;
	//		String newProcessName = "new" + this.currentDefinition;//processTName.getText().trim();
			RenameProcessRefactoring renameProcessRefactoring =
				new RenameProcessRefactoring(srcProcessName, newProcessName);
			cloneDefOfProcessToBeRenamed.apply(renameProcessRefactoring);
			
			// 4 - Embbed the cloneDefOfProcessToBeRenamed in a PProcess.
			//PProcess pProcess = embbedInPProcess(cloneDefOfProcessToBeRenamed, newProcessName);
			PArguments arguments;
			if (this.arguments != null) {
				arguments = (PArguments) this.arguments.clone();
			} else arguments = new ACspArguments();
			
			PProcess pProcess = embbedInPProcess(cloneDefOfProcessToBeRenamed,arguments);
			
			// 5 - Finishes the expansion by replacing the node by a non-renaming (BASE) process.
			node.replaceBy(new ACspBaseRenamingProcess(pProcess));
			
			 // Finish expanding.
	        expanding = false;
	        
	        if (this.directives.containsProcArguments(srcProcessName)){
				Hashtable<String,String> args = this.directives.getProcArguments(srcProcessName);
				for (String arg : args.keySet()){
					this.directives.putProcArguments(this.currentDefinition, arg, args.get(arg));
				}
			}
			if (this.directives.containsProcChannelInOut(srcProcessName)){
				Hashtable<String,String> chans = this.directives.getProcChannelInOut(srcProcessName);
				for (String chan : chans.keySet()){
					this.directives.putProcChannelInOut(this.currentDefinition, chan, chans.get(chan));
				}
			}
    	//}
    }

	private PProcess embbedInPProcess(ACspProcessDefinition cloneDefOfProcessToBeRenamed //, String newProcessName
										, PArguments arguments) {
		PExpression expression = cloneDefOfProcessToBeRenamed.getExpression();
		return new ACspBaseProcess(new ACspBaseExpr(
				new ACspBaseLogicalExpr(new ACspBaseRelationalExpr(
						new ACspBaseMathExpr(new ACspBaseMath1Expr(
						new ACspBaseSequenceExpr(new ACspBaseNotExpr(
						new ACspBaseMinusExpr(new ACspBaseLengthExpr(
						new ACspBaseDotExpr(new ACspBaseCallExpr(
							new ACspParenthesedValue(new TLPar(),
								expression, new TRPar())
						))))))))))));
	}

	private final class RenameProcessRefactoring extends ExtendedDepthFirstAdapter {
		private String oldName;
		private String newName;
	    
    	public RenameProcessRefactoring(String oldName, String newName) {
    		this.oldName = oldName;
    		this.newName = newName;
    	}

    	@Override
    	public void caseTName(TName node) {
    		if (node.getText().equals(oldName))
    			node.setText(newName);
    	}
	}

	@Override
    public void caseACspSubstitution(ACspSubstitution node) {
    	if (!expanding) {
            PEvent oldE = node.getOldEvent();
            PEvent newE = node.getNewEvent();
            String oldName = ((ACspEvent) oldE).getChannelName().getText().trim();
            String newName = ((ACspEvent) newE).getChannelName().getText().trim(); 
            if (this.channels.containsChannelType(oldName) && 
            		this.channels.containsChannelType(newName)) {
            	if (this.channels.isProjChannel(oldName) && !this.channels.isProjChannel(newName)){
            		this.channels.insertProjChannel(newName);
            	}
            	
            	if (this.channels.isCommChannel(oldName) && !this.channels.isCommChannel(newName)){
            		this.channels.insertCommChannel(newName);
            	}
    
            	List oldEType = this.channels.getChannelType(oldE.toString().trim());
            	List newEType = this.channels.getChannelType(newE.toString().trim());
            	
            	boolean eq = true;
            	if (oldEType != null && newEType != null ){
            		if (oldEType.size() != newEType.size()) eq = false;
            		else {
            			for (int i = 0; i < oldEType.size(); i++){
            				String elemOld = oldEType.get(i).toString().trim();
            				String elemNew = newEType.get(i).toString().trim();
            				if (!elemOld.equals(elemNew)) eq = false;
            			}
            		}
            	} else if (oldEType != newEType ){
            		eq = false;
            	}
            	
            	if (!eq){
            		throw new RuntimeException("The " + oldE.toString().trim() + " and " + newE.toString().trim() + " renaming channels shouldn't have differents types in " + this.currentDefinition);
            	}
            } else if (!((ACspEvent)oldE).getProjection().isEmpty()) {
    			throw new UnsupportedSubstitutionException("projection", this.currentDefinition);
    		} else if (!((ACspEvent)oldE).getComunication().isEmpty()){
    			throw new UnsupportedSubstitutionException("communication", this.currentDefinition);
    		} if (!((ACspEvent)newE).getProjection().isEmpty()) {
    			throw new UnsupportedSubstitutionException("projection", this.currentDefinition);
    		} else if (!((ACspEvent)newE).getComunication().isEmpty()){
    			throw new UnsupportedSubstitutionException("communication", this.currentDefinition);
    		} else if (!this.channels.containsChannelType(oldE.toString().trim())) {
            	throw new CouldNotInferTypeException(oldE.toString().trim(), this.currentDefinition);
            } else if (!this.channels.containsChannelType(newE.toString().trim())) {
            	throw new CouldNotInferTypeException(newE.toString().trim(), this.currentDefinition);
            } 
            
            List<String> newsE;
            if (this.renamings.containsKey(oldE.toString().trim())) {
            	newsE = this.renamings.get(oldE.toString().trim());
            } else {
            	newsE = new Vector <String> ();
            }
            newsE.add(newE.toString().trim());
            this.renamings.put(oldE.toString().trim(), newsE);

            if (this.directives.containsProcChannelInOut(this.processToBeRenamed)) {
                Hashtable channelInOut = this.directives.getProcChannelInOut(this.processToBeRenamed);
                if (channelInOut.containsKey(oldE.toString().trim())){
                    String inOut = (String)channelInOut.get(oldE.toString().trim()); // A_READ or A_WRITE
                    this.directives.putProcChannelInOut("new"+this.currentDefinition, newE.toString().trim(), inOut);
                }
            } 
        }
    }
    
    @Override
    public void caseACspExternalChoiceProcess(ACspExternalChoiceProcess node) {
    	node.getRightProcess().apply(this);
    	node.getLeftProcess().apply(this);
    }
    
    @Override
    public void caseACspPrefixProcess(ACspPrefixProcess node) {
    	if (!expanding) {
    		if(this.unnamedRenaming)
    			throw new UnsupportedConstructorException("Renaming with a unnamed process", currentDefinition);
    		else node.getProcess().apply(this);
    	}
    	else {
    		node.getProcess().apply(this);
        	PEvent event = ((ACspPrefixProcess) node).getEvent();
            TName tName = ((ACspEvent)event).getChannelName();
            if (renamings.containsKey(tName.getText().trim())) {
            	List <String> newNames = renamings.get(tName.getText().trim());
            	if (newNames.size() == 1) {
            		String strNewName = newNames.get(0);
            		if (!this.directives.containsProcChannelInOut(currentDefinition)
            				|| !this.directives.getProcChannelInOut(currentDefinition).containsKey(strNewName)){
            			String type = (String) this.directives.getProcChannelInOut(processToBeRenamed).get(tName.toString().trim());
            			this.directives.putProcChannelInOut(currentDefinition, strNewName, type);
            		}
            		TName newName = new TName(strNewName);
            		newName.setLine(tName.getLine());
            		newName.setPos(tName.getPos());
            		tName.replaceBy(newName);
            	} else {
                	Node ancestor = node.parent();
                	while (!(ancestor instanceof ACspExternalChoiceProcess)  // Caso eu tenha P = a -> SKIP [] b -> SKIP
                			&& !(ancestor instanceof ACspBaseChoiceProcess) 
                			&& !(ancestor instanceof ACspBaseReplicatedHelp)){ // Caso eu tenha P = a -> b -> SKIP
                		ancestor = ancestor.parent();
                	}
                	
                	if (ancestor instanceof ACspBaseReplicatedHelp) {
                		ancestor = (ACspBaseReplicatedProcess)((ACspBaseReplicatedHelp) ancestor).getReplicatedProcess();
                	}
                	
                	createExtChoice (ancestor, node, newNames);
            	}
            }
        }
    }
    
    private void createExtChoice(Node node, ACspPrefixProcess pref, List<String> newChannelNames){
    	PChoiceProcess extChoicProcess = null;
    	if (node instanceof ACspExternalChoiceProcess) {
    		extChoicProcess = ((ACspExternalChoiceProcess)node).getRightProcess();
    	}
    	
    	ListIterator<String> newChannelName = newChannelNames.listIterator();
    	while (newChannelName.hasNext()){
    		String name = newChannelName.next();
    		PEvent event = (PEvent)pref.getEvent().clone();
    		((ACspEvent) event).getChannelName().setText(name);
    		ACspPrefixProcess leftNode = new ACspPrefixProcess(event, new TArrow(),(PHelp) pref.getProcess().clone());
    		ACspBaseReplicatedProcess leftReplicatedProcess = new ACspBaseReplicatedProcess(leftNode);
    		ACspBaseSequentialProcess leftSequentialProcess = new ACspBaseSequentialProcess(leftReplicatedProcess);
    		ACspBaseInterruptProcess leftInterruptProcess = new ACspBaseInterruptProcess(leftSequentialProcess);
    		ACspBaseTimeoutProcess leftProcess = new ACspBaseTimeoutProcess(leftInterruptProcess);
    		if (extChoicProcess == null){
    			extChoicProcess = new ACspBaseChoiceProcess(leftProcess);
    		} else {
    			extChoicProcess = new ACspExternalChoiceProcess(leftProcess, new TExternalChoice(), extChoicProcess);
    		}
    	}
    	
    	if (node instanceof ACspExternalChoiceProcess || node instanceof ACspBaseChoiceProcess) {
    		node.replaceBy(extChoicProcess);
    	} else if (node instanceof ACspBaseReplicatedProcess) {
    		node.replaceBy(new ACspBaseReplicatedProcess(
    				new ACspBasePrefixProcess(new ACspBaseGuardedProcess(
    				new ACspBaseHidingProcess(new ACspBaseRenamingProcess(
					new ACspBaseProcess(new ACspBaseExpr(
					new ACspBaseLogicalExpr(new ACspBaseRelationalExpr(
					new ACspBaseMathExpr(new ACspBaseMath1Expr(
					new ACspBaseSequenceExpr(new ACspBaseNotExpr(
					new ACspBaseMinusExpr(new ACspBaseLengthExpr(
					new ACspBaseDotExpr(new ACspBaseCallExpr(
						new ACspParenthesedValue(new TLPar(),
							new ACspBaseExpression(new ACspBaseInterleaveProcess(
							new ACspBaseParallelProcess(extChoicProcess)))
						, new TRPar())
					))))))))))))))))));
    	}
    }
	private Integer extractProcessBranch(String pname){
		if (HandelC.getInstance().MAINP.equals(this.currentDefinition))
			return 0;
		int lastIndex = pname.lastIndexOf("_RN_"); 
		if (lastIndex < 0) return lastIndex;
		int index = lastIndex + 4;
		int branch = Integer.parseInt(pname.substring(index));
		return branch;
	}
	
	public void caseACspSharingParallelProcess(ACspSharingParallelProcess node) {
		analyseConcurrency(node);
    }
    
    public void caseACspAlphabetisedParallelProcess(ACspAlphabetisedParallelProcess node) {
    	analyseConcurrency(node);
    }
    
    public void caseACspInterleaveProcess(ACspInterleaveProcess node) {
    	analyseConcurrency(node);
    }
    
    private void analyseConcurrency(Node node){
        // Checking the branches
        Node left = null;
        Node right = null;

        if (node instanceof ACspSharingParallelProcess) {
            left = ((ACspSharingParallelProcess)node).getLeftProcess();
            right = ((ACspSharingParallelProcess)node).getRightProcess();
            //renameSyncSet((ACspSharingParallelProcess) node);
    	} else if (node instanceof ACspAlphabetisedParallelProcess) {
            left = ((ACspAlphabetisedParallelProcess)node).getLeftProcess();
            right = ((ACspAlphabetisedParallelProcess)node).getRightProcess();
    	} else if (node instanceof ACspInterleaveProcess) {
            left = ((ACspInterleaveProcess)node).getLeftProcess();
            right = ((ACspInterleaveProcess)node).getRightProcess();
    	}
        left.apply(this);
        if(this.currentBranch >= 0)
        	this.currentBranch += BranchUtil.getBranchesInProcess(procInfo, left);
        right.apply(this);
    }
    
    public void renameSyncSet(ACspSharingParallelProcess node) {
        PExpression syncExp = ((ACspSharingParallelProcess)node).getExpression();
        
        AlphabetsVisitor alphabets = new AlphabetsVisitor(this.procInfo, this.directives, this.constants);
        alphabets.setCurrentAnalysedProcess("");
        
        Set<String> channels = alphabets.analyseChannelSet(syncExp);

        Node left = ((ACspSharingParallelProcess)node).getLeftProcess();
        Node right = ((ACspSharingParallelProcess)node).getRightProcess();
        
        alphabets.resetChannels();
        left.apply(alphabets);
        Set<String> leftChannels = alphabets.getVisibleChannels();
        alphabets.resetChannels();

        right.apply(alphabets);
        Set<String> rightChannels = alphabets.getVisibleChannels();
        alphabets.resetChannels();

        rightChannels.retainAll(leftChannels); // rightUsedChannels intersection leftUsedChannels
        //rightChannels.retainAll(channels); // rightUsedChannels minus sync

        Map <String, List<String>> syncSet = new Hashtable<String, List<String>>();
        for (String c: rightChannels) {
        	int pos = c.indexOf("_BR_");
        	if (pos <= 0) pos = c.indexOf("_SYNC_");
			if (pos >= 0){
				String chan = c.substring(0, pos);
				if (channels.contains(chan)){
					if (syncSet.containsKey(chan)){
						List<String> list = syncSet.get(chan);
						list.add(c);
						syncSet.put(chan, list);
					} else {
						List<String> list = new Vector<String>();
						list.add(c);
						syncSet.put(chan, list);
					}
				}
			}
		}
        if (!syncSet.isEmpty()){
        	channels.removeAll(syncSet.keySet());
        	for (String chan : syncSet.keySet()) {
        		channels.addAll(syncSet.get(chan));
        	}
        	
        	node.setExpression(createACspBaseExpression(createPEventList(channels)));
        }
    }
    public void renameSyncSet(ACspAlphabetisedParallelProcess node) {
    	/*AlphabetsVisitor alphabets = new AlphabetsVisitor(this.procsInfo,this.directives, this.constants);
        alphabets.setCurrentAnalysedProcess(this.currentAnalysedProcess);

        // Checking the synchronisation channel set
        PExpression syncExpLeft = ((ACspAlphabetisedParallelProcess)node).getLeftExpr();
        PExpression syncExpRight = ((ACspAlphabetisedParallelProcess)node).getRightExpr();
        Set<String> channelsLeft = alphabets.analyseChannelSet(syncExpLeft);
        Set<String> channelsRight = alphabets.analyseChannelSet(syncExpRight);
        
        // Intersection
        channelsLeft.retainAll(channelsRight);
        
        return channelsLeft;*/
    	
    }
    
    private ACspEvent createACspEvent(String name){
    	return new ACspEvent(new TName(name), new Vector<PProjection>(), new Vector<PComunication>());
    }
    private ACspBaseEventList createACspBaseEventList(String name){
    	return new ACspBaseEventList(createACspEvent(name));
    }
    
    private PEventList createPEventList(Set<String> list){
    	PEventList evList = new ACspBaseEventList();
    	if (!list.isEmpty()){
    		Iterator<String> it = list.iterator();
    		evList = createACspBaseEventList(it.next());
    		while(it.hasNext()){
    			evList = new ACspLinkedEventList(createACspEvent(it.next()), new TComma(), evList);
    		}
    	}
    	return evList;
    }
    
    private ACspBaseExpression createACspBaseExpression(PEventList evList){
		//return new ACspBaseExpression(new ACspBaseInterleaveProcess(new ACspBasePara));
    	return new ACspBaseExpression(
			new ACspBaseInterleaveProcess(
			new ACspBaseParallelProcess(
			new ACspBaseChoiceProcess(
			new ACspBaseTimeoutProcess(
			new ACspBaseInterruptProcess(
			new ACspBaseSequentialProcess(
			new ACspBaseReplicatedProcess(
			new ACspBasePrefixProcess(
			new ACspBaseGuardedProcess(
			new ACspBaseHidingProcess(
			new ACspBaseRenamingProcess(
			new ACspBaseProcess(
			new ACspBaseExpr(
			new ACspBaseLogicalExpr(
			new ACspBaseRelationalExpr(
			new ACspBaseMathExpr(
			new ACspBaseMath1Expr(
			new ACspBaseSequenceExpr(
			new ACspBaseNotExpr(
			new ACspBaseMinusExpr(
			new ACspBaseLengthExpr(
			new ACspBaseDotExpr(
			new ACspBaseCallExpr(
			new ACspSetValue(
			new ACspProductionsSet(
					new TLProds(),
					evList,
					new TRProds()
			))))))))))))))))))))))))));
    }
}