package csp2hc.visitors;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import csp2hc.exception.UnsupportedConstructorException;
import csp2hc.messages.HandelC;
import csp2hc.util.CspUtil;

import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.node.ACspAlphabetisedParallelProcess;
import lmf.formula.csp.node.ACspEvent;
import lmf.formula.csp.node.ACspInterleaveProcess;
import lmf.formula.csp.node.ACspPrefixProcess;
import lmf.formula.csp.node.ACspProcessCallExpr;
import lmf.formula.csp.node.ACspProcessDefinition;
import lmf.formula.csp.node.ACspRenamingProcess;
import lmf.formula.csp.node.ACspSharingParallelProcess;
import lmf.formula.csp.node.ACspSubstitution;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.Start;
import lmf.formula.csp.node.TExtract;
import lmf.formula.csp.node.TName;

public class SubstitutionVisitor extends ExtendedDepthFirstAdapter {
	
	private int nextBranchId;
	
	private boolean isRenaming;
	
	private String renProcName;
	
	private Map<String,Map<Integer,List<ACspSubstitution>>> procSubstitutions;
	
	private List<List> changedSubstitution;
	
	private List<ACspSubstitution> substitutions;
	
	private DirectivesVisitor directives;
	
	private ChannelsVisitor channels;
	
	private ProcessesInfoVisitor procInfo;
	
	private String currentDefinition;
	
	private int parentBranch;
	
	private List<List> processParents;

	private boolean isConcurrency;
	
	String cspCode;
	
	Map<String,String> declaredChannels;
	
	public SubstitutionVisitor(DirectivesVisitor directives, ChannelsVisitor channels, ProcessesInfoVisitor procInfo){
		this.nextBranchId = 0;
		this.isRenaming = false;
		this.renProcName = "";
		this.procSubstitutions = new Hashtable<String,Map<Integer,List<ACspSubstitution>>>();
		this.changedSubstitution = new Vector<List>();
		this.directives = directives;
		this.channels = channels;
		this.procInfo = procInfo;
		this.currentDefinition = "";
		this.parentBranch = 0;
		this.processParents = new Vector<List>();
		this.cspCode = "";
		this.declaredChannels = new Hashtable<String, String>();
	}
	
	@Override
	public void caseStart(Start node) {
		node.getPSpecification().apply(this);
	}
	
	public void caseACspProcessDefinition(ACspProcessDefinition node) {
		String definition = node.getName().getText().trim();
		if (HandelC.MAINP.equals(definition)){
			this.currentDefinition = definition;
			this.parentBranch = 0;
			node.getExpression().apply(this);
		}
	}
	
	public void caseACspRenamingProcess(ACspRenamingProcess node) {
		this.isRenaming = true;
		node.getProcess().apply(this);
		this.isRenaming = false;
		this.substitutions = new Vector<ACspSubstitution>();
		node.getSubstitutionList().apply(this);
		
		List parent = new Vector();
		parent.add(this.renProcName);
		parent.add(nextBranchId);
		parent.add(currentDefinition);
		parent.add(parentBranch);
		this.processParents.add(parent);
		
		if (procSubstitutions.containsKey(this.renProcName)){
			Map<Integer,List<ACspSubstitution>> ps = procSubstitutions.get(this.renProcName);
			ps.put(nextBranchId, substitutions);
			procSubstitutions.put(renProcName, ps);
		} else {
			Map<Integer,List<ACspSubstitution>> ps = new Hashtable<Integer, List<ACspSubstitution>>();
			ps.put(nextBranchId, substitutions);
			procSubstitutions.put(renProcName, ps);
		}
	}
	
	public void caseACspProcessCallExpr(ACspProcessCallExpr node) {
		String pname = node.getName().getText().trim();
		isConcurrency = false;
		if (this.isRenaming)
			this.renProcName = pname;
		else{
			if (!pname.equals(CspUtil.getSkip()) 
	        		&& !pname.equals(CspUtil.getStop())
	        		&& !this.currentDefinition.equals(pname)) {
				String parentDefinition = this.currentDefinition;
				int parentBr = this.parentBranch;
				this.currentDefinition = pname;
				this.parentBranch = nextBranchId;
				this.procInfo.getProcDefinition(pname).apply(this);
				this.currentDefinition = parentDefinition;
				this.parentBranch = parentBr;
			}
		}
	}
	
	public void caseACspSubstitution(ACspSubstitution node) {
		this.substitutions.add((ACspSubstitution) node.clone());
	}
	
	public void caseACspSharingParallelProcess(ACspSharingParallelProcess node) {
        this.analyseConcurrency(node);
    }
    
    public void caseACspAlphabetisedParallelProcess(ACspAlphabetisedParallelProcess node) {
        this.analyseConcurrency(node);
    }
    
    public void caseACspInterleaveProcess(ACspInterleaveProcess node) {
        this.analyseConcurrency(node);
    }

    private void analyseConcurrency(Node node) {
        // Checking the branches
        Node left = null;
        Node right = null;
        if (node instanceof ACspSharingParallelProcess) {
            left = ((ACspSharingParallelProcess)node).getLeftProcess();
            right = ((ACspSharingParallelProcess)node).getRightProcess();
    	} else if (node instanceof ACspAlphabetisedParallelProcess) {
            left = ((ACspAlphabetisedParallelProcess)node).getLeftProcess();
            right = ((ACspAlphabetisedParallelProcess)node).getRightProcess();
    	} else if (node instanceof ACspInterleaveProcess) {
            left = ((ACspInterleaveProcess)node).getLeftProcess();
            right = ((ACspInterleaveProcess)node).getRightProcess();
    	}
        isConcurrency = true;
		left.apply(this);
		nextBranchId++;
		isConcurrency = true;
        right.apply(this);
    }
    @Override
    public void caseACspPrefixProcess(ACspPrefixProcess node) {
    	if (isRenaming)
    		throw new UnsupportedConstructorException("renaming with unnamed process", currentDefinition);
    	else if (isConcurrency)
    		throw new UnsupportedConstructorException("concurrency with unnamed process", currentDefinition);
    	else
    		node.getProcess().apply(this);
    }

	public boolean isRealSubstitution(String pname, int pbranch){
		if (this.procSubstitutions.containsKey(pname)){
			Map<Integer, List<ACspSubstitution>> ps = this.procSubstitutions.get(pname);
			if (ps.containsKey(pbranch)) return true;
		}
		return false;
	}
	
	public Map<String,List<String>> getSubstitutionList (String pname, int pbranch){
		Map<String,List<String>> subs = new Hashtable<String, List<String>>();
		if (isRealSubstitution(pname, pbranch)){
			for (ACspSubstitution node : this.procSubstitutions.get(pname).get(pbranch)) {
				String oldE = node.getOldEvent().toString().trim();
				String newE = node.getNewEvent().toString().trim();
				if (subs.containsKey(oldE)){
					List<String> s = subs.get(oldE);
					s.add(newE);
					subs.put(oldE, s);
				} else {
					List<String> s = new Vector<String>();
					s.add(newE);
					subs.put(oldE, s);
				}
			}
		}
		return subs;
	}
	
	public void changeSubstitutions(String pname, int pbranch, String cname, String suffix){
		if (isRealSubstitution(pname, pbranch)){
			for (ACspSubstitution node : this.procSubstitutions.get(pname).get(pbranch)) {
				String evName = ((ACspEvent) node.getNewEvent()).getChannelName().getText().trim();
				if (evName.equals(cname)){
					((ACspEvent) node.getNewEvent()).setChannelName(new TName(cname+suffix));
					Vector v = new Vector();
					v.add(pname);
					v.add(pbranch);
					this.changedSubstitution.add(v);
				}
			}
		}
	}
	public void changeSubstitutions(String pname, int pbranch, String cname, List<String> list){
		if (isRealSubstitution(pname, pbranch)){
			List<ACspSubstitution> toRemove = new Vector<ACspSubstitution>();
			List<ACspSubstitution> toAdd = new Vector<ACspSubstitution>();
			List<ACspSubstitution> toContinue = this.procSubstitutions.get(pname).get(pbranch); 
			for (ACspSubstitution node : toContinue) {
				String evName = ((ACspEvent) node.getNewEvent()).getChannelName().getText().trim();
				if (evName.equals(cname)){
					for (String suffix : list) {
						String oldEvName = ((ACspEvent) node.getOldEvent()).getChannelName().getText().trim();
						ACspEvent oldEvent = new ACspEvent();
						oldEvent.setChannelName(new TName(oldEvName));
						ACspEvent newEvent = new ACspEvent();
						newEvent.setChannelName(new TName(cname+suffix));
						toAdd.add(new ACspSubstitution(oldEvent, new TExtract(), newEvent));
					}
					toRemove.add(node);
				}
			}
			if (!toAdd.isEmpty()){
				toContinue.removeAll(toRemove);
				toContinue.addAll(toAdd);
				//Map<Integer,List<ACspSubstitution>> s = new Hashtable<Integer,List<ACspSubstitution>>();
				//Map<Integer,List<ACspSubstitution>> s = this.procSubstitutions.get(pname);
				//s.put(pbranch, toContinue);
				//this.procSubstitutions.put(pname, s);
				List l = new Vector();
				l.add(pname);
				l.add(pbranch);
				this.changedSubstitution.add(l);
			}
		}
	}
	public boolean isChangedSubst(String pname, int pbranch){
		boolean result = false;
		for (List element : this.changedSubstitution) {
			if (pname.equals(element.get(0)) && pbranch == Integer.parseInt(element.get(1).toString()))
				result = true;
		}
		return result;
	}
	
	public String getChangedSubstDecl (){
		//List<String> cspDeclarations = new Vector<String>();
		String cspDirectives = "";
		for (String pname : this.procSubstitutions.keySet()) {
			for (int pbranch : this.procSubstitutions.get(pname).keySet()) {
				if (isChangedSubst(pname, pbranch)){
					for (List<String> evts : getSubstEvents(this.procSubstitutions.get(pname).get(pbranch))) {
						String action = " ";
						if (this.directives.containsProcChannelInOut(pname)){
							Hashtable<String,String> actions = this.directives.getProcChannelInOut(pname);
							if (actions.containsKey(evts.get(0))){
								action = action + (actions.get(evts.get(0)).equals(HandelC.A_READ)?"in":"out");
							}
						}
						String channelType = "";
						if (this.channels.containsChannelType(evts.get(0))){
							List types = channels.getChannelType(evts.get(0));
							if (!types.isEmpty()){
								channelType = channelType + " : ";
								for (Object type:types){
									if (!channelType.equals(" : ")){
										channelType = channelType + " . ";
									}
									channelType = channelType + type.toString();
								}
							}
						}
						
						/*String cspDeclaration = "channel " + evts.get(1) + channelType + " \n";
						if (!cspDeclarations.contains(cspDeclaration))
							cspDeclarations.add(cspDeclaration);*/
						
						if (!this.declaredChannels.containsKey(evts.get(1))){
							this.declaredChannels.put(evts.get(1), channelType);
						}
						
						cspDirectives = cspDirectives + "--!! channel " + evts.get(1) + action  + " within " 
										+ getParent(pname, pbranch) + "_RN_" + getParentBranch(pname, pbranch) + " \n"; 
						cspDirectives = cspDirectives + "--!! channel " + evts.get(1) + action  + " within " 
										+ getParent(pname, pbranch) + " \n";
						
					}
				}
			}
		}
		for (String decl : this.declaredChannels.keySet()) {
			this.cspCode += "channel " + decl + this.declaredChannels.get(decl) + " \n";
		}
		this.cspCode = this.cspCode + cspDirectives;
		return this.cspCode;
	}
	
	public boolean isDeclaredChannel (String chan){
		return this.declaredChannels.containsKey(chan);
	}
	
	private Set <List<String>> getSubstEvents(List<ACspSubstitution> list){
		Set<List<String>> result = new HashSet<List<String>>();
		for (ACspSubstitution sub : list) {
			List l = new Vector<String>(2);
			l.add(((ACspEvent) sub.getOldEvent()).getChannelName().getText().trim());
			l.add(((ACspEvent) sub.getNewEvent()).getChannelName().getText().trim());
			result.add(l);
		}
		return result;
	}
	
	private String getParent(String pname, int pbranch){
		for (List group : this.processParents) {
			if (pname.equals(group.get(0)) && pbranch == (Integer)group.get(1))
				return (String)group.get(2);
		}
		return "";
	}
	
	private int getParentBranch(String pname, int pbranch){
		for (List group : this.processParents) {
			if (pname.equals(group.get(0)) && pbranch == (Integer)group.get(1))
				return (Integer)group.get(3);
		}
		return -1;
	}
}
