package csp2hc.core;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class MainMethodVisitor extends ExtendedDepthFirstAdapter {
    // Handel-C Code of the main method
    private String hcCode;
    
    // Visitor that has translation rules
    private ProcessesTranslatorVisitor translator;

    // Indicates if the system has been translated as a mutual recursion
    private boolean mutualRecursion;
    
    public MainMethodVisitor(ProcessesTranslatorVisitor translator, boolean mutualRec) {
        this.hcCode = "";
        this.translator = translator;
        this.mutualRecursion = mutualRec;
    }
    
    public String getHcCode(){ return this.hcCode; }
    public int getParallelismCounter(){ return this.translator.getParallelismCounter(); }
    
    public void caseACspDefinitionParagraph(ACspDefinitionParagraph node){
        this.hcCode = HandelC.generateComment(Messenger.MAIN_DECLARATION);
        this.hcCode = this.hcCode + HandelC.MAIN_DECL + HandelC.BLOCK_BEGIN;

        PDefinition definition = node.getDefinition();
        if (definition instanceof ACspProcessDefinition){
            this.translateProcessDefinition((ACspProcessDefinition)definition);
        } 

        this.hcCode = this.hcCode + HandelC.BLOCK_END;
    }
    
    // Translates the main process only
    private void translateProcessDefinition(ACspProcessDefinition definition) {
        String name = definition.getName().toString().trim();
        
        if (name.equals(HandelC.MAIN)) {
            translator.setCurrentDefinition(HandelC.MAIN);
            ACspParameters parameters = (ACspParameters)definition.getParameters();
            PExpression expression = definition.getExpression();
            expression.apply(translator);
        }
        this.hcCode = this.hcCode + translator.getHcCode();
    }
    
}
