package csp2hc.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.PushbackReader;

import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;

import lmf.formula.csp.lexer.Lexer;
import lmf.formula.csp.node.Start;
import lmf.formula.csp.parser.Parser;
import lmf.formula.csp.semantic.CspTypeChecker;

import csp2hc.util.*;

public class CspDisplay {

public static void main(String[] args) {
    JFileChooser jfc = new JFileChooser();
    Config config = Config.getInstance();
    String csp_files_dir = config.getConfig("CSP_FILES_DIR");
    jfc.setCurrentDirectory(new File(csp_files_dir));

        if (jfc.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
    		JFrame frame = new JFrame("CSP Syntax Trees");
            TreeView treeViewCst = new TreeView();
            JScrollPane scrollCst = new JScrollPane(treeViewCst);

            TreeView treeViewAst = new TreeView();
            JScrollPane scrollAst = new JScrollPane(treeViewAst);

            treeViewCst.setModel(new DefaultTreeModel(
            		new DefaultMutableTreeNode("Constructing the tree.")));
            treeViewAst.setModel(new DefaultTreeModel(
            		new DefaultMutableTreeNode("Constructing the tree.")));

            JTabbedPane tabs = new JTabbedPane();
            tabs.addTab("CST - Concrete Syntax Tree", scrollCst);
            tabs.addTab("AST - Abstract Syntax Tree", scrollAst);

            frame.getContentPane().add(tabs);

            frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    		frame.setSize(800, 400);
            frame.setVisible(true);

            try {
                FileInputStream in = new FileInputStream(jfc.getSelectedFile());
                InputStreamReader isr = new InputStreamReader(in);
                PushbackReader pb = new PushbackReader(isr, 1024);
                Lexer lexer = new Lexer(pb);
                Parser p = new Parser(lexer);

                Start tree = p.parse();

                tree.apply(new CspTypeChecker());
                tree.apply(new CspCstFactory(treeViewCst));
                tree.apply(new CspAstFactory(treeViewAst));
            } catch (Exception e) {
                treeViewCst.setRoot(new DefaultMutableTreeNode(e.getMessage()));
                treeViewAst.setRoot(new DefaultMutableTreeNode(e.getMessage()));
                e.printStackTrace();
            }
        }
	}

}