/*
 * StringUtil.java
 *
 * Created on 20 January 2006, 12:29
 */

package csp2hc.util;

import java.util.*;
import csp2hc.messages.*;

/**
 *
 * @author  marcel
 */
public class MathUtil {
    
    // Returns the number of bits needed to represent 
    // a cardinality number of elements
    public static int getNumberOfBits(int cardinality) {
        int answer = cardinality;
        if (answer > 0) {
            answer = (int)Math.ceil(Math.log(answer)/Math.log(2));
        }
        return answer;
    }

    // Returns the number of elements that can be represented with the given
    // numberOfBits
    public static int getCardinality(int numberOfBits) {
        int answer = numberOfBits;
        if (answer > 0) {
            answer = ((int)Math.pow(2,answer));
        }
        return answer;
    }
    
    // Represents the offset between the signed representation and the unsigned
    // one for a given numberOfBits that is used to represent the numbers
    public static int getOffset(int numberOfBits) {
        return MathUtil.getCardinality(numberOfBits) / 2;
    }

    // Returns the unsigned representation of a given signed number, given the 
    // numberOfBits used in the representation
    public static int getUnsignedInt(int signedInt, int numberOfBits) {
        int temp = (signedInt + MathUtil.getCardinality(numberOfBits) + MathUtil.getOffset(numberOfBits)) % MathUtil.getCardinality(numberOfBits);
        return temp;
    }

    /*
    // Returns the list containing the range of possible signed integers
    // that can be represented using the numberOfBits given
    public static List getSignedIntRange(int numberOfBits) {
        List numbers = new Vector();
        
        for (int i=0; i < MathUtil.getOffset(numberOfBits); i++) {
            numbers.add(new Integer(i));
        }

        for (int i = MathUtil.getOffset(numberOfBits); i>0; i--) {
            numbers.add(new Integer(0-i));
        }

        return numbers;
    }
     */

    // Returns the list containing the range of possible signed integers
    // that can be represented using the numberOfBits given
    public static List getSignedIntRangeOrderedBySignedRepresentation(int numberOfBits) {
        Vector positiveNumbers = new Vector();
        
        for (int i=0; i < MathUtil.getOffset(numberOfBits); i++) {
            positiveNumbers.add(new Integer(i));
        }
        
        Vector negativeNumbers = new Vector();
        for (int i = MathUtil.getOffset(numberOfBits); i>0; i--) {
            negativeNumbers.add(new Integer(0-i));
        }

        positiveNumbers.addAll(negativeNumbers);
        
        return positiveNumbers;
    }
    
    // Returns the list containing the range of possible unsigned integers
    // that can be represented using the numberOfBits given
    public static List getUnsignedIntRange(int numberOfBits) {
        List numbers = new Vector();
        
        for (int i=0; i < MathUtil.getCardinality(numberOfBits); i++) {
            numbers.add(new Integer(i));
        }
    
        return numbers;
    }

    public static String getIntName(int i) {
        String answer;
        if(i < 0) {
            answer = HandelC.NEG + Math.abs(i);
        } else {
            answer = i + "";
        }
        return answer;
    }
    
    
    
    public static void main(String args[]){
        int numberOfBits = 4;
        int cardinality = MathUtil.getCardinality(numberOfBits);
        numberOfBits = MathUtil.getNumberOfBits(cardinality);
        int offset = MathUtil.getOffset(numberOfBits);

        System.out.println("Number of bits = "+numberOfBits);
        System.out.println("Cardinality = "+cardinality);
        System.out.println("Offset = "+offset);
        
        System.out.println(MathUtil.getSignedIntRangeOrderedBySignedRepresentation(numberOfBits));
        
    }
    
}
