/*
 * HandelC.java
 *
 * Contains the Handel-C constructors used
 * Created on 18 January 2006, 18:17
 */

package csp2hc.messages;

/**
 *
 * @author  marcel oliveira
 */
public class HandelC {
    // File extensions
    public static String CSP_EXTENSION = "csp";
    public static String EXTENSION = "hcc";
    public static String DOT = ".";
    public static String TEMPORARY_FILE = "___temp";

    // Handel-C Keywords
    public static String CONSTANT = "const";
    public static String LINE_COMMENT = "//";
    public static String PAR_BEGIN = "(";
    public static String PAR_END = ")";
    public static String BLOCK_BEGIN = "{";
    public static String BLOCK_END = "}";
    public static String ARRAY_BEGIN = "[";
    public static String ARRAY_END = "]";
    public static String END_COMMAND = ";";
    public static String INPUT_COMMUNICATION = "?";
    public static String OUTPUT_COMMUNICATION = "!";
    public static String DELAY = "delay";
    public static String SPACE = " ";
    public static String SEP_ARGUMENTS = ",";
    public static String VOID = "void";
    public static String CHANNEL = "chan";
    public static String PARALLEL = "par";
    public static String SEQUENCE = "seq";
    public static String ALTERNATION = "prialt";
    public static String CASE = "case";
    public static String CASE_SEP = ":";
    public static String SWITCH = "switch";
    public static String DEFAULT = "default";
    public static String BREAK = "break";
    public static String TYPEDEF = "typedef";    
    public static String ENUM = "enum";    
    public static String UNSIGNED = "unsigned";    
    public static String ASSIGNMENT = "=";    
    public static String DEFINE = "#define";    
    public static String BINARY_PREFIX = "0b";    
    public static String WHILE = "while";    
    public static String IF = "if";    
    public static String CONDITIONAL_COND = "?";    
    public static String CONDITIONAL_SEP = ":";    
    public static String ELSE = "else";    
    public static String BOOLEAN = "boolean";    
    public static String LOGICAL_VALUES = "2";    
    public static String LOGICAL_TRUE = "true";    
    public static String LOGICAL_FALSE = "false";    
    public static String LOGICAL_AND = "&&";    
    public static String LOGICAL_OR = "||";    
    public static String LOGICAL_NOT = "!";    
    public static String LOGICAL_BITWISE_OR = "|";    
    public static String INT = "int";    
    public static String INTEGERL = "integer";    
    public static String MATH_ADD = "+";    
    public static String MATH_DIFF = "-";    
    public static String MATH_MULT = "*";    
    public static String MATH_POW = "**";    
    public static String MATH_DIV = "/";    
    public static String MATH_REMAINDER = "%";    
    public static String MATH_NEG = "-";    
    public static String MATH_ZERO = "0";    
    public static String MATH_ONE = "1";    
    public static String REL_EQUALS = "==";    
    public static String REL_NOT_EQUALS = "!=";    
    public static String REL_GREATER = ">";    
    public static String REL_GREATER_EQUALS = ">=";    
    public static String REL_LESS = "<";    
    public static String REL_LESS_EQUALS = "<=";    
    public static String MAIN = "main";
    public static String MACRO_EXP = "macro expr";
    public static String SELECT = "select";
    public static String INLINE = "inline";
    public static String STATIC = "static";
    public static String SEMAPHOR = "sema";
    public static String TRY_SEMAPHOR = "trysema";
    public static String RELEASE_SEMAPHOR = "releasesema";
    public static String REFERENCE = "*";
    public static String ADDRESS = "&";

    // Translator Keywords
    public static String SYNC = "sync";    
    public static String SYNC_OUT = "syncout";    
    public static String SYNC_IN = "syncin";    
    public static String UNDERSCORE = "_";
    public static String MACRO = "MACRO";
    public static String LOOKUP_TABLE = "LUT";
    public static String OFFSET = UNDERSCORE+"offset";
    public static String SET_SUFFIX = UNDERSCORE+"set";
    public static String EMPTY_SET_SUFFIX = "_nil";
    public static String TYPE_CARDINALITY = UNDERSCORE+"card";
    public static String NEG = "neg"+UNDERSCORE;    
    //public static String TYPE_SET = UNDERSCORE+"set";
    public static String LOCAL_COPY = UNDERSCORE+"local"+UNDERSCORE;
    public static String MUTUAL_RECURSION = "MUTUAL_REC";
    public static String PROGRAM_COUNTER = "PROGRAM_COUNTER";
    public static String KEEP_LOOPING = "KEEP_LOOPING";
    public static String CLOCK_DEF = "set clock = external \"clock1\" "+END_COMMAND;
    public static String COMMENT_LINE_BREAK = UNDERSCORE+UNDERSCORE+"LINE_BREAK";    
    public static String COMMENT_TAB = UNDERSCORE+UNDERSCORE+"TAB";    
    public static String COMMENT_SPACES = UNDERSCORE+UNDERSCORE+"SPACES"+UNDERSCORE;    
    public static String SYNC_TYPE = SYNC.toUpperCase();
    public static String SYNC_DEF = 
        TYPEDEF + SPACE + UNSIGNED + SPACE + INT + SPACE + MATH_ONE 
        + SPACE + SYNC_TYPE + END_COMMAND;    
    public static String SYNC_OUT_DEF = 
        CONSTANT + SPACE + SYNC_TYPE + SPACE + SYNC_OUT + SPACE + ASSIGNMENT 
        + MATH_ZERO + END_COMMAND;    
    public static String SYNC_IN_DEF = SYNC_TYPE + SPACE + SYNC_IN + END_COMMAND;    
    public static String TRUE_DEF = LOGICAL_TRUE + SPACE + MATH_ONE;    
    public static String FALSE_DEF = LOGICAL_FALSE + SPACE + MATH_ZERO;    
    public static String MAIN_DECL = VOID + SPACE + MAIN + PAR_BEGIN + PAR_END;
    public static String MUTUAL_REC_INDEX = MUTUAL_RECURSION + UNDERSCORE + "i";
    public static String SINGLETON = "singleton";
    public static String CHANNELS = "channels";
    public static String CHANNEL_MEMBER = "CHANNEL_MEMBER";
    public static String FORCED_SYNC = "FORCED_SYNC";
    public static String SET_UNION = "SET_UNION";
    public static String SEMAPHOR_SUFFIX = "_SEMA";
    public static String DONE_EVENTS = "DONE_EVENTS";
    public static String IS_EMPTY_CHANNEL_SET = "IS_EMPTY_CHANNEL_SET";
    public static String ENTER_SEMAPHOR = "ENTER_SEMAPHOR";
    public static String LEAVE_SEMAPHOR = "LEAVE_SEMAPHOR";
    /*
    public static String LEFT = "LEFT";
    public static String RIGHT = "RIGHT";
    public static String ENTER = "ENTER";
    public static String LEAVE = "LEAVE";
    public static String KILL = "KILL";
	*/
    public static String CHANNEL_TYPE = "CHANNEL_TYPE";
    
    public static String generateComment(String comment) {
        int length = comment.length();
        String beautyLine = HandelC.LINE_COMMENT + HandelC.SPACE;
        for (int i=0; i<length; i++) {
            beautyLine = beautyLine + HandelC.MATH_DIFF;
        }
        String commentLine = HandelC.LINE_COMMENT + HandelC.SPACE + comment;
        
        String result = HandelC.COMMENT_LINE_BREAK + beautyLine 
                        + HandelC.COMMENT_LINE_BREAK + commentLine 
                        + HandelC.COMMENT_LINE_BREAK + beautyLine 
                        + HandelC.COMMENT_LINE_BREAK;
        
        return result;
    }

    public static String generateDefine(String definition) {
        String define = HandelC.DEFINE + HandelC.SPACE + definition + HandelC.COMMENT_LINE_BREAK;
        return define;
    }
    
}
