package lmf.formula.csp.semantic;

import java.util.*;

import lmf.formula.csp.type.*;


	/// <summary>
	/// Summary description for CspClosureOperations.
	/// </summary>
	public class CspClosure
	{

		/// <summary>
		/// The set of values which will complete x.
		/// </summary>
		/// <param name="x">CspType</param>
		/// <returns>The set of values which will complete x.</returns>
		public ArrayList Extensions(CspType x)
		{
			return new ArrayList();
		}

		/// <summary>
		/// The set of values which begin with x.
		/// 
		/// Equivalences:
		/// 1 -  productions(x) =  { x.z | z<-extensions(x)}
		/// 2 -  {| x | ... |} = Union({ productions(x) | ... })
		/// </summary>
		/// <param name="x">CspType</param>
		/// <returns>The set of values which begin with x.</returns>
		public ArrayList Productions(CspType x)
		{
			ArrayList productions = new ArrayList();

			Iterator it = Extensions(x).iterator();
			CspType z;
			while (it.hasNext()) {
				z = (CspType) it.next();
				if (!(z instanceof CspDataType || z instanceof CspSubType
					|| z instanceof CspNameType))
				{
					productions.addAll(Productions(z));
				}
				else
				{
					productions.add(x + "." + z);
				}
			}

			return productions;
		}

	}