package lmf.formula.csp.semantic;

import java.util.Iterator;
import java.util.Stack;

import lmf.formula.csp.node.Node;
import lmf.formula.csp.semantic.exception.CspSemanticException;


/**
 * 
 * @author Joabe Jesus
 */
public class CspContextualTable {

	private Stack envsStack;

	private CspEnvironment env;

	public CspContextualTable() {
		this.envsStack = new Stack();
	}

	public void OpenScope(Node node) {
		if (node != null) {
			this.env = new CspEnvironment(node);
		}
	}

	public void openSubScope(Node node) {
		if (node != null) {
			this.envsStack.push(this.env);
			this.env = new CspEnvironment(node, this.env);
		}
	}

	public int getScopeLevel() {
		return this.envsStack.size() + 1;
	}

	public CspDeclaration getDeclaration(String name) {
		CspDeclaration declaration = null;

		if (this.env != null && name != null) {
			boolean found = false;
			Iterator it = this.env.getDeclarations();

			while (!found && it.hasNext()) {
				declaration = (CspDeclaration) it.next();
				if (declaration.getName().equals(name.trim())) {
					found = true;
				}
			}

			if (!found) {
				declaration = null;
			}
		}

		return declaration;
	}

	/**
	 * Binds a declaration in the current environment.
	 * 
	 * @param declaration
	 */
	public void bind(CspDeclaration declaration) {
		if (!this.env.isNameBound(declaration)) {
			// sets the scopeLevel of the entry
			declaration.setScopeLevel(this.getScopeLevel());
			// adds the entry to the environment
			this.env.Bind(declaration);
		} else {
			throw new CspSemanticException("Identifier '"
					+ declaration.getName() + "' already declared in this scope.");
		}
	}

	public void unbind(CspDeclaration declaration) {
		if (this.env != null && declaration != null) {
			this.env.unbind(declaration);
		}
	}

	public void closeScope() {
		if (this.envsStack.size() != 0) {
			this.env = (CspEnvironment) this.envsStack.pop();
		} else {
			this.env = null;
		}
	}

}