package lmf.formula.csp.semantic;

import java.util.ArrayList;

import lmf.formula.csp.analysis.DepthFirstAdapter;
import lmf.formula.csp.node.ACspAbstractDefinition;
import lmf.formula.csp.node.ACspChannel;
import lmf.formula.csp.node.ACspComplexBranch;
import lmf.formula.csp.node.ACspDatatypeDefinition;
import lmf.formula.csp.node.ACspNametypeDefinition;
import lmf.formula.csp.node.ACspSimpleBranch;
import lmf.formula.csp.node.ACspSubtypeDefinition;
import lmf.formula.csp.type.CspType;
import lmf.formula.csp.type.CspTypeFactory;


/**
 * 
 * @author Joabe Jesus
 */
public class CspIdentification extends DepthFirstAdapter {

	private CspContextualTable contextualTable;

	private CspTypeFactory typeFactory;

	private CspType branchType;

	public CspIdentification(CspContextualTable contextualTable,
			CspTypeFactory typeFactory) {
		this.contextualTable = contextualTable;
		this.typeFactory = typeFactory;
	}

	// DATATYPE
	public void inACspDatatypeDefinition(ACspDatatypeDefinition node) {
		String name = node.getName().getText();

		CspDeclaration decl = new CspDatatypeDeclaration(name, null,
				this.typeFactory.DataType(name));
		decl.setNode(node);
		this.contextualTable.bind(decl);

		this.branchType = decl.getCspType();
	}

	// SUBTYPE
	public void inACspSubtypeDefinition(ACspSubtypeDefinition node) {
		String name = node.getName().getText();

		CspDeclaration decl = new CspSubtypeDeclaration(name, this.typeFactory
				.SubType(name));
		decl.setNode(node);
		this.contextualTable.bind(decl);

		this.branchType = decl.getCspType();
	}

	// BRANCHES (TAGS)
	public void caseACspComplexBranch(ACspComplexBranch node) {
		String name = node.getTagName().getText();

		CspDeclaration decl = new CspTagDeclaration(name, this.branchType);
		decl.setNode(node);
		this.contextualTable.bind(decl);
	}

	public void CaseACspSimpleBranch(ACspSimpleBranch node) {
		String name = node.getTagName().getText();

		CspDeclaration decl = new CspTagDeclaration(name, this.branchType);
		decl.setNode(node);
		this.contextualTable.bind(decl);
	}

	// NAMETYPE
	public void CaseACspNametypeDefinition(ACspNametypeDefinition node) {
		String name = node.getName().getText();

		CspDeclaration decl = new CspNametypeDeclaration(name, this.typeFactory
				.NameType(name));
		decl.setNode(node);
		this.contextualTable.bind(decl);
	}

	// CHANNEL
	public void caseACspChannel(ACspChannel node) {
		String name = node.getName().getText();

		CspDeclaration decl = new CspChannelDeclaration(name, this.typeFactory
				.getAnyType());
		decl.setNode(node);
		this.contextualTable.bind(decl);
	}

	// ABSTRACT DEFINITION
	public void caseACspAbstractDefinition(ACspAbstractDefinition node) {
		String name = node.getName().getText();

		CspDeclaration decl = this.contextualTable.getDeclaration(name);

		if (!(decl instanceof CspFunctionDeclaration)) {
			decl = new CspFunctionDeclaration(name, new ArrayList(),
					this.typeFactory.getAnyType());
			decl.setNode(node);
			this.contextualTable.bind(decl);
		}
	}

}