package lmf.formula.csp.semantic;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Stack;

import lmf.formula.csp.analysis.DepthFirstAdapter;
import lmf.formula.csp.node.ACspAbstractDefinition;
import lmf.formula.csp.node.ACspAddMathExpr;
import lmf.formula.csp.node.ACspAlphabetisedParallelProcess;
import lmf.formula.csp.node.ACspAlphabetisedReplicatedProcess;
import lmf.formula.csp.node.ACspAndLogicalExpr;
import lmf.formula.csp.node.ACspBaseArgumentList;
import lmf.formula.csp.node.ACspBoolType;
import lmf.formula.csp.node.ACspBooleanValue;
import lmf.formula.csp.node.ACspCallPostfixExpr;
import lmf.formula.csp.node.ACspCatSequenceExpr;
import lmf.formula.csp.node.ACspChannel;
import lmf.formula.csp.node.ACspChannelDefinition;
import lmf.formula.csp.node.ACspChaosProcess;
import lmf.formula.csp.node.ACspClosedRange;
import lmf.formula.csp.node.ACspComplexBranch;
import lmf.formula.csp.node.ACspCompositeType;
import lmf.formula.csp.node.ACspComprehensionSequence;
import lmf.formula.csp.node.ACspComprehensionSet;
import lmf.formula.csp.node.ACspConstantDefinition;
import lmf.formula.csp.node.ACspConstrainedInputComunication;
import lmf.formula.csp.node.ACspConstraint;
import lmf.formula.csp.node.ACspDatatypeDefinition;
import lmf.formula.csp.node.ACspDeclaration;
import lmf.formula.csp.node.ACspDefinitionParagraph;
import lmf.formula.csp.node.ACspDiffMathExpr;
import lmf.formula.csp.node.ACspDivMath1Expr;
import lmf.formula.csp.node.ACspDotPostfixExpr;
import lmf.formula.csp.node.ACspEmptySequence;
import lmf.formula.csp.node.ACspEmptySet;
import lmf.formula.csp.node.ACspEqualRelationalExpr;
import lmf.formula.csp.node.ACspEvent;
import lmf.formula.csp.node.ACspExternalChoiceProcess;
import lmf.formula.csp.node.ACspExternalReplicatedProcess;
import lmf.formula.csp.node.ACspFunctionDefinition;
import lmf.formula.csp.node.ACspGtRelationalExpr;
import lmf.formula.csp.node.ACspGteRelationalExpr;
import lmf.formula.csp.node.ACspGuardedProcess;
import lmf.formula.csp.node.ACspHidingProcess;
import lmf.formula.csp.node.ACspIfExpr;
import lmf.formula.csp.node.ACspInputComunication;
import lmf.formula.csp.node.ACspIntType;
import lmf.formula.csp.node.ACspInterleaveProcess;
import lmf.formula.csp.node.ACspInterleaveReplicatedProcess;
import lmf.formula.csp.node.ACspInternalChoiceProcess;
import lmf.formula.csp.node.ACspInternalReplicatedProcess;
import lmf.formula.csp.node.ACspInterruptProcess;
import lmf.formula.csp.node.ACspLambdaExpr;
import lmf.formula.csp.node.ACspLengthPrefixExpr;
import lmf.formula.csp.node.ACspLetExpr;
import lmf.formula.csp.node.ACspLink;
import lmf.formula.csp.node.ACspLinkedArgumentList;
import lmf.formula.csp.node.ACspLinkedParallelProcess;
import lmf.formula.csp.node.ACspLinkedReplicatedProcess;
import lmf.formula.csp.node.ACspLtRelationalExpr;
import lmf.formula.csp.node.ACspLteRelationalExpr;
import lmf.formula.csp.node.ACspMinusPrefixExpr;
import lmf.formula.csp.node.ACspMulMath1Expr;
import lmf.formula.csp.node.ACspNametypeDefinition;
import lmf.formula.csp.node.ACspNotEqualRelationalExpr;
import lmf.formula.csp.node.ACspNotPrefixExpr;
import lmf.formula.csp.node.ACspNumberValue;
import lmf.formula.csp.node.ACspOpenRange;
import lmf.formula.csp.node.ACspOrLogicalExpr;
import lmf.formula.csp.node.ACspOutputComunication;
import lmf.formula.csp.node.ACspParam;
import lmf.formula.csp.node.ACspPrefixProcess;
import lmf.formula.csp.node.ACspProcessDefinition;
import lmf.formula.csp.node.ACspProductionsSet;
import lmf.formula.csp.node.ACspReferencedType;
import lmf.formula.csp.node.ACspRemainderMath1Expr;
import lmf.formula.csp.node.ACspRenamingProcess;
import lmf.formula.csp.node.ACspSequence;
import lmf.formula.csp.node.ACspSequentialProcess;
import lmf.formula.csp.node.ACspSequentialReplicatedProcess;
import lmf.formula.csp.node.ACspSet;
import lmf.formula.csp.node.ACspSharingParallelProcess;
import lmf.formula.csp.node.ACspSharingReplicatedProcess;
import lmf.formula.csp.node.ACspSimpleBranch;
import lmf.formula.csp.node.ACspSkipProcess;
import lmf.formula.csp.node.ACspSourceConstraint;
import lmf.formula.csp.node.ACspStopProcess;
import lmf.formula.csp.node.ACspSubstitution;
import lmf.formula.csp.node.ACspSubtypeDefinition;
import lmf.formula.csp.node.ACspTupleValue;
import lmf.formula.csp.node.ACspUntimedTimeoutProcess;
import lmf.formula.csp.node.ACspVariable;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.PArgumentList;
import lmf.formula.csp.node.PComunication;
import lmf.formula.csp.node.Start;
import lmf.formula.csp.node.Token;
import lmf.formula.csp.semantic.exception.CspSemanticException;
import lmf.formula.csp.semantic.exception.CspTypeException;
import lmf.formula.csp.type.CspBoolType;
import lmf.formula.csp.type.CspIntRangeType;
import lmf.formula.csp.type.CspIntType;
import lmf.formula.csp.type.CspParametrizedType;
import lmf.formula.csp.type.CspProcessType;
import lmf.formula.csp.type.CspSequenceType;
import lmf.formula.csp.type.CspSetType;
import lmf.formula.csp.type.CspTupleType;
import lmf.formula.csp.type.CspType;
import lmf.formula.csp.type.CspTypeFactory;


import csp2hc.util.LoggedStack;

/**
 * 
 * @author Joabe Jesus
 */
public class CspTypeChecker extends DepthFirstAdapter {

	private CspContextualTable contextualTable;

	private CspTypeFactory typeFactory;

	private CspType expectedType;

	private LoggedStack exprTypes;

	public CspType typeOf(String name) {
		CspType ret = null;

		CspDeclaration decl = this.contextualTable.getDeclaration(name);
		if (decl != null) {
			ret = decl.getCspType();
		}

		return ret;
	}

	private Token lastToken;

	public void defaultcase(Node node) {
		if (node instanceof Token) {
			this.lastToken = (Token) node;
		}
	}

	private String getErrorLocation() {
		int line = 0, pos = 0;
		if (lastToken != null) {
			line = this.lastToken.getLine();
			pos = this.lastToken.getPos();
		}
		return "[" + line + "," + pos + "]";
	}

	// / START
	public void inStart(Start node) {
		this.contextualTable = new CspContextualTable();

		this.typeFactory = new CspTypeFactory();
		this.exprTypes = new LoggedStack(false);

		this.contextualTable.OpenScope(node);

		this.BindPrelude();

		CspIdentification identification = new CspIdentification(
				this.contextualTable, this.typeFactory);
		node.apply(identification);
	}

	// / <summary>
	// / Binding the pre-defined CSP names.
	// / </summary>
	private void BindPrelude() {
		CspDeclaration decl;
		ArrayList paramTypes;

		/*
		 * decl = new CspNametypeDeclaration("int",
		 * this.typeFactory.getIntType()); this.contextualTable.Bind(decl); decl =
		 * new CspNametypeDeclaration("Bool", this.typeFactory.BoolType);
		 * this.contextualTable.Bind(decl); decl = new
		 * CspNametypeDeclaration("true", this.typeFactory.BoolType);
		 * this.contextualTable.Bind(decl); decl = new
		 * CspNametypeDeclaration("false", this.typeFactory.BoolType);
		 * this.contextualTable.Bind(decl);
		 * 
		 * decl = new CspFunctionDeclaration("SKIP",
		 * this.typeFactory.getProcessType()); this.contextualTable.Bind(decl);
		 * decl = new CspFunctionDeclaration("STOP",
		 * this.typeFactory.getProcessType()); this.contextualTable.Bind(decl);
		 * decl = new CspFunctionDeclaration("CHAOS",
		 * this.typeFactory.getProcessType()); this.contextualTable.Bind(decl);
		 */

		paramTypes = new ArrayList();
		paramTypes.add(this.typeFactory.getAnyType());
		decl = new CspFunctionDeclaration("union", paramTypes, this.typeFactory
				.getSetType());
		this.contextualTable.bind(decl);

		paramTypes = new ArrayList();
		paramTypes.add(this.typeFactory.getAnyType());
		decl = new CspFunctionDeclaration("intersect", paramTypes,
				this.typeFactory.getSetType());
		this.contextualTable.bind(decl);
	}

	public void outACspDefinitionParagraph(ACspDefinitionParagraph node) {
		if (node.getDefinition() instanceof ACspAbstractDefinition) {
			ACspAbstractDefinition absDef = (ACspAbstractDefinition) node.getDefinition();

			CspFunctionDeclaration funDecl = (CspFunctionDeclaration)
				this.contextualTable.getDeclaration(absDef.getName().getText());

			if (funDecl.getKind() == CspFunctionKind.CONSTANT) {
				node.setDefinition(new ACspConstantDefinition(absDef.getName(),
						absDef.getDefs(), absDef.getExpression()));
			} else if (funDecl.getKind() == CspFunctionKind.FUNCTION) {
				node.setDefinition(new ACspFunctionDefinition(absDef.getName(),
						absDef.getParameters(), absDef.getDefs(), absDef.getExpression()));
			} else if (funDecl.getKind() == CspFunctionKind.PROCESS) {
				node.setDefinition(new ACspProcessDefinition(absDef.getName(),
						absDef.getParameters(), absDef.getDefs(), absDef.getExpression()));
			}
		}
	}

	// DATATYPE
	// TODO
	public void outACspDatatypeDefinition(ACspDatatypeDefinition node) {

	}

	// BRANCH
	// TODO
	public void outACspComplexBranch(ACspComplexBranch node) {
		//String tagName = node.getTagName().getText();
	}

	public void outACspSimpleBranch(ACspSimpleBranch node) {
		//String tagName = node.getTagName().getText();
	}

	// SUBTYPE
	// TODO
	public void outACspSubtypeDefinition(ACspSubtypeDefinition node) {
		String name = node.getName().getText();
		this.contextualTable.getDeclaration(name);
	}

	// NAMETYPE
	public void outACspNametypeDefinition(ACspNametypeDefinition node) {
		String name = node.getName().getText();
		this.contextualTable.getDeclaration(name);
	}

	// CHANNEL
	public void caseACspChannelDefinition(ACspChannelDefinition node) {
		inACspChannelDefinition(node);

		if (node.getTypeSpecifier() != null) {
			node.getTypeSpecifier().apply(this);
			this.exprTypes.push(this.exprTypes.pop());
		} else {
			this.exprTypes.push(this.typeFactory.getUntypedChannelType());
		}

		if (node.getChannelList() != null) {
			node.getChannelList().apply(this);
		}
		if (node.getChannel() != null) {
			node.getChannel().apply(this);
		}
		outACspChannelDefinition(node);
	}

	public void outACspChannelDefinition(ACspChannelDefinition node) {
            this.exprTypes.pop();
	}

	public void outACspChannel(ACspChannel node) {
            String name = node.getName().getText();

            CspDeclaration decl = this.contextualTable.getDeclaration(name);

            decl.setCspType((CspType) this.exprTypes.peek());
        }

	// ABSTRACT DEFINITION
	public void inACspAbstractDefinition(ACspAbstractDefinition node) {
		this.expectedType = this.typeFactory.getAnyType();
	}

	public void outACspAbstractDefinition(ACspAbstractDefinition node) {
		String name = node.getName().getText();
                
		CspType type = (CspType) this.exprTypes.pop();

		// Function, Constant or Process
		CspFunctionDeclaration funDecl = (CspFunctionDeclaration) this.contextualTable
				.getDeclaration(name);

		funDecl.setCspType(type);
		// Console.WriteLine("Name: {0} - Type: {1}", name, type);

		if (type instanceof CspProcessType) {
			funDecl.setKind(CspFunctionKind.PROCESS);
		} else if (type instanceof CspParametrizedType) {
			funDecl.setKind(CspFunctionKind.FUNCTION);
		} else {
			funDecl.setKind(CspFunctionKind.CONSTANT);
		}

                ArrayList parameters = this.getParametersTypes(node);
		ArrayList declParams = funDecl.getParamTypes();
		if (parameters.size() == declParams.size()) {
			// Pattern matching
		} else {
			// New declaration with the same name? PROBE accepts!!!
		}
	}

	private ArrayList getParametersTypes(ACspAbstractDefinition node) {
		ArrayList parameters = new ArrayList();
		if (node.getParameters() != null) {
                    CspType paramsTypes = (CspType) this.exprTypes.pop();

                    if (paramsTypes instanceof CspTupleType) {
                        parameters.addAll(((CspTupleType) paramsTypes).getTypes());
                    } else {
                        parameters.add(paramsTypes);
                    }
		}
		return parameters;
	}

	// COMPOSITE TYPE
	public void inACspCompositeType(ACspCompositeType node) {
		this.expectedType = this.typeFactory.getAnyType();
	}

	public void outACspCompositeType(ACspCompositeType node) {
		CspType rightType = (CspType) this.exprTypes.pop();
		CspType leftType = (CspType) this.exprTypes.pop();

		this.exprTypes.push(this.typeFactory.CompositeType(leftType, rightType));
	}

	// REFERENCED TYPE
	public void outACspReferencedType(ACspReferencedType node) {
		String name = node.getName().getText();

		CspDeclaration decl = this.contextualTable.getDeclaration(name);
		if (decl instanceof CspTypeDeclaration) {
			if (decl.getCspType() == this.typeFactory.getAnyType()) {
				decl.setCspType(this.expectedType);
			}
			this.exprTypes.push(decl.getCspType());
		} else if (decl instanceof CspFunctionDeclaration) {
			CspFunctionDeclaration funDecl = (CspFunctionDeclaration) decl;
			if (funDecl.getKind() == CspFunctionKind.CONSTANT) {
				;
			}
		} else if (decl != null) {
			throw new CspSemanticException(getErrorLocation()
					+ " - Referenced name: '" + name + "' is not a Type.");
		} else {
			throw new CspSemanticException(getErrorLocation()
					+ " - Referenced Type: '" + name + "' was not declared.");
		}
	}

	// PRIMITIVE TYPES
	public void outACspBoolType(ACspBoolType node) {
		this.exprTypes.push(this.typeFactory.getBoolType());
	}

	public void outACspIntType(ACspIntType node) {
		this.exprTypes.push(this.typeFactory.getIntType());
	}

	// LET EXPRESSION
	public void inACspLetExpr(ACspLetExpr node) {
		this.contextualTable.openSubScope(node);
	}

	public void outACspLetExpr(ACspLetExpr node) {
		this.contextualTable.closeScope();
	}

	// DECLARATION
	// TODO
	public void outACspDeclaration(ACspDeclaration node) {
		;
	}

	// LAMBDA EXPRESSION
	// TODO
	public void inACspLambdaExpr(ACspLambdaExpr node) {
		this.contextualTable.openSubScope(node);
	}

	public void outACspLambdaExpr(ACspLambdaExpr node) {
		this.contextualTable.closeScope();
	}

	// PARAM
	// TODO
	public void caseACspParam(ACspParam node) {
		try {
                    super.caseACspParam(node);
		} catch (CspSemanticException e) {
			if (e.getMessage().startsWith("")) {
				String name = node.toString();
				CspDeclaration decl = new CspVariableDeclaration(name, null);
				this.contextualTable.bind(decl);
			}
		}
	}

	// IF EXPRESSION
	public void caseACspIfExpr(ACspIfExpr node) {
		inACspIfExpr(node);
		if (node.getIf() != null) {
			node.getIf().apply(this);
		}
		this.expectedType = this.typeFactory.getBoolType();
		if (node.getIfBranch() != null) {
			node.getIfBranch().apply(this);
		}
		if (node.getThen() != null) {
			node.getThen().apply(this);
		}
		this.expectedType = this.typeFactory.getAnyType();
		if (node.getThenBranch() != null) {
			node.getThenBranch().apply(this);
		}
		if (node.getElse() != null) {
			node.getElse().apply(this);
		}
		if (node.getElseBranch() != null) {
			node.getElseBranch().apply(this);
		}
		outACspIfExpr(node);
	}

	public void outACspIfExpr(ACspIfExpr node) {
		CspType elseType = (CspType) this.exprTypes.pop();
		CspType thenType = (CspType) this.exprTypes.pop();
		CspType condType = (CspType) this.exprTypes.pop();

		if (condType instanceof CspBoolType) {
			if (elseType == thenType) {
				this.exprTypes.push(thenType);
			} else {
				throw new CspTypeException(getErrorLocation(), thenType
						.toString(), elseType);
			}
		} else {
			throw new CspTypeException(getErrorLocation(), "Bool", condType);
		}
	}

	// UTIL
	private void checkUnary(Token op, CspType resultType) {
		this.lastToken = op;

		CspType AnyType = this.typeFactory.getAnyType();

		CspType type = (CspType) this.exprTypes.pop();
		if (expectedType == AnyType || type == expectedType) {
			this.exprTypes.push(type);
		} else {
			throw new CspTypeException(getErrorLocation(), expectedType
					.toString(), type);
		}
	}

	private void checkBinary(Token op, CspType resultType) {
		this.lastToken = op;

		CspType AnyType = this.typeFactory.getAnyType();

		CspType rightType = (CspType) this.exprTypes.pop();
		CspType leftType = (CspType) this.exprTypes.pop();

		if (expectedType == AnyType || leftType == expectedType) {
			if (expectedType == AnyType || rightType == expectedType) {
				this.exprTypes.push(resultType);
			} else {
				throw new CspTypeException(getErrorLocation(), expectedType
						.toString(), rightType);
			}
		} else {
			throw new CspTypeException(getErrorLocation(), expectedType
					.toString(), leftType);
		}
	}

	// PROCESSES

	// INTERLEAVE PROCESS
	public void inACspInterleaveProcess(ACspInterleaveProcess node) {
		this.expectedType = this.typeFactory.getProcessType();
	}

	public void outACspInterleaveProcess(ACspInterleaveProcess node) {
		checkBinary(node.getInterleave(), this.typeFactory.getProcessType());
	}

	// SHARING PARALLEL PROCESS
	public void inACspSharingParallelProcess(ACspSharingParallelProcess node) {
		this.expectedType = this.typeFactory.getProcessType();
	}

	public void outACspSharingParallelProcess(ACspSharingParallelProcess node) {
		CspType rightProcessType = (CspType) this.exprTypes.pop();
		CspType exprType = (CspType) this.exprTypes.pop();
		CspType leftProcessType = (CspType) this.exprTypes.pop();

		this.exprTypes.push(leftProcessType);
		this.exprTypes.push(rightProcessType);
		checkBinary(node.getRParallel(), this.typeFactory.getProcessType());
	}

	// LINKED PARALLEL PROCESS
	public void inACspLinkedParallelProcess(ACspLinkedParallelProcess node) {
		this.expectedType = this.typeFactory.getProcessType();
	}

	public void outACspLinkedParallelProcess(ACspLinkedParallelProcess node) {
		checkBinary(node.getLBracket(), this.typeFactory.getProcessType());
	}

	// LINK
	// TODO
	public void outACspLink(ACspLink node) {
	}

	// ALPHABETISED PARALLEL PROCESS
	public void inACspAlphabetisedParallelProcess(
			ACspAlphabetisedParallelProcess node) {
		this.expectedType = this.typeFactory.getProcessType();
	}

	public void outACspAlphabetisedParallelProcess(
			ACspAlphabetisedParallelProcess node) {
		CspType rightProcessType = (CspType) this.exprTypes.pop();
		CspType rightExprType = (CspType) this.exprTypes.pop();
		CspType leftExprType = (CspType) this.exprTypes.pop();
		CspType leftProcessType = (CspType) this.exprTypes.pop();

		this.exprTypes.push(leftProcessType);
		this.exprTypes.push(rightProcessType);
		checkBinary(node.getParallel(), this.typeFactory.getProcessType());
	}

	// INTERNAL CHOICE PROCESS
	public void inACspInternalChoiceProcess(ACspInternalChoiceProcess node) {
		this.expectedType = this.typeFactory.getProcessType();
	}

	public void outACspInternalChoiceProcess(ACspInternalChoiceProcess node) {
		checkBinary(node.getInternalChoice(), this.typeFactory.getProcessType());
	}

	// EXTERNAL CHOICE PROCESS
	public void inACspExternalChoiceProcess(ACspExternalChoiceProcess node) {
		this.expectedType = this.typeFactory.getProcessType();
	}

	public void outACspExternalChoiceProcess(ACspExternalChoiceProcess node) {
		checkBinary(node.getExternalChoice(), this.typeFactory.getProcessType());
	}

	// UNTIMED TIMEOUT PROCESS
	public void inACspUntimedTimeoutProcess(ACspUntimedTimeoutProcess node) {
		this.expectedType = this.typeFactory.getProcessType();
	}

	public void outACspUntimedTimeoutProcess(ACspUntimedTimeoutProcess node) {
		checkBinary(node.getUntimedTimeout(), this.typeFactory.getProcessType());
	}

	// INTERRUPT PROCESS
	public void inACspInterruptProcess(ACspInterruptProcess node) {
		this.expectedType = this.typeFactory.getProcessType();
	}

	public void outACspInterruptProcess(ACspInterruptProcess node) {
		checkBinary(node.getInterrupt(), this.typeFactory.getProcessType());
	}

	// SEQUENTIAL PROCESS
	public void inACspSequentialProcess(ACspSequentialProcess node) {
		this.expectedType = this.typeFactory.getProcessType();
	}

	public void outACspSequentialProcess(ACspSequentialProcess node) {
		checkBinary(node.getSemicolon(), this.typeFactory.getProcessType());
	}

	// PREFIX PROCESS
	public void caseACspPrefixProcess(ACspPrefixProcess node) {
		inACspPrefixProcess(node);
		this.expectedType = this.typeFactory.getAnyType(); // Line added.
		if (node.getEvent() != null) {
			node.getEvent().apply(this);
		}
		{
			Iterator it = node.getComunication().iterator();
			PComunication comm;
			while (it.hasNext()) {
				comm = (PComunication) it.next();
				comm.apply(this);
			}
		}
		if (node.getArrow() != null) {
			node.getArrow().apply(this);
		}
		this.expectedType = this.typeFactory.getProcessType(); // Line added.
		if (node.getProcess() != null) {
			node.getProcess().apply(this);
		}
		outACspPrefixProcess(node);
	}

	public void outACspPrefixProcess(ACspPrefixProcess node) {
		CspType type = (CspType) this.exprTypes.pop();

		if (type instanceof CspProcessType) {
			this.exprTypes.push(type);
		} else {
			throw new CspTypeException(getErrorLocation(), "Process", type);
		}
	}

	public void caseACspEvent(ACspEvent node) {
		String name = node.getChannelName().getText();

		CspDeclaration decl = this.contextualTable.getDeclaration(name);

		if (decl instanceof CspChannelDeclaration) {
			// node.getChannelProjection()
			// Channel Projection instanceof checked in its own methods

			// node.getComunication()
			// Comunications are checked in their own methods
		} else if (decl != null) {
			throw new CspSemanticException(getErrorLocation() +
					" - Referenced name: '" + name + "' is not a channel.");
		} else {
			throw new CspSemanticException(getErrorLocation() +
					" - Referenced channel: '" + name + "' was not declared.");
		}
	}

	// COMUNICATION
	// TODO
	public void caseACspInputComunication(ACspInputComunication node) {
	}

	public void caseACspConstrainedinputComunication(
			ACspConstrainedInputComunication node) {
	}

	public void caseACspOutputComunication(ACspOutputComunication node) {
	}

	// GUARDED PROCESS
	public void caseACspGuardedProcess(ACspGuardedProcess node) {
		inACspGuardedProcess(node);
		this.expectedType = this.typeFactory.getBoolType(); // Line added.
		if (node.getGuardExpr() != null) {
			node.getGuardExpr().apply(this);
		}
		if (node.getGuard() != null) {
			node.getGuard().apply(this);
		}
		this.expectedType = this.typeFactory.getProcessType(); // Line added.
		if (node.getProcess() != null) {
			node.getProcess().apply(this);
		}
		outACspGuardedProcess(node);
	}

	public void outACspGuardedProcess(ACspGuardedProcess node) {
		CspType nextType = (CspType) this.exprTypes.pop();
		CspType guardType = (CspType) this.exprTypes.pop();

		if (guardType instanceof CspBoolType) {
			if (nextType instanceof CspProcessType) {
				this.exprTypes.push(nextType);
			} else {
				throw new CspTypeException(getErrorLocation(), "Process",
						nextType);
			}
		} else {
			throw new CspTypeException(getErrorLocation(), "Bool", guardType);
		}
	}

	// REPLICATED
	// TODO
	public void outACspInterleaveReplicatedProcess(
			ACspInterleaveReplicatedProcess node) {
	}

	public void outACspSharingReplicatedProcess(
			ACspSharingReplicatedProcess node) {
	}

	public void outACspLinkedReplicatedProcess(ACspLinkedReplicatedProcess node) {
	}

	public void outACspAlphabetisedReplicatedProcess(
			ACspAlphabetisedReplicatedProcess node) {
	}

	public void outACspInternalReplicatedProcess(
			ACspInternalReplicatedProcess node) {
	}

	public void outACspExternalReplicatedProcess(
			ACspExternalReplicatedProcess node) {
	}

	public void outACspSequentialReplicatedProcess(
			ACspSequentialReplicatedProcess node) {
	}

	// VARIABLE LIST
	// TODO
	public void outACspVariable(ACspVariable node) {
	}

	// HIDING PROCESS
	// TODO
	public void outACspHidingProcess(ACspHidingProcess node) {
	}

	// RENAMING PROCESS
	// TODO
	public void outACspRenamingProcess(ACspRenamingProcess node) {
	}

	public void outACspSubstitution(ACspSubstitution node) {
	}

	// PRIMITIVE PROCESSES
	public void outACspSkipProcess(ACspSkipProcess node) {
		this.exprTypes.push(this.typeFactory.getProcessType());
	}

	public void outACspStopProcess(ACspStopProcess node) {
		this.exprTypes.push(this.typeFactory.getProcessType());
	}

	public void outACspChaosProcess(ACspChaosProcess node) {
		this.exprTypes.push(this.typeFactory.getProcessType());
	}

	// EXPRESSIONS ////////////////////////////////////////////////////////

	// LOGICAL EXPRESSIONS
	public void inACspAndLogicalExpr(ACspAndLogicalExpr node) {
		this.expectedType = this.typeFactory.getBoolType();
	}

	public void outACspAndLogicalExpr(ACspAndLogicalExpr node) {
		checkBinary(node.getAnd(), this.typeFactory.getBoolType());
	}

	public void inACspOrLogicalExpr(ACspOrLogicalExpr node) {
		this.expectedType = this.typeFactory.getBoolType();
	}

	public void outACspOrLogicalExpr(ACspOrLogicalExpr node) {
		checkBinary(node.getOr(), this.typeFactory.getBoolType());
	}

	// RELATIONAL EXPRESSIONS
	private void checkRelationalExpr(Token op) {
		CspType rightType = (CspType) this.exprTypes.pop();
		CspType leftType = (CspType) this.exprTypes.peek();
		this.exprTypes.push(rightType);

		this.expectedType = leftType;

		this.checkBinary(op, this.typeFactory.getBoolType());
	}

	public void inACspEqualRelationalExpr(ACspEqualRelationalExpr node) {
		this.expectedType = this.typeFactory.getAnyType();
	}

	public void outACspEqualRelationalExpr(ACspEqualRelationalExpr node) {
		checkRelationalExpr(node.getEqual());
	}

	public void inACspNotEqualRelationalExpr(ACspNotEqualRelationalExpr node) {
		this.expectedType = this.typeFactory.getAnyType();
	}

	public void outACspNotEqualRelationalExpr(ACspNotEqualRelationalExpr node) {
		checkRelationalExpr(node.getNEqual());
	}

	public void inACspGtRelationalExpr(ACspGtRelationalExpr node) {
		this.expectedType = this.typeFactory.getAnyType();
	}

	public void outACspGtRelationalExpr(ACspGtRelationalExpr node) {
		checkRelationalExpr(node.getGt());
	}

	public void inACspGteRelationalExpr(ACspGteRelationalExpr node) {
		this.expectedType = this.typeFactory.getAnyType();
	}

	public void outACspGteRelationalExpr(ACspGteRelationalExpr node) {
		checkRelationalExpr(node.getGeq());
	}

	public void inACspLtRelationalExpr(ACspLtRelationalExpr node) {
		this.expectedType = this.typeFactory.getAnyType();
	}

	public void outACspLtRelationalExpr(ACspLtRelationalExpr node) {
		checkRelationalExpr(node.getLt());
	}

	public void inACspLteRelationalExpr(ACspLteRelationalExpr node) {
		this.expectedType = this.typeFactory.getAnyType();
	}

	public void outACspLteRelationalExpr(ACspLteRelationalExpr node) {
		checkRelationalExpr(node.getLeq());
	}

	// MATH EXPRESSIONS
	public void inACspAddMathExpr(ACspAddMathExpr node) {
		this.expectedType = this.typeFactory.getIntType();
	}

	public void outACspAddMathExpr(ACspAddMathExpr node) {
		checkBinary(node.getPlus(), this.typeFactory.getIntType());
	}

	public void inACspDiffMathExpr(ACspDiffMathExpr node) {
		this.expectedType = this.typeFactory.getIntType();
	}

	public void outACspDiffMathExpr(ACspDiffMathExpr node) {
		checkBinary(node.getMinus(), this.typeFactory.getIntType());
	}

	public void inACspMulMath1Expr(ACspMulMath1Expr node) {
		this.expectedType = this.typeFactory.getIntType();
	}

	public void outACspMulMath1Expr(ACspMulMath1Expr node) {
		checkBinary(node.getMult(), this.typeFactory.getIntType());
	}

	public void inACspDivMath1Expr(ACspDivMath1Expr node) {
		this.expectedType = this.typeFactory.getIntType();
	}

	public void outACspDivMath1Expr(ACspDivMath1Expr node) {
		checkBinary(node.getDiv(), this.typeFactory.getIntType());
	}

	public void inACspRemainderMath1Expr(ACspRemainderMath1Expr node) {
		this.expectedType = this.typeFactory.getIntType();
	}

	public void outACspRemainderMath1Expr(ACspRemainderMath1Expr node) {
		checkBinary(node.getRemainder(), this.typeFactory.getIntType());
	}

	// SEQUENCE CONCATENATION - seq1 ^ seq2
	public void inACspCatSequenceExpr(ACspCatSequenceExpr node) {
		this.expectedType = this.typeFactory.getSequenceType();
	}

	public void outACspCatSequenceExpr(ACspCatSequenceExpr node) {
		CspType rightType = (CspType) this.exprTypes.pop();
		CspType leftType = (CspType) this.exprTypes.pop();

		if (leftType instanceof CspSequenceType) {
			if (rightType.equals(leftType)) {
				this.exprTypes.push(leftType);
			} else {
				throw new CspTypeException(getErrorLocation(), leftType
						.toString(), rightType);
			}
		} else {
			throw new CspTypeException(getErrorLocation(), "Sequence <?>",
					leftType);
		}
	}

	// PREFIX EXPRESSIONS: not e; -e; #e;
	public void inACspNotPrefixExpr(ACspNotPrefixExpr node) {
		this.expectedType = this.typeFactory.getBoolType();
	}

	public void outACspNotPrefixExpr(ACspNotPrefixExpr node) {
		CspType type = (CspType) this.exprTypes.pop();
		if (type instanceof CspBoolType) {
			this.exprTypes.push(type);
		} else {
			throw new CspTypeException(getErrorLocation(), "Bool", type);
		}
	}

	public void inACspMinusPrefixExpr(ACspMinusPrefixExpr node) {
		this.expectedType = this.typeFactory.getIntType();
	}

	public void outACspMinusPrefixExpr(ACspMinusPrefixExpr node) {
		CspType type = (CspType) this.exprTypes.pop();
		if (type instanceof CspIntType) {
			this.exprTypes.push(type);
		} else {
			throw new CspTypeException(getErrorLocation(), "int", type);
		}
	}

	public void inACspLengthPrefixExpr(ACspLengthPrefixExpr node) {
		this.expectedType = this.typeFactory.getSequenceType();
	}

	public void outACspLengthPrefixExpr(ACspLengthPrefixExpr node) {
		CspType type = (CspType) this.exprTypes.pop();
		if (type instanceof CspSequenceType) {
			this.exprTypes.push(this.typeFactory.getIntType());
		} else {
			throw new CspTypeException(getErrorLocation(), "Sequence <?>", type);
		}
	}

	// CALL
	// private CspFunctionDeclaration currentFunction;
	public void outACspCallPostfixExpr(ACspCallPostfixExpr node) {
            
            
		this.lastToken = node.getName();

		String name = node.getName().getText();

		CspDeclaration decl = this.contextualTable.getDeclaration(name);

                
                if (decl != null) {
			if (decl.getCspType() == this.typeFactory.getAnyType()) {
				decl.setCspType(this.expectedType);
			}
		}

		if (decl instanceof CspFunctionDeclaration) {
			if (node.getArguments() == null) {
				this.exprTypes.push(decl.getCspType());
			} else {
				if (decl.getCspType() != this.typeFactory.getAnyType()) {
                                    //MARCEL//////////////
                                    this.exprTypes.pop();
                                    //////////////////////
                                    this.exprTypes.push(decl.getCspType());
				} else {
                                    this.exprTypes.push(this.typeFactory.getProcessType());
				}
			}
		} else if (decl != null) {
			this.exprTypes.push(decl.getCspType());
		} else {
			throw new CspSemanticException(getErrorLocation()
					+ " - Referenced name: '" + name + "' was not declared.");
		}

		/*
		 * if (decl instanceof CspFunctionDeclaration) { CspFunctionDeclaration
		 * f = (CspFunctionDeclaration) decl;
		 * 
		 * if (f == this.currentFunction && f.getCspType() ==
		 * this.typeFactory.getAnyType()) { f.getCspType() = this.expectedType; }
		 * 
		 * ArrayList parameters = f.parameters; int counter = 0; ArrayList
		 * argsTypes = new ArrayList();
		 * 
		 * if (node.getArgs() != null) { PArgumentList args = (PArgumentList)
		 * node.getArgumentList(); if (args.getExpressao() != null) {
		 * CspVariableDeclaration v = (CspVariableDeclaration)
		 * parameters[counter]; CspType paramType = v.getCspType(); counter++;
		 * 
		 * this.expectedType = paramType;
		 * 
		 * args.getExpressao().apply(this);
		 * 
		 * CspType argType = (CspType) this.exprTypes.pop();
		 * argsTypes.Add(argType);
		 * 
		 * if (argType != paramType && paramType !=
		 * this.typeFactory.getAnyType()) { throw new
		 * CspSemanticException("Expecting '" + paramType + "' found '" +
		 * argType); }
		 * 
		 * this.expectedType = this.typeFactory.getAnyType(); }
		 * 
		 * Object[] temp = args.getProximos(); for (int i = 0; i < temp.Length;
		 * i++) { AProximoArgumento proximo = (AProximoArgumento) temp[i];
		 * 
		 * if (counter < parameters.Count) { CspVariableDeclaration v =
		 * (CspVariableDeclaration) parameters[counter]; CspType paramType =
		 * v.getCspType(); counter++;
		 * 
		 * this.expectedType = paramType;
		 * 
		 * proximo.getExpressao().apply(this);
		 * 
		 * CspType argType = (CspType) this.exprTypes.pop();
		 * argsTypes.Add(argType);
		 * 
		 * if (argType != paramType && paramType !=
		 * this.typeFactory.getAnyType()) { throw new
		 * CspSemanticException("Expecting '" + paramType + "', found '" +
		 * argType); }
		 * 
		 * this.expectedType = this.typeFactory.getAnyType(); } else { throw new
		 * CspSemanticException( "Too many arguments: " + counter + "."); } }
		 * 
		 * if (counter < parameters.Count) { throw new CspSemanticException(
		 * "Too few arguments: " + counter + "."); }
		 *  // If not pre-defined nor recursive. if (!predefined(f) && f !=
		 * this.currentFunction) { // Avoid loop if applying a recursive
		 * function if (this.currentFunction == null) { this.currentFunction =
		 * f; }
		 *  // Visits the function body, to discover the return type. // The
		 * parameter types are the args types. CspFunctionDeclaration copy =
		 * (CspFunctionDeclaration) f.node.clone();
		 *  // Open a frame this.contextualTable.OpenFrame(node); IEnumerator
		 * paramsIt = parameters.getEnumerator(); IEnumerator argsIt =
		 * argsTypes.getEnumerator(); while (paramsIt.MoveNext() &&
		 * argsIt.MoveNext()) { // are = CspVariableDeclaration param =
		 * (CspVariableDeclaration) paramsIt.Current; // Create the real
		 * parameter param = new CspVariableDeclaration(param.Name, null);
		 * param.getCspType() = (CspType) argsIt.Current;
		 * this.contextualTable.Bind(param); }
		 * 
		 * copy.getExpressao().apply(this);
		 * 
		 * f.getCspType() = (CspType) this.exprTypes.pop(); // Closes the frame
		 * this.contextualTable.CloseFrame(node);
		 * 
		 * this.currentFunction = null; }
		 * 
		 * this.exprTypes.push(f.getCspType()); } } else { throw new
		 * CspSemanticException("Identifier '" + name + "' not declared."); }
		 */
	}

	// ARGUMENTS
	public void outACspLinkedArgumentList(ACspLinkedArgumentList node) {
		CspType rightType = (CspType) this.exprTypes.pop();
		CspType leftType = (CspType) this.exprTypes.pop();

		this.exprTypes.push(this.typeFactory.ParametrizedType(leftType,
				rightType));
	}

	public void outACspBaseArgumentList(ACspBaseArgumentList node) {
		CspType type = (CspType) this.exprTypes.pop();

		this.exprTypes.push(this.typeFactory.ParametrizedType(type));
	}

	// RANGES
	public void inACspOpenRange(ACspOpenRange node) {
		this.expectedType = this.typeFactory.getIntRangeType();
	}

	public void outACspOpenRange(ACspOpenRange node) {
		this.checkUnary(node.getDdot(), this.typeFactory.getIntRangeType());
	}

	public void inACspClosedRange(ACspClosedRange node) {
		this.expectedType = this.typeFactory.getIntRangeType();
	}

	public void outACspClosedRange(ACspClosedRange node) {
		this.checkBinary(node.getDdot(), this.typeFactory.getIntRangeType());
	}

	// COMPOSITION
	public void inACspDotPostfixExpr(ACspDotPostfixExpr node) {
		this.expectedType = this.typeFactory.getAnyType();
	}

	public void outACspDotPostfixExpr(ACspDotPostfixExpr node) {
		CspType rightType = (CspType) this.exprTypes.pop();
		CspType leftType = (CspType) this.exprTypes.pop();

		this.exprTypes
				.push(this.typeFactory.CompositeType(leftType, rightType));
	}

	// PRIMITIVE VALUES
	public void outACspBooleanValue(ACspBooleanValue node) {
		this.exprTypes.push(this.typeFactory.getBoolType());
	}

	public void outACspNumberValue(ACspNumberValue node) {
		this.exprTypes.push(this.typeFactory.getIntType());
	}

	private CspType getUniqueTypeOf(PArgumentList node) {
		CspParametrizedType type = (CspParametrizedType) this.exprTypes.pop();

		Iterator it = type.getTypes().iterator();

		if (it.hasNext()) {
			CspType elemType = (CspType) type.getTypes().get(0);
			CspType t;
			while (it.hasNext()) {
				t = (CspType) it.next();
				if (elemType instanceof CspIntRangeType
						|| elemType instanceof CspIntType) {
					if (t instanceof CspIntRangeType || t instanceof CspIntType) {
						elemType = this.typeFactory.getIntType();
					} else {
						throw new CspTypeException(getErrorLocation(), elemType
								.toString(), t);
					}
				} else {
					if (t != elemType) {
						throw new CspTypeException(getErrorLocation(), elemType
								.toString(), t);
					}
				}
			}
			return elemType;
		} else {
			return this.typeFactory.getAnyType();
		}
	}

	// SETS
	public void outACspEmptySet(ACspEmptySet node) {
		this.exprTypes.push(this.typeFactory.getSetType());
	}

	public void outACspSet(ACspSet node) {
		CspType elemType = this.getUniqueTypeOf(node.getArgumentList());
		this.exprTypes.push(this.typeFactory.SetOf(elemType));
	}

	public void caseACspComprehensionSet(ACspComprehensionSet node) {
		inACspComprehensionSet(node);
		if (node.getRBrace() != null) {
			node.getRBrace().apply(this);
		}
		if (node.getConstraintList() != null) {
			node.getConstraintList().apply(this);
		}
		if (node.getBar() != null) {
			node.getBar().apply(this);
		}
		if (node.getArgumentList() != null) {
			node.getArgumentList().apply(this);
		}
		if (node.getLBrace() != null) {
			node.getLBrace().apply(this);
		}
		outACspComprehensionSet(node);
	}

	public void outACspComprehensionSet(ACspComprehensionSet node) {
		CspType elemType = this.getUniqueTypeOf(node.getArgumentList());
		this.exprTypes.push(this.typeFactory.SetOf(elemType));
	}

	public void outACspProductionsSet(ACspProductionsSet node) {
		// CspType elemType = (CspType)
		// this.exprTypes.pop()(node.getChannelList());
		// this.exprTypes.push((node, this.typeFactory.SetOf(elemType));
		this.exprTypes.push(this.typeFactory.getSetType());
	}

	// SEQUENCES
	public void outACspEmptySequence(ACspEmptySequence node) {
		this.exprTypes.push(this.typeFactory.getSequenceType());
	}

	public void outACspSequence(ACspSequence node) {
		CspType elemType = this.getUniqueTypeOf(node.getArgumentList());
		this.exprTypes.push(this.typeFactory.SequenceOf(elemType));
	}

	public void caseACspComprehensionSequence(ACspComprehensionSequence node) {
		inACspComprehensionSequence(node);
		if (node.getGt() != null) {
			node.getGt().apply(this);
		}
		if (node.getConstraintList() != null) {
			node.getConstraintList().apply(this);
		}
		if (node.getBar() != null) {
			node.getBar().apply(this);
		}
		if (node.getArgumentList() != null) {
			node.getArgumentList().apply(this);
		}
		if (node.getLt() != null) {
			node.getLt().apply(this);
		}
		outACspComprehensionSequence(node);
	}

	public void outACspComprehensionSequence(ACspComprehensionSequence node) {
		CspType elemType = this.getUniqueTypeOf(node.getArgumentList());
		this.exprTypes.push(this.typeFactory.SequenceOf(elemType));
	}

	// SOURCE
	public void outACspConstraint(ACspConstraint node) {
		CspType type = (CspType) this.exprTypes.pop();
		if (!(type instanceof CspBoolType)) {
			throw new CspTypeException(getErrorLocation(), "Bool", type);
		}
	}

	public void outACspSourceConstraint(ACspSourceConstraint node) {
		CspSetType setType = (CspSetType) this.exprTypes.pop();

		this.Match(node.getExpression(), setType.getElemType());
	}

	// TUPLE
	public void outACspTupleValue(ACspTupleValue node) {
		CspType rightType = (CspType) this.exprTypes.pop();
		CspType leftType = (CspType) this.exprTypes.pop();

		this.exprTypes.push(this.typeFactory.TupleType(leftType, rightType));
	}

	// / <summary>
	// / Pattern matching
	// / </summary>
	// / <param name="node"></param>
	// / <param name="type"></param>
	private void Match(Node node, CspType type) {
	}

}