package lmf.formula.csp.type;

import java.util.*;


/**
 * 
 * @author Joabe Jesus
 */
public class CspTypeFactory {

	private CspType anyType = new CspType();

	public CspType getAnyType() {
		return anyType;
	}

	// INT TYPE
	private CspIntType intType = new CspIntType();

	public CspIntType getIntType() {
		return intType;
	}

	// INT RANGE TYPE
	private CspIntRangeType intRangeType = new CspIntRangeType();

	public CspIntRangeType getIntRangeType() {
		return intRangeType;
	}

	// BOOL TYPE
	private CspBoolType boolType = new CspBoolType();

	public CspBoolType getBoolType() {
		return boolType;
	}

	// COMPOSITE TYPE
	private Hashtable composites = new Hashtable();

	public CspCompositeType CompositeType(CspType leftType, CspType rightType) {
		ArrayList types = new ArrayList();
		types.add(leftType);

		if (rightType instanceof CspCompositeType) {
			types.addAll(((CspCompositeType) rightType).getTypes());
		} else {
			types.add(rightType);
		}

		String key = leftType.getName();
		for (int i = 1; i < types.size(); i++) {
			key += "." + ((CspType) types.get(i)).getName();
		}

		CspCompositeType ret = (CspCompositeType) this.composites.get(key);
		if (ret == null) {
			ret = new CspCompositeType(key, types);
			this.composites.put(key, ret);
		}
		return ret;
	}

	// TUPLE TYPE
	private Hashtable tuples = new Hashtable();

	public CspTupleType TupleType(CspType leftType, CspType rightType) {
		CspParametrizedType argsType = ParametrizedType(leftType, rightType);
		String key = "(" + argsType.getName() + ")";

		CspTupleType ret = (CspTupleType) this.tuples.get(key);
		if (ret == null) {
			ret = new CspTupleType(key, argsType.getTypes());
			this.tuples.put(key, ret);
		}
		return ret;
	}

	public CspParametrizedType ParametrizedType(CspType type) {
		ArrayList types = new ArrayList();
		types.add(type);

		return new CspParametrizedType(type.getName(), types);
	}

	public CspParametrizedType ParametrizedType(CspType leftType,
			CspType rightType) {
		ArrayList types = new ArrayList();
		types.add(leftType);

		if (rightType instanceof CspParametrizedType) {
			types.addAll(((CspParametrizedType) rightType).getTypes());
		} else {
			types.add(rightType);
		}

		String key = leftType.getName();
		for (int i = 1; i < types.size(); i++) {
			key += ", " + ((CspType) types.get(i)).getName();
		}

		return new CspParametrizedType(key, types);
	}

	// SET TYPE
	private CspSetType setType = new CspSetType(new CspType());

	public CspSetType getSetType() {
		return setType;
	}

	private Hashtable sets = new Hashtable();

	public CspSetType SetOf(CspType elemType) {
		CspSetType ret = (CspSetType) this.sets.get(elemType);
		if (ret == null) {
			ret = new CspSetType(elemType);
			this.sets.put(elemType, ret);
		}
		return ret;
	}

	// SEQUENCE TYPE
	private CspSequenceType sequenceType = new CspSequenceType(new CspType());

	public CspSequenceType getSequenceType() {
		return sequenceType;
	}

	private Hashtable sequences = new Hashtable();

	public CspSequenceType SequenceOf(CspType elemType) {
		CspSequenceType ret = (CspSequenceType) this.sequences.get(elemType);
		if (ret == null) {
			ret = new CspSequenceType(elemType);
			this.sequences.put(elemType, ret);
		}
		return ret;
	}

	// DATATYPE
	private Hashtable datatypes = new Hashtable();

	public CspDataType DataType(String name) {
		CspDataType ret = (CspDataType) this.datatypes.get(name);
		if (ret == null) {
			ret = new CspDataType(name);
			this.datatypes.put(name, ret);
		}
		return ret;
	}

	// SUBTYPE
	private Hashtable subtypes = new Hashtable();

	public CspSubType SubType(String name) {
		CspSubType ret = (CspSubType) this.subtypes.get(name);
		if (ret == null) {
			ret = new CspSubType(name);
			this.subtypes.put(name, ret);
		}
		return ret;
	}

	// NAMETYPE
	private Hashtable nametypes = new Hashtable();

	public CspNameType NameType(String name) {
		CspNameType ret = (CspNameType) this.nametypes.get(name);
		if (ret == null) {
			ret = new CspNameType(name);
			this.nametypes.put(name, ret);
		}
		return ret;
	}

	// FUNCTION TYPE
	private Hashtable parametrizedTypes = new Hashtable();

	public CspParametrizedType ParametrizedType(String name,
			ArrayList paramTypes) {
		CspParametrizedType ret = (CspParametrizedType) this.parametrizedTypes
				.get(paramTypes);
		if (ret == null) {
			ret = new CspParametrizedType(name, paramTypes);
			this.parametrizedTypes.put(paramTypes, ret);
		}
		return ret;
	}

	// UNTYPED CHANNEL TYPE
	private CspUntypedChannelType untypedChannelType = new CspUntypedChannelType();

	public CspUntypedChannelType getUntypedChannelType() {
		return untypedChannelType;
	}

	// PROCESS TYPE
	private CspProcessType processType = new CspProcessType();

	public CspProcessType getProcessType() {
		return processType;
	}

}