package csp2hc;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PushbackReader;
import java.util.Properties;

import lmf.formula.csp.semantic.CspAnalyserListener;

import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;

import lmf.formula.csp.lexer.Lexer;
import lmf.formula.csp.lexer.LexerException;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.Start;
import lmf.formula.csp.parser.Parser;
import lmf.formula.csp.parser.ParserException;
import lmf.formula.csp.semantic.CspContextualAnalyser;
import lmf.formula.csp.semantic.exception.CspAnalyserException;
import csp2hc.core.ChannelsVisitor;
import csp2hc.core.ConstantsAndFunctionsVisitor;
import csp2hc.core.DirectivesVisitor;
import csp2hc.core.MainMethodVisitor;
import csp2hc.core.ProcessesInfoVisitor;
import csp2hc.core.ProcessesTranslatorVisitor;
import csp2hc.core.RestrictionsVisitor;
import csp2hc.core.TailRecursionVisitor;
import csp2hc.core.TypesVisitor;
import csp2hc.exception.InconsistentChannelProjectionUseException;
import csp2hc.exception.InvalidDefinitionException;
import csp2hc.exception.InvalidDirectiveException;
import csp2hc.exception.RepeatedConstantException;
import csp2hc.exception.UndefinedArgumentTypeException;
import csp2hc.exception.UndefinedChannelInOutException;
import csp2hc.exception.UndefinedChannelTypeException;
import csp2hc.exception.UndefinedIntBitsException;
import csp2hc.exception.UnsupportedConstructorException;
import csp2hc.exception.WrongNumberOfArgumentsException;
import csp2hc.gui.MyFileFilter;
import csp2hc.messages.HandelC;
import csp2hc.messages.Messenger;
import csp2hc.util.CodeFormatting;
import csp2hc.util.Config;
import csp2hc.util.CspAstFactory;
import csp2hc.util.CspCstFactory;
import csp2hc.util.TreeView;
import csp2hc.util.LineCounter;

public class Csp2HC implements CspAnalyserListener {

    private static final String fileSep = System.getProperty("file.separator");
   
    private Config config;
    private String csp_files_dir;
    private Start start;
    private CspAnalyserException error;
    
    // Stores the number of lines of the original file 
    static int currentLineCounter = 0;
    
    public Csp2HC() {
        this.loadConfigurations();
    }
    
	/** AnalysisListener methods */

	public void started() {
	}

	public void analysing(Node node) {
	}

	public void analysed(Node node) {
	}

	public void catchException(CspAnalyserException e) {
		this.error = e;
	}

	public void warning(int line, int column, String message) {
	}

	public void finished() {
	}
	/* End of methods */

    private void loadConfigurations(){
        Messenger.getInstance().start(Messenger.LOADING_CONFIG);
        try {
        	this.config = Config.getInstance();
        } catch (Exception e) {
        	e.printStackTrace();
        }
        this.csp_files_dir = config.getConfig("CSP_FILES_DIR");
        Messenger.getInstance().finish(Messenger.LOADING_CONFIG);
        
    }
    
    public String openFile() throws FileNotFoundException, IOException{
        Messenger.getInstance().start(Messenger.OPENING_FILE);
        String file = "";
        String fileName = "";

        int val = 0;
        JFileChooser chooser = new JFileChooser();
        MyFileFilter filter = new MyFileFilter(HandelC.CSP_EXTENSION);
        chooser.setFileFilter(filter);
        chooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
        chooser.setDialogType(JFileChooser.OPEN_DIALOG);
        chooser.setCurrentDirectory(new File(this.csp_files_dir));

        val = chooser.showDialog(null,Messenger.OPEN);
        if(val == JFileChooser.APPROVE_OPTION){
            fileName = chooser.getSelectedFile().toString();
            Messenger.getInstance().finish(Messenger.OPENING_FILE);
            this.rememberDirectory(fileName);
        } else {
            Messenger.getInstance().cancel(Messenger.OPENING_FILE);
        }
        return fileName;
    }

    public void saveFile(String text) throws FileNotFoundException, IOException{
        Messenger.getInstance().start(Messenger.SAVING_OUTPUT);
        String fileName = null;

        boolean choosed = false;
        while (!choosed) {
            fileName = null;
            int val = 0;
            JFileChooser chooser = new JFileChooser();

            MyFileFilter filter = new MyFileFilter(HandelC.EXTENSION);
            chooser.setFileFilter(filter);
            chooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
            chooser.setDialogType(JFileChooser.SAVE_DIALOG);
            chooser.setCurrentDirectory(new File(csp_files_dir));

            val = chooser.showDialog(null,Messenger.SAVE);
        
            if(val == JFileChooser.APPROVE_OPTION){
                File file = chooser.getSelectedFile(); 
                fileName = file.toString();
                
                if(!fileName.endsWith(HandelC.DOT+HandelC.EXTENSION)) {
                    fileName = fileName + HandelC.DOT + HandelC.EXTENSION;
                    file = new File(fileName);
                }
                
                if(file.exists()) {
                    int answer = JOptionPane.showConfirmDialog (chooser,
                        Messenger.OVERWRITE_1 + fileName + Messenger.OVERWRITE_2,
                        Messenger.OVERWRITE_TITLE,
                        JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
                    if (answer == JOptionPane.OK_OPTION){
                        choosed = true;
                    }
                } else {
                    choosed = true;
                }
            } else if (val == JFileChooser.CANCEL_OPTION){
                choosed = true;
            }
        }
        if (fileName != null) {
            FileWriter writer = new FileWriter(fileName,false);
            writer.write(text);
            writer.close();
            Messenger.getInstance().finish(Messenger.SAVING_OUTPUT);
        } else {
            Messenger.getInstance().cancel(Messenger.SAVING_OUTPUT);
        }
    }
    
    private void rememberDirectory(String fileName) throws FileNotFoundException, IOException{
        Messenger.getInstance().start(Messenger.STORING_DIRECTORY);
        Properties properties = new Properties();
        
        File f = new File("./config/config.properties");
        try{
            FileInputStream propFile = new FileInputStream(f);
            properties.load(propFile);
        } catch (FileNotFoundException e) {
            Messenger.getInstance().error(Messenger.NO_CONFIGURATION_FILE);
        }

        try{
            int    pos       = fileName.lastIndexOf(fileSep);
            String directory = ( pos<0 ? "./" : fileName.substring(0,pos) );

            if ( !fileSep.equals("/") ) {
                directory.replace(fileSep,"/");
            }
            FileOutputStream propFileOut = new FileOutputStream(f);
            properties.setProperty("CSP_FILES_DIR",directory);
            properties.store(propFileOut,"Configuration File");
            this.csp_files_dir = fileName;
        } catch (FileNotFoundException e) {
            Messenger.getInstance().error(Messenger.NO_CONFIGURATION_FILE);
        }
        Messenger.getInstance().finish(Messenger.STORING_DIRECTORY);
    }
    
    private Start parser(String fileName) throws FileNotFoundException, 
                                                IOException, 
                                                ParserException,
                                                LexerException{

        FileInputStream in = new FileInputStream(fileName);
        InputStreamReader isr = new InputStreamReader(in);
        PushbackReader pb = new PushbackReader(isr, 1024);
        Lexer lexer = new Lexer(pb);
        Parser p = new Parser(lexer);
        Start tree = p.parse();
        return tree;
    }
    
    public String translate(String fileName) throws FileNotFoundException, 
                                                    IOException,
                                                    ParserException,
                                                    LexerException{
                                                        
        /////////////////////////////////
        // PARSING AND CHECKING MAIN FILE
        /////////////////////////////////

    	Messenger.getInstance().start(Messenger.PARSING);
        Node start = this.parser(fileName);
        Messenger.getInstance().finish(Messenger.PARSING);

        Messenger.getInstance().start(Messenger.TYPE_CHECKING);
        
		CspContextualAnalyser analyser = new CspContextualAnalyser(this);
		analyser.setDebugging(true);
		analyser.setStopOnError(true);
		start.apply(analyser);

        Messenger.getInstance().finish(Messenger.TYPE_CHECKING);
        
        //////////////////////
        // LOAD DIRECTIVE INFO
        ////////////////////// 
        Messenger.getInstance().start(Messenger.LOADING_DIRECTIVES);
        DirectivesVisitor directives = new DirectivesVisitor();
        start.apply(directives);
        Messenger.getInstance().finish(Messenger.LOADING_DIRECTIVES);

        ///////////////////////////////
        // LOAD CONSTANTS AND FUNCTIONS
        ///////////////////////////////
        Messenger.getInstance().start(Messenger.LOADING_CONSTANTS_AND_FUNCTIONS);
        ConstantsAndFunctionsVisitor constants = new ConstantsAndFunctionsVisitor(directives);
        start.apply(constants);
        Messenger.getInstance().finish(Messenger.LOADING_CONSTANTS_AND_FUNCTIONS);

        /////////////////
        // LOAD DATATYPES
        /////////////////
        Messenger.getInstance().start(Messenger.LOADING_DATATYPES);
        TypesVisitor datatypes = new TypesVisitor(directives,constants);
        start.apply(datatypes);
        Messenger.getInstance().finish(Messenger.LOADING_DATATYPES);

        ///////////////////////////////////////////
        // LOAD PROCESSES ARGUMENTS AND DEFINITIONS
        ///////////////////////////////////////////
        Messenger.getInstance().start(Messenger.LOADING_PROCESSES_INFO);
        ProcessesInfoVisitor procInfo = new ProcessesInfoVisitor();
        start.apply(procInfo);
        Messenger.getInstance().finish(Messenger.LOADING_PROCESSES_INFO);
        
        /////////////////////////////////
        // PARSING AND CHECKING TEMP FILE
        /////////////////////////////////
        Messenger.getInstance().start(Messenger.GENERATING_MAIN);

        String tempFileName = fileName+HandelC.TEMPORARY_FILE;

        // Read Original File
        BufferedReader in = new BufferedReader(new FileReader(fileName));
        String original = "";
        String str;
        while ((str = in.readLine()) != null) {
            original = original + str + "\n";
        }
        in.close();

        // Write Temporary File (original file + main process)
        Messenger.getInstance().start(Messenger.CREATING_TEMP_FILE);
        String mainProc = HandelC.MAIN + HandelC.ASSIGNMENT + directives.getMain();
        String contents = original + mainProc;
        File file = new File(tempFileName);
        file.deleteOnExit();
        FileWriter writer = new FileWriter(file);
        writer.write(contents);
        writer.close();
        Messenger.getInstance().finish(Messenger.CREATING_TEMP_FILE);

        // Parser Temporary File
        Messenger.getInstance().start(Messenger.PARSING_TEMP_FILE);
        Node mainStart = this.parser(tempFileName);
        Messenger.getInstance().finish(Messenger.PARSING_TEMP_FILE);

        // Type Check Temporary File
        Messenger.getInstance().start(Messenger.TYPE_CHECKING_TEMP_FILE);

    	this.error = null;
        
        CspContextualAnalyser analyserMain = new CspContextualAnalyser(this);
        analyserMain.setDebugging(true);
        analyserMain.setStopOnError(true);
		mainStart.apply(analyserMain);
		
		if (this.error != null) {
			throw this.error;
		}

        Messenger.getInstance().finish(Messenger.TYPE_CHECKING_TEMP_FILE);

        ////////////////////////////
        // LOAD CHANNELS INFORMATION
        ////////////////////////////
        Messenger.getInstance().start(Messenger.LOADING_CHANNELS_INFO);
        ChannelsVisitor channels = new ChannelsVisitor(datatypes,directives,procInfo,constants);
        mainStart.apply(channels);
        Messenger.getInstance().finish(Messenger.LOADING_CHANNELS_INFO);
        
        /////////////////////
        // CHECK RESTRICTIONS
        ///////////////////// 
        Messenger.getInstance().start(Messenger.CHECKING_PRIMARY_RESTRICTIONS);
        RestrictionsVisitor restrictions = new RestrictionsVisitor(directives, datatypes, constants, channels, procInfo);
        mainStart.apply(restrictions);
        Messenger.getInstance().finish(Messenger.CHECKING_PRIMARY_RESTRICTIONS);

        ///////////////////////
        // CHECK TAIL RECURSION
        ///////////////////////
        Messenger.getInstance().start(Messenger.CHECKING_SECONDARY_RESTRICTIONS);
        TailRecursionVisitor recursion = new TailRecursionVisitor(procInfo,directives);
        mainStart.apply(recursion);
        Messenger.getInstance().finish(Messenger.CHECKING_SECONDARY_RESTRICTIONS);


        ///////////////////////////
        // TRANSLATE MAIN PROCESSES
        ///////////////////////////
        ProcessesTranslatorVisitor mainTranslator = 
            new ProcessesTranslatorVisitor(procInfo,directives,datatypes,
                                            recursion,channels,constants);
        
        MainMethodVisitor mainVisitor = new MainMethodVisitor(mainTranslator,recursion.getMutualRecursion(),channels.needsForcedInterleaving());
        mainStart.apply(mainVisitor);
        
        Messenger.getInstance().finish(Messenger.GENERATING_MAIN);

        //////////////////////
        // TRANSLATE PROCESSES
        ////////////////////// 
        String code = "";
        
        // Processes
        Messenger.getInstance().start(Messenger.TRANSLATING_PROCESSES);
        ProcessesTranslatorVisitor translator = 
            new ProcessesTranslatorVisitor(procInfo,directives,datatypes,
                                            recursion,channels,constants);
        
        // As we do not have parallel in the specification, we are allowed to consider
        // that the value of the parallelism counter will not be altered in
        // the translation of the specification
        translator.setParallelismCounter(mainVisitor.getParallelismCounter());
        start.apply(translator);
        Messenger.getInstance().finish(Messenger.TRANSLATING_PROCESSES);

        //////////////////
        // FORMAT CODE
        ////////////////// 
        String finalCode = translator.getHcCode() + mainVisitor.getHcCode();

        Messenger.getInstance().start(Messenger.FORMATTING_OUTPUT);
        code = CodeFormatting.format(finalCode);
        Messenger.getInstance().finish(Messenger.FORMATTING_OUTPUT);

        return code;
    }
    
    public void displayTree(String fileName) throws FileNotFoundException, 
                                                    IOException,
                                                    ParserException,
                                                    LexerException{
            JFrame frame = new JFrame("CSP Syntax Trees");
            TreeView treeViewCst = new TreeView();
            JScrollPane scrollCst = new JScrollPane(treeViewCst);

            TreeView treeViewAst = new TreeView();
            JScrollPane scrollAst = new JScrollPane(treeViewAst);

            treeViewCst.setModel(new DefaultTreeModel(
                    new DefaultMutableTreeNode("Constructing the tree.")));
            treeViewAst.setModel(new DefaultTreeModel(
                    new DefaultMutableTreeNode("Constructing the tree.")));

            JTabbedPane tabs = new JTabbedPane();
            tabs.addTab("CST - Concrete Syntax Tree", scrollCst);
            tabs.addTab("AST - Abstract Syntax Tree", scrollAst);

            frame.getContentPane().add(tabs);

            frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
            frame.setSize(800, 400);
            frame.setVisible(true);

            try {
                Start tree = this.parser(fileName);

				//tree.apply(new CspSemanticAnalyser(null));
				CspContextualAnalyser analyser = new CspContextualAnalyser(this);
				analyser.setDebugging(true);
				tree.apply(analyser);

                tree.apply(new CspCstFactory(treeViewCst));
                tree.apply(new CspAstFactory(treeViewAst));


            } catch (Exception e) {
                treeViewCst.setRoot(new DefaultMutableTreeNode(e.getMessage()));
                treeViewAst.setRoot(new DefaultMutableTreeNode(e.getMessage()));
                e.printStackTrace();
            }
    }
    
    private boolean askForNewTranslation(){
        int answer = JOptionPane.showConfirmDialog (null,
            Messenger.NEW_TRANSLATION,
            Messenger.NEW_TRANSLATION_TITLE,
            JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
        return (answer == JOptionPane.OK_OPTION);
    }

    private boolean askTryAgain(Throwable t){
        t.printStackTrace();
        String message = t.getMessage();
        int lineOfError = 0;
        int index =0;

    	/////////////////////////////
        // Changes the indication of location of error if it happened in the main process
        if (t instanceof CspAnalyserException){
    		lineOfError = ((CspAnalyserException)t).getLocation().getLine();
			index = message.indexOf("-")+1;
    	} else if (t instanceof ParserException){
    		lineOfError = ((ParserException)t).getToken().getLine();
    		index = message.indexOf("]")+1;
    	}
		if (this.currentLineCounter < lineOfError) {
			if (index > 0) {
				message = message.substring(index);
			}
			message = Messenger.IN_MAIN + message;
		}
    	/////////////////////////////
    	/////////////////////////////
		
        int answer = JOptionPane.showConfirmDialog (null,
            Messenger.TRY_AGAIN+message+Messenger.TRY_AGAIN_2,
            Messenger.TRY_AGAIN_TITLE,
            JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
        
        return (answer == JOptionPane.OK_OPTION);
    }

    public boolean showKnownException(Throwable e) {
        Messenger.getInstance().error(e.getMessage());
        return askTryAgain(e);
    }

    public boolean showUnknownException(Throwable e) {
        Messenger.getInstance().uerror(e.getMessage());
        return askTryAgain(e);
    }

    public static void main (String[] argv) {
        
    	boolean keepGoing = true;
        Csp2HC translator = new Csp2HC();
        
        
        while(keepGoing) {
            try {
                String fileName = translator.openFile();
                if (!fileName.equals("")){
                    String message = Messenger.TRANSLATING_FILE + fileName;
                    Messenger.getInstance().addComment(message);
                    
                    Csp2HC.currentLineCounter = LineCounter.countLines(new File(fileName));
                    
                    String hcCode = translator.translate(fileName);
                    translator.saveFile(hcCode);
                    keepGoing = translator.askForNewTranslation();
                } else {
                    keepGoing = false;
                }
            } catch (FileNotFoundException e) {
                keepGoing = translator.showKnownException(e);
            } catch (IOException e) {
                keepGoing = translator.showKnownException(e);
            } catch (ParserException e) {
                keepGoing = translator.showKnownException(e);
            } catch (LexerException e) {
                keepGoing = translator.showKnownException(e);
            } catch (CspAnalyserException e) {
                keepGoing = translator.showKnownException(e);
            } catch (UnsupportedConstructorException e) {
                keepGoing = translator.showKnownException(e);
            } catch (UndefinedChannelInOutException e) {
                keepGoing = translator.showKnownException(e);
            } catch (UndefinedArgumentTypeException e) {
                keepGoing = translator.showKnownException(e);
            } catch (UndefinedChannelTypeException e) {
                keepGoing = translator.showKnownException(e);
            } catch (UndefinedIntBitsException e) {
                keepGoing = translator.showKnownException(e);
            } catch (InvalidDefinitionException e) {
                keepGoing = translator.showKnownException(e);
            } catch (InvalidDirectiveException e) {
                keepGoing = translator.showKnownException(e);
            } catch (InconsistentChannelProjectionUseException e) {
                keepGoing = translator.showKnownException(e);
            } catch (RepeatedConstantException e) {
                keepGoing = translator.showKnownException(e);
            } catch (WrongNumberOfArgumentsException e) {
                keepGoing = translator.showKnownException(e);
            } catch (RuntimeException e) {
                keepGoing = translator.showUnknownException(e);
            }
        }
        Messenger.getInstance().message(Messenger.PROGRAM_TERMINATED);
        System.exit(0);
    }


}

