package csp2hc.core;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.util.InsertionOrderHashMap;
import csp2hc.exception.UnsupportedConstructorException;

public class InitialsExtractorVisitor extends ExtendedDepthFirstAdapter {
    private static int COMMUNICATION_INDEX = 0;
    private static int PROJECTION_INDEX = 1;
    private static int PROCESS_INDEX = 2;
    
    private InsertionOrderHashMap initials;
    
    public InitialsExtractorVisitor() {
        this.initials = new InsertionOrderHashMap();
    }

    public List keyList() { return this.initials.keyList(); }
    public LinkedList getCommunication(PEvent event) {
        return (LinkedList)((List)this.initials.get(event)).get(COMMUNICATION_INDEX);
    }
    public LinkedList getProjection(PEvent event) {
        return (LinkedList)((List)this.initials.get(event)).get(PROJECTION_INDEX);
    }
    public PHelp getProcess(PEvent event) {
    	PHelp p = (PHelp)((List)this.initials.get(event)).get(PROCESS_INDEX);
    	return p;
    }
    
    // returns the keys on th channels belonging to the given list of names
    public List getKeyNamesInList(List names) {
    	List keysInList = new Vector();
    	
        List keys = this.keyList();        
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            String name = ((ACspEvent)event).getChannelName().toString().trim();
            if (names.contains(name)) {
            	keysInList.add(name);
            }
        }
    	return keysInList;
    }

    // returns the keys on th channels not belongging to the given list of names
    public List getKeyNamesNotInList(List names) {
    	List keysNotInList = new Vector();
    	
        List keys = this.keyList();        
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            String name = ((ACspEvent)event).getChannelName().toString().trim();
            if (!names.contains(name)) {
            	keysNotInList.add(name);
            }
        }
    	return keysNotInList;
    }

    // returns the keys on th channels not belongging to the given list of names
    public List getKeyNames() {
    	return this.getKeyNamesNotInList(new Vector());
    }
    
    public void caseACspPrefixProcess(ACspPrefixProcess node){
        PEvent event = node.getEvent();
        LinkedList comm = ((ACspEvent)node.getEvent()).getComunication();
        LinkedList proj = ((ACspEvent)node.getEvent()).getProjection();
        PHelp proc = node.getProcess();
        
        Vector data = new Vector();
        data.add(COMMUNICATION_INDEX,comm);
        data.add(PROJECTION_INDEX,proj);
        data.add(PROCESS_INDEX,proc);
        this.initials.put(event,data);
    }
}
