package csp2hc.core;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class MainMethodVisitor extends ExtendedDepthFirstAdapter {
    // Handel-C Code of the main method
    private String hcCode;
    
    // Visitor that has translation rules
    private ProcessesTranslatorVisitor translator;

    // Indicates if the system has been translated as a mutual recursion
    private boolean mutualRecursion;

    // Indicates if we need to force interleaving
    private boolean useSemaphors;
    
    public MainMethodVisitor(ProcessesTranslatorVisitor translator, 
    						boolean mutualRec, 
    						boolean useSemaphors) {

        this.hcCode = "";
        this.translator = translator;
        this.mutualRecursion = mutualRec;
        this.useSemaphors = useSemaphors;
    }
    
    public String getHcCode(){ return this.hcCode; }
    public int getParallelismCounter(){ return this.translator.getParallelismCounter(); }
    
    public void caseStart(Start node){
        this.hcCode = HandelC.generateComment(Messenger.MAIN_DECLARATION);
        this.hcCode = this.hcCode + HandelC.MAIN_DECL + HandelC.BLOCK_BEGIN;

        // If we have forced interleaving
        if (this.useSemaphors) {
        	this.hcCode = this.hcCode + translator.getForcedSyncDecl() + HandelC.END_COMMAND;
        	this.hcCode = this.hcCode + HandelC.FORCED_SYNC + HandelC.SPACE 
        				+ HandelC.ASSIGNMENT + HandelC.SPACE 
        				+ this.translator.getEmptyChannelSet() + HandelC.END_COMMAND;
        }
        
        node.getPSpecification().apply(this);
        
        this.hcCode = this.hcCode + translator.getHcCode();
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
    }
    
    public void caseACspProcessDefinition(ACspProcessDefinition node){
        String name = node.getName().toString().trim();
        if (name.equals(HandelC.MAIN)) {
            this.translator.setCurrentDefinition(HandelC.MAIN);
            PExpression expression = node.getExpression();
            expression.apply(this.translator);
        }
    }
    
}
