package csp2hc.core;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;
import java.util.HashSet;

public class UsedChannelsVisitor extends ExtendedDepthFirstAdapter {
    // Visitor that has the information about the processes
    private ProcessesInfoVisitor procsInfo;

    // Stack of process definitions visited during analysis of 
    // a single process
    private Stack visitedProcesses;
    
    // Alphabet that is being constructed
    private Set usedChannels;

    // Synchronisation channel set that is being constructed
    private Set syncChannels;
    
    public UsedChannelsVisitor(ProcessesInfoVisitor procInfo) {
        this.procsInfo = procInfo;
        this.usedChannels = new HashSet();
        this.visitedProcesses = new Stack();
}
    
    public Set getUsedChannels() { return this.usedChannels; }
    public void resetUsedChannels() { this.usedChannels = new HashSet(); }
    
    public void caseACspEvent(ACspEvent node){
        String name = node.getChannelName().toString().trim();
        this.usedChannels.add(name);
    }
    
    public void caseACspProcessCallExpr(ACspProcessCallExpr node){
        String name = ((ACspProcessCallExpr)node).getName().toString().trim();

        // If it is a process name
        if (this.procsInfo.containsProcArguments(name)){
        	if (!this.visitedProcesses.contains(name)) {
                this.visitedProcesses.push(name.trim());
                this.procsInfo.getProcDefinition(name.trim()).apply(this);
                this.visitedProcesses.pop();
        	}
        }
    }

}
