package csp2hc.util;

/*
 * CodeFormatting.java
 *
 * Formats an unformatted code in order to make it more readable
 * Created on 15 March 2005, 13:53
 * @author Marcel Oliveira
 */

import csp2hc.messages.HandelC;
import csp2hc.messages.Messenger;

import java.util.StringTokenizer;
import java.io.File;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FileNotFoundException;
import java.io.IOException;

public class CodeFormatting {

    /*
     * Insert a given string after a given symbol
     */
    private static String insertString(String input, String insertString, String afterString){
        String output = "";
        StringTokenizer tokenizer = new StringTokenizer(input,afterString,true);
        while(tokenizer.hasMoreTokens()){
            String next = tokenizer.nextToken();
            if(next.equals(afterString)){
                output = output + next + insertString;
            } else {
                output = output + next.trim();
            }
        }
        return output;
    }

    /*
     * Insert line breaks after a given symbol
     */
    private static String insertLineBreaks(String input, String afterString){
        return insertString(input, Messenger.BREAK_LINE, afterString);
    }
    
    /*
     * Counts how many times c happens in str
     */
    private static int count(char c, String str) {
        int count = 0;
        for(int i=0; i<str.length(); i++){
            if(str.charAt(i) == c) {
                count++;
            }
        }
        return count;
    }
    
    /*
     * Insert tabs after a given symbol
     */
    private static String insertTabs(String input){
        String output = "";
        String done = "";
        String toDo = "";
        StringTokenizer tokenizer = new StringTokenizer(input,Messenger.BREAK_LINE,true);
        while(tokenizer.hasMoreTokens()){
            String next = tokenizer.nextToken();
            done = done + next;
            if(next.equals(Messenger.BREAK_LINE)){
                int numberOfOpenBlocks = count(HandelC.BLOCK_BEGIN.charAt(0),done);
                int numberOfCloseBlocks = count(HandelC.BLOCK_END.charAt(0),done);
                int numberOfTabs = numberOfOpenBlocks - numberOfCloseBlocks;
                output = output + next + StringUtil.repeat(Messenger.TAB.charAt(0),numberOfTabs);
            } else {
                output = output + next.trim();
            }
        }
        
        // Removing extra tabs that were inserted before closing blocks
        output = output.replaceAll(Messenger.TAB+HandelC.BLOCK_END,HandelC.BLOCK_END);
        return output;
    }

    /*
     * Insert line breaks after a given symbol
     */
    private static String removeDoubles(String input, String remove){
        while(input.contains(remove+remove)) {
            input = input.replace(remove+remove,remove);
        }
        return input;
    }
    
    public static String format(String input){
        String output = input;
        if (input != null) {
            // Removing double spaces
            while(input.contains(HandelC.SPACE+HandelC.SPACE)) {
                input = input.replaceAll(HandelC.SPACE+HandelC.SPACE,HandelC.SPACE);
            }
            // Removing double ;
            output = removeDoubles(input,HandelC.END_COMMAND);
            
            // Removing all line breaks
            output = output.replace(Messenger.BREAK_LINE,"");

            // Filling possible empty blocks
            output = output.replace(HandelC.BLOCK_BEGIN+HandelC.BLOCK_END,
                                    HandelC.BLOCK_BEGIN+HandelC.END_COMMAND+HandelC.BLOCK_END);

            // Insert line breaks after block begin
            output = insertLineBreaks(output,HandelC.BLOCK_BEGIN);
            
            // Insert line breaks after block end
            output = insertLineBreaks(output,HandelC.BLOCK_END);
            
            // Insert line breaks after semi-colon
            output = insertLineBreaks(output,HandelC.END_COMMAND);

            // Insert Indicated line breaks
            output = StringUtil.insertIndicatedBreakLines(output);            

            // Insert indetation into the code
            output = insertTabs(output);

            // Insert Indicated spaces
            output = StringUtil.insertIndicatedSpaces(output);            
            
        }
        return output;
    }

    public static void main(String args[]){
        String function = "select";
        int argNumber = 3;
        String input = "select (a, b, c)";
        //CodeFormatting.formatFunctionInvocation(input,function,argNumber,0);
        
        /*
        String input = "";
        BufferedReader reader = null;
        String line;
        try {
            reader = new BufferedReader(new FileReader(new File("source.txt")));
            line = reader.readLine();
            while (line != null) {
                input = input + line + Messenger.BREAK_LINE;
                line = reader.readLine();
            }
        } catch (FileNotFoundException fnfe) {
            fnfe.printStackTrace();
        } catch (IOException io) {
            io.printStackTrace();
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException exception) {
                    exception.printStackTrace();
                }
            }
        }
        String output = CodeFormatting.format(input);
        BufferedWriter writer = null;
        try {
            writer = new BufferedWriter(new FileWriter(new File("target.txt")));
            writer.write(output);
        } catch (FileNotFoundException fnfe) {
            fnfe.printStackTrace();
        } catch (IOException io) {
            io.printStackTrace();
        } finally {
            if (writer != null) {
                try {
                    writer.close();
                } catch (IOException exception) {
                    exception.printStackTrace();
                }
            }
        }
         */
        
        
    }
    
}
