package csp2hc.util;

import java.util.Stack;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeNode;

import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.node.ACspSpecification;
import lmf.formula.csp.node.EOF;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.Start;
import lmf.formula.csp.node.TSingleLineComment;
import lmf.formula.csp.node.Token;


public class CspAstFactory extends ExtendedDepthFirstAdapter {

	private TreeView treeView;

	private Stack parents = new Stack();

	public CspAstFactory(TreeView treeView) {
		this.treeView = treeView;
	}

	public void outStart(Start node) {
	}

	public void outACspSpecification(ACspSpecification node) {
		this.treeView.setRoot((TreeNode) parents.pop());
		this.treeView.expandAll();
	}

	/*
	 * As we come across non terminals, push them onto the stack
	 */
	public void defaultIn(Node node) {
		String name = node.getClass().getSimpleName();

		if (name.indexOf("ACspBase") == -1) // No  um ACspBase...
		{
			name = name.substring(4); // Remove o "ACsp"
			if (name.startsWith("Linked")) {
				if (!node.parent().getClass().getSimpleName().endsWith("List")) {
					name = name.substring(6); // Remove o "Linked"
					DefaultMutableTreeNode thisNode = new DefaultMutableTreeNode(
							name);
					// thisNode.ForeColor = System.Drawing.Color.DarkBlue;
					parents.push(thisNode);
				}
			} else {
				DefaultMutableTreeNode thisNode = new DefaultMutableTreeNode(
						name);
				// thisNode.ForeColor = System.Drawing.Color.Blue;
				parents.push(thisNode);
			}
		} else //  um ACspBase...
		{
			if (name.endsWith("List")
					&& !node.parent().getClass().getSimpleName().endsWith("List")) {
				name = name.substring(8); // Remove o "ACspBase"
				DefaultMutableTreeNode thisNode = new DefaultMutableTreeNode(
						name);
				// thisNode.ForeColor = System.Drawing.Color.DarkBlue;
				parents.push(thisNode);
			}
		}
	}

	/*
	 * As we leave a non terminal, it's parent is the node before it on the
	 * stack, so we pop it off and add it to that node
	 */
	public void defaultOut(Node node) {
		String name = node.getClass().getSimpleName();
		if (name.indexOf("ACspBase") == -1) {
			name = name.substring(4); // Remove o "ACsp"
			if (name.startsWith("Linked")) {
				if (!node.parent().getClass().getSimpleName().endsWith("List")) {
					DefaultMutableTreeNode thisNode = (DefaultMutableTreeNode) parents
							.pop();
					((DefaultMutableTreeNode) parents.peek()).add(thisNode);
				}
			} else {
				DefaultMutableTreeNode thisNode = (DefaultMutableTreeNode) parents
						.pop();
				((DefaultMutableTreeNode) parents.peek()).add(thisNode);
			}
		} else {
			if (name.endsWith("List")
					&& !node.parent().getClass().getSimpleName().endsWith("List")) {
				DefaultMutableTreeNode thisNode = (DefaultMutableTreeNode) parents
						.pop();
				((DefaultMutableTreeNode) parents.peek()).add(thisNode);
			}
		}
	}

	/*
	 * Terminals - our parent is always on the top of the stack, so we add
	 * ourselves to it
	 */
	public void defaultCase(Node node) {
		if (node instanceof Token && !(node instanceof TSingleLineComment)) {
			String temp = ((Token) node).getText();
			temp = temp.replaceAll("\n\r", "\n");
			temp = temp.replaceAll("\r\n", "\n");

			String[] temp2 = temp.split("\n");
			for (int i = 0; i < temp2.length; i++) {
				String t = temp2[i];
				DefaultMutableTreeNode thisNode = new DefaultMutableTreeNode(t);
				// thisNode.ForeColor = System.Drawing.Color.Red;
				((DefaultMutableTreeNode) parents.peek()).add(thisNode);
			}
		} else {
			DefaultMutableTreeNode thisNode = new DefaultMutableTreeNode(node
					.toString().trim());
			// thisNode.ForeColor = System.Drawing.Color.Red;
			((DefaultMutableTreeNode) parents.peek()).add(thisNode);
		}
	}

	public void caseEOF(EOF node) {
	}

}