/*
 * HandelC.java
 *
 * Contains the Handel-C constructors used
 * Created on 18 January 2006, 18:17
 */

package csp2hc.messages;

import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import csp2hc.exception.InvalidSelectBlockArgumentsException;
import csp2hc.util.ListUtil;
import csp2hc.util.StringUtil;

/**
 *
 * @author  marcel oliveira
 */
public class HandelC {
    // File extensions
    public static String CSP_EXTENSION = "csp";
    public static String EXTENSION = "hcc";
    public static String DOT = ".";
    public static String TEMPORARY_FILE = "___temp";

    // Handel-C Keywords
    public static String ADDRESS = "&";
    public static String ALTERNATION = "prialt";
    public static String ARRAY_BEGIN = "[";
    public static String ARRAY_END = "]";
    public static String ASSIGNMENT = "=";    
    public static String AUTO = "auto";
    public static String BINARY_PREFIX = "0b";    
    public static String BLOCK_BEGIN = "{";
    public static String BIT_RANGE = ":";
    public static String BLOCK_END = "}";
    public static String BREAK = "break";
    public static String CASE = "case";
    public static String CASE_SEP = ":";
    public static String CHAN_DECL = "chan";
    public static String CHANIN = "chanin";
    public static String CHANOUT = "chanout";
    public static String CHAR = "char";
    public static String CLOCK = "clock";
    public static String CONDITIONAL_COND = "?";    
    public static String CONDITIONAL_SEP = ":";    
    public static String CONSTANT = "const";
    public static String CONTINUE = "continue";
    public static String DEFAULT = "default";
    public static String DEFINE = "#define";    
    public static String DELAY = "delay";
    public static String DO = "do";
    public static String DOUBLE = "double";
    public static String END_COMMAND = ";";
    public static String ENUM = "enum";    
    public static String EXPR = "expr";
    public static String EXTERN = "extern";
    public static String EXTERNAL_DIVIDE = "external_divide";
    public static String FAMILY = "family";
    public static String FLOAT = "float";
    public static String FOR = "for";
    public static String GOTO = "goto";
    public static String HC_MACRO = "macro";
    public static String IFSELECT = "ifselect";
    public static String IN = "in";
    public static String INLINE = "inline";
    public static String INPUT_COMMUNICATION = "?";
    public static String INT = "int";    
    public static String INT_WIDTH = "intwidth";
    public static String INTERFACE = "interface";
    public static String INTERNAL = "internal";
    public static String INTERNAL_DIVIDE = "internal_divide";
    public static String LINE_COMMENT = "//";
    public static String LOGICAL_AND = "&&";    
    public static String LOGICAL_BITWISE_AND = "&";    
    public static String LOGICAL_BITWISE_NOT = "~";    
    public static String LOGICAL_BITWISE_OR = "|";    
    public static String LOGICAL_BITWISE_XOR = "^";    
    public static String LOGICAL_NOT = "!";    
    public static String LOGICAL_OR = "||";    
    public static String CONCAT = "@";    
    public static String TAKE_LSB = "<-";    
    public static String LOGICAL_VALUES = "2";    
    public static String LONG = "long";    
    public static String MACRO_EXP = HC_MACRO + " " + EXPR;
    public static String PROC = "proc";    
    public static String MACRO_PROC = HC_MACRO + " " + PROC;
    public static String MAIN = "main";
    public static String MATH_DIFF = "-";    
    public static String MATH_DIV = "/";    
    public static String MATH_MULT = "*";    
    public static String MATH_NEG = "-";    
    public static String MATH_ONE = "1";    
    public static String MATH_POW = "**";    
    public static String MATH_REMAINDER = "%";    
    public static String MATH_SUM = "+";    
    public static String MATH_ZERO = "0";    
    public static String MPRAM = "mpram";    
    public static String OUTPUT_COMMUNICATION = "!";
    public static String PAR_BEGIN = "(";
    public static String PAR_END = ")";
    public static String PARALLEL = "par";
    public static String PART = "part";    
    public static String RAM = "ram";    
    public static String REFERENCE = "*";
    public static String REGISTER = "register";    
    public static String REL_EQUALS = "==";    
    public static String REL_GREATER = ">";    
    public static String REL_GREATER_EQUALS = ">=";    
    public static String REL_LESS = "<";    
    public static String REL_LESS_EQUALS = "<=";    
    public static String REL_NOT_EQUALS = "!=";    
    public static String RELEASE_SEMAPHOR = "releasesema";
    public static String RESET = "reset";    
    public static String RETURN = "return";
    public static String ROM = "rom";    
    public static String SELECT = "select";
    public static String SEMAPHOR = "sema";
    public static String SEP_ARGUMENTS = ",";
    public static String SEQUENCE = "seq";
    public static String SHARED = "shared";    
    public static String SHORT = "short";    
    public static String SIGNAL = "signal";    
    public static String SIGNED = "signed";    
    public static String SIZEOF = "sizeof";    
    public static String SPACE = " ";
    public static String STATIC = "static";
    public static String STRUCT = "struct";    
    public static String SWITCH = "switch";
    public static String TRY = "try";    
    public static String TRY_SEMAPHOR = "trysema";
    public static String TYPEDEF = "typedef";    
    public static String UNDEFINED = "undefined";    
    public static String UNSIGNED = "unsigned";    
    public static String VOID = "void";
    public static String VOLATILE = "volatile";    
    public static String WHILE = "while";    
    public static String WIDTH = "width";    
    public static String WITH = "with";    
    public static String WOM = "wom";

    // Handel-C and CSP keywords
    public static String EXTERNAL = "external";
    public static String IF = "if";    
    public static String ELSE = "else";    
    public static String LOGICAL_TRUE = "true";    
    public static String LOGICAL_FALSE = "false";    
    public static String LET = "let";    
    public static String SET = "set";    
    public static String SEQ = "seq";    
    public static String UNION = "union";    
    
    // Translator Keywords
    public static String UNDERSCORE = "_";
    public static String DUMB_CLOCK = "clock1";
    public static String SYNC = "sync";    
    public static String SYNC_TYPE = SYNC.toUpperCase();
    public static String BOOLEAN = "boolean";    
    public static String BRANCH = "BRANCH";
    public static String BRANCH_ID = "BRANCH" + HandelC.UNDERSCORE + "ID";
    public static String CHANNEL = "CHANNEL";
    public static String CLOCK_DEF = "set clock = external \""+DUMB_CLOCK+"\" "+END_COMMAND;
    public static String COMMENT_LINE_BREAK = UNDERSCORE+UNDERSCORE+"LINE_BREAK";    
    public static String COMMENT_SPACES = UNDERSCORE+UNDERSCORE+"SPACES"+HandelC.UNDERSCORE;    
    public static String COMMENT_TAB = UNDERSCORE+UNDERSCORE+"TAB";    
    public static String CURRENT_SYNC_STATUS_LUT = "CURRENT_SYNC_STATUS_LUT";
    public static String DONE_EVENTS = "DONE_EVENTS";
    public static String FALSE_DEF = LOGICAL_FALSE + SPACE + MATH_ZERO;    
    public static String FORCED_SYNC_LUT = "FORCED_SYNC_LUT";
    public static String HAS_JOINED = "HAS_JOINED";
    public static String INEXISTENT_CHANNEL = "INEXISTENT_CHANNEL";    
    public static String INTEGERL = "integer";    
    public static String IS_EMPTY_SET = "IS_EMPTY_SET";
    public static String JOIN_SEMAPHOR = "JOIN_SEMAPHOR";
    public static String KEEP_LOOPING = "KEEP_LOOPING";
    public static String LEAVE_SEMAPHOR = "LEAVE_SEMAPHOR";
    public static String LOCAL_COPY = UNDERSCORE+"local"+UNDERSCORE;
    public static String LOCK_VAR = "LOCK_VARIABLE";
    public static String LOOKUP_TABLE = "LUT";
    public static String MACRO = "MACRO";
    public static String MAINP = "mainp";
    public static String MAY_JOIN = "MAY_JOIN";
    public static String MUTUAL_RECURSION = "MUTUAL_REC";
    public static String NEG = "neg"+ UNDERSCORE;    
    public static String OFFSET = UNDERSCORE+"offset";
    public static String PROGRAM_COUNTER = "PROGRAM_COUNTER";
    public static String SET_DIFF = "SET_DIFF";
    public static String SET_INTER = "SET_INTER";
    public static String SET_UNION = "SET_UNION";
    public static String SINGLETON = "singleton";
    public static String RANDOM = "random";
    public static String RANDOM_VAR = "random_var";
    public static String SYNC_DEF =  TYPEDEF + SPACE + UNSIGNED + SPACE + INT + SPACE + MATH_ONE + SPACE + HandelC.SYNC_TYPE + END_COMMAND; 
    public static String SYNC_IN = "syncin";    
    public static String SYNC_IN_DEF = SYNC_TYPE + SPACE + SYNC_IN + END_COMMAND;    
    public static String SYNC_OUT = "syncout";    
    public static String SYNC_OUT_DEF = CONSTANT + SPACE + SYNC_TYPE + SPACE + SYNC_OUT + SPACE + ASSIGNMENT + MATH_ZERO + END_COMMAND;
    public static String TRUE_DEF = LOGICAL_TRUE + SPACE + MATH_ONE;    
    public static String TRY_JOIN_SEMAPHOR = "TRY_JOIN_SEMAPHOR";
    
    // SUFFIXES OF KEYWORDS
    private static String SET_SUFFIX = UNDERSCORE+"set";
    private static String CARDINALITY = UNDERSCORE+"card";
    private static String EMPTY_SET_SUFFIX = UNDERSCORE+"nil";
    private static String SEMAPHOR_SUFFIX = UNDERSCORE+"SEMA";

    // Argument names for macros 
    public static String INT_ARG_I = "I";
    public static String SET_ARG_S = "S";
    public static String SET_ARG_T = "T";
    public static String BRANCH_ARG_B = "B";
    public static String CHANNEL_ARG_C = "C";

    // Set of Handel-C keywords
    private static String[] hcKeywords = new String[]{
    	ALTERNATION,
    	AUTO,
    	BREAK,
    	CASE,
    	CHAN_DECL,
    	CHANIN,
    	CHANOUT,
    	CHAR,
    	CLOCK,
    	CONSTANT,
    	CONTINUE,
    	DEFAULT,
    	DEFINE,
    	DELAY,
    	DO,
    	DOUBLE,
    	ENUM,
    	EXPR,
    	EXTERN,
    	EXTERNAL_DIVIDE,
    	FAMILY,
    	FLOAT,
    	FOR,
    	GOTO,
    	HC_MACRO,
    	IFSELECT,
    	IN,
    	INLINE,
    	INT,
    	INT_WIDTH,
    	INTERFACE,
    	INTERNAL,
    	INTERNAL_DIVIDE,
    	LONG,
    	MAIN,
    	PARALLEL,
    	PART,
    	PROC,
    	RAM,
    	REFERENCE,
    	REGISTER,
    	RELEASE_SEMAPHOR,
    	RESET,
    	RETURN,
    	ROM,
    	SELECT,
    	SEMAPHOR,
    	SEQUENCE,
    	SHARED,
    	SHORT,
    	SIGNAL,
    	SIGNED,
    	SIZEOF,
    	STATIC,
    	STRUCT,
    	SWITCH,
    	TRY,
    	TRY_SEMAPHOR,
    	TYPEDEF,
    	UNDEFINED,
    	UNSIGNED,
    	VOID,
    	VOLATILE,
    	WHILE,
    	WIDTH,
    	WITH,   
    	WOM    
    };

    // Set of translator's keywords
    private static String[] toolKeywords = new String[]{
    	BOOLEAN,
    	BRANCH,
    	CHANNEL,
    	CURRENT_SYNC_STATUS_LUT,
    	DUMB_CLOCK,
    	FORCED_SYNC_LUT,
    	HAS_JOINED,
    	INEXISTENT_CHANNEL,
    	INTEGERL,
    	IS_EMPTY_SET,
    	JOIN_SEMAPHOR,
    	KEEP_LOOPING,
    	LEAVE_SEMAPHOR,
    	LOCK_VAR,
    	MAY_JOIN,
    	MUTUAL_RECURSION,
    	INTEGERL + OFFSET,
    	PROGRAM_COUNTER,
    	RANDOM,
    	RANDOM_VAR,
    	SET_DIFF,
    	SET_INTER,
    	SET_UNION,
    	SYNC,
    	SYNC_IN,
    	SYNC_OUT,
    	SYNC_TYPE,
    	TRY_JOIN_SEMAPHOR
    };
    
    private Set<String> furtherToolKeywords;
    private Vector<String> furtherToolKeywordsV;
    
    private static HandelC instance = null;
    private HandelC() {
    	this.furtherToolKeywords = new HashSet<String>();
    	this.furtherToolKeywordsV = new Vector<String>();
    }
    public static HandelC getInstance() {
       if(instance == null) {
          instance = new HandelC();
       }
       return instance;
    }
    
    public static Set<String> getHCKeywords(){
    	
    	// Build set of keywords
    	Set<String> keywordsSet = new HashSet<String>();
    	for (int i = 0; i < hcKeywords.length; i++) {
    		keywordsSet.add(hcKeywords[i]);
    	}

    	return keywordsSet;
    }

    public void addToolKeyword(String word) {
    	HandelC.getInstance().furtherToolKeywords.add(word);
    	HandelC.getInstance().furtherToolKeywordsV.add(word);
    }
    
    public Set<String> getToolKeywords(){
    	
    	// Build set of keywords
    	Set<String> keywordsSet = new HashSet<String>();
		keywordsSet.add(this.BRANCH_ID);
    	for (int i = 0; i < toolKeywords.length; i++) {
    		keywordsSet.add(toolKeywords[i]);
			//System.out.println(toolKeywords[i]);
    	}
		keywordsSet.addAll(furtherToolKeywords);
    	
		/*
		System.out.println("---------------------------------------------------");
		System.out.println("HandelC :: getToolKeywords -> further tool keywords");
		System.out.println("---------------------------------------------------");
		for (int i=0; i < HandelC.getInstance().furtherToolKeywordsV.size();i++) {
			System.out.println(HandelC.getInstance().furtherToolKeywordsV.get(i));
		}
		*/
		
    	return keywordsSet;
    }
    
    public String generateComment(String comment) {
        int length = comment.length();
        String beautyLine = HandelC.LINE_COMMENT + HandelC.SPACE;
        for (int i=0; i<length; i++) {
            beautyLine = beautyLine + HandelC.MATH_DIFF;
        }
        String commentLine = HandelC.LINE_COMMENT + HandelC.SPACE + comment;
        
        String result = HandelC.COMMENT_LINE_BREAK + beautyLine 
                        + HandelC.COMMENT_LINE_BREAK + commentLine 
                        + HandelC.COMMENT_LINE_BREAK + beautyLine 
                        + HandelC.COMMENT_LINE_BREAK;
        
        return result;
    }

    public String generateDefine(String definition) {
    	String define = HandelC.DEFINE + HandelC.SPACE + definition + HandelC.COMMENT_LINE_BREAK;
        return define;
    }
    
    public String getSetVarOf(String var){
    	String name = var + SET_SUFFIX;
    	if(!var.equals("")) {
        	HandelC.getInstance().addToolKeyword(name);
    	}
    	return name;
    }

    public String getSetOf(String var){
    	String name = var + SET_SUFFIX;
    	return name;
    }
    
    public String getEmptySetVarOf(String var){
    	String name = HandelC.getInstance().getSetVarOf(var) + EMPTY_SET_SUFFIX; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }
    
    public String getCardVarOf(String var){
    	String name = var + CARDINALITY; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }

    public String getSemaVarOf(String var){
    	String name = var + SEMAPHOR_SUFFIX; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }
    
    public String getBranchName(int i){
    	String name = HandelC.BRANCH + HandelC.UNDERSCORE + i; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }
    
    public String getChannelName(String c) {
    	String name = HandelC.CHAN_DECL + HandelC.UNDERSCORE + c; 
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }
    
    public String getSetUnionCode(List<String> elements, String type) {
		String retorno = HandelC.getInstance().getEmptySetVarOf(type);
    	if (elements.size() > 0) {
    		String head = elements.get(0);
    		elements.remove(0);
			retorno = HandelC.SET_UNION + HandelC.PAR_BEGIN
					+ head + HandelC.SEP_ARGUMENTS + getSetUnionCode(elements, type) 
					+ HandelC.PAR_END;
		}
		return retorno;
    }
    
    // Returns the name of the look up table for a given tag in a given datatype
    public String getLookupTableName(String type, String tag) {
    	String name = type + HandelC.UNDERSCORE + tag + HandelC.UNDERSCORE + HandelC.LOOKUP_TABLE;
    	if(!tag.equals("")) {
    		this.addToolKeyword(name);
    	}
    	return name;
    }

    // Returns the name of the singleton sets look up table for a given type
    public String getSingletonSetsLookupTableName(String type) {
    	String name = this.getLookupTableName(type,HandelC.getInstance().getSetVarOf(""));
    	this.addToolKeyword(name);
    	return name;
    }
    
    
    // Returns a conditional block for a given list of conditions and values
    // with same cardinality using the given identation (identation spaces).
    private String getConditionalBlock(List conditions, List values, int identation) {
        String select = "";
        
        if (conditions.size() > 0 && conditions.size() == values.size()) {
            int newIdentation = identation
                                + HandelC.MACRO.length();
            select = select + StringUtil.getSpace(identation) + HandelC.PAR_BEGIN;
            select = select + ListUtil.head(conditions);
            select = select + HandelC.CONDITIONAL_COND;
            select = select + HandelC.SPACE + ListUtil.head(values);
            select = select + HandelC.CONDITIONAL_SEP;
            select = select + HandelC.COMMENT_LINE_BREAK;

            // Creating the if-then-else block
            if (conditions.size() > 1) {
                select = select + this.getConditionalBlock(ListUtil.tail(conditions), ListUtil.tail(values), newIdentation);
            } else {
                select = select + StringUtil.getSpace(newIdentation) + ListUtil.head(values);
            }
            select = select + HandelC.PAR_END;
        } else {
            throw new InvalidSelectBlockArgumentsException();
        }
        
        return select;
    }
    
    public String getDoneEventsVarName(int diff){
    	String name = HandelC.DONE_EVENTS + HandelC.UNDERSCORE + diff;
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }

    public String getHasJoinedVarName(int diff){
    	String name = HandelC.HAS_JOINED + HandelC.UNDERSCORE + diff;
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }


    public String getArgName(String function, String arg) {
    	String name = function.toUpperCase() + HandelC.UNDERSCORE + arg;
    	HandelC.getInstance().addToolKeyword(name);
    	return name;
    }

    
}
