package csp2hc.util;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import javax.swing.JTree;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;


public final class TreeView extends JTree {

	public void setRoot(TreeNode node) {
		setModel(new DefaultTreeModel(node));
		this.updateUI();
	}

	public void expandAll() {
    	this.expandAll(new TreePath(this.getModel().getRoot()));
    }

	public void expandAll(TreePath path) {
    	Collection tps = extremalPaths(getModel(), path, new HashSet());
    	Iterator it = tps.iterator();
        while (it.hasNext()) {
        	TreePath tp = (TreePath) it.next();
            expandPath(tp);
        }
    }

    /** The "extremal paths" of the tree model's subtree starting at
           path. The extremal paths are those paths that a) are non-leaves
           and b) have only leaf children, if any. It suffices to know
           these to know all non-leaf paths in the subtree, and those are
           the ones that matter for expansion (since the concept of expan-
           sion only applies to non-leaves).
           The extremal paths of a leaves is an empty collection.
           This method uses the usual collection filling idiom, i.e.
           clear and then fill the collection that it is given, and
           for convenience return it. The extremal paths are stored
           in the order in which they appear in pre-order in the
           tree model.
     */
    public Collection extremalPaths(TreeModel data,
    		TreePath path, Collection result) {
        result.clear();

        if (data.isLeaf(path.getLastPathComponent())) {
            return result; // should really be forbidden (?)
        }
        extremalPathsImpl(data, path, result);

        return result;
    }

    private void extremalPathsImpl(TreeModel data,
    		TreePath path, Collection result) {
        Object node = path.getLastPathComponent();

        boolean hasNonLeafChildren = false;
        int count = data.getChildCount(node);
        for (int i = 0; i < count; i++) {
            if (!data.isLeaf(data.getChild(node, i))) {
                hasNonLeafChildren = true;
            }
        }

        if (!hasNonLeafChildren) {
            result.add(path);
        } else {
            for (int i = 0; i < count; i++) {
                Object child = data.getChild(node, i);

                if (!data.isLeaf(child)) {
                    extremalPathsImpl(data, path.pathByAddingChild(child),
                    		result);
                }
            }
        }
    }

}