package csp2hc.visitors;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class ProcessesInfoVisitor extends ExtendedDepthFirstAdapter {
    // Processes Arguments -- processName:String -> parameters:ACspParameters
    public Hashtable processesArguments;

    // Processes Definition -- processName:String -> definition:PExpression
    public Hashtable processesDefinitions;
    
    public ProcessesInfoVisitor() {
        this.processesArguments = new Hashtable();
        this.processesDefinitions = new Hashtable();
    }
    
    public boolean containsProcArguments(String procName) {
        return this.processesArguments.containsKey(procName);
    }
    
    public ACspParameters getProcArguments(String procName){
        return (ACspParameters)this.processesArguments.get(procName);
    }

    public boolean containsProcDefinition(String procName) {
        return this.processesDefinitions.containsKey(procName);
    }
    
    public PExpression getProcDefinition(String procName){
    	if (this.processesDefinitions.containsKey(procName)) {
            return (PExpression)this.processesDefinitions.get(procName);
    	} else {
    		throw new UndefinedProcessDefinitionException(procName);
    	}
    }
    
    public Enumeration keys() {
        return this.processesDefinitions.keys();
    }
    
    public int size() {
        return this.processesDefinitions.size();
    }
    
    public boolean isArgumentOf(String argName, String procName) {
        boolean answer = false;

        if (this.containsProcArguments(procName)) {
            ACspParameters parameters = this.getProcArguments(procName);
            if (parameters.getParamList() != null){
                Node paramsList = ((ACspParameters)parameters).getParamList();

                if (paramsList instanceof ACspBaseParamList) {
                    String parameter = ((ACspBaseParamList)paramsList).getParam().toString().trim();
                    if (argName.equals(parameter)) {
                        answer = true;
                    }
                } else if (paramsList instanceof ACspLinkedParamList) {
                    String parameter = ((ACspLinkedParamList)paramsList).getParam().toString().trim();
                    if (argName.equals(parameter)) {
                        answer = true;
                    }
                    while((((ACspLinkedParamList)paramsList)).getNext() instanceof ACspLinkedParamList){
                        paramsList = ((ACspLinkedParamList)paramsList).getNext();
                        parameter = ((ACspLinkedParamList)paramsList).getParam().toString().trim();
                        if (argName.equals(parameter)) {
                            answer = true;
                        }
                    }
                    paramsList = ((ACspLinkedParamList)paramsList).getNext();
                    parameter = ((ACspBaseParamList)paramsList).getParam().toString().trim();
                    if (argName.equals(parameter)) {
                        answer = true;
                    }
                }
            }
        }
        return answer;
    }

    public void caseACspDefinitionParagraph(ACspDefinitionParagraph node){
        PDefinition definition = node.getDefinition();
        if (definition instanceof ACspProcessDefinition){
            TName name = ((ACspProcessDefinition)definition).getName();
            ACspParameters parameters = (ACspParameters)((ACspProcessDefinition)definition).getParameters();
            PExpression expression = ((ACspProcessDefinition)definition).getExpression();

            if (parameters == null) {
                parameters = new ACspParameters();
            }
            
            String pname = name.toString().trim();
            
            this.processesArguments.put(pname,parameters);
            this.processesDefinitions.put(pname,expression);
        }
    }
}
