/*
 * HandelC.java
 *
 * Contains the messages used in the system
 * Created on 18 January 2006, 18:17
 */

package csp2hc.messages;

import java.util.Date;
import java.util.Hashtable;

import csp2hc.gui.LogWindow;
import csp2hc.util.CspUtil;

/**
 *
 * @author  marcel oliveira
 */
public class Messenger {
    private static Messenger messenger;

    // Translation stages
    public static String TRANSLATION                   	  =  "TRANSLATION"                 	  ;
    public static String LOADING_CONFIG                   =  "LOADING_CONFIG"                 ;
    public static String OPENING_FILE                     =  "OPENING_FILE"                   ;
    public static String STORING_DIRECTORY                =  "STORING_DIRECTORY"              ;
    public static String PARSING                          =  "PARSING"                        ;
    public static String TYPE_CHECKING                    =  "TYPE_CHECKING"                  ;
    public static String LOADING_DIRECTIVES               =  "LOADING_DIRECTIVES"             ;
    public static String LOADING_CONSTANTS_AND_FUNCTIONS  =  "LOADING_CONSTANTS_AND_FUNCTIONS";
    public static String LOADING_PROCESSES_INFO           =  "LOADING_PROCESSES_INFO"         ;
    public static String LOADING_DATATYPES                =  "LOADING_DATATYPES"              ;
    public static String LOADING_CHANNELS_INFO            =  "LOADING_CHANNELS_INFO"          ;
    public static String CHECKING_PRIMARY_RESTRICTIONS    =  "CHECKING_PRIMARY_RESTRICTIONS"  ;
    public static String CHECKING_SECONDARY_RESTRICTIONS  =  "CHECKING_SECONDARY_RESTRICTIONS";
    public static String GENERATING_MAIN                  =  "GENERATING_MAIN"                ;
    public static String CREATING_TEMP_FILE               =  "CREATING_TEMP_FILE"             ;
    public static String PARSING_TEMP_FILE                =  "PARSING_TEMP_FILE"              ;
    public static String TYPE_CHECKING_TEMP_FILE          =  "TYPE_CHECKING_TEMP_FILE"        ;
    public static String TRANSLATING_PROCESSES            =  "TRANSLATING_PROCESSES"          ;
    public static String FORMATTING_OUTPUT                =  "FORMATTING_OUTPUT"              ;
    public static String SAVING_OUTPUT                    =  "SAVING_OUTPUT"                  ;
    public static String REMOVING_TEMP_FILE               =  "REMOVING_TEMP_FILE"             ;
    public static String ANALISYNG_PARALLELISM            =  "ANALISYNG_PARALLELISM"          ;
    public static String SEARCHING_KEYWORDS            	  =  "SEARCHING_KEYWORDS"          ;
    
    // Output Messages
    public static String TRANSLATING_FILE = "Translating file ";
    public static String PROCESSING_DIRECTIVE = "Processing directive at line ";
    public static String OK = "OK";
    public static String CANCELLED = "Cancelled";
    public static String SAVE = "Save";
    public static String OPEN = "Open";
    public static String OVERWRITE_1 = "The file ";
    public static String OVERWRITE_2 = " already exists. \nWould you like to replace it?";
    public static String OVERWRITE_TITLE = "Replace?";
    public static String NEW_TRANSLATION = "Would you like to translate another file?";
    public static String NEW_TRANSLATION_TITLE = "New translation?";
    public static String TYPE = "TYPE: ";
    public static String NO_CONFIGURATION_FILE = "Unable to find configuration file. Please check that the configurations are right in the file /config/config_en.properties...";
    public static String CALLS = " -> ";
    public static String TRY_AGAIN = "The following error occurred during translation: \n";
    public static String TRY_AGAIN_2 = ". \nWould you like to fix the problem and try again?";
    public static String TRY_AGAIN_TITLE = "Try again?";
    public static String PROGRAM_TERMINATED = "Program terminated";
    public static String IN_MAIN = "In process main - ";
    
    private Hashtable messages;
    private Hashtable timing;
    
    private LogWindow window;

    /** Creates a new instance of Debug */
    private Messenger() {
        window = new LogWindow();
        window.launchFrame();
        
        this.messages = new Hashtable();
        messages.put(LOADING_CONFIG,"Loading configurations");
        messages.put(OPENING_FILE,"Opening file");
        messages.put(STORING_DIRECTORY,"Storing directory");
        messages.put(PARSING,"Parsing");
        messages.put(TYPE_CHECKING,"Type checking");
        messages.put(LOADING_DIRECTIVES,"Loading directives");
        messages.put(LOADING_CONSTANTS_AND_FUNCTIONS,"Loading constants and functions");
        messages.put(LOADING_PROCESSES_INFO,"Loading processes information");
        messages.put(LOADING_DATATYPES,"Loading datatypes");
        messages.put(LOADING_CHANNELS_INFO,"Loading channels information");
        messages.put(CHECKING_PRIMARY_RESTRICTIONS,"Checking primary translator restrictions");
        messages.put(CHECKING_SECONDARY_RESTRICTIONS,"Checking secondary translator restrictions");
        messages.put(GENERATING_MAIN,"Generating method main");
        messages.put(CREATING_TEMP_FILE,"Creating temporary file");
        messages.put(PARSING_TEMP_FILE,"Parsing temporary file containing main process");
        messages.put(TYPE_CHECKING_TEMP_FILE,"Type checking temporary file containing main process");
        messages.put(TRANSLATING_PROCESSES,"Translating processes");
        messages.put(FORMATTING_OUTPUT,"Formatting output code");
        messages.put(SAVING_OUTPUT,"Saving file");
        messages.put(REMOVING_TEMP_FILE,"Removing temporary file");
        messages.put(ANALISYNG_PARALLELISM,"Analysing parallel compositios from main process");
        messages.put(TRANSLATION,"Translating file");
        messages.put(SEARCHING_KEYWORDS,"Searching keywords");
        
        this.timing = new Hashtable();
    }

    public LogWindow getWindow() { return this.window; }
    
    /** Creates a new instance of Debug */
    public static Messenger getInstance() {
        if (messenger == null) {
            messenger = new Messenger();
        }
        return messenger;
    }

    // Symbols
    public static String BREAK_LINE = "\n";    
    public static String RETURN = "\r";    
    public static String TAB = "\t";    
    public static String DETAILS = ": ";    
    public static String OPEN_REASON = "( ";    
    public static String CLOSE_REASON = ")";    
    public static String SPACE = " ";    

    // Warnings & Exceptions
    public static String HANDEL_C =  "Handel-C"                 	  ;
    public static String TRANSLATOR =  "Translator"                 	  ;
    public static String WARNING = "WARNING: ";    
    public static String ERROR = "ERROR: ";
    public static String UNKNOWN_ERROR = "UNKNOWN ERROR: ";
    public static String LINE_INFO = "at line ";
    public static String INVALID_ABSTRACT_DEFINITION = "Invalid expression used in the definition of ";
    public static String INVALID_DIRECTIVE = " is invalid";
    public static String EXPECTED = "expected ";
    public static String PROCESS_NAME = "process name.";
    public static String INT_CHOICE_TYPE = "internal choice type.";
    public static String DEFINITION_NAME = "definition name.";
    public static String INEXISTENT_DIRECTIVE = "Inexistent directive type.";
    public static String DIRECTIVE = "Directive ";
    public static String DIRECTIVE_MISSING = "Directive missing: ";
    public static String CHANNEL_TYPE_TOO_FEW_INFO = "too few information. It should contain channel_type <NAME> <HANDEL_C TYPE> ";
    public static String CHANNEL_IN_OUT_TOO_MUCH_INFO = "too much information. It should contain channel <NAME> <{in,out}> within <NAME> ";
    public static String CHANNEL_IN_OUT_TOO_FEW_INFO = "too few information. It should contain channel <NAME> <{in,out}> within <NAME> ";
    public static String CHANNEL_IN_OUT_WRONG_TYPE = "channel type should be either in or out ";
    public static String CHANNEL_TYPE_UNDEFINED_1 = "Channel ";
    public static String CHANNEL_TYPE_UNDEFINED_2 = " is used as an input channel in process ";
    public static String CHANNEL_TYPE_UNDEFINED_3 = " and its type has not been specified by any directive.";
    public static String CHANNEL_UNDEFINED_IN_PROCESS_1 = "Channel ";
    public static String CHANNEL_UNDEFINED_IN_PROCESS_2 = " is not defined as either input or output by any directive for process ";
    public static String ARGUMENT_TOO_MUCH_INFO = "too much information. It should contain only argument name and type";
    public static String ARGUMENT_TOO_FEW_INFO = "too few information. It should contain var <NAME> <HANDEL_C TYPE> within <NAME>";
    public static String ARGUMENT_UNDEFINED_IN_PROCESS_1 = "The type of the argument ";
    public static String ARGUMENT_UNDEFINED_IN_PROCESS_2 = " in definition ";
    public static String ARGUMENT_UNDEFINED_IN_PROCESS_3 = " is not defined by any directive";
    public static String INT_CHOICE_TOO_FEW_INFO = "too few information. It should contain int_choice at <{compiletime,runtime}> ";
    public static String INT_CHOICE_TOO_MUCH_INFO = "too much information. It should contain only internal choice type (either compiletime or runtime)";
    public static String INT_CHOICE_WRONG_TYPE = "internal choice type should be either compiletime or runtime ";
    public static String UNSUPPORTED_CONSTRUCTOR_1 = "Constructor ";
    public static String UNSUPPORTED_CONSTRUCTOR_2 = " not supported by translator in the definition of ";
    public static String NON_SIMPLE_DATATYPE = "non-simple datatype";
    public static String NON_TAIL_RECURSION = "non tail-recursive definition";
    public static String PARAMETRISED_MUTUAL_RECURSION = "mutual recursive definition with parametrised processes";
    public static String INVALID_DEFINITION = "Invalid definition: ";
    public static String INPUT_PREFIX_SAME_VAR_NAME = "input prefixing in an external choice with same input variable names - ";
    public static String WRONG_NUMBER_OF_ARGUMENT_1 = "Wrong number of arguments for invoking ";
    public static String WRONG_NUMBER_OF_ARGUMENT_2 = " in definition of ";
    public static String ALPHABETISED_PARALLEL = "alphabetised parallel";
    public static String ALPHABETISED_REPLICATED_PROCESS = "alphabetised replicated process";
    public static String LINKED_PARALLEL = "linked parallel";
    public static String INVALID_TYPE_REFERENCE = "type reference to neither a valid datatype, Int, nor Bool";
    public static String INT_BITS_TOO_MUCH_INFO = "too much information. It should contain int_bits <POSITIVE NATURAL>";
    public static String INT_BITS_TOO_FEW_INFO = "too few information. It should contain int_bits <POSITIVE NATURAL>";
    public static String INT_BITS_WRONG_TYPE = "number of bits should be a positive natural number";
    public static String INT_BITS_UNDEFINED = "The number of bits for the CSP "+CspUtil.getIntType()+" not declared by any directive and used in definition ";
    public static String UNDEFINED_PROCESS_ARGUMENTS_1 = "No information on the arguments of process ";
    public static String UNDEFINED_PROCESS_ARGUMENTS_2 = " was found.";
    public static String UNDEFINED_PROCESS_DEFINITION_1 = "The definition of process ";
    public static String UNDEFINED_PROCESS_DEFINITION_2 = " could not be found. It may be the case that a process is being identified as a function";
    public static String INCONSISTENT_CHANNEL_PROJECTION_USE_1 = "Inconsistent use of channel projection of channel ";
    public static String INCONSISTENT_CHANNEL_PROJECTION_USE_2 = " in process ";
    public static String MUTUAL_RECURSION_AND_PARALLELISM = "parallel composition in mutually recursive processes";
    public static String SPECIFICATION = "specification";
    public static String INVALID_DATATYPE_TAG_1 = "Unable to retrieve information for tag ";
    public static String INVALID_DATATYPE_TAG_2 = " used in the definition of type ";
    public static String INVALID_DATATYPE_TAG_3 = " in order to create its MACRO.";
    public static String MUTUALLY_RECURSIVE_DATATYPES = "mutually recursive datatypes";
    public static String REPEATED_CONSTANT_DEFINITION = "Repeated definition of constant ";
    public static String UNDEFINED_NAME_REFERENCE_1 = "Reference to name ";
    public static String UNDEFINED_NAME_REFERENCE_2 = " in the definition of ";
    public static String UNDEFINED_NAME_REFERENCE_3 = " is undefined.";
    public static String DATATYPE_TAG_WITH_CONSTANT = "Datatype defined using constructors with constant arguments";
    public static String SET_COMPREHENSION = "set comprehension";
    public static String SET_TYPE = "set type";
    public static String RANGE = "range";
    public static String CLOSED_RANGE = "closed range";
    public static String OPEN_RANGE = "open range";
    public static String SEQUENCE_VALUE = "sequence values";
    public static String SEQUENCE_CONCATENATION = "sequence concatenation";
    public static String SEQUENCE_LENGTH = "length";
    public static String TUPLE_VALUE = "tuple values";
    public static String TUPLE_TYPE = "tuple type";
    public static String SET_TYPE_TOO_FEW_INFO = "too few information. It should contain set_of <HANDEL_C TYPE> within <NAME>";
    public static String SET_TYPE_REDEFINED = "only one type for sets is currently allowed for each definition.";
    public static String SET_TYPE_UNDEFINED_IN_DEFINITION_1 = "The type of the set in definition ";
    public static String SET_TYPE_UNDEFINED_IN_DEFINITION_2 = " is not defined by any directive.";
    public static String PARALLEL_COMPOSITION_FORCED_INTERLEAVING_1 = "The events that should be interleaved must be a constant for all nested parallel composition. \nIn definition ";
    public static String PARALLEL_COMPOSITION_FORCED_INTERLEAVING_2 = " the events that should be interleaved were ";
    public static String PARALLEL_COMPOSITION_FORCED_INTERLEAVING_3 = " at some point \nand ";
    public static String PARALLEL_COMPOSITION_FORCED_INTERLEAVING_4 = " at some other point";
    public static String UNKNOWN_PARALLEL_COMPOSITION = "Unknown Parallel Composition";
    public static String RENAMING = "renaming";
    public static String FUNCTION = "function";
    public static String TRANSPARENT = "transparent";
    public static String SUBTYPE = "subtype";
    public static String NAMETYPE = "nametype";
    public static String EXTERNAL = "external";
    public static String INTERRUPT = "interrupt";
    public static String INTERNAL_CHOICE = "internal choice";
    public static String UNTIMED_TIMEOUT = "untimed timeout";
    public static String EXTERNAL_CHOICE_OF_NON_PREFIX_PROCESSES = "external choice on non-prefixed processes";
    public static String REPLICATED_SEQUENTIAL_COMPOSITION = "replicated sequential composition";
    public static String REPLICATED_EXTERNAL_CHOICE = "replicated external choice";
    public static String REPLICATED_INTERNAL_CHOICE = "replicated internal choice";
    public static String REPLICATED_INTERLEAVE = "replicated interleave";
    public static String REPLICATED_SHARING = "replicated sharing";
    public static String REPLICATED_ALPHABETISED_PARALLEL = "replicated alphabetised parallel";
    public static String REPLICATED_LINKED_PARALLEL = "replicated linked parallel";
    public static String SYNCHRONISATION_CHANNEL_SET_CONSTANT = "constant in synchronisation channel sets";
    public static String SYNCHRONISATION_CHANNEL_SET_FUNCTION = "function calls in synchronisation channel sets";
    public static String SYNCHRONISATION_CHANNEL_SET_PROJECTION = "projections in synchronisation channel sets";
    public static String SYNCHRONISATION_CHANNEL_SET_INPUT = "input communications in synchronisation channel sets";
    public static String SYNCHRONISATION_CHANNEL_SET_OUTPUT = "output communications in synchronisation channel sets";
    public static String SYNCHRONISATION_CHANNEL_SET_COMPREHENSION = "set comprehension in synchronisation channel sets";
    public static String TAIL_RECURSION_IN_PARALLEL_COMPOSITION = "tail recursion in parallel compositions";
    public static String EVENT_WITH_DIFFERENT_DIRECTIONS_ON_EXTERNAL_CHOICE_1 = "channel ";
    public static String EVENT_WITH_DIFFERENT_DIRECTIONS_ON_EXTERNAL_CHOICE_2 = " with different directions in external choice";
    public static String EVENT_WITH_MULTIPLE_INPUT_ON_EXTERNAL_CHOICE_1 = "input on channel ";
    public static String EVENT_WITH_MULTIPLE_INPUT_ON_EXTERNAL_CHOICE_2 = " happens more than once in external choice";
    public static String HIDING_CHANNEL_WITH_NO_PARALLEL_COMPOSITION_1 = "hiding channel ";
    public static String HIDING_CHANNEL_WITH_NO_PARALLEL_COMPOSITION_2 = " which is not in parallel with any other process";
    public static String HIDING_CHANNEL_WITH_NO_COMMUNICATION_1 = "hiding channels ";
    public static String HIDING_CHANNEL_WITH_NO_COMMUNICATION_2 = " with no communication between both branches of the parallel composition";
    public static String HIDING_CHANNEL_WITH_COMMUNICATION_1 = "hiding channel(s) ";
    public static String HIDING_CHANNEL_WITH_COMMUNICATION_2 = " which is(are) in the synchronisation channel set ";
    public static String HIDING_CHANNEL_WITH_COMMUNICATION_3 = " and used in the other branch of a parallel composition";
    public static String MULTI_SYNCHRONISATION = "multi-synchronisation on channel(s) ";
    public static String KEYWORD_1 = " keyword ";
    public static String KEYWORD_2 = " is present in the specification";
    public static String READER_AND_WRITE_PARALLEL_BRANCH_1 = "Channel ";
    public static String READER_AND_WRITE_PARALLEL_BRANCH_2 = " is being read from and written to by a parallel branch in the definition of ";
    public static String NUMBER_OF_WRITERS_1 = "There is a parallel composition on channel ";
    public static String NUMBER_OF_WRITERS_2 = " on which the number of writers is different from one";
    
    
    // Internal Error messages
    public static String INVALID_LIST_OPERATION = "Invalid list operation: ";
    public static String CART_PROD_INVALID_LIST_OF_LISTS = "Function cartesianProduct can only be applied to List of Lists";
    public static String HEAD_OF_EMPTY_LIST = "Function head can only be applied to non-empty lists";
    public static String TAIL_OF_EMPTY_LIST = "Function tail can only be applied to non-empty lists";
    public static String LAST_OF_EMPTY_LIST = "Function last can only be applied to non-empty lists";
    public static String INVALID_SELECT_BLOCK_ARGUMENTS = "Invalid arguments given to the creation of a select block: the given list of conditions is either empty or has a different number of elements than the list of values.";
        
    // Code Comments
    public static String INT_DECLARATION = "Declaration of the integer constants";    
    public static String BOOL_DECLARATION = "Declaration of the boolean constants";    
    public static String CLOCK_DECLARATION = "Declaration of the system's clock";    
    public static String DATATYPE_DECLARATION = "Declaration of constants related to the datatype ";    
    public static String SYNC_VALUE_DECL = "Declaration of auxiliary constants for communication through synchronisation channels";    
    public static String CHANNELS_DECL = "Declaration of the channels";    
    public static String FORCED_INTERLEAVING_DECLARATION = "Declaration of variables, macros and functions used to force the interleaving between process";    
    public static String EMPTY_CHANNEL_SET_DECL = "Declaration of the empty channel set";    
    public static String MUTUAL_REC_STATE_DECL = "Declaration of constants for each of the processes in the mutual recursion";    
    public static String MUTUAL_REC_AUX_VARS = "Declaration of auxiliary variables for the mutual recursion";    
    public static String MUTUAL_REC_LOCAL_VARS = "Declaration of the local copies of each the processes arguments";    
    public static String MUTUAL_REC_METHOD = "Declaration of method corresponding to the mutual recursion";    
    public static String PROCESSES_PROTOTYPES = "Processes Prototypes";    
    public static String PROCESS = "Process";    
    public static String MAIN_DECLARATION = "Declaration of the main method";    
    public static String TAG_MACRO_1 = "Macro expression representing the contructor ";    
    public static String TAG_MACRO_2 = " used in the datatype ";    
    public static String CONSTANTS_AND_FUNCTIONS = "Constants and Functions";    
    public static String INT_MACRO = "Auxiliary macro used to remove the sign from an integer";    
    public static String AUX_MACROS = "Auxiliary macros";    
    public static String AUX_SEMA_MACROS = "Auxiliary macros related to the use of sempahors to force interleaving of events";    
    public static String SINGLETON_SETS_MACRO_DECLARATION = "Macro that returns singleton sets for every possible value in the system";    
    public static String CONSTANTS_AND_FUNCTIONS_PROTOTYPES = "Constants and functions prototypes";    
    public static String FORCED_INTERLEAVING_SEMAPHORS = "Semaphors used to have access to the variables used to force the interleaving between processes";    
    
    public synchronized void start(String stage) {
        this.message((String)this.messages.get(stage));
        long now = System.currentTimeMillis();
        this.timing.put(stage,new Long(now)); 
    }
    
    public synchronized void finish(String stage) {
        long now = System.currentTimeMillis();
        long before = ((Long)this.timing.get(stage)).longValue(); 
        
        long duration = now-before;
        long miliseconds = duration % 1000;
        duration = duration / 1000;
        long seconds = duration % 60;
        duration = duration / 60;
        long minutes = duration % 60;
        duration = duration / 60;
        long hours = duration % 60;
        duration = duration / 60;
        String time = hours + "hrs : " + minutes + "mins : " + seconds + "secs : " + miliseconds + "ms";
        
        this.message(this.messages.get(stage)+Messenger.SPACE+Messenger.OK+" ( "+ time +" )");
        
    }

    public synchronized void cancel(String stage) {
        this.message((String)this.messages.get(stage)+" "+Messenger.CANCELLED);
    }
    
    public synchronized void error(String error) {
        this.message(Messenger.ERROR + error);
    }
     
    public synchronized void uerror(String uerror) {
        this.message(Messenger.UNKNOWN_ERROR + uerror);
    }
    
    public synchronized void message(String msg) {
        this.window.addText(msg + "..." + BREAK_LINE);
    }

    public void addComment(String comment) {
        int length = comment.length();
        String beautyLine = "";
        for (int i=0; i<length; i++) {
            beautyLine = beautyLine + HandelC.MATH_DIFF;
        }
        String result = BREAK_LINE + beautyLine 
                        + BREAK_LINE + comment
                        + BREAK_LINE + beautyLine;
        
        this.message(result);
    }
    
}
