package csp2hc.visitors;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class ConstantsAndFunctionsVisitor extends ExtendedDepthFirstAdapter {

    // Visitor that has the information about the directies
    protected DirectivesVisitor directives;

    // Name:List String
    private String hcCode;
    
    // Constants on the specification
    // Name:List String
    private Vector constants;

    // Arguments of functions
    // Name:String -> Arguments:List String
    private Hashtable arguments;
    
    // Indicates that we are translating a constant
    private boolean isConstant;

    // List of arguments of the current definition
    private List currentArguments;
    
    public ConstantsAndFunctionsVisitor(DirectivesVisitor directives) {
        this.directives = directives;
        this.hcCode = "";
        this.constants = new Vector();
        this.isConstant = false;
        this.arguments = new Hashtable();
    }

    public String getHcCode() {
        return this.hcCode;
    }
    
    public Enumeration getConstantsNames(){
        return this.constants.elements();
    }

    public boolean isConstant(String name){
        return this.constants.contains(name);
    }
    
    public boolean containsConstant(String name){
        return this.constants.contains(name);
    }

    public List getArguments(String name){
        return (List)this.arguments.get(name);
    }
    
    public void caseACspConstantDefinition(ACspConstantDefinition definition) {
        this.isConstant = true;
        
        String name = definition.getName().toString().trim();
        PExpression expression = definition.getExpression();
        
        ExpressionsVisitor visitor = 
        	new ExpressionsVisitor(this.directives,true);
        visitor.setCurrentDefinition(name);
        
        if (this.containsConstant(name)) {
            throw new RepeatedConstantException(name);
        } else {
            this.currentArguments = new Vector();
            this.hcCode = this.hcCode + HandelC.MACRO_EXP + HandelC.SPACE + name;

            this.hcCode = this.hcCode + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.PAR_BEGIN;
            expression.apply(visitor);
            this.hcCode = this.hcCode + visitor.getHcCode();
            this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.END_COMMAND;
            
            this.constants.add(name);
            this.arguments.put(name,this.currentArguments);
        }
        
        this.isConstant = false;
    }
    
    public void caseACspFunctionDefinition(ACspFunctionDefinition definition) {
        this.isConstant = true;
        
        String name = definition.getName().toString().trim();
        PExpression expression = definition.getExpression();
        
        ExpressionsVisitor visitor = new ExpressionsVisitor(this.directives,true);
        visitor.setCurrentDefinition(name);
        
        if (this.containsConstant(name)) {
            throw new RepeatedConstantException(name);
        } else {
            this.currentArguments = new Vector();
            this.hcCode = this.hcCode + HandelC.MACRO_EXP + HandelC.SPACE + name;

            if (expression instanceof ACspLambdaExpression) {
                ACspLambdaExpr lambda = (ACspLambdaExpr)((ACspLambdaExpression)expression).getLambdaExpr();
                PParamList params = lambda.getParamList();
                expression = lambda.getExpression();

                this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
                params.apply(this);
                this.hcCode = this.hcCode + HandelC.PAR_END;
            } else {
                PParameters params = definition.getParameters();

                this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
                params.apply(this);
                this.hcCode = this.hcCode + HandelC.PAR_END;
            }

            this.hcCode = this.hcCode + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.PAR_BEGIN;
            expression.apply(visitor);
            this.hcCode = this.hcCode + visitor.getHcCode();
            this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.END_COMMAND;
            
            this.constants.add(name);
            this.arguments.put(name,this.currentArguments);
        }
        
        this.isConstant = false;
    }

    public void caseACspLinkedParamList(ACspLinkedParamList node) {
        if (this.isConstant) {
            node.getParam().apply(this);
            this.hcCode = this.hcCode + HandelC.SEP_ARGUMENTS;
            node.getNext().apply(this);
        }
    }

    public void caseACspParam(ACspParam node) {
        if (this.isConstant) {
            String argument = node.getExpression().toString().trim();
            this.hcCode = this.hcCode + argument;
            this.currentArguments.add(argument);
        }
    }

    public void caseACspAssert(ACspAssert node) {
    }    
}

