package csp2hc.visitors;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class ExpressionsVisitor extends ExtendedDepthFirstAdapter {

    // Visitor that has the information about the directies
    protected DirectivesVisitor directives;

    // Resulting handel-c code
    protected String hcCode = "";

    // Current definition being read
    protected String currentDefinition;

    // Indicates that we are analysing a set
    boolean isSet;

    // Indicates that we are analysing an array index
    boolean isArrayIndex;
    
    // Indicates if this is being used to translate macro expressions
    boolean isMacroExpression;
    
    public ExpressionsVisitor() {}
    	
    public ExpressionsVisitor(DirectivesVisitor directives,
    						boolean isMacroExpression) {
        this.directives = directives;
        this.currentDefinition = "";
        this.hcCode = "";
        
        this.isSet = false;
        this.isArrayIndex = false;
        this.isMacroExpression = isMacroExpression;
    }
    
    public String getHcCode() { return this.hcCode; }

    public void setCurrentDefinition(String def){ 
        this.currentDefinition = def; 
    }
    
    // Booleans
    public void caseACspAndLogicalExpr(ACspAndLogicalExpr node) {
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.SPACE + HandelC.LOGICAL_AND + HandelC.SPACE;
        node.getRightExpr().apply(this);
    }

    public void caseACspOrLogicalExpr(ACspOrLogicalExpr node) {
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.SPACE + HandelC.LOGICAL_OR + HandelC.SPACE;
        node.getRightExpr().apply(this);
    }
    
    public void caseACspNotExpr(ACspNotExpr node) {
        this.hcCode = this.hcCode + HandelC.LOGICAL_NOT;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getExpression().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    // Numbers 

    public void caseACspNumberValue(ACspNumberValue node) {
        this.hcCode = this.hcCode + node.toString().trim();
    }

    public void caseACspAddMathExpr(ACspAddMathExpr node) {
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.MATH_SUM;
        node.getRightExpr().apply(this);
    }

    public void caseACspDiffMathExpr(ACspDiffMathExpr node) {
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.MATH_DIFF;
        node.getRightExpr().apply(this);
    }

    public void caseACspMulMath1Expr(ACspMulMath1Expr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.MATH_MULT;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    public void caseACspDivMath1Expr(ACspDivMath1Expr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.MATH_DIV;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    public void caseACspRemainderMath1Expr(ACspRemainderMath1Expr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.MATH_REMAINDER;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    public void caseACspMinusExpr(ACspMinusExpr node) {
        this.hcCode = this.hcCode + HandelC.MATH_NEG;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getExpression().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    // Relations
    public void caseACspEqualRelationalExpr(ACspEqualRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_EQUALS;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    public void caseACspNotEqualRelationalExpr(ACspNotEqualRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_NOT_EQUALS;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    public void caseACspGtRelationalExpr(ACspGtRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_GREATER;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    public void caseACspGteRelationalExpr(ACspGteRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_GREATER_EQUALS;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    public void caseACspLtRelationalExpr(ACspLtRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_LESS;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    public void caseACspLteRelationalExpr(ACspLteRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_LESS_EQUALS;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    
    // Tuple
    public void caseACspTupleValue(ACspTupleValue node) {
    }

    // Sets
    public void caseACspSetValue(ACspSetValue node) {
        this.isSet = true;
        node.getSet().apply(this);
        this.isSet = false;
    }

    public void caseACspSet(ACspSet node) {
    	PElementList elements = node.getElementList();
        String type = "";
        if (this.directives.containsSetType(this.currentDefinition)) {
            type = (String)this.directives.getSetType(this.currentDefinition);
        } else {
            throw new UndefinedEmptySetTypeException(this.currentDefinition);
        }
        
		List<String> elementsStr = new Vector<String>();
		List<String> luts = new Vector<String>();

		if (elements != null) {
			while(! (elements instanceof ACspBaseElementList)) {
				ACspLinkedElementList list = (ACspLinkedElementList)elements;
				PCollectionElement element = list.getCollectionElement();
				
				if (element instanceof ACspRangeCollectionElement) {
	    			// It could be a range but we do not support it yet
					throw new UnsupportedConstructorException(Messenger.RANGE,this.currentDefinition);
				} else if (element instanceof ACspCollectionElement){
					ExpressionsVisitor visitor = 
						new ExpressionsVisitor(this.directives,true);
					element.apply(visitor);
					String elementStr = visitor.getHcCode();
					elementsStr.add(visitor.getHcCode());
				}
				elements = list.getNext();
			}
			ACspBaseElementList list = (ACspBaseElementList)elements;
			PCollectionElement element = list.getCollectionElement();
			
			if (element instanceof ACspRangeCollectionElement) {
				// It could be a range but we do not support it yet
				throw new UnsupportedConstructorException(Messenger.RANGE,this.currentDefinition);
			} else if (element instanceof ACspCollectionElement){
				ExpressionsVisitor visitor = 
					new ExpressionsVisitor(this.directives,true);
				element.apply(visitor);
				elementsStr.add(visitor.getHcCode());
			}
			
			for(int i=0; i<elementsStr.size(); i++) {
				String expStr = elementsStr.get(i);
				String lut = HandelC.getInstance().getSingletonSetsLookupTableName(type);
				String newElement = lut + HandelC.ARRAY_BEGIN + expStr + HandelC.ARRAY_END;
				luts.add(newElement);
			}
		}
		
        this.hcCode = this.hcCode + HandelC.getInstance().getSetUnionCode(luts, type);
    }
    
    
    // Sequences
    public void caseACspSequenceValue(ACspSequenceValue node) {
    }

    public void caseACspArgument(ACspArgument node) {
        String macro = HandelC.getInstance().getSetVarOf(HandelC.SINGLETON) + HandelC.UNDERSCORE + HandelC.MACRO;
        if(this.isSet) {
            this.hcCode = this.hcCode + macro + HandelC.PAR_BEGIN;
        }
        node.getExpression().apply(this);
        if(this.isSet) {
            this.hcCode = this.hcCode + HandelC.PAR_END;
        }

    }

    public void caseACspLinkedArgumentList(ACspLinkedArgumentList node) {
        PArgument argument = node.getArgument();
        argument.apply(this);
        
        PArgumentList next = node.getNext();
        if(this.isSet) {
            if (next != null){
                this.hcCode = this.hcCode + HandelC.SPACE + HandelC.LOGICAL_BITWISE_OR + HandelC.SPACE;
                next.apply(this);
            }
        }
    }
    
    public void caseACspNameCallExpr(ACspNameCallExpr node){
        this.hcCode = this.hcCode + node.toString().trim();
    }
    
    public void caseACspConstantCallExpr(ACspConstantCallExpr node){
        if (this.isMacroExpression) {
            this.hcCode = this.hcCode + node.getName().getText().trim();
        }
    }
    
	public void caseACspProductionsSet(ACspProductionsSet node){
		// System.out.println("ExpressionsVisitor :: caseACspProductionsSet -> I was here!!!");
        // this.hcCode = this.hcCode + HandelC.getEmptySetVarOf(HandelC.CHANNEL);
    }
    
    public void caseACspIfExpr(ACspIfExpr node){
        if (this.isMacroExpression) {
            this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
            this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
            node.getCondition().apply(this);
            this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.CONDITIONAL_COND + HandelC.SPACE;
            this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
            node.getThenExpr().apply(this);
            this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.CONDITIONAL_SEP + HandelC.SPACE;
            this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
            node.getElseExpr().apply(this);
            this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.SPACE;
            this.hcCode = this.hcCode + HandelC.PAR_END;
        } else {
            this.hcCode = this.hcCode + HandelC.IF + HandelC.PAR_BEGIN;
            node.getCondition().apply(this);
            this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.BLOCK_BEGIN;
            node.getThenExpr().apply(this);
            this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.SPACE;
            this.hcCode = this.hcCode + HandelC.ELSE + HandelC.SPACE + HandelC.BLOCK_BEGIN;
            node.getElseExpr().apply(this);
            this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.SPACE;
        }
    }

    public void caseACspAssert(ACspAssert node) {
    }
    
}
