package csp2hc.visitors;

import java.io.PushbackReader;
import java.io.StringReader;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;
import java.util.Stack;
import java.util.Vector;

import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.lexer.Lexer;
import lmf.formula.csp.node.ACspAlphabetisedParallelProcess;
import lmf.formula.csp.node.ACspArguments;
import lmf.formula.csp.node.ACspAssert;
import lmf.formula.csp.node.ACspBaseArgumentList;
import lmf.formula.csp.node.ACspBaseParamList;
import lmf.formula.csp.node.ACspBaseProcess;
import lmf.formula.csp.node.ACspBaseRenamingProcess;
import lmf.formula.csp.node.ACspBaseReplicatedHelp;
import lmf.formula.csp.node.ACspBaseReplicatedProcess;
import lmf.formula.csp.node.ACspCollectionElement;
import lmf.formula.csp.node.ACspConstantCallExpr;
import lmf.formula.csp.node.ACspConstantDefinition;
import lmf.formula.csp.node.ACspDotExpr;
import lmf.formula.csp.node.ACspEvent;
import lmf.formula.csp.node.ACspExternalChoiceProcess;
import lmf.formula.csp.node.ACspFunctionCallExpr;
import lmf.formula.csp.node.ACspFunctionDefinition;
import lmf.formula.csp.node.ACspGuardedProcess;
import lmf.formula.csp.node.ACspIfProcess;
import lmf.formula.csp.node.ACspInputComunication;
import lmf.formula.csp.node.ACspInterleaveProcess;
import lmf.formula.csp.node.ACspInternalChoiceProcess;
import lmf.formula.csp.node.ACspLetExpr;
import lmf.formula.csp.node.ACspLinkedArgumentList;
import lmf.formula.csp.node.ACspLinkedParamList;
import lmf.formula.csp.node.ACspOutputComunication;
import lmf.formula.csp.node.ACspParameters;
import lmf.formula.csp.node.ACspPrefixProcess;
import lmf.formula.csp.node.ACspProcessCallExpr;
import lmf.formula.csp.node.ACspProcessDefinition;
import lmf.formula.csp.node.ACspProductionsSet;
import lmf.formula.csp.node.ACspProjection;
import lmf.formula.csp.node.ACspRenamingProcess;
import lmf.formula.csp.node.ACspSequentialProcess;
import lmf.formula.csp.node.ACspSharingParallelProcess;
import lmf.formula.csp.node.ACspSubstitution;
import lmf.formula.csp.node.ACspType;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.PArgument;
import lmf.formula.csp.node.PArgumentList;
import lmf.formula.csp.node.PChoiceProcess;
import lmf.formula.csp.node.PDefinition;
import lmf.formula.csp.node.PDefinitionList;
import lmf.formula.csp.node.PEvent;
import lmf.formula.csp.node.PExpr;
import lmf.formula.csp.node.PExpression;
import lmf.formula.csp.node.PHelp;
import lmf.formula.csp.node.PParam;
import lmf.formula.csp.node.PParamList;
import lmf.formula.csp.node.PProcess;
import lmf.formula.csp.node.PProjection;
import lmf.formula.csp.node.PReplicatedProcess;
import lmf.formula.csp.node.PType;
import lmf.formula.csp.node.Start;
import lmf.formula.csp.parser.Parser;
import lmf.formula.csp.type.CspBoolType;
import lmf.formula.csp.type.CspDataType;
import lmf.formula.csp.type.CspIntType;
import lmf.formula.csp.type.CspSetType;
import lmf.formula.csp.type.CspType;
import lmf.formula.csp.type.CspUnknownType;
import mathCollection.MathSet;
import csp2hc.exception.CouldNotInferTypeException;
import csp2hc.exception.UndefinedArgumentSetTypeException;
import csp2hc.exception.UndefinedArgumentTypeException;
import csp2hc.exception.UndefinedChannelInOutException;
import csp2hc.exception.UndefinedChannelTypeException;
import csp2hc.exception.UndefinedProcessArgumentsException;
import csp2hc.exception.UnsupportedConstructorException;
import csp2hc.exception.WrongNumberOfArgumentsException;
import csp2hc.messages.Directives;
import csp2hc.messages.HandelC;
import csp2hc.messages.Messenger;
import csp2hc.messages.Warning;
import csp2hc.util.Branch2BranchSetLUT;
import csp2hc.util.BranchActionsOnChannel;
import csp2hc.util.Config;
import csp2hc.util.CspUtil;
import csp2hc.util.ForcedInterleavingOnChannel;
import csp2hc.util.ListUtil;
import csp2hc.util.MathUtil;

public class ProcessesTranslatorVisitor extends ExpressionsVisitor {
    // Visitor that has the information about the processes
    private ProcessesInfoVisitor procsInfo;

    // Visitor that has the information about the types
    private DatatypesVisitor types;

    // Visitor that has the information about recursion
    private TailRecursionVisitor recursion;

    // Visitor that has the information about the channels
    private ChannelsVisitor channels;

    // Visitor that has the information about Parallelism properties of the system
    private ParallelismAnalyserVisitor parallelismAnalyser;

    // Visitor that has the information about the constants
	private ConstantsAndFunctionsVisitor constants;
	
	private MultiSyncVisitor multisyncs;
    
    // Current prefixing being read
    private String currentPrefixing;

    // Current List of the remaining types for the current prefixing
    private List currentPrefixingTypes;

    // Current type for the current prefixing
    private String currentType;
    
    // Identity of Processes used in the translation of Mutual Recursion
    private List<String> processesIds;
    
    // List of current input variables in context
    private List inputVariables;

    // List of input variables types
    private Hashtable<String,Stack<String>> inputVarTypes;
    
    // Mapping of input variable names to buses 
    private Hashtable<String, Stack<String>> inputVarRenaming;
    
    // Indicates if the translator is translating a channel projection
    private boolean isProjection;

    // Stack of number of arguments to be translated for each tag
    // Changes the values according to the translation
    // Stack of int
    private Stack tagsArgsLeftStack;

    // Stack of number of arguments to be translated for each tag
    // Never changes the values (either push or pop)
    // Stack of int
    private Stack tagsArgsTotalStack;

    // Stack of types of arguments to be translated for each tag
    // Stack of (List of String)
    private Stack tagsTypesStack;
    
    // Stack of sets of events whose interleaving must be forced
    private Stack forcedInterleaving;

    // Flag that indicates if we should use semaphors or not
    private Boolean useBranchIds;

    // List of warning created during translation
    private List<Warning> warnings;
    
    private boolean isRenaming; // AddIvan
	private Hashtable <String, List<String>> currentSubstitutionList;
	private boolean hasLetDefinition;

    // Indicates if the translator is translating a multi-sync communication
    private boolean isMultiSync = false; // AddIvan
    
    // Used to return the Handel-C code in Multi-sync translations 
    //private String hcMultiSyncCode = "";
	
    private static boolean IS_ARGUMENT = true;
    private static boolean IS_NOT_ARGUMENT = false;
    private static boolean IS_DECLARATION = true;
    private static boolean IS_NOT_DECLARATION = false;
    private static boolean IS_PRI_ALT = true;
    private static boolean IS_NOT_PRI_ALT = false;
    private static boolean USE_DEFAULT = true;
    private static boolean DO_NOT_USE_DEFAULT = false;
    
    public ProcessesTranslatorVisitor(ProcessesInfoVisitor procsInfo, 
                                        DirectivesVisitor directives, 
                                        DatatypesVisitor types, 
                                        TailRecursionVisitor recursion,
                                        ChannelsVisitor channels,
                                        ConstantsAndFunctionsVisitor constants,
                                        MultiSyncVisitor multisyncs, 
                                        ParallelismAnalyserVisitor parallelismAnalyser) {
        
        super(directives, false);
        
        this.procsInfo = procsInfo;
        this.types = types;
        this.recursion = recursion;
        this.channels = channels;
        this.constants = constants;
        this.multisyncs = multisyncs;
        this.parallelismAnalyser = parallelismAnalyser;
        
        this.currentPrefixing = "";
        this.currentPrefixingTypes = new Vector();
        this.currentType = "";
        
        // Creating Ids for each process
        this.processesIds = new Vector<String>();
        Enumeration<String> keys = this.procsInfo.keys();
        while(keys.hasMoreElements()) {
            this.processesIds.add(keys.nextElement());
        }
        
        this.inputVariables = new Vector();
        this.inputVarRenaming = new Hashtable<String, Stack<String>>();
        this.inputVarTypes = new Hashtable<String, Stack<String>>();

        this.isProjection = false;
        this.tagsArgsLeftStack = new Stack();
        this.tagsArgsTotalStack = new Stack();
        this.tagsTypesStack = new Stack();
        this.forcedInterleaving = new Stack();
        this.useBranchIds = 
        		this.parallelismAnalyser.isMultiSync();
        this.warnings = new Vector<Warning>();
    }

    private String getCurrentDefinition(){ 
        return this.currentDefinition;
    }
    
    public List<Warning> getWarnings() {
    	return this.warnings;
    }
    
    public void caseStart(Start node){
    	if (!this.recursion.getMutualRecursion()) {
            node.getPSpecification().apply(this);
        } else {
            this.translateProcessesDefinitionsMutualRecursion();
        }
        
        // Preamble of file
        if (this.recursion.getMutualRecursion()) {
            this.declareMutualRecConstants();
        } else {
            this.declareProcessesPrototypes();
        }
        
        if (this.parallelismAnalyser.isMultiSync()){
        	List <String> prototypes = new Vector<String>();
        	prototypes.addAll(this.declareMultiSyncControls());
        	prototypes.addAll(this.declareMultiSyncClients(node));
        	for (String prototype : prototypes){
        		this.hcCode = prototype + this.hcCode;
        	}
        	this.hcCode = HandelC.getInstance().generateComment("Multi-Synchronization Prototypes") + this.hcCode;
        	
        	// The multi-sync writer will be always Zero. 
        	//this.declareMultiSyncWriterLUT();
        	this.declareMultiSyncIdLUT();
        }
        
        // Check if there are any buses declared
    	if (this.channels.getInputBuses().isEmpty()) {
        	this.warnings.add(new Warning(Messenger.WARNING_NO_INPUT_BUSES));
    	}
    	if (this.channels.getOutputBuses().isEmpty()) {
        	this.warnings.add(new Warning(Messenger.WARNING_NO_OUTPUT_BUSES));
    	}

        this.declareBuses();
        this.declareConstantsAndFunctions();
        this.declareChannels();
        //this.declareAuxiliarySetMacros();

        if (this.parallelismAnalyser.isMultiSync()){
        	this.declareAuxiliaryMultiSyncMacros();
        	this.declareMultiSyncDatatypes();
        	this.declareMultiSyncConstants();
        	this.declareActionsOnChannelsConstants();
        	this.declareBranches();
        }
        this.declareChannelsDatatypes();
        this.declareDatatypes();
        this.declareSyncConstants();
        this.declareClock();
        
    }
        
    public void caseACspAssert(ACspAssert node) {
    }
    
    public void caseACspProcessDefinition(ACspProcessDefinition definition) {

    	String name = definition.getName().toString().trim();
    	
    	System.out.println("ProcessesTranslatorVisitor :: caseACspProcessDefinition -> Process "+name);
    	
    	this.setCurrentDefinition(name.toString().trim());
    	
    	if (name.equals(HandelC.MAINP)) {
    		PExpression expression = definition.getExpression();
    		this.declareMainFunction(expression);
    	} else if (!this.recursion.getMutualRecursion()) {
    		ACspParameters parameters = (ACspParameters)definition.getParameters();
    		PExpression expression = definition.getExpression();
    		if (parameters == null) {
    			parameters = new ACspParameters();
    		}
    		
    		this.hcCode = this.hcCode + HandelC.getInstance().generateComment(Messenger.PROCESS + HandelC.SPACE + name);
    		this.hcCode = this.hcCode + HandelC.INLINE + HandelC.SPACE;
    		this.hcCode = this.hcCode + HandelC.VOID + HandelC.SPACE + name;
    		this.hcCode = this.hcCode + this.translateFormalArguments(parameters,this.currentDefinition);
    		this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
    		
    		if(this.recursion.getTailRecursiveProcesses().contains(this.currentDefinition)) {
    			// It is a tail recursion
    			
    			// int 1 KEEP_LOOPING;
    			this.hcCode = this.hcCode + HandelC.BOOLEAN + HandelC.SPACE;
    			this.hcCode = this.hcCode + HandelC.KEEP_LOOPING;
    			this.hcCode = this.hcCode + HandelC.END_COMMAND;
    			
    			// Declare argument copies
    			this.hcCode = this.hcCode + this.translateFormalArgumentsCopies(parameters,this.currentDefinition,IS_DECLARATION,IS_NOT_ARGUMENT);
    			
    			// Initialise them
    			this.hcCode = this.hcCode + this.translateFormalArgumentsCopies(parameters,this.currentDefinition,IS_NOT_DECLARATION,IS_NOT_ARGUMENT);
    			
    			// KEEP_LOOPING = true;
    			this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_TRUE);
    			
    			// Loop
    			this.hcCode = this.hcCode + HandelC.WHILE ;
    			this.hcCode = this.hcCode + HandelC.PAR_BEGIN + HandelC.KEEP_LOOPING + HandelC.PAR_END;
    			this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
    			this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_FALSE);
    			this.hcCode = this.hcCode + HandelC.SEQUENCE + HandelC.BLOCK_BEGIN;
    			this.procsInfo.getProcDefinition(name).apply(this);
    			//expression.apply(this);
    			this.hcCode = this.hcCode + HandelC.BLOCK_END;
    			this.hcCode = this.hcCode + HandelC.BLOCK_END;
    		} else {
    			//expression.apply(this);
    			this.procsInfo.getProcDefinition(name).apply(this);
    		}
    		this.hcCode = this.hcCode + HandelC.BLOCK_END;
    		
    	}
        if (this.hasLetDefinition){
        	for(PDefinitionList letProcess: this.definitionInLet){
        		this.hasLetDefinition = false;
        		letProcess.apply(this);
        	}
        	this.hasLetDefinition = false;
        	this.definitionInLet.clear();
        }
    	
    }
    
    public void caseACspConstantDefinition(ACspConstantDefinition node) {
    	// DO NOT REMOVE THIS CODE
    	// If you remove it, then the visitor will visit the 
    	// constant definition, instead of doing nothing 
    }

    public void caseACspFunctionDefinition(ACspFunctionDefinition node) {
    	// DO NOT REMOVE THIS CODE
    	// If you remove it, then the visitor will visit the 
    	// function definition, instead of doing nothing 
    }
    
    public void caseACspConstantCallExpr(ACspConstantCallExpr node){
    	//String localHcCode = "";
		String name = node.getName().getText();
		//System.out.println("ProcessesTranslatorVisitor :: caseACspConstantCallExpr -> Constant = "+name);
    	String type = node.getCspType().getName();
		//System.out.println("ProcessesTranslatorVisitor :: caseACspConstantCallExpr -> Type = "+type);
		
    	if (node.getCspType().toString().equals(CspUtil.getBoolType())) {
    		// Is a boolean constant
    		this.hcCode = this.hcCode + node.toString().trim();
    	} else {
    		if (this.inputVariables.contains(name)) {
                // It is an input variable
    			
    			// It may be an input variable of a bus
    			if (this.inputVarRenaming.containsKey(name)) {
    				// in this case, references should be replaced by the bus variable name
    				name = ((Stack<String>)this.inputVarRenaming.get(name)).peek();
    			}
    			
    			this.hcCode = this.hcCode + name;
            } else if (this.procsInfo.isArgumentOf(name,this.currentDefinition)) {
                // It is an argument of this process
                if (this.recursion.getTailRecursiveProcesses().contains(this.currentDefinition)
                    || this.recursion.getMutualRecursion()) {
                    // We are translating a tail recursive process or a mutual recursion
                    // Replace this argument by its copy
                	this.hcCode = this.hcCode + this.copyName(this.currentDefinition,name,IS_NOT_DECLARATION);
                } else {
                    // Keep original name
                	this.hcCode = this.hcCode + name;
                }
            } else if (this.types.isATypeBranch(name)) {
                // this is a reference to an element of a datatype
            	this.hcCode = this.hcCode + name;
            } else if (this.types.isTag(name)) {
                // this is a reference to a tag of a datatype
                
                String parent = this.types.getTagParent(name);
                String lookupTableName = HandelC.getInstance().getLookupTableName(parent,name);
                
                Integer tagSize = new Integer(this.types.getTagValues(name).size());
                
                this.tagsArgsLeftStack.push(tagSize);
                this.tagsArgsTotalStack.push(tagSize);
                this.tagsTypesStack.push((List)this.types.getTagTypes(name));
                
                this.hcCode = this.hcCode + lookupTableName;
                this.openArrayAccessTypeConstructor();
            } else if (this.constants.containsConstant(name)){
                // this is a reference to a global constant
            	this.hcCode = this.hcCode + name;
            }
    	}
    	
    	/*if (!this.isMultiSync){
    		this.hcCode = this.hcCode + localHcCode;
    	} else {
    		this.hcMultiSyncCode = this.hcMultiSyncCode + localHcCode;
    	}*/
    }

    public void caseACspIfProcess(ACspIfProcess node) {
        this.hcCode = this.hcCode + HandelC.IF + HandelC.PAR_BEGIN;

        this.setIsExpression(true);
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN + HandelC.BOOLEAN + HandelC.PAR_END + HandelC.PAR_BEGIN;
        node.getCondition().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.setIsExpression(false);

        this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.BLOCK_BEGIN + HandelC.SPACE;
        node.getThenProcess().apply(this);
        this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.ELSE + HandelC.BLOCK_BEGIN + HandelC.SPACE;
        node.getElseProcess().apply(this);
        this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.END_COMMAND;
    }

    public void caseACspGuardedProcess(ACspGuardedProcess node) {
        this.hcCode = this.hcCode + HandelC.IF + HandelC.PAR_BEGIN;
        this.setIsExpression(true);
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN + HandelC.BOOLEAN + HandelC.PAR_END;
        node.getGuardExpr().apply(this);
        this.setIsExpression(false);

        this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.BLOCK_BEGIN + HandelC.SPACE;
        node.getProcess().apply(this);
        this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.ELSE + HandelC.BLOCK_BEGIN + HandelC.SPACE;
        this.hcCode = this.hcCode + this.getStopImpl();
        this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.END_COMMAND;
    }
    
    public void caseACspPrefixProcess(ACspPrefixProcess node){
    	this.hcCode = this.hcCode + HandelC.SEQUENCE;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
        
        PEvent event = node.getEvent();
        LinkedList comm = ((ACspEvent)node.getEvent()).getComunication();
        LinkedList proj = ((ACspEvent)node.getEvent()).getProjection();
        PHelp proc = node.getProcess();

    	String inputVar = this.translatePrefix(event,comm,proj,IS_NOT_PRI_ALT);

        this.hcCode = this.hcCode + HandelC.END_COMMAND;
        proc.apply(this);
        this.hcCode = this.hcCode + HandelC.END_COMMAND;
        this.hcCode = this.hcCode + HandelC.BLOCK_END;

        if (inputVar != null) {
        	this.inputVariables.remove(inputVar);
        	this.removeInputVarRenaming(inputVar);        
        	this.removeInputVarType(inputVar);
        }
    }
    
    public void caseACspPrefixProcess_withRenaming(ACspPrefixProcess node){
    	this.hcCode = this.hcCode + HandelC.SEQUENCE;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
        
        PEvent event = node.getEvent();
        LinkedList comm = ((ACspEvent)node.getEvent()).getComunication();
        LinkedList proj = ((ACspEvent)node.getEvent()).getProjection();
        PHelp proc = node.getProcess();

        String channel = ((ACspEvent)event).getChannelName().toString().trim();
    	
        String inputVar;
        
        boolean isForcedInterleaving = false;
        for (String next: this.parallelismAnalyser.getForcedInterleavingLUT().keySet()){
        	if (channel.equals(HandelC.getInstance().getChannel(next)))
        		isForcedInterleaving = true;
        }
        
    	if (isForcedInterleaving && 
			this.parallelismAnalyser.getRealSyncLUT().containsKey(HandelC.CHAN_DECL + HandelC.UNDERSCORE + channel) &&
			this.parallelismAnalyser.getRealSyncLUT().get(HandelC.CHAN_DECL + HandelC.UNDERSCORE + channel).size() > 1) {
	    	
    		inputVar = this.translatePrefix(event,comm,proj,proc,IS_NOT_PRI_ALT);

    	} else {
	    	inputVar = this.translatePrefix(event,comm,proj,IS_NOT_PRI_ALT);

	    	this.hcCode = this.hcCode + HandelC.END_COMMAND;
	        proc.apply(this);
	        this.hcCode = this.hcCode + HandelC.END_COMMAND;
	        
	    }
	    
        this.hcCode = this.hcCode + HandelC.BLOCK_END;

        if (inputVar != null) {
        	this.inputVariables.remove(inputVar);
        	this.removeInputVarRenaming(inputVar);        
        	this.removeInputVarType(inputVar);
        }
    }

	private String translatePrefix(PEvent event, LinkedList comm,
			LinkedList proj, PHelp proc, boolean isPriAlt) {

        String inputVar = null;
    	
    	String channel = ((ACspEvent)event).getChannelName().toString().trim();

        this.setIsExpression(true);
    	
    	// First of all, check if this is a Bus
    	if (this.channels.getBuses().contains(channel)) {
    		// Output buses 
    		if (this.channels.getOutputBuses().contains(channel)){
        		this.hcCode = this.hcCode + HandelC.getInstance().getOutBusVarName(channel)
							+ HandelC.ASSIGNMENT; 
        		
        		ListIterator commIter = comm.listIterator(0);
	            Object next = commIter.next();
	            if (next instanceof ACspOutputComunication) {
	                ((ACspOutputComunication)next).getExpression().apply(this);
	            }
        		this.hcCode = this.hcCode + HandelC.END_COMMAND; 
    		// Input buses 
    		} else if (this.channels.getInputBuses().contains(channel)){
    			String inputVarName = HandelC.getInstance().getInBusVarName(channel);
    			String inputBusMethodName = HandelC.getInstance().getInBusMethodName(channel);
    			String inputBusArgName = HandelC.getInstance().getBusArgName(channel);
    			
        		this.hcCode = this.hcCode + inputVarName 
        					+ HandelC.ASSIGNMENT + inputBusMethodName + HandelC.DOT + inputBusArgName
        					+ HandelC.END_COMMAND; 

        		// Replace the following reference to the input variable to the bus variable
    			inputVar = this.checkInput(comm, channel);
    		}
    	// It is not a bus, but a normal CSP channel
    	} else {
	        this.currentPrefixing = channel;
	        ListIterator commIter = comm.listIterator(0);
	        ListIterator projIter = proj.listIterator(0);
	
	        String channel_name = HandelC.getInstance().getChannelName(this.currentPrefixing);
	        String action = this.parallelismAnalyser.getCommunicationAction(this.currentDefinition, 
	        																channel_name, 
	        																comm);
	        
	        this.currentPrefixingTypes = ListUtil.clone(this.channels.getChannelType(this.currentPrefixing));
	
	        /* 
	         * The Restrictions Visitor guarantess that we have communications
	         * of the form c[.exp]*[?name | !exp] 
	        */
	
	        if (!isPriAlt) {
	            inputVar = this.declareInputVar(comm, channel);
	        } 
	        
	        String backupHcCode = this.hcCode;
	        this.hcCode = "";
	        
	        // Translate the projections
	        while(projIter.hasNext()) {
	            Object next = projIter.next();
	            ((ACspProjection)next).apply(this);
	            if(!projIter.hasNext()){
	            	this.closeArrayAccessChannelSync();
	            }
	        }
	        
	        // Translate the communication
	        if (!commIter.hasNext()) {
	            if (this.directives.containsProcChannelInOut(this.currentDefinition)) {
	                Hashtable channelInOut = this.directives.getProcChannelInOut(this.currentDefinition);
	                if (channelInOut.containsKey(this.currentPrefixing)){
	                    String inOut = (String)channelInOut.get(this.currentPrefixing);
	                    if (inOut.equals(HandelC.A_READ)) {
	                        this.hcCode = this.hcCode + HandelC.INPUT_COMMUNICATION;
	                        this.hcCode = this.hcCode + HandelC.SYNC_IN;
	                    } else if (inOut.equals(HandelC.A_WRITE)) {
	                        this.hcCode = this.hcCode + HandelC.OUTPUT_COMMUNICATION;
	                        this.hcCode = this.hcCode + HandelC.SYNC_OUT;
	                    }
	                } else {
	                    throw new UndefinedChannelInOutException(this.currentPrefixing,this.currentDefinition);
	                }
	            } else {
	                throw new UndefinedChannelInOutException(this.currentPrefixing,this.currentDefinition);
	            }
	
	        } else {
	            Object next = commIter.next();
	            if (next instanceof ACspInputComunication) {
	                ((ACspInputComunication)next).apply(this);
	            } else if (next instanceof ACspOutputComunication) {
	                ((ACspOutputComunication)next).apply(this);
	            }
	        }
	        
	        String translationOfProjAndComm = this.hcCode;
	        
	        this.hcCode = "";
	        this.setIsExpression(false);
	        proc.apply(this);
	        this.setIsExpression(true);
	        
	        String translationOfProcess = this.hcCode;
	        
	        this.hcCode = backupHcCode;
	        
	        if (!isPriAlt) {
	        	Set<Set<Integer>> setOfFriendsInMultiSync = this.parallelismAnalyser.getRealSyncLUT().get(HandelC.CHAN_DECL + HandelC.UNDERSCORE + channel);
	        	int numberOfBranches = this.parallelismAnalyser.getNumberOfBranches();
	        	String prefixTranslation = "";
	        	for (int branch = 0; branch < numberOfBranches; branch++){
	        		String cases = "";
	        		for (Set<Integer>friends:setOfFriendsInMultiSync){
	        			if (friends.contains(branch)){
	        				cases = cases + "case " + channel + HandelC.UNDERSCORE + friends.hashCode() + translationOfProjAndComm;
	        				cases = cases + " : { delay; } break; ";
	        			}
	        		}
	        		if (!cases.equals("")){
	        			if (!prefixTranslation.equals("")){
	        				prefixTranslation = prefixTranslation + "else ";
	        			}
	        			prefixTranslation = prefixTranslation + "if (BRANCH_ID == "+ branch +") { prialt { ";
	        			prefixTranslation = prefixTranslation + cases;
	        			prefixTranslation = prefixTranslation + "} } "; 
	        		}
	        	}
	        	this.hcCode = this.hcCode + prefixTranslation;     	
	        	this.hcCode = this.hcCode + HandelC.END_COMMAND;
	        	if (!translationOfProcess.equals("")){
	        		this.hcCode = this.hcCode + translationOfProcess + HandelC.END_COMMAND;
	        	}
	        } else {
	        	this.hcCode = this.hcCode + this.currentPrefixing + translationOfProjAndComm;
	        }
	        this.currentType = "";
    	}
        this.setIsExpression(false);
        return inputVar;
	}

	public void caseACspProjection(ACspProjection node) {
        if (this.tagsArgsLeftStack.isEmpty()) {
            // this is a real projection
            this.nextArrayAccessChannelSync();
            node.getCallExpr().apply(this);
        } else {
            // this may be a tag access
            // Check how many arguments left for the current tag
            int argsLeft = ((Integer)this.tagsArgsLeftStack.pop()).intValue();
            int argsTotal = ((Integer)this.tagsArgsTotalStack.pop()).intValue();
            
        	// If it is zero, we should close the parenthesis
            // and remove the top of the stack of total number of arguments
            if (argsLeft == 0) {
                // For every finished argument of a datatype tag
                while(argsLeft == 0) {
                    // Close the array invocation
                    this.closeArrayAccessTypeConstructor();

                	// If there are still arguments left
                    if (!tagsArgsLeftStack.isEmpty()) {

                        // Let us check the next one
                        argsLeft = ((Integer)this.tagsArgsLeftStack.pop()).intValue();

                        // We have finished one tag constructor, remove it from the top of the stack
                        this.tagsArgsTotalStack.pop();
                    }
                }

                // Furthermore, If there are any further tags being analysed
                // and there are yet arguments left
                if (argsLeft > 0) {
                    // we have done one more argument from the previous tag
                    // Reduce the top of the stack of analysed arguments by one
                    this.tagsArgsLeftStack.push(new Integer(argsLeft-1));
                    this.tagsArgsTotalStack.push(new Integer(argsTotal));
                } else {
                    // Otherwise, if we finished the all the arguments, just remove the 
                    // last tag from the tags stack
                	this.tagsArgsLeftStack.pop();
                    this.tagsArgsTotalStack.pop();
                }
            } else {
                // If it is not zero, we do not need to close any parenthesis
                // But we have done one more argument from the current tag
                // Reduce the top of the stack of analysed arguments by one
                this.tagsArgsLeftStack.push(new Integer(argsLeft - 1));
                this.tagsArgsTotalStack.push(new Integer(argsTotal));
            }
            
            // If we are analysing tags
            // And there are still other arguments to be analysed
            if (argsLeft < argsTotal) {
                this.nextArrayAccessTypeConstructor();
            }

            // If we are not analysing any tags anymore
            // this is actually a real projection
            if (this.tagsArgsLeftStack.isEmpty()) {
                this.nextArrayAccessChannelSync();
            }
            
            // Analyse the remaining of the projection
            node.getCallExpr().apply(this);

            // After analysing the projection, let us check if we have finished
            // any tag analysis
            
            // If there are tags to be analysed yet
            if (!this.tagsArgsLeftStack.isEmpty()) {
                // For every tag whose analysys has finished
                while (!this.tagsArgsLeftStack.isEmpty() && ((Integer)this.tagsArgsLeftStack.peek()).intValue() == 0) {
                    // Remove the corresponding value from the stack of total 
                    // number of arguments, and close its parenthesis
                    this.tagsArgsLeftStack.pop();
                    this.tagsArgsTotalStack.pop();
                    
                    this.closeArrayAccessTypeConstructor();
                    
                }
            }
        }
    }
    
    
    public void caseACspInputComunication(ACspInputComunication node) {
    	if (!this.isMultiSync) {
    		this.hcCode = this.hcCode + HandelC.INPUT_COMMUNICATION;
    		this.hcCode = this.hcCode + HandelC.SPACE; 
    	}
        node.getParam().apply(this);
        this.directives.putProcChannelInOut(this.currentDefinition,this.currentPrefixing,HandelC.A_WRITE); 
    }
    
    public void caseACspOutputComunication(ACspOutputComunication node) {
    	if (!this.isMultiSync) {
	        this.hcCode = this.hcCode + HandelC.OUTPUT_COMMUNICATION;
	        this.hcCode = this.hcCode + HandelC.SPACE;
    	}
	    node.getExpression().apply(this);
        this.directives.putProcChannelInOut(this.currentDefinition,this.currentPrefixing,HandelC.A_WRITE); 
    }

    private String getStopImpl() {
    	return HandelC.INEXISTENT_CHANNEL
				+ HandelC.INPUT_COMMUNICATION
					+ HandelC.SYNC_IN + HandelC.END_COMMAND;    
	}
    
    public void caseACspProcessCallExpr(ACspProcessCallExpr node) {
        String name = node.getName().getText();
 
        this.setIsExpression(true);
        
        if (name.equals(CspUtil.getSkip())) {
    	// It is SKIP
    	} else if (name.equals(CspUtil.getStop())) {
    	// It is STOP
            this.hcCode = this.hcCode + this.getStopImpl();
    	} else if (!this.recursion.getMutualRecursion()) {
            // It is not mutual recursion, behave normally
    		//System.out.println("TranslatorVisitor::caseACspProcessCallExpr: name "+name+", currentDefinition "+this.currentDefinition+", tailRecursiveProcess "+this.recursion.getTailRecursiveProcesses());
            if(name.equals(this.currentDefinition) && 
                this.recursion.getTailRecursiveProcesses().contains(this.currentDefinition)) {
                // It is a tail recursion invocation
                // Save values of the copies
                this.translateRealArgumentsCopies(node,this.currentDefinition);
                this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_TRUE);
            } else {
            	String hcCodeBackup = this.hcCode;
            	this.hcCode = "";
                this.translateRealArguments(node,this.currentDefinition);
                this.hcCode = this.hcCode + HandelC.END_COMMAND;
        		this.hcCode = hcCodeBackup + name + this.hcCode;
            }
        } else {
            // It is a mutual recursion

            // Save values of the copies
            this.translateRealArgumentsCopies(node,this.currentDefinition);

            // Change the value of the program counter
            if (this.isTranslatingMain()) {
                // Translating the main method Invoke MUTUAL_REC
                this.hcCode = this.hcCode + HandelC.MUTUAL_RECURSION 
                			+ HandelC.PAR_BEGIN;

                if (this.useBranchIds) {
                	this.hcCode = this.hcCode 
                				+ HandelC.BRANCH_ID + HandelC.MATH_SUM
                				+ this.parallelismAnalyser.getNodeBranchId(node);
	            	this.hcCode = this.hcCode + HandelC.SEP_ARGUMENTS;
                }

                this.hcCode = this.hcCode + name + HandelC.PAR_END + HandelC.END_COMMAND;
            } else {
                this.hcCode = this.hcCode + HandelC.PROGRAM_COUNTER;
                this.hcCode = this.hcCode + HandelC.SPACE;
                this.hcCode = this.hcCode + HandelC.ASSIGNMENT + HandelC.SPACE;
                this.hcCode = this.hcCode + name + HandelC.END_COMMAND;

                // Indicate the change by setting KEEP_LOOPING to true
                this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_TRUE);
            }
        }
        
        this.setIsExpression(false);
    }
    
    public void caseACspDotExpr(ACspDotExpr node){
        node.getCallExpr().apply(this);
    	
    	List projections = node.getProjection();
    	for (Iterator iter = projections.iterator(); iter.hasNext();) {
			PProjection element = (PProjection) iter.next();
			element.apply(this);
		}
    }
    
    private String keepLooping(String value) {
        String answer = HandelC.KEEP_LOOPING + HandelC.SPACE;
        answer = answer + HandelC.ASSIGNMENT + HandelC.SPACE + value + HandelC.END_COMMAND;
        
        return answer;
    }

    public void caseACspSharingParallelProcess(ACspSharingParallelProcess node) {
        this.translateConcurrency(node);
    }
    
    public void caseACspAlphabetisedParallelProcess(ACspAlphabetisedParallelProcess node) {
        this.translateConcurrency(node);
    }
    
    public void caseACspInterleaveProcess(ACspInterleaveProcess node) {
        this.translateConcurrency(node);
    }
    
    public void caseACspInternalChoiceProcess(ACspInternalChoiceProcess node) {
		if (this.directives.getIntChoice().equals(Directives.COMPILE_TIME)) {
			node.getLeftProcess().apply(this);
		} else if (this.directives.getIntChoice().equals(Directives.RUN_TIME)) {
			this.hcCode = this.hcCode + HandelC.SEQ + HandelC.BLOCK_BEGIN
						+ this.updateRandomVar() + HandelC.END_COMMAND
						+ HandelC.IF + HandelC.PAR_BEGIN
						+ HandelC.PAR_BEGIN 
						+ HandelC.RANDOM_VAR + HandelC.MATH_REMAINDER + "2" + HandelC.PAR_END 
						+ HandelC.REL_EQUALS + HandelC.MATH_ZERO + HandelC.PAR_END + HandelC.BLOCK_BEGIN;
			node.getLeftProcess().apply(this);
			this.hcCode = this.hcCode + HandelC.BLOCK_END 
						+ HandelC.ELSE + HandelC.BLOCK_BEGIN; 
			node.getRightProcess().apply(this);
			this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.BLOCK_END; 
		}
    }
    
    private String updateRandomVar() {
    	return HandelC.RANDOM + HandelC.PAR_BEGIN + HandelC.RANDOM_VAR 
    		+ HandelC.PAR_END;
    }
    
    public void caseACspExternalChoiceProcess(ACspExternalChoiceProcess node) {
    	this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
    	
        InitialsExtractorVisitor initialsExtractor = new InitialsExtractorVisitor();
        node.apply(initialsExtractor);
        List keys = initialsExtractor.keyList();
       
        // Declare the input variables (if any)
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            LinkedList comm = initialsExtractor.getCommunication(event);
            this.currentPrefixing = ((ACspEvent)event).getChannelName().toString().trim();
            this.declareInputVar(comm, this.currentPrefixing);
        }

        List guardedList = initialsExtractor.getGuardedKeyList(); 
		if (!guardedList.isEmpty()){
			List notGuardedList = initialsExtractor.getNotGuardedKeyList();
			this.translateGuardedCases_NEW(node,initialsExtractor, guardedList, notGuardedList);
		} else {
			this.translateCases(
					node,
					initialsExtractor,						// InitialsExtractorVisitor initials
					initialsExtractor.keyList()				// List unblockedEvents
					);
		}
		
    	this.hcCode = this.hcCode + HandelC.BLOCK_END;
    	
    	for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            LinkedList comm = initialsExtractor.getCommunication(event);
            String channel = ((ACspEvent)event).getChannelName().toString().trim();
            
            String inputVar = this.extractVarName(comm, channel);
            
        	if (inputVar != null) {
            	this.inputVariables.remove(inputVar);
            	this.removeInputVarRenaming(inputVar);        
            	this.removeInputVarType(inputVar);
            }
        }
    	
    	
    }
    
    private void translateOutputInMSync(List<String> msChannels, List<String> inChannels, List<String> outChannels, InitialsExtractorVisitor initials){
    	//String code = "";
		
    	Hashtable <String, String> channelProj = new Hashtable<String, String>();
    	
    	this.hcCode = this.hcCode +
			"sMessageValue Local_M"
			+ (msChannels.size() > 1 ? "["+msChannels.size()+"]; par {" : ";");
		
		List eventList = initials.keyList();
		int countMS = 0;
		for(int i = 0; i < eventList.size(); i++){
			PEvent event = (PEvent) eventList.get(i);
			String channelName = ((ACspEvent) event).getChannelName().getText().trim();
			if (msChannels.contains(channelName)) {
				this.hcCode = this.hcCode 
	    			+ (msChannels.size() > 1 ? "{Local_M["+countMS+"]" : "Local_M")
	    			+".M = (integer_message)(0 @ "
	    			+ "MULTI_SYNC_ID_LUT[" + HandelC.getInstance().getChannelName(channelName) + "][BRANCH_ID]"
	    			+");"
	    			+ "to_" + channelName;
				
				int length = this.hcCode.length();
				
				this.currentPrefixing = channelName;
	    		this.currentPrefixingTypes = ListUtil.clone(this.channels.getChannelType(this.currentPrefixing));
	    		LinkedList proj = initials.getProjection(event);
	    		ListIterator projIter = proj.listIterator(0);
	    		// Translate the projections
	    		while(projIter.hasNext()) {
	    			Object next = projIter.next();
	    			((ACspProjection)next).apply(this);
	    			if(!projIter.hasNext()){
	    				this.closeArrayAccessChannelSync();
	    			}
	    		}
	    		this.currentType = "";
	    		
	    		channelProj.put(channelName, this.hcCode.substring(length));
	    		
				this.hcCode = this.hcCode
	    			+ "[MULTI_SYNC_ID_LUT[" + HandelC.getInstance().getChannelName(channelName) + "][BRANCH_ID]]" 
	    			+ "!Local_M"
	    			+ (msChannels.size() > 1 ? "["+countMS+"];};" : ";");
				countMS++;
			}
		}
		this.hcCode = this.hcCode +(msChannels.size() > 1 ? "};" : "");
		
		this.hcCode = this.hcCode +
		"{"
		+"boolean KEEP_LOOPING;"
		+"unsigned "+ MathUtil.getUnsignedBitsForNumber(msChannels.size()) +" channelsInSync;"
		+"sBoolValue boolean_value;"
		+"sMessageValue message_value;"
		+"boolean isMSyncChosen;"
		+"unsigned "
		+ MathUtil.getNumberOfBitsUnsigned(eventList.size())
		+" chosenId;"
		+ "sChannelOfMultiSync withdraw ["+msChannels.size()+"];"
		+"channelsInSync = "+msChannels.size()+";"
		+"KEEP_LOOPING = true;"
		+"while (KEEP_LOOPING){"
		+"KEEP_LOOPING = false;"
		+ "isMSyncChosen = false;"
		+ "prialt{";
	    
		countMS = 0;
		int countNotMS = 0;
		String withdraw = "";
		for(int i = 0; i < eventList.size(); i++){
			PEvent event = (PEvent) eventList.get(i);
			String channelName = ((ACspEvent) event).getChannelName().getText().trim();
			if (msChannels.contains(channelName)) {
				this.hcCode = this.hcCode +
					"case " 
					+ "from_" + channelName + channelProj.get(channelName) + "[MULTI_SYNC_ID_LUT[" + HandelC.getInstance().getChannelName(channelName) + "][BRANCH_ID]]"
					+ "?boolean_value :{"
					+ "isMSyncChosen = true;"
					+ "chosenId = " + countMS + HandelC.END_COMMAND
					+ HandelC.BLOCK_END
					+ HandelC.END_COMMAND
					+ "break;";
				
				withdraw = withdraw +
					"withdraw["+countMS+"].to_sync = &to_"+channelName + channelProj.get(channelName)+ "[0];" +
					"withdraw["+countMS+"].from_sync = &from_"+channelName + channelProj.get(channelName)+ "[0];" +
					"withdraw["+countMS+"].id = MULTI_SYNC_ID_LUT[chan_"+channelName + "][BRANCH_ID];";
				countMS++;
			} else {
				this.hcCode = this.hcCode + HandelC.CASE + HandelC.SPACE;
                LinkedList comm = initials.getCommunication((PEvent) event);
                LinkedList proj = initials.getProjection((PEvent) event);
                this.translatePrefix((PEvent) event,comm,proj,IS_PRI_ALT);
                this.hcCode = this.hcCode + HandelC.SPACE + HandelC.CASE_SEP + HandelC.BLOCK_BEGIN
					+ "chosenId = " + (msChannels.size() + countNotMS) + ";"
					+ HandelC.BLOCK_END
					+ HandelC.END_COMMAND
					+ "break;";
                countNotMS++;
			}
		}
		
		this.hcCode = this.hcCode + "}"
			+ "par {" +withdraw + "};";

		this.hcCode = this.hcCode
			+"if (isMSyncChosen == true) {"
				+"chan sMessageValue * to_sync;"
				+"chan sBoolValue * from_sync;"
				+"SYNC value;"
				+"sBoolValue * synchronized;"
				+"sMessageValue * myId;"
				+"MSYNC_ID BR_mSyncId;"
				+"MSYNC_ID BR_writer;"
				+"synchronized = &boolean_value;"
				+"myId = &message_value;";
				
		countMS = 0;
		
		for(int i = 0; i < eventList.size(); i++){
			PEvent event = (PEvent) eventList.get(i);
			String channelName = ((ACspEvent) event).getChannelName().getText().trim();
			if (msChannels.contains(channelName)) {
				this.hcCode = this.hcCode +
					"if (chosenId == " + countMS + ") {"
					+ "BR_mSyncId = "
					+ "MULTI_SYNC_ID_LUT[" + HandelC.getInstance().getChannelName(channelName) + "][BRANCH_ID]"
					//+ ");"
					+ ";"
					+ "to_sync = & to_" + channelName + channelProj.get(channelName) + "[0];"
					+ "from_sync = & from_" + channelName + channelProj.get(channelName) + "[0];"
					+ "BR_writer = MULTI_SYNC_WRITER;"
					+ "value = syncout;"
					+ "} ";
				countMS++;
			}
		}
		
		this.hcCode = this.hcCode		
			+"	(*myId).M = (integer_message)(0 @ BR_mSyncId);"
			+"	to_sync[BR_mSyncId]!(*myId);"
			+"	from_sync[BR_mSyncId]?(*synchronized);"
			+"	if ((*synchronized).B == true) {"
			+"		if (BR_mSyncId == BR_writer) {"
			+"			(*myId).MV = value;"
			+"			to_sync[BR_mSyncId]!(*myId);"
			+"		} else {"
			+"			sBoolValue * receivedValue;"
			+"			receivedValue = &boolean_value;"
			+"			from_sync[BR_mSyncId]?(*receivedValue);"
			+"			value = (*receivedValue).BV;"
			+"		}"
			+ (msChannels.size() > 1 ? "withdraw"+msChannels.size()+"AfterSync(withdraw, chosenId);":"")
			+"	} else{"
			+"		KEEP_LOOPING = true;"
			+"	};"
			+"} else {"
			+"	withdraw"+msChannels.size()+"AfterNotSync(withdraw);"
			+"};";
			
		countMS = 0;
		countNotMS = 0;
		for(int i = 0; i < eventList.size(); i++){
			PEvent event = (PEvent) eventList.get(i);
			String channelName = ((ACspEvent) event).getChannelName().getText().trim();
			if (msChannels.contains(channelName)){
				this.hcCode = this.hcCode +
					"if (chosenId == " + countMS + "){";
				countMS++;
			} else {
				this.hcCode = this.hcCode +
					"if (chosenId == " + (msChannels.size() + countNotMS) + "){";
				countNotMS++;
			}
			
			int lengthOfHcCode = this.hcCode.length();
			Node process = initials.getProcess(event);
			process.apply(this);
			if (lengthOfHcCode == this.hcCode.length()){
				this.hcCode = this.hcCode + HandelC.DELAY;
			} 
			this.hcCode = this.hcCode + HandelC.END_COMMAND + HandelC.BLOCK_END;
		}	
		this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.END_COMMAND + HandelC.BLOCK_END;	
    }
    
    private void translateExternalChoiceWithMultiSync(
			ACspExternalChoiceProcess node,
			InitialsExtractorVisitor initialsExtractor) {
    	
    	boolean choiceWithMultiSync = false;
    	Set <String> multiSyncChannels = this.parallelismAnalyser.getMultiSyncChannels();
    	List <String> syncInChoice = new Vector<String>();
		for (Object event:initialsExtractor.keyList()){
			String eventName = ((ACspEvent) event).getChannelName().toString().trim();
			if (multiSyncChannels.contains(eventName)) { 
				syncInChoice.add(eventName);
				choiceWithMultiSync = true;
			}
		}
		if (syncInChoice.isEmpty()){
			this.translateCases(
    				node,
    				initialsExtractor,
    				initialsExtractor.keyList()/*,
    				initialsExtractor.keyList().size(),
					"",
					"",
					new Vector(),
					DO_NOT_USE_DEFAULT,
					new Vector()*/);
		} else {
			
			List <String> msChannels = this.multisyncs.getMSChannels(currentDefinition, node.hashCode());
			List <String> inChannels = this.multisyncs.getInputChannels(currentDefinition, node.hashCode());
			List <String> outChannels = this.multisyncs.getOutputChannels(currentDefinition, node.hashCode());
			
			if (false) {
			//if (outChannels.size() > 0) {
				//translateOutputInMSync(msChannels, inChannels, outChannels, initialsExtractor);
			} else {
			
				String hcMSChannels = "";
	    		String hcInputChannels = "";
	    		String hcOutputChannels = "";
	    		
	    		int countSync = 0;
	    		int countInput = 0, countOutput = 0;
	    		
	    		List procMSEvent = new Vector();
				List procInputEvent = new Vector();
				List procOutputEvent = new Vector();
				
				Hashtable<Integer, String> varMSEvent = new Hashtable<Integer, String>();
				Hashtable<Integer, String> varIntrEvent = new Hashtable<Integer, String>();
	
				for (Object event:initialsExtractor.keyList()){
					String eventName = ((ACspEvent) event).getChannelName().toString().trim();
					LinkedList comm = initialsExtractor.getCommunication((PEvent)event);
					LinkedList proj = initialsExtractor.getProjection((PEvent)event);
					Node process = initialsExtractor.getProcess((PEvent) event);
					
					this.currentPrefixing = eventName;
					this.currentPrefixingTypes = ListUtil.clone(this.channels.getChannelType(this.currentPrefixing));
					String hcCodeBackup = this.hcCode;
					this.hcCode = "";
					ListIterator projIter = proj.listIterator(0);
					// Translate the projections
					while(projIter.hasNext()) {
						Object next = projIter.next();
						((ACspProjection)next).apply(this);
						if(!projIter.hasNext()){
							this.closeArrayAccessChannelSync();
						}
					}
					String channelProj = this.hcCode;
//					this.hcCode = hcCodeBackup;
					this.currentType = "";
					
					boolean isInput = false;
					this.hcCode = "";
		    		String commMS = "";
		    		
		    		ListIterator commIter = comm.listIterator(0);
		    		
		    		// Translate the communication
			        if (!commIter.hasNext()) {
			        	commMS = "0";
			        } else {
			        	this.isMultiSync = true;
			            Object next = commIter.next();
			            if (next instanceof ACspInputComunication) {
			            	isInput = true;
			                ((ACspInputComunication)next).apply(this);
			            } else if (next instanceof ACspOutputComunication) {
			                ((ACspOutputComunication)next).apply(this);
			            }
			            this.isMultiSync = false;
//			            commMS = this.hcMultiSyncCode;
			            commMS = this.hcCode;
//			            this.hcMultiSyncCode = "";
			        }
			        
			        this.hcCode = hcCodeBackup;
			        
					if (msChannels.contains(eventName)){
						String hcPrefix = "seqOfSync" + ((msChannels.size() > 1)?"["+countSync+"]":"");
						hcMSChannels = hcMSChannels + 
							hcPrefix + ".to_sync = &to_"+eventName + channelProj +"[0]; "+
							hcPrefix + ".from_sync = &from_"+eventName + channelProj +"[0]; "+
							hcPrefix + ".id = MULTI_SYNC_ID_LUT[chan_"+eventName+"][BRANCH_ID]; " +
							hcPrefix + ".channel = chan_"+eventName+"; ";
							//(!isInput ? hcPrefix + ".value = (MSYNC_VALUE) "+commMS+"; " : "");
							//hcPrefix + ".value = ";
						countSync++;
						if (!isInput) {
							hcMSChannels = hcMSChannels + hcPrefix + ".value = (MSYNC_VALUE) "+commMS+"; ";
						} else {
							varMSEvent.put(procMSEvent.size(), commMS);
						}
						
						procMSEvent.add(process);
					} else if (inChannels.contains(eventName)){
						String hcPrefix = "seqOfInputNotSync" + ((inChannels.size() > 1)?"["+countInput+"]":"");
						hcInputChannels = hcInputChannels + 
							hcPrefix + ".interrupt = &"+eventName + channelProj +"; ";
						countInput++;
						varIntrEvent.put(procInputEvent.size(), commMS);
						procInputEvent.add(process);
					} else if (outChannels.contains(eventName)){
						String hcPrefix = "seqOfOutputNotSync" + ((outChannels.size() > 1)?"["+countOutput+"]":"");
						hcOutputChannels = hcOutputChannels + 
							hcPrefix + ".interrupt = &"+eventName + channelProj +"; " +
							hcPrefix + ".value = (MSYNC_INTR_VALUE) "+ commMS +"; ";
						countOutput++;
						procOutputEvent.add(process);
					}
				}
	    		
				String hcDeclaration = "";
				hcDeclaration = hcDeclaration + 
			    				"sChannelOfMultiSync seqOfSync " + 
			    				((msChannels.size() > 1)?"["+msChannels.size()+"]":"") +
			    				"; ";
				if (inChannels.size() > 0) {
					hcDeclaration = hcDeclaration + 
				    				"sInterruptOfMultiSync seqOfInputNotSync " + 
				    				((inChannels.size() > 1)?"["+inChannels.size()+"]":"") +
				    				"; ";
				}
				if (outChannels.size() > 0) {
					hcDeclaration = hcDeclaration + 
									"sInterruptOfMultiSync seqOfOutputNotSync " + 
				    				((outChannels.size() > 1)?"["+outChannels.size()+"]":"") +
				    				"; ";
				}
				
				this.hcCode = this.hcCode + 
						hcDeclaration +
	    				"integer_MS_CLIENT idMultSync;"+
//	    	    		"integer_value value;"+
	    	    		"par {"+
		    	    		hcMSChannels +
		    	    		hcInputChannels +
		    	    		hcOutputChannels +
	    	    			//"value = 0;"+
	    	    		"}";
	
				this.hcCode = this.hcCode + "idMultSync = MultiSyncClient_"+msChannels.size()+"x"+inChannels.size()+"x"+outChannels.size()+"("+ 
						(msChannels.size()>1 ? "":"&") + "seqOfSync"+
						(inChannels.size() == 1 ? ", &seqOfInputNotSync":"") + (inChannels.size()>1 ? ", seqOfInputNotSync":"") +
						(outChannels.size() == 1 ? ", &seqOfOutputNotSync":"") + (outChannels.size()>1?", seqOfOutputNotSync":"")+
						");";
				
				int sizeProcMSEvent = procMSEvent.size();
				int sizeProcInputEvent = procInputEvent.size();
				procInputEvent.addAll(procOutputEvent);
				procMSEvent.addAll(procInputEvent);
				
				for (int i = 0; i < initialsExtractor.keyList().size(); i++){ // Can it be 'i < procMSEvent.size()'? Needs to be tested! 
					if (i != 0) this.hcCode = this.hcCode + "else ";
					
					this.hcCode = this.hcCode + "if (idMultSync == " + i + ") {";
					Node process = (Node) procMSEvent.get(i);
					if (process.toString().trim().equals(CspUtil.getSkip())){
						this.hcCode = this.hcCode + HandelC.DELAY;
					} else {
						if (varMSEvent.containsKey(i)){
							String var = varMSEvent.get(i);
							String varType = null;
							if (this.inputVarTypes.containsKey(var)) {
								varType = ((Stack<String>)this.inputVarTypes.get(var)).peek();
							}
							this.hcCode = this.hcCode + var + " = "
											+ (varType != null ? "(" + varType + ")": "") 
											+ "seqOfSync" + (sizeProcMSEvent>1?"["+i+"]":"") + ".value;";
						} else if (varIntrEvent.containsKey(i - sizeProcMSEvent)){
							String var = varIntrEvent.get(i - sizeProcMSEvent);
							String varType = null;
							if (this.inputVarTypes.containsKey(var)) {
								varType = ((Stack<String>)this.inputVarTypes.get(var)).peek();
							}
							this.hcCode = this.hcCode + var + " = "
											+ (varType != null ? "(" + varType + ")": "") 
											+ "seqOfInputNotSync" + (sizeProcInputEvent>1?"["+(i - sizeProcMSEvent)+"]":"") + ".value;";
						}
						process.apply(this);
					}
					this.hcCode = this.hcCode + HandelC.END_COMMAND;
					this.hcCode = this.hcCode + "} ";
				}
			}
		}
	}

	public void caseACspExternalChoiceProcess_withRenaming(ACspExternalChoiceProcess node) {
    	this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
    	
        InitialsExtractorVisitor initialsExtractor = new InitialsExtractorVisitor();
        node.apply(initialsExtractor);
        List keys = initialsExtractor.keyList();
       
        // Declare the input variables (if any)
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            LinkedList comm = initialsExtractor.getCommunication(event);
            this.currentPrefixing = ((ACspEvent)event).getChannelName().toString().trim();
            this.declareInputVar(comm, this.currentPrefixing);
        }
        
        this.translateCases_withRenaming(node, initialsExtractor);
        
    	this.hcCode = this.hcCode + HandelC.BLOCK_END;
    }

/*    private void translateExternalChoiceUsingSemaphors(Node node, InitialsExtractorVisitor initials) {
        // Declare the channel set that stores the events that happened
    	int nextId = MathUtil.getNextInt();
    	String doneEvents = HandelC.getInstance().getDoneEventsVarName(nextId);
    	String hasJoined = HandelC.getInstance().getHasJoinedVarName(nextId);

    	this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
    	
    	// Declare the local DONE_EVENTS_i
    	this.hcCode = this.hcCode + HandelC.getInstance().getSetVarOf(HandelC.CHANNEL)
					+ HandelC.SPACE + doneEvents + HandelC.SPACE + HandelC.END_COMMAND;

    	// Declare the local HAS_JOINED_i
    	this.hcCode = this.hcCode + HandelC.BOOLEAN
					+ HandelC.SPACE + hasJoined 
					+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(HandelC.CHANNEL)+ HandelC.ARRAY_END
					+ HandelC.SPACE + HandelC.END_COMMAND;

    	// Initialise the local DONE_EVENTS_i
    	this.hcCode = this.hcCode + doneEvents + HandelC.SPACE 
					+ HandelC. ASSIGNMENT + HandelC.SPACE 
					+ HandelC.getInstance().getEmptySetVarOf(HandelC.CHANNEL)
					+ HandelC.END_COMMAND;

    	List<String> allChannels = this.channels.getChannels();

    	// Initialise the local HAS_JOINED_i
    	for (int i=0; i< allChannels.size(); i++) {
    		String channel = allChannels.get(i);
        	this.hcCode = this.hcCode + this.getHasJoinedElement(hasJoined, channel) 
        				+ HandelC.SPACE 
						+ HandelC. ASSIGNMENT + HandelC.SPACE 
						+ HandelC.LOGICAL_FALSE
						+ HandelC.END_COMMAND;
    	}
    	
    	// List of channels taking part in the choice
    	List<String> choiceChannels = initials.getKeyNames();

    	List<String> semaphorsTaken = new Vector<String>();
    	List<String> semaphorsTried = new Vector<String>();
    	List<PEvent> unblockedEvents = new Vector<PEvent>();
    	
    	// For every channel name C that is initial
    	for (int i=0; i<choiceChannels.size(); i++) {
    		String channel = choiceChannels.get(i);
    		
    		List remainingChoices = choiceChannels.subList(i+1, choiceChannels.size());
    		System.out.println("ProcessesTranslatorVisitor -> translateExternalChoiceUsingSemaphors -> Channel = "+channel);
    		System.out.println("ProcessesTranslatorVisitor -> translateExternalChoiceUsingSemaphors -> Remaining = "+remainingChoices);
    		
            // while(IS_EMPTY_SET(DONE_EVENTS_0) && !HAS_JOINED_0[chan_c1]){
    		this.hcCode = this.hcCode + HandelC.WHILE + HandelC.PAR_BEGIN + HandelC.IS_EMPTY_SET
    				+ HandelC.PAR_BEGIN + doneEvents + HandelC.PAR_END + HandelC.SPACE
    				+ HandelC.LOGICAL_AND + HandelC.SPACE
    				+ HandelC.LOGICAL_NOT + this.getHasJoinedElement(hasJoined, channel)
    				+ HandelC.PAR_END + HandelC.BLOCK_BEGIN;

    		// if this is not the first channel in the choice
    		// then the previous channel is already unblocked
    		if (i > 0) {
        		PEvent event = (PEvent)initials.keyList().get(i-1);
    			unblockedEvents.add(event);
    		}

    		this.translateCases(node, 
    							initials,
    							ListUtil.clone(unblockedEvents),
    							i + 1,
    							doneEvents,
    							hasJoined,
    							ListUtil.clone(semaphorsTaken),
    							USE_DEFAULT,
    							ListUtil.clone(semaphorsTried));

    		// Try again
			semaphorsTried.clear();

			//HAS_JOINED_0[chan_c1] = TRY_JOIN_SEMAPHOR(BRANCH_ID,chan_c1);
			
    		String action = this.getChannelAction(initials, channel, this.currentDefinition);
    		boolean triedToJoin = this.insertHasJoinedUpdate(node, semaphorsTaken, semaphorsTried, hasJoined, channel, action);
    		if (triedToJoin) {
    			semaphorsTried.add(channel);	
    		}

    		this.hcCode = this.hcCode + HandelC.BLOCK_END;

    		// Now, you have taken this channels's semaphor
			semaphorsTaken.add(channel);
			
    	}

    	// After taking all channels semaphors(if needed), offer all of them
    	
		// IF(IS_EMPTY_SET(DONE_EVENTS_0)){
    	this.hcCode = this.hcCode + HandelC.IF + HandelC.PAR_BEGIN + HandelC.IS_EMPTY_SET
					+ HandelC.PAR_BEGIN + doneEvents + HandelC.PAR_END + HandelC.PAR_END + HandelC.BLOCK_BEGIN;
    	
    	// Add the last channel's semaphor which was taken in the previous loop
		unblockedEvents.add((PEvent)initials.keyList().get(choiceChannels.size()-1));
		this.translateCases(node, 
							initials,
							ListUtil.clone(unblockedEvents),
							choiceChannels.size(),
							doneEvents,
							hasJoined,
							ListUtil.clone(semaphorsTaken),
							DO_NOT_USE_DEFAULT, 
							ListUtil.clone(semaphorsTried));
		this.hcCode = this.hcCode + HandelC.BLOCK_END;

    	this.hcCode = this.hcCode + HandelC.BLOCK_END;
    }*/

    /*
     * Translates the cases of an external choice
     * @param initials The extractor containing the information of the initials taking part in the choice
     * @param unblockedEvents The list of events that may happen in the choice
     * 		  If no semaphor is being used, this is the list of keys from the initial extractor
     * @param checked The number of events that we have checked if it is a forced event or not
     * 		  If no semaphor is being used, this is the the size of the list of keys from the initial extractor
     * @param doneEvents Name of the variable that is the set of channels that happened
     * 			this variable controls the loop of delay
     * 		  If no semaphor is being used, this is not used
     * @param hasJoined Name of the variable that indicates for each channels if it has joined
     * 			the semaphor
     * 		  If no semaphor is being used, this is not used
     * @param semaphorsTaken List of semaphors that have been taken; it is used to release the
     * 			semaphors when some event happens
     * 		  If no semaphor is being used, this is always the empty list
     * 		  Otherwise it is only initially empty
     */
    private void translateCases_withRenaming(Node node,
			InitialsExtractorVisitor initials) {
    	
    	Hashtable<String,Set<Set<Integer>>> realSyncLUT = this.parallelismAnalyser.getRealSyncLUT();
    	//ForcedInterleavingOnChannel :: Hashtable<String, Branch2BranchSetLUT>
    	ForcedInterleavingOnChannel forcedInterleavingLUT = this.parallelismAnalyser.getForcedInterleavingLUT();
    	
    	List<String> channelsWithRenaming = new Vector <String>();
    	ListIterator <String> channelNames = initials.getKeyNames().listIterator();
    	while (channelNames.hasNext()){
    		String nextChannel = channelNames.next();
    		if (forcedInterleavingLUT.containsKey(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel) &&
    				realSyncLUT.containsKey(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel) &&
    				realSyncLUT.get(HandelC.CHAN_DECL + HandelC.UNDERSCORE + nextChannel).size() > 1){
    			channelsWithRenaming.add(nextChannel);
    		}
    	}
    	
    	int numberOfBranches = this.parallelismAnalyser.getNumberOfBranches();
		String branchesTranslation [] = new String [numberOfBranches];
		for (int branch = 0; branch < numberOfBranches; branch++){
			branchesTranslation[branch] = "";
		}
    		
		ListIterator events = initials.keyList().listIterator();
    	while (events.hasNext()){
    		PEvent event = (PEvent)events.next();
    		LinkedList comm = initials.getCommunication(event);
            LinkedList proj = initials.getProjection(event);
            
            // Translate prefix
            this.setIsExpression(true);
        	
            String channel = ((ACspEvent)event).getChannelName().toString().trim();

            this.currentPrefixing = channel;
	        ListIterator commIter = comm.listIterator(0);
	        ListIterator projIter = proj.listIterator(0);
	
	        String channel_name = HandelC.getInstance().getChannelName(this.currentPrefixing);
	        String action = this.parallelismAnalyser.getCommunicationAction(this.currentDefinition, 
	        																channel_name, 
	        																comm);
	        
	        this.currentPrefixingTypes = ListUtil.clone(this.channels.getChannelType(this.currentPrefixing));
	
	        /* 
	         * The Restrictions Visitor guarantess that we have communications
	         * of the form c[.exp]*[?name | !exp] 
	        */
	
	        String backupHcCode = this.hcCode;
	        this.hcCode = "";
	        
	        // Translate the projections
	        while(projIter.hasNext()) {
	            Object next = projIter.next();
	            ((ACspProjection)next).apply(this);
	            if(!projIter.hasNext()){
	            	this.closeArrayAccessChannelSync();
	            }
	        }
	        
	        // Translate the communication
	        if (!commIter.hasNext()) {
	            if (this.directives.containsProcChannelInOut(this.currentDefinition)) {
	                Hashtable channelInOut = this.directives.getProcChannelInOut(this.currentDefinition);
	                if (channelInOut.containsKey(this.currentPrefixing)){
	                    String inOut = (String)channelInOut.get(this.currentPrefixing);
	                    if (inOut.equals(HandelC.A_READ)) {
	                        this.hcCode = this.hcCode + HandelC.INPUT_COMMUNICATION;
	                        this.hcCode = this.hcCode + HandelC.SYNC_IN;
	                    } else if (inOut.equals(HandelC.A_WRITE)) {
	                        this.hcCode = this.hcCode + HandelC.OUTPUT_COMMUNICATION;
	                        this.hcCode = this.hcCode + HandelC.SYNC_OUT;
	                    }
	                } else {
	                    throw new UndefinedChannelInOutException(this.currentPrefixing,this.currentDefinition);
	                }
	            } else {
	                throw new UndefinedChannelInOutException(this.currentPrefixing,this.currentDefinition);
	            }
	
	        } else {
	            Object next = commIter.next();
	            if (next instanceof ACspInputComunication) {
	                ((ACspInputComunication)next).apply(this);
	            } else if (next instanceof ACspOutputComunication) {
	                ((ACspOutputComunication)next).apply(this);
	            }
	        }
	        
	        String translationOfProjAndComm = this.hcCode;
	        
	        this.hcCode = "";
	        this.setIsExpression(false);
	        Node process = initials.getProcess(event);
            process.apply(this);
            this.setIsExpression(true);
	        
	        String translationOfProcess = this.hcCode;
	        
	        this.hcCode = backupHcCode;
	
	        
	        if (channelsWithRenaming.contains(channel)){
	        	Set<Set<Integer>> setOfFriendsInMultiSync = realSyncLUT.get(HandelC.CHAN_DECL + HandelC.UNDERSCORE + channel);
	        	
	        	//String choiceTranslation = "";
	        	for (int branch = 0; branch < numberOfBranches; branch++){
	        		//String cases = "";
	        		for (Set<Integer>friends:setOfFriendsInMultiSync){
	        			if (friends.contains(branch)){
	        				branchesTranslation[branch] = branchesTranslation[branch] + "case " + channel + HandelC.UNDERSCORE + friends.hashCode() + translationOfProjAndComm;
	        				if (translationOfProcess.equals("")){
	        					branchesTranslation[branch] = branchesTranslation[branch] + " : { delay; } break; ";
	        				} else {
	        					branchesTranslation[branch] = branchesTranslation[branch] + " : { " + translationOfProcess + "; } break; ";
	        				}
	        			}
	        		}
	        	}
	        } else {
	        	for (int branch = 0; branch < numberOfBranches; branch++){
	        		branchesTranslation[branch] = branchesTranslation[branch] + "case " + channel + translationOfProjAndComm;
	        		if (translationOfProcess.equals("")){
    					branchesTranslation[branch] = branchesTranslation[branch] + " : { delay; } break; ";
    				} else {
    					branchesTranslation[branch] = branchesTranslation[branch] + " : { " + translationOfProcess + "; } break; ";
    				}
	        	}
	        }
    	
    	}    
	    String choiceTranslation = "";
    	for (int branch = 0; branch < numberOfBranches; branch++){
    		if (!choiceTranslation.equals("")){
    			choiceTranslation = choiceTranslation + "else ";
			}
    		choiceTranslation = choiceTranslation + "if (BRANCH_ID == "+ branch +") { prialt { ";
    		choiceTranslation = choiceTranslation + branchesTranslation[branch];
    		choiceTranslation = choiceTranslation + "} } ";
    	}
    	
    	this.hcCode = this.hcCode + choiceTranslation;
    }
    
    private void translateGuardedCases_NEW(Node node, InitialsExtractorVisitor initials, List<PEvent> blockedEvents, List<PEvent> unblockedEvents){
    	List <List> allCases_blockedEvents = ListUtil.powerSet_orderedByCard(blockedEvents);
    	
    	boolean firstIF = true;
    	for (List<PEvent> blockedList : allCases_blockedEvents){
    		if (!firstIF){
    			this.hcCode = this.hcCode + HandelC.ELSE + HandelC.SPACE;
    		}
    		if (!blockedList.isEmpty()) {
    			this.hcCode = this.hcCode + HandelC.IF + HandelC.PAR_BEGIN;
    			ListIterator<PEvent> blockedIter = blockedList.listIterator();
    			while(blockedIter.hasNext()){
    				this.hcCode = this.hcCode + HandelC.LOGICAL_NOT;
    				this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
//    				this.hcCode = this.hcCode + initials.getGuard(blockedIter.next());
    				initials.getGuard(blockedIter.next()).apply(this);
    				this.hcCode = this.hcCode + HandelC.PAR_END;
    				if (blockedIter.hasNext()){
    					this.hcCode = this.hcCode + HandelC.SPACE + HandelC.LOGICAL_AND  + HandelC.SPACE;
    				}
    			}
    			this.hcCode = this.hcCode + HandelC.PAR_END;
    		}
		    this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
		    List <PEvent> newUnblockedEvents = new Vector<PEvent>();
		    for (PEvent event:blockedEvents){
		    	if (!blockedList.contains(event)){
		    		newUnblockedEvents.add(event);
		    	}
		    }
		    newUnblockedEvents.addAll(unblockedEvents);
		    if (!newUnblockedEvents.isEmpty()){
			    this.translateCases(
						node,									// No usa.
						initials,								// InitialsExtractorVisitor initials
						newUnblockedEvents//,					// List unblockedEvents
						//initials.getKeyNames().size(),		// int checked
						//"",									// String doneEvents
						//"",									// String hasJoined
						//new Vector(),							// List<String> semaphorsTaken
						//DO_NOT_USE_DEFAULT,					// boolean useDefault
						//new Vector()							// List<String> semaphorsTried
			    		);
		    } else {
		    	this.hcCode = this.hcCode + this.getStopImpl();
		    }
		    this.hcCode = this.hcCode + HandelC.BLOCK_END;
		    
		    firstIF = false;
    	}
    }
    
    private void translateGuardedCases(InitialsExtractorVisitor initials, List<PEvent> blockedEvents, List<PEvent> unblockedEvents, List<String> guardsExp, boolean isFirst){
    	//String code = "";
    	if (!blockedEvents.isEmpty()){
    		//for (int index = 0; index < blockedEvents.size(); index ++){//PEvent event:blockedEvents){
    			// Traduzir o equivalente ao valor da guarda negado
    			PEvent event = (PEvent) ListUtil.head(blockedEvents);
    			String guard = ((PExpr)initials.getGuard(event)).toString().trim();
    			List<String> guards = (ListUtil.clone(guardsExp));
    			guards.add(guard);
    			List<PEvent> blockeds = ListUtil.clone(blockedEvents);
    			blockeds.remove(event);
    			List<PEvent> unblockeds = ListUtil.clone(unblockedEvents);
    			translateGuardedCases(	initials, blockeds,  unblockeds, guards, isFirst);
    			
    			// Traduzir o equivalente ao valor da guarda positivo
    			blockeds = ListUtil.clone(blockedEvents);
    			event = (PEvent) ListUtil.head(blockeds);
    			blockeds.remove(event);
    			unblockeds = ListUtil.clone(unblockedEvents);
    			unblockeds.add(event);
    			translateGuardedCases(	initials, blockeds,  unblockeds, (ListUtil.clone(guardsExp)), false);
    		//}
    	} else{
    		if (!isFirst) {
    			this.hcCode = this.hcCode + HandelC.ELSE + HandelC.SPACE;
    		}
    		if (!guardsExp.isEmpty()) {
    			this.hcCode = this.hcCode + HandelC.IF + HandelC.PAR_BEGIN;
    			ListIterator<String> guardsIter = guardsExp.listIterator();
    			while(guardsIter.hasNext()){
    				this.hcCode = this.hcCode + HandelC.LOGICAL_NOT;
    				this.hcCode = this.hcCode + HandelC.PAR_BEGIN + guardsIter.next() + HandelC.PAR_END;
    				if (guardsIter.hasNext()){
    					this.hcCode = this.hcCode + HandelC.SPACE + HandelC.LOGICAL_AND  + HandelC.SPACE;
    				}
    			}
    			this.hcCode = this.hcCode + HandelC.PAR_END;
    		}
		    this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
		    if (!unblockedEvents.isEmpty()){
			    this.translateCases(
						null,									// No usa.
						initials,								// InitialsExtractorVisitor initials
						unblockedEvents//,						// List unblockedEvents
						//initials.getKeyNames().size(),		// int checked
						//"",									// String doneEvents
						//"",									// String hasJoined
						//new Vector(),							// List<String> semaphorsTaken
						//DO_NOT_USE_DEFAULT,					// boolean useDefault
						//new Vector()							// List<String> semaphorsTried
			    		);
		    } else {
		    	this.hcCode = this.hcCode + this.getStopImpl();
		    }
		    this.hcCode = this.hcCode + HandelC.BLOCK_END;
    	}
    	
    	
    }
    
    private void translateCases(Node node,
    							InitialsExtractorVisitor initials, 
    							List unblockedEvents) {

    	int unblockedMultiSyncCard = 0;
    	int unblockedInterruptInCard = 0;
    	int unblockedInterruptOutCard = 0;
    	//Set <String> multiSyncChannels = this.parallelismAnalyser.getMultiSyncChannels();
    	List <String> msChannels = this.multisyncs.getMSChannels(currentDefinition, node.hashCode());
    	List <String> inChannels = this.multisyncs.getInputChannels(currentDefinition, node.hashCode());
    	List <String> outChannels = this.multisyncs.getOutputChannels(currentDefinition, node.hashCode());
    	for (Object event:unblockedEvents){								// unblockedEvents precisa ser uma lista de PEvent
			String eventName = ((ACspEvent) event).getChannelName().toString().trim();
			if (msChannels.contains(eventName)) { 
				unblockedMultiSyncCard ++;
			} else if (inChannels.contains(eventName)) { 
				unblockedInterruptInCard ++;
			} else if (outChannels.contains(eventName)) { 
				unblockedInterruptOutCard ++;
			}
		}
    	
    	if (unblockedMultiSyncCard > 0) {
    		
			String hcMSChannels = "";
    		String hcInputChannels = "";
    		String hcOutputChannels = "";
    		
    		int countSync = 0;
    		int countInput = 0, countOutput = 0;
    		
    		List procMSEvent = new Vector();
			List procInputEvent = new Vector();
			List procOutputEvent = new Vector();
			
			Hashtable<Integer, String> varMSEvent = new Hashtable<Integer, String>();
			Hashtable<Integer, String> varIntrEvent = new Hashtable<Integer, String>();

			for (Object event:unblockedEvents){						// unblockedEvents precisa ser uma lista de PEvent
				String eventName = ((ACspEvent) event).getChannelName().toString().trim();
				LinkedList comm = initials.getCommunication((PEvent)event);
				LinkedList proj = initials.getProjection((PEvent)event);
				Node process = initials.getProcess((PEvent) event);
				
				this.currentPrefixing = eventName;
				this.currentPrefixingTypes = ListUtil.clone(this.channels.getChannelType(this.currentPrefixing));
				String hcCodeBackup = this.hcCode;
				this.hcCode = "";
				ListIterator projIter = proj.listIterator(0);
				// Translate the projections
				while(projIter.hasNext()) {
					Object next = projIter.next();
					((ACspProjection)next).apply(this);
					if(!projIter.hasNext()){
						this.closeArrayAccessChannelSync();
					}
				}
				String channelProj = this.hcCode;
//				this.hcCode = hcCodeBackup;
				this.currentType = "";
				
				boolean isInput = false;
				this.hcCode = "";
	    		String commMS = "";
	    		
	    		ListIterator commIter = comm.listIterator(0);
	    		
	    		// Translate the communication
		        if (!commIter.hasNext()) {
		        	commMS = "0";
		        } else {
		        	this.isMultiSync = true;
		            Object next = commIter.next();
		            if (next instanceof ACspInputComunication) {
		            	isInput = true;
		                ((ACspInputComunication)next).apply(this);
		            } else if (next instanceof ACspOutputComunication) {
		                ((ACspOutputComunication)next).apply(this);
		            }
		            this.isMultiSync = false;
//		            commMS = this.hcMultiSyncCode;
		            commMS = this.hcCode;
//		            this.hcMultiSyncCode = "";
		        }
		        
		        this.hcCode = hcCodeBackup;
		        
				if (msChannels.contains(eventName)){
					String hcPrefix = "seqOfSync" + ((unblockedMultiSyncCard > 1)?"["+countSync+"]":"");
					hcMSChannels = hcMSChannels + 
						hcPrefix + ".to_sync = &to_"+eventName + channelProj +"[0]; "+
						hcPrefix + ".from_sync = &from_"+eventName + channelProj +"[0]; "+
						hcPrefix + ".id = MULTI_SYNC_ID_LUT[chan_"+eventName+"][BRANCH_ID]; " +
						hcPrefix + ".channel = chan_"+eventName+"; ";
						//(!isInput ? hcPrefix + ".value = (MSYNC_VALUE) "+commMS+"; " : "");
						//hcPrefix + ".value = ";
					countSync++;
					if (!isInput) {
						hcMSChannels = hcMSChannels + hcPrefix + ".value = (MSYNC_VALUE) "+commMS+"; ";
					} else {
						varMSEvent.put(procMSEvent.size(), commMS);
					}
					
					procMSEvent.add(process);
				} else if (inChannels.contains(eventName)){
					String hcPrefix = "seqOfInputNotSync" + ((unblockedInterruptInCard > 1)?"["+countInput+"]":"");
					hcInputChannels = hcInputChannels + 
						hcPrefix + ".interrupt = &"+eventName + channelProj +"; ";
					countInput++;
					varIntrEvent.put(procInputEvent.size(), commMS);
					procInputEvent.add(process);
				} else if (outChannels.contains(eventName)){
					String hcPrefix = "seqOfOutputNotSync" + ((unblockedInterruptOutCard > 1)?"["+countOutput+"]":"");
					hcOutputChannels = hcOutputChannels + 
						hcPrefix + ".interrupt = &"+eventName + channelProj +"; " +
						hcPrefix + ".value = (MSYNC_INTR_VALUE) "+ commMS +"; ";
					countOutput++;
					procOutputEvent.add(process);
				}
			}
    		
			String hcDeclaration = "";
			hcDeclaration = hcDeclaration + 
		    				"sChannelOfMultiSync seqOfSync " + 
		    				((unblockedMultiSyncCard > 1)?"["+unblockedMultiSyncCard+"]":"") +
		    				"; ";
			if (inChannels.size() > 0) {
				hcDeclaration = hcDeclaration + 
			    				"sInterruptOfMultiSync seqOfInputNotSync " + 
			    				((unblockedInterruptInCard > 1)?"["+unblockedInterruptInCard+"]":"") +
			    				"; ";
			}
			if (outChannels.size() > 0) {
				hcDeclaration = hcDeclaration + 
								"sInterruptOfMultiSync seqOfOutputNotSync " + 
			    				((unblockedInterruptOutCard > 1)?"["+unblockedInterruptOutCard+"]":"") +
			    				"; ";
			}
			
			this.hcCode = this.hcCode + 
					hcDeclaration +
    				"integer_MS_CLIENT idMultSync;"+
//    	    	    		"integer_value value;"+
    	    		"par {"+
	    	    		hcMSChannels +
	    	    		hcInputChannels +
	    	    		hcOutputChannels +
    	    			//"value = 0;"+
    	    		"}";

			this.hcCode = this.hcCode + "idMultSync = MultiSyncClient_"+unblockedMultiSyncCard+"x"+unblockedInterruptInCard+"x"+unblockedInterruptOutCard+"("+ 
					(unblockedMultiSyncCard>1 ? "":"&") + "seqOfSync"+
					(unblockedInterruptInCard == 1 ? ", &seqOfInputNotSync":"") + (unblockedInterruptInCard>1 ? ", seqOfInputNotSync":"") +
					(unblockedInterruptOutCard == 1 ? ", &seqOfOutputNotSync":"") + (unblockedInterruptOutCard>1?", seqOfOutputNotSync":"")+
					");";
			
			int sizeProcMSEvent = procMSEvent.size();
			int sizeProcInputEvent = procInputEvent.size();
			procInputEvent.addAll(procOutputEvent);
			procMSEvent.addAll(procInputEvent);
			
			for (int i = 0; i < unblockedEvents.size(); i++){  
				if (i != 0) this.hcCode = this.hcCode + "else ";
				
				this.hcCode = this.hcCode + "if (idMultSync == " + i + ") {";
				Node process = (Node) procMSEvent.get(i);
				if (process.toString().trim().equals(CspUtil.getSkip())){
					this.hcCode = this.hcCode + HandelC.DELAY;
				} else {
					if (varMSEvent.containsKey(i)){
						String var = varMSEvent.get(i);
						if (!var.equals("0")){
							String varType = null;
							if (this.inputVarTypes.containsKey(var)) {
								varType = ((Stack<String>)this.inputVarTypes.get(var)).peek();
							}
							this.hcCode = this.hcCode + var + " = "
							+ (varType != null ? "(" + varType + ")": "") 
							+ "seqOfSync" + (sizeProcMSEvent>1?"["+i+"]":"") + ".value;";
						}
					} else if (varIntrEvent.containsKey(i - sizeProcMSEvent)){					//AddIvan: Modificado - Verificar 16.03 
						String var = varIntrEvent.get(i - sizeProcMSEvent);
						if (!var.equals("0")){
							String varType = null;
							if (this.inputVarTypes.containsKey(var)) {
								varType = ((Stack<String>)this.inputVarTypes.get(var)).peek();
							}
							this.hcCode = this.hcCode + var + " = "
											+ (varType != null ? "(" + varType + ")": "") 
											+ "seqOfInputNotSync" + (sizeProcInputEvent>1?"["+(i - sizeProcMSEvent)+"]":"") + ".value;";
						}
					}
					process.apply(this);
				}
				this.hcCode = this.hcCode + HandelC.END_COMMAND;
				this.hcCode = this.hcCode + "} ";
			}
    		
    	} else {
            // Translate the non-multi-sync cases
            this.hcCode = this.hcCode + HandelC.ALTERNATION;
            this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
            
            // Declare the cases
            for(int i=0; i < unblockedEvents.size(); i++) {
                this.hcCode = this.hcCode + HandelC.CASE + HandelC.SPACE;
                PEvent event = (PEvent)unblockedEvents.get(i);
                LinkedList comm = initials.getCommunication(event);
                LinkedList proj = initials.getProjection(event);
                
                this.translatePrefix(event,comm,proj,IS_PRI_ALT);

                this.hcCode = this.hcCode + HandelC.SPACE + HandelC.CASE_SEP + HandelC.BLOCK_BEGIN;
                Node process = initials.getProcess(event);
                if (!process.toString().trim().equals(CspUtil.getSkip()))
                	process.apply(this);
                else
                	this.hcCode = this.hcCode + HandelC.DELAY + HandelC.END_COMMAND;
                this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.END_COMMAND + HandelC.BREAK + HandelC.END_COMMAND;
            }
            
            this.hcCode = this.hcCode + HandelC.BLOCK_END;
    	}
    }

    public void caseACspSequentialProcess(ACspSequentialProcess node){
        
        this.hcCode = this.hcCode + HandelC.SEQUENCE;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;

        node.getLeftProcess().apply(this);
        this.hcCode = this.hcCode + HandelC.END_COMMAND;
        node.getRightProcess().apply(this);

        this.hcCode = this.hcCode + HandelC.BLOCK_END;
    }
    
    private String getArgumentDeclaration(PParam parameter, String process){
    	String argument = parameter.toString().trim();

    	String type = this.getArgumentType(parameter, process);
        String code = type + HandelC.SPACE + argument;
        return code;
    }

    private String getArgumentType(PParam parameter, String process){
    	String argument = parameter.toString().trim();
    	
    	// Trying to inference the type of the parameter
    	MaximalTypesInferenceVisitor inference = new MaximalTypesInferenceVisitor();
    	parameter.apply(inference);
    	CspType csptype = inference.getType();
    	
        String type = "";
        if ((csptype instanceof CspIntType)
        	|| (csptype instanceof CspBoolType)
        	|| (csptype instanceof CspDataType)) {
        	type = HandelC.getInstance().getHandelCTypeName(csptype.toString());
    	} else if (this.directives.containsProcArguments(process) 
	            && this.directives.getProcArguments(process).containsKey(argument)) {
            type = (String)this.directives.getProcArguments(process).get(argument);
        } else {
            if (csptype instanceof CspSetType) {
            	throw new UndefinedArgumentSetTypeException(argument,process);
            } else {
            	throw new UndefinedArgumentTypeException(argument,process);
            	
            }
        }
        return type;
    }

    // if declaration is true it declares the variable, otherwise it initialises it
    // if argument is true, it declares it as a function argument, otherwise
    // it declares it as a local variable
    private String getArgumentCopyDeclaration(PParam parameter, String process, 
                                        boolean declaration, boolean isArgument){
    	
    	String argument = parameter.toString().trim();

    	String type = this.getArgumentType(parameter, process);
    	
        String code = "";
        String copy = this.copyName(process,argument,declaration);
        if (declaration) {
            // Declare copy
            String decl = type + HandelC.SPACE + copy;
            if (isArgument) {
                // As a function argument
                code = code + HandelC.SEP_ARGUMENTS + HandelC.SPACE + decl;
            } else {
                // As a local variable
                code = code + decl + HandelC.END_COMMAND;
            }
        } else {
            // Initialise Copy
            code = code + copy + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE;
            code = code + argument + HandelC.END_COMMAND;
        }
        return code;
    }
    
    // Return the name of the copy of the local variable together with any
    // possible array access
    // Receives the process name P, the variable name N, and a boolean that indicates
    // if this is a declaration or not
    // Returns the name P_local_N[
    private String copyName(String procName, String varName, boolean declaration) {
        String copyName = procName + HandelC.LOCAL_COPY + varName;
        return copyName;
    }
    
    private void declareChannelsDatatypes(){
    	// Declaring the channels names chan_c
        List<String> names = this.channels.getChannelNames();
        
        DatatypesVisitor typesVisitor = new DatatypesVisitor(directives,this.constants);
        typesVisitor.declareType(HandelC.CHANNEL, names);
        this.hcCode = typesVisitor.getHcCode() + this.hcCode;
        
    }
    
    private void declareChannels(){
  
    	// Declaring the channel variables
    	Iterator keys = this.channels.getChannels().iterator();
        while(keys.hasNext()) {
            String chan = (String)keys.next();
            if (!this.parallelismAnalyser.isMultiSyncChannel(chan)){
            	String decl = this.channels.getChannelDeclaration(chan);
            	this.hcCode = decl + HandelC.END_COMMAND + this.hcCode;
            } else {
            	this.hcCode = this.multisyncs.getChannelDeclaration(chan) + this.hcCode;
            }
        }
        
        // Declaring the channel that finishes the multi-sync system
        if (this.parallelismAnalyser.isMultiSync()){
        	String endSystem = HandelC.CHAN_DECL + HandelC.SPACE 
				        	+ HandelC.SYNC_TYPE + HandelC.SPACE 
				        	+ "endSystem" + HandelC.END_COMMAND;
        	this.hcCode = endSystem + this.hcCode;
        }
        
        // Declaring the inexistent channel
        String inexistent = HandelC.CHAN_DECL + HandelC.SPACE 
        				+ HandelC.SYNC_TYPE + HandelC.SPACE 
        				+ HandelC.INEXISTENT_CHANNEL + HandelC.END_COMMAND;
        this.hcCode = inexistent + this.hcCode;
    	        
        this.hcCode = HandelC.getInstance().generateComment(Messenger.CHANNELS_DECL) + this.hcCode;
        
    }
    
    private void declareBuses(){
    	// Declaring buses
        MathSet buses = this.channels.getBuses();
        
        if (!buses.isEmpty()) {
	        String external_interface = "";
	        String simulate_interface = "";
	
	        this.hcCode = " #endif " + this.hcCode;
	        for (Object bus : buses) {
	            String decl_var = this.channels.getBusVarDeclaration((String)bus);
	            String decl_interface = this.channels.getBusInterfaceDeclaration((String)bus);
	            external_interface = decl_var + HandelC.END_COMMAND 
	            			+ decl_interface + HandelC.END_COMMAND + external_interface;
	            
	            simulate_interface =
	            	this.channels.getBusSimulationDeclaration((String)bus) + HandelC.END_COMMAND + simulate_interface;
			}
	        
	        this.hcCode = " #ifdef SIMULATE " + simulate_interface + " #else " + external_interface + this.hcCode;
        }
        
        this.hcCode = HandelC.getInstance().generateComment(Messenger.BUSES_DECL) + this.hcCode;

    }
    
    
    private void declareBranches(){
    	// Declaring the branches
        int numberOfBranches = this.parallelismAnalyser.getNumberOfBranches();
        List<String> newnames = new Vector<String>();
        for(int i=numberOfBranches-1; i >=0; i--) {
        	newnames.add(HandelC.getInstance().getBranchName(i));
        }
        DatatypesVisitor typesVisitor = new DatatypesVisitor(directives,this.constants);
        typesVisitor.declareType(HandelC.BRANCH, newnames);
        this.hcCode = typesVisitor.getHcCode() + this.hcCode;

    }

    private void declareProcessesPrototypes(){
        Enumeration keys = this.procsInfo.keys();
        
        while(keys.hasMoreElements()) {
        	
            String process = ((String)keys.nextElement()).trim();

            if (!process.equals(HandelC.MAINP)) {
                String declaration = HandelC.INLINE + HandelC.SPACE            
                					+ HandelC.VOID + HandelC.SPACE + process;
				if (this.procsInfo.containsProcArguments(process)) {
				   ACspParameters parameters = this.procsInfo.getProcArguments(process);
				   declaration = declaration + translateFormalArguments(parameters,process);
				}
				declaration = declaration + HandelC.END_COMMAND;
				this.hcCode = declaration + this.hcCode;
            }
        }
        this.hcCode = HandelC.getInstance().generateComment(Messenger.PROCESSES_PROTOTYPES) + this.hcCode;
    }

    private void declareDatatypes(){
        this.hcCode = this.types.getHcCode() + this.hcCode;
    }
    
    private void declareClock(){
        String declaration = HandelC.getInstance().generateComment(Messenger.CLOCK_DECLARATION);
        declaration = declaration + HandelC.CLOCK_DEF;
        declaration = declaration + HandelC.COMMENT_LINE_BREAK;
        
        this.hcCode = declaration + this.hcCode;
    }

    private void declareSynchronisationSemaphors(){
        String declaration = HandelC.getInstance().generateComment(Messenger.SYNCHRONISATION_SEMAPHORS_SIGNALS);
        declaration = declaration 
					+ HandelC.SEMAPHOR + HandelC.SPACE + HandelC.getInstance().getSemaVarOf(HandelC.CURRENT_SYNC_STATUS_LUT)
					+ HandelC.END_COMMAND;
        declaration = declaration 
					+ HandelC.SEMAPHOR + HandelC.SPACE + HandelC.getInstance().getSemaVarOf(HandelC.HAS_WRITER_LUT)
					+ HandelC.END_COMMAND;
        declaration = declaration 
					+ HandelC.SIGNAL + HandelC.SPACE 
					+ "<" + HandelC.BOOLEAN + ">" + HandelC.SPACE
					+ HandelC.WITHDRAW_SIGNALS
					+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(HandelC.CHANNEL) + HandelC.ARRAY_END 
					+ HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.BLOCK_BEGIN;
			        
			        // FIRST: look comment at declareForcedSyncLUT
			        List<String> channelNames = ListUtil.reverse(this.channels.getChannelNames()); 
			        
			        for (int c=0; c < channelNames.size(); c++) {
			        	String channelName = channelNames.get(c);
			        	declaration = declaration + HandelC.getInstance().generateComment(channelName);
			        	declaration = declaration + HandelC.LOGICAL_FALSE;
			        	if (c < channelNames.size() - 1) {
			            	declaration = declaration + HandelC.SEP_ARGUMENTS + HandelC.SPACE;
			        	}
			        }
			        declaration = declaration + HandelC.BLOCK_END + HandelC.END_COMMAND;
        
        this.hcCode = declaration + this.hcCode;
    }
    
    private void declareForcedSyncLUT(){
        String declaration = HandelC.STATIC + HandelC.SPACE 
        					+ HandelC.getInstance().getSetVarOf(HandelC.BRANCH) + HandelC.SPACE 
        					+ HandelC.FORCED_SYNC_LUT 
        					+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(HandelC.CHANNEL) + HandelC.ARRAY_END
        					+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(HandelC.BRANCH) + HandelC.ARRAY_END
        					+ HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.BLOCK_BEGIN;
        
        // DO NOT REMOVE COMMENT BELOW
        // Automatic declaration of the look up table
        // FIRST: Channel names are declared in the reverse order from the list of names
        //			returned by the channels visitor
        //			For instance, if the channels visitor returns [c2, c1, c]
        //			the channels names are declared
        //			#define chan_c 0
        //			#define chan_c1 1
        //			#define chan_c2 2
        
        List<String> channelNames = ListUtil.reverse(this.channels.getChannelNames()); 
        
        for (int c=0; c < channelNames.size(); c++) {
        	String channelName = channelNames.get(c);
        	declaration = declaration + HandelC.getInstance().generateComment(channelName);
        	declaration = declaration + HandelC.BLOCK_BEGIN + HandelC.SPACE;
        	Branch2BranchSetLUT branchesLUT = 
        			this.parallelismAnalyser.getForcedInterleavingLUT().get(channelName);
        	
            for (int branch=0; branch < this.parallelismAnalyser.getNumberOfBranches(); branch++) {
            	declaration = declaration + HandelC.getInstance().generateComment(HandelC.getInstance().getBranchName(branch));
            	Set<Integer> excludedBranches = branchesLUT.get(new Integer(branch));
                List excludedSetNames = new Vector();
                for (Integer excluded : excludedBranches) {
					excludedSetNames.add(HandelC.getInstance().getSetVarOf(HandelC.getInstance().getBranchName(excluded.intValue())));
				}
				declaration = declaration + HandelC.getInstance().getSetUnionCode(excludedSetNames,HandelC.BRANCH);
	        	if (branch < this.parallelismAnalyser.getNumberOfBranches() - 1) {
	            	declaration = declaration + HandelC.SEP_ARGUMENTS + HandelC.SPACE;
	        	}
            }
            
            declaration = declaration + HandelC.SPACE + HandelC.BLOCK_END;
        	if (c < channelNames.size() - 1) {
            	declaration = declaration + HandelC.SEP_ARGUMENTS + HandelC.SPACE;
        	}
        }
        
        declaration = declaration + HandelC.BLOCK_END + HandelC.END_COMMAND;

        this.hcCode = declaration + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.FORCED_INTERLEAVING_LUT) + this.hcCode;
    }

    private void declareMultiSyncLUT(){
        String declaration = HandelC.STATIC + HandelC.SPACE 
        					+ HandelC.getInstance().getSetVarOf(HandelC.BRANCH) + HandelC.SPACE 
        					+ HandelC.MULTI_SYNC_LUT 
        					+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(HandelC.CHANNEL) + HandelC.ARRAY_END
        					+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(HandelC.MAX_MULTI_SYNC_SET) + HandelC.ARRAY_END
        					+ HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.BLOCK_BEGIN;
        
        // FIRST: look comment at declareForcedSyncLUT
        List<String> channelNames = ListUtil.reverse(this.channels.getChannelNames()); 
        
        for (int c=0; c < channelNames.size(); c++) {
        	String channelName = channelNames.get(c);
        	declaration = declaration + HandelC.getInstance().generateComment(channelName);
        	declaration = declaration + HandelC.BLOCK_BEGIN + HandelC.SPACE;

        	Set<Set<Integer>> realSyncOnChannel = new HashSet<Set<Integer>>();
        	if (this.parallelismAnalyser.getRealSyncLUT().containsKey(channelName)) {
    			realSyncOnChannel = this.parallelismAnalyser.getRealSyncLUT().get(channelName);
        	}
        	
        	Iterator<Set<Integer>> allSyncs = realSyncOnChannel.iterator();
        	int card = 0;
            while (allSyncs.hasNext()) {
            	card++;
            	Set<Integer> sync = allSyncs.next();
                List<String> branches = new Vector<String>();
                for (Integer branch : sync) {
					branches.add(HandelC.getInstance().getSetVarOf(HandelC.getInstance().getBranchName(branch.intValue())));
				}
				declaration = declaration + HandelC.getInstance().getSetUnionCode(branches,HandelC.BRANCH);
	        	if (card < this.parallelismAnalyser.getMaxMultiSyncSetCard()) {
	            	declaration = declaration + HandelC.SEP_ARGUMENTS + HandelC.SPACE + HandelC.COMMENT_LINE_BREAK + HandelC.SPACE;
	        	}
            }

            // The remaining sets are empty
            while(card < this.parallelismAnalyser.getMaxMultiSyncSetCard()) {
            	card++;
				declaration = declaration + HandelC.getInstance().getEmptySetVarOf(HandelC.BRANCH);
	        	if (card < this.parallelismAnalyser.getMaxMultiSyncSetCard()) {
	            	declaration = declaration + HandelC.SEP_ARGUMENTS + HandelC.SPACE + HandelC.COMMENT_LINE_BREAK + HandelC.SPACE;
	        	}
            }
            
            declaration = declaration + HandelC.SPACE + HandelC.BLOCK_END;

        	if (c < channelNames.size() - 1) {
            	declaration = declaration + HandelC.SEP_ARGUMENTS + HandelC.SPACE;
        	}
        }
        
        declaration = declaration + HandelC.BLOCK_END + HandelC.END_COMMAND;

        this.hcCode = declaration + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MULTI_SYNC_LUT) + this.hcCode;
    }
    
    private void declareCurrentSyncStatusLUT(){
        String declaration = HandelC.getInstance().getSetVarOf(HandelC.BRANCH) + HandelC.SPACE 
        					+ HandelC.CURRENT_SYNC_STATUS_LUT 
        					+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(HandelC.CHANNEL) + HandelC.ARRAY_END
        					+ HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.BLOCK_BEGIN;
        
        // FIRST: look comment at declareForcedSyncLUT
        List<String> channelNames = ListUtil.reverse(this.channels.getChannelNames()); 
        
        for (int c=0; c < channelNames.size(); c++) {
        	String channelName = channelNames.get(c);
        	declaration = declaration + HandelC.getInstance().generateComment(channelName);
        	declaration = declaration + HandelC.getInstance().getEmptySetVarOf(HandelC.BRANCH);
        	if (c < channelNames.size() - 1) {
            	declaration = declaration + HandelC.SEP_ARGUMENTS + HandelC.SPACE;
        	}
        }
        declaration = declaration + HandelC.BLOCK_END + HandelC.END_COMMAND;

        this.hcCode = declaration + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.CURRENT_SYNC_LUT) + this.hcCode;
    }

    private void declareHasWriterLUT(){
        String declaration = HandelC.BOOLEAN + HandelC.SPACE 
        					+ HandelC.HAS_WRITER_LUT 
        					+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(HandelC.CHANNEL) + HandelC.ARRAY_END
        					+ HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.BLOCK_BEGIN;
        
        // FIRST: look comment at declareForcedSyncLUT
        List<String> channelNames = ListUtil.reverse(this.channels.getChannelNames()); 
        
        for (int c=0; c < channelNames.size(); c++) {
        	String channelName = channelNames.get(c);
        	declaration = declaration + HandelC.getInstance().generateComment(channelName);
        	declaration = declaration + HandelC.LOGICAL_FALSE;
        	if (c < channelNames.size() - 1) {
            	declaration = declaration + HandelC.SEP_ARGUMENTS + HandelC.SPACE;
        	}
        }
        declaration = declaration + HandelC.BLOCK_END + HandelC.END_COMMAND;

        this.hcCode = declaration + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.HAS_WRITER_LUT) + this.hcCode;
    }
    
    private void declareSyncConstants(){
    	/*
		typedef unsigned int 1 SYNC;
		const SYNC syncout =0;
		SYNC syncin;
    	 */
    	
        this.hcCode = HandelC.SYNC_DEF + HandelC.SYNC_OUT_DEF + HandelC.SYNC_IN_DEF + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.SYNC_VALUE_DECL) + this.hcCode;
        
    }
    
    private void declareActionsOnChannelsConstants() {
    	/*
		#define ACTION unsigned int 1
		#define A_READ 0
		#define A_WRITE 1
		#define MAX_MULTI_SYNC_card 2
    	 */
        String nextLine = HandelC.getInstance().generateDefine(HandelC.getInstance().getCardVarOf(HandelC.MAX_MULTI_SYNC_SET) 
																+ HandelC.SPACE 
																+ this.parallelismAnalyser.getMaxMultiSyncSetCard());
        this.hcCode = nextLine + this.hcCode;

		nextLine = HandelC.getInstance().generateDefine(HandelC.A_WRITE 
																+ HandelC.SPACE 
																+ HandelC.MATH_ONE);
        this.hcCode = nextLine + this.hcCode;
        
        nextLine = HandelC.getInstance().generateDefine(HandelC.A_READ 
																+ HandelC.SPACE 
																+ HandelC.MATH_ZERO);
        this.hcCode = nextLine + this.hcCode;
        
        nextLine = HandelC.getInstance().generateDefine(HandelC.ACTION 
																+ HandelC.SPACE 
																+ HandelC.UNSIGNED
																+ HandelC.SPACE 
																+ HandelC.INT
																+ HandelC.SPACE
																+ HandelC.MATH_ONE);
        this.hcCode = nextLine + this.hcCode;

        this.hcCode = HandelC.getInstance().generateComment(Messenger.ACTION_ON_CHANNELS) + this.hcCode;
    }
    
    private void declareAuxiliarySetMacros() {

        /*
        macro expr SET_INTER(SET_S,SET_T) = (SET_S & SET_T);
        */
    	String setInter = HandelC.MACRO_EXP + HandelC.SPACE 
				        + HandelC.SET_INTER + HandelC.PAR_BEGIN
				        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) + HandelC.SEP_ARGUMENTS 
				        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_T) + HandelC.PAR_END
				        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
				        + HandelC.PAR_BEGIN
				        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) 
				        + HandelC.SPACE + HandelC.LOGICAL_BITWISE_AND + HandelC.SPACE
				        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_T) 
				        + HandelC.PAR_END
				        + HandelC.END_COMMAND;
		this.hcCode = setInter + this.hcCode;

        /*
        macro expr SET_DIFF(SET_S,SET_T) = (SET_S & (~SET_T));
        */
    	String setDiff = HandelC.MACRO_EXP + HandelC.SPACE 
                        + HandelC.SET_DIFF + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_T) + HandelC.PAR_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) 
                        + HandelC.SPACE + HandelC.LOGICAL_BITWISE_AND + HandelC.SPACE
                        + HandelC.PAR_BEGIN
                        + HandelC.LOGICAL_BITWISE_NOT + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_T) 
                        + HandelC.PAR_END
                        + HandelC.PAR_END
                        + HandelC.END_COMMAND;
        this.hcCode = setDiff + this.hcCode;

        /*
        macro expr SET_UNION(SET_S,SET_T) = (SET_S | SET_T);
        */
    	String setUnion = HandelC.MACRO_EXP + HandelC.SPACE 
                        + HandelC.SET_UNION + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_T) + HandelC.PAR_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) 
                        + HandelC.SPACE + HandelC.LOGICAL_BITWISE_OR + HandelC.SPACE
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_T) 
                        + HandelC.PAR_END
                        + HandelC.END_COMMAND;
        this.hcCode = setUnion + this.hcCode;

        /*
        macro expr IS_EMPTY_SET(SET_S) = (SET_S == 0);
        */
    	String isEmptyChannelSet = HandelC.MACRO_EXP + HandelC.SPACE 
                        + HandelC.IS_EMPTY_SET + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) + HandelC.PAR_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) 
                        + HandelC.SPACE + HandelC.REL_EQUALS + HandelC.SPACE
                        + HandelC.MATH_ZERO 
                        + HandelC.PAR_END
                        + HandelC.END_COMMAND;
        this.hcCode = isEmptyChannelSet + this.hcCode;

        
        /*
         * 	ONLY IF WE HAVE INTERNAL CHOICE DECIDED AT RUNTIME
         *  unsigned 32 random_var = 1;
         *  macro proc random(random_arg) { random_arg = (random_arg<-21) @ (random_arg[12] ^ random_arg[30]) @ (random_arg[10:1]^random_arg[29:20]);} 
        */
        if (this.directives.getIntChoice().equals(Directives.RUN_TIME)) {
	    	String arg = HandelC.getInstance().getArgName(HandelC.INTEGERL, HandelC.INT_ARG_I);
	    	String random = HandelC.MACRO_PROC + HandelC.SPACE + HandelC.RANDOM 
	                        + HandelC.PAR_BEGIN + arg + HandelC.PAR_END
	                        + HandelC.SPACE + HandelC.BLOCK_BEGIN
	                        + arg + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
	                        + HandelC.PAR_BEGIN + arg + HandelC.TAKE_LSB + "21" + HandelC.PAR_END
	                        + HandelC.CONCAT
	                        + HandelC.PAR_BEGIN + arg + HandelC.ARRAY_BEGIN + "12" + HandelC.ARRAY_END 
	                        + HandelC.LOGICAL_BITWISE_XOR
	                        + arg + HandelC.ARRAY_BEGIN + "30" + HandelC.ARRAY_END
	                        + HandelC.PAR_END
	                        + HandelC.CONCAT
	                        + HandelC.PAR_BEGIN + arg + HandelC.ARRAY_BEGIN + "10" + HandelC.BIT_RANGE + "1" + HandelC.ARRAY_END 
	                        + HandelC.LOGICAL_BITWISE_XOR
	                        + arg + HandelC.ARRAY_BEGIN + "29" + HandelC.BIT_RANGE + "20" + HandelC.ARRAY_END
	                        + HandelC.PAR_END + HandelC.END_COMMAND + HandelC.BLOCK_END;
	        this.hcCode = random + this.hcCode;
	    	String declaration = HandelC.UNSIGNED + HandelC.SPACE 
					        + "32" + HandelC.SPACE
					        + HandelC.RANDOM_VAR 
					        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.MATH_ONE + HandelC.END_COMMAND;
	    	this.hcCode = declaration + this.hcCode;
        }
        
        this.hcCode = HandelC.getInstance().generateComment(Messenger.AUX_SET_MACROS) + this.hcCode;

    }

    private void declareAuxiliarySemaphorMacrosAndFunctions() {
    	this.declareConfirmReaders();
    	this.declareTryConfirmReaders();
    	this.declareSleep();
    	this.declareLeaveSemaphor();
    	//this.declareTryLeaveSemaphor();
    	this.declareJoinSemaphor();
    	this.declareTryJoinSemaphor();
    	this.declareLockVariable();
    	//this.declareMayLeave();
    	this.declareMayJoin();
    	this.declareIsAlready();
    	this.declareBranchMember();
    	this.declareIsSyncReady();
    	this.declareReadersReady();
    	this.declareMultiSyncReady();
    	//this.declareNoReaders();
    	this.declareNoWriter();
    	this.declareIsAvailable();
        this.hcCode = HandelC.getInstance().generateComment(Messenger.AUX_SEMA_MACROS) + this.hcCode;
    }
    
    private void declareConfirmReaders() {
        /*
		inline void CONFIRM_READERS(BRANCH WRITER_BRANCH_B,CHANNEL CHANNEL_C) {
		    boolean CONFIRMED;
		    chan boolean end;
		
		    boolean finished;
		    boolean withdraw_in_confirmation;
		    boolean withdraw_in_comm;
		    boolean withdraw_in_finished;
		
		    CONFIRMED = false;
		
		    while(!CONFIRMED) {
		        finished = false;
		        withdraw_in_confirmation = false;
		        withdraw_in_comm = false;
		        withdraw_in_finished = false;
		
		        par {
		            {
		                CONFIRMED = TRY_CONFIRM_READERS(WRITER_BRANCH_B,CHANNEL_C);
		                while(!CONFIRMED){
		                    CONFIRMED = TRY_CONFIRM_READERS(WRITER_BRANCH_B,CHANNEL_C);
		                    delay;
		                };
		                // Wait for last reader (if needed)
		                SLEEP(10);
		                end!withdraw[CHANNEL_C];
		            };
		            {
		                while(!finished) {
		                    prialt{
		                        case end?withdraw_in_comm :{
		                            par {
		                                finished = true;
		                                withdraw_in_finished = withdraw[CHANNEL_C];
		                            }
		                        };
		                        break;
		
		                        default:
		                            if (withdraw[CHANNEL_C]) {
		                                withdraw_in_confirmation = true;
		                            };
		                            delay;
		                        break;
		                    }
		                }
		            }
		        };
		        CONFIRMED =
		            !withdraw_in_confirmation
		            && !withdraw_in_comm
		            && !withdraw_in_finished
		            && !withdraw[CHANNEL_C];
		    }
		}
        */
        String timeForReaderReadyStr	= Config.getInstance().getConfig("TIME_FOR_READER_READY");
        int timeForReaderReady = Integer.parseInt(timeForReaderReadyStr);

    	String confirmReaders = HandelC.INLINE + HandelC.SPACE + HandelC.VOID + HandelC.SPACE
				+ HandelC.CONFIRM_READERS 
				+ HandelC.PAR_BEGIN + HandelC.BRANCH  + HandelC.SPACE 
				+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
				+ HandelC.SEP_ARGUMENTS + HandelC.CHANNEL + HandelC.SPACE 
				+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)	
		        + HandelC.PAR_END + HandelC.SPACE + HandelC.BLOCK_BEGIN
		        + HandelC.BOOLEAN + HandelC.SPACE + HandelC.CONFIRMED + HandelC.END_COMMAND
		        + HandelC.CHAN_DECL + HandelC.SPACE + HandelC.BOOLEAN + HandelC.SPACE + HandelC.CONFIRMED_END + HandelC.END_COMMAND
		        + HandelC.BOOLEAN + HandelC.SPACE + HandelC.CONFIRM_FINISHED + HandelC.END_COMMAND
		        + HandelC.BOOLEAN + HandelC.SPACE + HandelC.WITHDRAW_IN_CONF + HandelC.END_COMMAND
		        + HandelC.BOOLEAN + HandelC.SPACE + HandelC.WITHDRAW_IN_COMM + HandelC.END_COMMAND
		        + HandelC.BOOLEAN + HandelC.SPACE + HandelC.WITHDRAW_IN_FINISHED + HandelC.END_COMMAND
		        + HandelC.CONFIRMED + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.LOGICAL_FALSE + HandelC.END_COMMAND
		        + HandelC.WHILE + HandelC.PAR_BEGIN 
		        + HandelC.LOGICAL_NOT + HandelC.CONFIRMED + HandelC.PAR_END
		        + HandelC.BLOCK_BEGIN
		        + HandelC.CONFIRM_FINISHED + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.LOGICAL_FALSE + HandelC.END_COMMAND
		        + HandelC.WITHDRAW_IN_CONF + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.LOGICAL_FALSE + HandelC.END_COMMAND
		        + HandelC.WITHDRAW_IN_COMM + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.LOGICAL_FALSE + HandelC.END_COMMAND
		        + HandelC.WITHDRAW_IN_FINISHED + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.LOGICAL_FALSE + HandelC.END_COMMAND
		        + HandelC.PARALLEL
		        + HandelC.BLOCK_BEGIN
		        + HandelC.BLOCK_BEGIN
		        + HandelC.CONFIRMED + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
		        + HandelC.TRY_CONFIRM_READERS
		        + HandelC.PAR_BEGIN   
				+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
				+ HandelC.SEP_ARGUMENTS  
				+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)	
		        + HandelC.PAR_END		        
		        + HandelC.END_COMMAND
		        + HandelC.WHILE + HandelC.PAR_BEGIN 
		        + HandelC.LOGICAL_NOT + HandelC.CONFIRMED + HandelC.PAR_END
		        + HandelC.BLOCK_BEGIN
		        + HandelC.CONFIRMED + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
		        + HandelC.TRY_CONFIRM_READERS
		        + HandelC.PAR_BEGIN   
				+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
				+ HandelC.SEP_ARGUMENTS  
				+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)	
		        + HandelC.PAR_END		        
		        + HandelC.END_COMMAND
		        + HandelC.DELAY + HandelC.END_COMMAND
		        + HandelC.BLOCK_END + HandelC.END_COMMAND
		        + HandelC.SLEEP 
		        + HandelC.PAR_BEGIN + timeForReaderReady + HandelC.PAR_END + HandelC.END_COMMAND
		        + HandelC.CONFIRMED_END + HandelC.OUTPUT_COMMUNICATION 
		        + HandelC.WITHDRAW_SIGNALS + HandelC.ARRAY_BEGIN 
	        	+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)
	        	+ HandelC.ARRAY_END + HandelC.END_COMMAND
		        + HandelC.BLOCK_END + HandelC.END_COMMAND
		        + HandelC.BLOCK_BEGIN
		        + HandelC.WHILE + HandelC.PAR_BEGIN 
		        + HandelC.LOGICAL_NOT + HandelC.CONFIRM_FINISHED + HandelC.PAR_END
		        + HandelC.BLOCK_BEGIN
		        + HandelC.ALTERNATION
		        + HandelC.BLOCK_BEGIN
		        + HandelC.CASE + HandelC.SPACE + HandelC.CONFIRMED_END + HandelC.INPUT_COMMUNICATION 
		        + HandelC.WITHDRAW_IN_COMM + HandelC.CASE_SEP
		        + HandelC.BLOCK_BEGIN
		        + HandelC.PARALLEL + HandelC.BLOCK_BEGIN
		        + HandelC.CONFIRM_FINISHED + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.LOGICAL_TRUE + HandelC.END_COMMAND
		        + HandelC.WITHDRAW_IN_FINISHED + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
		        + HandelC.WITHDRAW_SIGNALS + HandelC.ARRAY_BEGIN 
	        	+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)
	        	+ HandelC.ARRAY_END + HandelC.END_COMMAND
		        + HandelC.BLOCK_END
		        + HandelC.BLOCK_END + HandelC.END_COMMAND + HandelC.BREAK + HandelC.END_COMMAND
	        	+ HandelC.DEFAULT + HandelC.CASE_SEP
	        	+ HandelC.SPACE
	        	+ HandelC.BLOCK_BEGIN
	        	+ HandelC.IF + HandelC.PAR_BEGIN 
	        	+ HandelC.WITHDRAW_SIGNALS + HandelC.ARRAY_BEGIN 
	        	+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)
	        	+ HandelC.ARRAY_END
	        	+ HandelC.PAR_END + HandelC.BLOCK_BEGIN
		        + HandelC.WITHDRAW_IN_CONF + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.LOGICAL_TRUE + HandelC.END_COMMAND
	        	+ HandelC.BLOCK_END + HandelC.END_COMMAND
	        	+ HandelC.DELAY + HandelC.END_COMMAND
	        	+ HandelC.BLOCK_END + HandelC.END_COMMAND 
	        	+ HandelC.BREAK + HandelC.END_COMMAND
	        	+ HandelC.BLOCK_END
		        + HandelC.BLOCK_END
		        + HandelC.BLOCK_END 
		        + HandelC.BLOCK_END + HandelC.END_COMMAND
		        + HandelC.CONFIRMED + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
		        + HandelC.LOGICAL_NOT + HandelC.WITHDRAW_IN_CONF + HandelC.SPACE
		        + HandelC.LOGICAL_AND + HandelC.SPACE + HandelC.LOGICAL_NOT + HandelC.WITHDRAW_IN_COMM + HandelC.SPACE 
		        + HandelC.LOGICAL_AND + HandelC.SPACE + HandelC.LOGICAL_NOT + HandelC.WITHDRAW_IN_FINISHED + HandelC.SPACE
		        + HandelC.LOGICAL_AND + HandelC.SPACE + HandelC.LOGICAL_NOT 
		        	+ HandelC.WITHDRAW_SIGNALS + HandelC.ARRAY_BEGIN 
		        	+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)
		        	+ HandelC.ARRAY_END
		        + HandelC.END_COMMAND
		        + HandelC.BLOCK_END
		        + HandelC.BLOCK_END;
		this.hcCode = confirmReaders + this.hcCode;
		this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_CONFIRM_READERS) + this.hcCode;
    	
    }
    
    private void declareTryConfirmReaders() {
        /*
		inline boolean TRY_CONFIRM_READERS(BRANCH WRITER_BRANCH_B,CHANNEL CHANNEL_C) {
		    boolean ready;
		    LOCK_VARIABLE(&CURRENT_SYNC_STATUS_LUT_SEMA);
		    ready = READERS_READY(WRITER_BRANCH_B,CHANNEL_C);
		    releasesema(CURRENT_SYNC_STATUS_LUT_SEMA);
		    return ready;
		}
        */
    	String tryConfirmReaders = HandelC.INLINE + HandelC.SPACE + HandelC.BOOLEAN + HandelC.SPACE
				+ HandelC.TRY_CONFIRM_READERS 
				+ HandelC.PAR_BEGIN + HandelC.BRANCH  + HandelC.SPACE 
				+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
				+ HandelC.SEP_ARGUMENTS + HandelC.CHANNEL + HandelC.SPACE 
				+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)	
		        + HandelC.PAR_END + HandelC.SPACE + HandelC.BLOCK_BEGIN
		        + HandelC.BOOLEAN + HandelC.SPACE + HandelC.ARE_READERS_READY + HandelC.END_COMMAND
		        + HandelC.LOCK_VAR + HandelC.PAR_BEGIN 
		        + HandelC.ADDRESS + HandelC.getInstance().getSemaVarOf(HandelC.CURRENT_SYNC_STATUS_LUT) 
		        + HandelC.PAR_END + HandelC.END_COMMAND
		        + HandelC.ARE_READERS_READY + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE
				+ HandelC.MULTI_SYNC_READY_WITH_ME
				+ HandelC.PAR_BEGIN
				+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
				+ HandelC.SEP_ARGUMENTS 
				+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
				+ HandelC.PAR_END + HandelC.END_COMMAND
		        + HandelC.RELEASE_SEMAPHOR + HandelC.PAR_BEGIN
		        + HandelC.getInstance().getSemaVarOf(HandelC.CURRENT_SYNC_STATUS_LUT) 
		        + HandelC.PAR_END + HandelC.END_COMMAND
		        + HandelC.RETURN + HandelC.SPACE + HandelC.ARE_READERS_READY + HandelC.END_COMMAND
		        + HandelC.BLOCK_END;
		this.hcCode = tryConfirmReaders + this.hcCode;
		this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_TRY_CONFIRM_READERS) + this.hcCode;
    	
    }
    
    private void declareSleep() {
        /*
		inline void SLEEP(unsigned int 10 time) {
		    while (time > 0) {
		        delay;
		        time--;
		    }
		}
        */
        String maxSleepStr	= Config.getInstance().getConfig("MAXIMUM_SLEEP_TIME");
        int maxSleep = Integer.parseInt(maxSleepStr);
    	int length = MathUtil.getUnsignedBitsForNumber(maxSleep);
    	
    	String sleep = HandelC.INLINE + HandelC.SPACE + HandelC.VOID + HandelC.SPACE 
                        + HandelC.SLEEP + HandelC.PAR_BEGIN
                        + HandelC.UNSIGNED + HandelC.SPACE + HandelC.INT + HandelC.SPACE 
                        + length + HandelC.SPACE + HandelC.SLEEP_TIME
                        + HandelC.PAR_END
                        + HandelC.BLOCK_BEGIN
                        + HandelC.WHILE + HandelC.PAR_BEGIN
                        + HandelC.SLEEP_TIME + HandelC.SPACE + HandelC.REL_GREATER + HandelC.SPACE + HandelC.MATH_ZERO
                        + HandelC.PAR_END + HandelC.BLOCK_BEGIN
                        + HandelC.DELAY + HandelC.END_COMMAND
                        + HandelC.SLEEP_TIME + HandelC.MATH_DEC + HandelC.END_COMMAND
                        + HandelC.BLOCK_END
                        + HandelC.BLOCK_END;
        this.hcCode = sleep + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_SLEEP) + this.hcCode;
    }
    
    private void declareLeaveSemaphor() {
        /*
		inline void LEAVE_SEMAPHOR(BRANCH BRANCH_B,CHANNEL CHANNEL_C,ACTION ACTION_A){
		    LOCK_VARIABLE(&CURRENT_SYNC_STATUS_LUT_SEMA);
		    LOCK_VARIABLE(&HAS_WRITER_LUT_SEMA);
		    CURRENT_SYNC_STATUS_LUT[CHANNEL_C] = SET_DIFF(CURRENT_SYNC_STATUS_LUT[CHANNEL_C],BRANCH__set_LUT[BRANCH_B]);
		    if(ACTION_A==A_WRITE){
		        HAS_WRITER_LUT[CHANNEL_C] = false;
		    }
		    releasesema(CURRENT_SYNC_STATUS_LUT_SEMA);
		    releasesema(HAS_WRITER_LUT_SEMA);
		}
        */
    	String leaveSema = HandelC.INLINE + HandelC.SPACE + HandelC.VOID + HandelC.SPACE 
                        + HandelC.LEAVE_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.BRANCH + HandelC.SPACE + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.CHANNEL + HandelC.SPACE + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
    					+ HandelC.SEP_ARGUMENTS 
    					+ HandelC.ACTION + HandelC.SPACE + HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A)	
                        + HandelC.PAR_END
                        + HandelC.BLOCK_BEGIN
                        + HandelC.LOCK_VAR + HandelC.PAR_BEGIN 
                        + HandelC.ADDRESS + HandelC.getInstance().getSemaVarOf(HandelC.CURRENT_SYNC_STATUS_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.LOCK_VAR + HandelC.PAR_BEGIN 
                        + HandelC.ADDRESS + HandelC.getInstance().getSemaVarOf(HandelC.HAS_WRITER_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.CURRENT_SYNC_STATUS_LUT + HandelC.ARRAY_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) + HandelC.ARRAY_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE
                        + HandelC.SET_DIFF + HandelC.PAR_BEGIN
                        + HandelC.CURRENT_SYNC_STATUS_LUT + HandelC.ARRAY_BEGIN  
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) + HandelC.ARRAY_END
                        + HandelC.SEP_ARGUMENTS
                        + HandelC.getInstance().getSingletonSetsLookupTableName(HandelC.BRANCH) + HandelC.ARRAY_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) + HandelC.ARRAY_END
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.IF 
                        + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A)
                        + HandelC.REL_EQUALS + HandelC.A_WRITE 
                        + HandelC.PAR_END + HandelC.BLOCK_BEGIN
                        + HandelC.HAS_WRITER_LUT 
                        + HandelC.ARRAY_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
                        + HandelC.ARRAY_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.LOGICAL_FALSE + HandelC.END_COMMAND
                        + HandelC.BLOCK_END
                        + HandelC.RELEASE_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getSemaVarOf(HandelC.CURRENT_SYNC_STATUS_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.RELEASE_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getSemaVarOf(HandelC.HAS_WRITER_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.BLOCK_END;
        this.hcCode = leaveSema + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_LEAVE_SEMAPHOR) + this.hcCode;
        
        // OLD ONE
        /*
		inline void LEAVE_SEMAPHOR(BRANCH BRANCH_B,CHANNEL CHANNEL_C,ACTION ACTION_A){
		    boolean HAS_LEAVED;
		    HAS_LEAVED = TRY_LEAVE_SEMAPHOR(BRANCH_B,CHANNEL_C,ACTION_A);
		    while(!HAS_LEAVED){
		        HAS_LEAVED = TRY_LEAVE_SEMAPHOR(BRANCH_B,CHANNEL_C,ACTION_A);
		        delay;
		    }
		}
        */
    	/*
    	String leaveSema = HandelC.INLINE + HandelC.SPACE + HandelC.VOID + HandelC.SPACE 
                        + HandelC.LEAVE_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.BRANCH + HandelC.SPACE + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.CHANNEL + HandelC.SPACE + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
    					+ HandelC.SEP_ARGUMENTS 
    					+ HandelC.ACTION + HandelC.SPACE + HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A)	
                        + HandelC.PAR_END
                        + HandelC.BLOCK_BEGIN
                        + HandelC.BOOLEAN + HandelC.SPACE + HandelC.HAS_LEFT + HandelC.END_COMMAND
                        + HandelC.HAS_LEFT + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.TRY_LEAVE_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.WHILE	+ HandelC.PAR_BEGIN 
                        + HandelC.LOGICAL_NOT + HandelC.HAS_LEFT + HandelC.PAR_END + HandelC.BLOCK_BEGIN
                        + HandelC.HAS_LEFT + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.TRY_LEAVE_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.DELAY + HandelC.END_COMMAND	
                        + HandelC.BLOCK_END
                        + HandelC.BLOCK_END;
        */
    }

    private void declareJoinSemaphor() {
        /*
		inline void JOIN_SEMAPHOR(BRANCH BRANCH_B,CHANNEL CHANNEL_C, ACTION ACTION_A){
		    boolean HAS_JOINED;
		    HAS_JOINED = TRY_JOIN_SEMAPHOR(BRANCH_B,CHANNEL_C,ACTION_A); 
		    while(!HAS_JOINED){
		        HAS_JOINED = TRY_JOIN_SEMAPHOR(BRANCH_B,CHANNEL_C,ACTION_A);
		        delay;
		    }
		}
        */
    	String joinSema = HandelC.INLINE + HandelC.SPACE + HandelC.VOID + HandelC.SPACE 
                        + HandelC.JOIN_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.BRANCH + HandelC.SPACE + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.CHANNEL + HandelC.SPACE + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
    					+ HandelC.SEP_ARGUMENTS 
    					+ HandelC.ACTION + HandelC.SPACE + HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A)	
                        + HandelC.PAR_END
                        + HandelC.BLOCK_BEGIN
                        + HandelC.BOOLEAN + HandelC.SPACE + HandelC.HAS_JOINED + HandelC.END_COMMAND
                        + HandelC.HAS_JOINED + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.TRY_JOIN_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.WHILE	+ HandelC.PAR_BEGIN 
                        + HandelC.LOGICAL_NOT + HandelC.HAS_JOINED + HandelC.PAR_END + HandelC.BLOCK_BEGIN
                        + HandelC.HAS_JOINED + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.TRY_JOIN_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.DELAY + HandelC.END_COMMAND	
                        + HandelC.BLOCK_END
                        + HandelC.BLOCK_END;
        this.hcCode = joinSema + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_JOIN_SEMAPHOR) + this.hcCode;
    }

    private void declareTryJoinSemaphor() {
        /*
		inline boolean TRY_JOIN_SEMAPHOR(BRANCH BRANCH_B,CHANNEL CHANNEL_C, ACTION ACTION_A){
		    boolean HAS_JOINED;
		    if(MAY_JOIN(BRANCH_B,CHANNEL_C,ACTION_A)){
		        LOCK_VARIABLE(&FORCED_SYNC_LUT_SEMA);
		        LOCK_VARIABLE(&CURRENT_SYNC_STATUS_LUT_SEMA);
		        LOCK_VARIABLE(&HAS_WRITER_LUT_SEMA);
		        if(MAY_JOIN(BRANCH_B,CHANNEL_C,ACTION_A)){
		            CURRENT_SYNC_STATUS_LUT[CHANNEL_C] = SET_UNION(CURRENT_SYNC_STATUS_LUT[CHANNEL_C],BRANCH__set_LUT[BRANCH_B]);
		            if (ACTION_A == A_WRITE) {
		                HAS_WRITER_LUT[CHANNEL_C] = true;
		            }
		            HAS_JOINED = true;
		        }
		        else{
		            HAS_JOINED = false;
		        }
		        releasesema(CURRENT_SYNC_STATUS_LUT_SEMA);
		        releasesema(FORCED_SYNC_LUT_SEMA);
		        releasesema(HAS_WRITER_LUT_SEMA);
		    }
		    else{
		        HAS_JOINED = false;
		    }
		    return HAS_JOINED;
		}
        */
    	String tryJoinSema = HandelC.INLINE + HandelC.SPACE + HandelC.BOOLEAN + HandelC.SPACE
    					+ HandelC.TRY_JOIN_SEMAPHOR 
    					+ HandelC.PAR_BEGIN + HandelC.BRANCH  + HandelC.SPACE 
    					+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
    					+ HandelC.SEP_ARGUMENTS + HandelC.CHANNEL + HandelC.SPACE 
    					+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)	
    					+ HandelC.SEP_ARGUMENTS + HandelC.ACTION + HandelC.SPACE 
    					+ HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A)	
                        + HandelC.PAR_END + HandelC.SPACE + HandelC.BLOCK_BEGIN
                        + HandelC.BOOLEAN + HandelC.SPACE + HandelC.HAS_JOINED + HandelC.END_COMMAND
                        + HandelC.IF + HandelC.PAR_BEGIN 
                        + HandelC.MAY_JOIN + HandelC.PAR_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)
                        + HandelC.SEP_ARGUMENTS 
    					+ HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A)	
                        + HandelC.PAR_END + HandelC.PAR_END + HandelC.BLOCK_BEGIN
                        + HandelC.LOCK_VAR + HandelC.PAR_BEGIN 
                        + HandelC.ADDRESS + HandelC.getInstance().getSemaVarOf(HandelC.CURRENT_SYNC_STATUS_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.LOCK_VAR + HandelC.PAR_BEGIN 
                        + HandelC.ADDRESS + HandelC.getInstance().getSemaVarOf(HandelC.HAS_WRITER_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.IF + HandelC.PAR_BEGIN 
                        + HandelC.MAY_JOIN + HandelC.PAR_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A)
                        + HandelC.PAR_END + HandelC.PAR_END + HandelC.BLOCK_BEGIN
                        + HandelC.CURRENT_SYNC_STATUS_LUT + HandelC.ARRAY_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) + HandelC.ARRAY_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE
                        + HandelC.SET_UNION + HandelC.PAR_BEGIN
                        + HandelC.CURRENT_SYNC_STATUS_LUT + HandelC.ARRAY_BEGIN  
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) + HandelC.ARRAY_END
                        + HandelC.SEP_ARGUMENTS
                        + HandelC.getInstance().getSingletonSetsLookupTableName(HandelC.BRANCH) + HandelC.ARRAY_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) + HandelC.ARRAY_END
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.IF 
                        + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A)
                        + HandelC.REL_EQUALS + HandelC.A_WRITE 
                        + HandelC.PAR_END + HandelC.BLOCK_BEGIN
                        + HandelC.HAS_WRITER_LUT 
                        + HandelC.ARRAY_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
                        + HandelC.ARRAY_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.LOGICAL_TRUE + HandelC.END_COMMAND
                        + HandelC.BLOCK_END
                        + HandelC.HAS_JOINED + HandelC.SPACE + HandelC.ASSIGNMENT 
                        + HandelC.SPACE + HandelC.LOGICAL_TRUE + HandelC.END_COMMAND 
                        + HandelC.BLOCK_END + HandelC.ELSE + HandelC.BLOCK_BEGIN
                        + HandelC.HAS_JOINED + HandelC.SPACE + HandelC.ASSIGNMENT 
                        + HandelC.SPACE + HandelC.LOGICAL_FALSE + HandelC.END_COMMAND 
                        + HandelC.BLOCK_END
                        + HandelC.RELEASE_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getSemaVarOf(HandelC.CURRENT_SYNC_STATUS_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.RELEASE_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getSemaVarOf(HandelC.HAS_WRITER_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.BLOCK_END + HandelC.ELSE + HandelC.BLOCK_BEGIN
                        + HandelC.HAS_JOINED + HandelC.SPACE + HandelC.ASSIGNMENT 
                        + HandelC.SPACE + HandelC.LOGICAL_FALSE + HandelC.END_COMMAND 
                        + HandelC.BLOCK_END
                        + HandelC.RETURN + HandelC.SPACE + HandelC.HAS_JOINED + HandelC.END_COMMAND
                        + HandelC.BLOCK_END;
        this.hcCode = tryJoinSema + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_TRY_JOIN_SEMAPHOR) + this.hcCode;
    }

    private void declareTryLeaveSemaphor() {
        /*
		inline boolean TRY_LEAVE_SEMAPHOR(BRANCH BRANCH_B,CHANNEL CHANNEL_C,ACTION ACTION_A){
		    boolean HAS_LEAVED;
		    if(MAY_LEAVE(BRANCH_B,CHANNEL_C,ACTION_A)){
		        LOCK_VARIABLE(&CURRENT_SYNC_STATUS_LUT_SEMA);
		        LOCK_VARIABLE(&HAS_WRITER_LUT_SEMA);
		        if(MAY_LEAVE(BRANCH_B,CHANNEL_C,ACTION_A)){
		            CURRENT_SYNC_STATUS_LUT[CHANNEL_C] = SET_DIFF(CURRENT_SYNC_STATUS_LUT[CHANNEL_C],BRANCH__set_LUT[BRANCH_B]);
		            if(ACTION_A==A_WRITE){
		                HAS_WRITER_LUT[CHANNEL_C] = false;
		            }
		            HAS_LEAVED = true;
		        }
		        else{
		            HAS_LEAVED = false;
		        }
		        releasesema(CURRENT_SYNC_STATUS_LUT_SEMA);
		        releasesema(HAS_WRITER_LUT_SEMA);
		    }
		    else{
		        HAS_LEAVED = false;
		    }
		    return HAS_LEAVED;
		}
        */
    	String tryLeaveSema = HandelC.INLINE + HandelC.SPACE + HandelC.BOOLEAN + HandelC.SPACE
    					+ HandelC.TRY_LEAVE_SEMAPHOR 
    					+ HandelC.PAR_BEGIN + HandelC.BRANCH  + HandelC.SPACE 
    					+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
    					+ HandelC.SEP_ARGUMENTS + HandelC.CHANNEL + HandelC.SPACE 
    					+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)	
    					+ HandelC.SEP_ARGUMENTS + HandelC.ACTION + HandelC.SPACE 
    					+ HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A)	
                        + HandelC.PAR_END + HandelC.SPACE + HandelC.BLOCK_BEGIN
                        + HandelC.BOOLEAN + HandelC.SPACE + HandelC.HAS_LEFT + HandelC.END_COMMAND
                        + HandelC.IF + HandelC.PAR_BEGIN 
                        + HandelC.MAY_LEAVE + HandelC.PAR_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)
                        + HandelC.SEP_ARGUMENTS 
    					+ HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A)	
                        + HandelC.PAR_END + HandelC.PAR_END + HandelC.BLOCK_BEGIN
                        + HandelC.LOCK_VAR + HandelC.PAR_BEGIN 
                        + HandelC.ADDRESS + HandelC.getInstance().getSemaVarOf(HandelC.CURRENT_SYNC_STATUS_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.LOCK_VAR + HandelC.PAR_BEGIN 
                        + HandelC.ADDRESS + HandelC.getInstance().getSemaVarOf(HandelC.HAS_WRITER_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.IF + HandelC.PAR_BEGIN 
                        + HandelC.MAY_LEAVE + HandelC.PAR_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A)
                        + HandelC.PAR_END + HandelC.PAR_END + HandelC.BLOCK_BEGIN
                        + HandelC.CURRENT_SYNC_STATUS_LUT + HandelC.ARRAY_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) + HandelC.ARRAY_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE
                        + HandelC.SET_DIFF + HandelC.PAR_BEGIN
                        + HandelC.CURRENT_SYNC_STATUS_LUT + HandelC.ARRAY_BEGIN  
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) + HandelC.ARRAY_END
                        + HandelC.SEP_ARGUMENTS
                        + HandelC.getInstance().getSingletonSetsLookupTableName(HandelC.BRANCH) + HandelC.ARRAY_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) + HandelC.ARRAY_END
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.IF 
                        + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A)
                        + HandelC.REL_EQUALS + HandelC.A_WRITE 
                        + HandelC.PAR_END + HandelC.BLOCK_BEGIN
                        + HandelC.HAS_WRITER_LUT 
                        + HandelC.ARRAY_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
                        + HandelC.ARRAY_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.LOGICAL_FALSE + HandelC.END_COMMAND
                        + HandelC.BLOCK_END
                        + HandelC.HAS_LEFT + HandelC.SPACE + HandelC.ASSIGNMENT 
                        + HandelC.SPACE + HandelC.LOGICAL_TRUE + HandelC.END_COMMAND 
                        + HandelC.BLOCK_END + HandelC.ELSE + HandelC.BLOCK_BEGIN
                        + HandelC.HAS_LEFT + HandelC.SPACE + HandelC.ASSIGNMENT 
                        + HandelC.SPACE + HandelC.LOGICAL_FALSE + HandelC.END_COMMAND 
                        + HandelC.BLOCK_END
                        + HandelC.RELEASE_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getSemaVarOf(HandelC.CURRENT_SYNC_STATUS_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.RELEASE_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getSemaVarOf(HandelC.HAS_WRITER_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.BLOCK_END + HandelC.ELSE + HandelC.BLOCK_BEGIN
                        + HandelC.HAS_LEFT + HandelC.SPACE + HandelC.ASSIGNMENT 
                        + HandelC.SPACE + HandelC.LOGICAL_FALSE + HandelC.END_COMMAND 
                        + HandelC.BLOCK_END
                        + HandelC.RETURN + HandelC.SPACE + HandelC.HAS_LEFT + HandelC.END_COMMAND
                        + HandelC.BLOCK_END;
        this.hcCode = tryLeaveSema + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_TRY_LEAVE_SEMAPHOR) + this.hcCode;
    }
    
    private void declareLockVariable() {
        /*
         * inline void LOCK_VARIABLE(sema *sema_s) {
    	 * 		while(trysema(*sema_s)==0) delay;
		 * }
         */
    	String lockVariable = HandelC.INLINE + HandelC.SPACE + HandelC.VOID + HandelC.SPACE 
				        + HandelC.LOCK_VAR + HandelC.PAR_BEGIN
				        + HandelC.SEMAPHOR + HandelC.SPACE + HandelC.REFERENCE 
				        + HandelC.getInstance().getArgName(HandelC.SEMAPHOR, HandelC.SET_ARG_S) + HandelC.PAR_END
				        + HandelC.BLOCK_BEGIN
				        + HandelC.WHILE	+ HandelC.PAR_BEGIN
				        + HandelC.TRY_SEMAPHOR + HandelC.PAR_BEGIN + HandelC.REFERENCE 
				        + HandelC.getInstance().getArgName(HandelC.SEMAPHOR, HandelC.SET_ARG_S) + HandelC.PAR_END
				        + HandelC.REL_EQUALS + HandelC.MATH_ZERO + HandelC.PAR_END
				        + HandelC.SPACE + HandelC.DELAY + HandelC.END_COMMAND
				        + HandelC.BLOCK_END;
		this.hcCode = lockVariable + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_LOCK_VARIABLE) + this.hcCode;
    }

    private void declareMayJoin() {
        /*
		macro expr MAY_JOIN(BRANCH_B,CHANNEL_C,ACTION_A) = 
		    IS_ALREADY(BRANCH_B,CHANNEL_C) || ((IS_AVAILABLE(BRANCH_B,CHANNEL_C) && IS_SYNC_READY(BRANCH_B,CHANNEL_C,ACTION_A)));
         */
    	String mayJoin = HandelC.MACRO_EXP + HandelC.SPACE + HandelC.MAY_JOIN 
						+ HandelC.PAR_BEGIN 
						+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A) 
						+ HandelC.PAR_END + HandelC.SPACE 
						+ HandelC.ASSIGNMENT + HandelC.SPACE 
						+ HandelC.PAR_BEGIN
						+ HandelC.PAR_BEGIN
						+ HandelC.IS_ALREADY
						+ HandelC.PAR_BEGIN
						+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
						+ HandelC.PAR_END
						+ HandelC.PAR_END
						+ HandelC.SPACE + HandelC.LOGICAL_OR + HandelC.SPACE
						+ HandelC.PAR_BEGIN
						+ HandelC.IS_AVAILABLE
						+ HandelC.PAR_BEGIN
						+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
						+ HandelC.PAR_END
						+ HandelC.SPACE + HandelC.LOGICAL_AND + HandelC.SPACE
						+ HandelC.IS_SYNC_READY
						+ HandelC.PAR_BEGIN 
						+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A) 
						+ HandelC.PAR_END 
						+ HandelC.PAR_END
						+ HandelC.PAR_END
    					+ HandelC.END_COMMAND;
    	this.hcCode = mayJoin + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_MAY_JOIN) + this.hcCode;
    }

    /*
    private void declareMayLeave() {
		macro expr MAY_LEAVE(BRANCH_B,CHANNEL_C,ACTION_A) = 
		    (!IS_ALREADY(BRANCH_B,CHANNEL_C)) || ((ACTION_A==A_READ)||(ACTION_A==A_WRITE && NO_READERS(BRANCH_B,CHANNEL_C)));
    	String mayLeave = HandelC.MACRO_EXP + HandelC.SPACE + HandelC.MAY_LEAVE 
						+ HandelC.PAR_BEGIN 
						+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A) 
						+ HandelC.PAR_END + HandelC.SPACE 
						+ HandelC.ASSIGNMENT + HandelC.SPACE 
						+ HandelC.PAR_BEGIN 
						+ HandelC.PAR_BEGIN
						+ HandelC.LOGICAL_NOT
						+ HandelC.IS_ALREADY
						+ HandelC.PAR_BEGIN
						+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
						+ HandelC.PAR_END
						+ HandelC.PAR_END
						+ HandelC.SPACE + HandelC.LOGICAL_OR + HandelC.SPACE
						+ HandelC.PAR_BEGIN
						+ HandelC.PAR_BEGIN
						+ HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A) 
						+ HandelC.REL_EQUALS 
						+ HandelC.A_READ  
						+ HandelC.PAR_END
						+ HandelC.LOGICAL_OR
						+ HandelC.PAR_BEGIN
						+ HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A) 
						+ HandelC.REL_EQUALS 
						+ HandelC.A_WRITE
						+ HandelC.SPACE + HandelC.LOGICAL_AND + HandelC.SPACE
						+ HandelC.NO_READERS
						+ HandelC.PAR_BEGIN
						+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
						+ HandelC.PAR_END
						+ HandelC.PAR_END
						+ HandelC.PAR_END
						+ HandelC.PAR_END
    					+ HandelC.END_COMMAND;
    	this.hcCode = mayLeave + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_MAY_LEAVE) + this.hcCode;
    }
     */

    private void declareIsAlready() {
        /*
		macro expr IS_ALREADY(BRANCH_B,CHANNEL_C) = 
		    (BRANCH_MEMBER(BRANCH_B,CURRENT_SYNC_STATUS_LUT[CHANNEL_C]));
         */
    	String isAlready = HandelC.MACRO_EXP + HandelC.SPACE + HandelC.IS_ALREADY 
						+ HandelC.PAR_BEGIN 
						+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
						+ HandelC.PAR_END + HandelC.SPACE 
						+ HandelC.ASSIGNMENT + HandelC.SPACE 
						+ HandelC.PAR_BEGIN
						+ HandelC.BRANCH_MEMBER
						+ HandelC.PAR_BEGIN
						+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
						+ HandelC.SEP_ARGUMENTS
                        + HandelC.CURRENT_SYNC_STATUS_LUT + HandelC.ARRAY_BEGIN  
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) + HandelC.ARRAY_END
						+ HandelC.PAR_END
						+ HandelC.PAR_END
    					+ HandelC.END_COMMAND;
    	this.hcCode = isAlready + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_IS_ALREADY) + this.hcCode;
    }

    private void declareBranchMember() {
        /*
		macro expr BRANCH_MEMBER(BRANCH_B,BRANCH_SET_S) = 
		    (SET_UNION(BRANCH__set_LUT[BRANCH_B],BRANCH_SET_S) == BRANCH_SET_S);
         */
    	String branchMember = HandelC.MACRO_EXP + HandelC.SPACE + HandelC.BRANCH_MEMBER 
								+ HandelC.PAR_BEGIN 
								+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
								+ HandelC.SEP_ARGUMENTS 
								+ HandelC.getInstance().getArgName(HandelC.BRANCH_SET, HandelC.SET_ARG_S) 
								+ HandelC.PAR_END + HandelC.SPACE 
								+ HandelC.ASSIGNMENT 
								+ HandelC.SPACE 
								+ HandelC.PAR_BEGIN
								+ HandelC.SET_UNION
								+ HandelC.PAR_BEGIN
								+ HandelC.getInstance().getSingletonSetsLookupTableName(HandelC.BRANCH)
								+ HandelC.ARRAY_BEGIN
								+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
								+ HandelC.ARRAY_END
								+ HandelC.SEP_ARGUMENTS
								+ HandelC.getInstance().getArgName(HandelC.BRANCH_SET, HandelC.SET_ARG_S)
								+ HandelC.PAR_END
								+ HandelC.REL_EQUALS 
								+ HandelC.getInstance().getArgName(HandelC.BRANCH_SET, HandelC.SET_ARG_S)
								+ HandelC.PAR_END
								+ HandelC.END_COMMAND;
			this.hcCode = branchMember + this.hcCode;
			this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_BRANCH_MEMBER) + this.hcCode;
    }
    
    private void declareIsSyncReady() {
        /*
		macro expr IS_SYNC_READY(BRANCH_B,CHANNEL_C,ACTION_A) = 
		    //NO_WRITER(CHANNEL_C) 
		    //&& 
		    ((ACTION_A==A_READ)||(ACTION_A==A_WRITE && READERS_READY(BRANCH_B,CHANNEL_C)));
         */
    	String readersReady = HandelC.MACRO_EXP + HandelC.SPACE + HandelC.IS_SYNC_READY 
						+ HandelC.PAR_BEGIN 
						+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A) 
						+ HandelC.PAR_END + HandelC.SPACE 
						+ HandelC.ASSIGNMENT + HandelC.SPACE 
						/*
						+ HandelC.NO_WRITER
						+ HandelC.PAR_BEGIN
						+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
						+ HandelC.PAR_END
						+ HandelC.SPACE + HandelC.LOGICAL_AND + HandelC.SPACE
						*/
						+ HandelC.PAR_BEGIN
						+ HandelC.PAR_BEGIN
						+ HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A) 
						+ HandelC.REL_EQUALS 
						+ HandelC.A_READ  
						+ HandelC.PAR_END
						+ HandelC.LOGICAL_OR
						+ HandelC.PAR_BEGIN
						+ HandelC.getInstance().getArgName(HandelC.ACTION, HandelC.ACTION_ARG_A) 
						+ HandelC.REL_EQUALS 
						+ HandelC.A_WRITE
						+ HandelC.SPACE + HandelC.LOGICAL_AND + HandelC.SPACE
						+ HandelC.MULTI_SYNC_READY_WITH_ME
						+ HandelC.PAR_BEGIN
						+ HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
						+ HandelC.PAR_END
						+ HandelC.PAR_END
						+ HandelC.PAR_END
    					+ HandelC.END_COMMAND;
    	this.hcCode = readersReady + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_IS_SYNC_READY) + this.hcCode;
    }
    
    
    private void declareReadersReady() {
        /*
		macro expr READERS_READY(WRITER_BRANCH_B, CHANNEL_C) = 
		    MULTI_SYNC_READY(CHANNEL_C, 
		    				SET_UNION(CURRENT_SYNC_STATUS_LUT[CHANNEL_C], 
		    						  BRANCH__set_LUT[WRITER_BRANCH_B]));
         */
    	String readersReady = HandelC.MACRO_EXP + HandelC.SPACE + HandelC.MULTI_SYNC_READY_WITH_ME 
						+ HandelC.PAR_BEGIN 
						+ HandelC.getInstance().getArgName(HandelC.WRITER_BRANCH, HandelC.BRANCH_ARG_B) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
						+ HandelC.PAR_END + HandelC.SPACE 
						+ HandelC.ASSIGNMENT + HandelC.SPACE 
						+ HandelC.MULTI_SYNC_READY
						+ HandelC.PAR_BEGIN
						+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.SET_UNION  
						+ HandelC.PAR_BEGIN
						+ HandelC.CURRENT_SYNC_STATUS_LUT	
						+ HandelC.ARRAY_BEGIN 
						+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
						+ HandelC.ARRAY_END
						+ HandelC.SEP_ARGUMENTS
						+ HandelC.getInstance().getSingletonSetsLookupTableName(HandelC.BRANCH)
						+ HandelC.ARRAY_BEGIN 
						+ HandelC.getInstance().getArgName(HandelC.WRITER_BRANCH, HandelC.BRANCH_ARG_B) 
						+ HandelC.ARRAY_END
						+ HandelC.PAR_END
						+ HandelC.PAR_END
    					+ HandelC.END_COMMAND;
    	this.hcCode = readersReady + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_READERS_READY) + this.hcCode;
    }
    
    private void declareMultiSyncReady() {
        /*
		macro expr MULTI_SYNC_READY(CHANNEL_C, BRANCH_SET_S) = 
		    (MULTI_SYNC_LUT[CHANNEL_C][0] == BRANCH_SET_S)
		    ||
		    ...
		    ||
		    (MULTI_SYNC_LUT[CHANNEL_C][n] == BRANCH_SET_S);
         */
    	String multiSyncReady = HandelC.MACRO_EXP + HandelC.SPACE + HandelC.MULTI_SYNC_READY 
						+ HandelC.PAR_BEGIN 
						+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
						+ HandelC.SEP_ARGUMENTS 
						+ HandelC.getInstance().getArgName(HandelC.BRANCH_SET, HandelC.SET_ARG_S) 
						+ HandelC.PAR_END + HandelC.SPACE 
						+ HandelC.ASSIGNMENT + HandelC.COMMENT_LINE_BREAK;
		for (int i=0; i < this.parallelismAnalyser.getMaxMultiSyncSetCard(); i++) {
			multiSyncReady = 
				multiSyncReady 
				+ HandelC.SPACE 
				+ HandelC.PAR_BEGIN
				+ HandelC.MULTI_SYNC_LUT  
				+ HandelC.ARRAY_BEGIN 
				+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) + HandelC.ARRAY_END
				+ HandelC.ARRAY_BEGIN + i + HandelC.ARRAY_END
				+ HandelC.REL_EQUALS 
				+ HandelC.getInstance().getArgName(HandelC.BRANCH_SET, HandelC.SET_ARG_S)
				+ HandelC.PAR_END;
			if (i != this.parallelismAnalyser.getMaxMultiSyncSetCard() -1 ) {
				multiSyncReady = multiSyncReady 
								+ HandelC.SPACE + HandelC.COMMENT_LINE_BREAK  
								+ Messenger.TAB + HandelC.LOGICAL_OR 
								+ HandelC.SPACE + HandelC.COMMENT_LINE_BREAK + HandelC.SPACE;
			}
		}
		
		multiSyncReady = multiSyncReady + HandelC.END_COMMAND;
		this.hcCode = multiSyncReady + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_MULTI_SYNC_READY) + this.hcCode;
    }

    private void declareNoWriter() {
        /*
		macro expr NO_WRITER(CHANNEL_C) = (!HAS_WRITER_LUT[CHANNEL_C]);
         */
    	String noWriter = HandelC.MACRO_EXP + HandelC.SPACE + HandelC.NO_WRITER 
			+ HandelC.PAR_BEGIN + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
			+ HandelC.PAR_END + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE
			+ HandelC.PAR_BEGIN
			+ HandelC.LOGICAL_NOT  
			+ HandelC.HAS_WRITER_LUT
			+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
			+ HandelC.ARRAY_END 
			+ HandelC.PAR_END + HandelC.END_COMMAND;
    	this.hcCode = noWriter + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_HAS_WRITER) + this.hcCode;
    }
    
    /*
    private void declareNoReaders() {
		macro expr NO_READERS(CHANNEL_C, WRITER_BRANCH_B) = 
		    (SET_DIFF(CURRENT_SYNC_STATUS_LUT[CHANNEL_C],BRANCH__set_LUT[WRITER_BRANCH_B])==0);
    	String noReaders = HandelC.MACRO_EXP + HandelC.SPACE + HandelC.NO_READERS 
			+ HandelC.PAR_BEGIN 
			+ HandelC.getInstance().getArgName(HandelC.WRITER_BRANCH, HandelC.BRANCH_ARG_B) 
			+ HandelC.SEP_ARGUMENTS
			+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)
			+ HandelC.PAR_END + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE
			+ HandelC.PAR_BEGIN
			+ HandelC.SET_DIFF
			+ HandelC.PAR_BEGIN
			+ HandelC.CURRENT_SYNC_STATUS_LUT
			+ HandelC.ARRAY_BEGIN
			+ HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C)
			+ HandelC.ARRAY_END
			+ HandelC.SEP_ARGUMENTS
			+ HandelC.getInstance().getSingletonSetsLookupTableName(HandelC.BRANCH)
			+ HandelC.ARRAY_BEGIN
			+ HandelC.getInstance().getArgName(HandelC.WRITER_BRANCH, HandelC.BRANCH_ARG_B) 
			+ HandelC.ARRAY_END
			+ HandelC.PAR_END
			+ HandelC.REL_EQUALS
			+ HandelC.MATH_ZERO
			+ HandelC.PAR_END + HandelC.END_COMMAND;
    	this.hcCode = noReaders + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_NO_READERS) + this.hcCode;
    }
     */
    
    private void declareIsAvailable() {
        /*
         * macro expr IS_AVAILABLE(BRANCH_B, CHANNEL_C) = 
         * 	(IS_EMPTY_SET(SET_INTER(FORCED_SYNC_LUT[CHANNEL_C][BRANCH_B],CURRENT_SYNC_STATUS[CHANNEL_C])));
         */
    	String isAvailable = HandelC.MACRO_EXP + HandelC.SPACE + HandelC.IS_AVAILABLE 
							+ HandelC.PAR_BEGIN + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) 
							+ HandelC.SEP_ARGUMENTS + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) 
							+ HandelC.PAR_END + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE
							+ HandelC.PAR_BEGIN
							+ HandelC.IS_EMPTY_SET + HandelC.PAR_BEGIN
							+ HandelC.SET_INTER + HandelC.PAR_BEGIN 
							+ HandelC.FORCED_SYNC_LUT 
							+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) + HandelC.ARRAY_END 
							+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getArgName(HandelC.BRANCH, HandelC.BRANCH_ARG_B) + HandelC.ARRAY_END + HandelC.SEP_ARGUMENTS
							+ HandelC.CURRENT_SYNC_STATUS_LUT
							+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getArgName(HandelC.CHANNEL, HandelC.CHANNEL_ARG_C) + HandelC.ARRAY_END 
							+ HandelC.PAR_END + HandelC.PAR_END  + HandelC.PAR_END  + HandelC.END_COMMAND;
		this.hcCode = isAvailable + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MACRO_IS_AVAILABLE) + this.hcCode;
    }
    
    private void declareConstantsAndFunctions(){
        this.hcCode = this.constants.getHcCode() + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.CONSTANTS_AND_FUNCTIONS) + this.hcCode;
        this.declareConstantsAndFunctionsPrototypes();
    }

    private void declareConstantsAndFunctionsPrototypes(){
        Enumeration keys = this.constants.getConstantsNames();
        
        while(keys.hasMoreElements()) {
            String name = (String)keys.nextElement();
            List args = this.constants.getArguments(name);
            
            // macro expr name = expression;
            String macro = HandelC.MACRO_EXP + HandelC.SPACE + name;
            
            if (args.size() > 0) {
                macro = macro + HandelC.PAR_BEGIN;
                for(int i=0; i < args.size(); i++){
                   macro = macro + args.get(i);
                   if (i < args.size() - 1) {
                       macro = macro + HandelC.SEP_ARGUMENTS;
                   }
                }
                macro = macro + HandelC.PAR_END;
            }
            macro = macro + HandelC.END_COMMAND;
            this.hcCode = macro + this.hcCode;
        }
        
        this.hcCode = HandelC.getInstance().generateComment(Messenger.CONSTANTS_AND_FUNCTIONS_PROTOTYPES) + this.hcCode;
    }
    
    
    private void declareMutualRecConstants(){
        for (int i=0; i<this.processesIds.size(); i++) {
        	String pname = this.processesIds.get(i);
        	if (!pname.equals(HandelC.MAINP)) {
	            String nextLine = HandelC.getInstance().generateDefine(this.processesIds.get(i) + HandelC.SPACE + i);
	            this.hcCode = nextLine + this.hcCode;
        	}
        }
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MUTUAL_REC_STATE_DECL) + this.hcCode;
    }
    
    private void declareAuxiliaryMultiSyncMacros() {
    	String flipFunction = HandelC.MACRO_EXP + HandelC.SPACE 
	        + "FLIP" + HandelC.PAR_BEGIN
	        + "x" + HandelC.PAR_END
	        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
	        + HandelC.MATH_NEG
	        + HandelC.PAR_BEGIN
	        + "x"
	        + HandelC.SPACE + HandelC.MATH_SUM + HandelC.SPACE
	        + "1" 
	        + HandelC.PAR_END
	        + HandelC.END_COMMAND;
    	this.hcCode = flipFunction + this.hcCode;
    	this.hcCode = HandelC.getInstance().generateComment("Auxiliary macros on Multi-sync") + this.hcCode;
    }
    
    private List<String> declareMultiSyncClients (Node node){
    	
    	List <String> prototypes = new Vector<String>();
    	List <Integer> withdrawAfterMS = new Vector<Integer>();
    	List <Integer> withdrawAfterNotMS = new Vector<Integer>();
    	for (List <Integer> client: this.multisyncs.getClients()){
    		if (client.size() == 3) { // ( channelsInMultiSync , interruptInMultiSyncInput , interruptInMultiSyncOutput)
    			int msChannels = client.get(0);
    			int inputInterrupts = client.get(1);
    			int outputInterrupts = client.get(2);
    			
    	    	prototypes.addAll(this.declareMultiSyncClient(msChannels,inputInterrupts, outputInterrupts));
    			if (!withdrawAfterMS.contains(msChannels)
    					&& msChannels > 1){
    				withdrawAfterMS.add(msChannels);
    			}
    			if ((outputInterrupts > 0 || inputInterrupts > 0)
    					&& !withdrawAfterNotMS.contains(msChannels)){
    				withdrawAfterNotMS.add(msChannels);
    			}
    		}
    	}
    	
    	for(int length:withdrawAfterMS){
    		prototypes.add(this.declareMultiSyncWithdraw(length,true));
    	}
    	for(int length:withdrawAfterNotMS){
    		prototypes.add(this.declareMultiSyncWithdraw(length,false));
    	}
    	
    	return prototypes;
    }
    
    private String declareMultiSyncWithdraw(int length, boolean isAfterMS) {
    	String prototype = ""; 
    	if (length > 0) {
    		prototype =
    			"inline void withdraw"+ length +"After" + 
    			(isAfterMS?"":"Not") +
    			"Sync (psChannelOfMultiSync channels" +
    			(isAfterMS?", integer_MS_CLIENT excluded)":")");
    		
    		String decl =
    			"par{ {"
    			+"sMessageValue message_value;"
    			+"message_value.M = FLIP(((integer_message)(0 @ (channels"
    			+(length > 1?"+index":"")
    			+")->id)));"
    			+(length > 1?"(channels+index)->to_sync[(channels+index)->id]":"(channels)->to_sync[(channels)->id]")
    			+"!message_value; } {"
    			+"sBoolValue boolean_value;"
    			+(length > 1?"(channels+index)->from_sync[(channels+index)->id]":"(channels)->from_sync[(channels)->id]")
    			+"?boolean_value;"
    			+"if (boolean_value.B == true) {"
    			+(length > 1?"(channels+index)->from_sync[(channels+index)->id]":"(channels)->from_sync[(channels)->id]")
    			+"?boolean_value; } } }";
    		
    		if (length > 1){
    			this.hcCode = 
    				prototype + HandelC.BLOCK_BEGIN + 
    				"integer_MS_CLIENT index;" +
    				"par (index = 0; index < " + length + "; index++){" +
    				(isAfterMS?"if (index == excluded) { delay; } else {" + decl + "}":decl) +
    				"} }" + this.hcCode;
    		} else {
    			this.hcCode = prototype + HandelC.BLOCK_BEGIN + 
    			decl + HandelC.BLOCK_END + this.hcCode;
    		}
    		prototype = prototype + HandelC.END_COMMAND; 
    	}
    	
		return prototype;
	}

	private List<String> declareMultiSyncClient (int cardMSync, int cardMSInterruptIn, int cardMSInterruptOut){
    	List <String> prototypes = new Vector<String>();
    	String prototype;
    	String syncCard = Integer.toString(cardMSync) + "x" + cardMSInterruptIn + "x" + cardMSInterruptOut;
    	
    	String client = ""; 
    	if (cardMSync>1) {
    		client = client +
    			"sMessageValue Local_M["+cardMSync+"];"
	    		+"unsigned "+ MathUtil.getUnsignedBitsForNumber(cardMSync) +" i; "
	    		+"par (i=0; i"+HandelC.REL_LESS + cardMSync + HandelC.END_COMMAND+" i++) {"
	    		+"{"
	    		+"Local_M[i].M = (integer_message)(0 @ (seqOfSync+i)->id);"
	    		+"(seqOfSync+i)->to_sync[(seqOfSync+i)->id]!Local_M[i];"
	    		+"};"
	    		+"}"
	    		+"return sync" + HandelC.UNDERSCORE + syncCard + "(seqOfSync"
		    	+(cardMSInterruptIn>0 ? ", seqOfInputNotSync":"")
		    	+(cardMSInterruptOut>0 ? ", seqOfOutputNotSync":"")
		    	+");"
	    		+"}\n";
    	} else {
    		client = client +
	    		"sMessageValue Local_M;"
	    		+"Local_M.M = (integer_message)(0 @ (seqOfSync)->id);"
	    		+"(seqOfSync)->to_sync[(seqOfSync)->id]!Local_M;"
	    		+"return sync" + HandelC.UNDERSCORE + syncCard + "(seqOfSync"
		    	+(cardMSInterruptIn>0 ? ", seqOfInputNotSync":"")
		    	+(cardMSInterruptOut>0 ? ", seqOfOutputNotSync":"")
		    	+");"
	    		+"}\n";
    	}
    	
    	prototype = "inline integer_MS_CLIENT MultiSyncClient" + HandelC.UNDERSCORE + syncCard + "("
	    	+"psChannelOfMultiSync seqOfSync"
	    	+(cardMSInterruptIn>0 ? ", psInterruptOfMultiSync seqOfInputNotSync":"")
	    	+(cardMSInterruptOut>0 ? ", psInterruptOfMultiSync seqOfOutputNotSync":"")
	    	+")";
    	client = prototype + HandelC.BLOCK_BEGIN + client;
    	prototypes.add(prototype+HandelC.END_COMMAND);
    	
    	prototype = 
    		"inline integer_MS_CLIENT sync" + HandelC.UNDERSCORE + syncCard + "("
			+"psChannelOfMultiSync seqOfSync"
			+(cardMSInterruptIn>0 ? ", psInterruptOfMultiSync seqOfInputNotSync":"")
	    	+(cardMSInterruptOut>0 ? ", psInterruptOfMultiSync seqOfOutputNotSync":"")
	    	+")";
    	
    	/*int chosenId_bits =
			(cardMSync>cardMSInterruptIn?
					((cardMSync>cardMSInterruptOut)?
							(MathUtil.getUnsignedBitsForNumber(cardMSync)):
							(MathUtil.getUnsignedBitsForNumber(cardMSInterruptOut))
					):
					((cardMSInterruptIn>cardMSInterruptOut)?
							(MathUtil.getUnsignedBitsForNumber(cardMSInterruptIn)):
							(MathUtil.getUnsignedBitsForNumber(cardMSInterruptOut))
					)
					);*/
		
    	String sync = prototype 
			+"{"
			+"boolean KEEP_LOOPING;"
			/*+"unsigned "+(int)Math.floor((Math.log(cardMSync)/Math.log(2.0))+1)+" channelsInSync;"*/
			+ "integer_MS_CLIENT channelsInSync;"
			+"sBoolValue boolean_value;"
			+"sMessageValue message_value;"
			+"boolean chosenInSeqOfSync"
			+(cardMSInterruptIn>0 ? ", chosenInputNotSync":"")
			+(cardMSInterruptOut>0 ? ", chosenOutputNotSync":"")
			+";"
			+"integer_MS_CLIENT" /*+"unsigned "+ chosenId_bits*/
			+" chosenId;"
			+"channelsInSync = "+cardMSync+";"
			+"KEEP_LOOPING = true;"
			+"while (KEEP_LOOPING){"
				+"KEEP_LOOPING = false;";
		    	
		    	String syncPrialt = "";
		    	if ((cardMSync + cardMSInterruptIn + cardMSInterruptOut) > 1) {
		    		for (int i = 0; i < cardMSync; i++) {
		    			syncPrialt +=
		    				"case (seqOfSync + " + i + ")->from_sync[(seqOfSync + " + i + ")->id]?boolean_value :{"
		    				+ "chosenInSeqOfSync = true;"
		    				+ "chosenId = " + i + HandelC.END_COMMAND
		    				+ HandelC.BLOCK_END
		    				+ HandelC.END_COMMAND
		    				+ "break;";
		    		}
		    		for (int i = 0; i < cardMSInterruptIn; i++) {
		    			syncPrialt +=
		    				"case *((seqOfInputNotSync + " + i + ")->interrupt)?(  (seqOfInputNotSync + " + i + ")-> value) :{"
		    				+ "chosenInputNotSync = true;"
		    				+ "chosenId = " + i + ";"
		    				+ HandelC.BLOCK_END
		    				+ HandelC.END_COMMAND
		    				+ "break;";
		    		}				
		    		for (int i = 0; i < cardMSInterruptOut; i++) {
		    			syncPrialt +=
		    				"case *((seqOfOutputNotSync + " + i + ")->interrupt)! ( (seqOfOutputNotSync + " + i + ")-> value) :{"
		    				+ "chosenOutputNotSync = true;"
		    				+ "chosenId = " + i + ";"
		    				+ HandelC.BLOCK_END
		    				+ HandelC.END_COMMAND
		    				+ "break;";
		    		}
		    		syncPrialt =
		    			"chosenInSeqOfSync = false;"
						+ (cardMSInterruptIn>0 ? "chosenInputNotSync = false;":"")
						+ (cardMSInterruptOut>0 ? "chosenOutputNotSync = false;":"")
		    			+ "prialt{" + syncPrialt + "	};";
		    	} else {
		    		syncPrialt = 
		    			"seqOfSync->from_sync[seqOfSync->id]?boolean_value;"
	    				+ "chosenInSeqOfSync = true;"
	    				+ "chosenId = 0" + HandelC.END_COMMAND;
		    	}
		    	
			sync += syncPrialt
				+"if (chosenInSeqOfSync == true) {"
					+"unsigned "+MathUtil.getUnsignedBitsForNumber(cardMSync)+" iOfWithdraw;"
					+"sBoolValue * synchronized;"
					+"sMessageValue * myId;"
					+"synchronized = &boolean_value;"
					+"myId = &message_value;"
					+"(*myId).M = (integer_message)(0 @ (seqOfSync+chosenId)->id);"
					+"(seqOfSync+chosenId)->to_sync[(seqOfSync+chosenId)->id]!(*myId);"
					+"(seqOfSync+chosenId)->from_sync[(seqOfSync+chosenId)->id]?(*synchronized);"
					+"if ((*synchronized).B == true) {"
						+"			seq {"
						+"					if ((seqOfSync+chosenId)->id == MULTI_SYNC_WRITER) {"
						+"						(*myId).MV = ((seqOfSync+chosenId)->value);"
						+"						(seqOfSync+chosenId)->to_sync[(seqOfSync+chosenId)->id]!(*myId);"
						+"					} else {"
						+"						sBoolValue * receivedValue;"
						+"						receivedValue = &boolean_value;"
						+"						(seqOfSync+chosenId)->from_sync[(seqOfSync+chosenId)->id]?(*receivedValue);"
						+"						(seqOfSync+chosenId)->value = (*receivedValue).BV;"
						+"					};"
						+ (cardMSync > 1 ? "withdraw"+ cardMSync +"AfterSync(seqOfSync, chosenId);" :"")
						+"			}"
					+"} else{"
						+"KEEP_LOOPING = true;"
					+"};"
					+""
				+"}";
			
			if (cardMSInterruptIn > 0 || cardMSInterruptOut > 0){
				sync +=	
					" else if ("
					+ ((cardMSInterruptIn > 0)?"chosenInputNotSync == true":"")
					+ ((cardMSInterruptIn > 0) && (cardMSInterruptOut > 0)?" || ":"")
					+ ((cardMSInterruptOut > 0)?"chosenOutputNotSync == true":"")
					+ ") {"
					+"		withdraw"+ cardMSync +"AfterNotSync(seqOfSync);"
					+"	};";
				
				int MS_CLIENT_bits = MathUtil.getUnsignedBitsForNumber(this.multisyncs.getLargerClient());
				sync +=	"}; " +
							"if (chosenInSeqOfSync == true){" +
								/*"return " + (MS_CLIENT_bits > chosenId_bits?"(integer_MS_CLIENT)(0 @ (chosenId))":"(chosenId)") +"; }" +*/
								"return " + "(chosenId)" +"; }" +
							/*((cardMSInterruptIn > 0)?" else if (chosenInputNotSync == true){ return " + (MS_CLIENT_bits > chosenId_bits?"(integer_MS_CLIENT)(0 @ (chosenId + channelsInSync))":"(chosenId + channelsInSync)") +";}":"") +*/
							((cardMSInterruptIn > 0)?" else if (chosenInputNotSync == true){ return " + "(chosenId + channelsInSync)" +";}":"") +
							/*((cardMSInterruptOut > 0)?" else { return " + (MS_CLIENT_bits > chosenId_bits?"(integer_MS_CLIENT)(0 @ (chosenId + channelsInSync + " + cardMSInterruptIn + "))":"(chosenId + channelsInSync + " + cardMSInterruptIn + ")") +"; }":"") +*/
							((cardMSInterruptOut > 0)?" else { return " + "(chosenId + channelsInSync + " + cardMSInterruptIn + ")" +"; }":"") +
						" }";
		
			} else {
				int MS_CLIENT_bits = MathUtil.getUnsignedBitsForNumber(this.multisyncs.getLargerClient());
				
				sync +=	"}; " +
							/*"return " + (MS_CLIENT_bits > chosenId_bits?"(integer_MS_CLIENT)(0 @ (chosenId))":"(chosenId)") +";" +*/
							"return (chosenId);" +
						" }";
				
			}
			
		prototypes.add(prototype + HandelC.END_COMMAND);
    	this.hcCode = "\n" + client + "\n" + sync + "\n" + this.hcCode;
    	return prototypes;
    }
    
    private List<String> declareMultiSyncClient_OLD (int cardSeqSync, int cardSeqNotSync){
    	List <String> prototypes = new Vector<String>();
    	String prototype;
    	String syncCard = Integer.toString(cardSeqSync) + "x" + cardSeqNotSync;
    	
    	String client = ""; 
    	if (cardSeqSync>1) {
    		client = client +
    			"sMessageValue Local_M["+cardSeqSync+"];"
    			+"unsigned "+ MathUtil.getUnsignedBitsForNumber(cardSeqSync) +" i; "
	    		+"par (i=0; i"+HandelC.REL_LESS + cardSeqSync + HandelC.END_COMMAND+" i++) {"
	    		+"{"
	    		+"Local_M[i].M = (integer_message)(0 @ (seqOfSync+i)->id);"
	    		+"(seqOfSync+i)->to_sync[(seqOfSync+i)->id]!Local_M[i];"
	    		+"};"
	    		+"}"
	    		+"return sync" + HandelC.UNDERSCORE + syncCard + "(seqOfSync,"
		    	+(cardSeqNotSync>0 ? "seqOfNotSync,":"")
		    	+"value);"
	    		+"}\n";
    	} else {
    		client = client +
	    		"sMessageValue Local_M;"
	    		+"Local_M.M = (integer_message)(0 @ (seqOfSync)->id);"
	    		+"(seqOfSync)->to_sync[(seqOfSync)->id]!Local_M;"
	    		+"return sync" + HandelC.UNDERSCORE + syncCard + "(seqOfSync,"
		    	+(cardSeqNotSync>0 ? "seqOfNotSync,":"")
		    	+"value);"
	    		+"}\n";
    	}
    	
    	prototype = "inline integer_MS_CLIENT MultiSyncClient" + HandelC.UNDERSCORE + syncCard + "("
	    	+"psChannelOfMultiSync seqOfSync,"
	    	+(cardSeqNotSync>0 ? "psInterruptOfMultiSync seqOfNotSync,":"")
	    	+"MSYNC_VALUE * value)";
    	client = prototype + HandelC.BLOCK_BEGIN + client;
    	prototypes.add(prototype+HandelC.END_COMMAND);
    	
    	prototype = 
    		"inline integer_MS_CLIENT sync" + HandelC.UNDERSCORE + syncCard + "("
			+"psChannelOfMultiSync seqOfSync,"
			+(cardSeqNotSync>0 ? "psInterruptOfMultiSync seqOfNotSync,":"")
			+"MSYNC_VALUE * value)";
    	String sync = prototype 
			+"{"
			+"boolean KEEP_LOOPING;"
			+"unsigned "+(int)Math.floor((Math.log(cardSeqSync)/Math.log(2.0))+1)+" channelsInSync;"
			+"sBoolValue boolean_value;"
			+"sMessageValue message_value;"
			+"boolean chosenInSeqOfSync"
			+(cardSeqNotSync>0 ? ",chosenInSeqOfNotSync;":";")
			+"unsigned "
			+(cardSeqSync>cardSeqNotSync?
					(int)Math.floor((Math.log(cardSeqSync)/Math.log(2.0))+1):
					(int)Math.floor((Math.log(cardSeqNotSync)/Math.log(2.0))+1))
			+" chosenId;"
			+"channelsInSync = "+cardSeqSync+";"
			+"KEEP_LOOPING = true;"
			+"while (KEEP_LOOPING){"
				+"KEEP_LOOPING = false;";
		    	
		    	String syncPrialt = "";
		    	if ((cardSeqSync + cardSeqNotSync) > 1) {
		    		for (int i = 0; i < cardSeqSync; i++) {
		    			syncPrialt +=
		    				"case (seqOfSync + " + i + ")->from_sync[(seqOfSync + " + i + ")->id]?boolean_value :{"
		    				+ "chosenInSeqOfSync = true;"
		    				+ "chosenId = " + i + HandelC.END_COMMAND
		    				+ HandelC.BLOCK_END
		    				+ HandelC.END_COMMAND
		    				+ "break;";
		    		}
		    		for (int i = 0; i < cardSeqNotSync; i++) {
		    			syncPrialt +=
		    				"case *((seqOfNotSync + " + i + ")->interrupt)?syncin :{"
		    				+ "chosenInSeqOfNotSync = true;"
		    				+ "chosenId = " + i + ";"
		    				+ HandelC.BLOCK_END
		    				+ HandelC.END_COMMAND
		    				+ "break;";
		    		}				
		    		syncPrialt =
		    			"chosenInSeqOfSync = false;"
						+ (cardSeqNotSync>0 ? "chosenInSeqOfNotSync = false;":"")
		    			+ "prialt{" + syncPrialt + "	};";
		    	} else {
		    		syncPrialt = 
		    			"seqOfSync->from_sync[seqOfSync->id]?boolean_value;"
	    				+ "chosenInSeqOfSync = true;"
	    				+ "chosenId = 0" + HandelC.END_COMMAND;
		    	}
		    	
			sync += syncPrialt
				+"if (chosenInSeqOfSync == true) {"
					+"unsigned "+(int)Math.floor((Math.log(cardSeqSync)/Math.log(2.0))+1)+" iOfWithdraw;"
					+"sBoolValue * synchronized;"
					+"sMessageValue * myId;"
					+"synchronized = &boolean_value;"
					+"myId = &message_value;"
					+"(*myId).M = (integer_message)(0 @ (seqOfSync+chosenId)->id);"
					+"(seqOfSync+chosenId)->to_sync[(seqOfSync+chosenId)->id]!(*myId);"
					+"(seqOfSync+chosenId)->from_sync[(seqOfSync+chosenId)->id]?(*synchronized);"
					+"if ((*synchronized).B == true) {"
						+"			seq {"
						+"				{"	
						+"					if (BRANCH__set_LUT[(seqOfSync+chosenId)->id] == MULTI_SYNC_WRITER) {"
						+"						(*myId).MV = (*value);"
						+"						(seqOfSync+chosenId)->to_sync[(seqOfSync+chosenId)->id]!(*myId);"
						+"					} else {"
						+"						sBoolValue * receivedValue;"
						+"						receivedValue = &boolean_value;"
						+"						(seqOfSync+chosenId)->from_sync[(seqOfSync+chosenId)->id]?(*receivedValue);"
						+"						(*value) = (*receivedValue).BV;"
						+"					}"
						+"				};"
						+"				{"
						+"					iOfWithdraw = 0;"
						+"					while (iOfWithdraw < channelsInSync){"
						+"						if (iOfWithdraw != chosenId) {"
						+"							par{"
						+"								{"
						+"									sMessageValue * myId;"
						+"									myId = &message_value;"
						+"									(*myId).M = FLIP(((integer_message)(0 @ (seqOfSync+iOfWithdraw)->id)));"
						+"									(seqOfSync+iOfWithdraw)->to_sync[(seqOfSync+iOfWithdraw)->id]!(*myId);"
						+"								};"
						+"								{"
						+"									sBoolValue * invite;"
						+"									invite = &boolean_value;"
						+"									(seqOfSync+iOfWithdraw)->from_sync[(seqOfSync+iOfWithdraw)->id]?(*invite);"
						+"									if ((*invite).B == true) {"
						+"										(seqOfSync+iOfWithdraw)->from_sync[(seqOfSync+iOfWithdraw)->id]?(*invite);"
						+"									}"
						+"								}"
						+"							};"
						+"						};"
						+"						iOfWithdraw++;"
						+"					};					"
						+"				};"
						+"			}"
					+"} else{"
						+"KEEP_LOOPING = true;"
					+"};"
					+""
				+"}";
			
			if (cardSeqNotSync > 0) {
				sync +=	
					" else if (chosenInSeqOfNotSync == true) {"
					+"		unsigned "+(int)Math.floor((Math.log(cardSeqSync)/Math.log(2.0))+1)+" iOfWithdraw;"
					+"		iOfWithdraw = 0;"
					+"		while (iOfWithdraw < channelsInSync){"
					+"			par{"
					+"				{"
					+"					sMessageValue * myId;"
					+"					myId = &message_value;"
					+"					(*myId).M = FLIP(((integer_message)(0 @ (seqOfSync+iOfWithdraw)->id)));"
					+"					(seqOfSync+iOfWithdraw)->to_sync[(seqOfSync+iOfWithdraw)->id]!(*myId);"
					+"				};"
					+"				{"
					+"					sBoolValue * invite;"
					+"					invite = &boolean_value;"
					+"					(seqOfSync+iOfWithdraw)->from_sync[(seqOfSync+iOfWithdraw)->id]?(*invite);"
					+"					if ((*invite).B == true) {"
					+"						(seqOfSync+iOfWithdraw)->from_sync[(seqOfSync+iOfWithdraw)->id]?(*invite);"
					+"					}"
					+"				}"
					+"			};"
					+"			iOfWithdraw++;"
					+"		};	"
					+"	};";
			}
			
			int differenceInSize = MathUtil.getUnsignedBitsForNumber(this.multisyncs.getLargerClient()) 
										- MathUtil.getUnsignedBitsForNumber(java.lang.Math.max(cardSeqSync,cardSeqNotSync)); 
			sync +=	"}; " +
						"if (chosenInSeqOfSync == true){" +
							"return " + (differenceInSize > 0?"(integer_MS_CLIENT)(0 @ (chosenId))":"(chosenId)") +";" +
						"} else return " + (differenceInSize > 0?"(integer_MS_CLIENT)(0 @ (chosenId + channelsInSync))":"(chosenId + channelsInSync)") +";" +
					" }";
		
			
		prototypes.add(prototype + HandelC.END_COMMAND);
    	this.hcCode = "\n" + client + "\n" + sync + "\n" + this.hcCode;
    	return prototypes;
    }

    private List<String> declareMultiSyncControls (){
    	List <String> prototypes = new Vector<String>();
    	prototypes.addAll(this.declareMultiSyncControl_Aux());
    	List <Integer> controls = this.parallelismAnalyser.getListOfControls();
    	ListIterator<Integer> controlsIter = controls.listIterator();
    	while (controlsIter.hasNext()){
    		int control = controlsIter.next();
    		prototypes.addAll(this.declareMultiSyncControl(control));
    	}
    	this.hcCode = HandelC.getInstance().generateComment("Processes MultiSyncControl") + this.hcCode;
    	return prototypes;
    }

    private Hashtable<Integer,List<Integer>> versionsOfMultiSyncClient (Node node){
    	
    	return new ExtendedDepthFirstAdapter() {
    		Hashtable<Integer,List<Integer>> clients;
            int countChanInMultiSync, countChanInterrupt = 0;
            boolean isExternalChoice = false;
            boolean checkedPrefix = false;
            ParallelismAnalyserVisitor parallelAnalyser;
            
            public Hashtable<Integer,List<Integer>> getClients(Node node, ParallelismAnalyserVisitor parallelAnalyser) {
            	this.parallelAnalyser = parallelAnalyser;
            	this.clients = new Hashtable<Integer, List<Integer>>();
                node.apply(this);
                return this.clients;
            }

            public void caseACspPrefixProcess(ACspPrefixProcess node) {
            	PEvent event = node.getEvent();
            	String channelName = ((ACspEvent)event).getChannelName().toString().trim();
 
          		if (this.parallelAnalyser.isMultiSyncChannel(channelName)){
          			List<Integer> client;
        			if (this.clients.containsKey(1)){
        				client = this.clients.get(1);
        				if (!client.contains(0)){
        					client.add(0);
        					this.clients.put(1, client);
        				}
        			} else {
        				client = new Vector <Integer> ();
        				client.add(0);
        				this.clients.put(1,client);
        			}
        		}
        		this.checkedPrefix = false;
        		node.getProcess().apply(this);
            }

            public void caseACspExternalChoiceProcess(ACspExternalChoiceProcess node) {
            	InitialsExtractorVisitor initialsExtractor = new InitialsExtractorVisitor();
                node.apply(initialsExtractor);
                List keys = initialsExtractor.keyList();
                
                this.countChanInMultiSync = 0;
        		this.countChanInterrupt = 0;
                for(int i=0; i < keys.size(); i++) {
                    PEvent event = (PEvent)keys.get(i);
                    String channelName = ((ACspEvent)event).getChannelName().toString().trim();
                    if (this.parallelAnalyser.isMultiSyncChannel(channelName))
            			this.countChanInMultiSync++;
            		else this.countChanInterrupt++;
                }
                if (this.countChanInMultiSync > 0) {
          			List<Integer> client;
        			if (this.clients.containsKey(this.countChanInMultiSync)){
        				client = this.clients.get(this.countChanInMultiSync);
        				if (!client.contains(this.countChanInterrupt)){
        					client.add(this.countChanInterrupt);
        					this.clients.put(this.countChanInMultiSync, client);
        				}
        			} else {
        				client = new Vector <Integer> ();
        				client.add(this.countChanInterrupt);
        				this.clients.put(this.countChanInMultiSync,client);
        			}
            	}
                for(int i=0; i < keys.size(); i++) {
                	PEvent event = (PEvent)keys.get(i);
                	Node process = initialsExtractor.getProcess(event);
                    process.apply(this);
                }
            }
        }.getClients(node, parallelismAnalyser);
    }
    
    private List<String> declareMultiSyncControl (int numberOfControl){
    	List <String> prototypes = new Vector<String>();
    	String prototype;
    	
    	String control = 
    			"MSYNC_ID numberOfClients;"
    				+ "boolean KEEP_LOOPING;"
					+ "MSYNC_ID Local_countdown;"
					+ "numberOfClients = " + numberOfControl + ";"
					+ "Local_countdown = numberOfClients;"
					+ "KEEP_LOOPING = true;"
					+ "while(KEEP_LOOPING){"
						+ "if (Local_countdown > 0 && Local_countdown <= numberOfClients) {"
							+ "sMessageValue nextoffer;"
							+ "prialt {";
    	
    	for (int i =0; i<numberOfControl; i++){
    		control += "case to_sync[" + i + "]?nextoffer :{"
						+ "delay;"
						+ "};break;";
			
    	}

    	control +=					"case *endController?syncin :{"
									+ "KEEP_LOOPING = false;"
									+ "delay;"
								+ "};break;"
							+ "};"
							+ "if (KEEP_LOOPING) {"
								+ "if (nextoffer.M >= 0) {"
									+ "Local_countdown = Local_countdown - 1;"
								+ "} else {"
									+ "sBoolValue response;"
									+ "MSYNC_ID nextflipped;"
									+ "nextflipped = (MSYNC_ID)(FLIP(nextoffer.M) <- bits_of_MSYNC_ID);"
									+ "response.B = false;"
									+ "from_sync[nextflipped]!response;"
									+ "Local_countdown = Local_countdown + 1;"
								+ "}"
							+ "}"
						+ "} else {"
							+ "Local_countdown = inviteMultiSyncControl" + numberOfControl + "(0, to_sync, from_sync);"
						+ "}"
					+ "}"
				+ "}";
    	prototype = "inline void MultiSyncControl" + numberOfControl + "("
			+ "chan sMessageValue * to_sync,"
			+ "chan sBoolValue * from_sync,"
			+ "chan SYNC * endController)";
    	control = prototype + HandelC.BLOCK_BEGIN + control;
		prototypes.add(prototype + HandelC.END_COMMAND);
    	
    	String invite = 
    			"MSYNC_ID numberOfClients;"
    			+ "boolean KEEP_LOOPING;"
				+ "numberOfClients = "+ numberOfControl +";"
				+ "KEEP_LOOPING = true;"
				+ "while(KEEP_LOOPING){"
					+ "KEEP_LOOPING = false;"
					+ "if (Local_i < numberOfClients){"
						+ "sBoolValue confirm_var;"
						+ "confirm_var.B = true;"
						+ "from_sync[Local_i]!confirm_var;"
						+ "Local_i = Local_i + 1;"
						+ "KEEP_LOOPING = true;"
					+ "} else {"
						+ "KEEP_LOOPING = false;"
					+ "}"
				+ "}"
				+ "return invitePrimeMultiSyncControl" + numberOfControl + "(0,to_sync,from_sync);"
			+ "}";
    	prototype = "inline MSYNC_ID inviteMultiSyncControl" + numberOfControl + "(MSYNC_ID Local_i,"
				     + "chan sMessageValue * to_sync, "
				     + "chan sBoolValue * from_sync)";
    	invite = prototype + HandelC.BLOCK_BEGIN + invite;
    	prototypes.add(prototype + HandelC.END_COMMAND);
			
    	String invitePrime = 
	    		"MSYNC_ID numberOfClients;"
				+ "boolean KEEP_LOOPING;"
				+ "MSYNC_ID Local_countdown;"
				+ "numberOfClients = "+ numberOfControl +";"
				+ "Local_countdown = numberOfClients;"
				+ "KEEP_LOOPING = true;"
				+ "while(KEEP_LOOPING){"
					+ "KEEP_LOOPING = false;"
					+ "if (Local_i >= 0 && Local_i < numberOfClients){"
						+ "sMessageValue nextcommit;"
						+ "prialt {";
    	
    	for (int i =0; i<numberOfControl; i++){
    		invitePrime += "case to_sync[" + i + "]?nextcommit :{"
							+ "delay;"
							+ "};break;";
			
    	}

    	invitePrime +=		"};"
						+ "if (nextcommit.M >= 0) {"
							+ "Local_i = Local_i + 1;"
							+ "Local_countdown = Local_countdown - 1;"
						+ "} else {"
							+ "Local_i = Local_i + 1;"
						+ "}"
						+ "KEEP_LOOPING = true;"
					+ "} else {"
						+ "KEEP_LOOPING = false;"
					+ "}"
				+ "}"
				+ "return syncMultiSyncControl(0,Local_countdown,numberOfClients,to_sync,from_sync);"
			+ "}";
    	prototype = "inline MSYNC_ID invitePrimeMultiSyncControl" + numberOfControl + "("
						+ "MSYNC_ID Local_i,"
						+ "chan sMessageValue * to_sync, "
						+ "chan sBoolValue * from_sync)";
    	invitePrime = prototype + HandelC.BLOCK_BEGIN + invitePrime;
    	prototypes.add(prototype + HandelC.END_COMMAND);
    	this.hcCode = control + invite + invitePrime + this.hcCode;
    	return prototypes;
    }
    	
    private List<String> declareMultiSyncControl_Aux (){
    	String prototype;
		List <String> prototypes = new Vector<String>();
		String sync = 
	    		"boolean KEEP_LOOPING;"
				+ "MSYNC_ID result;"
				+ "KEEP_LOOPING = true;"
				+ "while(KEEP_LOOPING){"
					+ "KEEP_LOOPING = false;"
					+ "if (Local_i < numberOfClients){"
						+ "sBoolValue confirm_var;"
						+ "confirm_var.B = (countdown==0);"
						+ "from_sync[Local_i]!confirm_var;"
						+ "if (countdown==0) {"
							+ "sMessageValue value;"
							+ "to_sync[Local_i]?value;"
							+ "result = syncVMultiSyncControl(Local_i + 1,countdown,value.MV,numberOfClients,from_sync);"
							+ "KEEP_LOOPING = false;"
						+ "} else {"
							+ "Local_i = Local_i + 1;"
							+ "KEEP_LOOPING = true;"
						+ "}"
					+ "} else {"
						+ "result = syncPrimeMultiSyncControl(countdown,numberOfClients);"
						+ "KEEP_LOOPING = false;"
					+ "}"
				+ "}"
				+ "return result;"
			+ "}";
    	
    	prototype = "inline MSYNC_ID syncMultiSyncControl("
						+ "MSYNC_ID Local_i,"
						+ "MSYNC_ID countdown,"
						+ "MSYNC_ID numberOfClients,"
						+ "chan sMessageValue * to_sync,"
						+ "chan sBoolValue * from_sync)";
		sync = prototype + HandelC.BLOCK_BEGIN + sync;
		prototypes.add(prototype + HandelC.END_COMMAND);
    	
    	String syncV =
    			"boolean KEEP_LOOPING;"
				+ "KEEP_LOOPING = true;"
				+ "while(KEEP_LOOPING){"
					+ "KEEP_LOOPING = false;"
					+ "if (Local_i < numberOfClients){"
						+ "sBoolValue pack_var;"
						+ "pack_var.B = (countdown==0);"
						+ "pack_var.BV = value;"
						+ "from_sync[Local_i]!pack_var;"
						+ "from_sync[Local_i]!pack_var;"
						+ "Local_i = Local_i + 1;"
						+ "KEEP_LOOPING = true;"
					+ "} else {"
						+ "KEEP_LOOPING = false;"
					+ "}"
				+ "}"
				+ "return syncPrimeMultiSyncControl(countdown,numberOfClients);"
			+ "}";
    	prototype = "inline MSYNC_ID syncVMultiSyncControl("
						+ "MSYNC_ID Local_i,"
						+ "MSYNC_ID countdown,"
						+ "MSYNC_VALUE value,"
						+ "MSYNC_ID numberOfClients, "
						+ "chan sBoolValue * from_sync)";
		syncV = prototype + HandelC.BLOCK_BEGIN + syncV;
		prototypes.add(prototype + HandelC.END_COMMAND);
			    	
    	String syncPrime =
    			"MSYNC_ID result;"
	    		+ "if (countdown > 0){"
	    			+ "result = countdown;"
	    		+ "} else {"
	    			+ "result = numberOfClients;"
	    		+ "}"
	    		+ "return result;"
	    	+ "}";
    	prototype = "inline MSYNC_ID syncPrimeMultiSyncControl(MSYNC_ID countdown, MSYNC_ID numberOfClients)";
		syncPrime = prototype + HandelC.BLOCK_BEGIN + syncPrime;
		prototypes.add(prototype + HandelC.END_COMMAND);
    	
    	this.hcCode = sync + syncV + syncPrime + this.hcCode;
    	return prototypes;
    }
    
    private void declareMultiSyncManagers (PExpression expression){
    	this.hcCode = this.hcCode + HandelC.getInstance().generateComment("Declaration of the controllers manager in Multi-sync");
    	this.declareControlsManager();
    	this.hcCode = this.hcCode + HandelC.getInstance().generateComment("Declaration of the processes manager in Multi-sync");
    	this.declareProcessManager(expression);
    }
    
    private void declareProcessManager(PExpression expression) {
		this.hcCode = this.hcCode + HandelC.INLINE + HandelC.SPACE + HandelC.VOID + HandelC.SPACE 
			+ "processManagerMultiSync" + HandelC.PAR_BEGIN + HandelC.PAR_END + HandelC.BLOCK_BEGIN;
		if (this.useBranchIds) {
            this.hcCode = this.hcCode + HandelC.BRANCH + HandelC.SPACE + HandelC.BRANCH_ID
            			+ HandelC.END_COMMAND;
            this.hcCode = this.hcCode + HandelC.BRANCH_ID + HandelC.SPACE + HandelC.ASSIGNMENT
							+ HandelC.SPACE + HandelC.MATH_ZERO + HandelC.END_COMMAND;
        }
        expression.apply(this);
        this.hcCode = this.hcCode+ "endSystem!syncout" + HandelC.END_COMMAND;
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
	}

	private void declareControlsManager() {
		String HANDELC_CHANNELMANAGER = "channelManager";
		String HANDELC_TO = "to";
		String HANDELC_FROM = "from";
		
		Set <String> channels = this.parallelismAnalyser.getMultiSyncChannels();
		Iterator<String> channel = channels.iterator();
		String hcChannelDeclaration = "", hcMultiSyncControl = "", hcChannelAcess = "";
		int countDecl = 0;
		while(channel.hasNext()){
			String channelName = channel.next();
			
			List channelType = null;
			int sizeChannelType = 0;
			String hcChannelType = "";
	        // Declaring the value communicated via this channel
	        if (this.channels.containsChannelType(channelName)) {
	            channelType = ListUtil.clone(this.channels.getChannelType(channelName));
	            sizeChannelType = channelType.size();
	            if (sizeChannelType> 0 && this.channels.isCommChannel(channelName)){
	            	sizeChannelType--;
	            }
	            if(this.channels.isProjChannel(channelName)){
	                for (int j = 0; j < sizeChannelType; j++) {
	                    PType type = (PType)channelType.get(j);
	                    
	                    String nodeTypeName = ((ACspType) type).getName().getText().trim(); 
	                	String hcType = HandelC.getInstance().getHandelCTypeName(nodeTypeName);
	                    String typeCard = HandelC.getInstance().getCardVarOf(hcType);
	                    
	                    hcChannelType = hcChannelType + HandelC.ARRAY_BEGIN;
	                    hcChannelType = hcChannelType + typeCard;
	                    hcChannelType = hcChannelType + HandelC.ARRAY_END;
	                }
	            }
	        }
			
			hcChannelDeclaration += HandelC.CHAN_DECL + HandelC.SPACE + HandelC.SYNC_TYPE + HandelC.SPACE + "end" + HandelC.UNDERSCORE + channelName + hcChannelType + HandelC.END_COMMAND;
			
			Set<Set<Integer>> syncs = this.parallelismAnalyser.getRealSyncLUT().get(HandelC.CHAN_DECL + HandelC.UNDERSCORE + channelName);
    		Iterator <Set<Integer>> sync = syncs.iterator();
    		int card = sync.next().size();
			
    		List <List<String>> typeBranches = new Vector<List<String>>();
    		for (int j = 0; j < sizeChannelType; j++) {
    			PType type = (PType)channelType.get(j);
    			String nodeTypeName = ((ACspType) type).getName().getText().trim(); 
            	String hcType = HandelC.getInstance().getHandelCTypeName(nodeTypeName);
                String typeCard = HandelC.getInstance().getCardVarOf(hcType);
                typeBranches.add(this.types.getTypeBranches(nodeTypeName));
    		}
    		
    		List<List<String>> typeBranchesDecl = (List<List<String>>) ListUtil.cartesianProduct(typeBranches);
    		if (!typeBranchesDecl.isEmpty()){
    			for (List <String> decl: typeBranchesDecl){
	    			String arrayDecl = "";
	    			for (String brnc: decl){
	    				arrayDecl = arrayDecl + HandelC.ARRAY_BEGIN + brnc + HandelC.ARRAY_END;
	    			}
	    			hcMultiSyncControl +=
	    				"MultiSyncControl" + card
	    				+ HandelC.PAR_BEGIN
	    				+ HandelC.ADDRESS + HANDELC_TO + HandelC.UNDERSCORE + channelName + arrayDecl + "[0]" + HandelC.SEP_ARGUMENTS
	    				+ HandelC.ADDRESS + HANDELC_FROM + HandelC.UNDERSCORE + channelName + arrayDecl + "[0]" + HandelC.SEP_ARGUMENTS
	    				+ HandelC.ADDRESS + HandelC.SPACE + "end" + HandelC.UNDERSCORE + channelName + arrayDecl 
	    				+ HandelC.PAR_END + HandelC.END_COMMAND ;
	    			hcChannelAcess = hcChannelAcess + "end" + HandelC.UNDERSCORE + channelName + arrayDecl + HandelC.OUTPUT_COMMUNICATION + HandelC.SYNC_OUT + HandelC.END_COMMAND;
	    			countDecl++;
	    		}
    		} else {
    			hcMultiSyncControl +=
    				"MultiSyncControl" + card
    				+ HandelC.PAR_BEGIN
    				+ HandelC.ADDRESS + HANDELC_TO + HandelC.UNDERSCORE + channelName + "[0]" + HandelC.SEP_ARGUMENTS
    				+ HandelC.ADDRESS + HANDELC_FROM + HandelC.UNDERSCORE + channelName + "[0]" + HandelC.SEP_ARGUMENTS
    				+ HandelC.ADDRESS + HandelC.SPACE + "end" + HandelC.UNDERSCORE + channelName
    				+ HandelC.PAR_END + HandelC.END_COMMAND ;
    			hcChannelAcess = hcChannelAcess + "end" + HandelC.UNDERSCORE + channelName + HandelC.OUTPUT_COMMUNICATION + HandelC.SYNC_OUT + HandelC.END_COMMAND;
    			countDecl++;
    		}
		}

		String controlsManager = 
			"inline void controllersManager () {"
			+ hcChannelDeclaration
			+ HandelC.PARALLEL + HandelC.BLOCK_BEGIN 
			+ HandelC.SEQ + HandelC.BLOCK_BEGIN
			+ "endSystem"+ HandelC.INPUT_COMMUNICATION + HandelC.SYNC_IN + HandelC.END_COMMAND
			+ HandelC.PARALLEL 
			+ HandelC.BLOCK_BEGIN 
			+ hcChannelAcess
			+ HandelC.BLOCK_END + HandelC.END_COMMAND
			+ HandelC.BLOCK_END + HandelC.END_COMMAND
			+ hcMultiSyncControl
			+ HandelC.BLOCK_END + HandelC.END_COMMAND
			+ HandelC.BLOCK_END;
		
		this.hcCode = this.hcCode + controlsManager;
    }
	
	private void declareMultiSyncConstants() {
		String nextline;
		
		nextline = HandelC.getInstance().generateDefine("MULTI_SYNC_WRITER 0 ");
		this.hcCode = nextline + this.hcCode;
		
		nextline = HandelC.getInstance().generateDefine("integer_MS_CLIENT unsigned "+ MathUtil.getUnsignedBitsForNumber(this.multisyncs.getLargerClient()));
		this.hcCode = nextline + this.hcCode;
		
		nextline = HandelC.getInstance().generateDefine("MSYNC_INTR_VALUE unsigned "+ this.multisyncs.getLargerCommInterrupt() +" // Value transfered in an interrupt of Multi-sync");
		this.hcCode = nextline + this.hcCode;
		
		nextline = HandelC.getInstance().generateDefine("MSYNC_VALUE unsigned "+ this.multisyncs.getLargerCommInMultiSync() +" // Value transfered in Multi-sync");
		this.hcCode = nextline + this.hcCode;
				
		nextline = HandelC.getInstance().generateDefine("integer_value SYNC");
		this.hcCode = nextline + this.hcCode;
		
		nextline = HandelC.getInstance().generateDefine("integer_message int (bits_of_MSYNC_ID+1)");
		this.hcCode = nextline + this.hcCode;
		
		nextline = HandelC.getInstance().generateDefine("MSYNC_ID unsigned bits_of_MSYNC_ID");
		this.hcCode = nextline + this.hcCode;
		
		int branchesInMSync = this.multisyncs.getLargerMultiSyncSet();
		nextline = HandelC.getInstance().generateDefine("bits_of_MSYNC_ID "+ MathUtil.getNumberOfBitsUnsigned(branchesInMSync+1)+ "");
		this.hcCode = nextline + this.hcCode;
				
		this.hcCode = HandelC.getInstance().generateComment("Declaration of constants for Multi-sync") + this.hcCode;
	}

	private void declareMultiSyncDatatypes() {
	
		String datatypes = ""
		+ "	struct MessageValue {"
		+ "		integer_message M;"
		+ "		MSYNC_VALUE MV;"
		+ "	};"
		
		+ "struct BoolValue {"
		+ "	boolean B;"
		+ "	MSYNC_VALUE BV;"
		+ "};"
		
		+ "typedef struct MessageValue sMessageValue;"
		+ "typedef struct BoolValue sBoolValue;"
		
		+ "struct ChannelOfMultiSync {"
		+ "	chan sMessageValue * to_sync;" 
		+ "	chan sBoolValue * from_sync;"
		+ "	MSYNC_ID id; "
		+ "	CHANNEL channel; "
		+ "	MSYNC_VALUE value; "
		+ "};"
		
		+ "typedef struct ChannelOfMultiSync sChannelOfMultiSync;"
		+ "typedef struct ChannelOfMultiSync * psChannelOfMultiSync;"
		
		+ "struct InterruptOfMultiSync {"
		+ "	chan MSYNC_INTR_VALUE * interrupt; "
		+ "	MSYNC_INTR_VALUE value; "
		+ "};"
		
		+ "typedef struct InterruptOfMultiSync sInterruptOfMultiSync;"
		+ "typedef struct InterruptOfMultiSync * psInterruptOfMultiSync;";
		
		this.hcCode = "\n" + datatypes + "\n" + this.hcCode ;
		this.hcCode = HandelC.getInstance().generateComment("Declaration of datatypes for Multi-sync") + this.hcCode;
	}
	
	private void declareMultiSyncIdLUT(){
        String declaration = HandelC.STATIC + HandelC.SPACE 
        					+ "MSYNC_ID" + HandelC.SPACE 
        					+ "MULTI_SYNC_ID_LUT" 
        					+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(HandelC.CHANNEL) + HandelC.ARRAY_END
        					+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(HandelC.BRANCH) + HandelC.ARRAY_END
        					+ HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.BLOCK_BEGIN;
        
        List<String> channelNames = ListUtil.reverse(this.channels.getChannelNames()); 
        
        BranchActionsOnChannel branchActionsOnChannel = this.parallelismAnalyser.getBranchActionsOnChannel();
        
        for (int c=0; c < channelNames.size(); c++) {
        	String channelName = channelNames.get(c);
        	declaration = declaration + HandelC.getInstance().generateComment(channelName);
        	declaration = declaration + HandelC.BLOCK_BEGIN + HandelC.SPACE;
        	
        	int index_multi_sync = 1;
            for (int branch=0; branch < this.parallelismAnalyser.getNumberOfBranches(); branch++) {
            	declaration = declaration + HandelC.getInstance().generateComment(HandelC.getInstance().getBranchName(branch));
            	
            	if (this.parallelismAnalyser.isMultiSyncChannel(HandelC.getInstance().getChannel(channelName))
            			&& branchActionsOnChannel.get(channelName).containsKey(branch)){
            		if (branchActionsOnChannel.get(channelName).get(branch).iterator().next().equals(HandelC.A_WRITE)){
            			declaration = declaration + 0;
            		} else {
            			declaration = declaration + index_multi_sync;
            			index_multi_sync++;
            		}
            	} else {
            		declaration = declaration + 0;
            	}
            	
	        	if (branch < this.parallelismAnalyser.getNumberOfBranches() - 1) {
	            	declaration = declaration + HandelC.SEP_ARGUMENTS + HandelC.SPACE;
	        	}
            }
            
            declaration = declaration + HandelC.SPACE + HandelC.BLOCK_END;
        	if (c < channelNames.size() - 1) {
            	declaration = declaration + HandelC.SEP_ARGUMENTS + HandelC.SPACE;
        	}
        }
        
        declaration = declaration + HandelC.BLOCK_END + HandelC.END_COMMAND;

        this.hcCode = declaration + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment("ID for Multi-sync") + this.hcCode;
    }
	
	@Deprecated
	private void declareMultiSyncWriterLUT(){
        String declaration = HandelC.STATIC + HandelC.SPACE 
        					+ "MSYNC_ID" + HandelC.SPACE 
        					+ "MULTI_SYNC_WRITER_LUT" 
        					+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(HandelC.CHANNEL) + HandelC.ARRAY_END
        					+ HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.BLOCK_BEGIN;
        
        List<String> channelNames = ListUtil.reverse(this.channels.getChannelNames()); 
                
        for (int c=0; c < channelNames.size(); c++) {
        	String channelName = channelNames.get(c);
        	declaration = declaration + HandelC.getInstance().generateComment(channelName);
        	if (this.parallelismAnalyser.isMultiSyncChannel(HandelC.getInstance().getChannel(channelName))){
        		declaration = declaration + 0; 
        	} else {
        		declaration = declaration + 0;
        	}
        	if (c < channelNames.size() - 1) {
            	declaration = declaration + HandelC.SEP_ARGUMENTS + HandelC.SPACE;
        	}
        }
        
        declaration = declaration + HandelC.BLOCK_END + HandelC.END_COMMAND;

        this.hcCode = declaration + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment("WRITER in Multi-sync") + this.hcCode;
    }
	
    
    private String getForcedSyncDecl() {
    	return HandelC.BRANCH + HandelC.SPACE + HandelC.BRANCH_ID;
    }
    
    private String translateFormalArguments(ACspParameters parameters, String process){
        String code = HandelC.PAR_BEGIN;

        // If we have forced interleaving, the first argument is the 
        // set FORCED_SYNC of forced interleaving events
        if (this.useBranchIds) {
            code = code + this.getForcedSyncDecl();
            if (parameters.getParamList() != null){
                code = code + HandelC.SEP_ARGUMENTS;
            }
        }
        if (parameters.getParamList() != null){
            // Writing the parameters
            Node paramsList = ((ACspParameters)parameters).getParamList();

            if (paramsList instanceof ACspBaseParamList) {
                PParam parameter = ((ACspBaseParamList)paramsList).getParam();
                code = code + this.getArgumentDeclaration(parameter,process);
            } else if (paramsList instanceof ACspLinkedParamList) {
                PParam parameter = ((ACspLinkedParamList)paramsList).getParam();
                code = code + this.getArgumentDeclaration(parameter,process);
                code = code + HandelC.SEP_ARGUMENTS;
                while((((ACspLinkedParamList)paramsList)).getNext() instanceof ACspLinkedParamList){
                    paramsList = ((ACspLinkedParamList)paramsList).getNext();
                    parameter = ((ACspLinkedParamList)paramsList).getParam();
                    code = code + this.getArgumentDeclaration(parameter,process);
                    code = code + HandelC.SEP_ARGUMENTS;
                }
                paramsList = ((ACspLinkedParamList)paramsList).getNext();
                parameter = ((ACspBaseParamList)paramsList).getParam();
                code = code + this.getArgumentDeclaration(parameter,process);
            }
        }
        code = code + HandelC.PAR_END;
        return code;
    }

    // if declaration is true it declares the variables, otherwise it initialises them
    // if argument is true, it declares it as a function argument, otherwise
    // it declares it as a local variable
    private String translateFormalArgumentsCopies(ACspParameters parameters, String process, 
                                                    boolean declaration, boolean argument){
        String code = "";
        if (parameters.getParamList() != null){
            // Writing the parameters
            Node paramsList = ((ACspParameters)parameters).getParamList();

            if (paramsList instanceof ACspBaseParamList) {
                PParam parameter = ((ACspBaseParamList)paramsList).getParam();
                code = this.getArgumentCopyDeclaration(parameter,process,declaration,argument);
            } else if (paramsList instanceof ACspLinkedParamList) {
                PParam parameter = ((ACspLinkedParamList)paramsList).getParam();
                code = this.getArgumentCopyDeclaration(parameter,process,declaration,argument);
                while((((ACspLinkedParamList)paramsList)).getNext() instanceof ACspLinkedParamList){
                    paramsList = ((ACspLinkedParamList)paramsList).getNext();
                    parameter = ((ACspLinkedParamList)paramsList).getParam();
                    code = code + this.getArgumentCopyDeclaration(parameter,process,declaration,argument);
                }
                paramsList = ((ACspLinkedParamList)paramsList).getNext();
                parameter = ((ACspBaseParamList)paramsList).getParam();
                code = code + this.getArgumentCopyDeclaration(parameter,process,declaration,argument);
            }
        }
        
        return code;
    }
    
    private void translateRealArguments(ACspProcessCallExpr node, String current){
        ACspArguments arguments = (ACspArguments)node.getArguments();
        String name = node.getName().getText();

        this.check(name,arguments,current);
        
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;

        // If we have forced interleaving, the first argument is the branch id
        if (this.useBranchIds) {
        	this.hcCode = this.hcCode + this.getBranchReference(node);
	        if (arguments != null && arguments.getArgumentList() != null){
	            	this.hcCode = this.hcCode + HandelC.SEP_ARGUMENTS;
	        }
        }
        
        if (arguments != null && arguments.getArgumentList() != null){
    		// Writing the parameters
            Node argsList = ((ACspArguments)arguments).getArgumentList();

            if (argsList instanceof ACspBaseArgumentList) {
                ((ACspBaseArgumentList)argsList).getArgument().apply(this);
            } else if (argsList instanceof ACspLinkedArgumentList) {
                ((ACspLinkedArgumentList)argsList).getArgument().apply(this);
                this.hcCode = this.hcCode + HandelC.SEP_ARGUMENTS;
                while((((ACspLinkedArgumentList)argsList)).getNext() instanceof ACspLinkedArgumentList){
                    argsList = ((ACspLinkedArgumentList)argsList).getNext();
                    ((ACspLinkedArgumentList)argsList).getArgument().apply(this);
                    this.hcCode = this.hcCode + HandelC.SEP_ARGUMENTS;
                }
                argsList = ((ACspLinkedArgumentList)argsList).getNext();
                ((ACspBaseArgumentList)argsList).getArgument().apply(this);
            }
        }
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    private void translateRealArgumentsCopies(ACspProcessCallExpr node, String current){
        ACspArguments arguments = (ACspArguments)node.getArguments();
        String name = node.getName().getText();
        
        this.check(name,arguments,current);
        
        if (arguments != null && arguments.getArgumentList() != null){
            if (this.procsInfo.containsProcArguments(name)) {
                ACspParameters parameters = this.procsInfo.getProcArguments(name);
                
                PArgumentList argsList = ((ACspArguments)arguments).getArgumentList();
                PParamList paramsList = parameters.getParamList();
                
                if (argsList instanceof ACspBaseArgumentList) {
                    this.translateRealArgumentCopy(name, 
                                                    ((ACspBaseArgumentList)argsList).getArgument(), 
                                                    ((ACspBaseParamList)paramsList).getParam());
                } else if (argsList instanceof ACspLinkedArgumentList) {
                    this.translateRealArgumentCopy(name, 
                                                    ((ACspLinkedArgumentList)argsList).getArgument(), 
                                                    ((ACspLinkedParamList)paramsList).getParam());
                    while((((ACspLinkedArgumentList)argsList)).getNext() instanceof ACspLinkedArgumentList){
                        argsList = ((ACspLinkedArgumentList)argsList).getNext();
                        paramsList = ((ACspLinkedParamList)paramsList).getNext();
                        this.translateRealArgumentCopy(name, 
                                                        ((ACspLinkedArgumentList)argsList).getArgument(), 
                                                        ((ACspLinkedParamList)paramsList).getParam());
                    }
                    argsList = ((ACspLinkedArgumentList)argsList).getNext();
                    paramsList = ((ACspLinkedParamList)paramsList).getNext();
                    this.translateRealArgumentCopy(name, 
                                                    ((ACspBaseArgumentList)argsList).getArgument(), 
                                                    ((ACspBaseParamList)paramsList).getParam());
                }
            } else {
                throw new UndefinedProcessArgumentsException(name);
            }
        }
    }

    private void translateRealArgumentCopy(String procName, PArgument arg, PParam param) {
        this.hcCode = this.hcCode + this.copyName(procName,param.toString().trim(),IS_NOT_DECLARATION) + HandelC.SPACE; 
        this.hcCode = this.hcCode + HandelC.ASSIGNMENT + HandelC.SPACE; 
        arg.apply(this);
        this.hcCode = this.hcCode + HandelC.END_COMMAND;
    }

    private String translateForcedSyncSet(){
    	Set<String> forcedChannels = (Set)this.forcedInterleaving.peek();
    	List setNames = new Vector();
    	for(String c:forcedChannels){
    		String singletonSet = HandelC.getInstance().getSetVarOf(HandelC.getInstance().getChannelName(c));
    		setNames.add(singletonSet);
    	}
    	return HandelC.getInstance().getSetUnionCode(setNames,HandelC.CHANNEL);
    }
    
    private void check(String name, ACspArguments arguments, String current){
        if (this.procsInfo.containsProcArguments(name)) {
            ACspParameters parameters = this.procsInfo.getProcArguments(name);
            if(!this.check(parameters,arguments)) {
                throw new WrongNumberOfArgumentsException(current,name);
            }
        } else {
            if (arguments != null && arguments.getArgumentList() != null){
                throw new WrongNumberOfArgumentsException(current,name);
            }
        }
    }

    private boolean check(ACspParameters parameters, ACspArguments arguments) {
        boolean answer = true;
        
        if (arguments != null && arguments.getArgumentList() != null 
            && parameters != null && parameters.getParamList() != null) {
            PArgumentList argsList = ((ACspArguments)arguments).getArgumentList();
            PParamList paramsList = parameters.getParamList();

            if ((paramsList instanceof ACspBaseParamList) && (argsList instanceof ACspBaseArgumentList)) {
                // do nothing!!! They are ok
            } else if ((paramsList instanceof ACspLinkedParamList) 
                        && (argsList instanceof ACspLinkedArgumentList)) {
                while( ((((ACspLinkedParamList)paramsList)).getNext() instanceof ACspLinkedParamList)
                        && ((((ACspLinkedArgumentList)argsList)).getNext() instanceof ACspLinkedArgumentList) ){
                    paramsList = ((ACspLinkedParamList)paramsList).getNext();
                    argsList = ((ACspLinkedArgumentList)argsList).getNext();
                }
                boolean moreParams = ((((ACspLinkedParamList)paramsList)).getNext() instanceof ACspLinkedParamList);
                boolean moreArgs = ((((ACspLinkedArgumentList)argsList)).getNext() instanceof ACspLinkedArgumentList);
                if (moreParams != moreArgs) {
                    answer = false;
                }
            } else {
                answer = false;
            }
        } else {
            boolean anyParams = (parameters != null && parameters.getParamList() != null);
            boolean anyArguments = (arguments != null && arguments.getArgumentList() != null);
            if (anyParams != anyArguments) {
                answer = false;
            }
        }
        return answer;
    }
    
    /*
     * Extract if there is a variable in the communication.
     * Returns the name of the variable if yes and null otherwise.
     */
    private String extractVarName(LinkedList comms, String channel) {
        String var = null;
        ListIterator iterator = comms.listIterator(0);
        
        while(iterator.hasNext()) {
            Object next = iterator.next();
            if (next instanceof ACspInputComunication) {
                var = ((ACspInputComunication)next).getParam().toString().trim();
            }
        }
        return var;
    }
    
    /*
     * Checks if there is an input in the list of communications.
     * Returns the name of the variable if yes and null otherwise.
     */
    private String checkInput(LinkedList comms, String channel) {
        String var = null;
        ListIterator iterator = comms.listIterator(0);
        
        while(iterator.hasNext()) {
            Object next = iterator.next();
            if (next instanceof ACspInputComunication) {
                var = ((ACspInputComunication)next).getParam().toString().trim();
                //AddIvan: If-else adicionado para evitar que haja a declarao repetida de variveis nas escolha externa.
                if (!this.inputVariables.contains(var)){
		            this.inputVariables.add(var);
		            this.includeInputVarRenaming(var, channel);
                } else {
                	var = null;
                }
            }
        }
        return var;
    }
    
    private void includeInputVarRenaming(String var, String channel) {
    	String renaming = var;
    	if (this.channels.getBuses().contains(channel)) {
    		renaming = HandelC.getInstance().getInBusVarName(channel);
    	}

    	Stack renamingStack = this.inputVarRenaming.remove(var);
    	if (renamingStack == null) {
    		renamingStack = new Stack<String>();
    	}
    	renamingStack.push(renaming);
		this.inputVarRenaming.put(var, renamingStack);
    }

    private void removeInputVarRenaming(String var) {
    	Stack renamingStack = this.inputVarRenaming.remove(var);
    	if (renamingStack != null) {
    		renamingStack.pop();
    		if (!renamingStack.isEmpty()) {
    			this.inputVarRenaming.put(var, renamingStack);
    		}
    	}
    }

    private void includeInputVarType(String var, String type) {
    	Stack typeStack = this.inputVarTypes.remove(var);
    	if (typeStack == null) {
    		typeStack = new Stack<String>();
    	}
    	typeStack.push(type);
		this.inputVarTypes.put(var, typeStack);
    }

    private void removeInputVarType(String var) {
    	Stack typeStack = this.inputVarTypes.remove(var);
    	if (typeStack != null) {
    		typeStack.pop();
    		if (!typeStack.isEmpty()) {
        		this.inputVarTypes.put(var, typeStack);
    		}
    	}
    }
    
    private String getTryJoinSema(Node node, String c, String action){
    	//TRY_JOIN_SEMAPHOR(BRANCH_ID,chan_c,action)
    	return HandelC.TRY_JOIN_SEMAPHOR + HandelC.PAR_BEGIN
		+ this.getBranchReference(node) 
		+ HandelC.SEP_ARGUMENTS + HandelC.getInstance().getChannelName(c) 
		+ HandelC.SEP_ARGUMENTS 
		+ action
		+ HandelC.PAR_END;
    }

    private String getJoinSema(Node node, String c, String action){
        //JOIN_SEMAPHOR(BRANCH_ID,chan_c,action);
        return HandelC.JOIN_SEMAPHOR + HandelC.PAR_BEGIN
			+ this.getBranchReference(node) 
			+ HandelC.SEP_ARGUMENTS 
			+ HandelC.getInstance().getChannelName(c)
			+ HandelC.SEP_ARGUMENTS 
			+ action
			+ HandelC.PAR_END;
    }

    private String getLeaveSema(Node node, String c, String action){
        //LEAVE_SEMAPHOR(BRANCH_ID,chan_c,action);
        return HandelC.LEAVE_SEMAPHOR + HandelC.PAR_BEGIN
			+ this.getBranchReference(node) 
			+ HandelC.SEP_ARGUMENTS 
			+ HandelC.getInstance().getChannelName(c)
			+ HandelC.SEP_ARGUMENTS 
			+ action
			+ HandelC.PAR_END;
    }

    private String getConfirmReaders(Node node, String c, String action){
        //CONFIRM_READERS(BRANCH_ID,chan_c);
        return HandelC.CONFIRM_READERS + HandelC.PAR_BEGIN
			+ this.getBranchReference(node) 
			+ HandelC.SEP_ARGUMENTS 
			+ HandelC.getInstance().getChannelName(c)
			+ HandelC.PAR_END;
    }

    private String getSendWithdrawSignal(String c){
        //withdraw[chan_c] = true;
        return HandelC.WITHDRAW_SIGNALS + HandelC.ARRAY_BEGIN
			+ HandelC.getInstance().getChannelName(c)
        	+ HandelC.ARRAY_END
        	+ HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
        	+ HandelC.LOGICAL_TRUE;
    }
    
    private String getChannelSemaphor(String chan) {
    	String semaphor = HandelC.getInstance().getSetVarOf(chan);
        return semaphor;
    }

    private String translatePrefix(PEvent event, LinkedList comm, LinkedList proj, 
    								boolean isPriAlt) {

        String inputVar = null;
    	
    	String channel = ((ACspEvent)event).getChannelName().toString().trim();

        this.setIsExpression(true);
    	
    	// First of all, check if this is a Bus
    	if (this.channels.getBuses().contains(channel)) {
    		// Output buses 
    		if (this.channels.getOutputBuses().contains(channel)){
    			
    			this.hcCode = this.hcCode + "#ifdef SIMULATE " + "simulate_" + channel + HandelC.OUTPUT_COMMUNICATION;
    			
    			String hcBackup = this.hcCode;
    			this.hcCode = "";
    			
    			ListIterator commIter = comm.listIterator(0);
	            Object next = commIter.next();
	            if (next instanceof ACspOutputComunication) {
	                ((ACspOutputComunication)next).getExpression().apply(this);
	            }
	            
	            String communication = this.hcCode;
	            this.hcCode = hcBackup;
	            
	            this.hcCode = this.hcCode + communication + HandelC.END_COMMAND + " #else ";
    			
        		this.hcCode = this.hcCode + HandelC.getInstance().getOutBusVarName(channel)
							+ HandelC.ASSIGNMENT + communication; 
        		this.hcCode = this.hcCode + HandelC.END_COMMAND;
        		
        		this.hcCode = this.hcCode + " #endif ";
    		// Input buses 
    		} else if (this.channels.getInputBuses().contains(channel)){
    			String inputVarName = HandelC.getInstance().getInBusVarName(channel);
    			String inputBusMethodName = HandelC.getInstance().getInBusMethodName(channel);
    			String inputBusArgName = HandelC.getInstance().getBusArgName(channel);
    			
    			this.hcCode = this.hcCode + "#ifdef SIMULATE " + "simulate_" + channel + HandelC.INPUT_COMMUNICATION
    						+ channel + "_value" + HandelC.END_COMMAND + " #else ";
    			
        		this.hcCode = this.hcCode + inputVarName 
        					+ HandelC.ASSIGNMENT + inputBusMethodName + HandelC.DOT + inputBusArgName
        					+ HandelC.END_COMMAND; 
        		
        		this.hcCode = this.hcCode + " #endif " ;

        		// Replace the following reference to the input variable to the bus variable
    			inputVar = this.checkInput(comm, channel);
    		}
    	// It is not a bus, but a normal CSP channel
    	} else if (this.parallelismAnalyser.isMultiSyncChannel(channel)) {
    		this.currentPrefixing = channel;
    		String channel_name = HandelC.getInstance().getChannelName(this.currentPrefixing);
 	        String action = this.parallelismAnalyser.getCommunicationAction(this.currentDefinition, 
 	        																channel_name, 
 	        																comm);
    		String commType = this.channels.getChannelCommType(this.currentPrefixing);
    		
    		
    		/* 
 	         * The Restrictions Visitor guarantess that we have communications
 	         * of the form c[.exp]*[?name | !exp] 
 	        */
 	
 	        if (!isPriAlt) {
 	            inputVar = this.declareInputVar(comm, channel);
 	        } 
    		
 	       String hcCodeBackup = this.hcCode;
 	       this.hcCode = "";
    		ListIterator projIter = proj.listIterator(0);
    		// Translate the projections
    		while(projIter.hasNext()) {
    			Object next = projIter.next();
    			((ACspProjection)next).apply(this);
    			if(!projIter.hasNext()){
    				this.closeArrayAccessChannelSync();
    			}
    		}
    		
    		String clientProj = this.hcCode;
    		//this.hcCode = hcCodeBackup;
    		
    		boolean isInput = false;
    		this.hcCode = "";
    		String commMS = "";
    		
    		ListIterator commIter = comm.listIterator(0);
    		
    		// Translate the communication
	        if (!commIter.hasNext()) {
	        	commMS = "0";
	        } else {
	        	this.isMultiSync = true;
	            Object next = commIter.next();
	            if (next instanceof ACspInputComunication) {
	            	isInput = true;
	                ((ACspInputComunication)next).apply(this);
	            } else if (next instanceof ACspOutputComunication) {
	                ((ACspOutputComunication)next).apply(this);
	            }
	            this.isMultiSync = false;
//	            commMS = this.hcMultiSyncCode;
	            commMS = this.hcCode;
//	            this.hcMultiSyncCode = "";
	        }
    		
	        this.hcCode = hcCodeBackup;
    		
    		if (!isPriAlt){
    			String client = "" +
	    		"sChannelOfMultiSync seqOfSync;"+
	    		//"psInterruptOfMultiSync seqOfNotSync;"+
//	    		"integer_value value;"+
	    		"par {"+
	    		"	seqOfSync.to_sync = &to_"+channel + clientProj +"[0];"+
	    		"	seqOfSync.from_sync = &from_"+channel + clientProj +"[0];"+
	    		"	seqOfSync.id = MULTI_SYNC_ID_LUT[chan_"+channel+"][BRANCH_ID]; "+
	    		"	seqOfSync.channel = chan_"+channel+"; "+
	    		"	seqOfSync.value = (MSYNC_VALUE) "+commMS+"; "+
//	    		"	value = 0;"+
	    		"}"+
	    		"MultiSyncClient_1x0x0(&seqOfSync);"
    			+(!isInput ? "" : commMS + " = ("+commType+") seqOfSync.value;");
	    		this.hcCode = this.hcCode + client;
    		} else {
    			String client = "" +
    			"	seqOfSync.to_sync = &to_"+channel + clientProj +"[0];"+
	    		"	seqOfSync.from_sync = &from_"+channel + clientProj +"[0];"+
	    		"	seqOfSync.id = MULTI_SYNC_ID_LUT[chan_"+channel+"][BRANCH_ID]; "+
	    		"	seqOfSync.channel = chan_"+channel+"; ";
    			this.hcCode = this.hcCode + client;
    			
    		}
    		this.currentType = "";
    	}else {
	        this.currentPrefixing = channel;
	        ListIterator commIter = comm.listIterator(0);
	        ListIterator projIter = proj.listIterator(0);
	
	        String channel_name = HandelC.getInstance().getChannelName(this.currentPrefixing);
	        String action = this.parallelismAnalyser.getCommunicationAction(this.currentDefinition, 
	        																channel_name, 
	        																comm);
        	this.currentPrefixingTypes = ListUtil.clone(this.channels.getChannelType(this.currentPrefixing));
	
	        /* 
	         * The Restrictions Visitor guarantess that we have communications
	         * of the form c[.exp]*[?name | !exp] 
	        */
	
	        if (!isPriAlt) {
	            inputVar = this.declareInputVar(comm, channel);
	        } 
	        
	        this.hcCode = this.hcCode + this.currentPrefixing;
	        
	        // Translate the projections
	        while(projIter.hasNext()) {
	            Object next = projIter.next();
	            ((ACspProjection)next).apply(this);
	            if(!projIter.hasNext()){
	            	this.closeArrayAccessChannelSync();
	            }
	        }
	        
	        // Translate the communication
	        if (!commIter.hasNext()) {
	        	if (this.directives.containsProcChannelInOut(this.currentDefinition)) {
	                Hashtable channelInOut = this.directives.getProcChannelInOut(this.currentDefinition);
	                if (channelInOut.containsKey(this.currentPrefixing)){
	                    String inOut = (String)channelInOut.get(this.currentPrefixing);
	                    if (inOut.equals(HandelC.A_READ)) {
	                        this.hcCode = this.hcCode + HandelC.INPUT_COMMUNICATION;
	                        this.hcCode = this.hcCode + HandelC.SYNC_IN;
	                    } else if (inOut.equals(HandelC.A_WRITE)) {
	                        this.hcCode = this.hcCode + HandelC.OUTPUT_COMMUNICATION;
	                        this.hcCode = this.hcCode + HandelC.SYNC_OUT;
	                    }
	                } else {
	                    throw new UndefinedChannelInOutException(this.currentPrefixing,this.currentDefinition);
	                }
	            } else {
	                throw new UndefinedChannelInOutException(this.currentPrefixing,this.currentDefinition);
	            }
	
	        } else {
	            Object next = commIter.next();
	            if (next instanceof ACspInputComunication) {
	                ((ACspInputComunication)next).apply(this);
	            } else if (next instanceof ACspOutputComunication) {
	                ((ACspOutputComunication)next).apply(this);
	            }
	        }
	
	        if (!isPriAlt && this.useBranchIds) {
	        	this.hcCode = this.hcCode + HandelC.END_COMMAND 
	        				//+ this.getLeaveSema(event, this.currentPrefixing,action) + HandelC.END_COMMAND
	        				;
	        }
	        
	        this.currentType = "";
    	}
        this.setIsExpression(false);

        return inputVar;
    }

    // Declares the input variable (if there is any)
    private String declareInputVar(LinkedList comm, String channel) {
        String var = this.checkInput(comm, channel);
        if (var != null) {
            if (this.channels.containsChannelType(this.currentPrefixing)) {
                String type = this.channels.getChannelCommType(this.currentPrefixing);
                this.hcCode = this.hcCode + type + HandelC.SPACE + var + HandelC.END_COMMAND;
                
                this.includeInputVarType(var, type);
                
            } else {
                throw new UndefinedChannelTypeException(this.currentPrefixing,this.currentDefinition);
            }
        }
        return var;
    }

    // Translates all the processes definitions as an action system like
    // Does not support parametrised processes
    private void translateProcessesDefinitionsMutualRecursion() {
        int numberOfProcesses = this.procsInfo.size();
        int numberOfBits = MathUtil.getNumberOfBitsSigned(numberOfProcesses);

        // Declare local copies for all arguments of every parametrised process
        this.hcCode = this.hcCode + HandelC.getInstance().generateComment(Messenger.MUTUAL_REC_LOCAL_VARS);
        
        //System.out.println("ProcessesTranslatorVisitor :: translateProcessesDefinitionsMutualRecursion -> INICIO");
        
        for (int i=0; i< this.processesIds.size(); i++) {
            String name = (String)this.processesIds.get(i);
            if (this.procsInfo.containsProcArguments(name)) {
                ACspParameters parameters = this.procsInfo.getProcArguments(name);
                if (parameters == null) {
                    parameters = new ACspParameters();
                }
                this.hcCode = this.hcCode + this.translateFormalArgumentsCopies(parameters,name,IS_DECLARATION,IS_NOT_ARGUMENT);
            }
        }

        //System.out.println("ProcessesTranslatorVisitor :: translateProcessesDefinitionsMutualRecursion -> 1");
        
        // inline void MUTUAL_REC(
        this.hcCode = this.hcCode + HandelC.getInstance().generateComment(Messenger.MUTUAL_REC_METHOD);
        this.hcCode = this.hcCode + HandelC.INLINE + HandelC.SPACE;
        this.hcCode = this.hcCode + HandelC.VOID + HandelC.SPACE;
        this.hcCode = this.hcCode + HandelC.MUTUAL_RECURSION + HandelC.PAR_BEGIN;

        // CHANNEL_TYPE_set FORCED_SYNC
        if(this.useBranchIds) {
        	this.hcCode = this.hcCode + this.getForcedSyncDecl() + HandelC.SEP_ARGUMENTS;
        }

        // int numberOfBits PROGRAM_COUNTER
        this.hcCode = this.hcCode + HandelC.INT + HandelC.SPACE;
        this.hcCode = this.hcCode + numberOfBits + HandelC.SPACE;
        this.hcCode = this.hcCode + HandelC.PROGRAM_COUNTER;

        // ){
        this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.BLOCK_BEGIN;
        
        // int 1 KEEP_LOOPING;
        this.hcCode = this.hcCode + HandelC.BOOLEAN + HandelC.SPACE;
        this.hcCode = this.hcCode + HandelC.KEEP_LOOPING;
        this.hcCode = this.hcCode + HandelC.END_COMMAND;

        // KEEP_LOOPING = true;
        this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_TRUE);
        
        // while(KEEP_LOOPING) {
        this.hcCode = this.hcCode + HandelC.WHILE;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        this.hcCode = this.hcCode + HandelC.KEEP_LOOPING;
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;

        // KEEP_LOOPING = false;
        this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_FALSE);
        
        // switch(PROGRAM_COUNTER) {
        this.hcCode = this.hcCode + HandelC.SWITCH + HandelC.PAR_BEGIN;
        this.hcCode = this.hcCode + HandelC.PROGRAM_COUNTER;
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;

        //System.out.println("ProcessesTranslatorVisitor :: translateProcessesDefinitionsMutualRecursion -> 2");
        
        // case PROC_NAME: 
        for (int i = 0; i < this.processesIds.size(); i++) {
        	String pname = this.processesIds.get(i).trim();

            //System.out.println("ProcessesTranslatorVisitor :: translateProcessesDefinitionsMutualRecursion -> Translating "+pname);
        	
        	if (!pname.equals(HandelC.MAINP)) {
                this.setCurrentDefinition(pname);

                this.hcCode = this.hcCode + HandelC.getInstance().generateComment(Messenger.PROCESS + HandelC.SPACE + this.currentDefinition);
                this.hcCode = this.hcCode + HandelC.CASE + HandelC.SPACE;
                this.hcCode = this.hcCode + this.currentDefinition + HandelC.SPACE + HandelC.CASE_SEP;
                this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
                if(this.procsInfo.containsProcDefinition(this.currentDefinition)) {
                    this.procsInfo.getProcDefinition(this.currentDefinition).apply(this);
                }
                this.hcCode = this.hcCode + HandelC.BREAK + HandelC.END_COMMAND;
                this.hcCode = this.hcCode + HandelC.BLOCK_END;
        	}
        }

        //System.out.println("ProcessesTranslatorVisitor :: translateProcessesDefinitionsMutualRecursion -> end ");
        
        // default: KEEP_LOOPING" = false;
        this.hcCode = this.hcCode + HandelC.DEFAULT + HandelC.CASE_SEP + HandelC.SPACE;
        this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_FALSE);

        // }}}
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
        
        // Declare the main process
        this.declareMainFunction(this.procsInfo.getProcDefinition(HandelC.MAINP));
    }
    
    // Indicates that we are translating the main method's process 
    private boolean isTranslatingMain() {
        return this.currentDefinition.equals(HandelC.MAINP);
    }

    // Translates both branches of a parallelism or interleaving
    private void translateConcurrency(Node node) {
    	Set newForcedInterleaving = null;
        Node left = null;
        Node right = null;
        
        if (node instanceof ACspSharingParallelProcess) {
            newForcedInterleaving = this.parallelismAnalyser.getForcedInterleavingChannels((ACspSharingParallelProcess)node, this.currentDefinition);
            left = ((ACspSharingParallelProcess)node).getLeftProcess();
            right = ((ACspSharingParallelProcess)node).getRightProcess();
    	} else if (node instanceof ACspAlphabetisedParallelProcess) {
            newForcedInterleaving = this.parallelismAnalyser.getForcedInterleavingChannels((ACspAlphabetisedParallelProcess)node, this.currentDefinition);
            left = ((ACspAlphabetisedParallelProcess)node).getLeftProcess();
            right = ((ACspAlphabetisedParallelProcess)node).getRightProcess();
    	} else if (node instanceof ACspInterleaveProcess) {
            newForcedInterleaving = this.parallelismAnalyser.getForcedInterleavingChannels((ACspInterleaveProcess)node, this.currentDefinition);
            left = ((ACspInterleaveProcess)node).getLeftProcess();
            right = ((ACspInterleaveProcess)node).getRightProcess();
    	}

        if (newForcedInterleaving == null || left == null || right == null) {
        	throw new UnsupportedConstructorException(Messenger.UNKNOWN_PARALLEL_COMPOSITION, this.currentDefinition);
        } else {
    		Set allChannels = new HashSet<String>(this.channels.getChannels());

    		// Push the new forced interleaving into the stack
    		this.forcedInterleaving.push(newForcedInterleaving);

            this.hcCode = this.hcCode + HandelC.PARALLEL;
            this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
            
            this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
            left.apply(this);
            this.hcCode = this.hcCode + HandelC.END_COMMAND;
            this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.END_COMMAND;

            this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
            right.apply(this);
            this.hcCode = this.hcCode + HandelC.END_COMMAND;
            this.hcCode = this.hcCode + HandelC.BLOCK_END;
            
            this.hcCode = this.hcCode + HandelC.BLOCK_END;

    		this.forcedInterleaving.pop();
        }
    }

    // Closes the access to the arrays of channels
    private void nextArrayAccessChannelSync() {
        this.closeArrayAccessChannelSync();
        this.openArrayAccessChannelSync();
    }

    // Closes the access to the arrays of channels
    private void closeArrayAccessChannelSync() {

        // this is the first access, do nothing
        if(!this.currentType.equals("")) {
            if(this.currentType.equals(CspUtil.getIntType())) {
                this.hcCode = this.hcCode + HandelC.PAR_END;
            }
            this.hcCode = this.hcCode + HandelC.ARRAY_END;
        }
        
    }

    // Opens the access to the arrays of channels
    private void openArrayAccessChannelSync() {

        this.hcCode = this.hcCode + HandelC.ARRAY_BEGIN;
        if (this.intArrayAccessChannelSync()) { 
        	this.hcCode = this.hcCode + this.getUnsignedCastForInteger(); 
        }
        
        this.removeCurrentPrefixingTypes();
    }
    
    // Indicates that we were accessing an array using integers 
    private boolean intArrayAccessChannelSync() {
        boolean answer = false;
        
        boolean empty = this.currentPrefixingTypes.isEmpty();
        boolean lastType = this.currentPrefixingTypes.size()==1;
        boolean isComm = this.channels.isCommChannel(this.currentPrefixing);
        
        boolean isInt = this.currentPrefixingTypes.get(0).toString().trim().equals(CspUtil.getIntType());
        
        if (!empty) {
            if (lastType) {
                if (!isComm && isInt) {
                    answer = true;
                }
            } else {
                if (isInt) {
                    answer = true;
                }
            }
        }
        
        return answer;
    }
    
    private void removeCurrentPrefixingTypes() {
        if (!this.currentPrefixingTypes.isEmpty()) {
            this.currentType = this.currentPrefixingTypes.get(0).toString().trim();
            this.currentPrefixingTypes.remove(0);
        }
    }

    private void openArrayAccessTypeConstructor(){
    	//String code = "";
    	this.hcCode = this.hcCode + HandelC.ARRAY_BEGIN;
//    	code = HandelC.ARRAY_BEGIN;
        if(!this.tagsTypesStack.isEmpty()) {
            String type = (String)ListUtil.head(((List)this.tagsTypesStack.peek()));
            if(type.equals(HandelC.INTEGERL)) {
            	this.hcCode = this.hcCode + this.getUnsignedCastForInteger();
//            	code = code + this.getUnsignedCastForInteger();
            }
        }
        /*if (!this.isMultiSync){
        	this.hcCode = this.hcCode + code;
    	} else{
    		this.hcMultiSyncCode = this.hcMultiSyncCode + code;
    	}*/
    }
    
    private void closeArrayAccessTypeConstructor(){
        if(!this.tagsTypesStack.isEmpty() && !(((List)this.tagsTypesStack.peek()).size() == 0)) {
            List types = (List)this.tagsTypesStack.pop();
            String type = (String)ListUtil.head(types);
            if(type.equals(HandelC.INTEGERL)) {
                this.hcCode = this.hcCode + HandelC.PAR_END;
            }
            if (types.size() > 1) {
                this.tagsTypesStack.push(ListUtil.tail(types));
            }
        }
        this.hcCode = this.hcCode + HandelC.ARRAY_END;
        
    }

    private void nextArrayAccessTypeConstructor(){
        this.closeArrayAccessTypeConstructor();
        this.openArrayAccessTypeConstructor();
    }

    private String getHasJoinedUpdate(Node node, String hasJoined, String chan, String action) {
    	return hasJoined + HandelC.ARRAY_BEGIN + HandelC.getInstance().getChannelName(chan) + HandelC.ARRAY_END
    				+ HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE
    				+ this.getTryJoinSema(node, chan,action) + HandelC.END_COMMAND;
    }

    private boolean insertHasJoinedUpdate(Node node, 
    									List<String> semaphorsTaken, 
    									List<String> semaphorsTried,
    									String hasJoined, 
    									String channel, 
    									String action) {
    	boolean retorno = false;
    	if (!semaphorsTaken.contains(channel) && !semaphorsTried.contains(channel)) {
    		this.hcCode = this.hcCode + this.getHasJoinedUpdate(node, hasJoined,channel,action);
    		retorno = true;
    	}
    	return retorno;
    }
    
    private String getHasJoinedElement(String hasJoined, String chan) {
    	return hasJoined + HandelC.ARRAY_BEGIN + HandelC.getInstance().getChannelName(chan) + HandelC.ARRAY_END;
    }

    private void declareMainFunction(PExpression expression){
    	//AddIvan: Main for MultiSync
    	if (this.parallelismAnalyser.isMultiSync()){
    		this.declareMultiSyncManagers(expression);
    		
    		this.hcCode = this.hcCode + HandelC.getInstance().generateComment(Messenger.MAIN_DECLARATION);
            this.hcCode = this.hcCode + HandelC.VOID + HandelC.SPACE + HandelC.MAIN + HandelC.PAR_BEGIN 
            			+ HandelC.PAR_END + HandelC.BLOCK_BEGIN;
            this.hcCode = this.hcCode + HandelC.PARALLEL + HandelC.BLOCK_BEGIN;
            this.hcCode = this.hcCode + "processManagerMultiSync" + HandelC.PAR_BEGIN + HandelC.PAR_END + HandelC.END_COMMAND;
            this.hcCode = this.hcCode + "controllersManager" + HandelC.PAR_BEGIN + HandelC.PAR_END + HandelC.END_COMMAND;
            this.hcCode = this.hcCode + HandelC.BLOCK_END;
            this.hcCode = this.hcCode + HandelC.BLOCK_END;
        	
    	} else {
	        this.hcCode = this.hcCode + HandelC.getInstance().generateComment(Messenger.MAIN_DECLARATION);
	        this.hcCode = this.hcCode + HandelC.VOID + HandelC.SPACE + HandelC.MAIN + HandelC.PAR_BEGIN 
	        			+ HandelC.PAR_END + HandelC.BLOCK_BEGIN;
	        this.setCurrentDefinition(HandelC.MAINP);
	        
	        if (this.useBranchIds) {
	            this.hcCode = this.hcCode + HandelC.BRANCH + HandelC.SPACE + HandelC.BRANCH_ID
	            			+ HandelC.END_COMMAND;
	            this.hcCode = this.hcCode + HandelC.BRANCH_ID + HandelC.SPACE + HandelC.ASSIGNMENT
								+ HandelC.SPACE + HandelC.MATH_ZERO + HandelC.END_COMMAND;
	        }
	        expression.apply(this);
	        this.hcCode = this.hcCode + HandelC.BLOCK_END;
    	}
    }

    private String getChannelAction(InitialsExtractorVisitor initials, String channel, String definition) {
	    LinkedList comm = initials.getCommunication(channel);
	    String channel_name = HandelC.getInstance().getChannelName(channel);
	    String action = this.parallelismAnalyser.getCommunicationAction(definition, 
	    																channel_name, 
	    																comm);
	    return action;
    }

    private String getBranchReference(Node node){
		return HandelC.BRANCH_ID + HandelC.MATH_SUM
				+ this.parallelismAnalyser.getNodeBranchId(node);
    }

	public void caseACspProductionsSet(ACspProductionsSet node){
    }

    public void caseACspCollectionElement(ACspCollectionElement node) {
    	MaximalTypesInferenceVisitor inference = new MaximalTypesInferenceVisitor();
    	node.apply(inference);
    	CspType type = inference.getType();

    	boolean isInteger = false;
    	String hcType = "";
		String name = node.toString().trim();
    	
    	if (type instanceof CspUnknownType) {
			// in this case, the type checker did not work 
			// The type is retrieved from the context
			// It is formed by the cases in which we know the type checker has problems
			// First, let us try the input variables
			if (this.inputVarTypes.containsKey(name)) {
				hcType = ((Stack<String>)this.inputVarTypes.get(name)).peek();
				isInteger = hcType.equals(HandelC.INTEGERL);
			} else {
				// It may be an argument of this process
	    		if (this.directives.containsProcArguments(this.currentDefinition) 
	        	        && this.directives.getProcArguments(this.currentDefinition).containsKey(name)) {
	        			hcType = ((String)this.directives.getProcArguments(this.currentDefinition).get(name)).trim();
	    				isInteger = hcType.equals(HandelC.INTEGERL);
        		} else {
					// Nothing worked, I give up
					throw new CouldNotInferTypeException(name,this.currentDefinition);
        		}
			}
    	} else {
        	isInteger = (type instanceof CspIntType);
        	hcType = HandelC.getInstance().getHandelCTypeName(type.toString());
    	}
    	
    	super.translateCollectionElement(hcType, isInteger, node);
    }
    
    private PHelp createProcessClone (PHelp process){
    	PHelp newProcess = new ACspBaseReplicatedHelp();
    	String processClone = "Help = p -> " + process.toString().trim();
        //System.out.println("\ncreateChannel: " + newPrefixProcess + "\n");
        Lexer l = new Lexer(new PushbackReader(
                new StringReader(processClone), 1024));
        Parser p = new Parser(l);
        try {
            Start t = p.parse();
            newProcess = 
            new ExtendedDepthFirstAdapter() {
            	PHelp proc;
            	PReplicatedProcess choiceWithParenthesis;
        		public PHelp getProcess(Start s) {
                    s.apply(this);
                    return this.proc;
                }
                @Override
                public void caseACspPrefixProcess(ACspPrefixProcess node) {
                    this.proc = node.getProcess();
                }
            }.getProcess(t);
        } catch (Throwable e) {
            e.printStackTrace();
        }
        return newProcess;
    }
    
    public void caseACspRenamingProcess(ACspRenamingProcess node) {
    	this.currentSubstitutionList = new Hashtable <String,List<String>> ();
    	node.getSubstitutionList().apply(this);
    	this.isRenaming = true;
    	node.getProcess().apply(this);
    	this.isRenaming = false;
    }
    
    public void caseACspSubstitution(ACspSubstitution node) {
	    if (this.currentSubstitutionList == null) {
	    	this.currentSubstitutionList = new Hashtable<String, List<String>>(10, 10);
        }
        PEvent oldE = node.getOldEvent();
        PEvent newE = node.getNewEvent();
        List<String> newsE;
        if (this.currentSubstitutionList.containsKey(oldE.toString().trim())) {
        	newsE = this.currentSubstitutionList.get(oldE.toString().trim());
        } else {
        	newsE = new Vector <String> ();
        }
        newsE.add(newE.toString().trim());
        this.currentSubstitutionList.put(oldE.toString().trim(), newsE);
    }
    
    public List<PDefinitionList> definitionInLet;
    
    public void caseACspLetExpr (ACspLetExpr node){
    	node.getExpression().apply(this);
    	if (this.definitionInLet != null){
    		this.definitionInLet.add(node.getDefinitionList());
    	} else {
    		this.definitionInLet = new Vector<PDefinitionList>();
    		this.definitionInLet.add(node.getDefinitionList());
    	}
    	this.hasLetDefinition = true;
    	//node.getDefinitionList().apply(this);
    }
	
}
